/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.aptformat;

import java.io.*;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Vector;
import java.net.URL;
import java.net.MalformedURLException;
import fr.pixware.util.FileUtil;
import fr.pixware.apt.parse.ParseException;
import com.xmlmind.xmledit.model.DocumentModel;
import fr.pixware.util.WorkInProgressObserver;
import com.xmlmind.structformat.*;
// Needed by main().
import com.xmlmind.xmledit.modeltool.DocumentWriter;
import com.xmlmind.xmledit.modelimpl.Document;

public class APTFormat implements StructuredFormat {
    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.aptformat.APTFormatRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    // -----------------------------------------------------------------------

    private static final String platformDefaultEncoding = 
        FileUtil.defaultEncoding();

    private static final String[] extensions = { "apt", "txt" };
    private static final String description = msg("description");
    private static final String[][] parameterUsage = {
	{"defaultEncoding", msg("string"), platformDefaultEncoding, 
         msg("defaultEncoding")}
    };

    private APTLoader aptLoader = new APTLoader();
    private APTWriter aptWriter = new APTWriter();
    private String defaultEncoding = platformDefaultEncoding;

    // -----------------------------------------------------------------------

    public String getMimeType() {
	return "text/x-apt";
    }

    public String[] getFileNameExtensions() {
	return extensions;
    }

    public String getDescription() {
	return description;
    }

    public String[][] getParameterUsage() {
	return parameterUsage;
    }

    public void setParameters(String[] parameters) 
	throws ParameterUsageException {
	int count = (parameters.length/2)*2;

	for (int i = 0; i < count; i += 2) {
	    String param = parameters[i];
	    String value = parameters[i+1];

	    if ("defaultEncoding".equals(param)) {
		value = value.trim();

		if (value.length() == 0) {
		    value = platformDefaultEncoding;
		} else {
		    if (!FileUtil.isKnownEncoding(value))
			throw new ParameterUsageException(
			    msg("invalidParam", param));
		}

		setDefaultEncoding(value);
	    } else {
		throw new ParameterUsageException(
		    msg("unknownParam", param));
	    }
	}
    }

    public String[] getParameters() {
	Vector parameters = new Vector();

	if (getDefaultEncoding() != platformDefaultEncoding) {
	    parameters.addElement("defaultEncoding");
	    parameters.addElement(getDefaultEncoding());
	}

	String[] result = new String[parameters.size()];
	parameters.copyInto(result);
	return result;
    }

    public void setDefaultEncoding(String defaultEncoding) {
	this.defaultEncoding = defaultEncoding;
    }

    public String getDefaultEncoding() {
	return defaultEncoding;
    }

    public void setWorkInProgressObserver(WorkInProgressObserver observer) {
	aptLoader.setObserver(observer);
    }

    public WorkInProgressObserver getWorkInProgressObserver() {
	return aptLoader.getObserver();
    }

    public void loadDocument(String fileName, DocumentModel document) 
	throws IOException, StructuredFormatException {
	URL url = FileUtil.fileToURL(fileName);
	InputStream in = new FileInputStream(fileName);

	try {
	    loadDocument(in, url, null, document);
	} finally {
	    in.close();
	}
    }

    public void loadDocument(InputStream in, URL url, String javaEncoding, 
			     DocumentModel document) 
	throws IOException, StructuredFormatException {
	InputStreamReader reader;
	if (javaEncoding == null)
	    reader = new InputStreamReader(in, defaultEncoding);
	else
	    reader = new InputStreamReader(in, javaEncoding);

	try {
	    aptLoader.loadDocument(new BufferedReader(reader), url, document);
	} catch (ParseException e) {
	    throw new StructuredFormatException(
		url.toExternalForm(), e.getLineNumber(), /*columnNumber*/ 1, 
		e.getMessage());
	}
    }

    public void saveDocument(DocumentModel document, String fileName) 
	throws IOException, StructuredFormatException {
	URL url = FileUtil.fileToURL(fileName);
	OutputStream out = new FileOutputStream(fileName);
	
	try {
	    saveDocument(document, out, url, null);
	} finally {
	    out.close();
	}
    }

    public void saveDocument(DocumentModel document, OutputStream out, 
			     URL url, String javaEncoding) 
	throws IOException, StructuredFormatException {
	OutputStreamWriter writer = new OutputStreamWriter(out);
	if (javaEncoding == null)
	    writer = new OutputStreamWriter(out, defaultEncoding);
	else
	    writer = new OutputStreamWriter(out, javaEncoding);

	WorkInProgressObserver observer = aptLoader.getObserver();
	if (observer != null)
	    observer.workInProgress(this, true);

	try {
	    aptWriter.writeDocument(document, new BufferedWriter(writer), url);
	} finally {
	    if (observer != null)
		observer.workInProgress(this, false);
	}
    }

    // -----------------------------------------------------------------------
    // Main
    // -----------------------------------------------------------------------

    public static void main(String[] args) throws Exception {
	if (args.length != 2) {
	    System.err.println("usage: java com.xmlmind.aptformat.APTFormat" + 
			       " apt_src_file apt_or_xml_dst_file");
	    System.exit(1);
	}

	APTFormat format = new APTFormat();

	Document document = new Document();
	format.loadDocument(args[0], document);

	String dstFileName = args[1];
	if (dstFileName.endsWith(".xml") || dstFileName.endsWith(".XML")) 
	    DocumentWriter.writeDocument(document, dstFileName,
					 false, -1, -1);
	else
	    format.saveDocument(document, dstFileName);
    }
}

