/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.aptformat;

import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Stack;
import java.util.EmptyStackException;
import java.util.Hashtable;
import fr.pixware.apt.parse.*;
import fr.pixware.util.WorkInProgressObserver;
import fr.pixware.util.WorkCanceledException;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.ElementUtil;
import com.xmlmind.xmledit.modeltool.StyleSheetInfo;
import com.xmlmind.xmledit.modelimpl.XPDocumentLoader;
import com.xmlmind.xmledit.modelimpl.Document;
import com.xmlmind.cssparser.ParseUtil;

public class APTLoader implements Sink {
    private Parser parser = new Parser();
    private WorkInProgressObserver observer = null;

    private DocumentTypeModel docType = null;
    private StyleSheetInfo[] styleSheetInfo = null;

    private StringBuffer buffer = new StringBuffer();
    private Stack stack = new Stack();
    private Hashtable anchorMap = new Hashtable();
    private int nextAnchorId = 0;
    private DocumentModel doc = null;
    private URL docURL = null;
    private int notifyObserverCountdown = 0;

    private static final String[] imageSuffixes = { ".gif", ".jpeg", ".jpg" };

    public void setObserver(WorkInProgressObserver observer) {
	this.observer = observer;
    }

    public WorkInProgressObserver getObserver() {
	return observer;
    }

    public void loadDocument(Reader in, URL url, DocumentModel document) 
	throws IOException, ParseException {
	notifyObserverCountdown = 0;
	if (observer != null && !observer.workInProgress(this, true))
	    throw new WorkCanceledException(this);

	if (docType == null)
	    loadResources();

	doc = document;
	docURL = url;
	ElementModel rootElement = doc.createElementModel("apt");
	stack.push(rootElement);

	try {
	    parser.parse(new ReaderSource(in), this);
	} finally {
	    popAll();
	    anchorMap.clear();
	    nextAnchorId = 0;
	    buffer = new StringBuffer();
	    doc = null;

	    if (observer != null)
		observer.workInProgress(this, false);
	}

	document.initializeDocument(rootElement, docType);
	document.putDocumentProperty(DocumentModel.URL_PROPERTY, url);
	document.putDocumentProperty(StyleSheetInfo.PROPERTY, styleSheetInfo);
    }

    private void loadResources() throws IOException {
	URL url = getClass().getResource("apt.css");
	StyleSheetInfo info = new StyleSheetInfo();
	info.content = ParseUtil.loadSource(url);
	info.url = url.toExternalForm();
	info.type = "text/css";
	info.title = "APT";
	info.alternate = false;
	styleSheetInfo = new StyleSheetInfo[] { info };

	Document template = new Document();
	(new XPDocumentLoader()).load(getClass().getResource("apt.xml"), 
				      template);
	docType = template.getDocumentType();
    }

    private void popAll() {
	while (!stack.empty())
	    stack.pop();
    }

    private String getAnchorId(String name) {
	String key = anchorNameToKey(name);
	String id = (String) anchorMap.get(key);
	if (id == null) {
	    id = "z" + nextAnchorId++;
	    anchorMap.put(key, id);
	}
	return id;
    }

    private static final String anchorNameToKey(String name) {
	int length = name.length();
	StringBuffer buffer = new StringBuffer(length);

	for (int i = 0; i < length; ++i) {
	    char c = name.charAt(i);
	    if (Character.isLetterOrDigit(c))
		buffer.append(Character.toLowerCase(c));
	}

	return buffer.toString();
    }

    public void head() throws ParseException {
	pushElement("head", null);
    }

    public void head_() throws ParseException {
	ElementModel head = topElement();

	popElement();

	if (head.getSubelementCount() == 0)
	    // Remove useless empty Head.
	    topElement().removeSubelement(head);
    }

    public void body() throws ParseException {
	pushElement("body", null);
    }

    public void body_() throws ParseException {
	popElement();
    }

    public void section1() throws ParseException {
	pushElement("section1", null);
    }

    public void section1_() throws ParseException {
	popElement();
    }

    public void section2() throws ParseException {
	pushElement("section2", null);
    }

    public void section2_() throws ParseException {
	popElement();
    }

    public void section3() throws ParseException {
	pushElement("section3", null);
    }

    public void section3_() throws ParseException {
	popElement();
    }

    public void section4() throws ParseException {
	pushElement("section4", null);
    }

    public void section4_() throws ParseException {
	popElement();
    }

    public void section5() throws ParseException {
	pushElement("section5", null);
    }

    public void section5_() throws ParseException {
	popElement();
    }

    public void list() throws ParseException {
	pushElement("list", null);
    }

    public void list_() throws ParseException {
	popElement();
    }

    public void listItem() throws ParseException {
	pushElement("listitem", null);
    }

    public void listItem_() throws ParseException {
	popElement();
    }

    public void numberedList(int numbering) throws ParseException {
	String numberingValue;
	switch (numbering) {
	case Sink.NUMBERING_UPPER_ALPHA:
	    numberingValue = "upperalpha";
	    break;
	case Sink.NUMBERING_LOWER_ALPHA:
	    numberingValue = "loweralpha";
	    break;
	case Sink.NUMBERING_UPPER_ROMAN:
	    numberingValue = "upperroman";
	    break;
	case Sink.NUMBERING_LOWER_ROMAN:
	    numberingValue = "lowerroman";
	    break;
	case Sink.NUMBERING_DECIMAL:
	default:
	    numberingValue = "decimal";
	}
	pushElement("numberedlist", 
		    new String[] { "numbering", numberingValue });
    }

    public void numberedList_() throws ParseException {
	popElement();
    }

    public void numberedListItem() throws ParseException {
	pushElement("numberedlistitem", null);
    }

    public void numberedListItem_() throws ParseException {
	popElement();
    }

    public void definitionList() throws ParseException {
	pushElement("definitionlist", null);
    }

    public void definitionList_() throws ParseException {
	popElement();
    }

    public void definitionListItem() throws ParseException {
	pushElement("definitionlistitem", null);
    }

    public void definitionListItem_() throws ParseException {
	popElement();
    }

    public void definition() throws ParseException {
	pushElement("definition", null);
    }

    public void definition_() throws ParseException {
	popElement();
    }

    public void figure() throws ParseException {
	pushElement("figure", null);
    }

    public void figure_() throws ParseException {
	popElement();
    }

    public void table() throws ParseException {
	pushElement("table", null);
    }

    public void table_() throws ParseException {
	popElement();
    }

    public void tableRows(int[] justification, boolean grid) 
	throws ParseException {
	String gridValue = grid? "yes" : "no";

	StringBuffer alignValue = new StringBuffer();
	for (int i = 0; i < justification.length; ++i) {
	    switch (justification[i]) {
	    case Parser.JUSTIFY_LEFT:
		alignValue.append('l');
		break;
	    case Parser.JUSTIFY_RIGHT:
		alignValue.append('r');
		break;
	    case Parser.JUSTIFY_CENTER:
	    default:
		alignValue.append('c');
	    }
	}

	String[] attributes = { 
	    "align", alignValue.toString(), "grid", gridValue 
	};
	pushElement("rows", attributes);
    }

    public void tableRows_() throws ParseException {
	popElement();
    }

    public void tableRow() throws ParseException {
	pushElement("row", null);
    }

    public void tableRow_() throws ParseException {
	popElement();
    }

    public void title() throws ParseException {
	pushElement("title", null);
    }

    public void title_() throws ParseException {
	popElement();
    }

    public void author() throws ParseException {
	pushElement("author", null);
    }

    public void author_() throws ParseException {
	popElement();
    }

    public void date() throws ParseException {
	pushElement("date", null);
    }

    public void date_() throws ParseException {
	popElement();
    }

    public void sectionTitle() throws ParseException {
	pushElement("title", null);
    }

    public void sectionTitle_() throws ParseException {
	popElement();
    }

    public void paragraph() throws ParseException {
	if (observer != null && (notifyObserverCountdown++ % 10) == 0) {
	    if (!observer.workInProgress(this, true))
		throw new WorkCanceledException(this);
	}

	pushElement("paragraph", null);
    }

    public void paragraph_() throws ParseException {
	popElement();
    }

    public void verbatim(boolean boxed) throws ParseException {
	String boxValue = boxed? "yes" : "no";
	pushElement("verbatim", new String[] { "box", boxValue });
    }

    public void verbatim_() throws ParseException {
	popElement();
    }

    public void definedTerm() throws ParseException {
	pushElement("definedterm", null);
    }

    public void definedTerm_() throws ParseException {
	popElement();
    }

    public void figureCaption() throws ParseException {
	pushElement("caption", null);
    }

    public void figureCaption_() throws ParseException {
	popElement();
    }

    public void tableCell() throws ParseException {
	pushElement("cell", null);
    }

    public void tableCell_() throws ParseException {
	popElement();
    }

    public void tableCaption() throws ParseException {
	pushElement("caption", null);
    }

    public void tableCaption_() throws ParseException {
	popElement();
    }

    public void figureGraphics(String name) throws ParseException {
	String fileValue = name;

	if (!(new File(name)).isAbsolute() &&
	    docURL.getProtocol().equals("file")) {
	    String dirName = docURL.getFile();
	    int slash = dirName.lastIndexOf('/');
	    if (slash < 0)
		dirName = null;
	    else
		dirName = dirName.substring(0, slash);

	    if (dirName != null && File.separatorChar != '/')
		dirName = dirName.replace('/', File.separatorChar);

	    String fileName = ((dirName == null)? name : 
			       (dirName + File.separatorChar + name));

	    for (int i = 0; i < imageSuffixes.length; ++i) {
		String imageSuffix = imageSuffixes[i];

		File file = new File(fileName + imageSuffix);
		if (file.isFile()) {
		    fileValue = name + imageSuffix;
		    break;
		}
	    }
	}

	newElement("graphics", new String[] { "file", fileValue });
    }

    public void horizontalRule() throws ParseException {
	newElement("horizontalrule", null);
    }

    public void pageBreak() throws ParseException {
	newElement("pagebreak", null);
    }

    public void anchor(String name) throws ParseException {
	pushElement("anchor", new String[] { "id", getAnchorId(name) });
    }

    public void anchor_() throws ParseException {
	popElement();
    }

    public void link(String name) throws ParseException {
	URL url;
	try {
	    url = new URL(name);
	} catch (MalformedURLException ignored) {
	    url = null;
	}

	if (url == null)
	    pushElement("link", new String[] { "id", getAnchorId(name) });
	else
	    pushElement("externallink", new String[] { "url", name });
    }

    public void link_() throws ParseException {
	popElement();
    }

    public void italic() throws ParseException {
	pushElement("italic", null);
    }

    public void italic_() throws ParseException {
	popElement();
    }

    public void bold() throws ParseException {
	pushElement("bold", null);
    }

    public void bold_() throws ParseException {
	popElement();
    }

    public void monospaced() throws ParseException {
	pushElement("monospaced", null);
    }

    public void monospaced_() throws ParseException {
	popElement();
    }

    public void lineBreak() throws ParseException {
	newElement("linebreak", null);
    }

    public void nonBreakingSpace() throws ParseException {
	buffer.append('\u00A0');
    }

    public void text(String text) throws ParseException {
	buffer.append(text);
    }

    private ElementModel newElement(String name, String[] attributes) {
	ElementModel parent = topElement();

	if (buffer.length() > 0) {
	    parent.appendSubelement(doc.createTextModel(buffer.toString()));
	    buffer = new StringBuffer();
	}

	ElementModel element = doc.createElementModel(name);
	if (attributes != null) {
	    for (int i = 0; i < attributes.length; i += 2) 
		element.putElementAttribute(attributes[i], attributes[i+1]);
	}
	parent.appendSubelement(element);
	return element;
    }

    private ElementModel topElement() {
	return (stack.empty()? null : (ElementModel) stack.peek());
    }

    private void pushElement(String name, String[] attributes) {
	stack.push(newElement(name, attributes));
    }

    private void popElement() {
	ElementModel top = topElement();

	if (buffer.length() > 0) {
	    top.appendSubelement(doc.createTextModel(buffer.toString()));
	    buffer = new StringBuffer();
	} else {
	    // If top is an empty mix, insert a Text in it ---

	    if (top.getSubelementCount() == 0 && 
		ElementUtil.hasMixedContent(top, docType)) 
		top.appendSubelement(doc.createTextModel(""));
	}

	stack.pop();
    } 
}

