/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.aptformat;

import java.io.*;
import java.net.URL;
import java.util.Vector;
import java.util.Stack;
import fr.pixware.apt.parse.Sink;
import fr.pixware.apt.convert.AlphaNumerals;
import fr.pixware.apt.convert.RomanNumerals;
import com.xmlmind.xmledit.model.*;

public class APTWriter {
    public static final int MAX_LINE_LENGTH = 78;

    private Stack numberingStack = new Stack();

    // -----------------------------------------------------------------------

    public void writeDocument(DocumentModel document, Writer out, 
			      URL url) throws IOException {
	try {
	    ElementModel root = document.getRootElement();
	    ElementModel child = firstChildElement(root);
	    if (isElement(child, "head")) {
		writeHead(child, out);
		child = nextChildElement(child);
	    }

	    matchElement(child, "body");
	    writeBody(child, out);

	    out.flush();
	} finally {
	    popAllNumberings();
	}
    }

    private void popAllNumberings() {
	while (!numberingStack.empty())
	    numberingStack.pop();
    }

    private static ElementModel firstChildElement(ElementModel element) {
	SubelementModel sub = element.getFirstSubelement();
	if (sub != null &&
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    throw new RuntimeException("expected an element, found some text");
	return (ElementModel) sub;
    }

    private static TextModel firstChildText(ElementModel element) {
	SubelementModel sub = element.getFirstSubelement();
	if (sub != null &&
	    sub.getSubelementType() != SubelementModel.TEXT)
	    throw new RuntimeException("expected some text, found " +
		((ElementModel) sub).getElementTypeName());
	return (TextModel) sub;
    }

    private static ElementModel nextChildElement(ElementModel element) {
	SubelementModel sub = element.getNextSubelement();
	if (sub != null &&
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    throw new RuntimeException("expected an element, found some text");
	return (ElementModel) sub;
    }

    private static boolean isElement(ElementModel element, 
				     String elementTypeName) {
	return (element != null && 
		element.getElementTypeName().equals(elementTypeName));
    }

    private static void matchElement(ElementModel element, 
				     String elementTypeName) {
	if (!isElement(element, elementTypeName)) {
	    if (element == null) {
		throw new RuntimeException("expected " + elementTypeName);
	    } else {
		throw new RuntimeException(
		    "expected " + elementTypeName + ", found " + 
		    element.getElementTypeName());
	    }
	}
    }

    private void writeHead(ElementModel element, Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);

	matchElement(child, "title");
	int lineWidth = writeTitleLines(blockLines(child), out);
	child = nextChildElement(child);

	if (isElement(child, "author")) {
	    writeTitleLines(blockLines(child), out);
	    child = nextChildElement(child);
	} else {
	    writeBorder("---", (MAX_LINE_LENGTH-lineWidth)/2, lineWidth, out);
	}

	if (isElement(child, "date"))
	    writeTitleLines(blockLines(child), out);
	
	out.write('\n');
    }

    private static int writeTitleLines(String[] lines, Writer out) 
	throws IOException {
	int count = lines.length;
	int maxWidth = 0;
	int i;

	for (i = 0; i < count; ++i) {
	    String line = lines[i];
	    int width = line.length();
	    if (width > maxWidth)
		maxWidth = width;
	}

	if (maxWidth < 3)
	    maxWidth = 3;
	else if (maxWidth > MAX_LINE_LENGTH-2) 
	    maxWidth = MAX_LINE_LENGTH-2;
	writeBorder("---", (MAX_LINE_LENGTH - maxWidth)/2, maxWidth, out);

	for (i = 0; i < count; ++i) {
	    String line = lines[i];
	    int width = line.length();
	    if (width > MAX_LINE_LENGTH-2)
		width = MAX_LINE_LENGTH-2;

	    writeChars(' ', (MAX_LINE_LENGTH - width)/2, out);
	    out.write(line);
	    out.write('\n');
	}

	return maxWidth;
    }

    private static void writeBorder(String border, int indent, int length, 
				    Writer out) throws IOException {
	writeChars(' ', indent, out);
	out.write(border.charAt(0));
	writeChars(border.charAt(1), length - 2, out);
	out.write(border.charAt(2));
	out.write('\n');
    }

    private static void writeChars(char c, int count, 
				   Writer out) throws IOException {
	for (int i = 0; i < count; ++i) 
	    out.write(c);
    }

    private static String[] blockLines(ElementModel element) {
	Vector lines = new Vector();
	StringBuffer line = new StringBuffer();

	line = blockLines(element, lines, line);

	if (line.length() > 0)
	    lines.addElement(line.toString());

	String[] result = new String[lines.size()];
	lines.copyInto(result);
	return result;
    }

    private static StringBuffer blockLines(ElementModel element, Vector lines, 
					   StringBuffer line) {
	String elemType = element.getElementTypeName();
	if (elemType.equals("italic")) {
	    line.append("<");
	} else if (elemType.equals("bold")) {
	    line.append("<<");
	} else if (elemType.equals("monospaced")) {
	    line.append("<<<");
	} else if (elemType.equals("anchor")) {
	    line.append("{");
	} else if (elemType.equals("link")) {
	    line.append("{{");
	} else if (elemType.equals("externallink")) {
	    line.append("{{{");
	    String url = element.getElementAttribute("url");
	    if (url == null)
		throw new RuntimeException(
		    "externallink with no url attribute");

	    line.append(url.trim());
	    line.append('}');
	} else if (elemType.equals("linebreak")) {
	    // Implicit "\\\n" at the end of each line but the last one.
	    lines.addElement(line.toString());
	    line = new StringBuffer();
	}

	SubelementModel child = element.getFirstSubelement();
	while (child != null) {
	    switch (child.getSubelementType()) {
	    case SubelementModel.TEXT:
		{
		    char[] chars = ((TextModel) child).getTextChars();
		    if (chars.length > 0) {
			for (int i = 0; i < chars.length; ++i) {
			    char c = chars[i];

			    switch (c) {
			    case '~': case '=': case '-': case '+': case '*':  
				if (line.length() == 0)
				    line.append('\\');
				line.append(c);
				break;
			    case '\\': case '<': case '>': case '{': case '}': 
			    case '[': case ']': case '|':
				line.append('\\');
				line.append(c);
				break;
			    case '\u00A0': // &nbsp;
				line.append("\\ ");
				break;
			    default:
				if (Character.isWhitespace(c))
				    line.append(' ');
				else
				    line.append(c);
			    }
			}
		    } else {
			// Empty PCDATA: leaving it as is may generate an
			// illegal APT document (ex. empty definedterm).
			line.append('?');
		    }
		}
		break;

	    case SubelementModel.ELEMENT:
		line = blockLines((ElementModel) child, lines, line);
		break;
	    }

	    child = child.getNextSubelement();
	}

	if (elemType.equals("italic")) {
	    line.append(">");
	} else if (elemType.equals("bold")) {
	    line.append(">>");
	} else if (elemType.equals("monospaced")) {
	    line.append(">>>");
	} else if (elemType.equals("anchor")) {
	    line.append("}");
	} else if (elemType.equals("link") ||
		   elemType.equals("externallink")) {
	    line.append("}}");
	}

	return line;
    }

    private void writeBody(ElementModel element, Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);

	child = writeChildBlocks(child, out);

	while (child != null) {
	    writeSection(child, 1, out);
	    child = nextChildElement(child);
	}
    }

    private ElementModel writeChildBlocks(ElementModel child, Writer out) 
	throws IOException {
	while (child != null) {
	    String elemType = child.getElementTypeName();

	    if (elemType.equals("paragraph")) {
		writeParagraph(child, 2, out);
	    } else if (elemType.equals("verbatim")) {
		writeVerbatim(child, out);
	    } else if (elemType.equals("list")) {
		writeList(child, 1, 2+4, out);
	    } else if (elemType.equals("numberedlist")) {
		writeNumberedList(child, 1, 2+4, out);
	    } else if (elemType.equals("definitionlist")) {
		writeDefinitionList(child, 1, 2+4, out);
	    } else if (elemType.equals("figure")) {
		writeFigure(child, out);
	    } else if (elemType.equals("table")) {
		writeTable(child, out);
	    } else if (elemType.equals("horizontalrule")) {
		writeHorizontalRule(child, out);
	    } else if (elemType.equals("pagebreak")) {
		writePageBreak(child, out);
	    } else {
		break;
	    }

	    child = nextChildElement(child);
	}

	return child;
    }

    private void writeSection(ElementModel element, int sectionLevel, 
			      Writer out) throws IOException {
	ElementModel child = firstChildElement(element);

	matchElement(child, "title");
	writeLeftBlockLines(blockLines(child), 0, 0, sectionLevel-1, 
			    true, out);
	child = nextChildElement(child);

	child = writeChildBlocks(child, out);

	while (child != null) {
	    writeSection(child, sectionLevel+1, out);
	    child = nextChildElement(child);
	}
    }

    private static void writeLeftBlockLines(String[] lines, int indent, 
					    int firstLineIndent, int bullets,
					    boolean underline,
					    Writer out) throws IOException {

	int lastLine = lines.length-1;
	for (int i = 0; i <= lastLine; ++i) {
	    String line = lines[i];
	    int column;

	    if (i == 0) {
		column = indent + firstLineIndent;
		writeChars(' ', column, out);
		if (bullets > 0) {
		    writeChars('*', bullets, out);
		    out.write(' ');
		    column += bullets+1;
		}
	    } else {
		column = indent;
		writeChars(' ', column, out);
	    }

	    int lineWidth = column + line.length();
	    if (lineWidth <= MAX_LINE_LENGTH) {
		out.write(line);
	    } else {
		lineWidth = MAX_LINE_LENGTH;
		writeWrappedLine(line, column, indent, out);
	    }

	    if (i == lastLine) {
		if (underline) {
		    out.write('\n');
		    writeChars('~', lineWidth, out);
		}
		out.write("\n\n");
	    } else {
		out.write("\\\n");
	    }
	}
    }

    private static void writeWrappedLine(String line, int column, int indent,
					 Writer out) throws IOException {
	boolean firstWrap = true;
	int endPos = line.length();
	int pos = 0;
	int remainPos = 0;
	int wrapPos = -1;
	char prevChar = '\n';
	char prevPrevChar = '\n';

	while (pos <= endPos) {
	    char c = (pos == endPos)? ' ' : line.charAt(pos);

	    if (c == ' ' && (prevChar != '\\' || prevPrevChar == '\\')) {
		if (column + (pos - remainPos) > MAX_LINE_LENGTH) {
		    if (firstWrap) {
			firstWrap = false;
		    } else {
			out.write('\n');
			writeChars(' ', indent, out);
		    }
		    column = indent;

		    if (wrapPos <= remainPos)
			// Too long to be wrapped.
			wrapPos = pos;
		    out.write(line.substring(remainPos, wrapPos));
		    remainPos = wrapPos + 1;
		} else {
		    wrapPos = pos;
		}
	    }

	    prevPrevChar = prevChar;
	    prevChar = c;
	    ++pos;
	}

	if (remainPos < endPos) {
	    if (!firstWrap) {
		out.write('\n');
		writeChars(' ', indent, out);
	    }
	    out.write(line.substring(remainPos, endPos));
	}
    }

    private void writeParagraph(ElementModel element, int indent, 
				Writer out) throws IOException {
	writeLeftBlockLines(blockLines(element), indent, 0, 0, false, out);
    }

    private void writeList(ElementModel element, int listNesting, 
			   int indent, Writer out) throws IOException {
	ElementModel child = firstChildElement(element);
	do {
	    matchElement(child, "listitem");
	    writeListItem(child, listNesting, indent, out);
	    child = nextChildElement(child);
	} while (child != null);

	breakList(element, indent, out);
    }

    private static void breakList(ElementModel element, int indent, 
				  Writer out) throws IOException {
	SubelementModel next = element.getNextSubelement();
	if (next == null || next.getSubelementType() == SubelementModel.TEXT)
	    return;

	String nextType = ((ElementModel) next).getElementTypeName();
	if (nextType.equals("verbatim") ||
	    nextType.equals("figure") ||
	    nextType.equals("table") ||
	    nextType.equals("horizontalrule") ||
	    nextType.equals("pagebreak")) {
	    // Just end current list and no more (current list may be nested
	    // in a parent list).
	    writeChars(' ', indent, out);
	    out.write("[]\n\n");
	}
    }

    private void writeListItem(ElementModel element, int listNesting, 
			       int indent, Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);

	matchElement(child, "paragraph");
	writeLeftBlockLines(blockLines(child), indent+2, -2, 1, false, out);
	child = nextChildElement(child);

	while (child != null) {
	    String elemType = child.getElementTypeName();

	    if (elemType.equals("paragraph")) {
		writeParagraph(child, indent+2, out);
	    } else if (elemType.equals("verbatim")) {
		writeVerbatim(child, out);
	    } else if (elemType.equals("list")) {
		writeList(child, listNesting+1, indent+2+4, out);
	    } else if (elemType.equals("numberedlist")) {
		writeNumberedList(child, listNesting+1, indent+2+4, out);
	    } else if (elemType.equals("definitionlist")) {
		writeDefinitionList(child, listNesting+1, indent+2+4, out);
	    } else if (elemType.equals("figure")) {
		writeFigure(child, out);
	    } else if (elemType.equals("table")) {
		writeTable(child, out);
	    } else if (elemType.equals("horizontalrule")) {
		writeHorizontalRule(child, out);
	    } else if (elemType.equals("pagebreak")) {
		writePageBreak(child, out);
	    }

	    child = nextChildElement(child);
	}
    }

    private void writeNumberedList(ElementModel element, 
				   int listNesting, int indent, 
				   Writer out) throws IOException {
	pushNumbering(element);

	ElementModel child = firstChildElement(element);
	do {
	    matchElement(child, "numberedlistitem");
	    writeNumberedListItem(child, listNesting, indent, out);
	    child = nextChildElement(child);
	} while (child != null);

	numberingStack.pop();

	breakList(element, indent, out);
    }

    private void pushNumbering(ElementModel element) {
	int numbering;

	String style = element.getElementAttribute("numbering");
	if ("decimal".equals(style)) 
	    numbering = Sink.NUMBERING_DECIMAL;
	else if ("loweralpha".equals(style)) 
	    numbering = Sink.NUMBERING_LOWER_ALPHA;
	else if ("upperalpha".equals(style)) 
	    numbering = Sink.NUMBERING_UPPER_ALPHA;
	else if ("lowerroman".equals(style)) 
	    numbering = Sink.NUMBERING_LOWER_ROMAN;
	else if ("upperroman".equals(style)) 
	    numbering = Sink.NUMBERING_UPPER_ROMAN;
	else
	    numbering = Sink.NUMBERING_DECIMAL;

	numberingStack.push(new int[] { numbering, 1 });
    }

    private void writeNumberedListItem(ElementModel element, 
				       int listNesting, int indent, 
				       Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);

	matchElement(child, "paragraph");
	writeLeftBlockLines(
	    mergeLines(new String[] { incrNumbering() } , blockLines(child)), 
	    indent+6, -6, 0, false, out);
	child = nextChildElement(child);

	while (child != null) {
	    String elemType = child.getElementTypeName();

	    if (elemType.equals("paragraph")) {
		writeParagraph(child, indent+6, out);
	    } else if (elemType.equals("verbatim")) {
		writeVerbatim(child, out);
	    } else if (elemType.equals("list")) {
		writeList(child, listNesting+1, indent+6+4, out);
	    } else if (elemType.equals("numberedlist")) {
		writeNumberedList(child, listNesting+1, indent+6+4, out);
	    } else if (elemType.equals("definitionlist")) {
		writeDefinitionList(child, listNesting+1, indent+6+4, out);
	    } else if (elemType.equals("figure")) {
		writeFigure(child, out);
	    } else if (elemType.equals("table")) {
		writeTable(child, out);
	    } else if (elemType.equals("horizontalrule")) {
		writeHorizontalRule(child, out);
	    } else if (elemType.equals("pagebreak")) {
		writePageBreak(child, out);
	    }

	    child = nextChildElement(child);
	}
    }

    private String incrNumbering() {
	StringBuffer label = new StringBuffer();
	label.append("[[");

	int[] top = (int[]) numberingStack.peek();

	switch (top[0]) {
	case Sink.NUMBERING_LOWER_ALPHA:
	    label.append(AlphaNumerals.toString(top[1], /*lower*/ true));
	    break;
	case Sink.NUMBERING_UPPER_ALPHA:
	    label.append(AlphaNumerals.toString(top[1], /*lower*/ false));
	    break;
	case Sink.NUMBERING_LOWER_ROMAN:
	    label.append(RomanNumerals.toString(top[1], /*lower*/ true));
	    break;
	case Sink.NUMBERING_UPPER_ROMAN:
	    label.append(RomanNumerals.toString(top[1], /*lower*/ false));
	    break;
	case Sink.NUMBERING_DECIMAL:
	default:
	    label.append(Integer.toString(top[1]));
	}
	++top[1];

	label.append("]]");
	return label.toString();
    }

    private void writeDefinitionList(ElementModel element, 
					    int listNesting, int indent, 
					    Writer out) throws IOException {
	ElementModel child = firstChildElement(element);
	do {
	    matchElement(child, "definitionlistitem");
	    writeDefinitionListItem(child, listNesting, indent, out);
	    child = nextChildElement(child);
	} while (child != null);

	breakList(element, indent, out);
    }

    private void writeDefinitionListItem(ElementModel element, 
					 int listNesting, int indent, 
					 Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);

	matchElement(child, "definedterm");

	String[] definedTermLines = blockLines(child);
	int last = definedTermLines.length-1;
	if (last < 0)
	    throw new RuntimeException("empty definedterm");
	definedTermLines[0] = "[" + definedTermLines[0];
	definedTermLines[last] = definedTermLines[last] + ']';

	child = nextChildElement(child);

	matchElement(child, "definition");
	writeDefinition(child, definedTermLines, listNesting, indent, out);
    }

    private void writeDefinition(ElementModel element, 
				 String[] definedTermLines, 
				 int listNesting, int indent, Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);

	matchElement(child, "paragraph");
	writeLeftBlockLines(mergeLines(definedTermLines, blockLines(child)), 
			    indent+4, -4, 0, false, out);
	child = nextChildElement(child);

	while (child != null) {
	    String elemType = child.getElementTypeName();

	    if (elemType.equals("paragraph")) {
		writeParagraph(child, indent+4, out);
	    } else if (elemType.equals("verbatim")) {
		writeVerbatim(child, out);
	    } else if (elemType.equals("list")) {
		writeList(child, listNesting+1, indent+4+4, out);
	    } else if (elemType.equals("numberedlist")) {
		writeNumberedList(child, listNesting+1, indent+4+4, out);
	    } else if (elemType.equals("definitionlist")) {
		writeDefinitionList(child, listNesting+1, indent+4+4, out);
	    } else if (elemType.equals("figure")) {
		writeFigure(child, out);
	    } else if (elemType.equals("table")) {
		writeTable(child, out);
	    } else if (elemType.equals("horizontalrule")) {
		writeHorizontalRule(child, out);
	    } else if (elemType.equals("pagebreak")) {
		writePageBreak(child, out);
	    }

	    child = nextChildElement(child);
	}
    }

    private static String[] mergeLines(String[] lines1, String[] lines2) {
	if (lines1.length == 0)
	    return lines2;

	if (lines2.length == 0)
	    return lines1;

	String[] merged = new String[lines1.length + lines2.length - 1];

	System.arraycopy(lines1, 0, merged, 0, lines1.length);
	merged[lines1.length-1] += ' ' + lines2[0];
	if (lines2.length > 1)
	    System.arraycopy(lines2, 1, merged, lines1.length, 
			     lines2.length-1);

	return merged;
    }

    private void writeVerbatim(ElementModel element, Writer out) 
	throws IOException {
	String boxValue = element.getElementAttribute("box");
	String border = ("yes".equals(boxValue))? "+-+" : "---";
	writeBorder(border, 0, MAX_LINE_LENGTH, out);

	TextModel text = firstChildText(element);
	char[] chars = text.getTextChars();
	out.write(chars);
	if (chars.length == 0 || chars[chars.length-1] != '\n')
	    out.write('\n');

	writeBorder(border, 0, MAX_LINE_LENGTH, out);
	out.write('\n');
    }

    private void writeFigure(ElementModel element, Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);
	
	matchElement(child, "graphics");
	String graphicsName = child.getElementAttribute("file");
	if (graphicsName == null)
	    throw new RuntimeException("graphics with no file attribute");

	// Remove extension.
	int dot = graphicsName.lastIndexOf('.');
	if (dot > 0)
	    graphicsName = graphicsName.substring(0, dot);
	graphicsName = "[" + graphicsName + ']';

	child = nextChildElement(child);

	if (isElement(child, "caption")) {
	    writeLeftBlockLines(
		mergeLines(new String[] { graphicsName }, blockLines(child)), 
		0, 0, 0, false, out);
	} else {
	    out.write(graphicsName);
	    out.write("\n\n");
	}
    }

    private void writeHorizontalRule(ElementModel element, Writer out) 
	throws IOException {
	writeChars('=', MAX_LINE_LENGTH, out);
	out.write("\n\n");
    }

    private void writePageBreak(ElementModel element, Writer out) 
	throws IOException {
	out.write("\f\n");
    }

    private void writeTable(ElementModel element, Writer out) 
	throws IOException {
	ElementModel child = firstChildElement(element);

	matchElement(child, "rows");
	String align = child.getElementAttribute("align");
	boolean grid = "yes".equals(child.getElementAttribute("grid"));

	Vector lines = cellLines(child);
	int columnCount = computeColumnCount(lines);
	int[] columnWidthes = computeColumnWidthes(lines, columnCount);
	align = checkAlign(align, columnCount);
	String rowSeparator = makeRowSeparator(columnWidthes, align);
	
	out.write(rowSeparator);
	int lineCount = lines.size();
	for (int i = 0; i < lineCount; ++i) {
	    Object line = lines.elementAt(i);
	    if (line instanceof Boolean) {
		out.write(rowSeparator);
		continue;
	    }

	    String[] cells = (String[]) line;
	    int cellCount = cells.length;
	    for (int j = 0; j < columnCount; ++j) {
		String cell = (j < cellCount)? cells[j] : "";

		if (j == 0) {
		    if (grid)
			out.write("| ");
		    else
			out.write("  ");
		} else {
		    out.write(" | ");
		}
		out.write(cell);
		int pad = columnWidthes[j] - cell.length();
		if (pad > 0)
		    writeChars(' ', pad, out);
	    }
	    out.write('\n');
	}

	child = nextChildElement(child);

	if (isElement(child, "caption"))
	    writeLeftBlockLines(blockLines(child), 0, 0, 0, false, out);
	else
	    out.write('\n');
    }

    private static Vector cellLines(ElementModel rows) {
	Vector lines = new Vector();
	Vector rowCellLines = new Vector();

	ElementModel row = firstChildElement(rows);
	while (row != null) {
	    matchElement(row, "row");
	    rowCellLines.removeAllElements();

	    ElementModel cell = firstChildElement(row);
	    while (cell != null) {
		matchElement(cell, "cell");
		rowCellLines.addElement(blockLines(cell));
		cell = nextChildElement(cell);
	    }

	    int columnCount = rowCellLines.size();
	    int lineIndex = 0;
	    for (;;) {
		boolean done = true;
		for (int i = 0; i < columnCount; ++i) {
		    String[] cellLines = (String[]) rowCellLines.elementAt(i);
		    if (lineIndex < cellLines.length) {
			done = false;
			break;
		    }
		}
		if (done)
		    break;

		String[] line = new String[columnCount];
		for (int i = 0; i < columnCount; ++i) {
		    String[] cellLines = (String[]) rowCellLines.elementAt(i);
		    if (lineIndex < cellLines.length) 
			line[i] = cellLines[lineIndex].trim();
		    else
			line[i] = "";
		}
		lines.addElement(line);

		++lineIndex;
	    }

	    // Add row separator.
	    lines.addElement(Boolean.TRUE);

	    row = nextChildElement(row);
	}

	return lines;
    }

    private static int computeColumnCount(Vector lines) {
	int columnCount = 0;

	int lineCount = lines.size();
	for (int i = 0; i < lineCount; ++i) {
	    Object line = lines.elementAt(i);
	    if (line instanceof Boolean)
		// Skip row separator.
		continue;

	    String[] cells = (String[]) line;
	    if (cells.length > columnCount)
		columnCount = cells.length;
	}

	return columnCount;
    }

    private static String checkAlign(String align, int columnCount) {
	if (align == null) 
	    align = "";

	StringBuffer buffer = new StringBuffer();

	int length = align.length();
	for (int i = 0; i < columnCount; ++i) {
	    if (i < length) {
		char al = align.charAt(i);

		switch (al) {
		case 'c': case 'r': case 'l':
		    buffer.append(al);
		    break;
		default:
		    buffer.append('c');
		}
	    } else {
		buffer.append('c');
	    }
	}

	return buffer.toString();
    }

    private static int[] computeColumnWidthes(Vector lines, int columnCount) {
	int[] columnWidthes = new int[columnCount];
	for (int j = 0; j < columnCount; ++j) 
	    columnWidthes[j] = 0;

	int lineCount = lines.size();
	for (int i = 0; i < lineCount; ++i) {
	    Object line = lines.elementAt(i);
	    if (line instanceof Boolean)
		// Skip row separator.
		continue;

	    String[] cells = (String[]) line;
	    int cellCount = cells.length;
	    for (int j = 0; j < columnCount; ++j) {
		if (j < cellCount) {
		    int w = cells[j].length();
		    if (w > columnWidthes[j])
			columnWidthes[j] = w;
		}
	    }
	}

	return columnWidthes;
    }

    private static String makeRowSeparator(int[] columnWidthes, String align) {
	StringBuffer buffer = new StringBuffer();
	
	buffer.append('*');
	for (int i = 0; i < columnWidthes.length; ++i) {
	    int columnWidth = columnWidthes[i];

	    buffer.append('-');
	    while (columnWidth > 0) {
		buffer.append('-');
		--columnWidth;
	    }
	    buffer.append('-');

	    switch (align.charAt(i)) {
	    case 'r':
		buffer.append(':');
		break;
	    case 'l':
		buffer.append('+');
		break;
	    default:
		buffer.append('*');
	    }
	}
	buffer.append('\n');
	
	return buffer.toString();
    }
}

