/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.attedit;

import java.util.EventObject;
import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.KeyStroke;
import javax.swing.table.TableCellEditor;
import javax.swing.event.EventListenerList;
import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;
import fr.pixware.util.QuickSort;
import com.xmlmind.xmledit.model.AttributeDefinitionModel;
import com.xmlmind.xmledit.modeltool.ElementAttributeChecker;

public class AttributeEditor implements TableCellEditor, 
                                        ActionListener, ItemListener {
    private JTextField field;
    private JComboBox combo;

    private EventListenerList listenerList = new EventListenerList();
    private ChangeEvent changeEvent = null;

    private Attribute editedAttr = new Attribute();
    private AttributesModel attrModel;
    private JComponent component;

    // -----------------------------------------------------------------------

    public AttributeEditor() {
	field = new JTextField();
	field.setActionCommand("ok");
	field.addActionListener(this);
	field.registerKeyboardAction(
	    this, "cancel", KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), 
	    JComponent.WHEN_FOCUSED);

	combo = new JComboBox();
	combo.setEditable(false);
	combo.addItemListener(this);
    }

    public Component getTableCellEditorComponent(JTable table, Object value,
						 boolean isSelected,
						 int row, int column) {
	attrModel = (AttributesModel) table.getModel();

	Attribute attr = (Attribute) value;
	editedAttr.name = attr.name;
	editedAttr.value = attr.value;
	editedAttr.definition = attr.definition;

	// This also prevents itemStateChanged from being called during
	// ComboBox configuration.
	component = null;

	if (attr.definition != null) {
	    String[] list = null;

	    switch (attr.definition.getAttributeType()) {
	    case AttributeDefinitionModel.IDREF:
		list = attrModel.getElementIds();
		break;
	    case AttributeDefinitionModel.ENTITY:
		list = attrModel.getUnparsedEntities();
		break;
	    case AttributeDefinitionModel.NOTATION:
	    case AttributeDefinitionModel.ENUMERATION:
		list = attr.definition.getAttributeEnumeratedValues();
		break;
	    }

	    if (list != null) {
		if (list.length > 1)
		    QuickSort.sort(list);

		combo.removeAllItems();

		String selectedValue = attr.value;
		int selectedIndex = -1;
		for (int i = 0; i < list.length; ++i) {
		    String item = list[i];

		    combo.addItem(item);
		    if (item.equals(selectedValue))
			selectedIndex = i;
		}

		if (selectedIndex < 0) {
		    // Value no specified or doesn't belong to the enumeration:
		    // adding a space makes it an invalid NAME or NMTOKEN
		    // and therefore makes it easy to detect and reject it.

		    if (selectedValue == null)
			combo.addItem(" -");
		    else
			combo.addItem(" " + selectedValue);
		    selectedIndex = list.length;
		}
		combo.setSelectedIndex(selectedIndex);
		component = combo;
	    }
	} 

	if (component == null) {
	    field.setText((attr.value == null)? "" : attr.value);
	    component = field;
	}

	component.setFont(table.getFont());
	return component;
    }

    public Object getCellEditorValue() {
        return editedAttr;
    }

    public boolean isCellEditable(EventObject e) {
    	return (e instanceof MouseEvent);
    }
    
    public boolean shouldSelectCell(EventObject e) { 
	return true; 
    }

    // -----------------------------------------------------------------------

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();

	if (command.equals("ok")) 
	    stopCellEditing();
	else if (command.equals("cancel")) 
	    cancelCellEditing();
    }

    public void itemStateChanged(ItemEvent e) {
	if (component != null && e.getStateChange() == ItemEvent.SELECTED)
	    stopCellEditing();
    }

    public boolean stopCellEditing() {
	if (component == field) {
	    editedAttr.value = field.getText();
	} else if (component == combo) {
	    editedAttr.value = (String) combo.getSelectedItem();
	    if (editedAttr.value.charAt(0) == ' ')
		// Ignore the original null or out of enumeration value.
		return false;
	}

	ElementAttributeChecker.Diagnostic d =
	    attrModel.normalizeAndCheck(editedAttr);
	if (d == null) {
	    fireEditingStopped();
	    return true;
	} else {
	    field.getToolkit().beep();
	    Dialogs.showError(field, d.message);
	    if (component == field)
		field.selectAll();
	    return false;
	}
    }

    public void cancelCellEditing() {
	fireEditingCanceled();
    }

    // -----------------------------------------------------------------------

    public void addCellEditorListener(CellEditorListener l) {
	listenerList.add(CellEditorListener.class, l);
    }

    public void removeCellEditorListener(CellEditorListener l) {
	listenerList.remove(CellEditorListener.class, l);
    }

    private void fireEditingStopped() {
	Object[] listeners = listenerList.getListenerList();

	for (int i = listeners.length-2; i >= 0;  i -= 2) {
	    if (listeners[i] == CellEditorListener.class) {
		if (changeEvent == null)
		    changeEvent = new ChangeEvent(this);

		((CellEditorListener) listeners[i+1]).editingStopped(
		    changeEvent);
	    }	       
	}
    }

    private void fireEditingCanceled() {
	Object[] listeners = listenerList.getListenerList();

	for (int i = listeners.length-2; i >= 0; i -= 2) {
	    if (listeners[i] == CellEditorListener.class) {
		if (changeEvent == null)
		    changeEvent = new ChangeEvent(this);

		((CellEditorListener) listeners[i+1]).editingCanceled(
		    changeEvent);
	    }	       
	}
    }
}
