/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.attedit;

import java.util.Enumeration;
import java.util.Hashtable;
import javax.swing.table.AbstractTableModel;
import fr.pixware.util.QuickSort;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.Traversal;
import com.xmlmind.xmledit.modeltool.XMLUtil;
import com.xmlmind.xmledit.modeltool.ElementAttributeChecker;

public class AttributesModel extends AbstractTableModel 
                             implements DocumentListener, Traversal.Handler {
    private ElementAttributeChecker attrChecker=new ElementAttributeChecker();
    private DocumentModel document = null;
    private ElementModel element = null;
    private boolean editable = true;

    private DocumentTypeModel documentType = null;
    private String[] unparsedEntities = null;
    private ElementTypeModel elementType = null;
    private Attribute[] attributes = new Attribute[0];

    private static class AttributeCompare implements QuickSort.Compare {
	public int compare(Object o1, Object o2) {
	    return ((Attribute) o1).name.compareTo(((Attribute) o2).name);
	}
    }
    private static AttributeCompare attributeCompare = new AttributeCompare();

    // -----------------------------------------------------------------------

    public void setDocument(DocumentModel doc) {
	attrChecker.setDocument(doc);

	if (document != null) {
	    setElement(null);

	    document.removeDocumentListener(this);
	    document = null;
	    documentType = null;
	    unparsedEntities = null;
	}

	if (doc != null) {
	    document = doc;
	    documentType = doc.getDocumentType();
	    document.addDocumentListener(this, /*hasData*/ false);
	}
    }

    public DocumentModel getDocument() {
	return document;
    }

    public DocumentTypeModel getDocumentType() {
	return documentType;
    }

    public String[] getUnparsedEntities() {
	if (unparsedEntities == null) {
	    if (documentType == null) {
		unparsedEntities = new String[0];
	    } else {
		unparsedEntities = 
		    new String[documentType.getUnparsedEntityCount()];
	
		Enumeration iter = documentType.getUnparsedEntities();
		int i = 0;
		while (iter.hasMoreElements()) {
		    UnparsedEntityModel entity = 
			(UnparsedEntityModel) iter.nextElement();
		    unparsedEntities[i++] = entity.getUnparsedEntityName();
		}
	    }
	}

	return unparsedEntities;
    }

    public String[] getElementIds() {
	if (documentType == null)
	    return new String[0];

	Hashtable map = attrChecker.getIdToElementMap();
	String[] list = new String[map.size()];

	Enumeration iter = map.keys();
	int i = 0;
	while (iter.hasMoreElements())
	    list[i++] = (String) iter.nextElement();

	return list;
    }

    public void setElement(ElementModel elem) {
	if (elem != element) {
	    element = elem;

	    if (element == null) {
		elementType = null;
		attributes = new Attribute[0];
	    } else {
		if (documentType != null)
		    elementType = documentType.getElementType(
			element.getElementTypeName());
		updateAttributes();
	    }

	    fireTableDataChanged();
	}
    }

    private void updateAttributes() {
	int count = element.getElementAttributeCount();
	if (elementType != null)
	    count += elementType.getAttributeDefinitionCount();
	
	Attribute[] newAttributes = new Attribute[count];
	if (count == 0) {
	    attributes = newAttributes;
	    return;
	}

	count = 0;
	Enumeration iter = element.getElementAttributes();
	while (iter.hasMoreElements()) {
	    String[] pair = (String[]) iter.nextElement(); 

	    Attribute attr = new Attribute();
	    attr.name = pair[0];
	    attr.value = pair[1];
	    newAttributes[count++] = attr;
	}

	if (elementType != null) {
	    int added = count;

	    iter = elementType.getAttributeDefinitions();
	    while (iter.hasMoreElements()) {
		AttributeDefinitionModel attrDef = 
		    (AttributeDefinitionModel) iter.nextElement(); 

		String name = attrDef.getAttributeName();
		int index = indexOfAttribute(newAttributes, added, name);
		if (index < 0) {
		    Attribute attr = new Attribute();
		    attr.name = name;
		    attr.definition = attrDef;
		    newAttributes[count++] = attr;
		} else {
		    newAttributes[index].definition = attrDef;
		}
	    }
	}

	if (count == newAttributes.length) {
	    attributes = newAttributes;
	} else {
	    attributes = new Attribute[count];
	    System.arraycopy(newAttributes, 0, attributes, 0, count);
	}
	if (count > 1)
	    QuickSort.sort(attributes, attributeCompare);
    }

    private static final int indexOfAttribute(Attribute[] attributes,
					      int attributeCount, 
					      String name) {
	for (int i = 0; i < attributeCount; ++i) {
	    if (name.equals(attributes[i].name))
		return i;
	}
	return -1;
    }

    public ElementModel getElement() {
	return element;
    }

    public void setEditable(boolean edit) {
	editable = edit;
    }

    public boolean isEditable() {
	return editable;
    }

    // -----------------------------------------------------------------------

    public int getColumnCount() {
	return 2;
    }

    public int getRowCount() {
	return attributes.length;
    }

    public boolean isCellEditable(int row, int col) {
	if (col == 0 || !editable)
	    return false;

	// Yes, it is possible to edit a fixed attribute for example to remove
	// it.
	return true;
    }

    public Object getValueAt(int row, int col) {
	return attributes[row];
    }

    public void setValueAt(Object value, int row, int col) {
	Attribute attr = (Attribute) value;

	// Name and value are supposed to have been normalized and checked
	// by the cell editor.
	apply(attr);
    }

    public 
    ElementAttributeChecker.Diagnostic normalizeAndCheck(Attribute attr) {
	ElementAttributeChecker.Diagnostic d = 
	    ElementAttributeChecker.checkAttributeName(element, attr.name);
	if (d != null)
	    return d;

	if (attr.value == null)
	    return null;

	// We cannot check before normalizing.
	if (attr.definition == null || 
	    attr.definition.getAttributeType()==AttributeDefinitionModel.CDATA)
	    attr.value = XMLUtil.normalizeAttributeStringValue(attr.value);
	else
	    attr.value = XMLUtil.normalizeAttributeValue(attr.value);
	    
	if (documentType == null)
	    return null;
	else
	    return attrChecker.checkAttribute(element, attr.name, attr.value);
    }

    public void apply(Attribute attr) {
	if (attr.value == null)
	    element.removeElementAttribute(attr.name);
	else
	    element.putElementAttribute(attr.name, attr.value);
    }

    // -----------------------------------------------------------------------

    public void elementAttributeChanged(ElementAttributeEvent event, 
					int listenerIndex) {
	if (element == null || event.getSource() != element)
	    return;

	int index = indexOfAttribute(attributes, attributes.length,
				     event.getName());
	if (index >= 0 && attributes[index].definition != null) {
	    // NewValue is null for a removed attribute.
	    attributes[index].value = event.getNewValue();
	    fireTableCellUpdated(index, 1);
	} else {
	    updateAttributes();
	    fireTableDataChanged();
	}
    }

    public void elementStructureChanged(ElementStructureEvent event, 
					int listenerIndex) {
	if (element == null)
	    return;

	switch (event.getId()) {
	case ElementStructureEvent.CHILD_REPLACED:
	case ElementStructureEvent.CHILD_REMOVED:
	    {
		SubelementModel oldChild = event.getOldChild();
		if (oldChild.getSubelementType() != SubelementModel.ELEMENT)
		    return;
		ElementModel removed = (ElementModel) oldChild;

		if (removed == element ||
		    Traversal.traverse(removed, this) != null)
		    setElement(null);
	    }
	    break;
	}
    }

    public void batchEditingStarted(BatchEditingEvent event, 
				    int listenerIndex) {}
    public void batchEditingCompleted(BatchEditingEvent event, 
				      int listenerIndex) {}
    public void textChanged(TextEvent event, int listenerIndex) {}
    public void documentEventHappened(DocumentEvent event, 
				      int listenerIndex) {}

    // -----------------------------------------------------------------------

    public Object enterElement(ElementModel elem) {
	if (elem == element)
	    return elem;
	else
	    return null;
    }

    public Object processText(TextModel text) { return null; }
    public Object leaveElement(ElementModel element) { return null; }
}
