/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.attedit;

import java.util.Hashtable;
import java.awt.Insets;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JList;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;
import javax.swing.ImageIcon;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import fr.pixware.util.QuickSort;
import fr.pixware.util.StringUtil;
import com.xmlmind.awtutil.ImageResource;

public class ItemsSelector extends JPanel 
                           implements ActionListener, ListSelectionListener {
    public static final int FIRST    = 0;
    public static final int PREVIOUS = 1;
    public static final int NEXT     = 2;
    public static final int LAST     = 3;

    private JList unselectedList;
    private JButton selectButton;
    private JButton unselectButton;
    private JList selectedList;
    private JButton firstButton;
    private JButton previousButton;
    private JButton nextButton;
    private JButton lastButton;

    private static class ItemCompare implements QuickSort.Compare {
	private Hashtable itemOrder = null;

	public void setItemOrder(Hashtable map) {
	    itemOrder = map;
	}

	public int compare(Object o1, Object o2) {
	    Integer rank = (Integer) itemOrder.get(o1);
	    int r1 = (rank == null)? Integer.MAX_VALUE : rank.intValue();
	    rank = (Integer) itemOrder.get(o2);
	    int r2 = (rank == null)? Integer.MAX_VALUE : rank.intValue();

	    return (r1 - r2);
	}
    }
    private ItemCompare itemCompare = new ItemCompare();

    // -----------------------------------------------------------------------

    public ItemsSelector(int rows, int columns, 
			 String unselectedTitle, String selectedTitle) {
        setLayout(new GridBagLayout());
        
        GridBagConstraints constraints = new GridBagConstraints();

	// Unselected list ---

        JPanel panel = new JPanel(new BorderLayout(0, 2));
        constraints.gridx = 0;
        constraints.gridy = 0;
	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
        add(panel, constraints);

	JLabel unselectedLabel = new JLabel(unselectedTitle);
        panel.add(unselectedLabel, BorderLayout.NORTH);
	
        unselectedList = new JList();
	unselectedList.setSelectionMode(
	    ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
	StringBuffer prototype = new StringBuffer();
	while (columns > 0) {
	    prototype.append('x');
	    --columns;
	}
	unselectedList.setPrototypeCellValue(prototype.toString());
	unselectedList.setVisibleRowCount(rows);
	unselectedList.addListSelectionListener(this);

	JScrollPane scroller = new JScrollPane(unselectedList);
        panel.add(scroller, BorderLayout.CENTER);

	// Select/unselect buttons ---

        panel = new JPanel(new GridLayout(2, 1, 0, 10));
        constraints.gridx = 1;
        constraints.gridy = 0;
        constraints.fill = GridBagConstraints.NONE;
        constraints.weightx = 0;
	constraints.weighty = 0;
        constraints.insets.left = 10;
        constraints.insets.right = 10;
        add(panel, constraints);

        selectButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "select.gif")));
	Insets margin = new Insets(1, 1, 1, 1);
	selectButton.setMargin(margin);
        selectButton.setActionCommand("select");
        selectButton.addActionListener(this);
        panel.add(selectButton);
          
        unselectButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "unselect.gif")));
	unselectButton.setMargin(margin);
        unselectButton.setActionCommand("unselect");
        unselectButton.addActionListener(this);
        panel.add(unselectButton);

	// Selected list ---

        panel = new JPanel(new BorderLayout(0, 2));
        constraints.gridx = 2;
        constraints.gridy = 0;
	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
        constraints.insets.left = 0;
        constraints.insets.right = 0;
        add(panel, constraints);

	JLabel selectedLabel = new JLabel(selectedTitle);
        panel.add(selectedLabel, BorderLayout.NORTH);
	
        selectedList = new JList();
	selectedList.setSelectionMode(
	    ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
	selectedList.setPrototypeCellValue(
	    unselectedList.getPrototypeCellValue());
	selectedList.setVisibleRowCount(Math.max(rows-2, 5));
	selectedList.addListSelectionListener(this);

	scroller = new JScrollPane(selectedList);
        panel.add(scroller, BorderLayout.CENTER);

	JPanel buttons = new JPanel(new FlowLayout(FlowLayout.CENTER, 2, 0));
	panel.add(buttons, BorderLayout.SOUTH);

	firstButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "first.gif")));
	firstButton.setMargin(margin);
	buttons.add(firstButton);
	firstButton.setActionCommand("first");
	firstButton.addActionListener(this);

	previousButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "previous.gif")));
	previousButton.setMargin(margin);
	buttons.add(previousButton);
	previousButton.setActionCommand("previous");
	previousButton.addActionListener(this);

	nextButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "next.gif")));
	nextButton.setMargin(margin);
	buttons.add(nextButton);
	nextButton.setActionCommand("next");
	nextButton.addActionListener(this);

	lastButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "last.gif")));
	lastButton.setMargin(margin);
	buttons.add(lastButton);
	lastButton.setActionCommand("last");
	lastButton.addActionListener(this);
    }

    public void setItems(String[] list, boolean sortList, String[] items) {
	if (sortList && list.length > 1) {
	    String[] sortedList = new String[list.length];

	    System.arraycopy(list, 0, sortedList, 0, list.length);
	    QuickSort.sort(sortedList);

	    list = sortedList;
	}

	Hashtable map = new Hashtable();
	for (int i = 0; i < list.length; ++i)
	    map.put(list[i], new Integer(i));
	itemCompare.setItemOrder(map);
	
	String[] unselectedItems = remove(list, items);
	if (unselectedItems.length > 1)
	    QuickSort.sort(unselectedItems, itemCompare);
	unselectedList.setListData(unselectedItems);
	unselectedList.ensureIndexIsVisible(0);

	selectedList.setListData(items);
	selectedList.ensureIndexIsVisible(0);

	cancelSelectable();
	cancelUnselectable();
    }

    private static String[] remove(String[] strings, String[] removed) {
	String[] newStrings = new String[strings.length];
	int count = 0;

	for (int i = 0; i < strings.length; ++i) {
	    String string = strings[i];
	    boolean found = false;

	    for (int j = 0; j < removed.length; ++j) {
		if (string.equals(removed[j])) {
		    found = true;
		    break;
		}
	    }

	    if (!found)
		newStrings[count++] = string;
	}

	if (count != strings.length) {
	    String[] newStrings2 = new String[count];
	    System.arraycopy(newStrings, 0, newStrings2, 0, count);
	    newStrings = newStrings2;
	}

	return newStrings;
    }

    private static String[] append(String[] strings, String[] added) {
	String[] newStrings = new String[strings.length + added.length];

	System.arraycopy(strings, 0, newStrings, 0, strings.length);
	System.arraycopy(added, 0, newStrings, strings.length, added.length);

	return newStrings;
    }

    private void cancelUnselectable() {
	selectedList.clearSelection();
	itemsUnselectable(null);
    }

    private void cancelSelectable() {
	unselectedList.clearSelection();
	itemsSelectable(null);
    }

    private void itemsSelectable(int[] indices) {
	if (indices == null || indices.length == 0) {
	    selectButton.setEnabled(false);
	} else {
	    selectButton.setEnabled(true);

	    cancelUnselectable();
	}
    }

    private void itemsUnselectable(int[] indices) {
	if (indices == null || indices.length == 0) {
	    unselectButton.setEnabled(false);

	    firstButton.setEnabled(false);
	    previousButton.setEnabled(false);
	    nextButton.setEnabled(false);
	    lastButton.setEnabled(false);
	} else {
	    unselectButton.setEnabled(true);

	    cancelSelectable();

	    if (indices.length == 1) {
		int index = indices[0];
		ListModel model = selectedList.getModel();
		int last = model.getSize() - 1;

		firstButton.setEnabled(index > 0);
		previousButton.setEnabled(index > 0);
		nextButton.setEnabled(index < last);
		lastButton.setEnabled(index < last);
	    } else {
		firstButton.setEnabled(false);
		previousButton.setEnabled(false);
		nextButton.setEnabled(false);
		lastButton.setEnabled(false);
	    }
	}
    }

    public String[] getItems() {
	return getListData(selectedList);
    }

    private static String[] getListData(JList list) {
	ListModel model = list.getModel();
	String[] items = new String[model.getSize()];

	for (int i = 0; i < items.length; ++i)
	    items[i] = (String) model.getElementAt(i);

	return items;
    }

    public void valueChanged(ListSelectionEvent e) {
	// MULTIPLE_INTERVAL_SELECTION allows the user to interactively
	// clear the selection that's why itemsXXXelectable() is called
	// whether there are selected indices or not.

	if (e.getSource() == unselectedList)
	    itemsSelectable(unselectedList.getSelectedIndices());
	else if (e.getSource() == selectedList)
	    itemsUnselectable(selectedList.getSelectedIndices());
    }

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();

	if (command.equals("select")) 
	    select();
	else if (command.equals("unselect")) 
	    unselect();
	else if (command.equals("first")) 
	    move(FIRST);
	else if (command.equals("previous")) 
	    move(PREVIOUS);
	else if (command.equals("next")) 
	    move(NEXT);
	else if (command.equals("last")) 
	    move(LAST);
    }

    private void select() {
	int[] indices = unselectedList.getSelectedIndices();
	if (indices == null || indices.length == 0)
	    return;
	String[] items = getListData(unselectedList, indices);

	String[] list = getListData(unselectedList);
	list = remove(list, items);
	unselectedList.setListData(list);

	list = getListData(selectedList);
	list = append(list, items);
	selectedList.setListData(list);
	selectedList.ensureIndexIsVisible(list.length-1);

	cancelSelectable();
    }

    private static String[] getListData(JList list, int[] indices) {
	ListModel model = list.getModel();
	String[] items = new String[indices.length];

	for (int i = 0; i < indices.length; ++i)
	    items[i] = (String) model.getElementAt(indices[i]);

	return items;
    }

    private void unselect() {
	int[] indices = selectedList.getSelectedIndices();
	if (indices == null || indices.length == 0)
	    return;
	String[] items = getListData(selectedList, indices);

	String[] list = getListData(selectedList);
	list = remove(list, items);
	selectedList.setListData(list);

	list = getListData(unselectedList);
	list = append(list, items);
	QuickSort.sort(list, itemCompare);
	unselectedList.setListData(list);
	unselectedList.ensureIndexIsVisible(
	    StringUtil.indexOf(list, items[items.length-1]));

	cancelUnselectable();
    }

    private void move(int where) {
	int srcIndex  = selectedList.getSelectedIndex();
	if (srcIndex < 0)
	    return;

	String[] items = getItems();
	if (srcIndex >= items.length)
	    return;

	int dstIndex = srcIndex;
	switch (where) {
	case FIRST:
	    dstIndex = 0;
	    break;
	case PREVIOUS:
	    dstIndex = srcIndex - 1;
	    if (dstIndex < 0)
		dstIndex = 0;
	    break;
	case NEXT:
	    dstIndex = srcIndex + 1;
	    if (dstIndex >= items.length)
		dstIndex = items.length - 1;
	    break;
	case LAST:
	    dstIndex = items.length - 1;
	    break;
	}

	if (dstIndex != srcIndex) {
	    move(items, srcIndex, dstIndex);
	    selectedList.setListData(items);
	}
	selectedList.setSelectedIndex(dstIndex);
	selectedList.ensureIndexIsVisible(dstIndex);
    }

    private static void move(String[] strings, int srcIndex, int dstIndex) {
	if (dstIndex < srcIndex) {
	    String src = strings[srcIndex];
	    for (int i = srcIndex-1; i >= dstIndex; --i)
		strings[i+1] = strings[i];
	    strings[dstIndex] = src;
	} else if (dstIndex > srcIndex) {
	    String src = strings[srcIndex];
	    for (int i = srcIndex+1; i <= dstIndex; ++i)
		strings[i-1] = strings[i];
	    strings[dstIndex] = src;
	}
    }
}
