/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.attedit;

import java.util.EventObject;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.event.TableModelListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import fr.pixware.util.StringUtil;
import com.xmlmind.awtutil.ImageResource;
import com.xmlmind.swingutil.ThinBorder;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.ElementAttributeChecker;
import com.xmlmind.xmledit.modeltool.XMLUtil;

public class JAttributesPane extends JPanel 
                             implements TableModelListener, 
                                        ListSelectionListener,
                                        ActionListener,
                                        ItemChecker {
    private AttributesModel attributesModel;
    private Table table;
    private JScrollPane scrollPane;
    private JButton removeButton;
    private JButton addButton;
    private JButton okButton;
    private JButton cancelButton;
    private JButton browseButton;
    private JTextField nameField;
    private JTextField valueField;
    private JLabel typeLabel;
    private JLabel presenceLabel;

    private Attribute editedAttr = new Attribute();

    private static class Table extends JTable {
	public Table(AttributesModel attributesModel) {
	    super(attributesModel);
	}

	public boolean editCellAt(int row, int column, EventObject e) {
	    // The default behavior is to stopCellEditing which is
	    // inappropriate when editing attributes.

	    TableCellEditor editor = getCellEditor();
	    if (editor != null)
		editor.cancelCellEditing();

	    return super.editCellAt(row, column, e);
	}
    }

    private static String attributeLabel = "attribute";
    private static String valueLabel = "value";
    private static String enumerationLabel = "enumeration";
    private static String defaultLabel = "default";
    static {
	try {
	    ResourceBundle resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.attedit.AttributesPaneRes");
	    attributeLabel = resourceBundle.getString(attributeLabel);
	    valueLabel = resourceBundle.getString(valueLabel);
	    enumerationLabel = resourceBundle.getString(enumerationLabel);
	    defaultLabel = resourceBundle.getString(defaultLabel);
	} catch (MissingResourceException ignored) {}
    }

    // -----------------------------------------------------------------------

    public JAttributesPane() {
	setLayout(new BorderLayout(0, 2));

	attributesModel = new AttributesModel();
	attributesModel.addTableModelListener(this);

	table = new Table(attributesModel); 
	table.setRowSelectionAllowed(true);
	table.setColumnSelectionAllowed(false);
	table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	table.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
	table.setShowGrid(true);

	table.getSelectionModel().addListSelectionListener(this);

	TableColumnModel colModel = table.getColumnModel();
	FontMetrics metrics = table.getFontMetrics(table.getFont());
	int colWidth = metrics.stringWidth("123456789012");

	TableColumn col = colModel.getColumn(0);
	col.setHeaderValue(attributeLabel);
	col.setPreferredWidth(colWidth);
	col.setResizable(true);
	AttributeRenderer attrRenderer = new AttributeRenderer();
	col.setCellRenderer(attrRenderer);
	AttributeEditor attrEditor = new AttributeEditor();
	col.setCellEditor(attrEditor);

	col = colModel.getColumn(1);
	col.setHeaderValue(valueLabel);
	col.setPreferredWidth(2*colWidth);
	col.setResizable(true);
	col.setCellRenderer(attrRenderer);
	col.setCellEditor(attrEditor);

	scrollPane = new JScrollPane(table);
	scrollPane.setPreferredSize(
	    new Dimension(3*(colWidth+5), 21*metrics.getHeight()));

	add(scrollPane, BorderLayout.CENTER);

	JPanel form = new JPanel(new GridBagLayout());
	form.setBorder(new CompoundBorder(new EtchedBorder(), 
					  new EmptyBorder(2, 2, 2, 2)));
	add(form, BorderLayout.SOUTH);

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 1;
        constraints.insets.left = 1;
        constraints.insets.right = 1;
        constraints.insets.bottom = 1;
        constraints.gridx = 0;
        constraints.gridy = 0;

	nameField = new JTextField();
	nameField.setEditable(false);
	constraints.fill = GridBagConstraints.HORIZONTAL;
	constraints.weightx = 1;
	form.add(nameField, constraints);
	++constraints.gridx;
	nameField.setActionCommand("ok");
	nameField.addActionListener(this);

	removeButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "remove.gif")));
	Insets margin = new Insets(1, 1, 1, 1);
	removeButton.setMargin(margin);
	removeButton.setEnabled(false);
	constraints.fill = GridBagConstraints.NONE;
	constraints.weightx = 0;
	form.add(removeButton, constraints);
	++constraints.gridx;
	removeButton.setActionCommand("remove");
	removeButton.addActionListener(this);

	addButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "add.gif")));
	addButton.setMargin(margin);
	addButton.setEnabled(false);
	form.add(addButton, constraints);
	++constraints.gridx;
	addButton.setActionCommand("ok");
	addButton.addActionListener(this);

	okButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "ok.gif")));
	okButton.setMargin(margin);
	okButton.setEnabled(false);
	form.add(okButton, constraints);
	++constraints.gridx;
	okButton.setActionCommand("ok");
	okButton.addActionListener(this);

	cancelButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "cancel.gif")));
	cancelButton.setMargin(margin);
	cancelButton.setEnabled(false);
	form.add(cancelButton, constraints);
	++constraints.gridy;
	constraints.gridx = 0;
	cancelButton.setActionCommand("cancel");
	cancelButton.addActionListener(this);

	valueField = new JTextField();
	valueField.setEditable(false);
	constraints.gridwidth = 4;
	constraints.fill = GridBagConstraints.HORIZONTAL;
	constraints.weightx = 1;
	form.add(valueField, constraints);
	constraints.gridx += constraints.gridwidth;
	valueField.setActionCommand("ok");
	valueField.addActionListener(this);

	browseButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "list.gif")));
	browseButton.setMargin(margin);
	browseButton.setEnabled(false);
	constraints.gridwidth = 1;
	constraints.fill = GridBagConstraints.NONE;
	constraints.weightx = 0;
	form.add(browseButton, constraints);
	++constraints.gridy;
	constraints.gridx = 0;
	browseButton.setActionCommand("browse");
	browseButton.addActionListener(this);

	JPanel info = new JPanel(new GridBagLayout());
	constraints.gridwidth = 5;
	constraints.fill = GridBagConstraints.HORIZONTAL;
	constraints.weightx = 1;
	form.add(info, constraints);

	constraints.fill = GridBagConstraints.HORIZONTAL;
        constraints.insets.top = 0;
        constraints.insets.left = 0;
        constraints.insets.right = 0;
        constraints.insets.bottom = 0;
        constraints.gridx = 0;
        constraints.gridy = 0;
	constraints.gridwidth = 1;

	typeLabel = new JLabel("ENTITIES");
	Font font = typeLabel.getFont();
	int fontSize = font.getSize();
	font = new Font(font.getFamily(), font.getStyle(),
			Math.max(8, fontSize-2));
	typeLabel.setFont(font);
	typeLabel.setBorder(
	    new ThinBorder(getBackground(), /*raised*/ false));
	typeLabel.setPreferredSize(typeLabel.getPreferredSize());
	typeLabel.setText("");
	constraints.weightx = 1;
	info.add(typeLabel, constraints);
	++constraints.gridx;

	presenceLabel = new JLabel("#REQUIRED");
	presenceLabel.setFont(font);
	presenceLabel.setBorder(
	    new ThinBorder(getBackground(), /*raised*/ false));
	presenceLabel.setPreferredSize(presenceLabel.getPreferredSize());
	presenceLabel.setText("");
	constraints.weightx = 2;
        constraints.insets.left = 2;
	info.add(presenceLabel, constraints);
    }

    public void setDocument(DocumentModel doc) {
	attributesModel.setDocument(doc);
    }

    public DocumentModel getDocument() {
	return attributesModel.getDocument();
    }

    public void setElement(ElementModel element) {
	attributesModel.setElement(element);
    }

    public ElementModel getElement() {
	return attributesModel.getElement();
    }

    public void setEditable(boolean editable) {
	attributesModel.setEditable(editable);	    
    }

    public boolean isEditable() {
	return attributesModel.isEditable();
    }

    public void tableChanged(TableModelEvent e) {
	clearForm();
    }

    public void valueChanged(ListSelectionEvent e) {
	int row = table.getSelectedRow();
	int col = table.getSelectedColumn();
	if (row < 0 || col < 0)
	    clearForm();
	else
	    fillForm(row, col);
    }

    private void clearForm() {
	nameField.setText("");
	removeButton.setEnabled(false);
	okButton.setEnabled(false);
	valueField.setText("");
	typeLabel.setText("");
	presenceLabel.setText("");

	if (attributesModel.getDocumentType() == null &&
	    attributesModel.getElement() != null) {
	    boolean editable = attributesModel.isEditable();

	    nameField.setEditable(editable);
	    addButton.setEnabled(editable);
	    cancelButton.setEnabled(editable);
	    valueField.setEditable(editable);
	    browseButton.setEnabled(editable);
	} else {
	    nameField.setEditable(false);
	    addButton.setEnabled(false);
	    cancelButton.setEnabled(false);
	    valueField.setEditable(false);
	    browseButton.setEnabled(false);
	}

	TableCellEditor editor = table.getCellEditor();
	if (editor != null)
	    editor.cancelCellEditing();

	table.clearSelection();

	table.requestFocus();
    }

    private void fillForm(int row, int col) {
	boolean editable = attributesModel.isEditable();

	Attribute attr = (Attribute) attributesModel.getValueAt(row, col);
	editedAttr.name = attr.name;
	editedAttr.value = attr.value;
	editedAttr.definition = attr.definition;

	nameField.setText(attr.name);
	nameField.setEditable(false);
	addButton.setEnabled(false);

	okButton.setEnabled(editable);
	cancelButton.setEnabled(editable);

	valueField.setEditable(editable);
	if (attr.value == null) {
	    removeButton.setEnabled(false);
	    valueField.setText("");
	} else {
	    removeButton.setEnabled(editable);
	    valueField.setText(attr.value);
	}

	boolean browse = false;

	if (attr.definition != null) {
	    switch (attr.definition.getAttributeType()) {
	    case AttributeDefinitionModel.CDATA:
		browse = false;
		typeLabel.setText("CDATA");
		break;
	    case AttributeDefinitionModel.ID:
		// User will want to see already defined IDs.
		browse = editable;
		typeLabel.setText("ID");
		break;
	    case AttributeDefinitionModel.IDREF:
		browse = editable;
		typeLabel.setText("IDREF");
		break;
	    case AttributeDefinitionModel.IDREFS:
		browse = editable;
		typeLabel.setText("IDREFS");
		break;
	    case AttributeDefinitionModel.ENTITY:
		browse = editable;
		typeLabel.setText("ENTITY");
		break;
	    case AttributeDefinitionModel.ENTITIES:
		browse = editable;
		typeLabel.setText("ENTITIES");
		break;
	    case AttributeDefinitionModel.NMTOKEN:
		browse = false;
		typeLabel.setText("NMTOKEN");
		break;
	    case AttributeDefinitionModel.NMTOKENS:
		browse = editable;
		typeLabel.setText("NMTOKENS");
		break;
	    case AttributeDefinitionModel.NOTATION:
		browse = editable;
		typeLabel.setText("NOTATION " + enumerationLabel);
		break;
	    case AttributeDefinitionModel.ENUMERATION:
		browse = editable;
		typeLabel.setText(enumerationLabel);
		break;
	    default:
		browse = false;
	    }

	    switch (attr.definition.getAttributePresence()) {
	    case AttributeDefinitionModel.DEFAULTED:
		presenceLabel.setText(
		    defaultLabel + " \"" + 
		    attr.definition.getAttributeDefaultValue() + '\"');
		break;
	    case AttributeDefinitionModel.REQUIRED:
		// Cannot remove a required attribute.
		removeButton.setEnabled(false);
		presenceLabel.setText("#REQUIRED");
		break;
	    case AttributeDefinitionModel.IMPLIED:
		presenceLabel.setText("#IMPLIED");
		break;
	    case AttributeDefinitionModel.FIXED:
		presenceLabel.setText(
		    "#FIXED \"" + attr.definition.getAttributeDefaultValue() +
		    '\"');
		break;
	    }
	}

	browseButton.setEnabled(browse);
    }

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();

	if (command.equals("cancel")) 
	    cancel();
	else if (command.equals("ok")) 
	    ok();
	if (command.equals("remove")) 
	    remove();
	else if (command.equals("browse")) 
	    browse();
    }

    private void remove() {
	editedAttr.value = null;
	attributesModel.apply(editedAttr);
	clearForm();
    }

    private void ok() {
	editedAttr.name = nameField.getText();
	editedAttr.value = valueField.getText();

	ElementAttributeChecker.Diagnostic d =
	    attributesModel.normalizeAndCheck(editedAttr);

	if (d == null) {
	    attributesModel.apply(editedAttr);
	    clearForm();
	} else {
	    getToolkit().beep();

	    JTextField field;
	    if (d.invalidReason == 
		ElementAttributeChecker.INVALID_ATTRIBUTE_NAME) 
		field = nameField;
	    else
		field = valueField;

	    Dialogs.showError(field, d.message);

	    field.selectAll();
	    field.requestFocus();
	}
    }

    private void cancel() {
	clearForm();
    }

    private void browse() {
	String newValue = null;
	String currentValue = valueField.getText();
	if (currentValue.length() > 0)
	    currentValue = XMLUtil.normalizeAttributeValue(currentValue);

	switch (editedAttr.definition.getAttributeType()) {
	case AttributeDefinitionModel.ID:
	    // User will want to see already defined IDs.
	case AttributeDefinitionModel.IDREF:
	    newValue = Dialogs.selectItem(
		this, attributesModel.getElementIds(), currentValue);
	    break;
	case AttributeDefinitionModel.IDREFS:
	    {
		String[] items;
		if (currentValue.length() > 0)
		    items = StringUtil.split(currentValue, ' ');
		else
		    items = StringUtil.EMPTY_LIST;
		items = Dialogs.selectItems(
		    this, attributesModel.getElementIds(), items);
		if (items != null)
		    newValue = StringUtil.join(items, ' ');
	    }
	    break;
	case AttributeDefinitionModel.ENTITY:
	    newValue = Dialogs.selectItem(
		this, attributesModel.getUnparsedEntities(), currentValue);
	    break;
	case AttributeDefinitionModel.ENTITIES:
	    {
		String[] items;
		if (currentValue.length() > 0)
		    items = StringUtil.split(currentValue, ' ');
		else
		    items = StringUtil.EMPTY_LIST;
		items = Dialogs.selectItems(
		    this, attributesModel.getUnparsedEntities(), items);
		if (items != null)
		    newValue = StringUtil.join(items, ' ');
	    }
	    break;
	case AttributeDefinitionModel.NMTOKENS:
	    {
		String[] items;
		if (currentValue.length() > 0)
		    items = StringUtil.split(currentValue, ' ');
		else
		    items = StringUtil.EMPTY_LIST;
		items = Dialogs.editList(this, items, this);
		if (items != null)
		    newValue = StringUtil.join(items, ' ');
	    }
	    break;
	case AttributeDefinitionModel.NOTATION:
	case AttributeDefinitionModel.ENUMERATION:
	    newValue = Dialogs.selectItem(
		this, editedAttr.definition.getAttributeEnumeratedValues(), 
		currentValue);
	    break;
	}

	if (newValue != null)
	    valueField.setText(newValue);
    }

    public boolean checkItem(String item) {
	return XMLUtil.isNmtoken(item);
    }
}

