/*
 * Copyright (c) 1997-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.awtutil;

import java.awt.*;

public class MultiLineLabel extends Component {
    public static final int CENTER = 0;
    public static final int LEFT   = 1;
    public static final int RIGHT  = 2;

    public static final int FLAT   = 0;
    public static final int SUNKEN = 1;
    public static final int RAISED = 2;

    private String[] lines = new String[0];
    private int alignment = LEFT;
    private int relief = FLAT;
    private int marginWidth = 2;
    private int marginHeight = 2;

    public MultiLineLabel() {
	this("");
    }

    public MultiLineLabel(String text) {
	this(text, LEFT);
    }

    public MultiLineLabel(String text, int align) {
	this(text, align, FLAT);
    }

    public MultiLineLabel(String text, int align, int relief) {
	setText(text);
	setAlignment(align);
	setRelief(relief);
    }

    // Let this component behave like native components
    // when these presentation attributes are changed.
    public void setForeground(Color color) {
        super.setForeground(color);
        repaint();
    }

    public void setBackground(Color color) {
        super.setBackground(color);
        repaint();
    }

    public void setFont(Font font) {
        super.setFont(font);
        invalidate();
        repaint();
    }

    public void setText(String text) {
	int length = text.length();
	int lineCount = 0;
	int prevPos = 0;
	int pos;
	while (prevPos < length && 
	       (pos = text.indexOf('\n', prevPos)) >= 0) {
	    prevPos = pos + 1;
	    ++lineCount;
	}
	if (prevPos != length)
	    ++lineCount;

	lines = new String[lineCount];
	if (lineCount > 0) {
	    lineCount = 0;
	    prevPos   = 0;
	    while (prevPos < length && 
		   (pos = text.indexOf('\n', prevPos)) >= 0) {
		lines[lineCount++] = text.substring(prevPos, pos);
		prevPos = pos + 1;
	    }
	    if (prevPos != length) 
		lines[lineCount] = text.substring(prevPos);
	}

	invalidate();
	repaint();
    }

    public String getText() {
	StringBuffer buffer = new StringBuffer();
	for (int i = 0; i < lines.length; ++i) {
	    buffer.append(lines[i]);
	    buffer.append('\n');
	}
    	return buffer.toString();
    }

    public void setAlignment(int align) {
	if (align < CENTER || align > RIGHT)
	    throw new IllegalArgumentException("invalid alignment");
	alignment = align;
	repaint();
    }

    public int getAlignment() {
    	return alignment;
    }

    public void setRelief(int relief) {
	if (relief < FLAT || relief > RAISED)
	    throw new IllegalArgumentException("invalid relief");
	this.relief = relief;
	invalidate();
	repaint();
    }

    public int getRelief() {
    	return relief;
    }

    public void setMarginWidth(int marginWidth) {
	if (marginWidth < 0)
	    marginWidth = 0;
	this.marginWidth = marginWidth;
	invalidate();
	repaint();
    }

    public int getMarginWidth() {
    	return marginWidth;
    }

    public void setMarginHeight(int marginHeight) {
	if (marginHeight < 0)
	    marginHeight = 0;
	this.marginHeight = marginHeight;
	invalidate();
	repaint();
    }

    public int getMarginHeight() {
    	return marginHeight;
    }

    public Dimension getMinimumSize() {
	return getPreferredSize();
    }
    
    public Dimension getPreferredSize() {
	FontMetrics fontMetrics = getFontMetrics(getFont());
	int lineHeight = fontMetrics.getHeight();
	int height = lineHeight * lines.length;
	int width = 0;

	for (int i = 0; i < lines.length; ++i) {
	    if (lines[i].length() > 0) {
		int lineWidth = fontMetrics.stringWidth(lines[i]);
		if (lineWidth > width)
		    width = lineWidth;
	    }
	}

	if (width == 0)
	    width = 1;
	if (height == 0)
	    height = lineHeight;

	int border = (relief == FLAT)? 0 : 1;
	width += 2*(border + marginWidth);
	height += 2*(border + marginHeight);

	return new Dimension(width, height);
    }

    public void paint(Graphics g) {
	Dimension dim = getSize();
	Dimension bestDim = getPreferredSize();
	int x, y, width, height;

	// Use largest dimensions ---

	if (dim.width >= bestDim.width) {
	    x = 0;
	    width = dim.width;
	} else {
	    x = (dim.width - bestDim.width) / 2;
	    width = bestDim.width;
	}
	if (dim.height >= bestDim.height) {
	    y = 0;
	    height = dim.height;
	} else {
	    y = (dim.height - bestDim.height) / 2;
	    height = bestDim.height;
	}

	// Draw border if any ---

	if (relief != FLAT) {
	    g.setColor(getBackground());
	    g.draw3DRect(x, y, width - 1, height - 1, (relief == RAISED));
	}

	// Adjust dimensions to the inside of the label ---

	int border = (relief == FLAT)? 0 : 1;
	x += border + marginWidth;
	y += border + marginHeight;
	width -= 2*(border + marginWidth);
	height -= 2*(border + marginHeight);

	// Draw lines ---

	FontMetrics fontMetrics = getFontMetrics(getFont());
	int maxAscent = fontMetrics.getMaxAscent();
	int lineHeight = fontMetrics.getHeight();

	g.setColor(getForeground());
	switch (alignment) {
	case CENTER:
	    for (int i = 0; i < lines.length; ++i) {
		if (lines[i].length() > 0) {
		    int lineWidth = fontMetrics.stringWidth(lines[i]);
		    g.drawString(lines[i], 
				 x + (width - lineWidth)/2, y + maxAscent);
		}
		y += lineHeight;
	    }
	    break;
	case RIGHT:
	    for (int i = 0; i < lines.length; ++i) {
		if (lines[i].length() > 0) {
		    int lineWidth = fontMetrics.stringWidth(lines[i]);
		    g.drawString(lines[i], 
				 x + width - lineWidth, y + maxAscent);
		}
		y += lineHeight;
	    }
	    break;
	default:
	    for (int i = 0; i < lines.length; ++i) {
		if (lines[i].length() > 0) 
		    g.drawString(lines[i], x, y + maxAscent);
		y += lineHeight;
	    }
	}
    }
}
