/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.cssparser;

import java.net.URL;
import java.io.*;
import java.util.Hashtable;
import java.util.StringTokenizer;
import fr.pixware.util.FileUtil;
import fr.pixware.util.HashtableUtil;

public class ParseUtil {
    public static String loadSource(URL url) throws IOException {
	InputStream in = url.openStream();
	String source = null;

	try {
	    source = loadSource(in);
	} finally {
	    in.close();
	}

	return source;
    }

    public static String loadSource(InputStream stream) throws IOException {
	BufferedInputStream in = new BufferedInputStream(stream);
	ByteArrayOutputStream out = new ByteArrayOutputStream();
	byte[] bytes = new byte[8192];
	int count;

	while ((count = in.read(bytes)) >= 0)
	    out.write(bytes, 0, count);
	
	bytes = out.toByteArray();
	return loadSource(bytes, 0, bytes.length);
    }

    // Java 1.1: we'll rely on the very incomplete mimeToJavaMap table. 
    // Java 2: the situation should have improved:
    //
    // Excerpt of the JDK1.3 documentation 
    // -----------------------------------
    //
    // The canonical name of an encoding is the name returned by the
    // getEncoding() methods of the InputStreamReader and OutputStreamWriter
    // classes.
    //
    // Encoding names generally follow the conventions documented in RFC2278:
    // IANA Charset Registration Procedures.
    //
    // An implementation of the Java platform must support the MIME-preferred
    // registry name for a supported encoding if there is one; for convenience
    // it may additionally support other registry names.
    //
    // The IANA MIME-preferred name of an encoding, if there is one, is often,
    // but not necessarily, its canonical name.
    //
    // Following IANA convention, the mapping from IANA registry names to
    // encodings is not case-sensitive.

    private static final String[] mimeToJavaList = {
	"iso-8859-1", 	"ISO8859_1",
	"iso-8859-2", 	"ISO8859_2",
	"iso-8859-3", 	"ISO8859_3",
	"iso-8859-4", 	"ISO8859_4",
	"iso-8859-5", 	"ISO8859_5",
	"iso-8859-6", 	"ISO8859_6",
	"iso-8859-7", 	"ISO8859_7",
	"iso-8859-8", 	"ISO8859_8",
	"iso-8859-9", 	"ISO8859_9",
	"iso-8859-13", 	"ISO8859_13",
	"iso-8859-15", 	"ISO8859_15_FDIS",
	"windows-1250", "Cp1250", 	// Windows Central European
	"windows-1251", "Cp1251", 	// Windows Cyrillic
	"windows-1252", "Cp1252", 	// Windows Latin-1 
	"windows-1253", "Cp1253", 	// Windows Greek
	"shift_jis", 	"SJIS",
	"utf-8", 	"UTF8",
	"us-ascii", 	"ASCII"
    };
    private static Hashtable mimeToJavaMap = new Hashtable(); 
    static {
	HashtableUtil.putAll(mimeToJavaMap, mimeToJavaList);
    }

    public static String loadSource(byte[] bytes, int offset, int count) 
	throws UnsupportedEncodingException {
	int guess = guessEncoding(bytes, offset, count);
	//System.err.println("encoding guess=" + guess);

	String encoding = null;
	switch (guess) {
	case UTF16_BE_BOM:
	    // Skip Byte Order Mark.
	    offset += 2;
	    count -= 2;
	case UTF16_BE:
	    encoding = "UnicodeBigUnmarked";
	    break;

	case UTF16_LE_BOM:
	    offset += 2;
	    count -= 2;
	case UTF16_LE:
	    encoding = "UnicodeLittleUnmarked";
	    break;

	case UTF8_BOM:
	    offset += 3;
	    count -= 3;
	    encoding = "UTF8";
	    break;

	case UCS4_BE_BOM:
	case UCS4_BE:
	case UCS4_LE_BOM:
	case UCS4_LE:
	case UCS4_2143_BOM:
	case UCS4_2143:
	case UCS4_3412_BOM:
	case UCS4_3412:
	case EBCDIC:
	    throw new UnsupportedEncodingException();
	}

	if (encoding == null && count > 8 && 
	    bytes[offset] == '@' && 
	    (bytes[offset+1] == 'c' || bytes[offset+1] == 'C') && 
	    (bytes[offset+2] == 'h' || bytes[offset+2] == 'H') && 
	    (bytes[offset+3] == 'a' || bytes[offset+3] == 'A') &&
	    (bytes[offset+4] == 'r' || bytes[offset+4] == 'R') &&
	    (bytes[offset+5] == 's' || bytes[offset+5] == 'S') &&
	    (bytes[offset+6] == 'e' || bytes[offset+6] == 'E') &&
	    (bytes[offset+7] == 't' || bytes[offset+7] == 'T')) {
	    // ASCII_COMPATIBLE, cannot guess or missing '@charset': 
	    // assume it is ASCII_COMPATIBLE.

	    byte quote = 0;
	    int first = 8;
	    loop: while (first < count) {
		switch (bytes[offset+first]) {
		case '\"': case '\'':
		    quote = bytes[offset+first];
		    ++first;
		    break loop;
		default:
		    ++first;
		}
	    }

	    int end = first;
	    while (end < count) {
		if (bytes[offset+end] == quote)
		    break;
		else
		    ++end;
	    }

	    if (end > first && end < count) {
		encoding = new String(bytes, offset+first, end-first, "ASCII");
		//System.err.println("charset=" + encoding);

		String javaVersion = System.getProperty("java.version");
		if (javaVersion != null && javaVersion.startsWith("1.1")) {
		    String javaEncoding = 
			(String) mimeToJavaMap.get(encoding.toLowerCase());
		    if (javaEncoding != null)
			encoding = javaEncoding;
		}
	    }
	}
	//System.err.println("encoding=" + encoding);

	if (encoding == null) 
	    encoding = FileUtil.defaultEncoding();
	return new String(bytes, offset, count, encoding);
    }

    // -----------------------------------------------------------------------

    public static final int UCS4_BE_BOM      = 0;
    public static final int UCS4_LE_BOM      = 1;
    public static final int UCS4_2143_BOM    = 2;
    public static final int UCS4_3412_BOM    = 3;
    public static final int UTF16_BE_BOM     = 4;
    public static final int UTF16_LE_BOM     = 5;
    public static final int UTF8_BOM         = 6;
    public static final int UCS4_BE          = 7;
    public static final int UCS4_LE          = 8;
    public static final int UCS4_2143        = 9;
    public static final int UCS4_3412        = 10;
    public static final int UTF16_BE         = 11;
    public static final int UTF16_LE         = 12;
    public static final int ASCII_COMPATIBLE = 13;
    public static final int EBCDIC           = 14;

    public static int guessEncoding(byte[] bytes, int offset, int count) {
	if (count < 4)
	    return -1;

	// '@'=0x40, 'c'=0x63|'C'=0x43, 'h'=0x68|'H'=0x48, 'a'=0x61|'A'=0x41

	int b0 = (bytes[offset] & 0xFF);
	int b1 = (bytes[offset+1] & 0xFF);
	int b2 = (bytes[offset+2] & 0xFF);
	int b3 = (bytes[offset+3] & 0xFF);

	if (b0 == 0x00 && b1 == 0x00 && b2 == 0xFE && b3 == 0xFF)
	    return UCS4_BE_BOM;
	else if (b0 == 0xFF && b1 == 0xFE && b2 == 0x00 && b3 == 0x00)
	    return UCS4_LE_BOM;
	else if (b0 == 0x00 && b1 == 0x00 && b2 == 0xFF && b3 == 0xFE)
	    return UCS4_2143_BOM;
	else if (b0 == 0xFE && b1 == 0xFF && b2 == 0x00 && b3 == 0x00)
	    return UCS4_3412_BOM;
	else if (b0 == 0xFE && b1 == 0xFF && b2 != 0x00 && b3 != 0x00)
	    return UTF16_BE_BOM;
	else if (b0 == 0xFF && b1 == 0xFE && b2 != 0x00 && b3 != 0x00)
	    return UTF16_LE_BOM;
	else if (b0 == 0xEF && b1 == 0xBB && b2 == 0xBF)
	    return UTF8_BOM;
	else if (b0 == 0x00 && b1 == 0x00 && b2 == 0x00 && b3 == 0x40)
	    return UCS4_BE;
	else if (b0 == 0x40 && b1 == 0x00 && b2 == 0x00 && b3 == 0x00)
	    return UCS4_LE;
	else if (b0 == 0x00 && b1 == 0x00 && b2 == 0x40 && b3 == 0x00)
	    return UCS4_2143;
	else if (b0 == 0x00 && b1 == 0x40 && b2 == 0x00 && b3 == 0x00)
	    return UCS4_3412;
	else if (b0 == 0x00 && b1 == 0x40 && 
		 b2 == 0x00 && (b3 == 0x63 || b3 == 0x43))
	    return UTF16_BE;
	else if (b0 == 0x40 && b1 == 0x00 && 
		 (b2 == 0x63 || b2 == 0x43) && b3 == 0x00)
	    return UTF16_LE;
	else if (b0 == 0x40 && 
		 (b1 == 0x63 || b1 == 0x43) && 
		 (b2 == 0x68 || b2 == 0x48) && 
		 (b3 == 0x61 || b3 == 0x41))
	    return ASCII_COMPATIBLE;
	else
	    return -1;
    }

    // -----------------------------------------------------------------------

    public static String lexemeToCaseSensitiveIdent(String lexeme) {
	return unescape(lexeme);
    }

    public static String lexemeToIdent(String lexeme) {
	return unescape(lexeme).toLowerCase();
    }

    public static String lexemeToString(String lexeme) {
	// Unescape also unescapes quoted quotes.
	String s = unescape(lexeme.substring(1, lexeme.length() - 1));
	if (s.indexOf('\\') < 0)
	    return s;

	// Trim quoted newlines ---

	StringBuffer buffer = new StringBuffer();
	int count = s.length();

	for (int i = 0; i < count; ++i) {
	    char c = s.charAt(i);

	    if (c == '\\' && i+1 < count) {
		switch (s.charAt(i+1)) {
		case '\r': 
		    if (i+2 < count && s.charAt(i+2) == '\n')
			i += 2;
		    else
			++i;
		    break;
		case '\n': case '\f':
		    ++i;
		    break;
		default:
		    buffer.append(c);
		}
	    } else {
		buffer.append(c);
	    }
	}

	return buffer.toString();
    }

    public static String lexemeToURI(String lexeme) {
	String uri = trim(lexeme.substring(4, lexeme.length() - 1));
	if (uri.length() > 0) {
	    switch (uri.charAt(0)) {
	    case '\"':
	    case '\'':
		uri = uri.substring(1, uri.length() - 1);
		break;
	    }
	}

	return unescape(uri);
    }

    public static String lexemeToUnit(String lexeme) {
	int length = lexeme.length();
	int end;
	loop: for (end = 0; end < length; ++end) {
	    switch (lexeme.charAt(end)) {
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	    case '.':
		break;
	    default:
		break loop;
	    }
	}

	if (end >= length)
	    return null;
	else
	    return lexeme.substring(end).toLowerCase();
    }

    public static double lexemeToNumber(String lexeme) {
	String unit = lexemeToUnit(lexeme);
	if (unit != null) 
	    lexeme = lexeme.substring(0, lexeme.length() - unit.length());

	double number;

	try {
	    number = Double.valueOf(lexeme).doubleValue();
	} catch (NumberFormatException e) {
	    number = Double.NaN;
	}

	return number;
    }

    public static int[] lexemeToUnicodeRange(String lexeme) {
	// Trim "U+".
	lexeme = lexeme.substring(2);

	String min, max;
	int dash = lexeme.indexOf('-');
	if (dash < 0) {
	    min = lexeme.replace('?', '0');
	    max = lexeme.replace('?', 'F');
	} else {
	    min = lexeme.substring(0, dash);
	    max = lexeme.substring(dash+1);
	}

	int[] range = new int[2];
	try {
	    range[0] = Integer.parseInt(min, 16);
	    range[1] = Integer.parseInt(max, 16);
	} catch (NumberFormatException e) {
	    return null;
	}
	if (range[0] > range[1])
	    // Equality is OK.
	    return null;

	return range;
    }

    public static int[] lexemeToHexColor(String lexeme) {
	// Trim '#'.
	String unescaped = unescape(lexeme.substring(1));

	int[] rgb = new int[3];
	switch (unescaped.length()) {
	case 3:
	    {
		String digit;
		try {
		    digit = unescaped.substring(0, 1);
		    rgb[0] = Integer.parseInt(digit+digit, 16);
		    digit = unescaped.substring(1, 2);
		    rgb[1] = Integer.parseInt(digit+digit, 16);
		    digit = unescaped.substring(2, 3);
		    rgb[2] = Integer.parseInt(digit+digit, 16);
		} catch (NumberFormatException e) {
		    return null;
		}
	    }
	    break;

	case 6:
	    {
		try {
		    rgb[0] = 
			Integer.parseInt(unescaped.substring(0, 2), 16);
		    rgb[1] = 
			Integer.parseInt(unescaped.substring(2, 4), 16);
		    rgb[2] = 
			Integer.parseInt(unescaped.substring(4, 6), 16);
		} catch (NumberFormatException e) {
		    return null;
		}
	    }
	    break;

	default:
	    return null;
	}

	return rgb;
    }

    public static String lexemeToFunction(String lexeme) {
	// Trim ')'.
	String unescaped = unescape(lexeme.substring(0, lexeme.length()-1));

	return unescaped.toLowerCase();
    }

    // -----------------------------------------------------------------------

    private static class NamedColor {
	public String name;
	public int[] rgb = new int[3];

	public NamedColor(String name, int r, int g, int b) {
	    this.name = name;
	    rgb[0] = r;
	    rgb[1] = g;
	    rgb[2] = b;
	}
    }

    private static final NamedColor black = 
    	new NamedColor("black", 0, 0, 0);
    private static final NamedColor silver = 
    	new NamedColor("silver", 192, 192, 192);
    private static final NamedColor gray = 
    	new NamedColor("gray", 128, 128, 128);
    private static final NamedColor white = 
    	new NamedColor("white", 255, 255, 255);
    private static final NamedColor maroon = 
    	new NamedColor("maroon", 128, 0, 0);
    private static final NamedColor red = 
    	new NamedColor("red", 255, 0, 0);
    private static final NamedColor purple = 
    	new NamedColor("purple", 128, 0, 128);
    private static final NamedColor fuchsia = 
    	new NamedColor("fuchsia", 255, 0, 255);
    private static final NamedColor green = 
    	new NamedColor("green", 0, 128, 0);
    private static final NamedColor lime = 
    	new NamedColor("lime", 0, 255, 0);
    private static final NamedColor olive = 
    	new NamedColor("olive", 128, 128, 0);
    private static final NamedColor yellow = 
    	new NamedColor("yellow", 255, 255, 0);
    private static final NamedColor navy = 
    	new NamedColor("navy", 0, 0, 128);
    private static final NamedColor blue = 
    	new NamedColor("blue", 0, 0, 255);
    private static final NamedColor teal = 
    	new NamedColor("teal", 0, 128, 128);
    private static final NamedColor aqua = 
    	new NamedColor("aqua", 0, 255, 255);

    private static final NamedColor[] namedColors = {
	black, silver, gray, white, 
	maroon, red, purple, fuchsia, 
	green, lime, olive, yellow,
	navy, blue, teal, aqua
    };

    private static Hashtable namedColorMap = new Hashtable();
    static {
	for (int i = 0; i < namedColors.length; ++i) 
	    namedColorMap.put(namedColors[i].name, namedColors[i]);
    }

    public static final int[] nameToRGB(String name) {
	NamedColor color = (NamedColor) namedColorMap.get(name.toLowerCase());
	if (color == null)
	    return null;
	else
	    return color.rgb;
    }

    public static final int[] functionArgumentsToRGB(ValueComponent[] args) {
	if (args.length != 3)
	    return null;

	int[] rgb = new int[3];
	for (int i = 0; i < 3; ++i) {
	    ValueComponent arg = args[i];

	    if ((i == 0 && arg.operator != ValueComponent.NONE) ||
		(i > 0 && arg.operator != ','))
		return null;

	    int value;
	    switch (arg.type) {
	    case ValueComponent.NUMBER:
		value = (int) arg.numericValue;
		break;
	    case ValueComponent.PERCENTAGE:
		value = (int) ((arg.numericValue * 255) / 100);
		break;
	    default:
		return null;
	    }

	    if (value < 0)
		value = 0;
	    else if (value > 255)
		value = 255;
	    rgb[i] = value;
	}

	return rgb;
    }

    // -----------------------------------------------------------------------

    public static String trim(String s) {
	int count = s.length();
	int first, last;

	for (first = 0; first < count; ++first) {
	    if (!isSpace(s.charAt(first)))
		break;
	}

	for (last = count-1; last >= 0; --last) {
	    if (!isSpace(s.charAt(last)))
		break;
	}

	return ((last >= first)? s.substring(first, last+1) : "");
    }

    public static String unescape(String s) {
	StringBuffer buffer = new StringBuffer();
	int count = s.length();

	for (int i = 0; i < count; ++i) {
	    char c = s.charAt(i);

	    if (c == '\\' && i+1 < count) {
		char escaped = s.charAt(i+1);

		if (isHexDigit(escaped)) {
		    ++i;
		    int begin = i;

		    while (i < count && isHexDigit(s.charAt(i))) {
			++i;
			if (i - begin == 6)
			    break;
		    }

		    try {
			int code = Integer.parseInt(s.substring(begin, i), 16);
			if (code > Character.MAX_VALUE)
			    c = '?';
			else
			    c = (char) code;
		    } catch (NumberFormatException e) {
			c = '?';
		    }

		    if (i < count && isSpace(s.charAt(i))) 
			++i;

		    --i;
		} else if ((escaped >= ' ' && escaped <= '~') ||
		           escaped >= '\200') {
		    ++i;
		    c = escaped;
		}
	    }

	    buffer.append(c);
	}

	return buffer.toString();
    }

    public static boolean isHexDigit(char c) {
	switch (c) {
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
	    return true;
	default:
	    return false;
	}
    }

    public static boolean isSpace(char c) {
	switch (c) {
	case ' ': case '\t': case '\r': case '\n': case '\f':
	    return true;
	default:
	    return false;
	}
    }

    // -----------------------------------------------------------------------

    public static String unescapeURI(String ascii) {
	int length = ascii.length();
	byte[] bytes = new byte[length];
	int byteCount = 0;
	
	for (int i = 0; i < length; ++i) {
	    char c = ascii.charAt(i);

	    if (c == '%' && 
		i+1 < length && isHexDigit(ascii.charAt(i+1)) &&
		i+2 < length && isHexDigit(ascii.charAt(i+2))) {
		try {
		    bytes[byteCount++] = (byte)
			Integer.parseInt(ascii.substring(i+1, i+3), 16);
		    i += 2;
		} catch (NumberFormatException cannotHappen1) {
		    bytes[byteCount++] = (byte) c;
		}
	    } else {
		bytes[byteCount++] = (byte) c;
	    }
	}

	String unescaped;
	try {
	    unescaped = new String(bytes, 0, byteCount, "UTF8");
	} catch (UnsupportedEncodingException cannotHappen2) {
	    unescaped = ascii;
	}
	return unescaped;
    }

    public static String escapeURI(String string) {
	StringBuffer escaped = new StringBuffer();

	int i = string.indexOf(':');
	if (i > 0) {
	    // The scheme is not escaped.
	    escaped.append(string.substring(0, i));
	    string = string.substring(i);
	}

	String reservedChars = ";/?:@&=+$,";

	StringTokenizer tokens = 
	    new StringTokenizer(string, reservedChars, true);
	while (tokens.hasMoreTokens()) {
	    String token = tokens.nextToken();
	    
	    if (reservedChars.indexOf(token.charAt(0)) >= 0) {
		// reserved char used as a separator, do not escape.
		escaped.append(token);
	    } else {
		try {
		    escapeURIComponent(token, escaped);
		} catch (UnsupportedEncodingException cannotHappen) {}
	    }
	}

	return escaped.toString();
    }

    private static void escapeURIComponent(String string, 
					   StringBuffer escaped) 
	throws UnsupportedEncodingException {

	char[] unicode = new char[1];

	int length = string.length();
	for (int i = 0; i < length; ++i) {
	    char c = string.charAt(i);
	    switch (c) {
	    case '-':
	    case '_':
	    case '.':
	    case '!':
	    case '~':
	    case '*':
	    case '\'':
	    case '(':
	    case ')':
		escaped.append(c);
		break;
	    default:
		if ((c >= 'a' && c <= 'z') || 
		    (c >= 'A' && c <= 'Z') || 
		    (c >= '0' && c <= '9')) {
		    escaped.append(c);
		} else {
		    unicode[0] = c;
		    byte[] bytes = 
			(new String(unicode, 0, 1)).getBytes("UTF8");

		    for (int j = 0; j < bytes.length; ++j) {
			String hex = Integer.toHexString(bytes[j] & 0xFF);

			escaped.append('%');
			if (hex.length() == 1)
			    escaped.append('0');
			escaped.append(hex);
		    }
		}
	    }
	}
    }

    // -----------------------------------------------------------------------

    public static void main(String[] args) throws Exception {
	if (args.length <= 1) {
	    System.err.println(
		"usage: java com.xmlmind.cssparser.ParseUtil" + 
		" uri|string s1 ... sN");
	    System.exit(1);
	}

	char mode = args[0].charAt(0);
	for (int i = 1; i < args.length; ++i) {
	    String arg = args[i];
	    String result;

	    switch (mode) {
	    case 'u':
		result = unescapeURI(arg);
		break;
	    default:
		result = unescape(arg);
	    }

	    System.out.println("'" + arg + "' = '" + result + "'");
	}
    }
}
