/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.cssparser;

import java.net.URL;
import java.net.MalformedURLException;
import java.io.File;
import java.io.IOException;
import java.util.Stack;
import java.util.Vector;
import fr.pixware.util.FileUtil;

public class Parser {
    private Handler handler;

    private Stack sources = new Stack();
    private Scanner scanner;
    private URL url;
    private String[] media;
    private int token;

    private static class Source {
	public Scanner scanner;
	public URL url;
	public String[] media;
	public int token;

	public Source(Scanner scanner, URL url, String[] media) {
	    this.scanner = scanner;
	    this.url = url;
	    this.media = media;
	    this.token = 0;
	}
    }

    // -----------------------------------------------------------------------

    public void setHandler(Handler handler) {
	this.handler = handler;
    }

    public Handler getHandler() {
	return handler;
    }

    public void parse(File file) 
	throws MalformedURLException, IOException {
	parse(file, null);
    }

    public void parse(File file, String[] media) 
	throws MalformedURLException, IOException {
	parse(FileUtil.fileToURL(file), media);
    }

    public void parse(URL url) throws IOException {
	parse(url, null);
    }

    public void parse(URL url, String[] media) throws IOException {
	parse(ParseUtil.loadSource(url), url, media);
    }

    public void parse(String source) {
	parse(source, null, null);
    }

    public void parse(String source, URL url, String[] media) {
	parse(source.toCharArray(), 0, source.length(), url, media);
    }

    public void parse(char[] source, int offset, int count,
		      URL url, String[] media) {
	if (media == null)
	    media = new String[0];

	try {
	    if (handler.beginStyleSheet(url, null, media)) {
		pushSource(source, offset, count, url, media);

		nextToken();
		stylesheet();

		handler.endStyleSheet(url, null, media);
	    }
	} finally {
	    while (scanner != null)
		popSource();
	}
    }

    private void nextToken() {
	token = scanner.getNextToken();
    }

    private void pushSource(String source, URL url, String[] media) {
	pushSource(source.toCharArray(), 0, source.length(), url, media);
    }

    private void pushSource(char[] source, int offset, int count,
			    URL url, String[] media) {
	if (!sources.empty()) {
	    Source top = (Source) sources.peek();
	    top.token = token;
	}

	Source newTop = new Source(new Scanner(source, offset, count), 
				   url, media);
	sources.push(newTop);

	scanner = newTop.scanner;
	this.url = newTop.url;
	this.media = newTop.media;
	token = 0;
    }

    private void popSource() {
	if (!sources.empty()) {
	    sources.pop();

	    if (sources.empty()) {
		scanner = null;
		url = null;
		media = null;
		token = 0;
	    } else {
		Source top = (Source) sources.peek();
		scanner = top.scanner;
		url = top.url;
		media = top.media;
		token = top.token;
	    }
	}
    }

    public URL getURL() {
	return url;
    }

    public String[] getMedia() {
	return media;
    }

    public int getLineNumber() {
	return scanner.getLineNumber();
    }

    public int getColumnNumber() {
	return scanner.getColumnNumber();
    }

    // -----------------------------------------------------------------------

    private void stylesheet() {
	if (token == Scanner.CHARSET_SYM) 
	    // Ignored.
	    skipStatement();

	skipSpaceBetweenBlocks();

	while (token == Scanner.IMPORT_SYM) {
	    import_();

	    skipSpaceBetweenBlocks();
	}

	for (;;) {
	    switch (token) {
	    case Scanner.EOF:
		return;

	    case Scanner.MEDIA_SYM:
		media();
		break;

	    case Scanner.PAGE_SYM:
		page();
		break;

	    case Scanner.FONT_FACE_SYM:
		font_face();
		break;

	    default:
		ruleset();
	    }

	    skipSpaceBetweenBlocks();
	}
    }

    private void import_() {
	// Skip IMPORT_SYM.
	nextToken();
	skipSpace();

	String spec;
	switch (token) {
	case Scanner.STRING:
	    spec = ParseUtil.lexemeToString(scanner.getLexeme());
	    break;
	case Scanner.URI:
	    spec = ParseUtil.lexemeToURI(scanner.getLexeme());
	    break;
	default:
	    reportError("expected " + Scanner.tokenName(Scanner.STRING) + 
			" or " + Scanner.tokenName(Scanner.URI) + 
			", found " + Scanner.tokenName(token));
	    skipStatement();
	    return;
	}
	if (spec.trim().length() == 0) {
	    reportError("empty import specification");
	    skipStatement();
	    return;
	}
	nextToken();
	skipSpace();

	String[] newMedia = mediaList();
	if (!matchToken(';')) {
	    skipStatement();
	    return;
	}
	nextToken();
	skipSpace();

	if (handler.beginStyleSheet(url, spec, newMedia)) {		
	    URL newURL;
	    try {
		newURL = new URL(url, spec);
	    } catch (MalformedURLException e1) {
		reportError("cannot convert \"" + spec + "\" to an URL", e1);
		// Nothing to skip.
		handler.endStyleSheet(url, spec, newMedia);
		return;
	    }

	    String newSource;
	    try {
		newSource = ParseUtil.loadSource(newURL);
	    } catch (IOException e2) {
		reportError("cannot load " + newURL.toExternalForm(), e2);
		// Nothing to skip.
		handler.endStyleSheet(url, spec, newMedia);
		return;
	    }

	    pushSource(newSource, newURL, newMedia);

	    nextToken();
	    stylesheet();
	    
	    popSource();

	    handler.endStyleSheet(url, spec, newMedia);
	}
    }

    private String[] mediaList() {
	Vector newMediaList = new Vector();

	for (;;) {
	    if (token == Scanner.IDENT) {
		newMediaList.addElement(
		    ParseUtil.lexemeToIdent(scanner.getLexeme()));
		nextToken();
		skipSpace();
	    } else {
		break;
	    }

	    if (token == ',') {
		nextToken();
		skipSpace();
	    } else {
		break;
	    }
	}

	String[] newMedia = new String[newMediaList.size()];
	newMediaList.copyInto(newMedia);

	return newMedia;
    }

    private void media() {
	// Skip MEDIA_SYM.
	nextToken();
	skipSpace();

	String[] newMedia = mediaList();
	if (!matchToken('{')) {
	    skipBlock();
	    return;
	}
	if (newMedia.length == 0) {
	    reportError("no media specified");
	    skipBlock();
	    return;
	}

	if (!handler.beginMedia(newMedia)) {
	    skipBlock();
	    return;
	}
	nextToken();
	skipSpace();

	loop: for (;;) {
	    switch (token) {
	    case Scanner.EOF:
	    case '}':
		break loop;

	    case Scanner.PAGE_SYM:
		// Not allowed by the standard but the sample css for HTML4
		// includes an @page inside a @media print?
		page();
		break;

	    case Scanner.FONT_FACE_SYM:
		// Not allowed by the standard.
		font_face();
		break;

	    default:
		ruleset();
	    }
	}

	handler.endMedia();

	if (!matchToken('}')) {
	    skipBlock();
	    return;
	}
	nextToken();
	skipSpace();
    }

    private void page() {
	// Skip PAGE_SYM.
	nextToken();
	skipSpace();

	String name = null;
	String pseudoPage = null;

	if (token == Scanner.IDENT) {
	    name = ParseUtil.lexemeToIdent(scanner.getLexeme());
	    nextToken();
	}
	if (token == ':') {
	    nextToken();
	    if (!matchToken(Scanner.IDENT)) {
		skipBlock();
		return;
	    }
	    pseudoPage = ParseUtil.lexemeToIdent(scanner.getLexeme());
	    nextToken();
	}
	skipSpace();

	if (!matchToken('{')) {
	    skipBlock();
	    return;
	}

	if (!handler.beginPage(name, pseudoPage)) {
	    skipBlock();
	    return;
	}
	nextToken();
	skipSpace();

	declarations();
	handler.endPage();

	if (!matchToken('}')) {
	    skipBlock();
	    return;
	}
	nextToken();
	skipSpace();
    }

    private void declarations() {
	for (;;) {
	    while (token == ';') {
		nextToken();
		skipSpace();
	    }

	    switch (token) {
	    case Scanner.EOF:
	    case '}':
		return;

	    default:
		declaration();
	    }
	}
    }

    private void font_face() {
	// Skip FONT_FACE_SYM.
	nextToken();
	skipSpace();

	if (!matchToken('{')) {
	    skipBlock();
	    return;
	}

	if (!handler.beginFontFace()) {
	    skipBlock();
	    return;
	}
	nextToken();
	skipSpace();

	declarations();
	handler.endFontFace();

	if (!matchToken('}')) {
	    skipBlock();
	    return;
	}
	nextToken();
	skipSpace();
    }

    private void ruleset() {
	// Check it is really a ruleset.
	handler.beginRule();
	handler.beginSelectors();
	loop: for (;;) {
	    if (!selector()) {
		skipBlock();
		handler.endSelectors();
		handler.endRule();
		return;
	    }

	    switch (token) {
	    case Scanner.EOF:
	    case '{':
		break loop;

	    case ',':
		nextToken();
		skipSpace();
		break;
	    }
	}
	handler.endSelectors();

	if (!matchToken('{')) {
	    skipBlock();
	    handler.endRule();
	    return;
	}

	handler.beginProperties();
	nextToken();
	skipSpace();

	declarations();
	handler.endProperties();
	handler.endRule();

	if (!matchToken('}')) {
	    skipBlock();
	    return;
	}
	nextToken();
	skipSpace();
    }

    private boolean selector() {
	Vector components = new Vector();

	SelectorComponent component = new SelectorComponent();
	component.combinator = SelectorComponent.NONE;
	switch (simple_selector(component)) {
	case -1:
	    return false;
	case 0:
	    reportError("expected simple selector, found " + 
			Scanner.tokenName(token));
	    return false;
	}
	components.addElement(component);

	loop: for (;;) {
	    component = new SelectorComponent();
	    switch (token) {
	    case '+': 
	    case '>':
		component.combinator = (char) token;
		nextToken();
		skipSpace();
		break;
	    default:
		component.combinator = SelectorComponent.NONE;
	    }

	    switch (simple_selector(component)) {
	    case -1:
		return false;
	    case 0:
		if (component.combinator != SelectorComponent.NONE) {
		    reportError("expected simple selector after " + 
				Scanner.tokenName(component.combinator) + 
				", found " + Scanner.tokenName(token));
		    return false;
		}
		break loop;
	    }
	    components.addElement(component);
	}

	if (!matchToken(',', '{'))
	    return false;

	SelectorComponent[] componentList = 
	    new SelectorComponent[components.size()];
	components.copyInto(componentList);
	handler.selector(componentList);
	return true;
    }

    private int simple_selector(SelectorComponent component) {
	switch (token) {
	case Scanner.IDENT:
	    component.elementName = 
		ParseUtil.lexemeToCaseSensitiveIdent(scanner.getLexeme());
	    nextToken();
	    break;
	case '*':
	    component.elementName = "*";
	    nextToken();
	    break;
	}

	Vector attribs = new Vector();
	Vector pseudos = new Vector();

	loop: for (;;) {
	    AttributeSelector attrib;
	    String pseudo;

	    switch (token) {
	    case Scanner.HASH:
		attrib = idAttrib();
		if (attrib == null)
		    return -1;
		attribs.addElement(attrib);
		break;
	    case '.':
		attrib = classAttrib();
		if (attrib == null)
		    return -1;
		attribs.addElement(attrib);
		break;
	    case '[':
		attrib = namedAttrib();
		if (attrib == null)
		    return -1;
		attribs.addElement(attrib);
		break;
	    case ':':
		pseudo = pseudo();
		if (pseudo == null)
		    return -1;
		pseudos.addElement(pseudo);
		break;
	    default:
		break loop;
	    }
	}
	skipSpace();

	if (component.elementName == null &&
	    attribs.size() == 0 &&
	    pseudos.size() == 0)
	    return 0;

	if (component.elementName == null)
	    component.elementName = "*";

	component.attributeSelectors = new AttributeSelector[attribs.size()];
	attribs.copyInto(component.attributeSelectors);

	component.pseudos = new String[pseudos.size()];
	pseudos.copyInto(component.pseudos);

	return 1;
    }

    private AttributeSelector idAttrib() {
	AttributeSelector attrib = new AttributeSelector();
	attrib.matcher = "#";
	attrib.value = ParseUtil.lexemeToCaseSensitiveIdent(
	    scanner.getLexeme().substring(1));
	nextToken();

	return attrib;
    }

    private AttributeSelector classAttrib() {
	// Skip '.'.
	nextToken();

	if (!matchToken(Scanner.IDENT))
	    return null;
	AttributeSelector attrib = new AttributeSelector();
	attrib.matcher = ".";
	attrib.value = 
	    ParseUtil.lexemeToCaseSensitiveIdent(scanner.getLexeme());
	nextToken();

	return attrib;
    }

    private AttributeSelector namedAttrib() {
	// Skip ']'.
	nextToken();
	skipSpace();
	
	if (!matchToken(Scanner.IDENT))
	    return null;
	AttributeSelector attrib = new AttributeSelector();
	attrib.name = 
	    ParseUtil.lexemeToCaseSensitiveIdent(scanner.getLexeme());
	nextToken();
	skipSpace();

	boolean hasValue = true;
	switch (token) {
	case '=':
	    attrib.matcher = "=";
	    break;
	case Scanner.INCLUDES:
	    attrib.matcher = "~=";
	    break;
	case Scanner.DASHMATCH:
	    attrib.matcher = "|=";
	    break;
	default:
	    hasValue = false;
	}
	if (hasValue) {
	    nextToken();
	    skipSpace();

	    if (!matchToken(Scanner.IDENT, Scanner.STRING)) 
		return null;
	    if (token == Scanner.IDENT) 
	       attrib.value = 
		   ParseUtil.lexemeToCaseSensitiveIdent(scanner.getLexeme());
	    else
	       attrib.value = ParseUtil.lexemeToString(scanner.getLexeme());
	    nextToken();
	    skipSpace();
	}

	if (!matchToken(']'))
	    return null;
	nextToken();
	    
	return attrib;
    }

    private String pseudo() {
	// Skip ':'.
	nextToken();

	if (!matchToken(Scanner.IDENT, Scanner.FUNCTION)) 
	    return null;

	String pseudo;

	if (token == Scanner.IDENT) {
	    pseudo = ParseUtil.lexemeToIdent(scanner.getLexeme());
	    nextToken();
	} else {
	    pseudo = ParseUtil.unescape(scanner.getLexeme()).toLowerCase();
	    nextToken();
	    skipSpace();

	    if (!matchToken(Scanner.IDENT))
		return null;
	    pseudo += 
		ParseUtil.lexemeToCaseSensitiveIdent(scanner.getLexeme());
	    nextToken();
	    skipSpace();

	    if (!matchToken(')'))
		return null;
	    pseudo += ')';
	    nextToken();
	}

	return pseudo;
    }

    private void declaration() {
	// Check it is really a declaration.
	if (!matchToken(Scanner.IDENT)) {
	    skipToEndOfDecl();
	    return;
	}
	String name = ParseUtil.lexemeToIdent(scanner.getLexeme());
	nextToken();
	skipSpace();

	if (!matchToken(':')) {
	    skipToEndOfDecl();
	    return;
	}
	nextToken();
	skipSpace();

	ValueComponent[] components = expr();
	if (components == null) {
	    skipToEndOfDecl();
	    return;
	}

	boolean important = false;
	if (token == Scanner.IMPORTANT_SYM) {
	    important = true;
	    nextToken();
	    skipSpace();
	}

	if (!matchToken(';', '}')) {
	    skipToEndOfDecl();
	    return;
	}
	handler.property(name, components, important);
    }

    private ValueComponent[] expr() {
	Vector components = new Vector();

	ValueComponent component = new ValueComponent();
	component.operator = ValueComponent.NONE;
	switch (term(component)) {
	case -1:
	    return null;
	case 0:
	    reportError("expected expression term, found " + 
			Scanner.tokenName(token));
	    return null;
	}
	components.addElement(component);

	loop: for (;;) {
	    component = new ValueComponent();
	    switch (token) {
	    case '/': 
	    case ',':
		component.operator = (char) token;
		nextToken();
		skipSpace();
		break;
	    default:
		component.operator = ValueComponent.NONE;
	    }

	    switch (term(component)) {
	    case -1:
		return null;
	    case 0:
		if (component.operator != ValueComponent.NONE) {
		    reportError("expected expression term after " + 
				Scanner.tokenName(component.operator) + 
				", found " + Scanner.tokenName(token));
		    return null;
		}
		break loop;
	    }
	    components.addElement(component);
	}

	ValueComponent[] componentList = new ValueComponent[components.size()];
	components.copyInto(componentList);
	return componentList;
    }

    private int term(ValueComponent component) {
	int signToken = Scanner.EOF;
	switch (token) {
	case '-': 
	case '+':
	    signToken = token;
	    nextToken();
	    break;
	}

	component.type = token;

	String lexeme = scanner.getLexeme();
	switch (token) {

	case Scanner.EMS: 
	case Scanner.EXS: 
	    component.type = Scanner.LENGTH;
	    /*FALLTHROUGH*/
	case Scanner.LENGTH: 
	case Scanner.PERCENTAGE: 
	case Scanner.DIMEN: 
	case Scanner.ANGLE: 
	case Scanner.TIME: 
	case Scanner.FREQ: 
	    component.unit = ParseUtil.lexemeToUnit(lexeme);
            /*FALLTHROUGH*/
	case Scanner.NUMBER: 
	    component.numericValue = ParseUtil.lexemeToNumber(lexeme);
	    if (Double.isNaN(component.numericValue)) {
		reportError("invalid " + Scanner.tokenName(token));
		return -1;
	    }
	    if (signToken == '-') {
		if (token == Scanner.TIME || token == Scanner.FREQ) {
		    reportError("negative " + Scanner.tokenName(token));
		    return -1;
		}
		component.numericValue = -component.numericValue;
	    }
	    break;

	case Scanner.IDENT: 
	    component.stringValue = ParseUtil.lexemeToIdent(lexeme);
	    break;

	case Scanner.STRING: 
	    component.stringValue = ParseUtil.lexemeToString(lexeme);
	    break;

	case Scanner.URI: 
	    component.stringValue = ParseUtil.lexemeToURI(lexeme);
	    break;

	case Scanner.UNICODERANGE: 
	    component.unicodeRange = ParseUtil.lexemeToUnicodeRange(lexeme);
	    if (component.unicodeRange == null) {
		reportError("invalid " + Scanner.tokenName(token));
		return -1;
	    }
	    break;

	case Scanner.HASH: 
	    component.hexColor = ParseUtil.lexemeToHexColor(lexeme);
	    if (component.hexColor == null) {
		reportError(Scanner.tokenName(token) + " is not a color");
		return -1;
	    }
	    break;

	case Scanner.FUNCTION: 
	    {
		nextToken();
		component.functionArguments = expr();
		if (component.functionArguments == null)
		    return -1;

		if (token != ')') {
		    reportError("expected " + Scanner.tokenName(')') + 
				", found " + Scanner.tokenName(token));
		    return -1;
		}

		component.functionName = ParseUtil.lexemeToFunction(lexeme);
	    }
	    break;

	default:
	    if (signToken != Scanner.EOF) {
		reportError("expected term after " + 
			    Scanner.tokenName(signToken) + ", found " +
			    Scanner.tokenName(token));
		return -1;
	    }
	    return 0;
	}

	nextToken();
	skipSpace();
	return 1;
    }

    // -----------------------------------------------------------------------

    private boolean matchToken(int expectedToken) {
	if (token != expectedToken) {
	    reportError("expected " + Scanner.tokenName(expectedToken) + 
			", found " + Scanner.tokenName(token));
	    return false;
	} else {
	    return true;
	}
    }

    private boolean matchToken(int expectedToken1, int expectedToken2) {
	if (token != expectedToken1 && token != expectedToken2) {
	    reportError("expected " + Scanner.tokenName(expectedToken1) + 
			" or " + Scanner.tokenName(expectedToken2) + 
			", found " + Scanner.tokenName(token));
	    return false;
	} else {
	    return true;
	}
    }

    private void reportError(String message) {
	reportError(message, null);
    }

    private void reportError(String message, Exception reason) {
	if (reason != null) {
	    String detail = reason.getMessage();
	    if (detail == null)
		detail = reason.getClass().getName();

	    message += ": " + detail;
	}

	handler.error(getURL(), getLineNumber(), getColumnNumber(), message);
    }

    // -----------------------------------------------------------------------

    private void skipSpaceBetweenBlocks() {
	for (;;) {
	    switch (token) {
	    case Scanner.S:
	    case Scanner.CDO:
	    case Scanner.CDC:
		nextToken();
		break;
	    default:
		return;
	    }
	}
    }

    private void skipSpace() {
	while (token == Scanner.S)
	    nextToken();
    }

    private void skipStatement() {
	int nesting = 0;
	for (;;) {
	    switch (token) {
	    case Scanner.EOF:
		return;

	    case ';':
		nextToken();
		if (nesting <= 0) {
		    skipSpace();
		    return;
		}
		break;

	    case '{':
		++nesting;
		nextToken();
		break;

	    case '}':
		--nesting;
		nextToken();
		break;

	    default:
		nextToken();
	    }
	}
    }

    private void skipBlock() {
	int nesting = 0;
	for (;;) {
	    switch (token) {
	    case Scanner.EOF:
		return;

	    case '{':
		++nesting;
		nextToken();
		break;

	    case '}':
		--nesting;
		nextToken();
		if (nesting <= 0) {
		    skipSpace();
		    return;
		}
		break;

	    default:
		nextToken();
	    }
	}
    }

    private void skipToEndOfDecl() {
	int nesting = 0;
	for (;;) {
	    switch (token) {
	    case Scanner.EOF:
		return;

	    case ';':
		if (nesting <= 0) 
		    return;
		nextToken();
		break;

	    case '{':
		++nesting;
		nextToken();
		break;

	    case '}':
		--nesting;
		if (nesting <= 0) 
		    return;
		nextToken();
		break;

	    default:
		nextToken();
	    }
	}
    }
}
