/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.cssparser;

import fr.pixware.util.FileUtil;

public class Scanner {
    public static final int EOF			= -1;
    public static final int S 			= -2;
    public static final int CDO 		= -3;
    public static final int CDC 		= -4;
    public static final int INCLUDES 		= -5;
    public static final int DASHMATCH 		= -6;
    public static final int STRING 		= -7;
    public static final int IDENT 		= -8;
    public static final int HASH 		= -9;
    public static final int IMPORT_SYM 		= -10;
    public static final int PAGE_SYM 		= -11;
    public static final int MEDIA_SYM 		= -12;
    public static final int FONT_FACE_SYM 	= -13;
    public static final int CHARSET_SYM 	= -14;
    public static final int ATKEYWORD 		= -15;
    public static final int IMPORTANT_SYM 	= -16;
    public static final int EMS 		= -17;
    public static final int EXS 		= -18;
    public static final int LENGTH 		= -19;
    public static final int ANGLE 		= -20;
    public static final int TIME 		= -21;
    public static final int FREQ 		= -22;
    public static final int DIMEN 		= -23;
    public static final int PERCENTAGE 		= -24;
    public static final int NUMBER 		= -25;
    public static final int URI 		= -26;
    public static final int FUNCTION 		= -27;
    public static final int UNICODERANGE 	= -28;

    private static final String[] tokenNames = {
	    null,
	    "end of data", 
	    "space",
	    "'<!--'",
	    "'-->'",
	    "'~='",
	    "'|='",
	    "string",
	    "identifier",
	    "'#'identifier",
	    "'@import'",
	    "'@page'",
	    "'@media'",
	    "'@font-face'",
	    "'@charset'",
	    "'@'identifier",
	    "'!important'",
	    "em",
	    "ex",
	    "length",
	    "angle",
	    "time",
	    "frequency",
	    "dimension",
	    "percentage",
	    "number",
	    "URL",
	    "function",
	    "Unicode range"
    };

    private char[] source;
    private int firstPos, endPos, pos, lexemePos;
    private int cachedPos, lineNum, charNum;

    // -----------------------------------------------------------------------

    public Scanner(String source) {
	this();
	setSourceChars(source.toCharArray(), 0, source.length());
    }

    public Scanner(char[] source, int offset, int count) {
	this();
	setSourceChars(source, offset, count);
    }

    public Scanner() {
	source = null;
	firstPos = endPos = pos = lexemePos = -1;
	cachedPos = lineNum = charNum = -1;
    }

    public void setSourceChars(char[] source, int offset, int count) {
	this.source = source;
	firstPos = offset;
	endPos = offset + count;
	lexemePos = pos = firstPos;
	cachedPos = lineNum = charNum = -1;
    }

    public char[] getSourceChars() {
	return source;
    }

    public int getSourceCharOffset() {
	return firstPos;
    }

    public int getSourceCharCount() {
	return (endPos - firstPos);
    }

    public int getNextToken() {
	if (pos >= endPos)
	    return EOF;

	int matched;
	char c = source[pos];

	switch (c) {
	case ' ': case '\t': case '\r': case '\n': case '\f':
	    lexemePos = pos;
	    skipSpace();
	    return S;

	case '/':
	    if (match("/*") && skipComment()) {
		return getNextToken();
	    } else {
		lexemePos = pos;
		++pos;
		return c;
	    }

	case '<':
	    lexemePos = pos;
	    if (match("<!--")) {
		pos += 4;
		return CDO;
	    } else {
		++pos;
		return c;
	    }

	case '-':
	    lexemePos = pos;
	    if (match("-->")) {
		pos += 3;
		return CDC;
	    } else {
		++pos;
		return c;
	    }

	case '~':
	    lexemePos = pos;
	    if (match("~=")) {
		pos += 2;
		return INCLUDES;
	    } else {
		++pos;
		return c;
	    }

	case '|':
	    lexemePos = pos;
	    if (match("|=")) {
		pos += 2;
		return DASHMATCH;
	    } else {
		++pos;
		return c;
	    }

	case '\"': case '\'': 
	    lexemePos = pos;
	    if (skipString(c)) {
		return STRING;
	    } else {
		++pos;
		return c;
	    }

	case '#':
	    lexemePos = pos;
	    ++pos;
	    if (matchNameStart()) {
		skipName();
		return HASH;
	    } else {
		return c;
	    }

	case '@':
	    if (match("@import")) {
		lexemePos = pos;
		pos += 7;
		return IMPORT_SYM;
	    } else if (match("@page")) {
		lexemePos = pos;
		pos += 5;
		return PAGE_SYM;
	    } else if (match("@media")) {
		lexemePos = pos;
		pos += 6;
		return MEDIA_SYM;
	    } else if (match("@font-face")) {
		lexemePos = pos;
		pos += 10;
		return FONT_FACE_SYM;
	    } else if (match("@charset")) {
		lexemePos = pos;
		pos += 8;
		return CHARSET_SYM;
	    } else {
		lexemePos = pos;
		++pos;
		if (matchIdentStart()) {
		    skipName();
		    return ATKEYWORD;
		} else {
		    return c;
		}
	    }

	case '!':
	    lexemePos = pos;
	    if (skipImportant()) {
		return IMPORTANT_SYM;
	    } else {
		++pos;
		return c;
	    }

	case '.': 
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	    lexemePos = pos;
	    if (c == '.' && 
		(pos+1 >= endPos || !Character.isDigit(source[pos+1]))) {
		++pos;
		return c;
	    }
	    skipNumber();
	    if (match("em")) {
		pos += 2;
		return EMS;
	    } else if (match("ex")) {
		pos += 2;
		return EXS;
	    } else if (match("cm") || match("mm") || match("in") ||
		       match("pt") || match("pc") || match("px")) {
		pos += 2;
		return LENGTH;
	    } else if (match("deg") || match("rad")) {
		pos += 3;
		return ANGLE;
	    } else if (match("grad")) {
		pos += 4;
		return ANGLE;
	    } else if (match("ms")) {
		pos += 2;
		return TIME;
	    } else if (match("s")) {
		++pos;
		return TIME;
	    } else if (match("hz")) {
		pos += 2;
		return FREQ;
	    } else if (match("khz")) {
		pos += 3;
		return FREQ;
	    } else if (match("%")) {
		++pos;
		return PERCENTAGE;
	    } else {
		if (matchIdentStart()) {
		    skipName();
		    return DIMEN;
		} else {
		    return NUMBER;
		}
	    }

	case 'u': case 'U':
	    lexemePos = pos;
	    if (match("url(")) {
		if (skipURI())
		    return URI;
	    } else if (match("u+")) {
		if (skipUnicodeRange())
		    return UNICODERANGE;
	    }
	    /*FALLTHROUGH*/

	default:
	    lexemePos = pos;
	    if (matchIdentStart()) {
		skipName();
		if (pos < endPos && source[pos] == '(') {
		    ++pos;
		    return FUNCTION;
		} else {
		    return IDENT;
		}
	    } else {
		++pos;
		return c;
	    }
	}
    }

    private boolean match(String keyword) {
	int count = keyword.length();
	int p = pos;

	for (int i = 0; i < count; ++i) {
	    if (p >= endPos ||
		Character.toLowerCase(source[p++]) != keyword.charAt(i))
		return false;
	}

	return true;
    }

    private boolean matchNameStart() {
	char c = source[pos];
	 
	if (c == '-' ||
	    (c >= '0' && c <= '9') ||
	    (c >= 'A' && c <= 'Z') ||
	    (c >= 'a' && c <= 'z') ||
	    c >= '\200')
	    return true;

	if (c == '\\' && pos+1 < endPos) {
	    char escaped = source[pos+1];

	    return (ParseUtil.isHexDigit(escaped) || 
		    ((escaped >= ' ' && escaped <= '~') || escaped >= '\200'));
	}

	return false;
    }

    private boolean matchIdentStart() {
	char c = source[pos];
	 
	if ((c >= 'A' && c <= 'Z') ||
	    (c >= 'a' && c <= 'z') ||
	    c >= '\200')
	    return true;

	if (c == '\\' && pos+1 < endPos) {
	    char escaped = source[pos+1];

	    return (ParseUtil.isHexDigit(escaped) || 
		    ((escaped >= ' ' && escaped <= '~') || escaped >= '\200'));
	}

	return false;
    }

    private void skipSpace() {
	while (pos < endPos) {
	    switch (source[pos]) {
	    case ' ': case '\t': case '\r': case '\n': case '\f':
		++pos;
		break;
	    default:
		return;
	    }
	}
    }

    private boolean skipComment() {
	int savedPos = pos;
	// Skip "/*"
	pos += 2;

	while (pos < endPos) {
	    switch (source[pos]) {
	    case '*':
		if (pos+1 < endPos && source[pos+1] == '/') {
		    pos += 2;
		    return true;
		}
		/*FALLTHROUGH*/
	    default:
		++pos;
	    }
	}

	pos = savedPos;
	return false;
    }

    private boolean skipString(char quote) {
	int savedPos = pos;
	// Skip quote.
	++pos;

	char quote2 = (quote == '\"')? '\'' : '\"';

	while (pos < endPos) {
	    char c = source[pos];

	    if (c == quote) {
		++pos;
		return true;
	    } else if (c == '\\') {
		++pos;
		if (pos < endPos) {
		    c = source[pos];

		    if (c == quote || c == quote2 || 
			c == '\n' || c == '\r' || c == '\f') {
			++pos;
			if (c == '\r' && pos < endPos && source[pos] == '\n')
			    ++pos;
		    }
		}
	    } else if (c == quote2 || 
		       (c >= '(' && c <= '~') ||
		       c == ' ' || c == '\t'  || c == '!' || c == '#' || 
		       c == '$' || c == '%' || c == '&' || 
		       c >= '\200') {
		++pos;
	    } else {
		break;
	    }
	}

	pos = savedPos;
	return false;
    }

    private void skipName() {
	while (pos < endPos) {
	    char c = source[pos];
	 
	    if (c == '-' ||
		(c >= '0' && c <= '9') ||
		(c >= 'A' && c <= 'Z') ||
		(c >= 'a' && c <= 'z') ||
		c >= '\200') {
		++pos;
	    } else if (c == '\\') {
		if (!skipEscape())
		    return;
	    } else {
		return;
	    }
	}
    }

    private boolean skipEscape() { 
	// Assume that (source[pos] == '\\').

	if (pos+1 < endPos) {
	    char escaped = source[pos+1];

	    if (ParseUtil.isHexDigit(escaped)) {
		++pos;
		int begin = pos;

		while (pos < endPos && ParseUtil.isHexDigit(source[pos])) {
		    ++pos;
		    if (pos - begin == 6)
			break;
		}

		if (pos < endPos && ParseUtil.isSpace(source[pos])) 
		    ++pos;

		return true;
	    } else if ((escaped >= ' ' && escaped <= '~') ||
		       escaped >= '\200') {
		pos += 2;
		return true;
	    }
	}

	return false;
    }

    private void skipNumber() {
	boolean dot = false;

	while (pos < endPos) {
	    char c = source[pos];
	 
	    if (c >= '0' && c <= '9') {
		++pos;
	    } else if (c == '.') {
		if (dot)
		    return;
		dot = true;
		++pos;
	    } else {
		return;
	    }
	}
    }

    private boolean skipURI() {
	int savedPos = pos;
	// Skip "url("
	pos += 4;

	skipSpace();

	if (pos < endPos) {
	    char c = source[pos];

	    if (c == '\"' || c == '\'') {
		skipString(c);
	    } else {
		boolean invalid = false;

		while (pos < endPos) {
		    c = source[pos];

		    if (c == ')' || ParseUtil.isSpace(c)) {
			break;
		    } else if (c == '\\') {
			if (!skipEscape()) {
			    invalid = true;
			    break;
			}
		    } else if ((c >= '*' && c <= '~') ||
			       c == '!' || c == '#' || c == '$' || 
			       c == '%' || c == '&' || 
			       c >= '\200') {
			++pos;
		    } else {
			invalid = true;
			break;
		    }
		}

		if (invalid) {
		    pos = savedPos;
		    return false;
		}
	    }
	}

	skipSpace();
	if (pos < endPos && source[pos] == ')') {
	    ++pos;
	    return true;
	} else {
	    pos = savedPos;
	    return false;
	}
    }

    private boolean skipUnicodeRange() {
	int savedPos = pos;
	// Skip "U+"
	pos += 2;

	int skipped = skipHexDigits(/*skipWildcard*/ true);
	if (skipped < 0) {
	    return true;
	} else if (skipped > 0) {
	    if (pos < endPos && source[pos] == '-') {
		++pos;

		skipped = skipHexDigits(/*skipWildcard*/ false);
		if (skipped > 0)
		    return true;
	    } else {
		return true;
	    }
	} 

	pos = savedPos;
	return false;
    }

    private int skipHexDigits(boolean skipWildcard) {
	boolean hasWildcards = false;
	int count = 0;

	while (pos < endPos) {
	    char c = source[pos];

	    if (ParseUtil.isHexDigit(c) || 
		(skipWildcard && c == '?')) {
		if (c == '?')
		    hasWildcards = true;

		++pos;
		++count;
		if (count == 6)
		    break;
	    } else {
		break;
	    }
	}

	return (hasWildcards? -count : count);
    }

    private boolean skipImportant() {
	int savedPos = pos;

	if (source[pos] != '!') 
	    return false;
	++pos;

	while (pos < endPos && ParseUtil.isSpace(source[pos]))
	    ++pos;

	if (!match("important")) {
	    pos = savedPos;
	    return false;
	}
	pos += 9;

	return true;
    }

    public String getLexeme() {
	return new String(source, lexemePos, pos - lexemePos);
    }

    public int getLineNumber() {
	if (cachedPos != pos)
	    locate(pos);
	return 1+lineNum;
    }

    public int getColumnNumber() {
	if (cachedPos != pos)
	    locate(pos);
	return 1+charNum;
    }

    private void locate(int searchedPos) {
	if (searchedPos > endPos)
	    searchedPos = endPos;

	lineNum = charNum = 0;
	int p = firstPos;

	while (p < searchedPos) {
	    char c = source[p];

	    switch (c) {
	    case '\r':
		if (p+1 < endPos && source[p+1] == '\n')
		    ++p;
	    case '\n': 
	    case '\f':
		++lineNum;
		charNum = 0;
		++p;
		break;
	    default:
		++charNum;
		++p;
	    }
	}
	
	cachedPos = searchedPos;
    }

    public static String tokenName(int token) {
	if (token >= 0 && token < Character.MAX_VALUE) {
	    return "'" + ((char) token) + "'";
	} else {
	    token = -token;
	    if (token >= 1 && token < tokenNames.length)
		return tokenNames[token];
	    else
		return "(unknown " + (-token) + ")";
	}
    }

    // -----------------------------------------------------------------------

    public static void main(String[] args) throws Exception {
	if (args.length != 2) {
	    System.err.println(
		"usage: java com.xmlmind.cssparser.Scanner in_file out_file");
	    System.exit(1);
	}

	Scanner scanner = new Scanner(FileUtil.loadString(args[0]));
	java.io.PrintWriter out = new java.io.PrintWriter(
	    new java.io.FileWriter(args[1]));
	boolean location = true;
	int token;

	while ((token = scanner.getNextToken()) != Scanner.EOF) {
	    out.print(Scanner.tokenName(token));
	    if (location) {
		out.print("\t\t" + scanner.getLineNumber() + 
			  ":" + scanner.getColumnNumber() + 
			  "\t'" + scanner.getLexeme() + "'");
	    }
	    out.println();
	}

	out.flush();
	out.close();
    }
}


