/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
import java.util.Vector;
import java.net.URL;
import java.io.*;
import fr.pixware.util.StringUtil;
import com.xmlmind.cssparser.*;

public class SampleHandler extends HandlerBase {
    private static final String[] ALL = { "all" };

    private String[] handledMedia = ALL;
    private boolean handleFontFace = true;
    private boolean handlePage = true;

    private PrintWriter out = null;
    private Vector errors = new Vector();

    private boolean firstSelector = false;

    // -----------------------------------------------------------------------

    public void setHandledMedia(String[] handledMedia) {
	if (handledMedia == null || handledMedia.length == 0)
	    handledMedia = ALL;

	this.handledMedia = new String[handledMedia.length]; 
	for (int i = 0; i < handledMedia.length; ++i)
	    this.handledMedia[i] = handledMedia[i].toLowerCase();
    }

    public String[] getHandledMedia() {
	return handledMedia;
    }

    public void setPageHandled(boolean handlePage) {
	this.handlePage = handlePage;
    }

    public boolean isPageHandled() {
	return handlePage;
    }

    public void setFontFaceHandled(boolean handleFontFace) {
	this.handleFontFace = handleFontFace;
    }

    public boolean isFontFaceHandled() {
	return handleFontFace;
    }

    public void setOutput(PrintWriter writer) {
	out = writer;
    }

    public PrintWriter getOutput() {
	return out;
    }

    public ParseError[] getErrors() {
	ParseError[] errorList = new ParseError[errors.size()];
	errors.copyInto(errorList);

	return errorList;
    }

    // -----------------------------------------------------------------------

    public boolean beginStyleSheet(URL baseURL, String importSpec, 
				   String[] media) {
	if (!containsHandledMedia(media))
	    return false;

	if (importSpec != null)
	    out.println("/* " + importSpec + " *************************/\n");

	if (media.length > 0)
	    beginMedia(media);

	return true;
    }

    private boolean containsHandledMedia(String[] media) {
	if (media == null || media.length == 0 ||
	    StringUtil.contains(media, "all") ||
	    StringUtil.contains(handledMedia, "all"))
	    return true;

	for (int i = 0; i < handledMedia.length; ++i) {
	    if (StringUtil.contains(media, handledMedia[i]))
		return true;
	}

	return false;
    }

    public void endStyleSheet(URL baseURL, String importSpec, String[] media) {
	if (media.length > 0)
	    endMedia();

	if (importSpec != null) 
	    out.println("/************************* " + importSpec + " */\n");
    }

    public boolean beginMedia(String[] media) {
	if (!containsHandledMedia(media))
	    return false;

	out.print("@media ");
	out.print(StringUtil.join(media, ", "));
	out.println(" {");
	return true;
    }

    public void endMedia() {
	out.println("}\n");
    }

    public boolean beginPage(String name, String pseudoPage) {
	if (!handlePage)
	    return false;

	out.print("@page");
	if (name != null) {
	    out.print(' ');
	    out.print(escapeIdent(name));
	}
	if (pseudoPage != null) {
	    if (name == null)
		out.print(' ');
	    out.print(':');
	    out.print(escapeIdent(pseudoPage));
	}
	out.println(" {");
	return true;
    }

    public void endPage() {
	out.println("}\n");
    }

    public boolean beginFontFace() {
	if (!handleFontFace)
	    return false;

	out.println("@font-face {");
	return true;
    }

    public void endFontFace() {
	out.println("}\n");
    }

    public void beginSelectors() {
	firstSelector = true;
    }

    public void selector(SelectorComponent[] components) {
	if (firstSelector) 
	    firstSelector = false;
	else
	    out.println(",");

	out.print(toString(components[0]));
	for (int i = 1; i < components.length; ++i) {
	    if (components[i].combinator == SelectorComponent.NONE)
		out.print(' ');
	    out.print(toString(components[i]));
	}
    }

    public void beginProperties() {
	out.println(" {");
    }

    public void property(String name, ValueComponent[] components, 
			 boolean important) {
	out.print("    ");
	out.print(escapeIdent(name));
	out.print(": ");

	out.print(toString(components[0]));
	for (int i = 1; i < components.length; ++i) {
	    if (components[i].operator == ValueComponent.NONE)
		out.print(' ');
	    out.print(toString(components[i]));
	}
	if (important)
	    out.print(" !important");
	out.println(';');
    }

    public void endProperties() {
	out.println("}\n");
    }

    public void error(URL url, int lineNumber, int columNumber, 
		      String message) {
	errors.addElement(
	    new ParseError(url, lineNumber, columNumber, message));
    }

    // -----------------------------------------------------------------------

    private static String toString(SelectorComponent compo) {
	StringBuffer buffer = new StringBuffer();

	if (compo.combinator != SelectorComponent.NONE)
	    buffer.append(compo.combinator);

	buffer.append(compo.elementName);

	for (int i = 0; i < compo.attributeSelectors.length; ++i)
	    buffer.append(toString(compo.attributeSelectors[i]));
	
	for (int i = 0; i < compo.pseudos.length; ++i) {
	    String pseudo = compo.pseudos[i];

	    buffer.append(':');

	    int parent = pseudo.indexOf('(');
	    if (parent < 0) {
		buffer.append(pseudo);
	    } else {
		// Something like lang(fr). Do not escape parenthesis!
		buffer.append(escapeIdent(pseudo.substring(0, parent)));
		buffer.append('(');
		buffer.append(escapeIdent(
		    pseudo.substring(parent+1, pseudo.length()-1)));
		buffer.append(')');
	    }
	}
	
	return buffer.toString();
    }

    private static String toString(AttributeSelector sel) {
	StringBuffer buffer = new StringBuffer();

	if (sel.name != null) {
	    buffer.append('[');
	    buffer.append(sel.name);
	}

	if (sel.matcher != null)
	    buffer.append(sel.matcher);

	if (sel.value != null) {
	    if (".".equals(sel.matcher) || "#".equals(sel.matcher)) 
		buffer.append(escapeIdent(sel.value));
	    else
		buffer.append(quote(sel.value));
	}

	if (sel.name != null) 
	    buffer.append(']');

	return buffer.toString();
    }

    private static String toString(ValueComponent compo) {
	StringBuffer buffer = new StringBuffer();

	if (compo.operator != ValueComponent.NONE)
	    buffer.append(compo.operator);

	switch (compo.type) {
	case ValueComponent.NUMBER:
	    buffer.append(compo.numericValue);
	    break;

	case ValueComponent.LENGTH:
	case ValueComponent.DIMENSION:
	case ValueComponent.PERCENTAGE:
	case ValueComponent.ANGLE:
	case ValueComponent.TIME:
	case ValueComponent.FREQUENCY:
	    buffer.append(compo.numericValue);
	    buffer.append(compo.unit);
	    break;

	case ValueComponent.IDENTIFIER:
	    buffer.append(escapeIdent(compo.stringValue));
	    break;

	case ValueComponent.STRING:
	    buffer.append(quote(compo.stringValue));
	    break;

	case ValueComponent.URI:
	    buffer.append("url(");
	    // URI are not escaped (i.e. ' ' escaped as "%20").
	    buffer.append(quote(compo.stringValue));
	    buffer.append(')');
	    break;

	case ValueComponent.UNICODE_RANGE:
	    buffer.append("U+");
	    buffer.append(toHexString(compo.unicodeRange[0], 4));
	    buffer.append('-');
	    buffer.append(toHexString(compo.unicodeRange[1], 4));
	    break;

	case ValueComponent.FUNCTION:
	    {
		int[] rgb;

		if (compo.functionName.equals("rgb") &&
		    (rgb = 
		     ParseUtil.functionArgumentsToRGB(compo.functionArguments))
		    != null) {
		    buffer.append(toHexColor(rgb));
		} else {
		    buffer.append(escapeIdent(compo.functionName));
		    buffer.append('(');
		    if (compo.functionArguments.length > 0) {
			ValueComponent[] args = compo.functionArguments;

			buffer.append(toString(args[0]));
			for (int i = 1; i < args.length; ++i) {
			    if (args[i].operator == ValueComponent.NONE)
				buffer.append(' ');
			    buffer.append(toString(args[i]));
			}
		    }
		    buffer.append(')');
		}
	    }
	    break;

	case ValueComponent.HEX_COLOR:
	    buffer.append(toHexColor(compo.hexColor));
	    break;
	}

	return buffer.toString();
    }

    // -----------------------------------------------------------------------

    private static String escapeIdent(String s) {
	StringBuffer buffer = new StringBuffer();

	int count = s.length();
	for (int i = 0; i < count; ++i) {
	    char c = s.charAt(i);

	    if (!(c == '-' ||
		  (c >= '0' && c <= '9') ||
		  (c >= 'A' && c <= 'Z') ||
		  (c >= 'a' && c <= 'z') ||
		  c >= '\200')) {
		buffer.append('\\');
	    }
	    buffer.append(c);
	}

	return buffer.toString();
    }

    private static String quote(String s) {
	StringBuffer buffer = new StringBuffer();
	buffer.append('\"');

	int count = s.length();
	for (int i = 0; i < count; ++i) {
	    char c = s.charAt(i);
	    if (c == '\"')
		buffer.append("\\\"");
	    else
		buffer.append(c);
	}

	buffer.append('\"');
	return buffer.toString();
    }

    private static String toHexColor(int[] rgb) {
	StringBuffer buffer = new StringBuffer();

	buffer.append('#');
	buffer.append(toHexString(rgb[0], 2));
	buffer.append(toHexString(rgb[1], 2));
	buffer.append(toHexString(rgb[2], 2));

	return buffer.toString();
    }

    private static String toHexString(int value, int width) {
	String hex = Integer.toHexString(value).toUpperCase();

	int length = hex.length();
	if (length < width) {
	    StringBuffer buffer = new StringBuffer();

	    while (length < width) {
		buffer.append('0');
		++length;
	    }
	    buffer.append(hex);

	    hex = buffer.toString();
	}

	return hex;
    }

    // -----------------------------------------------------------------------

    public static void main(String[] args) throws Exception {
	boolean handlePage = true;
	boolean handleFontFace = true;
	String[] handledMedia = ALL;

	if (args.length < 2) {
	    System.err.println(
		"usage: java com.xmlmind.cssparser.SampleHandler" +
		" in_file out_file ?+p|-p? ?+f|-f? ?media ... media?");
	    System.exit(1);
	}
	if (args.length >= 3 && args[2].equals("-p")) 
	    handlePage = false;
	if (args.length >= 4 && args[3].equals("-f")) 
	    handleFontFace = false;
	if (args.length >= 5) {
	    handledMedia = new String[args.length - 4];
	    System.arraycopy(args, 4, handledMedia, 0, handledMedia.length);
	}

	SampleHandler handler = new SampleHandler();
	handler.setPageHandled(handlePage);
	handler.setFontFaceHandled(handleFontFace);
	handler.setHandledMedia(handledMedia);
	PrintWriter out = 
	    new PrintWriter(new FileWriter(args[1]), /*flush*/ true);
	handler.setOutput(out);

	Parser parser = new Parser();
	parser.setHandler(handler);
	parser.parse(new File(args[0]));

	ParseError[] errors = handler.getErrors();
	if (errors.length > 0) {
	    out.println("/* ERRORS ****************************************");
	    for (int i = 0; i < errors.length; ++i) {
		out.println(errors[i]);
	    }
	    out.println("**************************************** ERRORS */");
	}

	out.flush();
	out.close();
    }
}

