/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.gadget;

import java.awt.Component;
import java.awt.Point;
import java.awt.AWTEvent;
import java.awt.event.*;

public class EventDispatcher {
    private static final int FOCUS_LOST      = 0;
    private static final int FOCUS_REQUESTED = 1;
    private static final int FOCUS_GAINED    = 2;

    private RootGadget root;

    // State variables.
    private int focusState = FOCUS_LOST;
    private Gadget focusGadget = null;

    private MouseEventTargets mousePressedTargets = new MouseEventTargets();
    private MouseEventTargets mouseReleasedTargets = new MouseEventTargets();
    private MouseEvent lastMouseEvent = null;
    private boolean mouseDragged = false;

    // Scratch variables.
    private Point point = new Point();

    // Platform dependant ``constants''.
    private static int popupTriggerEventModifiers = MouseEvent.BUTTON3_MASK;
    private static int popupTriggerEventId = MouseEvent.MOUSE_PRESSED;

    // -----------------------------------------------------------------------

    public EventDispatcher(RootGadget root) {
	this.root = root;
    }

    public void setFocusGadget(Gadget gadget) {
	if (gadget == focusGadget &&
	    (gadget == null || focusState == FOCUS_GAINED))
	    // Nothing to do.
	    return;

	if (focusGadget != null) {
	    if (focusState == FOCUS_GAINED)
		focusGadget.focusLost();
	    focusGadget = null;
	}
	    
	if (gadget != null) {
	    focusGadget = gadget;

	    switch (focusState) {
	    case FOCUS_LOST:
		focusState = FOCUS_REQUESTED;
		root.getPanel().requestFocus();
		break;

	    case FOCUS_GAINED:
		focusGadget.focusGained();
		break;

	    // Otherwise, FOCUS_REQUESTED: nothing to do.
	    }
	}
    }

    public Gadget getFocusGadget() {
	return focusGadget;
    }

    public void gadgetRemoved(Gadget gadget) {
	if (focusGadget == null) 
	    return;

	if (gadget == focusGadget ||
	    ((gadget instanceof GadgetContainer) && 
	     ((GadgetContainer) gadget).isAncestorOf(focusGadget)))
	    setFocusGadget(null);
    }

    public void dispatchEvent(AWTEvent e) {
	MouseEvent mouseEvent;

	switch (e.getID()) {
	case FocusEvent.FOCUS_GAINED:
	    focusState = FOCUS_GAINED;
	    if (focusGadget != null)
		focusGadget.focusGained();
	    break;

	case FocusEvent.FOCUS_LOST:
	    focusState = FOCUS_LOST;
	    if (focusGadget != null)
		focusGadget.focusLost();
	    break;

	case KeyEvent.KEY_PRESSED:
	case KeyEvent.KEY_RELEASED:
	case KeyEvent.KEY_TYPED:
	    if (focusGadget != null) {
		KeyEvent ke = (KeyEvent) e;
		if (focusGadget.handleKeyEvent(ke))
		    ke.consume();
	    }
	    break;

	case MouseEvent.MOUSE_PRESSED:
	    if (mousePressedTargets.size() > 0) {
		// Something weird happened. Synthesize a fake mouse released
		// event.

		MouseEvent fakeEvent = 
		    new MouseEvent((Component) lastMouseEvent.getSource(),
				   MouseEvent.MOUSE_RELEASED,
				   lastMouseEvent.getWhen(),
				   lastMouseEvent.getModifiers(),
				   lastMouseEvent.getX(), 
				   lastMouseEvent.getY(),
				   lastMouseEvent.getClickCount(),
				   /*isPopupTrigger*/ false);

		handleMouseEvent(fakeEvent, mousePressedTargets);
	    }
	    lastMouseEvent = null;
	    mouseDragged = false;

	    mouseEvent = (MouseEvent) e;
	    pushMouseEventTargets(mouseEvent, mousePressedTargets);
	    if (mousePressedTargets.size() > 0) {
		dispatchMouseEvent(mouseEvent, mousePressedTargets);
		lastMouseEvent = mouseEvent;
	    }
	    break;

	case MouseEvent.MOUSE_DRAGGED:
	    if (mousePressedTargets.size() > 0) {
		mouseEvent = (MouseEvent) e;
		handleMouseEvent(mouseEvent, mousePressedTargets);
		lastMouseEvent = mouseEvent;
		mouseDragged = true;
	    }
	    break;

	case MouseEvent.MOUSE_RELEASED:
	    if (mousePressedTargets.size() > 0) {
		mouseEvent = (MouseEvent) e;
		handleMouseEvent(mouseEvent, mousePressedTargets);

		if (mouseDragged ||
		    mouseEvent.getModifiers()!=lastMouseEvent.getModifiers()) {
		    // No mouse click if: 
		    //     * the mouse has been dragged;
		    //     * some modifiers have changed;
		    //     * the events are located on different gadgets.

		    mouseReleasedTargets.popAll();
		} else {
		    pushMouseEventTargets(mouseEvent, mouseReleasedTargets);
		    Gadget target = mousePressedTargets.top().gadget;
		    while (mouseReleasedTargets.size() > 0) {
			if (mouseReleasedTargets.top().gadget == target)
			    break;
			mouseReleasedTargets.pop();
		    }
		}

	        mousePressedTargets.popAll();
		lastMouseEvent = null;
		mouseDragged = false;
	    }
	    break;

	case MouseEvent.MOUSE_CLICKED:
	    if (mouseReleasedTargets.size() > 0) {
		mouseEvent = (MouseEvent) e;
		handleMouseEvent(mouseEvent, mouseReleasedTargets);
		mouseReleasedTargets.popAll();
	    }
	    break;
	}
    }

    private void pushMouseEventTargets(MouseEvent e, 
				       MouseEventTargets targets) {
	// Collect gadgets containing the mouse click ---
	targets.popAll();

	point.x = e.getX();
	point.y = e.getY();
	root.panelToGadget(point);

	int eX = point.x;
	int eY = point.y;
	// Root.x and root.y are both null.
	if (eX < 0 || eX >= root.width || eY < 0 || eY >= root.height)
	    return;

	GadgetContainer container = root;
	targets.push(root, eX, eY);

	loop: for (;;) {
	    Gadget[] children = container.children;

	    for (int i = 0; i < children.length; ++i) {
		Gadget child = children[i];

		if (eX >= child.x && eX < child.x+child.width &&
		    eY >= child.y && eY < child.y+child.height) {

		    eX -= child.x;
		    eY -= child.y;
		    targets.push(child, eX, eY);

		    if (child instanceof GadgetContainer) {
			// Search deeper.
			container = (GadgetContainer) child;
			continue loop;
		    } else {
			// Contained in a leaf gadget: stop searching.
			break loop;
		    }
		}
	    }

	    // Contained in a container gadget but not inside any of its
	    // children.
	    break;
	}
    }

    private void dispatchMouseEvent(MouseEvent event, 
				    MouseEventTargets targets) {
	int mouseX = event.getX();
	int mouseY = event.getY();

	while (targets.size() > 0) {
	    point.x = mouseX;
	    point.y = mouseY;
	    targets.panelToGadget(point);
	    event.translatePoint(point.x - mouseX, point.y - mouseY);

	    try {
		if (targets.top().gadget.handleMouseEvent(event)) {
		    event.consume();
		    break;
		}
	    } catch (Exception e) {
		e.printStackTrace();
	    }

	    event.translatePoint(mouseX - point.x, mouseY - point.y);
	    targets.pop();
	}
    }

    private void handleMouseEvent(MouseEvent event, 
				  MouseEventTargets targets) {
	int mouseX = event.getX();
	int mouseY = event.getY();

	point.x = mouseX;
	point.y = mouseY;
	targets.panelToGadget(point);
	event.translatePoint(point.x - mouseX, point.y - mouseY);

	try {
	    targets.top().gadget.handleMouseEvent(event);
	} catch (Exception e) {
	    e.printStackTrace();
	}

	event.translatePoint(mouseX - point.x, mouseY - point.y);
	event.consume();
    }

    // -----------------------------------------------------------------------

    private static final class MouseEventTarget {
	public Gadget gadget;
	public int eventX, eventY;
    }

    private static final class MouseEventTargets {
	private MouseEventTarget[] stack = new MouseEventTarget[0];
	private int pushed = 0;

	public void push(Gadget gadget, int eventX, int eventY) {
	    if (pushed == stack.length) {
		MouseEventTarget[] newStack = 
		    new MouseEventTarget[Math.max(10, 2*stack.length)];

		if (pushed > 0) 
		    System.arraycopy(stack, 0, newStack, 0, pushed);

		for (int i = pushed; i < newStack.length; ++i)
		    newStack[i] = new MouseEventTarget();

		stack = newStack;
	    }

	    MouseEventTarget target = stack[pushed++];
	    target.gadget = gadget;
	    target.eventX = eventX;
	    target.eventY = eventY;
	}

	public int size() {
	    return pushed;
	}

	public MouseEventTarget top() {
	    return (pushed > 0)? stack[pushed-1] : null;
	}

	public void pop() {
	    if (pushed > 0)
		stack[--pushed].gadget = null;
	}

	public void popAll() {
	    for (int i = 0; i < pushed; ++i)
		stack[i].gadget = null;
	    pushed = 0;
	}

	public void panelToGadget(Point point) {
	    if (pushed == 0) 
		return;

	    ((RootGadget) stack[0].gadget).panelToGadget(point);

	    int pX = point.x;
	    int pY = point.y;

	    for (int i = 1; i < pushed; ++i) {
		Gadget gadget = stack[i].gadget;
		pX -= gadget.x;
		pY -= gadget.y;
	    }

	    point.x = pX;
	    point.y = pY;
	}
    }

    // -----------------------------------------------------------------------

    public static final int getPopupTriggerEventModifiers() {
	return popupTriggerEventModifiers;
    }

    public static final int getPopupTriggerEventId() {
	return popupTriggerEventId;
    }

    private static void listPopupTriggerInfo(MouseEvent event) {
	if (event.isPopupTrigger()) {
	    int id = event.getID();
	    int modifiers = event.getModifiers();

	    System.err.print("PopupTrigger ID=" + id);
	    System.err.println(", modifiers=0x" + 
			       Integer.toHexString(modifiers));

	    switch (id) {
	    case MouseEvent.MOUSE_PRESSED:
		System.err.println("\tMOUSE_PRESSED");
		break;
	    case MouseEvent.MOUSE_RELEASED:
		System.err.println("\tMOUSE_RELEASED");
		break;
	    }

	    if ((modifiers & MouseEvent.BUTTON1_MASK) != 0)
		System.err.println("\tBUTTON1_MASK");
	    if ((modifiers & MouseEvent.BUTTON2_MASK) != 0)
		System.err.println("\tBUTTON2_MASK");
	    if ((modifiers & MouseEvent.BUTTON3_MASK) != 0)
		System.err.println("\tBUTTON3_MASK");
	    if ((modifiers & MouseEvent.CTRL_MASK) != 0)
		System.err.println("\tCTRL_MASK");
	    if ((modifiers & MouseEvent.SHIFT_MASK) != 0)
		System.err.println("\tSHIFT_MASK");
	    if ((modifiers & MouseEvent.META_MASK) != 0)
		System.err.println("\tMETA_MASK");
	    if ((modifiers & MouseEvent.ALT_MASK) != 0)
		System.err.println("\tALT_MASK");
	    if ((modifiers & MouseEvent.ALT_GRAPH_MASK) != 0)
		System.err.println("\tALT_GRAPH_MASK");
	}
    }
}


