/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.gadget;

import java.awt.*;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

public abstract class Gadget implements ActionListener {
    public /*read-only*/ GadgetContainer parent = null;
    public /*read-only*/ int x = 0;
    public /*read-only*/ int y = 0;
    public /*read-only*/ int width = 1;
    public /*read-only*/ int height = 1;

    public Point getLocation() {
	Point location = new Point();
	getLocation(location);
	return location;
    }

    public void getLocation(Point location) {
	location.x = x;
	location.y = y;
    }

    public Dimension getSize() {
	Dimension size = new Dimension();
	getSize(size);
	return size;
    }

    public void getSize(Dimension size) {
	size.width = width;
	size.height = height;
    }

    public Rectangle getBounds() {
	Rectangle bounds = new Rectangle();
	getBounds(bounds);
	return bounds;
    }

    public void getBounds(Rectangle bounds) {
	bounds.x = x;
	bounds.y = y;
	bounds.width = width;
	bounds.height = height;
    }

    public RootGadget getRoot() {
	Gadget gadget = this;
	for (;;) {
	    if (gadget.parent == null)
		break;
	    gadget = gadget.parent;
	}

	return (gadget instanceof RootGadget)? (RootGadget) gadget : null;
    }

    public EventDispatcher getEventDispatcher() {
	RootGadget root = getRoot();
	return (root == null)? null : root.getEventDispatcher();
    }

    public Container getPanel() {	
	RootGadget root = getRoot();
	return (root == null)? null : root.getPanel();
    }

    public FontMetrics getFontMetrics(Font font) {	
	Container panel = getPanel();
	return ((panel == null)? 
		Toolkit.getDefaultToolkit().getFontMetrics(font) : 
		panel.getFontMetrics(font));
    }

    public Toolkit getToolkit() {	
	Container panel = getPanel();
	return ((panel == null)? 
		Toolkit.getDefaultToolkit() : panel.getToolkit());
    }

    public Gadget[] getAncestors() {
	int count = 0;
	Gadget gadget = parent;
	while (gadget != null) {
	    ++count;
	    gadget = gadget.parent;
	}

	Gadget[] ancestors = new Gadget[count];

	gadget = parent;
	while (gadget != null) {
	    ancestors[--count] = gadget;
	    gadget = gadget.parent;
	}

	return ancestors;
    }

    public Gadget gadgetToRoot(Point point) {
	int pX = point.x;
	int pY = point.y;
	Gadget gadget = this;
	for (;;) {
	    if (gadget.parent == null)
		break;

	    pX += gadget.x;
	    pY += gadget.y;
	    gadget = gadget.parent;
	}
	point.x = pX;
	point.y = pY;

	return gadget;
    }

    public void gadgetToPanel(Point point) {
	Gadget gadget = gadgetToRoot(point);

	if (gadget instanceof RootGadget) 
	    ((RootGadget) gadget).gadgetToPanel(point);
    }

    public void panelToGadget(Point point) {
	Gadget[] ancestors = getAncestors();
	if (ancestors.length == 0) 
	    // This gadget has no parent yet. 
	    return;

	int i = 0;
	if (ancestors[0] instanceof RootGadget) {
	    ((RootGadget) ancestors[0]).panelToGadget(point);
	    i = 1;
	}

	int pX = point.x;
	int pY = point.y;

	for (; i < ancestors.length; ++i) {
	    Gadget ancestor = ancestors[i];
	    pX -= ancestor.x;
	    pY -= ancestor.y;
	}

	point.x = pX - this.x;
	point.y = pY - this.y;
    }

    public void makeRectangleVisible(int rx, int ry, int rw, int rh) {
	Gadget gadget = this;
	for (;;) {
	    if (gadget.parent == null)
		break;

	    rx += gadget.x;
	    ry += gadget.y;
	    gadget = gadget.parent;
	}

	if (gadget instanceof RootGadget) 
	    ((RootGadget) gadget).makeRectangleVisible(rx, ry, rw, rh);
    }

    // -----------------------------------------------------------------------
    // Layout
    // -----------------------------------------------------------------------

    /**
     * Invoked by the client to request a gadget to relayout and 
     * repaint itself.
     *
     * <p>This default implementation assumes that if, after the relayout,
     * this gadget has not changed its width and height, a repaint is
     * sufficient and its parent doesn't need to be notified.
     */
    public void refresh() {
	int oldWidth = width;
	int oldHeight = height;

	relayout(true);

	if (width == oldWidth && height == oldHeight) {
	    repaint(0, 0, width, height);
	} else {
	    if (parent != null)
		parent.paintRelayout(this, oldWidth, oldHeight, 
				     GadgetContainer.NO_SIDE, -1);
	}
    }

    public void relayout() {
	relayout(true);
    }

    /**
     * Invoked by the client to request a gadget to relayout itself.
     * 
     * <p>To improve performance, non stretchable gadgets may redefine this
     * method as layout(-1, -1, recursive).
     *
     * @param recursive If true, recursively layout children too.  
     */
    public void relayout(boolean recursive) {
	if (parent != null) {
	    Dimension constraints = new Dimension();
	    parent.getSizeConstraints(this, constraints);
	    layout(constraints.width, constraints.height, recursive);
	}
    }

    public void layout(int constrainedWidth, int constrainedHeight) {
	layout(constrainedWidth, constrainedHeight, true);
    }

    /**
     * Invoked by the parent to request its child to relayout itself.
     *
     * <p>Child responds by recomputing and updating its geometry.
     *
     * <p>A stretchable child must remember the constrainedWidth and/or 
     * constrainedHeight values for future relayouts initiated by itself 
     * (for example when its font has been changed).
     * 
     * @param constrainedWidth  A stretchable child might be able to adapt its
     *                          geometry to this constrained width. 
     *                          A negative or null value means not constrained.
     * @param constrainedHeight A stretchable child might be able to adapt its
     *                          geometry to this constrained height. 
     *                          A negative or null value means not constrained.
     * @param recursive         If true, recursively layout children too.
     */
    public abstract void layout(int constrainedWidth, int constrainedHeight,
				boolean recursive);

    // -----------------------------------------------------------------------
    // Paint
    // -----------------------------------------------------------------------

    public void repaint() {
	repaint(0, 0, width, height);
    }

    public void repaint(int x, int y, int with, int height) {
	int x0 = x;
	int y0 = y;
	Gadget gadget = this;
	for (;;) {
	    if (gadget.parent == null)
		break;

	    x0 += gadget.x;
	    y0 += gadget.y;
	    gadget = gadget.parent;
	}

	if (gadget instanceof RootGadget)
	    ((RootGadget) gadget).repaint(x0, y0, width, height);
	// Otherwise this gadget has no parent yet. Do no repaint.
    }

    /**
     * You cannot assume that any of the attributes of the Graphics object
     * (background, foreground, font, etc) is properly set.
     */
    public abstract void paint(Graphics g);

    /**
     * You cannot assume that any of the attributes of the Graphics object
     * (background, foreground, font, etc) is properly set.
     */
    public Graphics getGraphics() {
	int x0 = 0;
	int y0 = 0;
	Gadget gadget = this;
	for (;;) {
	    if (gadget.parent == null)
		break;

	    x0 += gadget.x;
	    y0 += gadget.y;
	    gadget = gadget.parent;
	}

	Graphics g = null;

	if (gadget instanceof RootGadget) {
	    g = ((RootGadget) gadget).getGraphics();
	    g.translate(x0, y0);
	    g.clipRect(0, 0, width, height);
	}
	// Otherwise this gadget has no parent yet. 

	return g;
    }

    public void print(Graphics g) {
	paint(g);
    }

    public int[] getNonBreakableVerticalRanges() {
	return new int[] { 0, height-1 };
    }

    // -----------------------------------------------------------------------
    // Event handling
    // -----------------------------------------------------------------------

    public boolean hasFocus() {
	EventDispatcher eventDispatcher = getEventDispatcher();
	return (eventDispatcher != null && 
		eventDispatcher.getFocusGadget() == this);
    }

    public void requestFocus() {
	EventDispatcher eventDispatcher = getEventDispatcher();
	if (eventDispatcher != null)
	    eventDispatcher.setFocusGadget(this);
    }

    public void focusGained() {}

    public void focusLost() {}

    public boolean handleKeyEvent(KeyEvent e) {
	return false;
    }

    public boolean handleMouseEvent(MouseEvent e) {
	return false;
    }

    public void showMenu(MenuSpec menu, int x, int y) {
	int x0 = x;
	int y0 = y;
	Gadget gadget = this;
	for (;;) {
	    if (gadget.parent == null)
		break;

	    x0 += gadget.x;
	    y0 += gadget.y;
	    gadget = gadget.parent;
	}

	if (gadget instanceof RootGadget)
	    ((RootGadget) gadget).showMenu(menu, x0, y0, this);
	// Otherwise this gadget has no parent yet. Do not show the menu.
    }

    public void actionPerformed(ActionEvent e) {}
}
