/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.gadget;

import java.awt.*;

public abstract class GadgetContainer extends Gadget {
    public static final int NO_SIDE     = 0;
    public static final int TOP_SIDE    = 1;
    public static final int BOTTOM_SIDE = 2;
    public static final int LEFT_SIDE   = 3;
    public static final int RIGHT_SIDE  = 4;

    protected Gadget[] children = new Gadget[0];

    public void addChild(Gadget gadget) {
	insertChild(children.length, gadget);
    }

    public void insertChild(int index, Gadget gadget) {
	Gadget[] newChildren = new Gadget[children.length+1];

	if (index > 0)
	    System.arraycopy(children, 0, newChildren, 0, index);

	int tail = children.length - index;
	if (tail > 0) 
	    System.arraycopy(children, index, newChildren, index+1, tail);

	newChildren[index] = gadget;
	children = newChildren;

	gadget.parent = this;

	Container panel = getPanel();
	if (panel != null) 
	    ((GadgetHost) panel).gadgetAdded(gadget);
    }

    public void removeChild(Gadget gadget) {
	int index = indexOfChild(gadget);
	if (index >= 0)
	    removeChild(index);
    }

    public void removeChild(int index) {
	Gadget gadget = children[index];
	Gadget[] newChildren = new Gadget[children.length-1];

	if (index > 0)
	    System.arraycopy(children, 0, newChildren, 0, index);

	int first = index + 1;
	int tail = children.length - first;
	if (tail > 0) 
	    System.arraycopy(children, first, newChildren, index, tail);

	children = newChildren;

	gadget.parent = null;

	EventDispatcher eventDispatcher = getEventDispatcher();
	if (eventDispatcher != null)
	    eventDispatcher.gadgetRemoved(gadget);

	Container panel = getPanel();
	if (panel != null)
	    ((GadgetHost) panel).gadgetRemoved(gadget);
    }

    public void replaceChild(Gadget oldGadget, Gadget newGadget) {
	replaceChild(indexOfChild(oldGadget), newGadget);
    }

    public void replaceChild(int index, Gadget newGadget) {
	Gadget oldGadget = children[index];
	children[index] = newGadget;

	oldGadget.parent = null;
	newGadget.parent = this;

	EventDispatcher eventDispatcher = getEventDispatcher();
	if (eventDispatcher != null)
	    eventDispatcher.gadgetRemoved(oldGadget);

	GadgetHost panel = (GadgetHost) getPanel();
	if (panel != null) {
	    panel.gadgetRemoved(oldGadget);
	    panel.gadgetAdded(newGadget);
	}
    }

    public Gadget[] getChildren() {
	return children;
    }

    public int indexOfChild(Gadget gadget) {
	for (int i = 0; i < children.length; ++i) {
	    if (children[i] == gadget)
		return i;
	}
	return -1;
    }

    public Gadget getChild(int index) {
	return children[index];
    }

    public int getChildCount() {
	return children.length;
    }

    public boolean isAncestorOf(Gadget gadget) {
	for (int i = 0; i < children.length; ++i) {
	    Gadget child = children[i];

	    if (child == gadget ||
		((child instanceof GadgetContainer) && 
		 ((GadgetContainer) child).isAncestorOf(gadget)))
		return true;
	}

	return false;
    }

    public int indexOfChildAt(int childX, int childY) {
	for (int i = 0; i < children.length; ++i) {
	    Gadget child = children[i];

	    if (childX >= child.x && childX < child.x+child.width &&
		childY >= child.y && childY < child.y+child.height)
		return i;
	}

	return -1;
    }

    public Gadget findGadgetAt(Point point) {
	int childX = point.x;
	int childY = point.y;

	if (childX < 0 || childX >= width || childY < 0 || childY >= height) 
	    return null;
	Gadget target = this;

	GadgetContainer container = this;
	loop: for (;;) {
	    Gadget[] children = container.children;

	    for (int i = 0; i < children.length; ++i) {
		Gadget child = children[i];

		if (childX >= child.x && childX < child.x+child.width &&
		    childY >= child.y && childY < child.y+child.height) {

		    childX -= child.x;
		    childY -= child.y;
		    target = child;

		    if (child instanceof GadgetContainer) {
			// Search deeper.
			container = (GadgetContainer) child;
			continue loop;
		    } else {
			// Contained in a leaf gadget: stop searching.
			break loop;
		    }
		}
	    }

	    // Contained in a container gadget but not inside any of its
	    // children.
	    break;
	}

	point.x = childX;
	point.y = childY;
	return target;
    }

    public abstract void getSizeConstraints(Gadget child, 
					    Dimension constraints);

    public void paintRelayout(Gadget child, int prevWidth, int prevHeight) {
	paintRelayout(child, prevWidth, prevHeight, NO_SIDE, -1);
    }

    /**
     * Invoked by the child when its geometry (width, height, but also for
     * some gadgets baseline, reference point, etc) has changed.
     *
     * <p>Parent must at least respond by repainting its child
     * but generally the parent has to ``refresh'' itself.
     *
     * <p>This default implementation is a kind of light-weight refresh()
     * where the relayout is not recursive. It should be redefined by a 
     * faster, more container specific, version.
     *
     * <p>This default implementation assumes that if, after the relayout,
     * this container has not changed its width and height, a repaint is
     * sufficient and its parent doesn't need to be notified.
     * 
     * @param child 	    The child that invoked the method.
     * @param prevWidth     Child width before the relayout.
     * @param prevHeight    Child height before the relayout.
     * @param damagedSide   The side if any that has expanded or shrinked.
     * @param damagedLength If damagedSide is TOP_SIDE, it is sufficient to
     *                      repaint (0, 0, width, damagedLength).
     *                      If damagedSide is BOTTOM_SIDE, 
     *                      it is sufficient to repaint 
     *                      (0, height-damagedLength, width, damagedLength).
     *                      If damagedSide is LEFT_SIDE, it is sufficient to 
     *                      repaint (0, 0, damagedLength, height).
     *                      If damagedSide is RIGHT_SIDE, 
     *                      it is sufficient to repaint 
     *                      (width-damagedLength, 0, damagedLength, height).
     *                      If damagedSide is NO_SIDE, damagedLength is 
     *                      meaningless.
     */
    public void paintRelayout(Gadget child, int prevWidth, int prevHeight,
			      int damagedSide, int damagedLength) {
	int oldWidth = width;
	int oldHeight = height;

	relayout(false);

	if (width == oldWidth && height == oldHeight) {
	    repaint(0, 0, width, height);
	} else {
	    if (parent != null)
		parent.paintRelayout(this, oldWidth, oldHeight, NO_SIDE, -1);
	}
    }

    public void paint(Graphics g) {
	Rectangle clip = g.getClipBounds();

	int count = children.length;
	for (int i = 0; i < count; ++i) {
	    Gadget child = children[i];

	    if (clip == null || 
		// Child intersects clip.
		(child.x < clip.x+clip.width && 
		 child.y < clip.y+clip.height &&
		 child.x+child.width > clip.x && 
		 child.y+child.height > clip.y)) {

		Graphics g2 = g.create(child.x, child.y, 
				       child.width, child.height);
		try {
		    child.paint(g2);
		} finally {
		    g2.dispose();
		}
	    }
	}
    }

    public void print(Graphics g) {
	Rectangle clip = g.getClipBounds();

	int count = children.length;
	for (int i = 0; i < count; ++i) {
	    Gadget child = children[i];

	    if (clip == null || 
		// Child intersects clip.
		(child.x < clip.x+clip.width && 
		 child.y < clip.y+clip.height &&
		 child.x+child.width > clip.x && 
		 child.y+child.height > clip.y)) {

		Graphics g2 = g.create(child.x, child.y, 
				       child.width, child.height);
		try {
		    child.print(g2);
		} finally {
		    g2.dispose();
		}
	    }
	}
    }

    public int[] getNonBreakableVerticalRanges() {
	// Can break anywhere.
	return null;
    }
}
