/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.gadget;

import java.util.StringTokenizer;
import java.util.Vector;
import java.awt.*;
import java.awt.print.*;
import fr.pixware.util.QuickSort;

public class GadgetPrinter implements Printable {
    private double screenResolution = -1;

    private PrintInProgressObserver observer = null;

    private String headerBegin = null;
    private String headerMiddle = null;
    private String headerEnd = null;
    private Font headerFont = new Font("SansSerif", Font.PLAIN, 12);
    private Color headerColor = Color.black;
    private boolean underlineHeader = false;
    private int headerGap = -1; // auto

    private String footerBegin = null;
    private String footerMiddle = null;
    private String footerEnd = null;
    private Font footerFont = headerFont;
    private Color footerColor = headerColor;
    private boolean overlineFooter = false;
    private int footerGap = -1;

    // Printing state.
    private Gadget printedGadget;
    private Container gadgetPanel;
    private double pageLeft, pageTop;
    private int pageWidth, pageHeight;
    private Dimension savedConstraints = new Dimension();
    private int printedLeft, printedTop, printedWidth, printedHeight;
    private double bodyTop;
    private int bodyMaxHeight;
    private int[] boundaries;

    private static final class Atom {
	public int y1, y2;

	public Atom(int y1, int y2) {
	    this.y1 = y1;
	    this.y2 = y2;
	}
    }

    private static class AtomCompare implements QuickSort.Compare {
	public int compare(Object o1, Object o2) {
	    return ((Atom) o1).y1 - ((Atom) o2).y1;
	}
    }
    private static AtomCompare atomCompare = new AtomCompare();

    // -----------------------------------------------------------------------

    public void setObserver(PrintInProgressObserver observer) {
	this.observer = observer;
    }

    public PrintInProgressObserver getObserver() {
	return observer;
    }

    public void setScreenResolution(double res) {
	screenResolution = res;
    }

    public double getScreenResolution() {
	return screenResolution;
    }

    public void setHeaderBegin(String header) {
	headerBegin = header;
    }

    public String getHeaderBegin() {
	return headerBegin;
    }

    public void setHeaderMiddle(String header) {
	headerMiddle = header;
    }

    public String getHeaderMiddle() {
	return headerMiddle;
    }

    public void setHeaderEnd(String header) {
	headerEnd = header;
    }

    public String getHeaderEnd() {
	return headerEnd;
    }

    public void setHeaderFont(Font font) {
	headerFont = font;
    }

    public Font getHeaderFont() {
	return headerFont;
    }

    public void setHeaderColor(Color color) {
	headerColor = color;
    }

    public Color getHeaderColor() {
	return headerColor;
    }

    public void setHeaderUnderlined(boolean underline) {
	underlineHeader = underline;
    }

    public boolean isHeaderUnderlined() {
	return underlineHeader;
    }

    public void setHeaderGap(int gap) {
	headerGap = gap;
    }

    public int getHeaderGap() {
	return headerGap;
    }

    public void setFooterBegin(String footer) {
	footerBegin = footer;
    }

    public String getFooterBegin() {
	return footerBegin;
    }

    public void setFooterMiddle(String footer) {
	footerMiddle = footer;
    }

    public String getFooterMiddle() {
	return footerMiddle;
    }

    public void setFooterEnd(String footer) {
	footerEnd = footer;
    }

    public String getFooterEnd() {
	return footerEnd;
    }

    public void setFooterFont(Font font) {
	footerFont = font;
    }

    public Font getFooterFont() {
	return footerFont;
    }

    public void setFooterColor(Color color) {
	footerColor = color;
    }

    public Color getFooterColor() {
	return footerColor;
    }

    public void setFooterOverlined(boolean overline) {
	overlineFooter = overline;
    }

    public boolean isFooterOverlined() {
	return overlineFooter;
    }

    public void setFooterGap(int gap) {
	footerGap = gap;
    }

    public int getFooterGap() {
	return footerGap;
    }

    public void begin(Gadget gadget, PageFormat pageFormat) {
	printedGadget = gadget;
	gadgetPanel = gadget.getPanel();

	if (screenResolution <= 0)
	    screenResolution = gadgetPanel.getToolkit().getScreenResolution();

	// Convert pt to pixels.
	pageLeft = ((pageFormat.getImageableX() * screenResolution) / 72);
	pageTop = ((pageFormat.getImageableY() * screenResolution) / 72);
	pageWidth = (int) 
	    ((pageFormat.getImageableWidth() * screenResolution) / 72);
	pageHeight = (int) 
	    ((pageFormat.getImageableHeight() * screenResolution) / 72);

	RootGadget rootGadget = ((GadgetHost) gadgetPanel).getRoot();
	if (rootGadget.getChildCount() > 0) {
	    rootGadget.getSizeConstraints(rootGadget.getChild(0), 
					  savedConstraints);
	    Insets insets = gadgetPanel.getInsets();
	    if (savedConstraints.width > 0)
		savedConstraints.width += (insets.left+insets.right);
	    if (savedConstraints.height > 0)
		savedConstraints.height += (insets.top+insets.bottom);

	    // Insets are not printed.
	    rootGadget.layout(pageWidth + (insets.left+insets.right), -1);
	    // Needed to move embedded Components.
	    rootGadget.repaint();
	}

	printedWidth = printedGadget.width;
	printedHeight = printedGadget.height;
	Point origin = new Point(0, 0);
	printedGadget.gadgetToPanel(origin);
	printedLeft = origin.x;
	printedTop = origin.y;

	splitInPages();

	// Normally this is not useful because we use print not paint.
	javax.swing.RepaintManager.currentManager(
	    gadgetPanel).setDoubleBufferingEnabled(false);

	if (observer != null)
	    observer.printInProgress(this, -1, boundaries.length);
    }

    private void splitInPages() {
	bodyTop = pageTop;
	bodyMaxHeight = pageHeight;

	if (headerBegin != null || headerMiddle != null || headerEnd != null) {
	    FontMetrics metrics = gadgetPanel.getFontMetrics(headerFont);
	    int lineHeight = lineHeight(metrics);

	    bodyMaxHeight -= lineHeight;
	    bodyTop += lineHeight;

	    if (headerGap <= 0)
		headerGap = lineHeight;
	    bodyMaxHeight -= headerGap;
	    bodyTop += headerGap;
	}

	if (footerBegin != null || footerMiddle != null || footerEnd != null) {
	    FontMetrics metrics = gadgetPanel.getFontMetrics(footerFont);
	    int lineHeight = lineHeight(metrics);

	    bodyMaxHeight -= lineHeight;

	    if (footerGap <= 0)
		footerGap = lineHeight;
	    bodyMaxHeight -= footerGap;
	}

	computeBoundaries();
    }

    private static final int lineHeight(FontMetrics metrics) {
	return (int) Math.rint(metrics.getHeight()*1.5);
    }

    private void computeBoundaries() {
	Atom[] atoms = groupAtoms(listAtoms());

	int[] yList = new int[100];
	int yListSize = 0;

	int y = printedTop;
	int endY = printedTop + printedHeight;
	do {
	    yList = addInt(yList, yListSize, y);
	    ++yListSize;

	    int maxY = y + bodyMaxHeight;
	    if (maxY >= endY) {
		y = endY;
	    } else {
		y = checkBoundary(maxY, atoms, y);
		if (y < 0)
		    y = maxY;
	    }
	 } while (y < endY);

	boundaries = new int[yListSize];
	System.arraycopy(yList, 0, boundaries, 0, yListSize);
    }

    private static final int checkBoundary(int y, Atom[] atoms, int boundary) {
	Atom splitAtom = findSplitAtom(atoms, y);
	if (splitAtom == null)
	    return y;
	
	y = splitAtom.y1;
	if (y <= boundary)
	    return -1;
	else
	    return y;
    }

    private static final Atom findSplitAtom(Atom[] atoms, int y) {
	int atomCount = atoms.length;
	for (int i = 0; i < atomCount; ++i) {
	    Atom atom = atoms[i];

	    if (atom.y1 >= y)
		return null;

	    if (y > atom.y1 && y <= atom.y2)
		return atom;
	}

	return null;
    }

    private static final int[] addInt(int[] list, int listSize, int item) {
	if (listSize >= list.length) {
	    int[] newList = new int[Math.max(10, 2*list.length)];
	    System.arraycopy(list, 0, newList, 0, list.length);
	    list = newList;
	}

	list[listSize] = item;
	return list;
    }

    private Atom[] listAtoms() {
	Vector atoms = new Vector();

	listAtoms(printedGadget, printedTop, atoms);

	Atom[] list = new Atom[atoms.size()];
	atoms.copyInto(list);
	QuickSort.sort(list, atomCompare);

	return list;
    }

    private void listAtoms(Gadget gadget, int y0, Vector atoms) {
	int[] ranges = gadget.getNonBreakableVerticalRanges();
	if (ranges != null) {
	    int count = (ranges.length/2)*2;

	    for (int i = 0; i < count; i += 2) {
		int y1 = ranges[i];
		int y2 = ranges[i+1];

		if (y2 >= y1)
		    atoms.addElement(new Atom(y0+y1, y0+y2));
	    }
	}

	if (gadget instanceof GadgetContainer) {
	    Gadget[] children = ((GadgetContainer) gadget).getChildren();

	    for (int i = 0; i < children.length; ++i) {
		Gadget child = children[i];
		listAtoms(child, y0+child.y, atoms);
	    }
	}
    }

    private Atom[] groupAtoms(Atom[] atoms) {
	Vector groups = new Vector();

	Atom group = null;
	for (int i = 0; i < atoms.length; ++i) {
	    Atom atom = atoms[i];

	    if (group == null || atom.y1 > group.y2) {
		group = new Atom(atom.y1, atom.y2);
		groups.addElement(group);
	    } else {
		group.y2 = Math.max(atom.y2, group.y2);
	    }
	}

	Atom[] list = new Atom[groups.size()];
	groups.copyInto(list);

	return list;
    }

    public int print(Graphics graphics, PageFormat pageFormat, int pageIndex)
	throws PrinterException {
	if (pageIndex >= boundaries.length)
	    return NO_SUCH_PAGE;

	if (observer != null) {
	    if (!observer.printInProgress(this, pageIndex, boundaries.length))
		throw new PrintCanceledException(this);
	}

	Graphics2D g2 = (Graphics2D) graphics;
	// This will map pixels to pt.
	double scale = 72.0/screenResolution;
	g2.scale(scale, scale);
	g2.setClip((int) pageLeft, (int) pageTop, pageWidth, pageHeight);
	g2.translate(pageLeft, 0);

	if (headerBegin != null || headerMiddle != null || headerEnd != null) {
	    g2.setColor(headerColor);

	    g2.setFont(headerFont);
	    FontMetrics metrics = g2.getFontMetrics();
	    int ascent = metrics.getAscent();

	    g2.translate(0, pageTop);

	    if (headerBegin != null) {
		String text = 
		    substituteVars(headerBegin, pageIndex, boundaries.length);
		g2.drawString(text, 0, ascent);
	    }

	    if (headerMiddle != null) {
		String text = 
		   substituteVars(headerMiddle, pageIndex, boundaries.length);
		g2.drawString(text, (pageWidth - metrics.stringWidth(text))/2,
			      ascent);
	    }

	    if (headerEnd != null) {
		String text = 
		    substituteVars(headerEnd, pageIndex, boundaries.length);
		g2.drawString(text, pageWidth - metrics.stringWidth(text), 
			      ascent);
	    }

	    if (underlineHeader) {
		int y = lineHeight(metrics)-1;
		g2.drawLine(0, y, pageWidth-1, y);
	    }
		
	    g2.translate(0, -pageTop);
	}

	if (footerBegin != null || footerMiddle != null || footerEnd != null) {
	    g2.setColor(footerColor);

	    g2.setFont(footerFont);
	    FontMetrics metrics = g2.getFontMetrics();
	    int lineHeight = lineHeight(metrics);
	    // ``lineHeight - (ascent+descent) + ascent''
	    int baseLine = lineHeight - metrics.getDescent();

	    g2.translate(0, pageTop + pageHeight - lineHeight);

	    if (overlineFooter)
		g2.drawLine(0, 0, pageWidth-1, 0);

	    if (footerBegin != null) {
		String text = 
		    substituteVars(footerBegin, pageIndex, boundaries.length);
		g2.drawString(text, 0, baseLine);
	    }

	    if (footerMiddle != null) {
		String text = 
		   substituteVars(footerMiddle, pageIndex, boundaries.length);
		g2.drawString(text, (pageWidth - metrics.stringWidth(text))/2,
			      baseLine);
	    }

	    if (footerEnd != null) {
		String text = 
		    substituteVars(footerEnd, pageIndex, boundaries.length);
		g2.drawString(text, pageWidth - metrics.stringWidth(text), 
			      baseLine);
	    }

	    g2.translate(0, -(pageTop + pageHeight - lineHeight));
	}

	int splitY = boundaries[pageIndex];
	g2.translate(-printedLeft, bodyTop - splitY);

	int bodyHeight = ((pageIndex+1 < boundaries.length)? 
			  boundaries[pageIndex+1] - splitY : 
			  (printedTop + printedHeight) - splitY);
	graphics.setClip(printedLeft, splitY, printedWidth, bodyHeight);

	gadgetPanel.print(graphics);

	return PAGE_EXISTS;
    }

    private static final String substituteVars(String text, 
					       int pageIndex, int pageCount) {
	if (text.indexOf('%') < 0)
	    return text;

	StringBuffer result = new StringBuffer();

	StringTokenizer tokens = new StringTokenizer(text, "%", true);
	boolean isVar = false;

	while (tokens.hasMoreTokens()) {
	    String token = tokens.nextToken();
	    int tokenLength = token.length();
	    char c = token.charAt(0);

	    switch (c) {
	    case '%':
		if (isVar) {
		    isVar = false;
		    result.append('%');
		} else {
		    isVar = true;
		}
		break;

	    case 'I':
	    case 'C':
		if (isVar) {
		    isVar = false;

		    String value;
		    switch (c) {
		    case 'I':
			value = Integer.toString(1+pageIndex);
			break;
		    case 'C':
			value = Integer.toString(pageCount);
			break;
		    default:
			value = "???";
		    }

		    result.append(value);
		    if (tokenLength > 1)
			result.append(token.substring(1));
		} else {
		    result.append(token);
		}
		break;

	    default:
		if (isVar) {
		    isVar = false;
		    result.append('%');
		}
		result.append(token);
	    }
	}
	if (isVar)
	    result.append('%');

	return result.toString();
    }

    public void end() {
	javax.swing.RepaintManager.currentManager(
	    gadgetPanel).setDoubleBufferingEnabled(true);

	RootGadget rootGadget = ((GadgetHost) gadgetPanel).getRoot();
	if (rootGadget.getChildCount() > 0) {
	    rootGadget.layout(savedConstraints.width, savedConstraints.height);
	    rootGadget.repaint();
	}

	if (observer != null)
	    observer.printInProgress(this, boundaries.length, 
				     boundaries.length);

	boundaries = null;
    }
}
