/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.gadget;

import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.JPanel;
import javax.swing.RepaintManager;
import javax.swing.JPopupMenu;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.event.PopupMenuListener;
import javax.swing.event.PopupMenuEvent;

public class JGadgetPanel extends JPanel implements GadgetHost {
    protected RootGadget root;
    protected EventDispatcher eventDispatcher;
    
    private boolean dispatchEvents = true;
    private ComponentWrapper[] componentWrappers = new ComponentWrapper[0];

    private JPopupMenu popupMenu = null;
    private boolean popupMenuPopped = false;
    private boolean popupMenuHidden = true;

    // Scratch variables.
    private Insets insets = new Insets(0, 0, 0, 0);
    private Rectangle rect = new Rectangle();
    private Point point = new Point();

    // -----------------------------------------------------------------------

    public JGadgetPanel() {
	super(/*layout manager*/ null, /*doubleBuffered*/ true);

	root = createRootGadget();
	eventDispatcher = root.getEventDispatcher();

	enableEvents(AWTEvent.FOCUS_EVENT_MASK|
		     AWTEvent.KEY_EVENT_MASK|
		     AWTEvent.MOUSE_EVENT_MASK|
		     AWTEvent.MOUSE_MOTION_EVENT_MASK);
    }

    protected RootGadget createRootGadget() {
	return new RootGadget(this);
    }

    public RootGadget getRoot() {
	return root;
    }

    public Gadget findGadgetAt(int x, int y) {
	point.x = x;
	point.y = y;
	return findGadgetAt(point);
    }

    public Gadget findGadgetAt(Point point) {
	Insets insets = getInsets();

	point.x -= insets.left;
	point.y -= insets.top;
	return root.findGadgetAt(point);
    }

    // -----------------------------------------------------------------------
    // Component
    // -----------------------------------------------------------------------

    public void setEnabled(boolean enabled) {
	super.setEnabled(enabled);
	dispatchEvents = enabled;
    }

    public Insets getInsets() {
	// Here we try not to create a new Insets object each time getInsets()
	// is called.
	return super.getInsets(insets);
    }

    public Dimension getMinimumSize() {
	return getPreferredSize();
    }
    
    public Dimension getMaximumSize() {
	return getPreferredSize();
    }
    
    public Dimension getPreferredSize() {
	Dimension dim = new Dimension();
	root.getSize(dim);

	Insets insets = getInsets();
	dim.width += insets.left + insets.right;
	dim.height += insets.top + insets.bottom;
	return dim;
    }

    protected void paintComponent(Graphics g) {
	// In Swing, update() has been redefined in order NOT to clear the
	// background. Super.paintComponent() calls PanelUI.update() which
	// clears the background.
	super.paintComponent(g);

	Insets insets = getInsets();
	g.translate(insets.left, insets.top);

	root.paint(g);

	// The root has not modified g. It has just copied it to paint its
	// children.
	g.translate(-insets.left, -insets.top);
    }

    protected void paintChildren(Graphics g) {
	Rectangle clip = g.getClipBounds();

	int count = componentWrappers.length;
	for (int i = 0; i < count; ++i) {
	    ComponentWrapper wrapper = componentWrappers[i];
	    wrapper.getComponent().getBounds(rect);

	    if (clip == null || 
		// Possibly obsolete component position intersects clip.
		(rect.x < clip.x+clip.width && 
		 rect.y < clip.y+clip.height &&
		 rect.x+rect.width > clip.x && 
		 rect.y+rect.height > clip.y))
		wrapper.updateComponentLocation();
	}

	super.paintChildren(g);
    }

    protected void printComponent(Graphics g) {
	Insets insets = getInsets();
	g.translate(insets.left, insets.top);

	root.print(g);

	g.translate(-insets.left, -insets.top);
    }

    protected void processEvent(AWTEvent e) {
	// A displayed PopupMenu ``grabs'' all InputEvents.
	if (dispatchEvents && (popupMenuHidden || !(e instanceof InputEvent)))
	    eventDispatcher.dispatchEvent(e);

	// Calling super.processEvent() just after a PopupMenu is displayed
	// means immediatly canceling it!
	if (popupMenuPopped)
	    popupMenuPopped = false;
	else
	    super.processEvent(e);
    }

    // -----------------------------------------------------------------------
    // GadgetHost
    // -----------------------------------------------------------------------

    public void gadgetAdded(Gadget gadget) {
	if (gadget instanceof ComponentWrapper) {
	    addComponentWrapper((ComponentWrapper) gadget);
	    return;
	}

	if (gadget instanceof GadgetContainer) {
	    Gadget[] children = ((GadgetContainer) gadget).getChildren();
	    for (int i = 0; i < children.length; ++i)
		gadgetAdded(children[i]);
	}
    }

    public void gadgetRemoved(Gadget gadget) {
	if (gadget instanceof ComponentWrapper) {
	    removeComponentWrapper((ComponentWrapper) gadget);
	    return;
	}

	if (gadget instanceof GadgetContainer) {
	    Gadget[] children = ((GadgetContainer) gadget).getChildren();
	    for (int i = 0; i < children.length; ++i)
		gadgetRemoved(children[i]);
	}
    }

    private void addComponentWrapper(ComponentWrapper wrapper) {
	ComponentWrapper[] newWrappers = 
	    new ComponentWrapper[componentWrappers.length+1];

	if (componentWrappers.length > 0)
	    System.arraycopy(componentWrappers, 0, 
			     newWrappers, 0, componentWrappers.length);
	newWrappers[componentWrappers.length] = wrapper;

	componentWrappers = newWrappers;

	add(wrapper.getComponent());
    }

    private void removeComponentWrapper(ComponentWrapper wrapper) {
	int index;
	for (index = 0; index < componentWrappers.length; ++index) {
	    if (componentWrappers[index] == wrapper)
		break;
	}
	if (index == componentWrappers.length)
	    return;

	ComponentWrapper[] newWrappers = 
	    new ComponentWrapper[componentWrappers.length-1];

	if (index > 0)
	    System.arraycopy(componentWrappers, 0, newWrappers, 0, index);

	int first = index + 1;
	int tail = componentWrappers.length - first;
	if (tail > 0) 
	    System.arraycopy(componentWrappers, first, 
			     newWrappers, index, tail);

	componentWrappers = newWrappers;

	remove(wrapper.getComponent());
    }

    public void getVisibleRectangle(Rectangle visibleRect) {
	computeVisibleRect(visibleRect);
    }    

    public void makeRectangleVisible(int rx, int ry, int rw, int rh) {
	rect.x = rx;
	rect.y = ry;
	rect.width = rw;
	rect.height = rh;
	scrollRectToVisible(rect);
    }

    public void showMenu(MenuSpec menu, int x, int y, ActionListener handler) {
	JPopupMenu popup = getPopupMenu();

	// The top menu label is ignored.

	MenuItemSpec[] items = menu.items;
	for (int i = 0; i < items.length; ++i) {
	    MenuItemSpec item = items[i];

	    if (item instanceof MenuSpec) {
		JMenu subMenu = new JMenu(item.label);
		popup.add(subMenu);
		subMenu.setEnabled(item.enabled);

		addMenuItems(subMenu, (MenuSpec) item, handler);
	    } else {
		if (item.label == null) {
		    popup.addSeparator();
		} else {
		    JMenuItem menuItem = popup.add(item.label);
		    menuItem.setEnabled(item.enabled);
		    if (item.command != null) {
			menuItem.setActionCommand(item.command);
			menuItem.addActionListener(handler);
		    }
		}
	    }
	}

	showPopupMenu(x, y);
    }

    private static void addMenuItems(JMenu popup, MenuSpec menu, 
				     ActionListener handler) {
	MenuItemSpec[] items = menu.items;
	for (int i = 0; i < items.length; ++i) {
	    MenuItemSpec item = items[i];

	    if (item instanceof MenuSpec) {
		JMenu subMenu = new JMenu(item.label);
		popup.add(subMenu);
		subMenu.setEnabled(item.enabled);

		addMenuItems(subMenu, (MenuSpec) item, handler);
	    } else {
		if (item.label == null) {
		    popup.addSeparator();
		} else {
		    JMenuItem menuItem = popup.add(item.label);
		    menuItem.setEnabled(item.enabled);
		    if (item.command != null) {
			menuItem.setActionCommand(item.command);
			menuItem.addActionListener(handler);
		    }
		}
	    }
	}
    }

    protected JPopupMenu getPopupMenu() {
	if (popupMenu == null) {
	    popupMenu = new JPopupMenu();
	    popupMenu.addPopupMenuListener(new PopupMenuListener() {
		public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
		    JGadgetPanel.this.popupMenuHidden = false;
		}

		public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
		    JGadgetPanel.this.popupMenuHidden = true;
		}

		public void popupMenuCanceled(PopupMenuEvent e) {}
	    });
	} else {
	    popupMenu.removeAll();
	}

	return popupMenu;
    }

    protected void showPopupMenu(int x, int y) {
	popupMenu.show(this, x, y);
	popupMenuPopped = true;
    }
}
