/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.gadget;

import java.awt.*;
import java.awt.event.ActionListener;
import java.lang.reflect.Method;

public class RootGadget extends GadgetContainer {
    protected Container panel;
    protected EventDispatcher eventDispatcher;
    protected int constrainedWidth = -1;
    protected int constrainedHeight = -1;

    // Scratch variables.
    private Rectangle r1 = new Rectangle();
    private Rectangle r2 = new Rectangle();
    private Rectangle r3 = new Rectangle();
    
    public RootGadget(Container panel) {
	this.panel = panel;
	this.eventDispatcher = new EventDispatcher(this);
    }

    public Container getPanel() {
	return panel;
    }

    public EventDispatcher getEventDispatcher() {
	return eventDispatcher;
    }

    public void gadgetToPanel(Point point) {
	Insets insets = panel.getInsets();
	point.x += insets.left;
	point.y += insets.top;
    }

    public void panelToGadget(Point point) {
	Insets insets = panel.getInsets();
	point.x -= insets.left;
	point.y -= insets.top;
    }

    public Rectangle getVisibleRectangle() {
	((GadgetHost) panel).getVisibleRectangle(r1);

	Insets insets = panel.getInsets();
	r1.x -= insets.left;
	r1.y -= insets.top;

	r2.x = 0;
	r2.y = 0;
	r2.width = width;
	r2.height = height;

	intersection(r1, r2, r3);

	return r3;
    }    

    public void makeRectangleVisible(int rx, int ry, int rw, int rh) {
	Insets insets = panel.getInsets();
	((GadgetHost) panel).makeRectangleVisible(
	    rx+insets.left, ry+insets.top, rw, rh);
    }

    public void insertChild(int index, Gadget gadget) {
	if (children.length > 0)
	    removeChild(0);
	super.insertChild(0, gadget);
    }

    public void refresh() {
	int oldWidth = width;
	int oldHeight = height;

	layout(constrainedWidth, constrainedHeight, true);

	// If the gadget has shrinked, a part of its old area needs 
	// to be cleared.
	int w = (width > oldWidth)? width : oldWidth;
	int h = (height > oldHeight)? height : oldHeight;
	repaint(0, 0, w, h);
    }

    public void relayout(boolean recursive) {
	layout(constrainedWidth, constrainedHeight, recursive);
    }

    public void getSizeConstraints(Gadget child, Dimension constraints) {
	Insets insets = panel.getInsets();

	if (constrainedWidth > 0)
	    constraints.width = constrainedWidth-(insets.left+insets.right);
	else
	    constraints.width = -1;

	if (constrainedHeight > 0)
	    constraints.height = constrainedHeight-(insets.top+insets.bottom);
	else
	    constraints.height = -1;
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	int oldWidth = width;
	int oldHeight = height;

	this.constrainedWidth = constrainedWidth;
	this.constrainedHeight = constrainedHeight;

	if (children.length == 0) {
	    width = 1;
	    height = 1;
	} else {
	    Gadget child = children[0];

	    if (recursive) {
		Insets insets = panel.getInsets();
		if (constrainedWidth > 0)
		    constrainedWidth -= insets.left+insets.right;
		if (constrainedHeight > 0)
		    constrainedHeight -= insets.top+insets.bottom;

		child.layout(constrainedWidth, constrainedHeight, true);
	    }

	    width = child.width;
	    height = child.height;
	    child.x = child.y = 0;
	}

	if (width != oldWidth || height != oldHeight)
	    ((GadgetHost) panel).revalidate();
    }

    public void paintRelayout(Gadget child, int prevWidth, int prevHeight,
			      int damagedSide, int damagedLength) {
	// Simpler, non recursive, layout(): [1.], [2.].
	int oldWidth = width;
	int oldHeight = height;

	// [1.] No need to move child.
	width = child.width;
	height = child.height;

	// [2.] Dimensions have almost certainly changed so always revalidate.
	((GadgetHost) panel).revalidate();

	// If the gadget has shrinked, a part of its old area needs 
	// to be cleared.
	int w = (width > oldWidth)? width : oldWidth;
	int h = (height > oldHeight)? height : oldHeight;

	switch (damagedSide) {
	case RIGHT_SIDE:
	    repaint(width-damagedLength, 0, damagedLength+(w-width), h);
	    break;
	case BOTTOM_SIDE:
	    repaint(0, height-damagedLength, w, damagedLength+(h-height));
	    break;
	default:
	    repaint(0, 0, w, h);
	}

//System.out.println("" + damagedSide + "," + damagedLength);
    }

    public void repaint(int x, int y, int width, int height) {
	Insets insets = panel.getInsets();
	int x0 = insets.left;
	int y0 = insets.top;

	r1.x = x0+x;
	r1.y = y0+y;
	r1.width = width;
	r1.height = height;

	((GadgetHost) panel).getVisibleRectangle(r2);

	intersection(r1, r2, r3);

	if (r3.width > 0) 
	    panel.repaint(r3.x, r3.y, r3.width, r3.height);
    }

    protected static final void intersection(Rectangle r1, Rectangle r2, 
					     Rectangle inter) {
	int x1 = Math.max(r1.x, r2.x);
	int x2 = Math.min(r1.x + r1.width, r2.x + r2.width);
	int y1 = Math.max(r1.y, r2.y);
	int y2 = Math.min(r1.y + r1.height, r2.y + r2.height);
        if ((x2 - x1) < 0 || (y2 - y1) < 0) {
            // Width or height is negative: no intersection.
            inter.x = 0;
            inter.y = 0;
            inter.width = 0;
            inter.height = 0;
        } else {
            inter.x = x1;
            inter.y = y1;
            inter.width = x2 - x1;
            inter.height = y2 - y1;
	}
    }

    public Graphics getGraphics() {
	Insets insets = panel.getInsets();
	int x0 = insets.left;
	int y0 = insets.top;

	Graphics g = panel.getGraphics();
	if (g != null) {
	    g.translate(x0, y0);
	    g.clipRect(0, 0, width, height);
	}
	return g;
    }

    public void showMenu(MenuSpec menu, int x, int y, ActionListener handler) {
	Insets insets = panel.getInsets();
	int x0 = insets.left;
	int y0 = insets.top;

	((GadgetHost) panel).showMenu(menu, x0+x, y0+y, handler);
    }
}
