/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */

import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.xmlmind.awtutil.AWTUtil;
import com.xmlmind.gadget.*;

public class TestContainer extends GadgetContainer {
    private Color background = null;
    private int marginWidth = 5;
    private int marginHeight = 5;

    public TestContainer() {
	relayout();
    }

    public void setBackground(Color color) {
	background = color;
    }

    public Color getBackground() {
	return background;
    }

    public void setMarginWidth(int margin) {
	marginWidth = margin;
    }

    public int getMarginWidth() {
	return marginWidth;
    }

    public void setMarginHeight(int margin) {
	marginHeight = margin;
    }

    public int getMarginHeight() {
	return marginHeight;
    }

    public void getSizeConstraints(Gadget child, Dimension constraints) {
	if (parent != null) {
	    parent.getSizeConstraints(this, constraints);
	    if (constraints.width > 0)
		constraints.width -= 2*marginWidth;
	    constraints.height = -1;
	}
    }

    public void layout(int constrainedWidth, int constrainedHeight, 
		       boolean recursive) {
	if (children.length == 0) {
	    width = 2*marginWidth;
	    height = 2*marginHeight;
	} else {
	    if (constrainedWidth > 0)
		constrainedWidth -= 2*marginWidth;

	    width = 0;
	    height = marginHeight;

	    for (int i = 0; i < children.length; ++i) {
		Gadget child = children[i];

		if (recursive)
		    child.layout(constrainedWidth, -1, true);

		child.x = marginWidth;
		child.y = height;

		if (child.width > width)
		    width = child.width;
		height += child.height + marginHeight;
	    }

	    width += 2*marginWidth;
	}

	if (width == 0)
	    width = 1;
	if (height == 0)
	    height = 1;
    }

    public void paintRelayout(Gadget child, int prevWidth, int prevHeight,
			      int damagedSide, int damagedLength) {

	if (child.height == prevHeight &&
	    2*marginWidth + prevWidth < width && 
	    2*marginWidth + child.width <= width) {
	    // Same height and the new width cannot have changed the width of
	    // the container: just repaint.

	    int fromX;
	    if (damagedSide == RIGHT_SIDE)
		fromX = marginWidth + child.width - damagedLength;
	    else
		fromX = marginWidth;
	    repaint(fromX, child.y, width-fromX, child.height);

//System.out.println("repaint " + fromX);
	    return;
	}

	if (child.width == prevWidth) {
	    // Same width: optimized relayout.

	    int delta = child.height - prevHeight;

	    for (int i = indexOfChild(child)+1; i < children.length; ++i) {
		children[i].y += delta;
	    }
		
	    height += delta;

	    int fromY;
	    if (damagedSide == BOTTOM_SIDE)
		fromY = child.y + child.height - damagedLength;
	    else
		fromY = child.y;
	    parent.paintRelayout(this, width, height-delta, 
				 BOTTOM_SIDE, height-fromY);

//System.out.println("paintRelayout " + fromY);
	    return;
	}
	
	// Width must probably be updated: non-optimized paintRelayout.

	super.paintRelayout(child, prevWidth, prevHeight,
			    damagedSide, damagedLength);

//System.out.println("super.paintRelayout");
    }

    public void paint(Graphics g) {
	if (background != null) {
	    g.setColor(background);
	    g.fillRect(0, 0, width, height);
	}

	super.paint(g);
    }

    public boolean handleMouseEvent(MouseEvent e) {
	int modifiers = e.getModifiers();
	boolean captured = true;

	if (modifiers == (MouseEvent.SHIFT_MASK|MouseEvent.BUTTON1_MASK) ||
	    modifiers == MouseEvent.BUTTON1_MASK) {

	    if (e.getID() == MouseEvent.MOUSE_CLICKED) 
	        addChildAt(e.getX(), e.getY(), /*before*/ true);

	} else if (modifiers == (AWTUtil.MOUSE_MODIFIER|
				 MouseEvent.BUTTON1_MASK)) {

	    if (e.getID() == MouseEvent.MOUSE_CLICKED) 
	        removeChildAt(e.getX(), e.getY());

	} else if (modifiers == (AWTUtil.MOUSE_MODIFIER|MouseEvent.SHIFT_MASK|
				 MouseEvent.BUTTON1_MASK)) {

	    if (e.getID() == MouseEvent.MOUSE_CLICKED) 
	        addChildAt(e.getX(), e.getY(), /*before*/ false);

	} else {
	    captured = false;
	}

	return captured;
    }

    private void addChildAt(int childX, int childY, boolean before) {
	if (children.length == 0) {
	    addChild(new TestGadget());
	    refresh();
	    return;
	}

	int childIndex = indexOfChildAt(childX, childY);
	if (childIndex < 0) {
	    getToolkit().beep();
	    return;
	}

	insertChild((before? childIndex : childIndex+1), new TestGadget());
	refresh();
    }

    private void removeChildAt(int childX, int childY) {
	int childIndex = indexOfChildAt(childX, childY);
	if (childIndex < 0) {
	    getToolkit().beep();
	    return;
	}

	removeChild(childIndex);
	refresh();
    }

    // -----------------------------------------------------------------------

    public static void main(String[] args) {
	JFrame frame = new JFrame();
   	frame.addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        System.exit(0);
   	    }
   	});

	JGadgetPanel panel = new JGadgetPanel();
	final RootGadget root = panel.getRoot();

	TestContainer container = new TestContainer();
	container.setBackground(Color.blue);

	for (int i = 1; i <= 9; ++i)
	    container.addChild(new TestGadget("Initial child #" + i +
					      ".\nClick to type."));

	TestStretch label = 
	    new TestStretch(TestStretch.sample1 + TestStretch.sample2);
	label.setBackground(Color.lightGray);
	label.setForeground(Color.blue);
	container.addChild(label);

	JButton button = new JButton("Stretchable");
	ComponentWrapper wrapper = new ComponentWrapper(button);
	wrapper.setXStretchable(true);
	wrapper.setYStretchable(true);
	container.addChild(wrapper);

	button = new JButton("Not stretchable");
	wrapper = new ComponentWrapper(button);
	wrapper.setXStretchable(false);
	wrapper.setYStretchable(false);
	container.addChild(wrapper);

	root.addChild(container);
	root.refresh();

	JScrollPane scroller = new JScrollPane(panel);
	frame.getContentPane().add(scroller);

	final JViewport viewport = scroller.getViewport();
	viewport.addComponentListener(new ComponentAdapter() {
	    private Dimension dim = new Dimension();

	    public void componentResized(ComponentEvent e) {
		viewport.getSize(dim);
		root.layout(dim.width, dim.height);
		root.repaint();
	    }
	});

	frame.setBounds(0, 0, 200, 200);
	frame.show();
    }
}

