/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */

import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import com.xmlmind.gadget.*;

public class TestGadget extends Gadget {
    private static final int FIRST_KEY_CHAR1 = 0x0020;
    private static final int LAST_KEY_CHAR1 = 0x007E;
    private static final int FIRST_KEY_CHAR2 = 0x00A0;
    private static final int LAST_KEY_CHAR2 = 0x00FF;

    private Vector lines = new Vector();
    private Color background = Color.white;
    private Color foreground = Color.black;
    private Font font = new Font("SansSerif", Font.PLAIN, 12);
    private boolean hasFocus = false;

    public TestGadget() {
	this("");
    }

    public TestGadget(String string) {
	setText(string);
	relayout();
    }

    public void setText(String string) {
	lines.removeAllElements();

	int sep = 0;
	int nextSep;
	while ((nextSep = string.indexOf('\n', sep)) >= 0) {
	    String line = 
		(sep == nextSep)? "" : string.substring(sep, nextSep);

	    lines.addElement(new StringBuffer(line));
	    sep = nextSep + 1;
	} 
	lines.addElement(new StringBuffer(string.substring(sep)));
    }

    public void setBackground(Color color) {
	background = color;
    }

    public Color getBackground() {
	return background;
    }

    public void setForeground(Color color) {
	foreground = color;
    }

    public Color getForeground() {
	return foreground;
    }

    public void setFont(Font font) {
	this.font = font;
    }

    public Font getFont() {
	return font;
    }

    public void relayout(boolean recursive) {
	layout(-1, -1, recursive);
    }

    public void layout(int constrainedWidth, int constrainedHeight, 
		       boolean recursive) {
	FontMetrics metrics = getFontMetrics(font);
	int fontHeight = metrics.getAscent() + metrics.getDescent();
	int lineSkip = metrics.getLeading();
	int lineCount = lines.size();

	height = lineCount*fontHeight + (lineCount-1)*lineSkip;
	width = 0;
	for (int i = 0; i < lineCount; ++i) {
	    String line = ((StringBuffer) lines.elementAt(i)).toString();
	    int w = metrics.stringWidth(line);
	    if (w > width)
		width = w;
	}

	width += 4;
	height += 4;
    }

    public void paint(Graphics g) {
	FontMetrics metrics = getFontMetrics(font);
	int fontAscent = metrics.getAscent();
	int fontDescent = metrics.getDescent();
	int lineHeight = metrics.getHeight();
	int lineCount = lines.size();

	g.setColor(background);
	g.fillRect(0, 0, width, height);

	g.setColor(foreground);
	g.setFont(font);

	int y0 = 2 + fontAscent;
	for (int i = 0; i < lineCount; ++i) {
	    String line = ((StringBuffer) lines.elementAt(i)).toString();
	    g.drawString(line, 2, y0);

	    if (hasFocus && i == lineCount-1) {
		int caretX = 2 + metrics.stringWidth(line);
		g.drawLine(caretX, y0-fontAscent, caretX, y0+fontDescent-1);
	    }

	    y0 += lineHeight;
	}
    }

    public void focusGained() {
	hasFocus = true;
	repaint();
	makeCaretVisible(null, 0);
    }

    public void focusLost() {
	hasFocus = false;
	repaint();
    }

    public boolean handleKeyEvent(KeyEvent e) {
	boolean handled = false;

	switch (e.getID()) {
	case KeyEvent.KEY_PRESSED:
	    if (e.getKeyCode() == KeyEvent.VK_BACK_SPACE) {
		deleteChar();
		handled = true;
	    }
	    break;
	    
	case KeyEvent.KEY_TYPED:
	    appendChar(e.getKeyChar());
	    handled = true;
	    break;
	}

	return handled;
    }

    private void appendChar(char c) {
	FontMetrics metrics = getFontMetrics(font);
	int lineHeight = metrics.getHeight();
	boolean showCaret = true;

	if (c == '\n' || c == '\r') {
	    lines.addElement(new StringBuffer());

	    // Optimized refresh: a newly added empty line doesn't change the
	    // width.

	    height += lineHeight;
	    parent.paintRelayout(this, width, height-lineHeight, 
				 GadgetContainer.BOTTOM_SIDE, 2*lineHeight+2);
	} else if ((c >= FIRST_KEY_CHAR1 && c <= LAST_KEY_CHAR1) ||
		   (c >= FIRST_KEY_CHAR2 && c <= LAST_KEY_CHAR2)) {
	    StringBuffer line = (StringBuffer) lines.lastElement();
	    line.append(c);

	    int newLineWidth = 4+metrics.stringWidth(line.toString());
	    if (newLineWidth <= width) {
		// Optimized refresh: if, after adding last char, the last
		// line is still narrower than the gadget, the width of the
		// gadget hasn't changed and suffice to repaint last line.

		repaint(0, height-lineHeight-2, width, lineHeight+2);
	    } else {
		if (lines.size() == 1) {
		    // Optimized refresh: appending a char to a single line
		    // doesn't change the height and width computation is very
		    // easy.

		    int delta = metrics.charWidth(c);
		    width += delta;
		    parent.paintRelayout(this, width-delta, height, 
					 GadgetContainer.RIGHT_SIDE, delta+2);
		} else {
		    refresh();
		}
	    }
	} else {
	    showCaret = false;
	}

	if (showCaret) 
	    makeCaretVisible(metrics, lineHeight);
    }

    private void deleteChar() {
	FontMetrics metrics = getFontMetrics(font);
	int lineHeight = metrics.getHeight();
	boolean showCaret = true;

	int lineCount = lines.size();
	int lastLine = lineCount - 1;
	StringBuffer line = (StringBuffer) lines.elementAt(lastLine);
	int length = line.length();

	if (length > 0) {
	    int prevLineWidth = 4+metrics.stringWidth(line.toString());

	    char c = line.charAt(length-1);
	    line.setLength(length-1);

	    if (prevLineWidth < width) {
		// Optimized refresh: if, before removing last char, the last
		// line was narrower than the gadget, the width of the gadget
		// hasn't changed and suffice to repaint last line.

		repaint(0, height-lineHeight-2, width, lineHeight+2);
	    } else {
		if (lineCount == 1) {
		    // Optimized refresh: deleting last char to a single line
		    // doesn't change the height and width computation is very
		    // easy.

		    int delta = metrics.charWidth(c);
		    width -= delta;
		    parent.paintRelayout(this, width+delta, height, 
					 GadgetContainer.RIGHT_SIDE, 2);
		} else {
		    refresh();
		}
	    }
	} else {
	    if (lineCount > 1) {
		// Always keep at least an empty line.
		lines.removeElementAt(lastLine);

		// Optimized refresh: a removed empty line doesn't change the
		// width.

		height -= lineHeight;
		parent.paintRelayout(this, width, height+lineHeight, 
				     GadgetContainer.BOTTOM_SIDE, 
				     lineHeight+2);
	    } else {
		getToolkit().beep();
		showCaret = false;
	    }
	}

	if (showCaret)
	    makeCaretVisible(metrics, lineHeight);
    }

    private void makeCaretVisible(FontMetrics metrics, int lineHeight) {
	if (metrics == null) {
	    metrics = getFontMetrics(font);
	    lineHeight = metrics.getHeight();
	}

	StringBuffer lastLine = (StringBuffer) lines.lastElement();
	int lastLineWidth = metrics.stringWidth(lastLine.toString());
	makeRectangleVisible(2+lastLineWidth, height-lineHeight-2, 
			     2, lineHeight+2);
    }

    public boolean handleMouseEvent(MouseEvent e) {
	if (e.getModifiers() == MouseEvent.BUTTON1_MASK) {
	    if (e.getID() == MouseEvent.MOUSE_CLICKED) 
		requestFocus();
	    return true;
	} else {
	    return false;
	}
    }

    // -----------------------------------------------------------------------

    public static void main(String[] args) {
	JFrame frame = new JFrame();
   	frame.addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        System.exit(0);
   	    }
   	});

	JPanel workArea = (JPanel) frame.getContentPane();

	final JGadgetPanel panel = new JGadgetPanel();
	final RootGadget root = panel.getRoot();
	root.addChild(new TestGadget("Click to type"));
	root.refresh();
	workArea.add(panel, BorderLayout.CENTER);

	final JCheckBox toggle = new JCheckBox("JGadgetPanel enabled", true);
	workArea.add(toggle, BorderLayout.NORTH);

	toggle.addChangeListener(new ChangeListener() {
	    public void stateChanged(ChangeEvent e) {
		panel.setEnabled(toggle.isSelected());
	    }
	});

	frame.pack();
	frame.show();
    }
}

