/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.xmlmind.gadget.*;

public class TestStretch extends Gadget {
    private String text;
    private Color background = Color.white;
    private Color foreground = Color.black;
    private Font font = new Font("SansSerif", Font.PLAIN, 12);
    private int margin = 2;

    public TestStretch() {
	this("");
    }

    public TestStretch(String text) {
	this.text = text;
	relayout();
    }

    public void setText(String text) {
	this.text = text;
    }

    public String getText() {
	return text;
    }

    public void setBackground(Color color) {
	background = color;
    }

    public Color getBackground() {
	return background;
    }

    public void setForeground(Color color) {
	foreground = color;
    }

    public Color getForeground() {
	return foreground;
    }

    public void setFont(Font font) {
	this.font = font;
    }

    public Font getFont() {
	return font;
    }

    public void setMargin(int margin) {
	this.margin = margin;
    }

    public int getMargin() {
	return margin;
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	paintText(null, constrainedWidth);
    }

    public void paint(Graphics g) {
	g.setColor(background);
	g.fillRect(0, 0, width, height);
	
	g.setFont(font);
	g.setColor(foreground);

	Dimension constraints = new Dimension();
	if (parent != null)
	    parent.getSizeConstraints(this, constraints);
	paintText(g, constraints.width);

//System.out.println("paintText");
    }

    private void paintText(Graphics g, int constrainedWidth) {
	FontMetrics metrics = getFontMetrics(font);

	int textLength = text.length();
	if (textLength == 0) {
	    if (g == null) {
		width = 2*margin;
		height = 2*margin + 
		    metrics.getAscent() + metrics.getDescent();
	    }
	    return;
	}

	int lineY = margin + metrics.getAscent();
	int lineHeight = metrics.getHeight();
	int widest = 0;
	int spaceWidth = metrics.charWidth(' ');

	int maxWidth = ((constrainedWidth > 0)? (constrainedWidth - 2*margin) :
			Integer.MAX_VALUE);
	if (maxWidth <= 0)
	    maxWidth = 1;

	char[] lineChars = new char[textLength];
	int lineCharCount = 0;
	int lineBegin = 0;
	int lineWidth = 0;
	int wordFirst = -1;
	int wordLast = -1;

	for (int i = 0; i < textLength; ++i) {
	    switch (text.charAt(i)) {
	    case ' ': case '\n': case '\r': case '\t':
		if (wordFirst >= 0) {
		    int adding;

		    if (lineCharCount == 0) {
			// First word: no space separator.
			adding = 0;
		    } else {
			lineChars[lineCharCount] = ' ';
			adding = 1;
		    }

		    text.getChars(wordFirst, wordLast+1, 
				  lineChars, lineCharCount+adding);
		    adding += wordLast-wordFirst+1;
		    wordFirst = wordLast = -1;

		    int w = 
			metrics.charsWidth(lineChars, lineCharCount, adding);

		    if (lineWidth + w > maxWidth && lineCharCount != 0) {
			// The first word may happen to be longer than
			// maxWidth!

			if (g == null) {
			    if (lineWidth > widest)
				widest = lineWidth;
			} else {
			    g.drawChars(lineChars, lineBegin,
					lineCharCount-lineBegin, 
					margin, lineY);
			}
			lineY += lineHeight;

			// Do not begin a line with the space separator
			// we have added.
			lineBegin = lineCharCount + 1;
			lineWidth = w - spaceWidth;
		    } else {
			lineWidth += w;
		    }

		    lineCharCount += adding;
		}
		break;

	    default:
		if (wordFirst < 0) 
		    wordFirst = wordLast = i;
		else
		    ++wordLast;
	    }
	}

	// Add last word if any ---

	if (wordFirst >= 0) {
	    int adding;

	    if (lineCharCount == 0) {
		adding = 0;
	    } else {
		lineChars[lineCharCount] = ' ';
		adding = 1;
	    }

	    text.getChars(wordFirst, wordLast+1, 
			  lineChars, lineCharCount+adding);
	    adding += wordLast-wordFirst+1;

	    int w = metrics.charsWidth(lineChars, lineCharCount, adding);

	    if (lineWidth + w > maxWidth && lineCharCount != 0) {
		if (g == null) {
		    if (lineWidth > widest)
			widest = lineWidth;
		} else {
		    g.drawChars(lineChars, lineBegin, lineCharCount-lineBegin, 
				margin, lineY);
		}
		lineY += lineHeight;

		lineBegin = lineCharCount + 1;
		lineWidth = w - spaceWidth;
	    } else {
		lineWidth += w;
	    }

	    lineCharCount += adding;
	}

	// Paint last line ---

	if (lineBegin < lineCharCount) {
	    if (g == null) {
		if (lineWidth > widest)
		    widest = lineWidth;
	    } else {
		g.drawChars(lineChars, lineBegin, lineCharCount-lineBegin, 
			    margin, lineY);
	    }
	    lineY += lineHeight;
	}

	// Commit size if not painting ---

	if (g == null) {
	    width = widest + 2*margin;
	    height = lineY - metrics.getAscent() + margin;
	}
    }

    // -----------------------------------------------------------------------

    public static final String sample1 = 
    "Copyright (c) 1999,2000 Pixware.\n\n" +
    "This file is part of the Pixware APT tools.\n" +
    "For conditions of distribution and use," +
    " see the accompanying legal.txt file.\n\n";

    public static final String sample2 = 
    "Pixware APT tools. Generate, parse, convert APT documents.\n" +
    "Copyright (c) 1999,2000 Pixware\n\n" +
    "This program is free software; you can redistribute it" + 
    " and/or modify it\n" +
    "under the terms of the GNU General Public License Version 2" + 
    " as published by\n" +
    "the Free Software Foundation.\n\n" +
    "This program is distributed in the hope that it will be useful,\n" +
    "but WITHOUT ANY WARRANTY; without even the implied warranty of\n" +
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n" +
    "GNU General Public License for more details.\n\n" +
    "You should have received a copy of the GNU General Public License\n" +
    "along with this program; if not, write to the Free Software\n" +
    "Foundation, Inc., 59 Temple Place, Suite 330, Boston," + 
    " MA	02111-1307	USA\n\n";

    public static void main(String[] args) {
	JFrame frame = new JFrame();
   	frame.addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        System.exit(0);
   	    }
   	});

	JGadgetPanel panel = new JGadgetPanel();
	final RootGadget root = panel.getRoot();
	root.addChild(new TestStretch(sample1));
	root.refresh();

	final JPanel workArea = (JPanel) frame.getContentPane();
	workArea.add(panel);

	workArea.addComponentListener(new ComponentAdapter() {
	    private Dimension dim = new Dimension();

	    public void componentResized(ComponentEvent e) {
		workArea.getSize(dim);
		root.layout(dim.width, dim.height);
		root.repaint();
	    }
	});
	
	frame.setBounds(0, 0, 200, 200);
	frame.show();
    }
}
