/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.javadocformat;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import com.xmlmind.xmledit.model.*;

public class ElementContent {
    private static ElementContent instance = null;

    private Hashtable emptyElements = new Hashtable();
    private Hashtable structureElements = new Hashtable();
    private Hashtable possibleChild = new Hashtable();
    private Hashtable blockOrFlow = new Hashtable();

    public static ElementContent get(DocumentTypeModel docType) {
	if (instance == null)
	    instance = new ElementContent(docType);
	return instance;
    }

    private ElementContent(DocumentTypeModel docType) {
	Enumeration iter = docType.getElementTypes();
	Vector elemTypeNames = new Vector();

	while (iter.hasMoreElements()) {
	    ElementTypeModel elemType = (ElementTypeModel) iter.nextElement();
	    String elemTypeName = elemType.getElementTypeName();

	    if (elemType.getElementContentType() != 
		ElementTypeModel.UNDECLARED)
		elemTypeNames.addElement(elemTypeName);
	}

	String[] declaredElemTypeNames = new String[elemTypeNames.size()];
	elemTypeNames.copyInto(declaredElemTypeNames);

	iter = docType.getElementTypes();
	while (iter.hasMoreElements()) {
	    ElementTypeModel elemType = (ElementTypeModel) iter.nextElement();
	    String parentName = elemType.getElementTypeName();

	    switch (elemType.getElementContentType()) {
	    case ElementTypeModel.EMPTY:
		emptyElements.put(parentName, parentName);
		break;

	    case ElementTypeModel.ANY:
		// ANY means any declared type or #PCDATA.
	    case ElementTypeModel.UNDECLARED:
		{
		    // Note that this type of element type can contain itself.
		    for (int i = 0; i < declaredElemTypeNames.length; ++i) {
			String key = 
			    newKey(parentName, declaredElemTypeNames[i]);
			possibleChild.put(key, key);
		    }

		    if (!blockOrFlow.containsKey(parentName))
			blockOrFlow.put(parentName, 
					canTrim(docType, elemType));
		}
		break;

	    case ElementTypeModel.MIXED:
		{
		    ElementTypeModel[] mix = elemType.getMixedElementTypes();

		    for (int i = 0; i < mix.length; ++i) {
			String key = 
			    newKey(parentName, mix[i].getElementTypeName());
			possibleChild.put(key, key);
		    }

		    // A flow is a mix which can contain <p>.
		    if (isPossibleChild(parentName, "p") &&
			!blockOrFlow.containsKey(parentName))
			blockOrFlow.put(parentName, 
					canTrim(docType, elemType));
		}
		break;

	    case ElementTypeModel.ELEMENT:
		{
		    structureElements.put(parentName, parentName);

		    ElementContentModel contentModel = 
			elemType.getElementContent();
		    Vector list = new Vector();
		    addElementContentChildTypeNames(contentModel, list);
		    String[] childTypeNames = new String[list.size()];
		    list.copyInto(childTypeNames);

		    for (int i = 0; i < childTypeNames.length; ++i) {
			String key = newKey(parentName, childTypeNames[i]);
			possibleChild.put(key, key);
		    }

		    // A block is an element which can be a child of 
		    // <blockquote>
		    if ("blockquote".equals(parentName)) {
			for (int i = 0; i < childTypeNames.length; ++i) {
			    String childName = childTypeNames[i];

			    if (!blockOrFlow.containsKey(childName))
				blockOrFlow.put(childName, 
						canTrim(docType, childName));
			}
		    }
		}
		break;
	    }
	}
    }

    private static final 
    void addElementContentChildTypeNames(ElementContentModel contentModel, 
					 Vector list) {
	switch (contentModel.getListType()) {
	case ElementContentModel.NONE:
	    {
		ElementTypeModel elemType = contentModel.getElementType();
		String elemTypeName = elemType.getElementTypeName();

		// Even UNDECLARED are added to the list.
		if (list.indexOf(elemTypeName) < 0)
		    list.addElement(elemTypeName);
	    }
	    break;

	case ElementContentModel.CHOICE:
	case ElementContentModel.SEQUENCE:
	    {
		ElementContentModel[] seq = contentModel.getList(); 

		for (int i = 0; i < seq.length; ++i)
		    addElementContentChildTypeNames(seq[i], list);
	    }
	    break;
	}
    }

    private static final Boolean canTrim(DocumentTypeModel docType, 
					 String elemName) {
	ElementTypeModel elemType = docType.getElementType(elemName);

	if (elemType == null)
	    return Boolean.TRUE;
	else
	    return canTrim(docType, elemType);
    }

    private static final Boolean canTrim(DocumentTypeModel docType, 
					 ElementTypeModel elemType) {
	switch (elemType.getElementContentType()) {
	case ElementTypeModel.MIXED:
	case ElementTypeModel.ANY:
            // ANY means any declared type or #PCDATA.
	case ElementTypeModel.UNDECLARED:
	    break;
	default:
	    return Boolean.FALSE;
	}

	AttributeDefinitionModel attrDef = 
	    elemType.getAttributeDefinition("xml:space");

	if (attrDef == null)
	    return Boolean.TRUE;
	else
	    return ((attrDef.getAttributePresence() == 
		     AttributeDefinitionModel.FIXED && 
		     "preserve".equals(attrDef.getAttributeDefaultValue()))?
		    Boolean.FALSE : Boolean.TRUE);
    }

    public boolean isPossibleChild(String parentName, String childName) {
	return possibleChild.containsKey(newKey(parentName, childName));
    }

    private static final String newKey(String parentName, String childName) {
	StringBuffer key = new StringBuffer();

	key.append(parentName);
	key.append(' ');
	key.append(childName);

	return key.toString();
    }

    public boolean isEmpty(String name) {
	return emptyElements.containsKey(name);
    }

    public boolean isStructure(String name) {
	return structureElements.containsKey(name);
    }

    public boolean isBlockOrFlow(String name) {
	return blockOrFlow.containsKey(name);
    }

    public boolean isTrimableTextContainer(String name) {
	Boolean trim = (Boolean) blockOrFlow.get(name);

	if (trim == null)
	    return false;
	else
	    return trim.booleanValue();
    }
}

