/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.javadocformat;

import java.io.*;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Stack;
import com.xmlmind.xmledit.modeltool.XMLUtil;

public class Filter extends Reader {
    private LineNumberReader in;
    private ElementContent elementContent;
    private int tabWidth;

    private String nextLine;
    private int lineNumber;
    private char[] buffer;
    private int bufferOffset;

    private static final char[] LAST_BUFFER = "</java>".toCharArray();
    private static final char[] END_BUFFER = new char[0];

    private static final char spaces[] = 
    "                                                          ".toCharArray();

    // Note that is array is sorted.
    private static final String[] knownTags = {
	"param", 
	"exception", 
	"throws", 
	"return", 
	"see", 
	"author", 
	"version",
	"since", 
	"deprecated", 
	"serialField", 
	"serialData", 
	"serial"
    };

    // Javadoc comment to XML filter state ---

    private Tokenizer scanner;
    private int lookAhead;
    private Stack stack;
    private StringBuffer xml;

    // -----------------------------------------------------------------------

    public Filter(Reader in, String internalSubset,
		  ElementContent elementContent, int tabWidth) {
	this.in = new LineNumberReader(in);
	this.elementContent = elementContent;
	this.tabWidth = tabWidth;

	StringBuffer firstBuffer = new StringBuffer();
	firstBuffer.append("<?xml version='1.0'?>\n<!DOCTYPE java [\n");
	firstBuffer.append(internalSubset);
	firstBuffer.append("\n]>\n<java>");

	nextLine = null;
	lineNumber = -1;
	buffer = firstBuffer.toString().toCharArray();
	bufferOffset = 0;
    }

    public int getLineNumber() {
	return lineNumber;
    }

    public int read() throws IOException {
	int c;

	if (bufferOffset >= buffer.length)
	    doRead();
	    
	if (bufferOffset < buffer.length)
	    c = buffer[bufferOffset++];
	else
	    c = -1;

	return c;
    }

    public int read(char chars[], int offset, int length) throws IOException {
	int totalCount = 0;

	while (length > 0) {
	    int count = Math.min(length, buffer.length - bufferOffset);
	    if (count == 0) {
		doRead();
		count = Math.min(length, buffer.length - bufferOffset);
	    }

	    if (count > 0) {
		System.arraycopy(buffer, bufferOffset, chars, offset, count);
		bufferOffset += count;

		offset += count;
		length -= count;

		totalCount += count;
	    } else {
		break;
	    }
	}

	return (totalCount == 0)? -1 : totalCount;
    }

    public void close() throws IOException {
	in.close();
    }

    // -----------------------------------------------------------------------
    // Buffering
    // -----------------------------------------------------------------------

    private void doRead() throws IOException {
	if (buffer == END_BUFFER)
	    return;

	bufferOffset = 0;

	if (buffer == LAST_BUFFER) {
	    buffer = END_BUFFER;
	    return;
	}

	if (nextLine == null) {
	    nextLine = in.readLine();
	    lineNumber = in.getLineNumber();
	}

	String line = nextLine;
	if (line == null) {
	    buffer = LAST_BUFFER;
	    return;
	}

	String trimmed = trimCommentStart(line);
	if (trimmed != line) {
	    line = trimmed;
	    StringBuffer doc = new StringBuffer();

	    // Support for something like "/** xxx */".
	    while ((trimmed = trimCommentEnd(line)) == line) {
		doc.append(replaceAtTag(trimCommentLine(line)));
		doc.append('\n');

		line = in.readLine();
		if (line == null)
		    throw new InvalidFormatException(
			msg("syntaxError2", "'*/'", msg("eof")));
	    }

	    line = trimmed;

	    // Support for something like "   * xxx */".
	    // Note that first line, which begins with "/**", has already
	    // been trimmed.
	    if (doc.length() > 0)
		line = trimCommentLine(line);

	    doc.append(replaceAtTag(line));
	    doc.append('\n');

	    nextLine = null;

	    buffer = filterDoc(doc.toString()).toCharArray();

	} else {
	    StringBuffer processed = new StringBuffer();

	    for (;;) {
		processSourceLine(line, processed);

		line = in.readLine();
		if (line == null)
		    break;

		lineNumber = in.getLineNumber();
		if (trimCommentStart(line) != line)
		    break;
	    }

	    nextLine = line;

	    buffer = processed.toString().toCharArray();
	}
    }

    private static final String trimCommentStart(String line) {
	int length = line.length();

	for (int pos = 0; pos < length; ++pos) {
	    char c = line.charAt(pos);

	    if (!Character.isWhitespace(c)) {
		if (c == '/' && pos+2 < length && 
		    line.charAt(pos+1) == '*' && line.charAt(pos+2) == '*')
		    return (pos+3 < length)? line.substring(pos+3) : "";
		else
		    return line;
	    }
	}

	return line;
    }

    private static final String trimCommentEnd(String line) {
	int length = line.length();

	for (int pos = length-1; pos >= 0; --pos) {
	    char c = line.charAt(pos);

	    if (!Character.isWhitespace(c)) {
		if (c == '/' && pos-1 >= 0 && line.charAt(pos-1) == '*')
		    return (pos-1 > 0)? line.substring(0, pos-1) : "";
		else
		    return line;
	    }
	}
	
	return line;
    }

    private static final String trimCommentLine(String line) {
	int length = line.length();

	for (int pos = 0; pos < length; ++pos) {
	    char c = line.charAt(pos);

	    if (!Character.isWhitespace(c)) {
		if (c == '*')
		    return (pos+1 < length)? line.substring(pos+1) : "";
		else
		    return line.substring(pos);
	    }
	}
	
	return "";
    }

    private static final String replaceAtTag(String line) {
	// If an @ (at) tag does not start at the beginning of a line,
	// it is treated as normal text.

	int length = line.length();

	for (int pos = 0; pos < length; ++pos) {
	    char c = line.charAt(pos);

	    if (!Character.isWhitespace(c)) {
		if (c == '@') {
		    for (int i = 0; i < knownTags.length; ++i) {
			String tag = knownTags[i];

			if (line.indexOf(tag, pos+1) == pos+1) {
			    StringBuffer replaced = new StringBuffer();

			    if (pos > 0)
				replaced.append(line.substring(0, pos));

			    replaced.append('<');
			    replaced.append(tag);
			    replaced.append('>');

			    pos += 1 + tag.length();
			    if (pos < length)
				replaced.append(line.substring(pos));

			    return replaced.toString();
			}
		    }
		    return line;
		} else {
		    return line;
		}
	    }
	}

	return line;
    }

    private void processSourceLine(String line, StringBuffer processed) {
	int length = line.length();
	boolean untabify = (tabWidth > 1);
	int prevColumn = -1;
	int column = 0;

	for (int i = 0; i < length; ++i) {
	    char c = line.charAt(i);

	    if (untabify && c == '\t') {
		prevColumn = column;
		column = ((column+1 + tabWidth-1) / tabWidth) * tabWidth;
		processed.append(spaces, 0, column - prevColumn);
	    } else {
		++column;
		switch (c) {
		case '\f':
		    processed.append("<formfeed/>");
		    break;
		case '<':
		    processed.append("&#60;");
		    break;
		case '>':
		    processed.append("&#62;");
		    break;
		case '&':
		    processed.append("&#38;");
		    break;
		default:
		    processed.append(c);
		}
	    }
	}
	processed.append('\n');
    }

    // -----------------------------------------------------------------------
    // Javadoc comment to XML filter
    // -----------------------------------------------------------------------

    private String filterDoc(String doc) throws InvalidFormatException {
	scanner = new Tokenizer(doc);
	stack = new Stack();
	xml = new StringBuffer();
	stack.push("doc");
	xml.append("<doc>");
	stack.push("description");
	xml.append("<description>");

	lookAhead = scanner.nextToken();
	loop: for (;;) {
	    switch (lookAhead) {
	    case Tokenizer.EOF:
		break loop;

	    case Tokenizer.START_TAG:
	        {
		    String elementName = scanner.getElementName();

		    if (elementName.equals("param")) {
			filterParam();
		    } else if (elementName.equals("exception") ||
			       elementName.equals("throws")) {
			filterException();
		    } else if (elementName.equals("see")) {
			filterSee();
		    } else if (elementName.equals("serialField")) {
			filterSerialField();
		    } else {
			filterStartTag();
		    }
	        }
		break;

	    case Tokenizer.END_TAG:
		filterEndTag();
		break;

	    default:
		filterText();
	    }
	}

	while (!stack.empty()) {
	    xml.append(endTag((String) stack.peek()));
	    stack.pop();
	}

	String filtered = xml.toString();
	scanner = null;
	xml = null;
	stack = null;
	return filtered;
    }

    private String top() {
	return stack.empty()? null : (String) stack.peek();
    }

    private static final String endTag(String name) {
	StringBuffer buffer = new StringBuffer();
	buffer.append("</");
	buffer.append(name);
	buffer.append(">");
	return buffer.toString();
    }

    private void filterParam() throws InvalidFormatException {
	autoEndTags("param");
	stack.push("param");
	xml.append("<param>");

	lookAhead = scanner.nextToken();
	if (lookAhead == Tokenizer.TEXT && isXMLSpace(scanner.getText()))
	    lookAhead = scanner.nextToken();

	if (lookAhead != Tokenizer.TEXT)
	    throw new InvalidFormatException(
		msg("syntaxError2", msg("paramname"), lookAheadAsString()));

	String[] parts = splitDescription(scanner.getText());

	xml.append("<paramname>");
	appendText(parts[0]);
	xml.append("</paramname>");

	stack.push("paramdesc");
	xml.append("<paramdesc>");
	if (parts[1] != null)
	    filterText(parts[1]);
	else
	    lookAhead = scanner.nextToken();
    }

    private void autoEndTags(String childName) {
	while (!stack.empty()) {
	    String parentName = (String) stack.peek();
	    if (elementContent.isPossibleChild(parentName, childName))
		break;

	    xml.append(endTag(parentName));
	    stack.pop();
	}
    }

    private static final boolean isXMLSpace(String s) {
	int length = s.length();

	for (int i = 0; i < length; ++i) {
	    if (!XMLUtil.isXMLSpace(s.charAt(i)))
		return false;
	}

	return true;
    }

    private String lookAheadAsString() {
	if (lookAhead == Tokenizer.EOF)
	    return msg("eoc");

	StringBuffer buffer = new StringBuffer();

	switch (lookAhead) {
	case Tokenizer.START_TAG:
	    buffer.append('<');
	    buffer.append(scanner.getElementName().toLowerCase());
	    break;
	case Tokenizer.END_TAG:
	    buffer.append("</");
	    buffer.append(scanner.getElementName().toLowerCase());
	    buffer.append('>');
	    break;
	case Tokenizer.TEXT:
	    buffer.append('\"');
	    buffer.append(scanner.getText());
	    buffer.append('\"');
	    break;
	}

	return buffer.toString();
    }

    private static final String[] splitDescription(String s) {
	String[] parts = new String[2];
	int length = s.length();

	int pos1 = 0;
	while (pos1 < length) {
	    if (!XMLUtil.isXMLSpace(s.charAt(pos1)))
		break;
	    ++pos1;
	}

	int pos2 = pos1+1;
	while (pos2 < length) {
	    if (XMLUtil.isXMLSpace(s.charAt(pos2)))
		break;
	    ++pos2;
	}

	int pos3 = pos2+1;
	while (pos3 < length) {
	    if (!XMLUtil.isXMLSpace(s.charAt(pos3)))
		break;
	    ++pos3;
	}

	if (pos1 < length)
	    parts[0] = s.substring(pos1, pos2);
	if (pos3 < length)
	    parts[1] = s.substring(pos3);

	return parts;
    }

    private void filterException() throws InvalidFormatException {
	autoEndTags("exception");
	stack.push("exception");
	xml.append("<exception>");

	lookAhead = scanner.nextToken();
	if (lookAhead == Tokenizer.TEXT && isXMLSpace(scanner.getText()))
	    lookAhead = scanner.nextToken();

	if (lookAhead != Tokenizer.TEXT)
	    throw new InvalidFormatException(
		msg("syntaxError2", msg("exceptionname"), 
		    lookAheadAsString()));

	String[] parts = splitDescription(scanner.getText());

	xml.append("<exceptionname>");
	appendText(parts[0]);
	xml.append("</exceptionname>");

	stack.push("exceptiondesc");
	xml.append("<exceptiondesc>");
	if (parts[1] != null)
	    filterText(parts[1]);
	else
	    lookAhead = scanner.nextToken();
    }

    private void filterSerialField() throws InvalidFormatException {
	autoEndTags("serialfield");
	stack.push("serialfield");
	xml.append("<serialfield>");

	lookAhead = scanner.nextToken();
	if (lookAhead == Tokenizer.TEXT && isXMLSpace(scanner.getText()))
	    lookAhead = scanner.nextToken();

	if (lookAhead != Tokenizer.TEXT)
	    throw new InvalidFormatException(
		msg("syntaxError2", msg("fieldname"), lookAheadAsString()));

	String[] parts = splitDescription(scanner.getText());
	if (parts[1] == null)
	    throw new InvalidFormatException(
		msg("syntaxError2", msg("fieldtype"), msg("eoc")));

	xml.append("<fieldname>");
	appendText(parts[0]);
	xml.append("</fieldname>");

	parts = splitDescription(parts[1]);
	xml.append("<fieldtype>");
	appendText(parts[0]);
	xml.append("</fieldtype>");

	stack.push("fielddesc");
	xml.append("<fielddesc>");
	if (parts[1] != null)
	    filterText(parts[1]);
	else
	    lookAhead = scanner.nextToken();
    }

    private void processLink(String text) throws InvalidFormatException {
	processLink("link", text);
    }

    private void processLink(String elementName, String text) 
	throws InvalidFormatException {
	String[] parts = splitLink(text);
	if (parts[0] == null)
	    throw new InvalidFormatException(
		msg("syntaxError2", "package.class#member", msg("eoc")));

	autoEndTags(elementName);
	xml.append('<');
	xml.append(elementName);
	if (parts[1] != null) {
	    xml.append(" label=\"");
	    appendText(parts[1].trim());
	    xml.append('\"');
	}
	xml.append('>');
	appendText(parts[0]);
	xml.append("</");
	xml.append(elementName);
	xml.append('>');
    }

    private static final String[] splitLink(String s) {
	String[] parts = new String[2];
	int length = s.length();

	int pos1 = 0;
	while (pos1 < length) {
	    if (!XMLUtil.isXMLSpace(s.charAt(pos1)))
		break;
	    ++pos1;
	}

	int pos2 = pos1+1;
	while (pos2 < length) {
	    char c = s.charAt(pos2);

	    if (XMLUtil.isXMLSpace(c))
		break;

	    if (c == '(') {
		// Spaces are allowed between '(' and ')' such as in
		// 'String#indexOf(int, int)'.

		int nextPos2 = s.indexOf(')', pos2+1);
		if (nextPos2 >= 0)
		    pos2 = nextPos2;
	    } 

	    ++pos2;
	}

	int pos3 = pos2+1;
	while (pos3 < length) {
	    if (!XMLUtil.isXMLSpace(s.charAt(pos3)))
		break;
	    ++pos3;
	}

	if (pos1 < length)
	    parts[0] = s.substring(pos1, pos2);
	if (pos3 < length)
	    parts[1] = s.substring(pos3);

	return parts;
    }

    private void filterSee() throws InvalidFormatException {
	lookAhead = scanner.nextToken();
	if (lookAhead == Tokenizer.TEXT && isXMLSpace(scanner.getText()))
	    lookAhead = scanner.nextToken();

	switch (lookAhead) {
	case Tokenizer.TEXT:
	    {
		String text = scanner.getText().trim();
		int length = text.length();

		if (length > 0 && text.charAt(0) == '\"') {
		    // Unquote text.
		    text = text.substring(1);
		    --length;
		    if (length > 0 && text.charAt(length-1) == '\"')
			text = text.substring(0, length-1);
			
		    autoEndTags("seeref");
		    xml.append("<seeref>");
		    appendText(text);
		    xml.append("</seeref>");
		} else {
		    processLink("see", text);
		}

		lookAhead = scanner.nextToken();
	    }
	    break;

	case Tokenizer.START_TAG:
	    {
		String elementName = scanner.getElementName().toLowerCase();
		if (!elementName.equals("a"))
		    throw new InvalidFormatException(
			msg("syntaxError2", "<a", lookAheadAsString()));

		String[] attributes = scanner.getElementAttributes();
		boolean hasHref = false;
		for (int i = 0; i < attributes.length; i += 2) {
		    String attrName = attributes[i].toLowerCase();
		    if (attrName.equals("href")) {
			hasHref = true;
			break;
		    }
		}
		if (!hasHref)
		    throw new InvalidFormatException(
			msg("syntaxError2", "<a href=\"URL#value\"", 
			    lookAheadAsString()));

		filterStartTag("seehref", attributes);
		// LookAhead is after <a>.

		loop: for (;;) {
		    switch (lookAhead) {
		    case Tokenizer.EOF:
			break loop;
		    case Tokenizer.START_TAG:
			filterStartTag();
			break;
		    case Tokenizer.END_TAG:
		        {
			    String elemName = 
				scanner.getElementName().toLowerCase();
			    if (elemName.equals("a"))
				break loop;

			    filterEndTag();
			}
			break;
		    default:
			filterText();
		    }
		}
		if (lookAhead == Tokenizer.EOF)
		    throw new InvalidFormatException(
			msg("syntaxError2", "</a>", msg("eoc")));
		    
		filterEndTag("seehref");
		// LookAhead is after </a>.
		if (lookAhead == Tokenizer.TEXT && 
		    isXMLSpace(scanner.getText()))
		    lookAhead = scanner.nextToken();
	    }
	    break;

	default:
	    throw new InvalidFormatException(
		msg("syntaxError2", msg("reference"), lookAheadAsString()));
	}
    }

    private void filterStartTag() throws InvalidFormatException {
	filterStartTag(scanner.getElementName().toLowerCase(),
		       scanner.getElementAttributes());
    }

    private void filterStartTag(String elementName, String[] attributes) 
	throws InvalidFormatException {
	boolean isEmptyElement = elementContent.isEmpty(elementName);

	autoEndTags(elementName);
	if (!isEmptyElement)
	    stack.push(elementName);
	xml.append('<');
	xml.append(elementName);
	
	for (int i = 0; i < attributes.length; i += 2) {
	    String attrName = attributes[i].toLowerCase();
	    String attrValue = attributes[i+1];
	    if (attrValue == null)
		// Convert '<dl compact>' to '<dl compact="compact">'.
		attrValue = attrName;

	    xml.append(' ');
	    xml.append(attrName);
	    xml.append('=');
	    xml.append('\"');
	    appendText(attrValue);
	    xml.append('\"');
	}

	xml.append(isEmptyElement? "/>" : ">");

	lookAhead = scanner.nextToken();
    }

    private void filterEndTag() throws InvalidFormatException {
	filterEndTag(scanner.getElementName().toLowerCase());
    }

    private void filterEndTag(String elementName) 
	throws InvalidFormatException {
	// Auto-close tags.
	while (!stack.empty()) {
	    String top = (String) stack.peek();
	    if (elementName.equals(top))
		break;

	    xml.append(endTag(top));
	    stack.pop();
	}

	if (stack.empty())
	    throw new InvalidFormatException(
		msg("syntaxError1", endTag(elementName)));

	xml.append(endTag(elementName));
	stack.pop();

	lookAhead = scanner.nextToken();
    }

    private void filterText() throws InvalidFormatException {
	filterText(scanner.getText());
    }

    private void filterText(String text) throws InvalidFormatException {
	int pos = text.indexOf("{@link");

	if (pos < 0) {
	    appendText(text);
	} else {
	    int textLength = text.length();

	    if (pos > 0)
		appendText(text.substring(0, pos));

	    while (pos >= 0) {
		int endPos = text.indexOf('}', pos+6);
		if (endPos < 0)
		    throw new InvalidFormatException(
			msg("syntaxError2", "'}'", msg("eoc")));

		processLink(text.substring(pos+6, endPos));

		if (endPos+1 < textLength) {
		    pos = text.indexOf("{@link", endPos+1);
		    if (pos < 0) {
			appendText(text.substring(endPos+1));
		    } else {
			if (pos > endPos+1)
			    appendText(text.substring(endPos+1, pos));
		    }
		} else {
		    pos = -1;
		}
	    }
	}

	lookAhead = scanner.nextToken();
    }

    private void appendText(String text) {
	int length = text.length();

	for (int i = 0; i < length; ++i) {
	    char c = text.charAt(i);

	    switch (c) {
	    case '\'':
		xml.append("&#39;");
		break;
	    case '\"':
		xml.append("&#34;");
		break;
	    case '<':
		xml.append("&#60;");
		break;
	    case '>':
		xml.append("&#62;");
		break;
	    case '&':
		{
		    boolean isEntity = false;

		    int semicol = text.indexOf(';', i+1);
		    if (semicol > i+1) {
			if (text.charAt(i+1) == '#') {

			    if (i+2 < semicol) {
				if (i+3 < semicol && text.charAt(i+2) == 'x') {
				    try {
					Integer.parseInt(
					    text.substring(i+3, semicol), 16);
					isEntity = true;
				    } catch (NumberFormatException e) {}
				} else {
				    try {
					Integer.parseInt(
					    text.substring(i+2, semicol));
					isEntity = true;
				    } catch (NumberFormatException e) {}
				}
			    }

			} else {
			    if (XMLUtil.isName(text.substring(i+1, semicol)))
				isEntity = true;
			}
		    }

		    if (isEntity)
			xml.append(c);
		    else
			xml.append("&#38;");
		}
		break;
	    default:
		xml.append(c);
	    }
	}
    }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.javadocformat.FilterRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    private static String msg(String id, Object arg0, Object arg1) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0, arg1 });
    }
}
