/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.javadocformat;

import java.io.*;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Vector;
import java.net.URL;
import java.net.MalformedURLException;
import fr.pixware.util.FileUtil;
import fr.pixware.util.StringUtil;
import fr.pixware.util.WorkInProgressObserver;
import com.xmlmind.xmledit.model.DocumentModel;
import com.xmlmind.structformat.*;
// Needed by main().
import com.xmlmind.xmledit.modeltool.DocumentWriter;
import com.xmlmind.xmledit.modelimpl.Document;

public class Format implements StructuredFormat {
    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.javadocformat.FormatRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    // -----------------------------------------------------------------------

    private static final String platformDefaultEncoding = 
        FileUtil.defaultEncoding();

    private static final String[] extensions = { "java" };
    private static final String description = msg("description");
    private static final String[][] parameterUsage = {
	{"untabifyWidth", msg("int"), "8", msg("untabifyWidth")},
	{"tabifyWidth", msg("int"), "8", msg("tabifyWidth")},
	{"lineSeparator", msg("string"), "\\n", msg("lineSeparator")},
	{"maxLineLength", msg("positiveInt"), "78", msg("maxLineLength")},
	{"defaultEncoding", msg("string"), platformDefaultEncoding, 
         msg("defaultEncoding")}
    };

    private Loader loader = new Loader();
    private Formatter formatter = new Formatter();
    private String defaultEncoding = platformDefaultEncoding;

    // -----------------------------------------------------------------------

    public String getMimeType() {
	return "text/x-java";
    }

    public String[] getFileNameExtensions() {
	return extensions;
    }

    public String getDescription() {
	return description;
    }

    public String[][] getParameterUsage() {
	return parameterUsage;
    }

    public void setParameters(String[] parameters) 
	throws ParameterUsageException {
	int count = (parameters.length/2)*2;

	for (int i = 0; i < count; i += 2) {
	    String param = parameters[i];
	    String value = parameters[i+1];

	    if ("untabifyWidth".equals(param) ||
		"tabifyWidth".equals(param)) {
		int number = -1;
		
		try {
		    number = Integer.parseInt(value);
		} catch (NumberFormatException e) {
		    throw new ParameterUsageException(
			msg("invalidParam", param));
		}

		if (param.charAt(0) == 'u')
		    setUntabifyWidth(number);
		else
		    setTabifyWidth(number);

	    } else if ("lineSeparator".equals(param)) {

		value = StringUtil.replaceAll(value, "\\n", "\n");
		value = StringUtil.replaceAll(value, "\\r", "\r");
		if (!"\n".equals(value) && 
		    !"\r\n".equals(value) && 
		    !"\r".equals(value))
		    throw new ParameterUsageException(
			msg("invalidParam", param));

		setLineSeparator(value);

	    } else if ("maxLineLength".equals(param)) {
		int number = -1;
		
		try {
		    number = Integer.parseInt(value);
		} catch (NumberFormatException e) {
		    number = -1;
		}

		if (number <= 0)
		    throw new ParameterUsageException(
			msg("invalidParam", param));

		setMaxLineLength(number);
	    } else if ("defaultEncoding".equals(param)) {
		value = value.trim();

		if (value.length() == 0) {
		    value = platformDefaultEncoding;
		} else {
		    if (!FileUtil.isKnownEncoding(value))
			throw new ParameterUsageException(
			    msg("invalidParam", param));
		}

		setDefaultEncoding(value);
	    } else {
		throw new ParameterUsageException(
		    msg("unknownParam", param));
	    }
	}
    }

    public String[] getParameters() {
	Vector parameters = new Vector();

	if (getUntabifyWidth() != Loader.DEFAULT_UNTABIFY_WIDTH) {
	    parameters.addElement("untabifyWidth");
	    parameters.addElement(Integer.toString(getUntabifyWidth()));
	}

	if (getTabifyWidth() != Formatter.DEFAULT_TABIFY_WIDTH) {
	    parameters.addElement("tabifyWidth");
	    parameters.addElement(Integer.toString(getTabifyWidth()));
	}

	if (!getLineSeparator().equals(Formatter.DEFAULT_LINE_SEPARATOR)) {
	    String lineSeparator = 
		StringUtil.replaceAll(getLineSeparator(), "\n", "\\n");
	    lineSeparator = 
		StringUtil.replaceAll(lineSeparator, "\r", "\\r");

	    parameters.addElement("lineSeparator");
	    parameters.addElement(lineSeparator);
	}

	if (getMaxLineLength() != Formatter.DEFAULT_MAX_LINE_LENGTH) {
	    parameters.addElement("maxLineLength");
	    parameters.addElement(Integer.toString(getMaxLineLength()));
	}

	if (getDefaultEncoding() != platformDefaultEncoding) {
	    parameters.addElement("defaultEncoding");
	    parameters.addElement(getDefaultEncoding());
	}

	String[] result = new String[parameters.size()];
	parameters.copyInto(result);
	return result;
    }

    public void setUntabifyWidth(int tabWidth) {
	loader.setUntabifyWidth(tabWidth);
    }

    public int getUntabifyWidth() {
	return loader.getUntabifyWidth();
    }

    public void setTabifyWidth(int tabWidth) {
	formatter.setTabifyWidth(tabWidth);
    }

    public int getTabifyWidth() {
	return formatter.getTabifyWidth();
    }

    public void setLineSeparator(String lineSeparator) {
	formatter.setLineSeparator(lineSeparator);
    }

    public String getLineSeparator() {
	return formatter.getLineSeparator();
    }

    public void setMaxLineLength(int maxLineLength) {
	formatter.setMaxLineLength(maxLineLength);
    }

    public int getMaxLineLength() {
	return formatter.getMaxLineLength();
    }

    public void setDefaultEncoding(String defaultEncoding) {
	this.defaultEncoding = defaultEncoding;
    }

    public String getDefaultEncoding() {
	return defaultEncoding;
    }

    public void setWorkInProgressObserver(WorkInProgressObserver observer) {
	loader.setObserver(observer);
    }

    public WorkInProgressObserver getWorkInProgressObserver() {
	return loader.getObserver();
    }

    public void loadDocument(String fileName, DocumentModel document) 
	throws IOException, StructuredFormatException {
	URL url = FileUtil.fileToURL(fileName);
	InputStream in = new FileInputStream(fileName);

	try {
	    loadDocument(in, url, null, document);
	} finally {
	    in.close();
	}
    }

    public void loadDocument(InputStream in, URL url, String javaEncoding, 
			     DocumentModel document) 
	throws IOException, StructuredFormatException {
	InputStreamReader reader;
	if (javaEncoding == null)
	    reader = new InputStreamReader(in, defaultEncoding);
	else
	    reader = new InputStreamReader(in, javaEncoding);

	loader.loadDocument(reader, url, document);
    }

    public void saveDocument(DocumentModel document, String fileName) 
	throws IOException, StructuredFormatException {
	URL url = FileUtil.fileToURL(fileName);
	OutputStream out = new FileOutputStream(fileName);

	try {
	    saveDocument(document, out, url, null);
	} finally {
	    out.close();
	}
    }

    public void saveDocument(DocumentModel document, OutputStream out, 
			     URL url, String javaEncoding) 
	throws IOException, StructuredFormatException {
	OutputStreamWriter writer = new OutputStreamWriter(out);
	if (javaEncoding == null)
	    writer = new OutputStreamWriter(out, defaultEncoding);
	else
	    writer = new OutputStreamWriter(out, javaEncoding);

	WorkInProgressObserver observer = loader.getObserver();
	if (observer != null)
	    observer.workInProgress(this, true);

	try {
	    formatter.writeDocument(document, writer, url);
	} finally {
	    if (observer != null)
		observer.workInProgress(this, false);
	}
    }

    // -----------------------------------------------------------------------
    // Main
    // -----------------------------------------------------------------------

    public static void main(String[] args) throws Exception {
	if (args.length != 2) {
	    System.err.println("usage: java com.xmlmind.javadocformat.Format" +
			       " java_src_file java_or_xml_dst_file");
	    System.exit(1);
	}

	Format format = new Format();

	Document document = new Document();
	try {
	    format.loadDocument(args[0], document);
	} catch (StructuredFormatException e) {
	    System.err.println(e.getLocation() + " line " + 
			       e.getLineNumber() + ", col. " + 
			       e.getColumnNumber() + ": " + e.getMessage());
	    System.exit(2);
	}

	String dstFileName = args[1];
	if (dstFileName.endsWith(".xml") || dstFileName.endsWith(".XML")) 
	    DocumentWriter.writeDocument(document, dstFileName,
					 false, -1, -1);
	else
	    format.saveDocument(document, dstFileName);
    }
}
