/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.javadocformat;

import java.io.*;
import java.net.URL;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Enumeration;
import fr.pixware.util.StringUtil;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.XMLUtil;
import com.xmlmind.xmledit.modeltool.ElementUtil;

public class Formatter {
    public static final int DEFAULT_TABIFY_WIDTH = 8;
    public static final String DEFAULT_LINE_SEPARATOR = "\n";
    public static final int DEFAULT_MAX_LINE_LENGTH = 78;

    private int tabWidth = DEFAULT_TABIFY_WIDTH;
    private String lineSeparator = DEFAULT_LINE_SEPARATOR;
    private int maxLineLength = DEFAULT_MAX_LINE_LENGTH;
    private ElementContent elementContent = null;

    // Formatting transient state.
    private int docIndent;
    private char requestedSpace;
    private int column;

    private static final String[] javadocTagNames = {
	"param", "@param",
	"exception", "@exception", 
	"return", "@return", 
	"see", "@see", 
	"seeref", "@see", 
	"seehref", "@see", 
	"author", "@author", 
	"version", "@version",
	"since", "@since", 
	"deprecated", "@deprecated", 
	"serialfield", "@serialField", 
	"serialdata", "@serialData", 
	"serial", "@serial"
    };
    // Do not add empty element names here.
    private static final String[] omittedEndTagNames = {
	"p",
	"dt",
	"dd",
	"li",
	"tr",
	"th",
	"td"
    };
    private static final String[] noAttributeValueNames = {
	"compact",
	"nowrap",
	"noshade",
	"ismap",
	"nohref",
	"checked",
	"multiple",
	"selected"
    };
    private static Hashtable javadocTags = new Hashtable();
    private static Hashtable omittedEndTags = new Hashtable();
    private static Hashtable noAttributeValues = new Hashtable();
    static {
	for (int i = 0; i < javadocTagNames.length; i += 2) {
	    String tag = javadocTagNames[i];
	    String literal = javadocTagNames[i+1];
	    javadocTags.put(tag, literal);
	    omittedEndTags.put(tag, tag);
	}

	for (int i = 0; i < omittedEndTagNames.length; ++i) {
	    String tag = omittedEndTagNames[i];
	    omittedEndTags.put(tag, tag);
	}

	for (int i = 0; i < noAttributeValueNames.length; ++i) {
	    String value = noAttributeValueNames[i];
	    noAttributeValues.put(value, value);
	}
    }

    // -----------------------------------------------------------------------

    public void setTabifyWidth(int tabWidth) {
	this.tabWidth = tabWidth;
    }

    public int getTabifyWidth() {
	return tabWidth;
    }

    public void setLineSeparator(String lineSeparator) {
	this.lineSeparator = lineSeparator;
    }

    public String getLineSeparator() {
	return lineSeparator;
    }

    public void setMaxLineLength(int maxLineLength) {
	this.maxLineLength = maxLineLength;
    }

    public int getMaxLineLength() {
	return maxLineLength;
    }

    public void writeDocument(DocumentModel document, Writer writer, URL url) 
	throws IOException {
	if (elementContent == null)
	    elementContent = ElementContent.get(document.getDocumentType());

	BufferedWriter out = new BufferedWriter(writer);

	SubelementModel sub = 
	    document.getRootElement().getFirstSubelement();

	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		writeSourceLines(((TextModel) sub).getTextChars(), out); 
	    } else {
		ElementModel elem = (ElementModel) sub;

		if (elem.getElementTypeName().equals("formfeed"))
		    out.write('\f');
		else
		    writeDoc(elem, out);
	    }

	    sub = sub.getNextSubelement();
	}

	out.flush();
    }

    // -----------------------------------------------------------------------

    private void writeSourceLines(char[] lines, Writer out) 
	throws IOException {
	if (tabWidth > 1) {
	    int charCount = lines.length;
	    int begin = 0;

	    for (int l = 0; l < charCount; ++l) {
		if (lines[l] == '\n') {
		    writeSourceLine(lines, begin, l, out);
		    out.write(lineSeparator);
		    begin = l+1;
		}
	    }

	    if (begin < charCount)
		writeSourceLine(lines, begin, charCount, out);
	} else {
	    out.write(lines);
	}
    }

    private void writeSourceLine(char[] line, int begin, int end, Writer out) 
	throws IOException {
	int indent = 0;
	for (int i = begin; i < end; ++i) {
	    if (line[i] != ' ')
		break;
	    ++indent;
	}
	if (indent > 0) {
	    writeIndentation(indent, out);
	    begin += indent;
	}

	out.write(line, begin, end - begin);
    }

    private void writeIndentation(int count, Writer out) throws IOException {
	if (tabWidth > 1) {
	    int tabCount = count / tabWidth;
	    count = count % tabWidth;

	    while (tabCount > 0) {
		out.write('\t');
		--tabCount;
	    }
	}

	while (count > 0) {
	    out.write(' ');
	    --count;
	}
    }

    private void writeDoc(ElementModel doc, Writer out) throws IOException {
	// Top line being '/**', min. indent cannot be 0.
	docIndent = findDocIndentation(doc) + 1;
	requestedSpace = 0;
	column = docIndent + 2; // docIndent + "* ";

	writeIndentation(docIndent-1, out);
	out.write("/**");

	SubelementModel sub = doc.getFirstSubelement();

	if (sub.getSubelementType() == SubelementModel.ELEMENT) {
	    ElementModel child = (ElementModel) sub;
	    String childName = child.getElementTypeName();
	    
	    if ("description".equals(childName)) {
		requestNewLine();
		writeBreakableElement(child, true, true, out);
		
		sub = sub.getNextSubelement();
		if (sub != null) {
		    // Add a blank line to separate description from javadoc
		    // tags.
		    requestNewLine();
		    writeText("", out);
		}
	    }
	}

	while (sub != null) {
	    requestNewLine();

	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		// Should not happen if valid.
		TextModel text = (TextModel) sub;
		writeWords(text.getTextChars(), out);
	    } else {
		ElementModel child = (ElementModel) sub;
		String childName = child.getElementTypeName();

		if (javadocTags.containsKey(childName)) {
		    if ("param".equals(childName))
			writeParam(child, out);
		    else if ("exception".equals(childName))
			writeException(child, out);
		    else if ("see".equals(childName))
			writeSee(child, out);
		    else if ("seeref".equals(childName))
			writeSeeRef(child, out);
		    else if ("seehref".equals(childName))
			writeSeeHref(child, out);
		    else if ("serialfield".equals(childName))
			writeSerialField(child, out);
		    else
			writeAnyJavadoc(child, out);
		} else {
		    // Should not happen if valid.
		    writeElement(child, out);
		}
	    }

	    sub = sub.getNextSubelement();
	}

	out.write(lineSeparator);
	writeIndentation(docIndent, out);
	out.write("*/");
	out.write(lineSeparator);
    }

    private static final int findDocIndentation(ElementModel doc) {
	int indent = 0;

	SubelementModel sub = doc.getNextSubelement();
	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		char[] chars = ((TextModel) sub).getTextChars();
		boolean isWhiteSpace = true;

		for (int i = 0; i < chars.length; ++i) {
		    char c = chars[i];

		    if (!XMLUtil.isXMLSpace(c)) {
			isWhiteSpace = false;
			break;
		    }

		    if (c == '\n')
			indent = 0;
		    else
			++indent;
		}

		if (isWhiteSpace)
		    // Just white space: continue.
		    indent = 0;
		else
		    break;
	    }

	    sub = sub.getNextSubelement();
	}

	return indent;
    }

    private void requestNewLine() {
	requestedSpace = '\n';
    }

    private void requestSpace() {
	requestedSpace = ' ';
    }

    private void writeWords(char[] chars, Writer out) throws IOException {
	StringTokenizer words = new StringTokenizer(escapeText(chars), 
						    " \t\r\n", true);
	while (words.hasMoreTokens()) {
	    String word = words.nextToken();

	    if (XMLUtil.isXMLSpace(word.charAt(0)))
		requestSpace();
	    else
		writeText(word, out);
	}
    }

    private static final String escapeText(char[] chars) {
	StringBuffer escaped = new StringBuffer();
	escapeText(chars, escaped);
	return escaped.toString();
    }
    
    private static final void escapeText(char[] chars, StringBuffer escaped) {
	// ISO Latin 1 entities are not escaped.
	for (int i = 0; i < chars.length; ++i) {
	    char c = chars[i];

	    switch (c) {
	    case '<':
		escaped.append("&lt;");
		break;
	    case '>':
		escaped.append("&gt;");
		break;
	    case '&':
		escaped.append("&amp;");
		break;
	    case 0x00A0:
		escaped.append("&nbsp;");
		break;
	    default:
		escaped.append(c);
	    }
	}
    }

    private void writeText(String text, Writer out) throws IOException {
	int length = text.length();

	writeRequestedSpace(length, out);

	out.write(text);
	column += length;
    }

    private void writeRequestedSpace(int length, Writer out) 
	throws IOException {
	switch (requestedSpace) {
	case ' ':
	    if (column + 1 + length <= maxLineLength) {
		out.write(' ');
		++column;
		requestedSpace = 0;
		break;
	    }
	    /*FALLTHROUGH*/
	case '\n':
	    out.write(lineSeparator);
	    writeIndentation(docIndent, out);
	    out.write("* ");
	    column = docIndent + 2;
	    requestedSpace = 0;
	    break;
	}
    }

    private void writeElement(ElementModel element, Writer out) 
	throws IOException {
	String elementName = element.getElementTypeName();
	
	if ("link".equals(elementName)) {
	    writeLink(element, out);
	} else {
	    if (!writeUnbreakableElement(element, out))
		writeBreakableElement(element, false, false, out);
	}
    }

    private boolean writeUnbreakableElement(ElementModel element, Writer out) 
	throws IOException {
	if (element.getSubelementCount() != 1 ||
	    element.getElementAttributeCount() > 0)
	    return false;

	SubelementModel sub = element.getFirstSubelement();
	if (sub.getSubelementType() != SubelementModel.TEXT)
	    return false;

	char[] chars = ((TextModel) sub).getTextChars();

	for (int i = 0; i < chars.length; ++i) {
	    if (XMLUtil.isXMLSpace(chars[i])) 
		return false;
	}

	StringBuffer unbreakable = new StringBuffer();
	unbreakable.append('<');
	String elementName = element.getElementTypeName();
	unbreakable.append(elementName);
	unbreakable.append('>');
	escapeText(chars, unbreakable);
	if (!omittedEndTags.containsKey(elementName)) {
	    unbreakable.append("</");
	    unbreakable.append(elementName);
	    unbreakable.append('>');
	}

	writeText(unbreakable.toString(), out);
	return true;
    }

    private void writeBreakableElement(ElementModel element, 
				       boolean omitStartTag,
				       boolean omitEndTag,
				       Writer out) throws IOException {
	if (!omitStartTag)
	    writeStartTag(element, out);

	boolean hasPreserveSpace = ElementUtil.hasPreserveSpace(element);
	String elementName = element.getElementTypeName();
	boolean isStructure = elementContent.isStructure(elementName);
	boolean prevChildIsBlockOrFlow = false;

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		TextModel text = (TextModel) sub;

		if (hasPreserveSpace) {
		    writeLines(text.getTextChars(), out);
		} else {
		    if (isStructure) {
			// Should not happen if valid.
			requestNewLine();
		    } else {
			if (prevChildIsBlockOrFlow)
			    requestNewLine();
			prevChildIsBlockOrFlow = false;
		    }

		    writeWords(text.getTextChars(), out);
		}
	    } else {
		ElementModel child = (ElementModel) sub;
		
		if (!hasPreserveSpace) {
		    if (isStructure) {
			requestNewLine();
		    } else {
			boolean childIsBlockOrFlow = 
			    elementContent.isBlockOrFlow(
				child.getElementTypeName());

			if (prevChildIsBlockOrFlow || childIsBlockOrFlow)
			    requestNewLine();
			prevChildIsBlockOrFlow = childIsBlockOrFlow;
		    }
		}

		writeElement(child, out);
	    }

	    sub = sub.getNextSubelement();
	}

	if (!omitEndTag)
	    omitEndTag = (omittedEndTags.containsKey(elementName) || 
			  elementContent.isEmpty(elementName));

	if (!omitEndTag) {
	    if (!hasPreserveSpace) {
		if (isStructure) {
		    requestNewLine();
		} else {
		    if (prevChildIsBlockOrFlow)
			requestNewLine();
		}
	    }

	    StringBuffer endTag = new StringBuffer();
	    endTag.append("</");
	    endTag.append(elementName);
	    endTag.append(">");
	    writeText(endTag.toString(), out);
	}
    }

    private void writeStartTag(ElementModel element, Writer out) 
	throws IOException {
	String elementName = element.getElementTypeName();

	writeText("<" + elementName, out);
	writeAttributes(element, out);
	writeText(">", out);
    }

    private void writeAttributes(ElementModel element, Writer out) 
	throws IOException {
	Enumeration iter = element.getElementAttributes();
	while (iter.hasMoreElements()) {
	    String[] pair = (String[]) iter.nextElement();

	    requestSpace();

	    StringBuffer attr = new StringBuffer();
	    attr.append(pair[0]);
	    if (!noAttributeValues.containsKey(pair[0])) {
		String attrValue = pair[1];
		int length = attrValue.length();

		attr.append("=\"");
		for (int i = 0; i < length; ++i) {
		    char c = attrValue.charAt(i);

		    switch (c) {
		    case '\"':
			attr.append("&quot;");
			break;
		    case '\'':
			attr.append("&apos;");
			break;
		    case '<':
			attr.append("&lt;");
			break;
		    case '>':
			attr.append("&gt;");
			break;
		    case '&':
			attr.append("&amp;");
			break;
		    case 0x00A0:
			attr.append("&nbsp;");
			break;
		    default:
			attr.append(c);
		    }
		}
		attr.append('\"');
	    }

	    writeText(attr.toString(), out);
	}
    }

    private void writeLines(char[] text, Writer out) throws IOException {
	String[] lines = StringUtil.split(escapeText(text), '\n');

	for (int i = 0; i < lines.length; ++i) {
	    String line = lines[i];

	    if (i == 0) {
		writeRequestedSpace(line.length(), out);
	    } else {
		out.write(lineSeparator);
		writeIndentation(docIndent, out);
		out.write('*');
		column = docIndent + 1;
	    }

	    out.write(line);
	    column += line.length();
	}
    }

    private void writeLink(ElementModel element, Writer out) 
	throws IOException {
	writeText("{@link", out);

	requestSpace();
	SubelementModel sub = element.getFirstSubelement();
	if (sub != null && sub.getSubelementType() == SubelementModel.TEXT) {
	    TextModel text = (TextModel) sub;
	    writeText(escapeText(text.getTextChars()), out);
	}

	String label = element.getElementAttribute("label");
	if (label != null) {
	    requestSpace();
	    writeWords(label.toCharArray(), out);
	}
	
	writeText("}", out);
    }

    private void writeSee(ElementModel element, Writer out) 
	throws IOException {
	writeText("@see", out);

	requestSpace();
	SubelementModel sub = element.getFirstSubelement();
	if (sub != null && sub.getSubelementType() == SubelementModel.TEXT) {
	    TextModel text = (TextModel) sub;
	    writeText(escapeText(text.getTextChars()), out);
	}

	String label = element.getElementAttribute("label");
	if (label != null) {
	    requestSpace();
	    writeWords(label.toCharArray(), out);
	}
    }

    private void writeSeeRef(ElementModel element, Writer out) 
	throws IOException {
	writeText("@see", out);

	requestSpace();
	writeText("\"", out);
	SubelementModel sub = element.getFirstSubelement();
	if (sub != null && sub.getSubelementType() == SubelementModel.TEXT) {
	    TextModel text = (TextModel) sub;
	    writeWords(text.getTextChars(), out);
	}
	writeText("\"", out);
    }

    private void writeSeeHref(ElementModel element, Writer out) 
	throws IOException {
	writeText("@see", out);

	requestSpace();
	writeText("<a", out);
	writeAttributes(element, out);
	writeText(">", out);
	writeBreakableElement(element, true, true, out);
	writeText("</a>", out);
    }

    private void writeParam(ElementModel element, Writer out) 
	throws IOException {
	writeText("@param", out);

	SubelementModel sub = element.getFirstSubelement();
	if (sub == null ||
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    return;

	// paramname
	requestSpace();
	writeBreakableElement((ElementModel) sub, true, true, out);

	sub = sub.getNextSubelement();
	if (sub == null ||
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    return;

	// paramdesc
	requestSpace();
	writeBreakableElement((ElementModel) sub, true, true, out);
    }

    private void writeException(ElementModel element, Writer out) 
	throws IOException {
	writeText("@exception", out);

	SubelementModel sub = element.getFirstSubelement();
	if (sub == null ||
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    return;

	// exceptionname
	requestSpace();
	writeBreakableElement((ElementModel) sub, true, true, out);

	sub = sub.getNextSubelement();
	if (sub == null ||
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    return;

	// exceptiondesc
	requestSpace();
	writeBreakableElement((ElementModel) sub, true, true, out);
    }

    private void writeSerialField(ElementModel element, Writer out) 
	throws IOException {
	writeText("@serialField", out);

	SubelementModel sub = element.getFirstSubelement();
	if (sub == null ||
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    return;

	// fieldname
	requestSpace();
	writeBreakableElement((ElementModel) sub, true, true, out);

	sub = sub.getNextSubelement();
	if (sub == null ||
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    return;

	// fieldtype
	requestSpace();
	writeBreakableElement((ElementModel) sub, true, true, out);

	sub = sub.getNextSubelement();
	if (sub == null ||
	    sub.getSubelementType() != SubelementModel.ELEMENT)
	    return;

	// fielddesc
	requestSpace();
	writeBreakableElement((ElementModel) sub, true, true, out);
    }

    private void writeAnyJavadoc(ElementModel element, Writer out) 
	throws IOException {
	String elementName = element.getElementTypeName();
	String literal = (String) javadocTags.get(elementName);

	writeText(literal, out);
	requestSpace();
	writeBreakableElement(element, true, true, out);
    }
}

