/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.javadocformat;

import java.io.*;
import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.net.URL;
import com.jclark.xml.sax.ReaderInputStream;
import com.jclark.xml.parse.OpenEntity;
import com.jclark.xml.parse.NotWellFormedException;
import fr.pixware.util.FileUtil;
import fr.pixware.util.WorkInProgressObserver;
import com.xmlmind.structformat.StructuredFormatException;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.StyleSheetInfo;
import com.xmlmind.xmledit.modelimpl.XPDocumentLoader;
import com.xmlmind.xmledit.modelimpl.Document;
import com.xmlmind.cssparser.ParseUtil;

public class Loader {
    public static final int DEFAULT_UNTABIFY_WIDTH = 8;
    
    private XPDocumentLoader docLoader = new XPDocumentLoader();
    private WorkInProgressObserver observer = null;
    private int tabWidth = DEFAULT_UNTABIFY_WIDTH;

    private StyleSheetInfo[] styleSheetInfo = null;
    private String internalSubset = null;
    private ElementContent elementContent = null;
    
    // -----------------------------------------------------------------------

    public void setObserver(WorkInProgressObserver observer) {
	this.observer = observer;
	docLoader.setObserver(observer);
    }

    public WorkInProgressObserver getObserver() {
	return observer;
    }

    public void setUntabifyWidth(int tabWidth) {
	this.tabWidth = tabWidth;
    }

    public int getUntabifyWidth() {
	return tabWidth;
    }

    public void loadDocument(Reader in, URL url, DocumentModel document) 
	throws IOException, StructuredFormatException {
	if (styleSheetInfo == null)
	    loadResources();

	Filter filter = new Filter(in, internalSubset, elementContent, 
				   tabWidth);
	OpenEntity entity = 
	    new OpenEntity(new ReaderInputStream(filter), 
			   url.toExternalForm(), url);

	try {
	    docLoader.load(entity, document);
	} catch (InvalidFormatException e1) {
	    throw new StructuredFormatException(
		url.toExternalForm(), filter.getLineNumber(),
		/*columnNumber*/ 1, e1.getMessage());
	} catch (NotWellFormedException e2) {
	    throw new StructuredFormatException(
		e2.getEntityLocation(), /*lineNum*/ -1, /*colNum*/ -1, 
		msg("xmlError", e2.getLineNumber(), e2.getColumnNumber(), 
		    e2.getMessageWithoutLocation()));
	}

	trim(document.getRootElement());

	document.putDocumentProperty(DocumentModel.URL_PROPERTY, url);
	document.putDocumentProperty(StyleSheetInfo.PROPERTY, styleSheetInfo);
    }

    private void loadResources() throws IOException {
	URL url = getClass().getResource("javadoc.css");
	loadStyleSheet(url);

	url = getClass().getResource("javadoc.xml");
	loadInternalSubset(url);
	loadDTD(url);
    }

    private void loadStyleSheet(URL url) throws IOException {
	StyleSheetInfo info = new StyleSheetInfo();
	info.content = ParseUtil.loadSource(url);
	info.url = url.toExternalForm();
	info.type = "text/css";
	info.title = "Javadoc";
	info.alternate = false;
	styleSheetInfo = new StyleSheetInfo[] { info };
    }

    private void loadInternalSubset(URL url) throws IOException {
	String template = FileUtil.loadString(url);

	int begin = template.indexOf("<!DOCTYPE java [");
	int end = template.lastIndexOf("]>");

	internalSubset = template.substring(begin+16, end);
    }

    private void loadDTD(URL url) throws IOException {
	Document template = new Document();
	docLoader.load(url, template);
	elementContent = ElementContent.get(template.getDocumentType());
    }

    public void trim(ElementModel element) {
	boolean canTrim = elementContent.isTrimableTextContainer(
	    element.getElementTypeName());

	SubelementModel previous = null;
	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    SubelementModel next = sub.getNextSubelement();

	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		if (canTrim) {
		    TextModel text = (TextModel) sub;

		    if (isBlockOrFlow(elementContent, previous)) {
			// Remember that consecutive white spaces have been
			// compressed. Suffice to remove the first white space
			// if any.

			if (text.getTextLength() > 0 && 
			    text.getTextChar(0) == ' ')
			    text.deleteText(0, 1);
		    }

		    if (isBlockOrFlow(elementContent, next)) {
			int last = text.getTextLength()-1;
			if (last >= 0 && text.getTextChar(last) == ' ')
			    text.deleteText(last, 1);
		    }

		    if (text.getTextLength() == 0 &&
			element.getSubelementCount() > 1)
			element.removeSubelement(text);
		}
	    } else {
		trim((ElementModel) sub);
	    }

	    previous = sub;
	    sub = next;
	}
    }

    private static final boolean isBlockOrFlow(ElementContent elementContent,
					       SubelementModel sub) {
	if (sub == null)
	    return true;

	if (sub.getSubelementType() == SubelementModel.TEXT)
	    return false;

	ElementModel child = (ElementModel) sub;
	return elementContent.isBlockOrFlow(child.getElementTypeName());
    }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.javadocformat.LoaderRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id, 
			      int arg0, int arg1, Object arg2) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { 
	    new Integer(arg0), new Integer(arg1), arg2 
	});
    }

    // -----------------------------------------------------------------------
    // Test
    // -----------------------------------------------------------------------

    public static void main(String[] args) throws IOException {
	if (args.length != 1) {
	    System.err.println(
		"usage: java com.xmlmind.javadocformat.Loader java_file");
	    System.exit(1);
	}

	Loader loader = new Loader();
	loader.loadResources();

	FileReader in = new FileReader(args[0]);
	Filter filter = 
	    new Filter(in, loader.internalSubset, loader.elementContent, 
		       loader.tabWidth);
	
	char[] chars = new char[8192];
	int count;

	try {
	    while ((count = filter.read(chars, 0, chars.length)) != -1) {
		System.out.print(new String(chars, 0, count));
		System.out.flush();
	    }
	} catch (InvalidFormatException e) {
	    System.err.println(args[0] + ", line " + filter.getLineNumber() + 
			       ": " + e.getMessage());
	}

	filter.close();
    }
}
