/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.javadocformat;

import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.util.Vector;
import com.xmlmind.xmledit.modeltool.XMLUtil;

public class Tokenizer {
    public static final int EOF       = -1;
    public static final int START_TAG = 0;
    public static final int END_TAG   = 1;
    public static final int TEXT      = 2;

    private String source;
    private int pos;
    private int endPos;

    private int token;
    private String elementName;
    private String[] elementAttributes;
    private String text;

    private static final String[] NO_ATTRIBUTES = new String[0];
    private static final String NO_VALUE = "";

    // -----------------------------------------------------------------------

    public Tokenizer(String source) {
	this.source = source;
	pos = 0;
	endPos = source.length();
    }

    public int nextToken() throws InvalidFormatException {
	// EOF ----------

	if (pos >= endPos) {
	    text = null;
	    elementName = null;
	    elementAttributes = null;
	    token = EOF;
	    return token;
	}

	// TEXT ----------
	// Note that comments ('<!') and PIs ('<?') are treated as text.

	int startText = pos;
	char c;

	while (pos < endPos) {
	    if (source.charAt(pos) == '<' && pos+1 < endPos &&
		((c = source.charAt(pos+1)) == '/' || isNameStartChar(c)))
		break;

	    ++pos;
	}

	if (pos > startText) {
	    text = source.substring(startText, pos);
	    elementName = null;
	    elementAttributes = null;
	    token = TEXT;
	    return token;
	}

	// END_TAG ----------

	if (source.charAt(pos+1) == '/') {
	    pos += 2; // skip '</'
	    int startName = pos;

	    while (pos < endPos) {
		if (source.charAt(pos) == '>')
		    break;

		++pos;
	    }
	    if (pos >= endPos)
		throw new InvalidFormatException(msg("invalidEndTag1"));

	    elementName = source.substring(startName, pos).trim();
	    if (!XMLUtil.isName(elementName))
		throw new InvalidFormatException(
		    msg("invalidEndTag2", elementName));

	    ++pos; // skip '>'
	    text = null;
	    elementAttributes = null;
	    token = END_TAG;
	    return token;
	}

	// START_TAG ----------

	++pos; // skip '<'
	int startName = pos;

	++pos;
	while (pos < endPos) {
	    if ((c = source.charAt(pos)) == '>' || XMLUtil.isXMLSpace(c))
		break;

	    ++pos;
	}
	if (pos >= endPos)
	    throw new InvalidFormatException(msg("invalidStartTag1"));

	elementName = source.substring(startName, pos);
	if (!XMLUtil.isName(elementName))
	    throw new InvalidFormatException(
		msg("invalidStartTag2", elementName));

	int startAttributes = pos;
	while (pos < endPos) {
	    if (source.charAt(pos) == '>')
		break;

	    ++pos;
	}
	if (pos >= endPos)
	    throw new InvalidFormatException(msg("invalidStartTag1"));

	if (pos > startAttributes)
	    elementAttributes = 
		parseAttributes(source.substring(startAttributes, pos));
	else
	    elementAttributes = NO_ATTRIBUTES;
	    
	++pos; // skip '>'
	text = null;
	token = START_TAG;
	return token;
    }

    private static final boolean isNameStartChar(char c) {
	switch (c) {
	case '_': case ':':
	    break;
	default:
	    if (!Character.isLetter(c))
		return false;
	}
	return true;
    }

    private static final String[] parseAttributes(String text) 
	throws InvalidFormatException {
	Vector list = new Vector();
	int length = text.length();
	int pos = 0;

	for (;;) {
	    // Skip space ---

	    while (pos < length) {
		if (!XMLUtil.isXMLSpace(text.charAt(pos)))
		    break;
		++pos;
	    }

	    if (pos >= length)
		break;

	    // Attribute name ---

	    char c;

	    int startName = pos;
	    while (pos < length) {
		if ((c = text.charAt(pos)) == '=' || XMLUtil.isXMLSpace(c))
		    break;

		++pos;
	    }

	    String attrName;
	    if (pos > startName)
		attrName = text.substring(startName, pos);
	    else
		attrName = "";
	    if (!XMLUtil.isName(attrName))
		throw new InvalidFormatException(
		    msg("invalidAttributeName", attrName));

	    // Attribute value ---

	    String attrValue = NO_VALUE;

	    if (pos < length && text.charAt(pos) == '=') {
		if (pos+1 >= length)
		    throw new InvalidFormatException(msg("noAttributeValue"));
		++pos; // skip '='
		c = text.charAt(pos);

		switch (c) {
		case '\"': 
		case '\'':
		    {
			char quote = c;

			++pos; // skip quote
			int startValue = pos;
			while (pos < length) {
			    if (text.charAt(pos) == quote)
				break;

			    ++pos;
			}
			if (pos >= length)
			    throw new InvalidFormatException(
				msg("missingQuote"));

			attrValue = text.substring(startValue, pos);
			++pos; // skip quote
		    }
		    break;

		default:
		    {
			if (XMLUtil.isXMLSpace(c))
			    throw new InvalidFormatException(
				msg("noAttributeValue"));

			int startValue = pos;
			while (pos < length) {
			    if (XMLUtil.isXMLSpace(text.charAt(pos)))
				break;

			    ++pos;
			}

			attrValue = text.substring(startValue, pos);
		    }
		}
	    }

	    list.addElement(attrName);
	    list.addElement(attrValue);
	}
	
	String[] attributes = new String[list.size()];
	list.copyInto(attributes);
	for (int i = 1; i < attributes.length; i += 2) {
	    if (attributes[i] == NO_VALUE)
		attributes[i] = null;
	}
	    
	return attributes;
    }

    public int getToken() {
	return token;
    }

    public String getElementName() {
	return elementName;
    }

    public String[] getElementAttributes() {
	return elementAttributes;
    }

    public String getText() {
	return text;
    }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.javadocformat.TokenizerRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    public static String tokenName(int token) {
	switch (token) {
	case EOF:
	    return msg("eof");
	case START_TAG:
	    return msg("startTag");
	case END_TAG:
	    return msg("endTag");
	case TEXT:
	    return msg("text");
	default:
	    return "???" + token;
	}
    }

    // -----------------------------------------------------------------------
    // Test
    // -----------------------------------------------------------------------

    public static void main(String[] args) throws java.io.IOException {
	if (args.length != 1) {
	    System.err.println(
		"usage: java com.xmlmind.javadocformat.Tokenizer xml_file");
	    System.exit(1);
	}

	String content = fr.pixware.util.FileUtil.loadString(args[0]);
	Tokenizer tokens = new Tokenizer(content);
	int token = tokens.nextToken();

	while (token != EOF) {
	    switch (token) {
	    case START_TAG:
		{
		    System.out.print("<" + tokens.getElementName());

		    String[] attributes = tokens.getElementAttributes();
		    for (int i = 0; i < attributes.length; i += 2) {
			String attrName = attributes[i];
			String attrValue = attributes[i+1];

			System.out.print(' ');
			System.out.print(attrName);
			if (attrValue != null) {
			    System.out.print("=\"");
			    System.out.print(attrValue);
			    System.out.print('\"');
			}
		    }

		    System.out.print('>');
		}
		break;

	    case END_TAG:
		System.out.print("</" + tokens.getElementName() + ">");
		break;

	    case TEXT:
		System.out.print(tokens.getText());
		break;
	    }

	    token = tokens.nextToken();
	}

	System.out.flush();
    }
}
