/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.namere;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Hashtable;

public final class NameRE {
    // -----------------------------------------------------------------------
    // Exception
    // -----------------------------------------------------------------------

    public static class InvalidParticleException extends Exception {
	private Particle particle;

	public InvalidParticleException(Particle particle, String message) {
	    super(message);
	    this.particle = particle;
	}

	public Particle getParticle() {
	    return particle;
	}
    }

    // -----------------------------------------------------------------------
    // Program instructions
    // -----------------------------------------------------------------------

    private static final int MATCH_NAME              = 0;
    private static final int MATCH_ALL_NAMES         = 1;
    private static final int CHOICE                  = 2;
    private static final int INIT_UNCONDITIONAL_LOOP = 3;
    private static final int UNCONDITIONAL_LOOP      = 4;
    private static final int NOTHING                 = 5;
    private static final int INIT_LOOP               = 6;
    private static final int LOOP                    = 7;
    private static final int FAIL                    = 8;

    private static final String[] opcodeNames = {
	"MATCH_NAME",
	"MATCH_ALL_NAMES",
	"CHOICE",
	"INIT_UNCONDITIONAL_LOOP",
	"UNCONDITIONAL_LOOP",
	"NOTHING",
	"INIT_LOOP",
	"LOOP",
	"FAIL"
    };

    public static class Program {
	public byte opcode;
	public Program next;

	public Program(int opcode) {
	    this.opcode = (byte) opcode;
	}

	public final Program tail() {
	    Program p = this;
	    while (p.next != null)
		p = p.next;
	    return p;
	}

	protected static void toString(Program start, Program end,
				       StringBuffer buffer, Hashtable labels, 
				       int indent) {
	    while (start != end) {
		start.toString(buffer, labels, indent);
		start = start.next;
	    }
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    String label = getLabel(this, labels);
	    buffer.append(label);
	    appendSpace(buffer, (8 - label.length()) + indent);
	    buffer.append(opcodeNames[opcode]);
	    buffer.append(" -> ");
	    buffer.append(getLabel(next, labels));
	    buffer.append('\n');
	}

	protected static String getLabel(Program program, Hashtable labels) {
	    if (program == null)
		return "/";

	    String label = (String) labels.get(program);
	    if (label == null) {
		label = Integer.toString(labels.size());
		labels.put(program, label);
	    }
	    return label;
	}

	protected static void appendSpace(StringBuffer buffer, int space) {
	    while (space > 0) {
		buffer.append(' ');
		--space;
	    }
	}
    }

    private static final class MatchName extends Program {
	public String name;

	public MatchName(String name) {
	    super(MATCH_NAME);
	    this.name = name;
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    super.toString(buffer, labels, indent);
	    appendSpace(buffer, 8+indent);
	    buffer.append(name);
	    buffer.append('\n');
	}
    }

    private static final class MatchAllNames extends Program {
	public String[] list;
	public boolean[] optional;

	public MatchAllNames(String[] list, boolean[] optional) {
	    super(MATCH_ALL_NAMES);
	    this.list = list;
	    this.optional = optional;
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    super.toString(buffer, labels, indent);
	    appendSpace(buffer, 8+indent);
	    for (int i = 0; i < list.length; ++i) {
		if (i > 0)
		    buffer.append(' ');
		buffer.append(list[i]);
		if (optional[i])
		    buffer.append('?');
	    }
	    buffer.append('\n');
	}
    }

    private static final class Choice extends Program {
	public Program[] choices;

	public Choice() {
	    super(CHOICE);
	    choices = new Program[0];
	}

	public void add(Program program) {
	    Program[] choices2 = new Program[choices.length+1];

	    System.arraycopy(choices, 0, choices2, 0, choices.length);
	    choices2[choices.length] = program;

	    choices = choices2;
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    super.toString(buffer, labels, indent);
	    for (int i = 0; i < choices.length; ++i)
		Program.toString(choices[i], next, buffer, labels, indent+4);
	}
    }

    private static final class UnconditionalLoop extends Program {
	public int start;
	public Program begin;

	public UnconditionalLoop(Program begin) {
	    super(UNCONDITIONAL_LOOP);
	    this.begin = begin;
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    super.toString(buffer, labels, indent);
	    appendSpace(buffer, 8+indent);
	    buffer.append("begin=");
	    buffer.append(getLabel(begin, labels));
	    buffer.append('\n');
	}
    }

    private static final class InitUnconditionalLoop extends Program {
	public UnconditionalLoop loop;

	public InitUnconditionalLoop() {
	    super(INIT_UNCONDITIONAL_LOOP);
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    super.toString(buffer, labels, indent);
	    appendSpace(buffer, 8+indent);
	    buffer.append("loop=");
	    buffer.append(getLabel(loop, labels));
	    buffer.append('\n');
	}
    }

    private static final class Loop extends Program {
	public int counter;
	public int maxCounter;
	public int start;
	public Program begin;

	public Loop(int maxCounter, Program begin) {
	    super(LOOP);
	    this.maxCounter = maxCounter;
	    this.begin = begin;
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    super.toString(buffer, labels, indent);
	    appendSpace(buffer, 8+indent);
	    buffer.append("max=");
	    buffer.append(maxCounter);
	    buffer.append(" begin=");
	    buffer.append(getLabel(begin, labels));
	    buffer.append('\n');
	}
    }

    private static final class InitLoop extends Program {
	public Loop loop;

	public InitLoop() {
	    super(INIT_LOOP);
	}

	protected void toString(StringBuffer buffer, Hashtable labels, 
				int indent) {
	    super.toString(buffer, labels, indent);
	    appendSpace(buffer, 8+indent);
	    buffer.append("loop=");
	    buffer.append(getLabel(loop, labels));
	    buffer.append('\n');
	}
    }

    // -----------------------------------------------------------------------
    // NameRE
    // -----------------------------------------------------------------------

    private Program program;
    private boolean[] matched = null;

    public NameRE(Particle particle) throws InvalidParticleException {
	program = compile(particle);

	int length = matchedArrayLength(program, null, 0);
	if (length > 0)
	    matched = new boolean[length];
    }

    private static int matchedArrayLength(Program program, Program end, 
					  int maxLength) {
	while (program != end) {
	    switch (program.opcode) {
	    case MATCH_ALL_NAMES:
	        {
		    int length = ((MatchAllNames) program).list.length;

		    if (length > maxLength)
			maxLength = length;
		}
		break;
	    case CHOICE:
	        {
		    Program[] choices = ((Choice) program).choices;
		    Program continuation = program.next;

		    for (int i = 0; i < choices.length; ++i) {
			int length = 
			    matchedArrayLength(choices[i], continuation, 
					       maxLength);
			if (length > maxLength)
			    maxLength = length;
		    }
	        }
		break;
	    }

	    program = program.next;
	}

	return maxLength;
    }

    public String toString() {
	StringBuffer buffer = new StringBuffer();
	Hashtable labels = new Hashtable();

	Program.toString(program, null, buffer, labels, 0);
	    
	return buffer.toString();
    }

    // -----------------------------------------------------------------------
    // Compiler
    // -----------------------------------------------------------------------

    private static Program compile(Particle particle) 
	throws InvalidParticleException {
	Program program = doCompile(particle);

	fixUpChoices(program, null);

	return program;
    }

    private static void fixUpChoices(Program program, Program end) {
	while (program != end) {
	    if (program.opcode == CHOICE)
		fixUpChoice((Choice) program);

	    program = program.next;
	}
    }

    private static void fixUpChoice(Choice choice) {
	Program[] choices = choice.choices;
	Program continuation = choice.next;

	for (int i = 0; i < choices.length; ++i)
	    choices[i].tail().next = continuation;

	for (int i = 0; i < choices.length; ++i)
	    fixUpChoices(choices[i], continuation);
    }

    private static Program doCompile(Particle particle) 
	throws InvalidParticleException {
	int minOccurs = particle.getMinOccurs();
	int maxOccurs = particle.getMaxOccurs();
	if (minOccurs < 0 || 
	    maxOccurs == 0 ||
	    (maxOccurs != Particle.UNBOUNDED && minOccurs > maxOccurs))
	    throw new InvalidParticleException(particle,
					       msg("invalidMinMaxOccurs"));

	Program program = null;
	
	if (minOccurs > 0) {
	    Program sub = compileParticle(particle);

	    if (minOccurs == 1) {
		program = sub;
	    } else {
		InitLoop initLoop = new InitLoop();
		program = initLoop;
		program.next = sub;

		Loop loop = new Loop(minOccurs-1, sub);
		initLoop.loop = loop;
		sub.tail().next = loop;
	    }
	}

	Program program2 = null;

	if (maxOccurs == Particle.UNBOUNDED) {
	    InitUnconditionalLoop initLoop = new InitUnconditionalLoop();
	    program2 = initLoop;
	    Choice choice = new Choice();
	    program2.next = choice;

	    Program sub = compileParticle(particle);
	    UnconditionalLoop loop = new UnconditionalLoop(choice);
	    initLoop.loop = loop;
	    sub.tail().next = loop;

	    choice.add(sub);
	    choice.add(new Program(NOTHING));
	} else {
	    int optionalCount = maxOccurs - minOccurs;

	    if (optionalCount == 1) {
		Choice choice = new Choice();
		program2 = choice;

		choice.add(compileParticle(particle));
		choice.add(new Program(NOTHING));
	    } else if (optionalCount > 1) {
		InitLoop initLoop = new InitLoop();
		program2 = initLoop;
		Choice choice = new Choice();
		program2.next = choice;

		Program sub = compileParticle(particle);
		Loop loop = new Loop(optionalCount, choice);
		initLoop.loop = loop;
		loop.next = new Program(FAIL);
		sub.tail().next = loop;

		choice.add(sub);
		choice.add(new Program(NOTHING));
	    }
	}

	if (program == null)
	    program = program2;
	else
	    program.tail().next = program2;

	return program;
    }

    private static Program compileParticle(Particle particle) 
	throws InvalidParticleException {
	switch (particle.getType()) {
	case Particle.CHOICE:
	    return compileChoiceParticle((ChoiceParticle) particle);

	case Particle.SEQUENCE:
	    return compileSequenceParticle((SequenceParticle) particle);

	case Particle.ALL:
	    return compileAllParticle((AllParticle) particle);

	case Particle.ELEMENT:
	    return new MatchName(((ElementParticle)particle).getElementName());

	default:
	    throw new RuntimeException("internal error: unknown particle " + 
				       particle.getType());
	}
    }

    private static Program compileChoiceParticle(ChoiceParticle particle) 
	throws InvalidParticleException {
	Particle[] particles = particle.getParticles();
	int particleCount = particles.length;
	if (particleCount == 0)
	    return new Program(NOTHING);

	if (particleCount == 1)
	    return doCompile(particles[0]);

	Choice choice = new Choice();

	for (int i = 0; i < particleCount; ++i)
	    choice.add(doCompile(particles[i]));

	return choice;
    }

    private static Program compileSequenceParticle(SequenceParticle particle) 
	throws InvalidParticleException {
	Particle[] particles = particle.getParticles();
	int particleCount = particles.length;
	if (particleCount == 0)
	    return new Program(NOTHING);

	Program program = doCompile(particles[0]);

	Program tail = program.tail();
	for (int i = 1; i < particleCount; ++i) {
	    Program sub = doCompile(particles[i]);
	    tail.next = sub;
	    tail = sub.tail();
	}

	return program;
    }

    private static Program compileAllParticle(AllParticle particle) 
	throws InvalidParticleException {
	Particle[] particles = particle.getParticles();
	int particleCount = particles.length;
	if (particleCount == 0)
	    return new Program(NOTHING);

	// Do not try to optimize the case with 1 particle.

	String[] list = new String[particleCount];
	boolean[] optional = new boolean[particleCount];

	for (int i = 0; i < particleCount; ++i) {
	    Particle p = particles[i];
	    int minOccurs = p.getMinOccurs();
	    int maxOccurs = p.getMaxOccurs();

	    if (p.getType() != Particle.ELEMENT ||
		minOccurs > 1 || maxOccurs != 1)
		throw new InvalidParticleException(particle, 
						   msg("badAllParticle"));

	    list[i] = ((ElementParticle) p).getElementName();
	    optional[i] = (minOccurs == 0);
	}
	    
	return new MatchAllNames(list, optional);
    }

    // -----------------------------------------------------------------------
    // Matcher
    // -----------------------------------------------------------------------

    public boolean match(String[] names, int nameCount) {
	// *** ALL NAMES MUST HAVE BEEN INTERNED ***
	return match(names, 0, nameCount, program);
    }

    private boolean match(String[] names, int start, int end,
			  Program program) {
	while (program != null) {
	    Program next = program.next;

	    switch (program.opcode) {
	    case MATCH_NAME:
		if (start == end || ((MatchName) program).name != names[start])
		    return false;
		++start;
		break;

	    case MATCH_ALL_NAMES:
	        {
		    MatchAllNames allNames = (MatchAllNames) program;
		    String[] list = allNames.list;
		    int nameCount = list.length;
		    boolean[] optional = allNames.optional;

		    for (int i = 0; i < nameCount; ++i)
			matched[i] = false;

		    for (int i = 0; i < nameCount; ++i) {
			if (start == end)
			    break;

			String name = names[start];
			boolean found = false;

			for (int j = 0; j < nameCount; ++j) {
			    if (!matched[j] && list[j] == name) {
				matched[j] = found = true;
				break;
			    }
			}
			
			if (!found)
			    break;

			++start;
		    }

		    for (int i = 0; i < nameCount; ++i) {
			if (!matched[i] && !optional[i])
			    return false;
		    }
		}
		break;

	    case CHOICE:
	        {
		    Program[] choices = ((Choice) program).choices;
		    for (int i = 0; i < choices.length; ++i) {
			Program choice = choices[i];

			if (choice.opcode == MATCH_NAME &&
			    (start == end || 
			     ((MatchName) choice).name != names[start]))
			    continue;

			if (match(names, start, end, choice))
			    return true;
		    }

		    return false;
	        }

	    case UNCONDITIONAL_LOOP:
	        {		    
		    UnconditionalLoop loop = (UnconditionalLoop) program;

		    if (start > loop.start) {
			loop.start = start;
			next = loop.begin;
		    }
		}
		break;

	    case INIT_UNCONDITIONAL_LOOP:
		((InitUnconditionalLoop) program).loop.start = start;
		break;

	    case NOTHING:
		break;

	    case INIT_LOOP:
	        {
		    Loop loop = ((InitLoop) program).loop;
		    loop.start = start;
		    loop.counter = 0;
		}
		break;

	    case LOOP:
	        {		    
		    Loop loop = (Loop) program;

		    if (start > loop.start &&
			loop.counter < loop.maxCounter) {
			loop.start = start;
			++loop.counter;
			next = loop.begin;
		    }
		}
		break;

	    case FAIL:
		return false;

	    default:
		throw new RuntimeException("internal error: unknown opcode " + 
					   program.opcode);
	    }

	    program = next;
	}

	return (start == end);
    }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.namere.NameRERes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    // -----------------------------------------------------------------------
    // Test
    // -----------------------------------------------------------------------

    public static void main(String[] args) throws Exception {
	if (args.length < 2) {
	    System.err.println(
		"usage: java com.xmlmind.namere.NameRE" + 
		" t|f nameRE name ... name");
	    System.exit(1);
	}

	boolean expected = (args[0].length() > 0 && args[0].charAt(0) == 't');

	String[] names = new String[args.length-2];
	for (int i = 2; i < args.length; ++i)
	    names[i-2] = args[i].intern();

	ParticleParser parser = new ParticleParser(false);

	Particle particle;
	if (args[1].indexOf("#PCDATA") >= 0)
	    particle = parser.parseMixedContent(args[1]);
	else
	    particle = parser.parseElementContent(args[1]);

	NameRE re = new NameRE(particle);

	if (re.match(names, names.length) != expected) {
	    System.err.println(re);
	    System.exit(2);
	}
    }
}
