/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.namere;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;
import fr.pixware.util.LinearHashtable;

public final class ParticleParser {
    // -----------------------------------------------------------------------
    // Exception
    // -----------------------------------------------------------------------

    public static class ParseException extends Exception {
	public ParseException(String message) {
	    super(message);
	}
    }

    // -----------------------------------------------------------------------
    // Particle parser
    // -----------------------------------------------------------------------

    private boolean strictSyntax;

    private String lookahead;

    private static final int PARENTHESE              = 0;
    private static final int CONNECTOR_OR_PARENTHESE = 1;
    private static final int PCDATA                  = 2;
    private static final int NAME                    = 3;
    private static final int TERM                    = 4;
    private static final int OCCUR                   = 5;
    private static final int END                     = 6;

    private static final Particle[] NO_PARTICLES = new Particle[0];

    public ParticleParser(boolean strictSyntax) {
	this.strictSyntax = strictSyntax;
    }

    public boolean isStrictSyntax() {
	return strictSyntax;
    }

    public Particle parseMixedContent(String spec) throws ParseException {
	StringTokenizer tokens = new StringTokenizer(spec, "(|)*", true);
	int state = PARENTHESE;
	LinearHashtable names = new LinearHashtable();
	boolean hasNames = false;

	while (tokens.hasMoreTokens()) {
	    String token = tokens.nextToken();
	    char c;

	    switch (state) {
	    case PARENTHESE:
		if (!token.equals("("))
		    throw new ParseException(
			msg("unexpectedToken", "(", token));
		state = PCDATA;
		break;

	    case CONNECTOR_OR_PARENTHESE:
		if (token.length() != 1 ||
		    ((c = token.charAt(0)) != '|' && c != ')'))
		    throw new ParseException(
			msg("unexpectedToken", "|)", token));
		state = (c == '|')? NAME : (hasNames? OCCUR : END);
		break;

	    case OCCUR:
		if (!token.equals("*"))
		    throw new ParseException(
			msg("unexpectedToken", "*", token));
		state = END;
		break;

	    case PCDATA:
		if (!token.equals("#PCDATA"))
		    throw new ParseException(
			msg("unexpectedToken", "#PCDATA", token));
		state = CONNECTOR_OR_PARENTHESE;
		break;

	    case NAME:
		{
		    if (!XMLUtil.isName(token))
			throw new ParseException(
			    msg("unexpectedToken", msg("elementName"), token));

		    if (names.put(token, token) != null)
			throw new ParseException(msg("duplicateElementName", 
						     token));
		    hasNames = true;

		    state = CONNECTOR_OR_PARENTHESE;
		}
		break;

	    case END:
		throw new ParseException(
			msg("unexpectedToken", msg("endOfSpec"), token));

	    default:
		throw new RuntimeException("internal error: unknown state " + 
					   state);
	    }
	}

	if (state != END)
	    throw new ParseException(msg("unexpectedEndOfSpec"));

	if (!hasNames) {
	    return new SequenceParticle(NO_PARTICLES, 1, 1);
	} else {
	    Particle[] particles = new Particle[names.size()];
	    Enumeration iter = names.elements();
	    int i = 0;
	    while (iter.hasMoreElements())
		particles[i++] = 
		    new ElementParticle((String) iter.nextElement(), 1, 1);

	    return new ChoiceParticle(particles, 0, Particle.UNBOUNDED);
	}
    }

    public Particle parseElementContent(String spec) throws ParseException {
	StringTokenizer tokens = new StringTokenizer(spec, "(|,&)?*+{}", true);
	lookahead = tokens.nextToken();
	Particle parsed = parseElementContent(tokens);

	if (lookahead != null)
	    throw new ParseException(
		msg("unexpectedToken", msg("endOfSpec"), lookahead));

	return parsed;
    }

    private Particle parseElementContent(StringTokenizer tokens) 
	throws ParseException {
	String elementName = null;
	Particle[] particles = null;
	int particleType;

	if (lookahead.charAt(0) == '(') {
	    int state = TERM;
	    lookahead = nextToken(tokens);
	    char connector = '\0';
	    Vector list = new Vector();
	    char c;

	    loop: for (;;) {
		switch (state) {
		case TERM:
		    list.addElement(parseElementContent(tokens));
		    state = CONNECTOR_OR_PARENTHESE;
		    if (lookahead == null)
			throw new ParseException(
			    msg("unexpectedToken", "|,&)", msg("endOfSpec")));
		    break;

		case CONNECTOR_OR_PARENTHESE:
		    if (lookahead.length() != 1)
			throw new ParseException(
			    msg("unexpectedToken", "|,&)", lookahead));

		    c = lookahead.charAt(0);
		    switch (c) {
		    case '|': case ',': case '&':
			if (connector == '\0')
			    connector = c;
			else if (c != connector) 
			    throw new ParseException(
				msg("unexpectedToken", connector + ")", 
				    lookahead));
			state = TERM;
			lookahead = nextToken(tokens);
			break;
		    case ')':
			state = END;
			break loop;
		    default:
			throw new ParseException(
			    msg("unexpectedToken", "|,&)", lookahead));
		    }
		    break;

		default:
		    throw new RuntimeException(
			"internal error: unknown state " + state);
		}
	    }

	    if (state != END)
		throw new ParseException(msg("unexpectedEndOfSpec"));

	    switch (connector) {
	    case '|':
		particleType = Particle.CHOICE;
		break;
	    case '&':
		if (strictSyntax)
		    throw new ParseException(msg("strictSyntaxError", "&"));
		particleType = Particle.ALL;
		break;
	    case ',':
		particleType = Particle.SEQUENCE;
		break;
	    case '\0':
		{
		    // Parse (foo)?*+ as [ELEMENT,foo,n,m] rather than 
		    // [SEQUENCE,[ELEMENT,foo,1,1],n,m].

		    Particle p = (Particle) list.elementAt(0);

		    if (p.getMinOccurs() == 1 &&
			p.getMaxOccurs() == 1 && 
			(p instanceof ElementParticle)) {
			elementName = ((ElementParticle) p).getElementName();
			particleType = Particle.ELEMENT;
		    } else {
			particleType = Particle.SEQUENCE;
		    }
		}
		break;
	    default:
		throw new RuntimeException(
		    "internal error: unknown connector " + connector);
	    }

	    if (particleType != Particle.ELEMENT) {
		particles = new Particle[list.size()];
		list.copyInto(particles);
	    }
	} else {
	    if (!XMLUtil.isName(lookahead))
		throw new ParseException(msg("unexpectedToken", 
					     msg("elementName"), lookahead));
	    elementName = lookahead;
	    particleType = Particle.ELEMENT;
	}

	int minOccurs = 1;
	int maxOccurs = 1;

	lookahead = nextTokenIfAny(tokens);
	if (lookahead != null) {
	    switch (lookahead.charAt(0)) {
	    case '?':
		minOccurs = 0;
		maxOccurs = 1;
		lookahead = nextTokenIfAny(tokens);
		break;
	    case '*':
		minOccurs = 0;
		maxOccurs = Particle.UNBOUNDED;
		lookahead = nextTokenIfAny(tokens);
		break;
	    case '+':
		minOccurs = 1;
		maxOccurs = Particle.UNBOUNDED;
		lookahead = nextTokenIfAny(tokens);
		break;
	    case '{':
		if (strictSyntax)
		    throw new ParseException(msg("strictSyntaxError","{m,n}"));
		lookahead = nextToken(tokens);
		minOccurs = parseOccur(lookahead);

		lookahead = nextToken(tokens);
		switch (lookahead.charAt(0)) {
		case ',':
		    {
			lookahead = nextToken(tokens);
			if (lookahead.charAt(0) == '}') {
			    maxOccurs = Particle.UNBOUNDED;
			} else {
			    maxOccurs = parseOccur(lookahead);

			    lookahead = nextToken(tokens);
			    if (lookahead.charAt(0) != '}')
				throw new ParseException(
				    msg("unexpectedToken", "}", lookahead));
			}
		    }
		    break;
		case '}':
		    maxOccurs = minOccurs;
		    break;
		default:
		    throw new ParseException(
			msg("unexpectedToken", ",}", lookahead));
		}

		if (maxOccurs == 0 ||
		    (maxOccurs != Particle.UNBOUNDED && minOccurs > maxOccurs))
		    throw new ParseException(msg("invalidMinMaxOccurs"));

		lookahead = nextTokenIfAny(tokens);
		break;
	    }
	}

	if (particleType == Particle.ELEMENT) {
	    return new ElementParticle(elementName, minOccurs, maxOccurs);
	} else {
	    if (particles.length == 1 && minOccurs == 1 && maxOccurs == 1) {
		// Parse (xxx) as xxx rather than [SEQUENCE,xxx,1,1].
		return particles[0];
	    } else {
		switch (particleType) {
		case Particle.CHOICE:
		    return new ChoiceParticle(particles, minOccurs, maxOccurs);
		case Particle.SEQUENCE:
		    return new SequenceParticle(particles, 
						minOccurs, maxOccurs);
		case Particle.ALL:
		    return new AllParticle(particles, minOccurs, maxOccurs);
		default:
		    throw new RuntimeException(
			"internal error: unknown particle type " + 
			particleType);
		}
	    }
	}
    }

    private static String nextTokenIfAny(StringTokenizer tokens) {
	return tokens.hasMoreTokens()? tokens.nextToken() : null;
    }

    private static String nextToken(StringTokenizer tokens) 
	throws ParseException {
	if (!tokens.hasMoreTokens())
	    throw new ParseException(msg("unexpectedEndOfSpec"));
	return tokens.nextToken();
    }

    private static int parseOccur(String s) throws ParseException {
	int i;

	try {
	    i = Integer.parseInt(s);
	} catch (NumberFormatException e) {
	    i = -1;
	}

	if (i < 0)
	    throw new ParseException(msg("unexpectedToken", msg("occur"), s));

	return i;
    }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.namere.ParticleParserRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    private static String msg(String id, Object arg0, Object arg1) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0, arg1 });
    }

    // -----------------------------------------------------------------------
    // Test
    // -----------------------------------------------------------------------

    public static void main(String[] args) {
	if (args.length < 2 || args[0].length() != 1)
	    usage();

	boolean mixedContent = false;
	boolean strictSyntax = true;
	switch (args[0].charAt(0)) {
	case 'm':
	    mixedContent = true;
	    break;
	case 'e':
	    break;
	case 'E':
	    strictSyntax = false;
	    break;
	default:
	    usage();
	}
	
	ParticleParser parser = new ParticleParser(strictSyntax);

	for (int i = 1; i < args.length; ++i) {
	    System.out.print("'");
	    System.out.print(args[i]);
	    System.out.print("'");

	    try {
		if (mixedContent)
		    parser.parseMixedContent(args[i]);
		else
		    parser.parseElementContent(args[i]);
		System.out.println(" OK");
	    } catch (ParseException e) {
		System.out.println(" *** error: " + e.getMessage());
	    }
	}
    }

    private static void usage() {
	System.err.println(
	    "usage: java com.xmlmind.namere.ParticleParser" + 
	    " m|e|E spec ... spec");
	System.exit(1);
    }
}
