/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.structformat;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.awt.Frame;
import fr.pixware.util.WorkInProgressObserver;
import com.xmlmind.xmledit.model.DocumentModel;

/**
 * This interface should be implemented to provide support for a non-XML
 * structured format in XMLmind XML editor. Examples of such non-XML
 * structured formats are: HTML, Javadoc comments, <tt>/etc/password</tt>,
 * <tt>/etc/printcap</tt>, <tt>httpd.conf</tt> (Apache configuration file),
 * etc.
 * <p>A class that implements <code>StructuredFormat</code>, called a non-XML
 * format plug-in, must be able to convert the non-XML format to XML when the
 * document is being loaded and to convert the XML back to the non-XML format
 * each time the document is being saved.
 * <p>The <code>StructuredFormat</code> implementor generally begins his work
 * by writing a DTD for an XML markup that can faithfully represent the
 * non-XML structured format he needs to support.
 * <p>The <code>StructuredFormat</code> implementor will finish his work by
 * bundling his Java code, the DTD and optionnally a CSS stylesheet in a jar
 * file. The jar file must be put somewhere in the CLASSPATH in order to be
 * dynamically loaded by XXE if needed to.
 * <p>The last step is to register the plug-in with XXE. For that, the XXE
 * user will have to use the Non-XML Formats dialog box.
 */
public interface StructuredFormat {
    /**
     * Returns the MIME type of this structured format.
     * 
     * @return the MIME type of this structured format plug-in
     */
    String getMimeType();

    /**
     * Returns one or several file name extensions customarily used for this
     * structured format.
     * 
     * @return one or several file name extensions customarily used for this
     * structured format plug-in
     */
    String[] getFileNameExtensions();

    /**
     * Returns a description of this structured format plug-in.
     * 
     * @return a few line description of the functionalities and limitations
     * of this structured format plug-in; this description can also include
     * information about the author, the home page of the plug-in, etc.
     */
    String getDescription();

    /**
     * Returns a description of the parameters used by this structured format
     * plug-in.
     * 
     * @return a description of the parameters supported by this plug-in if
     * any, or <code>null</code> otherwise. This description is a list of
     * arrays of Strings with one array of Strings for each parameter. Each
     * array contains 4 Strings:
     * <ol>
     * <li>the parameter name,
     * <li>the parameter type,
     * <li>the parameter default value,
     * <li>one line description for the parameter.
     * </ol>
     * <p>Example:
     * <pre>    private static final String[][] parameterUsage = {
     *        {"maxLineLength", "positive integer", "78", 
     *         "The maximum length for a line generated by the plug-in"}
     *    };</pre>
     */
    String[][] getParameterUsage();

    /**
     * Configure this structured format plug-in.
     * 
     * @param parameters an even number of Strings: parameter name followed by
     * the parameter value
     * @exception ParameterUsageException if a parameter is unknown or if its
     * value is invalid
     */
    void setParameters(String[] parameters) throws ParameterUsageException;

    /**
     * Returns the current configuration of this structured format plug-in.
     * 
     * @return an even number of Strings: parameter name followed by the
     * parameter value or <code>null</code> if this plug-in cannot be
     * configured. A parameter which value is equal to the plug-in default
     * value need not be returned in this array.
     */
    String[] getParameters();

    /**
     * Register with this structured format plug-in an object used to monitor
     * the action of loading or saving a document.
     * 
     * @param observer An object implementing the interface {@link
     * com.xmlmind.xmledit.modeltool.WorkInProgressObserver
     * WorkInProgressObserver}
     */
    void setWorkInProgressObserver(WorkInProgressObserver observer);

    /**
     * Returns the object used to monitor the action of loading or saving a
     * document using this structured format plug-in.
     * 
     * @return the object used to monitor the action of loading or saving a
     * document using this structured format plug-in if any, or
     * <code>null</code> otherwise
     */
    WorkInProgressObserver getWorkInProgressObserver();

    /**
     * Loads (as XML) the content of an <code>InputStream</code> into a newly
     * created empty document.
     * 
     * @param in the <code>InputStream</code> being read from
     * @param url the location of the <code>InputStream</code>
     * @param javaEncoding the Java name (that is Java's "UTF8" not IANA's
     * "UTF-8") of the encoding of the content to be loaded or
     * <code>null</code>, which means that this encoding is unknown
     * @param document the blank XML document to be initialized
     * @exception IOException if there is an IO problem
     * @exception StructuredFormatException if the source contains syntax
     * errors
     */
    void loadDocument(InputStream in, URL url, String javaEncoding, 
		      DocumentModel document) 
	throws IOException, StructuredFormatException;

    /**
     * Saves (as the non-XML structured format implemented by this plug-in)
     * the content of a document to an <code>OutputStream</code>. This XML
     * document has, of course, previously been initialized using {@link
     * #loadDocument loadDocument}: i.e. it is not an arbitrary XML document.
     * 
     * @param document the XML document to be saved
     * @param out the <code>OutputStream</code> being written to
     * @param url the location of the <code>OutputStream</code>
     * @param javaEncoding the Java name (that is Java's "UTF8" not IANA's
     * "UTF-8") of the encoding to be used when saving the document to a text
     * file or <code>null</code>, which means to use the plug-in default
     * encoding (generally the platform default encoding)
     * @exception IOException if there is an IO problem
     * @exception StructuredFormatException If the XML document has invalid
     * content which prevents it from being saved.
     */
    void saveDocument(DocumentModel document, OutputStream out, URL url, 
		      String javaEncoding) 
	throws IOException, StructuredFormatException;
}
