/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.structformat;

/**
 * Thrown by {@link StructuredFormat#loadDocument loadDocument} when a syntax
 * error is found in the source document.
 * <p>May also be thrown by {@link StructuredFormat#saveDocument saveDocument}
 * if the content of the document being edited is invalid and therefore cannot
 * be properly saved.
 */
public class StructuredFormatException extends Exception {
    private String location;
    private int lineNumber;
    private int columnNumber;
    private String message;

    /**
     * Constructs an exception with all the mandatory info about the error.
     * 
     * @param location the name of the source document (typically a file: or
     * http: URL name)
     * @param lineNumber the line number where the error occurred or -1 if
     * unknown; first line is line #1
     * @param columnNumber the column number where the error occurred or -1 if
     * unknown; first column is column #1
     * @param message a message describing the error; this message does not
     * include the location
     */
    public StructuredFormatException(String location, int lineNumber, 
				     int columnNumber, String message) {
	this.location = location;
	this.lineNumber = lineNumber;
	this.columnNumber = columnNumber;
	this.message = message;
    }

    /**
     * Returns the name of the source document where the error occurred.
     * 
     * @return the name of the source document (typically a file: or http: URL
     * name)
     */
    public String getLocation() {
	return location;
    }

    /**
     * Returns the line number where the error occurred.
     * 
     * @return the line number where the error occurred or -1 if unknown;
     * first line is line #1
     */
    public int getLineNumber() {
	return lineNumber;
    }

    /**
     * Returns the column number where the error occurred.
     * 
     * @return the column number where the error occurred or -1 if unknown;
     * first column is column #1
     */
    public int getColumnNumber() {
	return columnNumber;
    }

    /**
     * Returns a message describing the error.
     * 
     * @return a message describing the error; this message does not include
     * the location
     */
    public String getMessage() {
	return message;
    }
}
