/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.swingutil;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.io.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;

public class FileChooserUtil {
    private static ResourceBundle resourceBundle = null;

    public static String getOpenFile(Frame frame) {
	return getOpenFile(frame, null);
    }

    public static String getOpenFile(Frame frame, String title) {
	return getOpenFile(frame, title, null);
    }

    public static String getOpenFile(Frame frame, String title, 
				     String templateFileName) {
	return getOpenFile(frame, title, templateFileName, /*filtered*/ true);
    }

    public static String getOpenFile(Frame frame, String title, 
				     String templateFileName,
				     boolean filtered) {
	return getFile(frame, title, templateFileName, filtered,
                       /*open*/ true, /*dir*/ false);
    }

    public static String getSaveFile(Frame frame) {
	return getSaveFile(frame, null);
    }

    public static String getSaveFile(Frame frame, String title) {
	return getSaveFile(frame, title, null);
    }

    public static String getSaveFile(Frame frame, String title, 
				     String templateFileName) {
	return getSaveFile(frame, title, templateFileName, /*filtered*/ true);
    }

    public static String getSaveFile(Frame frame, String title, 
				     String templateFileName, 
				     boolean filtered) {
	return getFile(frame, title, templateFileName, filtered, 
                       /*open*/ false, /*dir*/ false);
    }

    public static String getOpenDir(Frame frame) {
	return getOpenDir(frame, null);
    }

    public static String getOpenDir(Frame frame, String title) {
	return getOpenDir(frame, title, null);
    }

    public static String getOpenDir(Frame frame, String title, 
				    String templateFileName) {
	return getFile(frame, title, templateFileName, /*filtered*/ false, 
                       /*open*/ true, /*dir*/ true);
    }

    public static String getSaveDir(Frame frame) {
	return getSaveDir(frame, null);
    }

    public static String getSaveDir(Frame frame, String title) {
	return getSaveDir(frame, title, null);
    }

    public static String getSaveDir(Frame frame, String title, 
				    String templateFileName) {
	return getFile(frame, title, templateFileName, /*filtered*/ false,
                       /*open*/ false, /*dir*/ true);
    }

    private static String getFile(Frame frame, String title, 
				  String templateFileName, boolean filtered,
				  boolean open, boolean dir) {
	JFileChooser chooser = new JFileChooser();

	if (resourceBundle == null) {
	    try {
		resourceBundle = ResourceBundle.getBundle(
		    "com.xmlmind.swingutil.FileChooserUtilRes");
	    } catch (MissingResourceException e) {
		resourceBundle = null;
	    }
	}
	if (title == null) 
	    title = MenuUtil.localizedString((open? "open" : "save"),
					     resourceBundle);
	chooser.setDialogTitle(title);

	chooser.setDialogType(open? JFileChooser.OPEN_DIALOG : 
			      JFileChooser.SAVE_DIALOG);

	chooser.setFileSelectionMode(dir? JFileChooser.DIRECTORIES_ONLY :
				     JFileChooser.FILES_ONLY);

	//FIXME: JFileChooser not properly localized.
	Object[] info = MenuUtil.localizedInfo("ok", resourceBundle);
	Character mnemo = (Character) info[1];
	chooser.setApproveButtonText((String) info[0]);
	if (mnemo != null)
	    chooser.setApproveButtonMnemonic(mnemo.charValue());
	chooser.setApproveButtonToolTipText(null);

	if (templateFileName != null && templateFileName.length() > 0) {
	    File curFile = new File(templateFileName);
	    String dirName = curFile.getParent();
	    String baseName = curFile.getName();
	    File curDir;

	    if (dirName != null && 
		(curDir = new File(dirName)).isDirectory()) {
		if (open)
		    chooser.setCurrentDirectory(curDir);
		else
		    chooser.setSelectedFile(curFile);
	    }

	    if (!dir) {
		int dot = baseName.lastIndexOf('.');
		if (dot > 0) {
		    // Note that in file ~/.profile, .profile is not an
		    // extension.

		    String ext = baseName.substring(dot);
		    final String ext1 = ext.toLowerCase();
		    final String ext2 = ext.toUpperCase();

		    FileFilter filter = new FileFilter() {
			public boolean accept(File file) {
			    String name = file.getName();
			    return (name.endsWith(ext1)||name.endsWith(ext2));
			}

			public String getDescription() {
			    return ext2.substring(1) + " Files (*" + 
				ext1 + ')';
			}
		    };
		    chooser.addChoosableFileFilter(filter);

		    if (filtered)
			chooser.setFileFilter(filter);
		    else
			chooser.setFileFilter(
			    chooser.getAcceptAllFileFilter());
		}
	    }
	}

	int option = chooser.showDialog(frame, null);

	File selectedFile;
	if (option == JFileChooser.APPROVE_OPTION && 
	    (selectedFile = chooser.getSelectedFile()) != null) {
	    if (!open && selectedFile.exists()) {
		//FIXME: JOptionPane not properly localized.
		String dlgTitle = MenuUtil.localizedString("overwrite.title", 
							   resourceBundle);
		String format = 
		    MenuUtil.localizedString("overwrite.question", 
					     resourceBundle);

		String question = MessageFormat.format(format, new Object[] { 
		    selectedFile.getPath() });

		option = JOptionPane.showConfirmDialog(
		    frame, question, dlgTitle,
		    JOptionPane.OK_CANCEL_OPTION);

		if (option == JOptionPane.CANCEL_OPTION)
		    return null;
	    }
	} else {
	    return null;
	}

	return selectedFile.getPath();
    }

    // -----------------------------------------------------------------------

    public static String truncateFileName(String fileName) {
	return truncateFileName(fileName, 40);
    }

    public static String truncateFileName(String fileName, int maxLength) {
	if (fileName.length() <= maxLength) 
	    return fileName;

	String[] elements = 
	    FileChooserUtil.split(fileName, File.separatorChar);
	if (elements.length <= 2)
	    return fileName;

	// On Unix, elements[0] is the empty string when the path is absolute.
	int j = 0;
	StringBuffer head = new StringBuffer(maxLength);
        head.append(elements[j]);
	++j;

	int k = elements.length - 1;
	StringBuffer tail = new StringBuffer(maxLength);
        tail.append(elements[k]);
	--k;

	int length = head.length() + 5 + tail.length();
	boolean addTail = true;
	while (j < k && length < maxLength) {
	    if (addTail) {
		tail.insert(0, File.separatorChar);
		tail.insert(0, elements[k]);
		--k;
		addTail = false;
	    } else {
		head.append(File.separatorChar);
		head.append(elements[j]);
		++j;
		addTail = true;
	    }
	    length = head.length() + 5 + tail.length();
	}

	head.append(File.separatorChar);
	head.append("...");
	head.append(File.separatorChar);
	head.append(tail.toString());

	return head.toString();
    }

    public static String[] split(String string, char separatorChar) {
	// Count elements ---

	int elementCount = 0;
	int sep = 0;
	while ((sep = string.indexOf(separatorChar, sep)) >= 0) {
	    ++elementCount;
	    ++sep;
	} 
	++elementCount;

	// Build element array ---

	String[] elements = new String[elementCount];

	elementCount = 0;
	sep = 0;
	int nextSep;
	while ((nextSep = string.indexOf(separatorChar, sep)) >= 0) {
	     elements[elementCount++] = 
		 (sep == nextSep)? "" : string.substring(sep, nextSep);
	     sep = nextSep + 1;
	} 
	elements[elementCount++] = string.substring(sep);

	return elements;
    }
}


