/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.swingutil;

import java.util.Hashtable;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.xmlmind.awtutil.AWTUtil;

public class MenuUtil {
    public static class MenuItemSpec {
	public Action action;
	public int accelKeyCode;
	public int accelModifiers;

	public MenuItemSpec() {
	    this(null, 0, 0);
	}
	public MenuItemSpec(Action action) {
	    this(action, 0, 0);
	}
	public MenuItemSpec(Action action, int accelKeyCode) {
	    this(action, accelKeyCode, AWTUtil.KEY_MODIFIER);
	}
	public MenuItemSpec(Action action, int accelKeyCode, 
			    int accelModifiers) {
	    this.action = action;
	    this.accelKeyCode = accelKeyCode;
	    this.accelModifiers = accelModifiers;
	}
    }

    public static class MenuSpec extends MenuItemSpec {
	public String name;
	public MenuItemSpec[] items; 

	public MenuSpec(String name, MenuItemSpec[] items) {
	    super();
	    this.name = name;
	    this.items = items;
	}
    }
    
    public static Hashtable addMenus(JMenuBar bar, MenuSpec[] specs, 
				     ResourceBundle localizedNames) {
	Hashtable map = new Hashtable();

	for (int i = 0; i < specs.length; ++i) {
	    MenuSpec spec = specs[i];
	    Object[] info = localizedInfo(spec.name, localizedNames);

	    JMenu menu = new JMenu((String) info[0]);
	    if (spec.name.equals("help")) 
		//NYI!!! bar.setHelpMenu(menu);
		bar.add(menu);
	    else
		bar.add(menu);

	    map.put("menu:" + spec.name, menu);

	    if (info[1] != null)
		menu.setMnemonic(((Character) info[1]).charValue());

	    addMenuItems(menu, map, spec.name, spec.items, localizedNames);
	}

	return map;
    }

    public static void addMenuItems(JMenu menu, Hashtable map, String prefix, 
				    MenuItemSpec[] itemSpecs, 
				    ResourceBundle localizedNames) {
	for (int i = 0; i < itemSpecs.length; ++i) {
	    MenuItemSpec itemSpec = itemSpecs[i];

	    if (itemSpec instanceof MenuSpec) {
		MenuSpec spec = (MenuSpec) itemSpec;
		String fullName = prefix + "." + spec.name;
		Object[] info = localizedInfo(fullName, localizedNames);

		JMenu subMenu = new JMenu((String) info[0]);
		menu.add(subMenu);

		map.put("menu:" + fullName, subMenu);

		if (info[1] != null)
		    subMenu.setMnemonic(((Character) info[1]).charValue());

		addMenuItems(subMenu, map, fullName, spec.items, 
			     localizedNames);
	    } else {
		if (itemSpec.action == null) {
		    menu.addSeparator();
		} else {
		    Action action = itemSpec.action;
		    String fullName = 
			prefix + "." + (String) action.getValue(Action.NAME);
		    Object[] info = localizedInfo(fullName, localizedNames);

		    JMenuItem item = menu.add(action);

		    map.put("action:" + fullName, action);
		    map.put("menuItem:" + fullName, item);

		    item.setText((String) info[0]);
		    if (info[1] != null)
			item.setMnemonic(((Character) info[1]).charValue());

		    if (itemSpec.accelKeyCode != 0) 
			item.setAccelerator(
			    KeyStroke.getKeyStroke(itemSpec.accelKeyCode, 
						   itemSpec.accelModifiers));
		}
	    }
	}
    }

    public static JMenu getMenu(Hashtable map, String name) {
	return (JMenu) map.get("menu:" + name);
    }

    public static JMenuItem getMenuItem(Hashtable map, String name) {
	return (JMenuItem) map.get("menuItem:" + name);
    }

    public static Action getAction(Hashtable map, String name) {
	return (Action) map.get("action:" + name);
    }

    // -----------------------------------------------------------------------

    public static String getAcceleratorText(JMenuItem item) {
	KeyStroke keyStroke = item.getAccelerator();
	if (keyStroke == null)
	    return null;

	int modifiers = keyStroke.getModifiers();
	int keyCode = keyStroke.getKeyCode();

	StringBuffer buffer = new StringBuffer();

	buffer.append(KeyEvent.getKeyModifiersText(modifiers));
	buffer.append('-');
	buffer.append(KeyEvent.getKeyText(keyCode));

	return buffer.toString();
    }

    // -----------------------------------------------------------------------

    public static String localizedString(String id, 
					 ResourceBundle localizedNames) {
	Object[] info = localizedInfo(id, localizedNames);
	return (String) info[0];
    }

    public static Object[] localizedInfo(String id, 
					 ResourceBundle localizedNames) {
	String text = id;
	try {
	    if (localizedNames != null)
		text = localizedNames.getString(id);
	} catch (MissingResourceException e) {
	    text = id;
	}

	Character mnemo = null;
	int i = text.indexOf('_');
	if (i >= 0 && i+1 < text.length()) {
	    mnemo = new Character(text.charAt(i+1));
	    text = text.substring(0, i) + text.substring(i+1);
	}
	
	return new Object[] { text, mnemo };
    }
}


