/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit;

import java.awt.Graphics;
import java.awt.Rectangle;
import com.xmlmind.xmledit.model.DocumentModel;
import com.xmlmind.xmledit.model.TextModel;
import com.xmlmind.xmledit.modeltool.TextLocations;
import com.xmlmind.xmledit.modeltool.TextLocation;
import com.xmlmind.xmledit.modeltool.TextLocationListener;
import com.xmlmind.xmledit.modeltool.TextLocationEvent;

public abstract class Caret implements TextLocationListener {
    public static final String POINT = "point";
    public static final String MARK  = "mark";

    protected DocumentView docView;
    protected TextLocations locations;
    protected TextLocation point;
    protected TextLocation mark;

    protected Caret(DocumentView documentView) {
	docView = documentView;

	locations = new TextLocations();
	locations.addTextLocationListener(this);
    }

    public DocumentView getDocumentView() {
	return docView;
    }

    public void setDocument(DocumentModel document) {
	mark = null;
	point = null;

	locations.setDocument(document);

	if (document != null) {
	    point = locations.add(POINT);
	    point.setCursor(true);
	}
    }

    public DocumentModel getDocument() {
	return locations.getDocument();
    }

    public TextLocations getLocationManager() {
	return locations;
    }

    public TextLocation mark() {
	if (mark == null) {
	    TextModel text = point.getText();
	    if (text != null) {
		mark = locations.add(MARK);
		mark.moveTo(text, point.getOffset());
	    }
	}

	return mark;
    }

    public void unmark() {
	if (mark != null) {
	    locations.remove(MARK);
	    // mark set to null by callback.
	}
    }

    public TextLocation getMark() {
	return mark;
    }

    public TextLocation getPoint() {
	return point;
    }

    public static boolean isPoint(TextLocation location) {
	return (location.getId() == POINT);
    }

    public static boolean isMark(TextLocation location) {
	return (location.getId() == MARK);
    }

    public abstract Rectangle getBounds();
    public abstract Rectangle refresh();
    public abstract void paintSelection(Graphics g);
    public abstract void paintCaret(Graphics g);

    public abstract void setBlinking(boolean blink);
    public abstract boolean isBlinking();

    public void textLocationAdded(TextLocationEvent event) {
	// Adding a TextLocation cannot cause a context change because
	// the TextLocation is no yet bound to a text+offset.
    }

    public void textLocationMoved(TextLocationEvent event) {
	TextLocation source = event.getSource();

	if (source == point) {
	    docView.clearMagicPosition();

	    docView.refreshCaret();

	    if (event.getText() != event.getPreviousText()) {
		// Point moved to another text.
		docView.notifyContextChangeListeners();
	    } else {
		if (mark != null) {
		    // Do we have a text selection now?
		    TextModel markText = mark.getText();
		    int markOffset = mark.getOffset();

		    boolean hadSelection = 
			(markText != event.getPreviousText() ||
			 markOffset != event.getPreviousOffset());
		    boolean hasSelection = (markText != event.getText() ||
					    markOffset != event.getOffset());
		    if (hasSelection != hadSelection)
			docView.notifyContextChangeListeners();
		}
	    }
	} else if (source == mark) {
	    docView.refreshCaret();

	    // Do we have a text selection now?
	    TextModel pointText = point.getText();
	    int pointOffset = point.getOffset();

	    boolean hadSelection = 
		(event.getPreviousText() != null &&
		 (event.getPreviousText() != pointText ||
		  event.getPreviousOffset() != pointOffset));
	    boolean hasSelection = (event.getText() != pointText ||
				    event.getOffset() != pointOffset);
	    if (hasSelection != hadSelection)
		docView.notifyContextChangeListeners();
	}
    }

    public void textLocationRemoved(TextLocationEvent event) {
	TextLocation source = event.getSource();

	if (source == point) {
	    docView.refreshCaret();

	    docView.notifyContextChangeListeners();
	} else if (source == mark) {
	    mark = null;

	    docView.refreshCaret();

	    docView.notifyContextChangeListeners();
	}
    }
}

