/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit;

import java.awt.*;
import java.awt.event.*;
import com.xmlmind.awtutil.Bitmap;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.gadget.GadgetContainer;
import com.xmlmind.xmledit.model.*;

public class CharView extends Gadget implements TextView, TextViewPart {
    public static class Settings extends CommonSettings {
	public Image wrappedIcon;

	public static final Bitmap wrappedBitmap = 
	    new Bitmap(9, 7, new byte[] {
		(byte)0xfe, (byte)0x00, (byte)0xee, (byte)0x00, (byte)0xee, 
		(byte)0x00, (byte)0x82, (byte)0x00, (byte)0xee, (byte)0x00, 
		(byte)0xee, (byte)0x00, (byte)0xfe, (byte)0x00
            });

	public Settings() {
	    background = Color.white;
	    foreground = Color.black;
	    font = new Font("Monospaced", Font.PLAIN, 10);
	    pad = 4;
	    borderColor = Color.gray;
	    borderRelief = RELIEF_SUNKEN;

	    wrappedIcon = wrappedBitmap.getImage(foreground);
	}
    }

    // -----------------------------------------------------------------------

    public static final Settings DEFAULT_SETTINGS = new Settings();

    private TextModel text;
    private Settings settings;
    private int[] lineOffsets = new int[0];

    // -----------------------------------------------------------------------

    public CharView(TextModel text) {
	this(text, DEFAULT_SETTINGS);
    }

    public CharView(TextModel text, Settings settings) {
	this.text = text;
	this.settings = settings;
    }

    public SubelementModel getModel() {
	return text;
    }

    public void getGadgets(GadgetList gadgets, int viewIndex) {
	gadgets.appendGadget(this);
    }

    public TextModel getText() {
	return text;
    }

    public SubelementView getView() {
	return this;
    }

    public TextView getTextView() {
	return this;
    }

    public Settings getSettings() {
	return settings;
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	relayoutLines(constrainedWidth, 0);
    }

    private int relayoutLines(int constrainedWidth, int from) {
	// Find the line where relayout begins ---

	int[] newOffsets = new int[10];
	int newOffsetCount = 0;
	
	if (lineOffsets.length <= 1) {
	    from = 0;
	} else {
	    int i = 1;
	    for (;;) {
		if (i == lineOffsets.length || lineOffsets[i] > from) {
		    from = lineOffsets[i-1];
		    break;
		}

		newOffsets = addLineOffset(newOffsets, newOffsetCount, 
					   lineOffsets[i-1]);
		++newOffsetCount;

		++i;
	    }
	}

	int fromLine = newOffsetCount;

	// Split text into lines ---
	
	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();

	FontMetrics metrics = getFontMetrics(settings.font);
	int charWidth = metrics.charWidth('x');

	int iconWidth = settings.wrappedIcon.getWidth(null);
	int maxWidth = ((constrainedWidth > 0)? 
			(constrainedWidth - 2*settings.pad - iconWidth) : 
			Integer.MAX_VALUE); 
	if (maxWidth <= 0)
	    maxWidth = 1;

	int lineBegin = from;
	int lineWidth = 0;
	int widest = 0;

	for (int i = from; i < textLength; ++i) {
	    char c = textChars[i];

	    if (c == '\n') {
		newOffsets = addLineOffset(newOffsets, newOffsetCount, 
					   lineBegin);
		++newOffsetCount;

		if (lineWidth > widest)
		    widest = lineWidth;

		lineBegin = i+1;
		lineWidth = 0;
	    } else { 
		if (lineWidth + charWidth > maxWidth && lineWidth > 0) {
		    newOffsets = addLineOffset(newOffsets, newOffsetCount, 
					       lineBegin);
		    ++newOffsetCount;

		    if (lineWidth > widest)
			widest = lineWidth;

		    lineBegin = i;
		    lineWidth = charWidth;
		} else {
		    lineWidth += charWidth;
		}
	    }
	}

	// Even when last line ends with a '\n', there is an empty line after
	// it.
	newOffsets = addLineOffset(newOffsets, newOffsetCount, lineBegin);
	++newOffsetCount;
	if (lineWidth > widest)
	    widest = lineWidth;

	// Save line layout ---

	lineOffsets = new int[newOffsetCount];
	System.arraycopy(newOffsets, 0, lineOffsets, 0, newOffsetCount);

	// Recompute size ---

	width = 2*settings.pad + iconWidth + widest;
	height = 2*settings.pad + (metrics.getAscent()+metrics.getDescent()) +
	    ((newOffsetCount-1)*metrics.getHeight());

	// This gives refresh() more latitude to just repaint() instead of
	// doing paintRelayout().
	if (constrainedWidth > 0 && width < constrainedWidth)
	    width = constrainedWidth;

	return fromLine;
    }

    private static final int[] addLineOffset(int[] newOffsets, 
					     int newOffsetCount,
					     int lineBegin) {
	if (newOffsetCount + 1 > newOffsets.length) {
	    int[] nextOffsets = new int[Math.max(newOffsetCount + 1, 
						 2*newOffsets.length)];
	    System.arraycopy(newOffsets, 0, nextOffsets, 0, newOffsetCount);
	    newOffsets = nextOffsets;
	}

	newOffsets[newOffsetCount] = lineBegin;
	return newOffsets;
    }

    public void paint(Graphics g) {
	if (settings.background != null) {
	    g.setColor(settings.background);
	    g.fillRect(0, 0, width, height);
	}

	if (settings.borderColor != null && settings.pad > 1) {
	    g.setColor(settings.borderColor);

	    switch (settings.borderRelief) {
	    case Settings.RELIEF_RAISED:
		g.draw3DRect(0, 0, width-1, height-1, true);
		break;
	    case Settings.RELIEF_SUNKEN:
		g.draw3DRect(0, 0, width-1, height-1, false);
		break;
	    default:
		g.drawRect(0, 0, width-1, height-1);
	    }
	}
    }

    public void paintText(Graphics g) {
	g.setColor(settings.foreground);
	g.setFont(settings.font);

	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();
	int lineCount = lineOffsets.length;
	FontMetrics metrics = g.getFontMetrics(settings.font);
	int lineHeight = metrics.getHeight();
	int pad = settings.pad;

	Rectangle clip = g.getClipBounds();

	// Translate clip.y to origin at leading before first line.
	int clipY = clip.y - pad + metrics.getLeading();
	int firstLine = clipY / lineHeight;
	if (firstLine < 0)
	    firstLine = 0;
	else if (firstLine >= lineCount)
	    firstLine = lineCount-1;

	clipY += clip.height-1;
	int lastLine = clipY / lineHeight;
	if (lastLine < 0)
	    lastLine = 0;
	else if (lastLine >= lineCount)
	    lastLine = lineCount-1;

	Image icon = settings.wrappedIcon;
	int iconHeight = icon.getHeight(null);
	int iconX = width - pad - icon.getWidth(null);
	int lineY = pad + metrics.getAscent() + firstLine*lineHeight;

	for (int i = firstLine; i <= lastLine; ++i) {
	    int firstChar = lineOffsets[i];
	    if (firstChar >= textLength)
		// Empty last line.
		break;

	    boolean isLastLine = (i == lineCount-1);
	    int endChar;
	    if (isLastLine)
		endChar = textLength;
	    else
		endChar = lineOffsets[i+1];

	    int lineCharCount;
	    boolean drawIcon;
	    if (textChars[endChar-1] == '\n') {
		// Do not draw '\n'.
		lineCharCount = (endChar-firstChar) - 1;
		drawIcon = false;
	    } else {
		lineCharCount = (endChar-firstChar);
		drawIcon = !isLastLine;
	    }

	    if (lineCharCount > 0)
		g.drawChars(textChars, firstChar, lineCharCount, pad, lineY);

	    if (drawIcon)
		g.drawImage(icon, iconX, lineY - iconHeight, null);

	    lineY += lineHeight;
	}
    }

    public void print(Graphics g) {
	if (settings.borderColor != null && settings.pad > 1) {
	    g.setColor(settings.borderColor);
	    g.drawRect(0, 0, width-1, height-1);
	}
    }

    public int gadgetToModel(int viewX, int viewY) {
	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();
	int lineCount = lineOffsets.length;
	FontMetrics metrics = getFontMetrics(settings.font);
	int charWidth = metrics.charWidth('x');
	int lineHeight = metrics.getHeight();
	int pad = settings.pad;

	// Translate viewY to origin at leading before first line.
	viewY = viewY - pad + metrics.getLeading();
	int lineNum = viewY / lineHeight;
	if (lineNum < 0)
	    lineNum = 0;
	else if (lineNum >= lineCount)
	    lineNum = lineCount-1;

	viewX = viewX - pad;

	int firstChar = lineOffsets[lineNum];
	if (firstChar >= textLength)
	    // Empty last line.
	    return textLength;

	if (viewX <= 0)
	    return firstChar;

	int endChar;
	if (lineNum == lineCount-1)
	    endChar = textLength;
	else
	    endChar = lineOffsets[lineNum+1];

	if (textChars[endChar-1] == '\n')
	    // '\n' is not drawn.
	    --endChar;

	if (viewX >= charWidth*(endChar-firstChar))
	    // Before (not drawn) '\n' or before first char of next line!
	    return endChar;

	return (firstChar + (int) Math.rint((double)viewX/(double)charWidth));
    }

    public int modelToGadget(int offset, int count, RectangleList rectangles) {
	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();
	int lineCount = lineOffsets.length;
	FontMetrics metrics = getFontMetrics(settings.font);
	int charWidth = metrics.charWidth('x');
	int lineHeight = metrics.getHeight();
	int fontHeight = metrics.getAscent()+metrics.getDescent();
	int pad = settings.pad;

	if (offset < 0)
	    offset = 0;
	else if (offset > textLength)
	    offset = textLength;

	if (count < 0)
	    count = 0;
	else if (offset + count > textLength)
	    count = textLength - offset;

	int i;
	for (i = 0; i < lineCount; ++i) {
	    int firstChar = lineOffsets[i];
	    if (firstChar >= textLength) {
		// Empty last line.
		rectangles.appendRectangle(pad, pad+i*lineHeight, 
					   0, fontHeight);
		return 1;
	    }

	    int endChar;
	    if (i == lineCount-1)
		endChar = textLength;
	    else
		endChar = lineOffsets[i+1];

	    if (textChars[endChar-1] == '\n')
		// '\n' is not drawn.
		--endChar;

	    if (offset <= endChar) {
		if (count == 0) {
		    int x1 = pad + charWidth*(offset-firstChar);
		    rectangles.appendRectangle(x1, pad + i*lineHeight, 
					       0, fontHeight);
		    return 1;
		} else {
		    break;
		}
	    }
	}

	int appended = 0;
	boolean addEmptyLastLine = (lineOffsets[lineCount-1] >= textLength &&
				    offset + count >= textLength);

	for (; i < lineCount; ++i) {
	    int firstChar = lineOffsets[i];

	    int nextFirstChar, endChar;
	    if (i == lineCount-1) 
		endChar = nextFirstChar = textLength;
	    else
		endChar = nextFirstChar = lineOffsets[i+1];

	    if (textChars[endChar-1] == '\n')
		// '\n' is not drawn.
		--endChar;

	    int x1 = ((offset > firstChar)? 
		      pad + charWidth*(offset-firstChar) : pad);
	    int x2 = x1 + charWidth*((offset + count <= endChar)? 
				     count : endChar-offset);

	    rectangles.appendRectangle(x1, pad + i*lineHeight, 
				       x2-x1, fontHeight);
	    ++appended;

	    int prevOffset = offset;
	    offset = nextFirstChar;
	    count -= (nextFirstChar - prevOffset);
	    if (count <= 0) 
		break;
	}

	if (addEmptyLastLine) {
	    rectangles.appendRectangle(pad, pad + (lineCount-1)*lineHeight, 
				       0, fontHeight);
	    ++appended;
	}

	return appended;
    }

    public boolean acceptsControlChar(char c) {
	switch (c) {
	case '\t':
	case '\n':
	    return true;
	default:
	    return false;
	}
    }

    public void charInserted(char c, int offset) {
	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();
	int lineNum = lineNumberOfChar(lineOffsets, offset);
	int firstChar = lineOffsets[lineNum];

	if (c == '\n') {
	    if (offset == firstChar && 
		offset > 0 && textChars[offset-1] != '\n')
		// Erase wrapped icon.
		refreshFrom(offset-1);
	    else
		// Add a line after this one.
		refreshFrom(offset);
	    return;
	}

	int lineCount = lineOffsets.length;
	FontMetrics metrics = getFontMetrics(settings.font);
	int charWidth = metrics.charWidth('x');

	int endChar;
	if (lineNum == lineCount-1) 
	    endChar = textLength;
	else
	    endChar = lineOffsets[lineNum+1] + /*ins. char*/ 1;

	if (textChars[endChar-1] == '\n')
	    // '\n' is not drawn.
	    --endChar;

	if (charWidth*(endChar-firstChar) <= 
	    width - 2*settings.pad - settings.wrappedIcon.getWidth(null)) {

	    // Update char offsets for lines following this one.
	    for (int i = lineNum+1; i < lineCount; ++i)
		++lineOffsets[i];

	    // Just repaint this line.
	    int fromY = settings.pad + (lineNum*metrics.getHeight());
	    int fontHeight = metrics.getAscent()+metrics.getDescent();
	    repaint(0, fromY, width, fontHeight);

	    return;
	} 

	refreshFrom(offset);
    }

    private static final int lineNumberOfChar(int[] lineOffsets, int offset) {
	int i = 1;
	for (;;) {
	    if (i == lineOffsets.length || lineOffsets[i] > offset)
		return i-1;

	    ++i;
	}
    }

    public void charDeleted(char c, int offset) {
	int textLength = text.getTextLength();
	if (textLength == 0) {
	    refreshFrom(0);
	    return;
	}

	char[] textChars = text.getTextChars();
	int lineNum = lineNumberOfChar(lineOffsets, offset);
	int firstChar = lineOffsets[lineNum];
	if (offset == firstChar && 
	    offset > 0 && textChars[offset-1] != '\n') {
	    // Possibly remove a line corresponding to wrapped segment.
	    refreshFrom(offset-1);
	    return;
	}

	int lineCount = lineOffsets.length;
	int endChar;

	if (lineNum == lineCount-1) 
	    endChar = textLength;
	else
	    endChar = lineOffsets[lineNum+1] - /*del. char*/ 1;
	if (firstChar == endChar) {
	    // Remove empty line.
	    refreshFrom(offset);
	    return;
	}

	if (textChars[endChar-1] == '\n') {
	    // Update char offsets for lines following this one.
	    for (int i = lineNum+1; i < lineCount; ++i)
		--lineOffsets[i];

	    // Just repaint this line.
	    FontMetrics metrics = getFontMetrics(settings.font);
	    int fromY = settings.pad + (lineNum*metrics.getHeight());
	    int fontHeight = metrics.getAscent()+metrics.getDescent();
	    repaint(0, fromY, width, fontHeight);

	    return;
	}
	
	refreshFrom(offset);
    }

    private void refreshFrom(int from) {
	if (parent == null)
	    return;

	int prevWidth = width;
	int prevHeight = height;

	Dimension constraints = new Dimension();
	parent.getSizeConstraints(this, constraints);
	int fromLine = relayoutLines(constraints.width, from);

	if (width == prevWidth && height == prevHeight) {
	    int lineHeight = getFontMetrics(settings.font).getHeight();
	    int fromY = settings.pad + (fromLine*lineHeight);

	    repaint(0, fromY, width, height-fromY);
	} else {
	    int damagedSide, damagedLength;

	    if (width == prevWidth) {
		int lineHeight = getFontMetrics(settings.font).getHeight();
		int fromY = settings.pad + (fromLine*lineHeight);

		damagedSide = GadgetContainer.BOTTOM_SIDE;
		damagedLength = height-fromY;
	    } else {
		damagedSide = GadgetContainer.NO_SIDE;
		damagedLength = -1;
	    }

	    parent.paintRelayout(this, prevWidth, prevHeight, 
				 damagedSide, damagedLength);
	}
    }
}
