/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.Timer;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.xmledit.model.DocumentModel;
import com.xmlmind.xmledit.model.TextModel;

public class DefaultCaret extends Caret implements ActionListener {
    private Color selectionColor = Color.gray;
    private Color caretColor = Color.black;
    private Color caretColor2 = Color.red;
    private int blinkDelay = 500;

    private RectangleListImpl selectionRectList = new RectangleListImpl();
    private Rectangle caretRect = new Rectangle();
    private Rectangle nextRect = new Rectangle();
    private Rectangle currentRect = new Rectangle();

    private Timer blinkTimer = null;
    private Rectangle blinkRect = new Rectangle();
    private Color blinkColor = caretColor2;

    public DefaultCaret(DocumentView documentView) {
	super(documentView);
    }

    public void setDocument(DocumentModel document) {
	blinkRect.height = -1;
	if (blinkTimer != null) {
	    blinkTimer.stop();
	    blinkTimer = null;
	}

	super.setDocument(document);

	if (blinkDelay > 0 && document != null) {
	    blinkTimer = new Timer(blinkDelay, this);
	    blinkTimer.setCoalesce(true);

	    if (docView.isEditable() && docView.hasFocus())
		blinkTimer.start();
	}
    }

    public void actionPerformed(ActionEvent e) {
	if (blinkRect.height > 0) {
	    Graphics g = docView.getGraphics();

	    try {
		if (blinkColor == caretColor2)
		    blinkColor = caretColor;
		else
		    blinkColor = caretColor2;
		g.setColor(blinkColor);

		g.drawLine(blinkRect.x, blinkRect.y, 
			   blinkRect.x, blinkRect.y + blinkRect.height - 1);
	    } finally {
		g.dispose();
	    }
	}
    }

    public void setSelectionColor(Color color) {
	selectionColor = color;
	repaint();
    }

    private void repaint() {
	if (currentRect.width > 0)
	    docView.repaint(currentRect.x, currentRect.y, 
			    currentRect.width, currentRect.height);
    }

    public Color getSelectionColor() {
	return selectionColor;
    }

    public void setCaretColor(Color color) {
	caretColor = color;
	repaint();
    }

    public Color getCaretColor() {
	return caretColor;
    }

    public void setCaretColor2(Color color) {
	caretColor2 = color;
	repaint();
    }

    public Color getCaretColor2() {
	return caretColor2;
    }

    public void setBlinkDelay(int delay) {
	boolean hadTimer;
	boolean wasRunning;

	blinkRect.height = -1;
	if (blinkTimer == null) {
	    hadTimer = false;
	    wasRunning = false;
	} else {
	    hadTimer = true;
	    wasRunning = blinkTimer.isRunning();
	    if (wasRunning)
		blinkTimer.stop();
	    blinkTimer = null;
	}

	blinkDelay = delay;

	if (blinkDelay > 0 && hadTimer) {
	    blinkTimer = new Timer(blinkDelay, this);
	    blinkTimer.setCoalesce(true);
	    if (wasRunning)
		blinkTimer.start();
	}
    }

    public int getBlinkDelay() {
	return blinkDelay;
    }

    public void setBlinking(boolean blink) {
	if (blinkTimer != null) {
	    boolean isRunning = blinkTimer.isRunning();

	    if (blink && !isRunning) {
		blinkTimer.start();
		// Redraw in the right color.
		repaint();
	    } else if (!blink && isRunning) {
		blinkTimer.stop();
		repaint();
	    }
	}
    }

    public boolean isBlinking() {
	return (blinkTimer != null && blinkTimer.isRunning());
    }

    public Rectangle getBounds() {
	return currentRect;
    }

    public Rectangle refresh() {
	blinkRect.height = -1;

	if (currentRect.width > 0) {
	    docView.repaint(currentRect.x, currentRect.y, 
			    currentRect.width, currentRect.height);
	    currentRect.x = currentRect.y = 
		currentRect.width = currentRect.height = 0;
	}

	relayout();
	if (nextRect.width > 0) 
	    docView.repaint(nextRect.x, nextRect.y,
			    nextRect.width, nextRect.height);
	return nextRect;
    }

    public void paintSelection(Graphics g) {
	if (nextRect.width <= 0)
	    relayout();

	int rectangleCount = selectionRectList.size;
	if (rectangleCount > 0) {
	    Rectangle clip = g.getClipBounds();
	    
	    g.setColor(selectionColor);

	    Rectangle[] rectangles = selectionRectList.list;
	    for (int i = 0; i < rectangleCount; ++i) {
		Rectangle rect = rectangles[i];

		if (clip == null || 
		    // Rect intersects clip.
		    (rect.x < clip.x+clip.width && 
		     rect.y < clip.y+clip.height &&
		     rect.x+rect.width > clip.x && 
		     rect.y+rect.height > clip.y)) 
		    g.fillRect(rect.x, rect.y, rect.width, rect.height);
	    }
	}
    }

    public void paintCaret(Graphics g) {
	if (caretRect.width > 0) {
	    g.setColor(caretColor);

	    g.drawLine(caretRect.x, caretRect.y, 
		       caretRect.x, caretRect.y + caretRect.height - 1);

	    blinkRect.x = caretRect.x;
	    blinkRect.y = caretRect.y;
	    blinkRect.width = caretRect.width;
	    blinkRect.height = caretRect.height;
	} else {
	    blinkRect.height = -1;
	}

	currentRect.x = nextRect.x;
	currentRect.y = nextRect.y;
	currentRect.width = nextRect.width;
	currentRect.height = nextRect.height;

	nextRect.x = nextRect.y = nextRect.width = nextRect.height = 0;

	selectionRectList.removeAll();

	caretRect.x = caretRect.y = caretRect.width = caretRect.height = 0;
    }

    private void relayout() {
	selectionRectList.removeAll();
	caretRect.x = caretRect.y = caretRect.width = caretRect.height = 0;
	nextRect.x = nextRect.y = nextRect.width = nextRect.height = 0;

	TextModel pointText = point.getText();
	int pointOffset = point.getOffset();

	TextModel markText;
	int markOffset;
	if (mark == null) {
	    markText = pointText;
	    markOffset = pointOffset;
	} else {
	    markText = mark.getText();
	    markOffset = mark.getOffset();
	}

	if (markText != pointText || markOffset != pointOffset) {
	    docView.modelToView(markText, markOffset, pointText, pointOffset, 
				selectionRectList);

	    int rectangleCount = selectionRectList.size;
	    Rectangle[] rectangles = selectionRectList.list;
	    for (int i = 0; i < rectangleCount; ++i) {
		Rectangle rectangle = rectangles[i];

		if (rectangle.height > 0) {
		    if (rectangle.width <= 0)
			rectangle.width = 1;

		    union(nextRect, rectangle);
		}
	    }

	    docView.modelToView(pointText, pointOffset, caretRect);
	    if (caretRect.height > 0) {
		caretRect.width = 1;
		union(nextRect, caretRect);
		caretRect.x = caretRect.y = 
		    caretRect.width = caretRect.height = 0;
	    }
	} else if (pointText != null) {
	    docView.modelToView(pointText, pointOffset, caretRect);

	    if (caretRect.height > 0) {
		caretRect.width = 1;

		nextRect.x = caretRect.x;
		nextRect.y = caretRect.y;
		nextRect.width = caretRect.width;
		nextRect.height = caretRect.height;
	    }
	}
    }

    private static final void union(Rectangle u, Rectangle r) {
	if (r.width <= 0 || r.height <= 0)
	    return;

	if (u.width <= 0 || u.height <= 0) {
	    u.x = r.x;
	    u.y = r.y;
	    u.width = r.width;
	    u.height = r.height;
	    return;
	}

	int x1 = Math.min(u.x, r.x);
	int x2 = Math.max(u.x + u.width, r.x + r.width);
	int y1 = Math.min(u.y, r.y);
	int y2 = Math.max(u.y + u.height, r.y + r.height);

	u.x = x1;
	u.y = y1;
	u.width = x2 - x1;
	u.height = y2 - y1;
    }
}

