/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit;

import java.util.Stack;
import java.util.EmptyStackException;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Container;
import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;
import fr.pixware.util.StringUtil;
import fr.pixware.util.FileUtil;
import com.xmlmind.awtutil.AWTUtil;
import com.xmlmind.awtutil.Clipboard;
import com.xmlmind.gadget.*;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.*;

public class DocumentView extends RootGadget 
                          implements DocumentListener, RectangleList {

    public static final int PREVIOUS_CHAR    = 0;
    public static final int NEXT_CHAR        = 1;
    public static final int PREVIOUS_WORD    = 2;
    public static final int NEXT_WORD        = 3;
    public static final int PREVIOUS_TEXT    = 4;
    public static final int NEXT_TEXT        = 5;
    public static final int PREVIOUS_ELEMENT = 6;
    public static final int NEXT_ELEMENT     = 7;
    public static final int TEXT_BEGIN       = 8;
    public static final int TEXT_END         = 9;
    public static final int ELEMENT_BEGIN    = 10;
    public static final int ELEMENT_END      = 11;
    public static final int DOCUMENT_BEGIN   = 12;
    public static final int DOCUMENT_END     = 13;
    public static final int PREVIOUS_LINE    = 14;
    public static final int NEXT_LINE        = 15;
    public static final int LINE_BEGIN       = 16;
    public static final int LINE_END         = 17;

    public static final int TEXT_PARENT      = 0;
    public static final int PARENT_OR_TEXT   = 1;
    public static final int PARENT           = 2;
    public static final int CHILD            = 3;
    public static final int CHILD_OR_NONE    = 4;
    public static final int PREVIOUS_SIBLING = 5;
    public static final int NEXT_SIBLING     = 6;

    public static final int PASTE_BEFORE = 0;
    public static final int PASTE_INTO   = 1;
    public static final int PASTE_AFTER  = 2;

    public static final int PREVIOUS = 0;
    public static final int NEXT     = 1;
    public static final int BEGIN    = 2;
    public static final int END      = 3;

    private ViewFactory viewFactory;
    private ElementStructureEditor structureEditor;
    private boolean editable = true;
    private boolean textContainerImplicitlySelected = true;

    private ContextChangeListener[] contextChangeListeners = 
        new ContextChangeListener[0];
    private final ContextChangeEvent contextChangeEvent = 
        new ContextChangeEvent(this);

    private DocumentModel document = null;
    private Caret caret = null;
    private SubelementSelection subelementSelection = null;

    private boolean batchEditing = false;
    private Gadget refreshRoot = null;

    private static final int IMMEDIATE  = 0;
    private static final int NO_PENDING = 1;
    private static final int PENDING    = 2;
    private int notifyListenersState = IMMEDIATE;
    private int refreshCaretState = IMMEDIATE;
    private int refreshSubelementSelectionState = IMMEDIATE;
    private int actionNesting = 0;

    private boolean extendingSelection = false;

    private TextLineParts textLineParts = new TextLineParts();
    private int magicX = -1;

    // Character to String conversion.
    private static final int SPACE_CHAR = 0x0020;
    private static final int FIRST_KEY_CHAR1 = SPACE_CHAR+1;
    private static final int LAST_KEY_CHAR1 = 0x007E;
    private static final int NBSP_CHAR = 0x00A0;
    private static final int FIRST_KEY_CHAR2 = NBSP_CHAR;
    private static final int LAST_KEY_CHAR2 = 0x00FF;
    private char[] chars1 = new char[1];
    private static String[] charStrings = new String[256];
    static {
	char[] oneChar = new char[1];
	for (int i = 0; i < charStrings.length; ++i) {
	    oneChar[0] = (char) i;
	    charStrings[i] = new String(oneChar, 0, 1);
	}
    }

    private ViewTreeDetacher detachViewTree = new ViewTreeDetacher();
    private ViewTreeAttacher attachViewTree = new ViewTreeAttacher();

    // Scratch variables.
    private Rectangle rect = new Rectangle();
    private TextOffset textOffset = new TextOffset();
    private Point point = new Point();
    private Rectangle appendRect = null;
    private TextRangeList rangeList = new TextRangeList();
    private GadgetListImpl gadgetList = new GadgetListImpl();
    private Gadget[] gadgetListGadgets = null;

    // -----------------------------------------------------------------------

    public DocumentView(Container panel) {
	super(panel);

	viewFactory = createViewFactory();
	structureEditor = createElementStructureEditor();
    }

    protected ViewFactory createViewFactory() {
	return new DefaultViewFactory(this);
    }

    protected ElementStructureEditor createElementStructureEditor() {
	return new DefaultElemStructEditor();
    }

    protected Caret createCaret() {
	return new DefaultCaret(this);
    }

    protected SubelementSelection createSubelementSelection() {
	return new DefaultSubelementSelection(this);
    }

    public void setDocument(DocumentModel doc) {
	if (document != null) {
	    if (getChildCount() > 0)
		removeChild(0);

	    caret.setDocument(null);
	    caret = null;
	    subelementSelection.setDocument(null);
	    subelementSelection = null;
	    viewFactory.setDocument(null);
	    structureEditor.setDocument(null);
	    document.removeDocumentListener(this);
	    document = null;

	    batchEditing = false;
	    refreshRoot = null;

	    actionNesting = 0;
	    notifyListenersState = IMMEDIATE;
	    refreshCaretState = IMMEDIATE;
	    refreshSubelementSelectionState = IMMEDIATE;

	    magicX = -1;

	    // View data has been removed from the model as a side effect of
	    // removeDocumentListener.
	}

	if (doc != null) {
	    document = doc;
	    document.addDocumentListener(this, /*hasData*/ true);
	    structureEditor.setDocument(document);
	    viewFactory.setDocument(document);
	    subelementSelection = createSubelementSelection();
	    subelementSelection.setDocument(document);
	    caret = createCaret();
	    caret.setDocument(document);

	    ElementModel element = document.getRootElement();
	    attachViewTree.run(element);

	    int viewIndex = document.documentListenerIndex(this);
	    ElementView view = 
		(ElementView) element.getDocumentListenerData(viewIndex);

	    gadgetList.removeAll();
	    view.getGadgets(gadgetList, viewIndex);
	    if (gadgetList.size > 0)
		addChild(gadgetList.list[gadgetList.size-1]);
	}

	notifyContextChangeListeners();
    }

    public DocumentModel getDocument() {
	return document;
    }

    public ViewFactory getViewFactory() {
	return viewFactory;
    }

    public ElementStructureEditor getElementStructureEditor() {
	return structureEditor;
    }

    public Caret getCaret() {
	return caret;
    }

    public SubelementSelection getSubelementSelection() {
	return subelementSelection;
    }

    // -----------------------------------------------------------------------

    public void makeLocationVisible(TextLocation location) {
	TextModel text = location.getText();
	if (text != null)
	    makeLocationVisible(text, location.getOffset());
    }

    public void makeLocationVisible(TextModel text, int offset) {
	modelToView(text, offset, rect);
	if (rect.height > 0) 
	    makeRectangleVisible(
		rect.x, rect.y, 
		((rect.width <= 0)? 1 : rect.width), rect.height);
    }

    public SubelementView getSubelementView(SubelementModel sub) {
	int viewIndex = document.documentListenerIndex(this);
	return (SubelementView) sub.getDocumentListenerData(viewIndex);
    }

    public Gadget getGadgetContainingView(SubelementView view) {
	gadgetList.removeAll();
	int viewIndex = document.documentListenerIndex(this);
	view.getGadgets(gadgetList, viewIndex);
	if (gadgetList.size == 0)
	    return null;

	Gadget gadget = gadgetList.list[0];
	for (int i = 1; i < gadgetList.size; ++i) {
	    Gadget gadget2 = gadgetList.list[i];
	    
	    gadget = findCommonAncestor(gadget, gadget2);
	    if (gadget == null)
		return null;
	}

	return gadget;
    }

    public void modelToView(SubelementModel sub, RectangleListImpl rectList) {
	rectList.removeAll();

	int viewIndex = document.documentListenerIndex(this);
	SubelementView view = 
	    (SubelementView) sub.getDocumentListenerData(viewIndex);

	if (view != null) {
	    gadgetList.removeAll();
	    view.getGadgets(gadgetList, viewIndex);

	    for (int i = 0; i < gadgetList.size; ++i) {
		Gadget gadget = gadgetList.list[i];

		point.x = point.y = 0;
		gadget.gadgetToRoot(point);
		if (point.x >= width || 
		    point.y >= height ||
		    point.x+gadget.width <= 0 || 
		    point.y+gadget.height <= 0)
		    // Gadget out of sight (for example collapsed).
		    continue;

		rectList.appendRectangle(point.x, point.y, 
					 gadget.width, gadget.height);
	    }
	}
    }

    public void modelToView(TextLocation location, Rectangle r) {
	TextModel text = location.getText();
	if (text == null) {
	    r.x = r.y = r.width = r.height = 0;
	} else {
	    modelToView(text, location.getOffset(), r);
	}
    }

    public void modelToView(TextModel text, int offset, Rectangle r) {
	r.x = r.y = r.width = r.height = 0;
	appendRect = r;

	int viewIndex = document.documentListenerIndex(this);
	TextView view = (TextView) text.getDocumentListenerData(viewIndex);

	if (view != null) {
	    gadgetList.removeAll();
	    view.getGadgets(gadgetList, viewIndex);

	    for (int i = 0; i < gadgetList.size; ++i) {
		Gadget gadget = gadgetList.list[i];
		if (!(gadget instanceof TextViewPart))
		    continue;

		point.x = point.y = 0;
		gadget.gadgetToRoot(point);
		if (point.x >= width || 
		    point.y >= height ||
		    point.x+gadget.width <= 0 || 
		    point.y+gadget.height <= 0)
		    // Gadget out of sight (for example collapsed).
		    continue;

		if (((TextViewPart) gadget).modelToGadget(offset, 0, 
							  this) > 0) {
		    r.x += point.x;
		    r.y += point.y;
		    break;
		}
		// Otherwise, the offset of interest is not rendered by this
		// TextViewPart. Try another TextViewPart.
	    }
	}
    }

    public void appendRectangle(int x, int y, int width, int height) {
	appendRect.x = x;
	appendRect.y = y;
	appendRect.width = width;
	appendRect.height = height;
    }

    public void modelToView(TextLocation location1, TextLocation location2, 
			    RectangleListImpl rectList) {
	TextModel text1 = location1.getText();
	TextModel text2 = location2.getText();

	if (text1 == null || text2 == null) {
	    rectList.removeAll();
	} else {
	    modelToView(text1, location1.getOffset(), 
			text2, location2.getOffset(), rectList);
	}
    }

    public void modelToView(TextModel text1, int offset1, 
			    TextModel text2, int offset2, 
			    RectangleListImpl rectList) {
	rectList.removeAll();

	int viewIndex = document.documentListenerIndex(this);

	structureEditor.get(text1, offset1, text2, offset2, rangeList);
	TextRange[] ranges = rangeList.list;
	int rangeCount = rangeList.size;

	for (int i = 0; i < rangeCount; ++i) {
	    TextRange range = ranges[i];
	    TextView view = (TextView) 
		range.text.getDocumentListenerData(viewIndex);
	    if (view == null)
		continue;

	    gadgetList.removeAll();
	    view.getGadgets(gadgetList, viewIndex);

	    for (int j = 0; j < gadgetList.size; ++j) {
		Gadget gadget = gadgetList.list[j];
		if (!(gadget instanceof TextViewPart))
		    continue;

		point.x = point.y = 0;
		gadget.gadgetToRoot(point);
		if (point.x >= width || 
		    point.y >= height ||
		    point.x+gadget.width <= 0 || 
		    point.y+gadget.height <= 0)
		    // Gadget out of sight (for example collapsed).
		    continue;

		int appendRectCount = ((TextViewPart) gadget).modelToGadget(
		    range.offset, range.count, rectList);
		if (appendRectCount > 0) {
		    Rectangle[] rectangles = rectList.list;
		    int rectCount = rectList.size;

		    for (int k = rectCount-appendRectCount; k<rectCount; ++k) {
			Rectangle rect = rectangles[k];
			rect.x += point.x;
			rect.y += point.y;
		    }
		}
		// Otherwise, the text range of interest doesn't intersect the
		// range displayed by this TextViewPart. Try another
		// TextViewPart.
	    }
	}
    }

    public void viewToModel(int viewX, int viewY, TextOffset model) {
	point.x = viewX;
	point.y = viewY;
	Gadget target = findGadgetAt(point);

	if (target == null || !(target instanceof TextViewPart)) {
	    model.text = null;
	    model.offset = 0;
	} else {
	    TextViewPart part = (TextViewPart) target;
	    model.text = part.getTextView().getText();
	    model.offset = part.gadgetToModel(point.x, point.y);
	}
    }

    // -----------------------------------------------------------------------

    public void paint(Graphics g) {
	super.paint(g);

	if (caret != null) {
	    // Contains a document.
	    subelementSelection.paint(g);

	    caret.paintSelection(g);
	}

	paintText(this, g);

	if (caret != null)
	    caret.paintCaret(g);
    }

    private static void paintText(GadgetContainer container, Graphics g) {
	Rectangle clip = g.getClipBounds();

	Gadget[] children = container.getChildren();
	int childCount = children.length;
	int containerWidth = container.width;

	for (int i = 0; i < childCount; ++i) {
	    Gadget child = children[i];

	    if (child instanceof TextViewPart) {
		if (clip == null || 
		    // Child intersects clip.
		    (child.x < clip.x+clip.width && 
		     child.y < clip.y+clip.height &&
		     child.x+child.width > clip.x && 
		     child.y+child.height > clip.y)) {

		    // Do not clip width too much otherwise slanted letters
		    // such as a 'd' in bold+italic are slightly cut.

		    Graphics g2 = 
			g.create(child.x, child.y, 
				 containerWidth - child.x, child.height);
		    try {
			((TextViewPart) child).paintText(g2);
		    } finally {
			g2.dispose();
		    }
		}
	    } else if (child instanceof GadgetContainer) {
		if (clip == null || 
		    (child.x < clip.x+clip.width && 
		     child.y < clip.y+clip.height &&
		     child.x+child.width > clip.x && 
		     child.y+child.height > clip.y)) {

		    Graphics g2 = g.create(child.x, child.y, 
					   child.width, child.height);
		    try {
			paintText((GadgetContainer) child, g2);
		    } finally {
			g2.dispose();
		    }
		}
	    }
	}
    }

    public void print(Graphics g) {
	super.print(g);
	paintText(this, g);
	// Do not paint marks.
    }

    // -----------------------------------------------------------------------

    public void setEditable(boolean editable) {
	this.editable = editable;

	if (caret != null)
	    caret.setBlinking(editable);
    }

    public boolean isEditable() {
	return editable;
    }

    public void beginAction() {
	if (actionNesting == 0) {
	    notifyListenersState = NO_PENDING;
	    refreshCaretState = NO_PENDING;
	    refreshSubelementSelectionState = NO_PENDING;
	}
	++actionNesting;
    }

    public void endAction() {
	--actionNesting;
	if (actionNesting <= 0) {
	    actionNesting = 0;

	    if (refreshSubelementSelectionState == NO_PENDING) {
		refreshSubelementSelectionState = IMMEDIATE;
	    } else {
		refreshSubelementSelectionState = IMMEDIATE;
		refreshSubelementSelection();
	    }

	    if (refreshCaretState == NO_PENDING) {
		refreshCaretState = IMMEDIATE;
	    } else {
		refreshCaretState = IMMEDIATE;
		refreshCaret();
	    }

	    if (notifyListenersState == NO_PENDING) {
		notifyListenersState = IMMEDIATE;
	    } else {
		notifyListenersState = IMMEDIATE;
		notifyContextChangeListeners();
	    }
	}
    }

    public void selectAll() {
	if (caret == null)
	    // No document.
	    return;

	TextLocation mark = caret.mark();
	if (mark == null)
	    // No text element.
	    return;

	beginAction();

	subelementSelection.setSubelement(null);

	TextLocation point = caret.getPoint();
	mark.moveToDocumentBegin();
	point.moveToDocumentEnd();

	endAction();
    }

    public void selectText(TextRange range) {
	selectText(range.text, range.offset,
		   range.text, range.offset + range.count);
    }

    public void selectText(TextOffset markLocation, TextOffset pointLocation) {
	selectText(markLocation.text, markLocation.offset,
		   pointLocation.text, pointLocation.offset);
    }

    public void selectText(TextModel markText, int markOffset,
			   TextModel pointText, int pointOffset) {
	if (caret == null)
	    // No document.
	    return;

	if (markText.getDocument() == null || pointText.getDocument() == null)
	    // Detached from document.
	    return;
	// The document contains at least one #PCDATA and point is on it.

	beginAction();

	caret.unmark();
	subelementSelection.setSubelement(null);

	TextLocation point = caret.getPoint();
	if (markText != pointText || markOffset != pointOffset) {
	    TextLocation mark = caret.mark();
	    mark.moveTo(markText, markOffset);
	}
	point.moveTo(pointText, pointOffset);

	endAction();
    }

    public void selectSubelement(SubelementModel sub) {
	if (caret == null)
	    // No document.
	    return;

	if (sub != null && sub.getDocument() == null)
	    // Detached from document.
	    return;

	beginAction();

	caret.unmark();
	subelementSelection.setSubelement(sub);
	if (sub != null)
	    ensurePointIsInside(sub);

	endAction();
    }

    public boolean ensurePointIsInside(SubelementModel sub) {
	if (caret == null)
	    // No document.
	    return false;

	if (sub.getDocument() == null)
	    // Detached from document.
	    return false;

	SubelementModel pointAncestor = caret.getPoint().getText();

	while (pointAncestor != null) {
	    if (pointAncestor == sub)
		return true;

	    pointAncestor = pointAncestor.getParentElement();
	}

	return movePointInto(sub, false);
    }

    public boolean movePointInto(SubelementModel sub, boolean selectIfNoText) {
	if (caret == null)
	    // No document.
	    return false;

	if (sub.getDocument() == null)
	    // Detached from document.
	    return false;

	TextModel firstText = 
	    (TextModel) Traversal.traverse(sub, Traversal.textFinder);
	if (firstText == null && !selectIfNoText)
	    // Sub does not contain text.
	    return false;

	beginAction();

	caret.unmark();
	if (firstText == null)
	    subelementSelection.setSubelement(sub);
	else
	    caret.getPoint().moveTo(firstText, 0);

	endAction();

	return true;
    }

    public boolean ensurePointIsVisible() {
	if (caret == null)
	    // No document.
	    return false;

	Rectangle r1 = caret.getBounds();
	if (r1.height <= 0)
	    // No text ==> no point ==> no bounds.
	    return false;

	Rectangle r2 = getVisibleRectangle();
	if (r2.height <= 0)
	    return false;

	intersection(r1, r2, rect);
	if (rect.height > 0)
	    return true;

	int viewX = r2.x + (r2.width/2);
	int endViewY = r2.y + r2.height;
	int stepSize = r2.height/10;
	if (stepSize == 0)
	    stepSize = 1;
	int subStepCount = 2;
	int subStepSize = stepSize/subStepCount;
	if (subStepSize == 0)
	    subStepSize = 1;
	int viewY;

	for (int i = 0; i < subStepCount; ++i) {
	    viewY = r2.y + i*subStepSize;

	    while (viewY < endViewY) {
		pickText(viewX, viewY, textOffset);
		if (textOffset.text != null) {
		    caret.getPoint().moveTo(textOffset.text,textOffset.offset);
		    return true;
		}

		viewY += stepSize;
	    }
	}

	return false;
    }

    public ElementModel refreshSelected() {
	if (caret == null)
	    // No document.
	    return null;

	ContextCommandInfo info = new ContextCommandInfo();
	doGetContextCommandInfo(/*editable*/ false, /*clipboardContent*/ null, 
				info);

	ElementModel refreshed = null;

	switch (info.context) {
	case ContextCommandInfo.SUBELEMENT_SELECTION:
	    refreshed = rebuildSubelementView(info.subelement);
	    // Note that the mark have not been erased/repainted because the
	    // model has not changed, so we need to manually repaint them.
	    refreshSubelementSelection();
	    refreshCaret();
	    break;
	case ContextCommandInfo.POINT_IN_ELEMENT:
	    refreshed = rebuildSubelementView(info.pointText);
	    refreshCaret();
	    break;
	}

	return refreshed;
    }

    private ElementModel rebuildSubelementView(SubelementModel sub) {
	if (document == null)
	    return null;

	if (sub.getDocument() == null)
	    // Detached from document.
	    return null;

	// This is not an action. This is more like refreshCaret and
	// refreshSubelementSelection.

	ElementModel element;
	if (sub.getSubelementType() == SubelementModel.ELEMENT)
	    element = (ElementModel) sub;
	else
	    element = sub.getParentElement();

	rebuildElementView(element, document.documentListenerIndex(this));
	return element;
    }

    // -----------------------------------------------------------------------

    public boolean handleMouseEvent(MouseEvent e) {
	int modifiers = e.getModifiers();
	if ((modifiers & MouseEvent.BUTTON1_MASK) == 0)
	    return false;

	int clickCount = e.getClickCount();

	switch (e.getID()) {
	case MouseEvent.MOUSE_PRESSED:
	    switch (clickCount) {
	    case 1:
		// Just in case!
		requestFocus();

		if (modifiers == MouseEvent.BUTTON1_MASK)
		    beginSelection(e.getX(), e.getY());
		break;
	    case 2:
		if (modifiers == MouseEvent.BUTTON1_MASK)
		    selectWord();
		break;
	    case 3:
		if (modifiers == MouseEvent.BUTTON1_MASK)
		    selectText();
		break;
	    }
	    break;
	case MouseEvent.MOUSE_DRAGGED:
	    if (modifiers == MouseEvent.BUTTON1_MASK)
		extendSelection(e.getX(), e.getY());
	    break;
	case MouseEvent.MOUSE_RELEASED:
	    // clickCount is null during and after a mouse drag.
	    if (clickCount <= 1 && 
		modifiers == MouseEvent.BUTTON1_MASK)
		endSelection(e.getX(), e.getY());
	    break;
	case MouseEvent.MOUSE_CLICKED:
	    if (clickCount == 1) {
		if (modifiers == (MouseEvent.BUTTON1_MASK|
				  MouseEvent.SHIFT_MASK))
		    selectTo(e.getX(), e.getY());
		else if (modifiers == (MouseEvent.BUTTON1_MASK|
				       AWTUtil.MOUSE_MODIFIER))
		    selectSubelement(e.getX(), e.getY());
	    }
	    break;
	}

	return true;
    }

    public void beginSelection(int pointX, int pointY) {
	if (caret == null)
	    // No document.
	    return;

	beginAction();

	pickText(pointX, pointY, textOffset);
	if (textOffset.text != null) {
	    extendingSelection = true;

	    caret.unmark();
	    subelementSelection.setSubelement(null);

	    caret.getPoint().moveTo(textOffset.text, textOffset.offset);
	}
    }

    private void pickText(int viewX, int viewY, TextOffset model) {
	TextViewPart part = null;

	point.x = viewX;
	point.y = viewY;
	Gadget target = findGadgetAt(point);

	if (target != null) {
	    if (target instanceof TextViewPart) {
		part = (TextViewPart) target;
	    } else if (target instanceof GadgetContainer) {
		viewX = point.x;
		viewY = point.y;
		GadgetContainer container = (GadgetContainer) target;

		for (;;) {
		    Gadget[] children = container.getChildren();
		    Gadget closestChild = null;
		    int smallestDistance = Integer.MAX_VALUE;

		    for (int i = 0; i < children.length; ++i) {
			Gadget child = children[i];

			if (viewY >= child.y && viewY < child.y+child.height &&
			    ((child instanceof TextViewPart) ||
			     (child instanceof GadgetContainer))) {
			    int distance = 0;
			    if (viewX < child.x)
				distance = child.x - viewX;
			    else if (viewX >= child.x+child.width)
				distance = viewX - (child.x+child.width);

			    if (closestChild == null ||
				distance < smallestDistance) {
				closestChild = child;
				smallestDistance = distance;
			    }
			}
		    }

		    if (closestChild == null)
			break;
		    viewX -= closestChild.x;
		    viewY -= closestChild.y;

		    if (closestChild instanceof TextViewPart) {
			// Found something.
			point.x = viewX;
			point.y = viewY;
			part = (TextViewPart) closestChild;
			break;
		    }

		    // Search deeper in the hierarchy.
		    container = (GadgetContainer) closestChild;
		}
	    }
	}

	if (part == null) {
	    model.text = null;
	    model.offset = 0;
	} else {
	    // We really want the offset to be inside part.
	    Gadget gadget = (Gadget) part;
	    int x = point.x;
	    if (x < 0)
		x = 0;
	    else if (x >= gadget.width)
		x = gadget.width-1;

	    model.text = part.getTextView().getText();
	    model.offset = part.gadgetToModel(x, point.y);
	}
    }

    public void extendSelection(int pointX, int pointY) {
	if (extendingSelection) {
	    selectTo(pointX, pointY);

	    if (refreshCaretState == PENDING) {
		refreshCaretState = IMMEDIATE;
		refreshCaret();
		refreshCaretState = NO_PENDING;
	    }
	}
    }

    public void endSelection(int pointX, int pointY) {
	if (extendingSelection) {
	    extendingSelection = false;	

	    // Remove mark if it is useless ---

	    TextLocation mark = caret.getMark();
	    if (mark != null) {
		TextLocation point = caret.getPoint();

		if (mark.getText() == point.getText() &&
		    mark.getOffset() == point.getOffset())
		    caret.unmark();
	    }

	    endAction();
	} else {
	    endAction();

	    selectSubelement(pointX, pointY);
	}
    }

    public void selectTo(int pointX, int pointY) {
	if (caret == null)
	    // No document.
	    return;

 	pickText(pointX, pointY, textOffset);
	if (textOffset.text != null) {
	    beginAction();

	    if (!extendingSelection)
		// Little optimization useful when extending selection.
		subelementSelection.setSubelement(null);

	    // No-op if mark already set.
	    caret.mark();
	    caret.getPoint().moveTo(textOffset.text, textOffset.offset);

	    endAction();
	}
    }

    public void selectWord() {
	if (caret == null)
	    // No document.
	    return;

	beginAction();

	caret.unmark();

	TextLocation mark = caret.mark();
	if (mark != null) {
	    subelementSelection.setSubelement(null);

	    mark.moveToWordBegin();
	    caret.getPoint().moveToWordEnd();
	}
	// Else, no text element => cannot mark.

	endAction();
    }

    public void selectText() {
	if (caret == null)
	    // No document.
	    return;

	beginAction();

	caret.unmark();

	TextLocation mark = caret.mark();
	if (mark != null) {
	    subelementSelection.setSubelement(null);

	    TextLocation point = caret.getPoint();
	    TextModel pointText = point.getText();

	    mark.moveTo(pointText, 0);
	    point.moveTo(pointText, pointText.getTextLength());
	}

	endAction();
    }

    public void selectSubelement(int x, int y) {
	if (caret == null)
	    // No document.
	    return;

	point.x = x;
	point.y = y;
	Gadget gadget = findGadgetAt(point);
	if (gadget != null && (gadget instanceof SubelementViewPart)) {
	    beginAction();

	    caret.unmark();

	    SubelementModel sub = 
		((SubelementViewPart) gadget).getView().getModel();
	    subelementSelection.setSubelement(sub);
	    ensurePointIsInside(sub);

	    endAction();
	}
    }

    // -----------------------------------------------------------------------

    public void focusGained() {
	if (caret != null)
	    caret.setBlinking(true);
    }

    public void focusLost() {
	if (caret != null)
	    caret.setBlinking(false);
    }

    public boolean handleKeyEvent(KeyEvent e) {
	// KeyStrokes handled by a JScrollPane are: PAGE_UP, PAGE_DOWN,
	// CTRL_PAGE_UP (page left), CTRL_PAGE_DOWN (page right), CTRL_HOME
	// (top, left), CTRL_END (bottom, right). Do not interfere with these
	// bindings!

	boolean handled = false;

	switch (e.getID()) {
	case KeyEvent.KEY_PRESSED:
	    {
		int modifiers = e.getModifiers();

		switch (e.getKeyCode()) {
		case KeyEvent.VK_UP:
		    if (modifiers == 0) {
			movePointTo(PREVIOUS_LINE);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(PREVIOUS_LINE);
			handled = true;
		    } else if (modifiers == AWTUtil.KEY_MODIFIER) {
			selectSubelement(PARENT_OR_TEXT);
			handled = true;
		    } else if (modifiers == 
			       (KeyEvent.SHIFT_MASK|AWTUtil.KEY_MODIFIER)) {
			selectSubelement(PREVIOUS_SIBLING);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_DOWN:
		    if (modifiers == 0) {
			movePointTo(NEXT_LINE);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(NEXT_LINE);
			handled = true;
		    } else if (modifiers == AWTUtil.KEY_MODIFIER) {
			selectSubelement(CHILD_OR_NONE);
			handled = true;
		    } else if (modifiers == 
			       (KeyEvent.SHIFT_MASK|AWTUtil.KEY_MODIFIER)) {
			selectSubelement(NEXT_SIBLING);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_LEFT:
		    if (modifiers == 0) {
			movePointTo(PREVIOUS_CHAR);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(PREVIOUS_CHAR);
			handled = true;
		    } else if (modifiers == AWTUtil.KEY_MODIFIER) {
			movePointTo(PREVIOUS_WORD);
			handled = true;
		    } else if (modifiers == 
			       (KeyEvent.SHIFT_MASK|AWTUtil.KEY_MODIFIER)) {
			selectTo(PREVIOUS_WORD);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_RIGHT:
		    if (modifiers == 0) {
			movePointTo(NEXT_CHAR);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(NEXT_CHAR);
			handled = true;
		    } else if (modifiers == AWTUtil.KEY_MODIFIER) {
			movePointTo(NEXT_WORD);
			handled = true;
		    } else if (modifiers == 
			       (KeyEvent.SHIFT_MASK|AWTUtil.KEY_MODIFIER)) {
			selectTo(NEXT_WORD);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_TAB:
		    if (modifiers == 0) {
			if (!insertControlChar('\t'))
			    movePointTo(NEXT_TEXT);
			handled = true;
		    } else if (modifiers == AWTUtil.KEY_MODIFIER) {
			movePointTo(NEXT_TEXT);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK ||
			       (modifiers == 
				(KeyEvent.SHIFT_MASK|AWTUtil.KEY_MODIFIER))) {
			movePointTo(PREVIOUS_TEXT);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_HOME:
		    if (modifiers == 0) {
			movePointTo(TEXT_BEGIN);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(TEXT_BEGIN);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_END:
		    if (modifiers == 0) {
			movePointTo(TEXT_END);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(TEXT_END);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_BACK_SPACE:
		    if (modifiers == 0) {
			deleteChar(/*backward*/ true);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_DELETE:
		    if (modifiers == 0) {
			deleteChar(/*backward*/ false);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_ESCAPE:
		    if (modifiers == 0) {
			cancelSelection();
			handled = true;
		    }
		    break;

		case KeyEvent.VK_SPACE:
		    if (modifiers == AWTUtil.KEY_MODIFIER) {
			insertString(charStrings[NBSP_CHAR]);
			handled = true;
		    } else {
			insertString(charStrings[SPACE_CHAR]);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_ENTER:
		    if (modifiers == 0) {
			if (insertControlChar('\n'))
			    handled = true;
		    }
		    break;

		case KeyEvent.VK_F11:
		    if (modifiers == 0) {
			movePointTo(LINE_BEGIN);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(LINE_BEGIN);
			handled = true;
		    }
		    break;

		case KeyEvent.VK_F12:
		    if (modifiers == 0) {
			movePointTo(LINE_END);
			handled = true;
		    } else if (modifiers == KeyEvent.SHIFT_MASK) {
			selectTo(LINE_END);
			handled = true;
		    }
		    break;
		}
	    }
	    break;
	    
	case KeyEvent.KEY_TYPED:
	    {
		// This, for example, discards 'F' in menu mnemonic sequence
		// "Alt-F S".  
		// Note that Ctrl+Alt sequences are said to generate valid
		// Unicode characters on some PC keyboards.

		int modifiers = e.getModifiers();

		if ((modifiers & KeyEvent.ALT_MASK) == 
		    (modifiers & KeyEvent.CTRL_MASK)) {
		    char c = e.getKeyChar();

//System.err.println("char='"+ c +"' (0x"+ Integer.toString((int) c, 16) +")");

		    if ((c >= FIRST_KEY_CHAR1 && c <= LAST_KEY_CHAR1) ||
			(c >= FIRST_KEY_CHAR2 && c <= LAST_KEY_CHAR2)) {
			insertString(charStrings[c]);
			handled = true;
		    } else if ((c > LAST_KEY_CHAR2 && c <= 0xD7FF) ||
			       (c >= 0xE000 && c <= 0xFFFD)) {
			chars1[0] = c;
			insertString(new String(chars1, 0, 1));
			handled = true;
		    }
		}
	    }
	    break;
	}

	return handled;
    }

    public void movePointTo(int where) {
	if (caret == null)
	    // No document.
	    return;

	TextLocation point = caret.getPoint();
	if (point.getText() != null) {
	    beginAction();

	    caret.unmark();

	    switch (where) {
	    case PREVIOUS_CHAR:
		point.moveToPreviousChar();
		break;
	    case NEXT_CHAR:
		point.moveToNextChar();
		break;
	    case PREVIOUS_WORD:
		point.moveToPreviousWord();
		break;
	    case NEXT_WORD:
		point.moveToNextWord();
		break;
	    case PREVIOUS_TEXT:
		point.moveToPreviousText();
		break;
	    case NEXT_TEXT:
		point.moveToNextText();
		break;
	    case PREVIOUS_ELEMENT:
		point.moveToPreviousElement();
		break;
	    case NEXT_ELEMENT:
		point.moveToNextElement();
		break;
	    case TEXT_BEGIN:
		point.moveToTextBegin();
		break;
	    case TEXT_END:
		point.moveToTextEnd();
		break;
	    case ELEMENT_BEGIN:
		point.moveToElementBegin();
		break;
	    case ELEMENT_END:
		point.moveToElementEnd();
		break;
	    case DOCUMENT_BEGIN:
		point.moveToDocumentBegin();
		break;
	    case DOCUMENT_END:
		point.moveToDocumentEnd();
		break;
	    case PREVIOUS_LINE:
		movePointToLine(point, PREVIOUS);
		break;
	    case NEXT_LINE:
		movePointToLine(point, NEXT);
		break;
	    case LINE_BEGIN:
		movePointToLine(point, BEGIN);
		break;
	    case LINE_END:
		movePointToLine(point, END);
		break;
	    }

	    endAction();
	}
    }

    public void selectTo(int where) {
	if (caret == null)
	    // No document.
	    return;

	if (caret.mark() == null)
	    // No text element.
	    return;

	beginAction();

	subelementSelection.setSubelement(null);

	TextLocation point = caret.getPoint();

	switch (where) {
	case PREVIOUS_CHAR:
	    point.moveToPreviousChar();
	    break;
	case NEXT_CHAR:
	    point.moveToNextChar();
	    break;
	case PREVIOUS_WORD:
	    point.moveToPreviousWord();
	    break;
	case NEXT_WORD:
	    point.moveToNextWord();
	    break;
	case PREVIOUS_TEXT:
	    point.moveToPreviousText();
	    break;
	case NEXT_TEXT:
	    point.moveToNextText();
	    break;
	case PREVIOUS_ELEMENT:
	    point.moveToPreviousElement();
	    break;
	case NEXT_ELEMENT:
	    point.moveToNextElement();
	    break;
	case TEXT_BEGIN:
	    point.moveToTextBegin();
	    break;
	case TEXT_END:
	    point.moveToTextEnd();
	    break;
	case ELEMENT_BEGIN:
	    point.moveToElementBegin();
	    break;
	case ELEMENT_END:
	    point.moveToElementEnd();
	    break;
	case DOCUMENT_BEGIN:
	    point.moveToDocumentBegin();
	    break;
	case DOCUMENT_END:
	    point.moveToDocumentEnd();
	    break;
	case PREVIOUS_LINE:
	    movePointToLine(point, PREVIOUS);
	    break;
	case NEXT_LINE:
	    movePointToLine(point, NEXT);
	    break;
	case LINE_BEGIN:
	    movePointToLine(point, BEGIN);
	    break;
	case LINE_END:
	    movePointToLine(point, END);
	    break;
	}

	endAction();
    }

    private boolean movePointToLine(TextLocation point, int where) {
	int what;
	switch (where) {
	case PREVIOUS:
	    what = TextLineParts.PREVIOUS;
	    break;
	case NEXT:
	    what = TextLineParts.NEXT;
	    break;
	default:
	    what = TextLineParts.CURRENT;
	}
	int partCount = 
	    textLineParts.get(what, point.getText(), point.getOffset(), 
			      document.documentListenerIndex(this));
	if (partCount == 0)
	    return false;
	TextLineParts.Part[] parts = textLineParts.getParts();
	TextLineParts.Part part = null;

	switch (where) {
	case PREVIOUS:
	case NEXT:
	    {
		if (magicX < 0) {
		    modelToView(point, rect);
		    magicX = rect.x;
		}

		int closest = Integer.MAX_VALUE;
		for (int i = 0; i < partCount; ++i) {
		    TextLineParts.Part p = parts[i];

		    int distance = 0;
		    if (magicX < p.x)
			distance = p.x - magicX;
		    else if (magicX >= p.x+p.gadget.width)
			distance = magicX - (p.x+p.gadget.width);

		    if (distance == 0) {
			part = p;
			break;
		    }

		    if (part == null || distance < closest) {
			part = p;
			closest = distance;
		    }
		}

		// When the point is moved, magicX is automatically cleared.
		int nextMagicX = magicX;

		// We really want the offset to be inside part.gadget.
		int x = magicX - part.x;
		if (x < 0)
		    x = 0;
		else if (x >= part.gadget.width)
		    x = part.gadget.width-1;

		TextViewPart vp = (TextViewPart) part.gadget;
		TextModel textModel = vp.getTextView().getText();
		int textOffset = vp.gadgetToModel(x, 0);
		if (!point.moveTo(textModel, textOffset))
		    return false;

		magicX = nextMagicX;
		return true;
	    }

	case BEGIN:
	    {
		int leftmost = Integer.MAX_VALUE;
		for (int i = 0; i < partCount; ++i) {
		    TextLineParts.Part p = parts[i];

		    if (p.x < leftmost) {
			part = p;
			leftmost = p.x;
		    }
		}

		TextViewPart vp = (TextViewPart) part.gadget;
		TextModel textModel = vp.getTextView().getText();
		int textOffset = vp.gadgetToModel(0, 0);
		if (!point.moveTo(textModel, textOffset))
		    return false;

		return true;
	    }

	case END:
	    {
		int rightmost = Integer.MIN_VALUE;
		for (int i = 0; i < partCount; ++i) {
		    TextLineParts.Part p = parts[i];

		    if (p.x > rightmost) {
			part = p;
			rightmost = p.x;
		    }
		}

		TextViewPart vp = (TextViewPart) part.gadget;
		TextModel textModel = vp.getTextView().getText();
		int textOffset = vp.gadgetToModel(part.gadget.width-1, 0);
		if (!point.moveTo(textModel, textOffset))
		    return false;

		return true;
	    }

	default:
	    return false;
	}
    }

    public void cancelSelection() {
	if (caret == null)
	    // No document.
	    return;
	
	beginAction();

	caret.unmark();
	subelementSelection.setSubelement(null);

	endAction();
    }

    public boolean insertControlChar(char c) {
	if (!editable)
	    return false;

	if (caret == null)
	    // No document.
	    return false;

	TextModel text = caret.getPoint().getText();
	if (text == null)
	    // No text element.
	    return false;

	int viewIndex = document.documentListenerIndex(this);
	TextView view = (TextView) text.getDocumentListenerData(viewIndex);
	if (view == null || !view.acceptsControlChar(c))
	    return false;

	// Not an action. Calls insertString which is an action.
	insertString(charStrings[c]);
	return true;
    }

    public void insertString(String s) {
	if (!editable)
	    return;

	if (caret == null)
	    // No document.
	    return;

	beginAction();

	TextLocation point = caret.getPoint();
	TextLocation mark = caret.getMark();
	boolean deleteSelection = (mark != null &&
				   (mark.getText() != point.getText() || 
				    mark.getOffset() != point.getOffset()));

	if (deleteSelection) {
	    document.beginBatchEditing();

	    structureEditor.delete(mark.getText(), mark.getOffset(),
				   point.getText(), point.getOffset());
	    // Mark is not always removed with text.
	    caret.unmark();

	    TextModel text = point.getText();
	    if (text != null)
		text.insertText(point.getOffset(), s);

	    document.endBatchEditing();
	} else {
	    TextModel text = point.getText();
	    if (text != null)
		text.insertText(point.getOffset(), s);
	}

	endAction();
    }

    public void deleteChar(boolean backward) {
	if (!editable)
	    return;

	if (caret == null)
	    // No document.
	    return;

	beginAction();

	TextLocation point = caret.getPoint();
	TextLocation mark = caret.getMark();
	boolean deleteSelection = (mark != null &&
				   (mark.getText() != point.getText() || 
				    mark.getOffset() != point.getOffset()));

	if (deleteSelection) {
	    structureEditor.delete(mark.getText(), mark.getOffset(),
				   point.getText(), point.getOffset());
	    // Mark is not always removed with text.
	    caret.unmark();
	} else {
	    TextModel text = point.getText();
	    if (text != null) {
		int offset = point.getOffset();

		if (backward) {
		    if (offset == 0)
			movePointTo(PREVIOUS_CHAR);
		    else
			text.deleteText(offset-1, 1);
		} else {
		    if (offset == text.getTextLength()) {
			movePointTo(NEXT_CHAR);
		    } else {
			text.deleteText(offset, 1);

			// When deleting the last char. of a right aligned
			// line, the caret is not properly refreshed because
			// (a) there is no text location move, (b) the caret
			// is outside the line at its right.
			refreshCaret();
		    }
		}
	    }
	}

	endAction();
    }

    public void selectSubelement(int which) {
	if (caret == null)
	    // No document.
	    return;

	SubelementModel sub;

	beginAction();

	switch (which) {
	case PARENT_OR_TEXT:
	    if (subelementSelection.getSubelement() != null) {
		subelementSelection.getMark().moveToParent();
		break;
	    }
	    /*FALLTHROUGH*/
	case TEXT_PARENT:
	    {
		TextModel text = caret.getPoint().getText();
		if (text == null) 
		    // No text element.
		    break;

		caret.unmark();
		subelementSelection.setSubelement(text);
	    }
	    break;

	case PARENT:
	    subelementSelection.getMark().moveToParent();
	    break;

	case CHILD:
	    subelementSelection.getMark().moveToChild();
	    break;

	case CHILD_OR_NONE:
	    if (!subelementSelection.getMark().moveToChild()) {
		sub = subelementSelection.getSubelement();

		subelementSelection.setSubelement(null);

		if (sub != null &&
		    sub.getSubelementType() == SubelementModel.TEXT &&
		    caret.getPoint().getText() != sub)
		    caret.getPoint().moveTo((TextModel) sub, 0);
	    }
	    break;

	case PREVIOUS_SIBLING:
	    sub = subelementSelection.getSubelement();
	    if (sub == null && textContainerImplicitlySelected) {
		TextModel text = caret.getPoint().getText();
		if (text != null)
		    sub = text.getParentElement();
	    }
	    if (sub != null) {
		sub = sub.getPreviousSubelement();
		if (sub != null)
		    subelementSelection.setSubelement(sub);
	    }
	    break;

	case NEXT_SIBLING:
	    sub = subelementSelection.getSubelement();
	    if (sub == null && textContainerImplicitlySelected) {
		TextModel text = caret.getPoint().getText();
		if (text != null)
		    sub = text.getParentElement();
	    }
	    if (sub != null) {
		sub = sub.getNextSubelement();
		if (sub != null)
		    subelementSelection.setSubelement(sub);
	    }
	    break;
	}

	sub = subelementSelection.getSubelement();
	if (sub != null)
	    ensurePointIsInside(sub);

	endAction();
    }

    // -----------------------------------------------------------------------
    // API used by Caret and SubelementSelection (which are logically parts
    // of a DocumentView i.e. they could be inner classes)
    // -----------------------------------------------------------------------

    /*package*/ void refreshCaret() {
	if (caret == null)
	    // No document.
	    return;

	switch (refreshCaretState) {
	case NO_PENDING:
	    refreshCaretState = PENDING;
	    break;
	case PENDING:
	    break;
	case IMMEDIATE:
	default:
	    {
		Rectangle r = caret.refresh();
		if (r.height > 0)
		    makeRectangleVisible(r.x, r.y, 
					 ((r.width <= 0)? 1 : r.width), 
					 r.height);
	    }
	}
    }

    /*package*/ void refreshSubelementSelection() {
	if (subelementSelection == null)
	    // No document.
	    return;

	switch (refreshSubelementSelectionState) {
	case NO_PENDING:
	    refreshSubelementSelectionState = PENDING;
	    break;
	case PENDING:
	    break;
	case IMMEDIATE:
	default:
	    {
		Rectangle r = subelementSelection.refresh();
		if (r.height > 0)
		    makeRectangleVisible(r.x, r.y, 
					 ((r.width <= 0)? 1 : r.width), 
					 r.height);
	    }
	}
    }

    /*package*/ void clearMagicPosition() {
	magicX = -1;
    }

    /*package*/ void notifyContextChangeListeners() {
	switch (notifyListenersState) {
	case NO_PENDING:
	    notifyListenersState = PENDING;
	    break;
	case PENDING:
	    break;
	case IMMEDIATE:
	default:
	    {
		for (int i = 0; i < contextChangeListeners.length; ++i)
		    contextChangeListeners[i].contextChanged(
			contextChangeEvent);
	    }
	}
    }

    // -----------------------------------------------------------------------

    public void setTextContainerImplicitlySelected(boolean select) {
	textContainerImplicitlySelected = select;
    }

    public boolean isTextContainerImplicitlySelected() {
	return textContainerImplicitlySelected;
    }

    public void addContextChangeListener(ContextChangeListener listener) {
	if (indexOfContextChangeListener(listener) >= 0)
	    return;

	ContextChangeListener[] newListeners = 
	    new ContextChangeListener[contextChangeListeners.length + 1];

	if (contextChangeListeners.length > 0)
	    System.arraycopy(contextChangeListeners, 0, 
			     newListeners, 0, contextChangeListeners.length);
	newListeners[contextChangeListeners.length] = listener;

	contextChangeListeners = newListeners;
    }

    public void removeContextChangeListener(ContextChangeListener listener) {
	int listenerIndex = indexOfContextChangeListener(listener);
	if (listenerIndex < 0)
	    return;

	ContextChangeListener[] newListeners = 
	    new ContextChangeListener[contextChangeListeners.length - 1];

	if (listenerIndex > 0)
	    System.arraycopy(contextChangeListeners, 0, 
			     newListeners, 0, listenerIndex);

	int first = listenerIndex + 1;
	int tail = contextChangeListeners.length - first;
	if (tail > 0) 
	    System.arraycopy(contextChangeListeners, first, 
			     newListeners, listenerIndex, tail);

	contextChangeListeners = newListeners;
    }

    private int indexOfContextChangeListener(ContextChangeListener listener) {
	for (int i = 0; i < contextChangeListeners.length; ++i) {
	    if (contextChangeListeners[i] == listener)
		return i;
	}
	return -1;
    }

    public void getContext(ContextCommandInfo info) {
	doGetContextCommandInfo(/*editable*/ false, null, info);

	info.commands = 0;

	if (info.context == ContextCommandInfo.TEXT_SELECTION) {
	    boolean inTextOrder = 
		structureEditor.checkTextOrder(info.markText, info.markOffset, 
					      info.pointText,info.pointOffset);

	    info.markAfterPoint = !inTextOrder;
	}
    }

    public void getContextCommandInfo(ContextCommandInfo info) {
	String s = Clipboard.getString(panel);
	if (s != null && s.length() == 0)
	    s = null;
	doGetContextCommandInfo(editable, s, info);
    }

    private void doGetContextCommandInfo(boolean editable, 
					 String clipboardContent, 
					 ContextCommandInfo info) {
	info.clear();
	if (caret == null)
	    // No document.
	    return;

	TextLocation mark = caret.getMark();
	TextLocation point = caret.getPoint();

	// getContextCommandInfo may be called in some transient states
	// where point, mark or subelementSelection are located on subelements
	// being destroyed. That's why markText, pointText, sub attachment to
	// the document are checked.

	TextModel markText, pointText;
	int markOffset, pointOffset;
	SubelementModel sub;

	if (mark != null && 
	    ((markText = mark.getText()) != (pointText = point.getText()) || 
	     mark.getOffset() != point.getOffset()) &&
	    markText.getDocument() != null && 
	    pointText.getDocument() != null) {
	    markOffset = mark.getOffset();
	    pointOffset = point.getOffset();

	    if (editable) 
		info.markAfterPoint =  
		    structureEditor.getCommandInfo(markText, markOffset,
						   pointText, pointOffset,
						   clipboardContent, info);
	    else
		info.commands |= CommandInfo.COPY;

	    info.context = ContextCommandInfo.TEXT_SELECTION;
	    info.markText = markText;
	    info.markOffset = markOffset;
	    info.pointText = pointText;
	    info.pointOffset = pointOffset;
	} else if ((sub = subelementSelection.getSubelement()) != null &&
		   sub.getDocument() != null) {

	    if (editable) {
		structureEditor.getCommandInfo(sub, clipboardContent, info);

		if (sub.getSubelementType() == SubelementModel.ELEMENT &&
		    (pointText = point.getText()) != null &&
		    pointText.getDocument() != null) {

		    if (structureEditor.canSplit((ElementModel) sub, 
						 pointText, point.getOffset()))
			info.commands |= CommandInfo.SPLIT;

		}
	    } else {
		info.commands |= CommandInfo.COPY;
	    }

	    info.context = ContextCommandInfo.SUBELEMENT_SELECTION;
	    info.subelement = sub;
	} else if ((pointText = point.getText()) != null &&
		   pointText.getDocument() != null) {
	    pointOffset = point.getOffset();

	    if (editable) {
		structureEditor.getCommandInfo(pointText, pointOffset, 
					       clipboardContent, info);

		if (textContainerImplicitlySelected) {
		    int commands = info.commands;
		    String[] insertArguments = info.insertArguments;

		    ElementModel textContainer = pointText.getParentElement();
		    structureEditor.getCommandInfo(textContainer, 
						   clipboardContent, info);

		    if ((commands & CommandInfo.PASTE) == 0)
			info.commands &= ~CommandInfo.PASTE;
		    else
			info.commands |= CommandInfo.PASTE;

		    if ((commands & CommandInfo.INSERT) == 0) {
			info.commands &= ~CommandInfo.INSERT;
			info.insertArguments = null;
		    } else {
			info.commands |= CommandInfo.INSERT;
			info.insertArguments = insertArguments;
		    }

		    if (structureEditor.canSplit(textContainer,
						 pointText, pointOffset))
			info.commands |= CommandInfo.SPLIT;
		}
	    } else {
		if (textContainerImplicitlySelected)
		    info.commands |= CommandInfo.COPY;
	    }

	    info.context = (textContainerImplicitlySelected? 
			    ContextCommandInfo.POINT_IN_ELEMENT : 
			    ContextCommandInfo.POINT);
	    info.pointText = pointText;
	    info.pointOffset = pointOffset;
	}
    }

    public void delete() {
	if (!editable || caret == null)
	    return;

	beginAction();

	TextLocation mark = caret.getMark();
	TextLocation point = caret.getPoint();
	if (mark != null && (mark.getText() != point.getText() || 
			     mark.getOffset() != point.getOffset())) {
	    structureEditor.delete(mark.getText(), mark.getOffset(),
				   point.getText(), point.getOffset());
	    // Mark is not always removed with text.
	    caret.unmark();
	} else if (subelementSelection.getSubelement() != null) {
	    structureEditor.delete(subelementSelection.getSubelement());
	} else if (textContainerImplicitlySelected && 
		   point.getText() != null) {
	    structureEditor.delete(point.getText().getParentElement());
	}

	endAction();
    }

    public void copy() {
	beginAction();

	String s = doCopy();
	if (s != null && s.length() > 0) {
	    Clipboard.putString(panel, s);

	    // Clipboard updated (even if model is not) ==> force context
	    // change.
	    notifyContextChangeListeners();
	}

	endAction();
    }

    public String doCopy() {
	if (caret == null)
	    return null;

	beginAction();

	String copy = null;

	TextLocation mark = caret.getMark();
	TextLocation point = caret.getPoint();
	if (mark != null && (mark.getText() != point.getText() || 
			     mark.getOffset() != point.getOffset())) {
	    copy = structureEditor.copy(mark.getText(), mark.getOffset(),
					point.getText(), point.getOffset());
	    // Yes, copying the selection clears it.
	    caret.unmark();
	} else if (subelementSelection.getSubelement() != null) {
	    copy = structureEditor.copy(subelementSelection.getSubelement());
	    subelementSelection.setSubelement(null);
	} else if (textContainerImplicitlySelected && 
		   point.getText() != null) {
	    copy = structureEditor.copy(point.getText().getParentElement());
	}

	endAction();
	return copy;
    }

    public void cut() {
	beginAction();

	String s = doCut();
	if (s != null && s.length() > 0) {
	    Clipboard.putString(panel, s);

	    // Clipboard updated ==> force context change.
	    notifyContextChangeListeners();
	}

	endAction();
    }

    public String doCut() {
	if (!editable || caret == null)
	    return null;

	beginAction();

	String cut = null;

	TextLocation mark = caret.getMark();
	TextLocation point = caret.getPoint();
	if (mark != null && (mark.getText() != point.getText() || 
			     mark.getOffset() != point.getOffset())) {
	    cut = structureEditor.cut(mark.getText(), mark.getOffset(),
				      point.getText(), point.getOffset());
	    // Mark is not always removed with text.
	    caret.unmark();
	} else if (subelementSelection.getSubelement() != null) {
	    cut = structureEditor.cut(subelementSelection.getSubelement());
	} else if (textContainerImplicitlySelected && 
		   point.getText() != null) {
	    cut = structureEditor.cut(point.getText().getParentElement());
	}

	endAction();
	return cut;
    }

    public void paste() {
	String s = Clipboard.getString(panel);
	if (s != null && s.length() > 0) 
	    doPaste(s);
    }

    public void doPaste(String clipboardContent) {
	if (!editable || caret == null)
	    return;

	beginAction();

	TextLocation mark = caret.getMark();
	TextLocation point = caret.getPoint();
	if (mark != null && (mark.getText() != point.getText() || 
			     mark.getOffset() != point.getOffset())) {
	    document.beginBatchEditing();

	    structureEditor.delete(mark.getText(), mark.getOffset(),
				   point.getText(), point.getOffset());
	    // Mark is not always removed with text.
	    caret.unmark();

	    if (point.getText() != null)
		structureEditor.paste(point.getText(), point.getOffset(), 
				      clipboardContent);

	    document.endBatchEditing();
	} else if (subelementSelection.getSubelement() != null) {
	    // Paste *into* the subelement.
	    structureEditor.paste(subelementSelection.getSubelement(), 
				  clipboardContent);
	    subelementSelection.setSubelement(null);
	} else if (point.getText() != null) {
	    structureEditor.paste(point.getText(), point.getOffset(), 
				  clipboardContent);
	}

	endAction();
    }

    public void pasteBefore() {
	String s = Clipboard.getString(panel);
	if (s != null && s.length() > 0) 
	    doPasteBefore(s);
    }

    public void doPasteBefore(String clipboardContent) {
	if (!editable || caret == null)
	    return;
	
	beginAction();

	TextLocation point = caret.getPoint();

	if (subelementSelection.getSubelement() != null) {
	    structureEditor.pasteBefore(subelementSelection.getSubelement(), 
					clipboardContent);
	    subelementSelection.setSubelement(null);
	} else if (textContainerImplicitlySelected && 
		   point.getText() != null) {
	    structureEditor.pasteBefore(point.getText().getParentElement(), 
					clipboardContent);
	}

	endAction();
    }

    public void pasteAfter() {
	String s = Clipboard.getString(panel);
	if (s != null && s.length() > 0)
	    doPasteAfter(s);
    }

    public void doPasteAfter(String clipboardContent) {
	if (!editable || caret == null)
	    return;

	beginAction();

	TextLocation point = caret.getPoint();

	if (subelementSelection.getSubelement() != null) {
	    structureEditor.pasteAfter(subelementSelection.getSubelement(), 
				       clipboardContent);
	    subelementSelection.setSubelement(null);
	} else if (textContainerImplicitlySelected && 
		   point.getText() != null) {
	    structureEditor.pasteAfter(point.getText().getParentElement(), 
				       clipboardContent);
	}

	endAction();
    }

    public void insert(String elementTypeName) {
	if (!editable || caret == null)
	    return;

	beginAction();

	SubelementModel inserted = null;
	TextLocation point;

	if (subelementSelection.getSubelement() != null) {
	    // Insert *into* the subelement.
	    inserted = 
		structureEditor.insert(subelementSelection.getSubelement(), 
				       elementTypeName);
	    subelementSelection.setSubelement(null);
	} else if ((point = caret.getPoint()).getText() != null) {
	    inserted = 
		structureEditor.insert(point.getText(), point.getOffset(), 
				       elementTypeName);
	}

	if (inserted != null)
	    movePointInto(inserted, /*selectIfNoText*/ true);

	endAction();
    }

    public void insertBefore(String elementTypeName) {
	if (!editable || caret == null)
	    return;

	beginAction();

	SubelementModel inserted = null;
	TextLocation point;

	if (subelementSelection.getSubelement() != null) {
	    inserted = 
	      structureEditor.insertBefore(subelementSelection.getSubelement(),
					   elementTypeName);
	    subelementSelection.setSubelement(null);
	} else if (textContainerImplicitlySelected && 
		   (point = caret.getPoint()).getText() != null) {
	    inserted = 
	       structureEditor.insertBefore(point.getText().getParentElement(),
					    elementTypeName);
	}

	if (inserted != null)
	    movePointInto(inserted, /*selectIfNoText*/ true);

	endAction();
    }

    public void insertAfter(String elementTypeName) {
	if (!editable || caret == null)
	    return;

	beginAction();

	SubelementModel inserted = null;
	TextLocation point;

	if (subelementSelection.getSubelement() != null) {
	    inserted = 
	       structureEditor.insertAfter(subelementSelection.getSubelement(),
					   elementTypeName);
	    subelementSelection.setSubelement(null);
	} else if (textContainerImplicitlySelected && 
		   (point = caret.getPoint()).getText() != null) {
	    inserted = 
		structureEditor.insertAfter(point.getText().getParentElement(),
					    elementTypeName);
	}

	if (inserted != null)
	    movePointInto(inserted, /*selectIfNoText*/ true);

	endAction();
    }

    public void convert(String elementTypeName) {
	if (!editable || caret == null)
	    return;

	beginAction();

	TextLocation mark = caret.getMark();
	TextLocation point = caret.getPoint();
	if (mark != null && (mark.getText() != point.getText() || 
			     mark.getOffset() != point.getOffset())) {
	    structureEditor.convert(mark.getText(), mark.getOffset(),
				    point.getText(), point.getOffset(),
				    elementTypeName);
	    // Mark is not always removed with text.
	    caret.unmark();
	} else if (subelementSelection.getSubelement() != null) {
	    structureEditor.convert(subelementSelection.getSubelement(), 
				    elementTypeName);
	} else if (textContainerImplicitlySelected && 
		   point.getText() != null) {
	    structureEditor.convert(point.getText().getParentElement(), 
				    elementTypeName);
	}

	endAction();
    }

    public void split() {
	if (!editable || caret == null)
	    return;

	TextLocation point = caret.getPoint();
	TextModel pointText = point.getText();
	if (pointText == null)
	    return;

	beginAction();

	SubelementModel sub = subelementSelection.getSubelement();
	if (sub != null && 
	    sub.getSubelementType() == SubelementModel.ELEMENT) {
	    structureEditor.split((ElementModel) sub,
				  pointText, point.getOffset());
	    subelementSelection.setSubelement(null);
	} else if (textContainerImplicitlySelected) {
	    structureEditor.split(pointText.getParentElement(),
				  pointText, point.getOffset());
	}

	endAction();
    }

    public void replace(String elementTypeName) {
	if (!editable || caret == null)
	    return;

	beginAction();

	SubelementModel replacement = null;
	TextLocation point;

	if (subelementSelection.getSubelement() != null) {
	    replacement = 
	       structureEditor.replace(subelementSelection.getSubelement(),
				       elementTypeName);
	    subelementSelection.setSubelement(null);
	} else if (textContainerImplicitlySelected && 
		   (point = caret.getPoint()).getText() != null) {
	    replacement = 
		structureEditor.replace(point.getText().getParentElement(),
					elementTypeName);
	}

	if (replacement != null)
	    movePointInto(replacement, /*selectIfNoText*/ true);

	endAction();
    }

    public void join() {
	if (!editable || caret == null)
	    return;

	beginAction();

	SubelementModel sub = subelementSelection.getSubelement();
	if (sub == null && textContainerImplicitlySelected) {
	    TextModel pointText = caret.getPoint().getText();
	    if (pointText != null)
		sub = pointText.getParentElement();
	}

	if (sub != null)
	    structureEditor.join(sub);

	endAction();
    }

    // -----------------------------------------------------------------------

    public boolean insertAtSelected(boolean text, boolean before) {
	if (!editable || document == null)
	    return false;

	// Not an action. Calls insertBefore or insertAfter which are actions.

	ContextCommandInfo info = new ContextCommandInfo();
	info.needArguments = (before? 
			      CommandInfo.INSERT_BEFORE_ARGUMENTS : 
			      CommandInfo.INSERT_AFTER_ARGUMENTS);

	doGetContextCommandInfo(/*editable*/ true, /*clipboardContent*/ null, 
				info);

	if ((before && (info.commands & CommandInfo.INSERT_BEFORE) == 0) ||
	    (!before && (info.commands & CommandInfo.INSERT_AFTER) == 0))
	    return false;

	String elemTypeName = null;

	if (text) {
	    elemTypeName = "#PCDATA";
	} else {
	    ElementModel element = null;

	    switch (info.context) {
	    case ContextCommandInfo.SUBELEMENT_SELECTION:
		if (info.subelement.getSubelementType() == 
		    SubelementModel.ELEMENT)
		    element = (ElementModel) info.subelement;
		break;
	    case ContextCommandInfo.POINT_IN_ELEMENT:
		element = info.pointText.getParentElement();
		break;
	    }

	    if (element != null)
		elemTypeName = element.getElementTypeName();
	}

	if (elemTypeName != null) {
	    if (before) {
		// The argument list is null when there is no DTD.
		if (info.insertBeforeArguments == null ||
		    StringUtil.contains(info.insertBeforeArguments, 
					elemTypeName)) {
		    insertBefore(elemTypeName);
		    return true;
		}
	    } else {
		if (info.insertAfterArguments == null ||
		    StringUtil.contains(info.insertAfterArguments, 
					elemTypeName)) {
		    insertAfter(elemTypeName);
		    return true;
		}
	    }
	}

	return false;
    }

    public boolean splitSelected() {
	if (!editable)
	    return false;

	if (document == null)
	    return false;

	// Not an action. Calls split which is an action.

	ContextCommandInfo info = new ContextCommandInfo();
	info.needArguments = ContextCommandInfo.NO_ARGUMENTS;

	doGetContextCommandInfo(/*editable*/ true, /*clipboardContent*/ null, 
				info);

	if ((info.commands & CommandInfo.SPLIT) != 0) {
	    split();
	    return true;
	}

	return false;
    }

    public boolean pasteAtSelected(String content, int where) {
	if (!editable)
	    return false;

	if (document == null)
	    return false;

	if (content == null || content.length() == 0)
	    return false;

	// Not an action. Calls doPaste which is an action.

	ContextCommandInfo info = new ContextCommandInfo();
	info.needArguments = ContextCommandInfo.NO_ARGUMENTS;

	doGetContextCommandInfo(/*editable*/ true, content, info);

	switch (where) {
	case PASTE_BEFORE:
	    if ((info.commands & CommandInfo.PASTE_BEFORE) != 0) {
		doPasteBefore(content);
		return true;
	    }
	    break;
	case PASTE_INTO:
	    if ((info.commands & CommandInfo.PASTE) != 0) {
		doPaste(content);
		return true;
	    }
	    break;
	case PASTE_AFTER:
	    if ((info.commands & CommandInfo.PASTE_AFTER) != 0) {
		doPasteAfter(content);
		return true;
	    }
	    break;
	}

	return false;
    }

    // -----------------------------------------------------------------------

    public void batchEditingStarted(BatchEditingEvent event, 
				    int listenerIndex) {
	beginAction();

	batchEditing = true;
    }

    public void batchEditingCompleted(BatchEditingEvent event, 
				      int listenerIndex) {
	batchEditing = false;

	if (refreshRoot != null) {
	    if (refreshRoot.getRoot() == null)
		// Just in case refreshRoot has been detached from 
		// the gadget tree.
		refreshRoot = this;

////////////////////////////////////////////////////////////////////////////
//if (refreshRoot == this)
//    System.err.println("*** Refreshing DocumentView after batchEditing");
////////////////////////////////////////////////////////////////////////////

	    refreshRoot.refresh();
	    refreshRoot = null;
	}

	endAction();
    }

    public void elementAttributeChanged(ElementAttributeEvent event, 
					int listenerIndex) {
	ElementModel element = event.getElementSource();
	String attrName = event.getName();

	ElementModel parent = element.getParentElement();
	ElementView parentView = null;
	if (parent != null) 
	    parentView = 
		(ElementView) parent.getDocumentListenerData(listenerIndex);
	else
	    parentView = null;

	if (viewFactory.isElementViewAttributeDependant(element, attrName, 
							parentView)) {

////////////////////////////////////////////////////////////////////////////
//System.err.println("rebuildElementView " + element.getElementTypeName() + 
//" on attribute " + attrName + " change");
////////////////////////////////////////////////////////////////////////////

	    rebuildElementView(element, listenerIndex);
	}
    }

    private void rebuildElementView(ElementModel element, int listenerIndex) {
	Gadget refreshed1;
	ElementModel parent = element.getParentElement();
	if (parent == null) {
	    // Root element.
	    if (getChildCount() > 0)
		removeChild(0);
	    refreshed1 = this;
	} else {
	    refreshed1 = removeChildViewGadgets(parent, element, 
						listenerIndex);
	}

	detachViewTree.run(element);
	Gadget refreshed2 = attachViewTree.run(element);

	refreshGadget(refreshed1, refreshed2);
    }

    private Gadget removeChildViewGadgets(ElementModel parent, 
					  SubelementModel sub, int viewIndex) {
	SubelementView view = 
	    (SubelementView) sub.getDocumentListenerData(viewIndex);
	if (view == null)
	    return null;

	ElementView parentView = 
	    (ElementView) parent.getDocumentListenerData(viewIndex);
	if (parentView == null)
	    // Should never happen.
	    return this;

	return parentView.removeChildViewGadgets(view, viewIndex);
    }

    private void refreshGadget(Gadget refreshed1, Gadget refreshed2) {
	if (refreshed1 == null) {
	    if (refreshed2 != null) {
		refreshGadget(refreshed2);
	    }
	} else {
	    if (refreshed2 == null) {
		refreshGadget(refreshed1);
	    } else {
		refreshed1 = findCommonAncestor(refreshed1, refreshed2);
		if (refreshed1 == null)
		    refreshGadget(this);
		else
		    refreshGadget(refreshed1);
	    }
	}
    }

    public void elementStructureChanged(ElementStructureEvent event, 
					int listenerIndex) {
	ElementModel element = event.getElementSource();
	SubelementModel oldChild;

	Gadget refreshed1, refreshed2;
	switch (event.getId()) {
	case ElementStructureEvent.CHILD_ADDED:
	    refreshed1 = attachViewTree.run(event.getNewChild());
	    if (refreshed1 != null)
		refreshGadget(refreshed1);
	    break;

	case ElementStructureEvent.CHILD_REPLACED:
	    oldChild = event.getOldChild();
	    refreshed1 = removeChildViewGadgets(element, oldChild,
						listenerIndex);
	    detachViewTree.run(oldChild);
	    refreshed2 = attachViewTree.run(event.getNewChild());
	    refreshGadget(refreshed1, refreshed2);
	    break;

	case ElementStructureEvent.CHILD_REMOVED:
	    oldChild = event.getOldChild();
	    refreshed1 = removeChildViewGadgets(element, oldChild, 
						listenerIndex);
	    detachViewTree.run(oldChild);
	    if (refreshed1 != null)
		refreshGadget(refreshed1);
	    break;
	}

	// InsertAfter does not change marks but the editing context may
	// have changed.
	notifyContextChangeListeners();
    }

    public void documentEventHappened(DocumentEvent event, 
				      int listenerIndex) {}

    public void textChanged(TextEvent event, int listenerIndex) {
	TextModel text = event.getTextSource();

	TextView view = (TextView) text.getDocumentListenerData(listenerIndex);
	if (view == null)
	    return;

	Gadget refreshed;
	if (batchEditing) {
	    refreshed = findRefreshedGadget(view, listenerIndex);
	    if (refreshed != null)
		refreshGadget(refreshed);
	} else {
	    String s;

	    switch (event.getId()) {
	    case TextEvent.TEXT_INSERTED:
		s = event.getNewString();
		if (s.length() == 1) {
		    view.charInserted(s.charAt(0), event.getOffset());
		} else {
		    refreshed = findRefreshedGadget(view, listenerIndex);
		    if (refreshed != null)
			refreshed.refresh();
		}
		break;

	    case TextEvent.TEXT_DELETED:
		s = event.getOldString();
		if (s.length() == 1) {
		    view.charDeleted(s.charAt(0), event.getOffset());
		} else {
		    refreshed = findRefreshedGadget(view, listenerIndex);
		    if (refreshed != null)
			refreshed.refresh();
		}
		break;

	    default:
		refreshed = findRefreshedGadget(view, listenerIndex);
		if (refreshed != null)
		    refreshed.refresh();
	    }
	}
    }

    private Gadget findRefreshedGadget(SubelementView view, int viewIndex) {
	gadgetList.removeAll();
	view.getGadgets(gadgetList, viewIndex);
	int gadgetCount = gadgetList.size;
	Gadget[] gadgets = gadgetList.list;

	switch (gadgetCount) {
	case 0:
	    return null;
	case 1:
	    return gadgets[0];
	default:
	    {
		int i;

		// Common parent? --- 

		Gadget commonAncestor = gadgets[0].parent;
		for (i = 1; i < gadgetCount; ++i) {
		    if (gadgets[i].parent != commonAncestor)
			break;
		}

		if (i < gadgetCount) {
		    // No common parent: find common ancestor ---

		    commonAncestor = gadgets[0];
		    for (i = 1; i < gadgetCount; ++i) {
			commonAncestor = findCommonAncestor(commonAncestor, 
							    gadgets[i]);
			if (commonAncestor == null)
			    break;
		    }
		}

		return ((commonAncestor == null)? this : commonAncestor);
	    }
	}
    }

    private void refreshGadget(Gadget gadget) {
	if (!batchEditing) {

////////////////////////////////////////////////////////
//if (gadget == this)
//    System.err.println("*** Refreshing DocumentView");
////////////////////////////////////////////////////////

	    gadget.refresh();
	    return;
	}

	if (refreshRoot == null) {
	    refreshRoot = gadget;
	    return;
	}

	if (refreshRoot == gadget || refreshRoot == this) 
	    return;

	Gadget commonAncestor = findCommonAncestor(refreshRoot, gadget);
	if (commonAncestor == null) {
	    // Gadget or refreshRoot has been detached from the gadget tree.
	    // In either case, we have lost track of what really needs to be
	    // refreshed. Therefore refresh everything.
	    refreshRoot = this;
	} else {
	    refreshRoot = commonAncestor;
	}
    }

    private static final Gadget findCommonAncestor(Gadget g1, Gadget g2) {
	// Is g1 an ancestor of g2? ---

	Gadget ancestor2 = g2;
	while (ancestor2 != null) {
	    if (ancestor2 == g1)
		return g1;

	    ancestor2 = ancestor2.parent;
	}

	// G1 is not an ancestor of g2. Do find common ancestor ---

	Gadget[] ancestors1 = g1.getAncestors();

	ancestor2 = g2;
	while (ancestor2 != null) {
	    for (int i = ancestors1.length-1; i >= 0; --i) {
		if (ancestor2 == ancestors1[i]) 
		    return ancestor2;
	    }

	    ancestor2 = ancestor2.parent;
	}

	// No common ancestor: generally because either g1 or g2 has been
	// detached from the gadget tree.
	return null;
    }

    // -----------------------------------------------------------------------

    private class ViewTreeDetacher extends Traversal.HandlerBase {
	private int viewIndex;

	public void run(SubelementModel sub) {
	    viewIndex = document.documentListenerIndex(DocumentView.this);

	    SubelementView view = 
		(SubelementView) sub.getDocumentListenerData(viewIndex);
	    if (view == null)
		return;

	    Traversal.traverse(sub, this);
	}

	public Object enterElement(ElementModel element) {
	    ElementView view = 
		(ElementView) element.getDocumentListenerData(viewIndex);
	    if (view == null)
		return Traversal.LEAVE_ELEMENT;

	    return null;
	}

	public Object processText(TextModel text) {
	    text.setDocumentListenerData(viewIndex, null);

	    return null;
	}

	public Object leaveElement(ElementModel element) {
	    element.setDocumentListenerData(viewIndex, null);

	    return null;
	}
    }

    private class ViewTreeAttacher extends Traversal.HandlerBase {
	private Stack stack = new Stack();
	private int viewIndex;
	private Gadget refreshed;

	public Gadget run(SubelementModel sub) {
	    viewIndex = document.documentListenerIndex(DocumentView.this);

	    ElementModel parent = sub.getParentElement();
	    if (parent != null) {
		ElementView parentView = 
		    (ElementView) parent.getDocumentListenerData(viewIndex);
		if (parentView == null)
		    // Should never happen.
		    return null;

		stack.push(parentView);
	    }

	    refreshed = null;

	    try {
		Traversal.traverse(sub, this);
	    } finally {
		while (!stack.empty())
		    stack.pop();
	    }

	    Gadget refresh = refreshed;
	    refreshed = null;
	    return refresh;
	}

	public Object enterElement(ElementModel element) {
	    ElementView parentView = 
		(stack.empty()? null : (ElementView) stack.peek());
	    ElementView view = 
		DocumentView.this.viewFactory.createElementView(element, 
								parentView);
	    if (view == null)
		return Traversal.LEAVE_ELEMENT;
	    stack.push(view);

	    return null;
	}

	public Object processText(TextModel text) {
	    ElementView parentView = 
		(stack.empty()? null : (ElementView) stack.peek());
	    TextView view = 
		DocumentView.this.viewFactory.createTextView(text, parentView);

	    if (view != null) {
		text.setDocumentListenerData(viewIndex, view);
		// parentView cannot be null: it is not possible to have
		// a TextModel as the root element.
		refreshed = parentView.addChildViewGadgets(view, viewIndex);
	    }

	    return null;
	}

	public Object leaveElement(ElementModel element) {
	    ElementView view = (ElementView) stack.peek();
	    stack.pop();

	    ElementView parentView = 
		(stack.empty()? null : (ElementView) stack.peek());

	    element.setDocumentListenerData(viewIndex, view);
	    if (parentView == null) {
		// Attaching root element view.
		GadgetListImpl list = DocumentView.this.gadgetList;
		list.removeAll();
		view.getGadgets(list, viewIndex);

		if (DocumentView.this.getChildCount() > 0)
		    DocumentView.this.removeChild(0);
		if (list.size > 0)
		    DocumentView.this.addChild(list.list[list.size-1]);

		refreshed = DocumentView.this;
	    } else {
		refreshed = parentView.addChildViewGadgets(view, viewIndex);
	    }

	    return null;
	}
    }
}
