/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit;

import java.awt.Rectangle;
import java.awt.Point;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.xmledit.model.TextModel;
import com.xmlmind.xmledit.model.ElementModel;
import com.xmlmind.xmledit.modeltool.Traversal;

public class TextLineParts implements Traversal.Handler, RectangleList {
    public static final int CURRENT  = 0;
    public static final int PREVIOUS = 1;
    public static final int NEXT     = 2;

    public static class Part {
	public int x, y;      // Gadget origin in root gadget coordinates.
	public Gadget gadget; // implements TextViewPart.
    }
    
    private Part[] parts = new Part[0];
    private int partCount = 0;
    private Rectangle bounds = new Rectangle();

    private int viewIndex;
    private GadgetListImpl gadgetList = new GadgetListImpl();
    private GadgetListImpl gadgetList2 = new GadgetListImpl();
    private Point point = new Point();
    private boolean crossLine;
    private boolean backwards;

    // -----------------------------------------------------------------------

    public int get(int what, TextModel text, int offset, int viewIndex) {
	int i;

	bounds.x = bounds.y = bounds.width = bounds.height = 0;
	partCount = 0;

	TextView view = (TextView) text.getDocumentListenerData(viewIndex);
	if (view == null)
	    return 0;

	gadgetList.removeAll();
	view.getGadgets(gadgetList, viewIndex);

	// Find gadget which contains offset.
	int gadgetIndex = -1;
	for (i = 0; i < gadgetList.size; ++i) {
	    Gadget gadget = gadgetList.list[i];
	    if (!(gadget instanceof TextViewPart))
		continue;

	    if (((TextViewPart) gadget).modelToGadget(offset, 0, this) > 0) {
		gadgetIndex = i;
		break;
	    }
	}
	if (gadgetIndex < 0)
	    return 0;

	this.viewIndex = viewIndex;
	switch (what) {
	case PREVIOUS:
	    {
		crossLine = false;
		for (i = gadgetIndex; i < gadgetList.size; ++i) {
		    if (!processGadget(gadgetList.list[i]))
			break;
		}
		if (i >= gadgetList.size) {
		    // We don't have the end of current line: add more.
		    backwards = false;
		    Traversal.traverseAfter(text, this);
		}

		crossLine = true;
		for (i = gadgetIndex-1; i >= 0; --i) {
		    if (!processGadget(gadgetList.list[i]))
			break;
		}
		// We have no parts of previous line (crossline==true),
		// OR we only have the end of previous line: add more.
		if (i < 0) {
		    backwards = true;
		    Traversal.traverseBefore(text, this);
		}

		if (crossLine) {
		    // Line not crossed.
		    bounds.x = bounds.y = bounds.width = bounds.height = 0;
		    partCount = 0;
		}
	    }
	    break;

	case NEXT:
	    {
		crossLine = false;
		for (i = gadgetIndex; i >= 0; --i) {
		    if (!processGadget(gadgetList.list[i]))
			break;
		}
		if (i < 0) {
		    // We don't have the beginning of current line: add more.
		    backwards = true;
		    Traversal.traverseBefore(text, this);
		}

		crossLine = true;
		for (i = gadgetIndex+1; i < gadgetList.size; ++i) {
		    if (!processGadget(gadgetList.list[i]))
			break;
		}
		// We have no parts of next line (crossline==true),
		// OR we only have the beginning of next line: add more.
		if (i >= gadgetList.size) {
		    backwards = false;
		    Traversal.traverseAfter(text, this);
		}

		if (crossLine) {
		    // Line not crossed.
		    bounds.x = bounds.y = bounds.width = bounds.height = 0;
		    partCount = 0;
		}
	    }
	    break;

	default:
	    {
		crossLine = false;
		for (i = gadgetIndex; i < gadgetList.size; ++i) {
		    if (!processGadget(gadgetList.list[i]))
			break;
		}
		if (i >= gadgetList.size) {
		    // We don't have the end of current line: add more.
		    backwards = false;
		    Traversal.traverseAfter(text, this);
		}

		for (i = gadgetIndex-1; i >= 0; --i) {
		    if (!processGadget(gadgetList.list[i]))
			break;
		}
		if (i < 0) {
		    // We don't have the beginning of current line: add more.
		    backwards = true;
		    Traversal.traverseBefore(text, this);
		}
	    }
	}

	return partCount;
    }

    private boolean processGadget(Gadget gadget) {
	if (!(gadget instanceof TextViewPart))
	    return true;

	point.x = point.y = 0;
	gadget.gadgetToRoot(point);

	if (bounds.height <= 0) {
	    bounds.x = point.x;
	    bounds.y = point.y;
	    bounds.width = gadget.width;
	    bounds.height = gadget.height;
	    addPart(point, gadget);
	    return true;
	}

	if ((point.y >= bounds.y && point.y < bounds.y+bounds.height)  ||
	    (point.y+gadget.height > bounds.y && 
	     point.y+gadget.height <= bounds.y+bounds.height) ||
	    (bounds.y >= point.y && bounds.y < point.y+gadget.height) ||
	    (bounds.y+bounds.height > point.y && 
	     bounds.y+bounds.height <= point.y+gadget.height)) {
	    union(bounds, point.x, point.y, gadget.width, gadget.height);
	    addPart(point, gadget);
	    return true;
	}

	// The line and the gadget are vertically disjointed ---

	if (!crossLine)
	    // Stop traversal, we are done.
	    return false;

	// Get rid of current line and begin a new one.
	bounds.x = point.x;
	bounds.y = point.y;
	bounds.width = gadget.width;
	bounds.height = gadget.height;
	partCount = 0;
	addPart(point, gadget);
	crossLine = false;
	return true;
    }

    private void addPart(Point point, Gadget gadget) {
	if (partCount == parts.length) {
	    Part[] newParts = new Part[Math.max(10, 2*parts.length)];
	    
	    System.arraycopy(parts, 0, newParts, 0, parts.length);
	    for (int i = parts.length; i < newParts.length; ++i)
		newParts[i] = new Part();

	    parts = newParts;
	}

	Part part = parts[partCount++];
	part.x = point.x;
	part.y = point.y;
	part.gadget = gadget;
    }

    private static final void union(Rectangle bounds, 
				    int x, int y, int width, int height) {
	if (width <= 0 || height <= 0)
	    return;

	if (bounds.width <= 0 || bounds.height <= 0) {
	    bounds.x = x;
	    bounds.y = y;
	    bounds.width = width;
	    bounds.height = height;
	    return;
	}

	int x1 = Math.min(bounds.x, x);
	int x2 = Math.max(bounds.x + bounds.width, x + width);
	int y1 = Math.min(bounds.y, y);
	int y2 = Math.max(bounds.y + bounds.height, y + height);

	bounds.x = x1;
	bounds.y = y1;
	bounds.width = x2 - x1;
	bounds.height = y2 - y1;
    }

    public Part[] getParts() {
	return parts;
    }

    public int getPartCount() {
	return partCount;
    }

    public Rectangle getBounds() {
	return bounds;
    }

    // -----------------------------------------------------------------------

    public Object processText(TextModel text) {
	TextView view = (TextView) text.getDocumentListenerData(viewIndex);
	if (view == null)
	    return null;

	gadgetList2.removeAll();
	view.getGadgets(gadgetList2, viewIndex);

	if (backwards) {
	    for (int i = gadgetList2.size-1; i >= 0; --i) {
		if (!processGadget(gadgetList2.list[i]))
		    return text;
	    }
	} else {
	    for (int i = 0; i < gadgetList2.size; ++i) {
		if (!processGadget(gadgetList2.list[i]))
		    return text;
	    }
	}

	return null;
    }

    public Object enterElement(ElementModel element) { return null; }
    public Object leaveElement(ElementModel element) { return null; }

    // -----------------------------------------------------------------------

    public void appendRectangle(int x, int y, int width, int height) {
	// Don't care.
    }
}
