/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit;

import java.util.Enumeration;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.awt.*;
import java.awt.event.*;
import fr.pixware.util.QuickSort;
import com.xmlmind.awtutil.Bitmap;
import com.xmlmind.gadget.*;
import com.xmlmind.xmledit.model.*;

public class TreeView extends GadgetContainer 
                      implements ElementView, ElementViewPart {

    public static class Settings extends CommonSettings {
	public Image collapsedIcon;
	public Image expandedIcon;
	public Image notCollapsibleIcon;

	public boolean showAttributes;
	public Color attributeColor;
	public Font attributeNameFont;
	public Font attributeValueFont;
	public int attributeHorizontalPad;
	public int attributeVerticalPad;

	public static final Bitmap collapsedBitmap = 
	    new Bitmap(13, 13, new byte[] {
		(byte)0x08, (byte)0x00, (byte)0x18, (byte)0x00, (byte)0x38, 
		(byte)0x00, (byte)0x78, (byte)0x00, (byte)0xf8, (byte)0x00, 
		(byte)0xf8, (byte)0x01, (byte)0xf8, (byte)0x03, (byte)0xf8, 
		(byte)0x01, (byte)0xf8, (byte)0x00, (byte)0x78, (byte)0x00, 
		(byte)0x38, (byte)0x00, (byte)0x18, (byte)0x00, (byte)0x08, 
		(byte)0x00
            });
	public static final Bitmap expandedBitmap = 
	    new Bitmap(13, 13, new byte[] {
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, 
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0xff, (byte)0x1f, 
		(byte)0xfe, (byte)0x0f, (byte)0xfc, (byte)0x07, (byte)0xf8, 
		(byte)0x03, (byte)0xf0, (byte)0x01, (byte)0xe0, (byte)0x00, 
		(byte)0x40, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, 
		(byte)0x00
            });

	public Settings() {
	    background = Color.lightGray;
	    foreground = Color.black;
	    font = new Font("SansSerif", Font.BOLD, 10);
	    pad = 4;
	    borderColor = Color.gray;
	    borderRelief = RELIEF_FLAT;

	    collapsedIcon = collapsedBitmap.getImage(foreground);
	    expandedIcon = expandedBitmap.getImage(foreground);
	    notCollapsibleIcon = expandedBitmap.getImage(Color.gray);

	    showAttributes = true;
	    attributeColor = new Color(0, 0, 128);
	    attributeNameFont = new Font("SansSerif", Font.BOLD,10);
	    attributeValueFont = new Font("SansSerif",Font.PLAIN,10);
	    attributeHorizontalPad = 2;
	    attributeVerticalPad = 2;
	}
    }

    // -----------------------------------------------------------------------

    public static final Settings DEFAULT_SETTINGS = new Settings();

    private ElementModel element;
    private Settings settings;
    private boolean collapsed = false;

    private static final int OUT_OF_SIGHT = Integer.MIN_VALUE/2;

    private static MenuSpec menuSpec;
    static {
	String collapseTreeLabel = "collapseTree";
	String expandTreeLabel = "expandTree";

	try {
	    ResourceBundle resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmledit.TreeViewRes");
	    collapseTreeLabel = resourceBundle.getString(collapseTreeLabel);
	    expandTreeLabel = resourceBundle.getString(expandTreeLabel);
	} catch (MissingResourceException ignored) {}

	menuSpec = new MenuSpec(null, new MenuItemSpec[] {
	    new MenuItemSpec(collapseTreeLabel, "collapseTree"),
	    new MenuItemSpec(expandTreeLabel, "expandTree"),
	});
    }

    private static class AttributeCompare implements QuickSort.Compare {
	public int compare(Object o1, Object o2) {
	    return (((String[]) o1)[0]).compareTo(((String[]) o2)[0]);
	}
    }
    private static AttributeCompare attributeCompare = new AttributeCompare();

    // -----------------------------------------------------------------------

    public TreeView(ElementModel element) {
	this(element, DEFAULT_SETTINGS);
    }

    public TreeView(ElementModel element, Settings settings) {
	this.element = element;
	this.settings = settings;
    }

    public SubelementModel getModel() {
	return element;
    }

    public void getGadgets(GadgetList gadgets, int viewIndex) {
	gadgets.appendGadget(this);
    }

    public ElementModel getElement() {
	return element;
    }

    public Gadget addChildViewGadgets(SubelementView childView, 
				      int viewIndex) {
	SubelementModel target = childView.getModel();
	int index = -1;

	SubelementModel sub = element.getFirstSubelement();
	int i = 0;
	while (sub != null) {
	    if (sub == target) {
		index = i;
		break;
	    }

	    sub = sub.getNextSubelement();
	    ++i;
	}

	if (index < 0)
	    return null;

	// Assumes that childView is a WordView, CharView or TreeView.
	insertChild(index, (Gadget) childView);

	return this;
    }

    public Gadget removeChildViewGadgets(SubelementView childView, 
					 int viewIndex) {
	// Assumes that childView is a WordView, CharView or TreeView.
	removeChild((Gadget) childView);

	return this;
    }

    public SubelementView getView() {
	return this;
    }

    public ElementView getElementView() {
	return this;
    }

    public void setSettings(Settings settings) {
	this.settings = settings;
    }

    public Settings getSettings() {
	return settings;
    }

    public void setCollapsed(boolean collapsed) {
	this.collapsed = collapsed;
    }

    public boolean isCollapsed() {
	return collapsed;
    }

    public void getSizeConstraints(Gadget child, Dimension constraints) {
	if (parent != null) {
	    parent.getSizeConstraints(this, constraints);
	    if (constraints.width > 0)
		constraints.width -= 
		    3*settings.pad + settings.collapsedIcon.getWidth(null);
	    constraints.height = -1;
	}
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	FontMetrics metrics = getFontMetrics(settings.font);
	int textHeight = metrics.getAscent() + metrics.getDescent();
	int pad = settings.pad;
	// All icons are assumed to have the same size.
	Image icon = settings.collapsedIcon;

	int childX = 2*pad + icon.getWidth(null);
	int childY = 2*pad + Math.max(icon.getHeight(null), textHeight);
	int maxWidth = metrics.stringWidth(element.getElementTypeName());

	if (collapsed) {
	    for (int i = 0; i < children.length; ++i) {
		Gadget child = children[i];

		// Even when collapsed, the CharView and WordView children
		// are queried for model<->view translation by the 
		// DocumentView.

		if (recursive)
		    child.layout(constrainedWidth, -1, true);

		child.x = child.y = OUT_OF_SIGHT;
	    }
	} else {
	    if (constrainedWidth > 0)
		constrainedWidth -= (childX+pad);

	    if (settings.showAttributes) {
		int attrCount = element.getElementAttributeCount();
		if (attrCount > 0) {
		    int nameWidth = 0;
		    int valueWidth = 0;

		    FontMetrics nameMetrics = 
			getFontMetrics(settings.attributeNameFont);
		    FontMetrics valueMetrics = 
			getFontMetrics(settings.attributeValueFont);
		    int attHeight = Math.max(
			(nameMetrics.getAscent()+nameMetrics.getDescent()),
			(valueMetrics.getAscent()+valueMetrics.getDescent()));

		    childY +=
			attrCount*(attHeight+settings.attributeVerticalPad) +
			pad;

		    Enumeration iter = element.getElementAttributes();
		    while (iter.hasMoreElements()) {
			String[] attr = (String[]) iter.nextElement();

			int w = nameMetrics.stringWidth(attr[0]);
			if (w > nameWidth)
			    nameWidth = w;

			w = nameMetrics.stringWidth(attr[1]);
			if (w > valueWidth)
			    valueWidth = w;
		    }

		    int attrWidth = 
		      nameWidth + settings.attributeHorizontalPad + valueWidth;
			
		    if (constrainedWidth > 0 && attrWidth > constrainedWidth)
			attrWidth = constrainedWidth;

		    if (attrWidth > maxWidth)
			maxWidth = attrWidth;
		}
	    }

	    for (int i = 0; i < children.length; ++i) {
		Gadget child = children[i];

		if (recursive)
		    child.layout(constrainedWidth, -1, true);

		child.x = childX;
		child.y = childY;

		childY += child.height + pad;

		if (child.width > maxWidth)
		    maxWidth = child.width;
	    }
	}

	width = childX + maxWidth + pad;
	height = childY;
    }

    public void paintRelayout(Gadget child, int prevWidth, int prevHeight,
			      int damagedSide, int damagedLength) {
	if (collapsed)
	    return;

	if (child.height == prevHeight &&
	    child.x + prevWidth + settings.pad < width && 
	    child.x + child.width + settings.pad <= width) {
	    // Same height and the new width cannot have changed the width of
	    // the container: just repaint.

	    int fromX;
	    if (damagedSide == RIGHT_SIDE)
		fromX = child.x + child.width - damagedLength;
	    else
		fromX = child.x;
	    repaint(fromX, child.y, width-fromX, child.height);
	    return;
	}

	if (child.width == prevWidth) {
	    // Same width: optimized relayout.

	    int delta = child.height - prevHeight;

	    for (int i = indexOfChild(child)+1; i < children.length; ++i) {
		children[i].y += delta;
	    }
		
	    height += delta;

	    int fromY;
	    if (damagedSide == BOTTOM_SIDE)
		fromY = child.y + child.height - damagedLength;
	    else
		fromY = child.y;
	    parent.paintRelayout(this, width, height-delta, 
				 BOTTOM_SIDE, height-fromY);
	    return;
	}
	
	// Width must probably be updated: non-optimized paintRelayout.

	super.paintRelayout(child, prevWidth, prevHeight,
			    damagedSide, damagedLength);
    }

    public void paint(Graphics g) {
	// Paint box ---

	int pad = settings.pad;

	// Workaround Graphics.fillRect() bug when the rectangle to paint is
	// huge.
	boolean small = (width <= 1000 && height <= 1000);
	Rectangle clip = null;

	if (settings.background != null) {
	    g.setColor(settings.background);

	    if (small) {
		g.fillRect(0, 0, width, height);
	    } else {
		clip = g.getClipBounds();
		g.fillRect(clip.x, clip.y, clip.width, clip.height);
	    }
	}

	if (settings.borderColor != null && pad > 1) {
	    g.setColor(settings.borderColor);

	    if (small) {
		switch (settings.borderRelief) {
		case Settings.RELIEF_RAISED:
		    g.draw3DRect(0, 0, width-1, height-1, true);
		    break;
		case Settings.RELIEF_SUNKEN:
		    g.draw3DRect(0, 0, width-1, height-1, true);
		    break;
		default:
		    g.drawRect(0, 0, width-1, height-1);
		}
	    } else {
		Color bg1;
		Color bg2;
		switch (settings.borderRelief) {
		case Settings.RELIEF_RAISED:
		    bg1 = settings.borderColor.darker();
		    bg2 = settings.borderColor.brighter();
		    break;
		case Settings.RELIEF_SUNKEN:
		    bg1 = settings.borderColor.brighter();
		    bg2 = settings.borderColor.darker();
		    break;
		default:
		    bg1 = bg2 = settings.borderColor;
		}

		if (clip == null)
		    clip = g.getClipBounds();

		if (clip.y + clip.height == height) {
		    g.setColor(bg1);
		    g.drawLine(clip.x,height-1, 
			       clip.x+clip.width-1, height-1);
		}
		if (clip.y == 0) {
		    g.setColor(bg2);
		    g.drawLine(clip.x, 0, clip.x+clip.width-1, 0);
		}
		if (clip.x + clip.width == width) {
		    g.setColor(bg1);
		    g.drawLine(width-1, clip.y, 
			       width-1, clip.y+clip.height-1);
		}
		if (clip.x == 0) {
		    g.setColor(bg2);
		    g.drawLine(0, clip.y, 0, clip.y+clip.height-1);
		}
	    }
	}

	// Paint content ---

	Image icon = currentIcon();
	FontMetrics metrics = g.getFontMetrics(settings.font);
	int ascent = metrics.getAscent();
	int textHeight = ascent + metrics.getDescent();
	int iconHeight = icon.getHeight(null);
	int lineHeight = Math.max(textHeight, iconHeight);

	g.drawImage(icon, pad, pad + (lineHeight-iconHeight)/2, null);

	g.setColor(settings.foreground);
	g.setFont(settings.font);
	int x0 = 2*pad + icon.getWidth(null);
	g.drawString(element.getElementTypeName(), 
		     x0, pad + (lineHeight-textHeight)/2 + ascent);

	if (!collapsed) {
	    if (settings.showAttributes) {
		int attrCount = element.getElementAttributeCount();
		if (attrCount > 0) {
		    int y0 = 2*pad + lineHeight;
		    Graphics g2 = g.create(x0, y0, 
					   width - x0 - pad, height - y0);

		    try {
			paintAttributes(g2, attrCount);
		    } finally {
			g2.dispose();
		    }
		}
	    }

	    super.paint(g);
	}
    }

    private void paintAttributes(Graphics g, int attrCount) {
	Font nameFont = settings.attributeNameFont;
	FontMetrics nameMetrics = g.getFontMetrics(nameFont);

	String[][] attributes = new String[attrCount][];
	int i = 0;

	int nameWidth = 0;
	Enumeration iter = element.getElementAttributes();
	while (iter.hasMoreElements()) {
	    String[] attr = (String[]) iter.nextElement();

	    int w = nameMetrics.stringWidth(attr[0]);
	    if (w > nameWidth)
		nameWidth = w;

	    attributes[i++] = (String[]) attr.clone();
	}

	if (attrCount >= 2) 
	    QuickSort.sort(attributes, 0, attrCount-1, attributeCompare);

	int x1 = nameWidth + settings.attributeHorizontalPad;

	Font valueFont = settings.attributeValueFont;
	FontMetrics valueMetrics = g.getFontMetrics(valueFont);
	int ascent = Math.max(nameMetrics.getAscent(),
			      valueMetrics.getAscent());
	int y0 = ascent;

	int yIncr = ascent + 
	    Math.max(nameMetrics.getDescent(), valueMetrics.getDescent()) + 
	    settings.attributeVerticalPad;

	g.setColor(settings.attributeColor);

	for (i = 0; i < attrCount; ++i) {
	    String[] attr = attributes[i];
	    String name = attr[0];

	    g.setFont(nameFont);
	    g.drawString(name, nameWidth - nameMetrics.stringWidth(name), y0);

	    g.setFont(valueFont);
	    g.drawString(attr[1], x1, y0);

	    y0 += yIncr;
	}
    }

    public void print(Graphics g) {
	int pad = settings.pad;

	Image icon = currentIcon();
	FontMetrics metrics = g.getFontMetrics(settings.font);
	int ascent = metrics.getAscent();
	int textHeight = ascent + metrics.getDescent();
	int iconHeight = icon.getHeight(null);
	int lineHeight = Math.max(textHeight, iconHeight);

	g.drawImage(icon, pad, pad + (lineHeight-iconHeight)/2, null);

	g.setColor(settings.foreground);
	g.setFont(settings.font);
	int x0 = 2*pad + icon.getWidth(null);
	g.drawString(element.getElementTypeName(), 
		     x0, pad + (lineHeight-textHeight)/2 + ascent);

	if (!collapsed) {
	    if (settings.showAttributes) {
		int attrCount = element.getElementAttributeCount();
		if (attrCount > 0) {
		    int y0 = 2*pad + lineHeight;
		    Graphics g2 = g.create(x0, y0, 
					   width - x0 - pad, height - y0);

		    try {
			paintAttributes(g2, attrCount);
		    } finally {
			g2.dispose();
		    }
		}
	    }

	    super.print(g);
	}
    }

    public int[] getNonBreakableVerticalRanges() {
	int lastY;
	if (collapsed || children.length == 0)
	    lastY = height-1;
	else
	    lastY = children[0].y-1;

	return new int[] { 0, lastY };
    }

    private Image currentIcon() {
	Image icon;

	if (notCollapsible()) 
	    icon = settings.notCollapsibleIcon;
	else 
	    icon = collapsed? settings.collapsedIcon : settings.expandedIcon;

	return icon;
    }

    private boolean notCollapsible() {
	return (children.length == 0 && 
		(!settings.showAttributes ||
		 element.getElementAttributeCount() == 0));
    }

    public boolean handleMouseEvent(MouseEvent e) {
	if (notCollapsible())
	    return false;

	int eX = e.getX();
	int eY = e.getY();
	if (eX >= 2*settings.pad + settings.collapsedIcon.getWidth(null) ||
	    eY >= 2*settings.pad + settings.collapsedIcon.getHeight(null))
	    return false;

	int modifiers = e.getModifiers();
	if (modifiers == MouseEvent.BUTTON1_MASK) {
	    if (e.getID() == MouseEvent.MOUSE_CLICKED) {
		collapsed = !collapsed;
		refresh();
	    }
	    return true;
	} else if ((modifiers &
		    EventDispatcher.getPopupTriggerEventModifiers()) != 0) {
	    if (e.getID() == EventDispatcher.getPopupTriggerEventId()) 
		showMenu(menuSpec, eX, eY);
	    return true;
	} else {
	    return false;
	}
    }

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	if (command == null)
	    return;

	if (command.equals("collapseTree")) {
	    collapseTree(this, true);
	    refresh();
	} else if (command.equals("expandTree")) {
	    collapseTree(this, false);
	    refresh();
	}
    }

    private static void collapseTree(GadgetContainer container, 
				     boolean collapse) {
	Gadget[] gadgets = container.getChildren();
	for (int i = 0; i < gadgets.length; ++i) {
	    if (gadgets[i] instanceof GadgetContainer)
		collapseTree((GadgetContainer) gadgets[i], collapse);
	}

	if (container instanceof TreeView)
	    ((TreeView) container).setCollapsed(collapse);
    }
}
