/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit;

import java.awt.*;
import java.awt.event.*;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.gadget.GadgetContainer;
import com.xmlmind.xmledit.model.*;

public class WordView extends Gadget implements TextView, TextViewPart {
    public static class Settings extends CommonSettings {
	public Settings() {
	    background = Color.white;
	    foreground = Color.black;
	    font = new Font("SansSerif", Font.PLAIN, 10);
	    pad = 4;
	    borderColor = Color.gray;
	    borderRelief = RELIEF_SUNKEN;
	}
    }

    // -----------------------------------------------------------------------

    public static final Settings DEFAULT_SETTINGS = new Settings();

    private TextModel text;
    private Settings settings;
    private int[] lineLayout = new int[0];

    // -----------------------------------------------------------------------

    public WordView(TextModel text) {
	this(text, DEFAULT_SETTINGS);
    }

    public WordView(TextModel text, Settings settings) {
	this.text = text;
	this.settings = settings;
    }

    public SubelementModel getModel() {
	return text;
    }

    public void getGadgets(GadgetList gadgets, int viewIndex) {
	gadgets.appendGadget(this);
    }

    public TextModel getText() {
	return text;
    }

    public boolean acceptsControlChar(char c) {
	return false;
    }

    public SubelementView getView() {
	return this;
    }

    public TextView getTextView() {
	return this;
    }

    public Settings getSettings() {
	return settings;
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	relayoutLines(constrainedWidth, 0);
    }

    private int relayoutLines(int constrainedWidth, int from) {
	// Find the line where relayout begins ---

	int[] newLayout = new int[10];
	int newLayoutLength = 0;
	
	if (lineLayout.length <= 2) {
	    from = 0;
	} else {
	    int i = 2;
	    for (;;) {
		if (i == lineLayout.length || lineLayout[i] > from) {
		    from = lineLayout[i-2];
		    break;
		}

		newLayout = addLineLayout(newLayout, newLayoutLength, 
					  lineLayout[i-2], lineLayout[i-1]);
		newLayoutLength += 2;

		i += 2;
	    }
	}

	int fromLine = newLayoutLength/2;

	// Split text into lines ---
	
	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();

	FontMetrics metrics = getFontMetrics(settings.font);
	int[] charWidths = metrics.getWidths();
	int maxWidth = 
	    ((constrainedWidth > 0)? 
	     (constrainedWidth - 2*settings.pad) : Integer.MAX_VALUE); 
	if (maxWidth <= 0)
	    maxWidth = 1;

	int lineBegin = from;
	int lineWidth = 0;
	int wordFirst = -1;
	int wordWidth = 0;

	for (int i = from; i < textLength; ++i) {
	    char c = textChars[i];
	    int w = (c < 256)? charWidths[c] : metrics.charWidth(c);

	    switch (c) {
	    case ' ': case '\n': case '\r': case '\t':
		if (wordFirst >= 0) {
		    wordFirst = -1;
		    lineWidth += wordWidth;
		    wordWidth = 0;
		}

		if (lineWidth + w > maxWidth && lineWidth > 0) {
		    newLayout = addLineLayout(newLayout, newLayoutLength, 
					      lineBegin, lineWidth);
		    newLayoutLength += 2;

		    lineBegin = i+1;
		    lineWidth = 0;
		} else {
		    lineWidth += w;
		}
		break;

	    default:
		if (wordFirst < 0)
		    wordFirst = i;
		wordWidth += w;

		if (lineWidth + wordWidth > maxWidth && lineWidth > 0) {
		    newLayout = addLineLayout(newLayout, newLayoutLength, 
					      lineBegin, lineWidth);
		    newLayoutLength += 2;

		    lineBegin = wordFirst;
		    lineWidth = 0;
		}
	    }
	}

	// Add last line if any ---

	if (lineWidth + wordWidth > 0) {
	    newLayout= addLineLayout(newLayout, newLayoutLength, 
				     lineBegin, lineWidth + wordWidth);
	    newLayoutLength += 2;
	}

	// Save line layout ---

	lineLayout = new int[newLayoutLength];
	System.arraycopy(newLayout, 0, lineLayout, 0, newLayoutLength);

	// Recompute size ---

	int widest = 0;
	for (int i = 1; i < newLayoutLength; i += 2) {
	    if (newLayout[i] > widest)
		widest = newLayout[i];
	}

	width = 2*settings.pad + widest;
	height = 2*settings.pad + (metrics.getAscent()+metrics.getDescent());
	int lineCount = newLayoutLength / 2;
	if (lineCount > 1)
	    height += (lineCount-1)*metrics.getHeight();

	// This gives refresh() more latitude to just repaint() instead of
	// doing paintRelayout().
	if (constrainedWidth > 0 && width < constrainedWidth)
	    width = constrainedWidth;

	return fromLine;
    }

    private static final int[] addLineLayout(int[] newLayout, 
					     int newLayoutLength,
					     int lineBegin, int lineWidth) {
	if (newLayoutLength + 2 > newLayout.length) {
	    int[] nextLayout = new int[Math.max(newLayoutLength + 2, 
						2*newLayout.length)];
	    System.arraycopy(newLayout, 0, nextLayout, 0, newLayoutLength);
	    newLayout = nextLayout;
	}

	newLayout[newLayoutLength] = lineBegin;
	newLayout[newLayoutLength+1] = lineWidth;
	return newLayout;
    }

    public void paint(Graphics g) {
	if (settings.background != null) {
	    g.setColor(settings.background);
	    g.fillRect(0, 0, width, height);
	}

	if (settings.borderColor != null && settings.pad > 1) {
	    g.setColor(settings.borderColor);

	    switch (settings.borderRelief) {
	    case Settings.RELIEF_RAISED:
		g.draw3DRect(0, 0, width-1, height-1, true);
		break;
	    case Settings.RELIEF_SUNKEN:
		g.draw3DRect(0, 0, width-1, height-1, false);
		break;
	    default:
		g.drawRect(0, 0, width-1, height-1);
	    }
	}
    }

    public void paintText(Graphics g) {
	int lineCount = lineLayout.length / 2;
	if (lineCount == 0)
	    return;

	g.setColor(settings.foreground);
	g.setFont(settings.font);

	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();

	FontMetrics metrics = g.getFontMetrics(settings.font);
	int lineHeight = metrics.getHeight();
	int pad = settings.pad;

	Rectangle clip = g.getClipBounds();

	// Translate clip.y to origin at leading before first line.
	int clipY = clip.y - pad + metrics.getLeading();
	int firstLine = clipY / lineHeight;
	if (firstLine < 0)
	    firstLine = 0;
	else if (firstLine >= lineCount)
	    firstLine = lineCount-1;

	clipY += clip.height-1;
	int lastLine = clipY / lineHeight;
	if (lastLine < 0)
	    lastLine = 0;
	else if (lastLine >= lineCount)
	    lastLine = lineCount-1;

	int lineY = pad + metrics.getAscent() + firstLine*lineHeight;

	for (int i = firstLine; i <= lastLine; ++i) {
	    int firstChar = lineLayout[2*i];
	    int endChar;
	    if (i < lineCount-1)
		// -1: do not draw the space where the word break happens!
		endChar = lineLayout[2*(i+1)] - 1;
	    else
		endChar = textLength;

	    g.drawChars(textChars, firstChar, endChar-firstChar, pad, lineY);
	    lineY += lineHeight;
	}
    }

    public void print(Graphics g) {
	if (settings.borderColor != null && settings.pad > 1) {
	    g.setColor(settings.borderColor);
	    g.drawRect(0, 0, width-1, height-1);
	}
    }

    public int gadgetToModel(int viewX, int viewY) {
	int lineCount = lineLayout.length / 2;
	if (lineCount == 0)
	    return 0;

	FontMetrics metrics = getFontMetrics(settings.font);
	int lineHeight = metrics.getHeight();
	int pad = settings.pad;

	// Translate viewY to origin at leading before first line.
	viewY = viewY - pad + metrics.getLeading();
	int lineNum = viewY / lineHeight;
	if (lineNum < 0)
	    lineNum = 0;
	else if (lineNum >= lineCount)
	    lineNum = lineCount-1;

	viewX = viewX - pad;

	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();

	int firstChar = lineLayout[2*lineNum];
	int endChar;
	if (lineNum < lineCount-1)
	    // -1: end of line is before the space where the word break 
	    // happens.
	    endChar = lineLayout[2*(lineNum+1)] - 1;
	else
	    endChar = textLength;
	    
	int offset = firstChar;
	int count = endChar-firstChar;
	while (count > 1) {
	    int halfCount = count/2;
	    int w = metrics.charsWidth(textChars, offset, halfCount);
	    if (viewX < w) {
		count = halfCount;
	    } else {
		viewX -= w;
		offset += halfCount;
		count -= halfCount;
	    }
	}
	if (viewX >= metrics.charWidth(textChars[offset])/2)
	    ++offset;

	return offset;
    }

    public int modelToGadget(int offset, int count, 
			     RectangleList rectangles) {
	FontMetrics metrics = getFontMetrics(settings.font);
	int lineHeight = metrics.getHeight();
	int fontHeight = metrics.getAscent()+metrics.getDescent();
	int pad = settings.pad;

	int lineCount = lineLayout.length / 2;
	if (lineCount == 0) {
	    rectangles.appendRectangle(pad, pad, 0, fontHeight);
	    return 1;
	}

	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();

	if (offset < 0)
	    offset = 0;
	else if (offset > textLength)
	    offset = textLength;

	if (count < 0)
	    count = 0;
	else if (offset + count > textLength)
	    count = textLength - offset;

	int i;
	for (i = 0; i < lineCount; ++i) {
	    int endChar;
	    if (i < lineCount-1)
		// -1: the space where the word break happens is not drawn.
		endChar = lineLayout[2*(i+1)] - 1;
	    else
		endChar = textLength;

	    if (offset <= endChar) {
		if (count == 0) {
		    int firstChar = lineLayout[2*i];
		    int x1 = pad + metrics.charsWidth(textChars, firstChar, 
						      offset-firstChar);
		    rectangles.appendRectangle(x1, pad + i*lineHeight, 
					       0, fontHeight);
		    return 1;
		} else {
		    break;
		}
	    }
	}

	int appended = 0;

	for (; i < lineCount; ++i) {
	    int firstChar = lineLayout[2*i];
	    int nextFirstChar, endChar;
	    if (i < lineCount-1) {
		nextFirstChar = lineLayout[2*(i+1)];
		endChar = nextFirstChar - 1;
	    } else {
		nextFirstChar = endChar = textLength;
	    }

	    int x1 = ((offset > firstChar)? 
		      (pad + metrics.charsWidth(textChars, firstChar, 
						offset-firstChar)) : pad);
	    int x2 = x1 + metrics.charsWidth(textChars, offset, 
					     ((offset + count <= endChar)? 
					      count : endChar-offset));

	    rectangles.appendRectangle(x1, pad + i*lineHeight, 
				       x2-x1, fontHeight);
	    ++appended;

	    int prevOffset = offset;
	    offset = nextFirstChar;
	    count -= (nextFirstChar - prevOffset);
	    if (count <= 0)
		break;
	}

	return appended;
    }

    public void charInserted(char c, int offset) {
	if (lineLayout.length == 0) {
	    refreshFrom(0);
	    return;
	}

	int layoutIndex = layoutIndexOfChar(lineLayout, offset);

	if (layoutIndex != 0 && isXMLSpace(c)) {
	    // Inserting a space in the first word of a line means
	    // refreshing from previous line (if any).

	    int wordEnd = 
		findWordEnd(text.getTextChars(), text.getTextLength(),
			    lineLayout[layoutIndex]);

	    // Why '<=' and not '<' ? Splitting a word with a space means
	    // appending a space after first word.
	    if (offset <= wordEnd) {
		refreshFrom(lineLayout[layoutIndex-2]);
		return;
	    }
	}

	FontMetrics metrics = getFontMetrics(settings.font);
	int charWidth = metrics.charWidth(c);
	int lineWidth = lineLayout[layoutIndex+1];

	if (lineWidth + charWidth <= width - 2*settings.pad) {
	    // Update line width.
	    lineLayout[layoutIndex+1] += charWidth;

	    // Update char offsets for lines following this one.
	    for (int i = layoutIndex+2; i < lineLayout.length; i += 2)
		++lineLayout[i];

	    // Just repaint this line.
	    int fromY = settings.pad + ((layoutIndex/2)*metrics.getHeight());
	    int fontHeight = metrics.getAscent()+metrics.getDescent();
	    repaint(0, fromY, width, fontHeight);

	    return;
	}

	refreshFrom(offset);
    }

    private static final int layoutIndexOfChar(int[] lineLayout, int offset) {
	if (lineLayout.length <= 2) 
	    return 0;

	int i = 2;
	for (;;) {
	    if (i == lineLayout.length || lineLayout[i] > offset)
		return i-2;

	    i += 2;
	}
    }

    private static final boolean isXMLSpace(char c) {
	switch (c) {
	case ' ': case '\n': case '\r': case '\t':
	    return true;
	default:
	    return false;
	}
    }

    private static final int findWordEnd(char[] chars, int charCount, 
					 int from) {
	// Skip space before word.
	while (from < charCount && isXMLSpace(chars[from]))
	    ++from;

	// Skip word.
	while (from < charCount && !isXMLSpace(chars[from]))
	    ++from;

	return from;
    }

    public void charDeleted(char c, int offset) {
	int textLength = text.getTextLength();
	if (textLength == 0) {
	    // LineLayout need to be zeroed.
	    refreshFrom(0);
	    return;
	}

	int layoutIndex = layoutIndexOfChar(lineLayout, offset);
	char[] textChars = text.getTextChars();

	if (layoutIndex != 0) {
	    // Deleting a char in the first word of a line means refreshing
	    // from previous line (if any).

	    int wordEnd = findWordEnd(textChars, textLength,
				      lineLayout[layoutIndex]);

	    // Why '<=' and not '<' ? If the last char of the word was
	    // deleted, this char was at wordEnd.
	    if (offset <= wordEnd) {
		refreshFrom(lineLayout[layoutIndex-2]);
		return;
	    }
	}

	if (layoutIndex < lineLayout.length-2 && 
	    offset == lineLayout[layoutIndex+2]-1) {
	    // If the white space that separates 2 lines has been deleted, do
	    // not try to optimize.
	    refreshFrom(offset);
	    return;
	}

	FontMetrics metrics = getFontMetrics(settings.font);
	int charWidth = metrics.charWidth(c);
	int lineWidth = lineLayout[layoutIndex+1];
	int wordWidth;

	if (layoutIndex == lineLayout.length-2) {
	    // Last line always satisfies our conditions.
	    wordWidth = width;
	} else {
	    // Does the space or word+space that follows this line fit in this
	    // line?

	    int firstChar = lineLayout[layoutIndex+2];
	    int nextSpace = findNextSpace(textChars, textLength, firstChar);

	    wordWidth = metrics.charsWidth(
		textChars, firstChar, 
		((nextSpace >= textLength)? 
		 textLength-firstChar : nextSpace-firstChar+1));
	}

	if (lineWidth - charWidth + wordWidth >= width - 2*settings.pad) {
	    // Update line width.
	    lineLayout[layoutIndex+1] -= charWidth;

	    // Update char offsets for lines following this one.
	    for (int i = layoutIndex+2; i < lineLayout.length; i += 2)
		--lineLayout[i];

	    // Just repaint this line.
	    int fromY = settings.pad + ((layoutIndex/2)*metrics.getHeight());
	    int fontHeight = metrics.getAscent()+metrics.getDescent();
	    repaint(0, fromY, width, fontHeight);

	    return;
	}

	refreshFrom(offset);
    }

    private static final int findNextSpace(char[] chars, int charCount, 
					   int from) {
	// Skip word.
	while (from < charCount && !isXMLSpace(chars[from]))
	    ++from;

	return from;
    }

    private void refreshFrom(int from) {
	if (parent == null)
	    return;

	int prevWidth = width;
	int prevHeight = height;

	Dimension constraints = new Dimension();
	parent.getSizeConstraints(this, constraints);
	int fromLine = relayoutLines(constraints.width, from);

	if (width == prevWidth && height == prevHeight) {
	    int lineHeight = getFontMetrics(settings.font).getHeight();
	    int fromY = settings.pad + (fromLine*lineHeight);

	    repaint(0, fromY, width, height-fromY);
	} else {
	    int damagedSide, damagedLength;

	    if (width == prevWidth) {
		int lineHeight = getFontMetrics(settings.font).getHeight();
		int fromY = settings.pad + (fromLine*lineHeight);

		damagedSide = GadgetContainer.BOTTOM_SIDE;
		damagedLength = height-fromY;
	    } else {
		damagedSide = GadgetContainer.NO_SIDE;
		damagedLength = -1;
	    }

	    parent.paintRelayout(this, prevWidth, prevHeight, 
				 damagedSide, damagedLength);
	}
    }
}
