/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modelimpl;

import java.util.Enumeration;
import fr.pixware.util.LinearHashtable;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.Traversal;

public class Document implements DocumentModel {
    private ElementModel root;
    private DocumentTypeModel docType;
    private LinearHashtable properties = new LinearHashtable();

    private DocumentListener[] listenersNoData = new DocumentListener[0];
    private DocumentListener[] listeners = new DocumentListener[0];

    private TextEvent textEvent = new TextEvent();
    private ElementStructureEvent elementStructureEvent = 
        new ElementStructureEvent();
    private ElementAttributeEvent elementAttributeEvent = 
        new ElementAttributeEvent();
    private BatchEditingEvent batchEditingEvent = 
        new BatchEditingEvent();
    private int batchEditingNesting = 0;
    
    public ElementModel createElementModel(String typeName) {
	return new Element(typeName, listeners.length);
    }

    public TextModel createTextModel(String text) {
	return new Text(text, listeners.length);
    }

    public void initializeDocument(ElementModel root, 
				   DocumentTypeModel docType) {
	if (this.root != null)
	    throw new RuntimeException("already initialized");
	if (root == null)
	    throw new IllegalArgumentException("null root");

	this.root = root;
	((Element) root).setDocument(this);
	this.docType = docType;
    }

    public ElementModel getRootElement() {
	return root;
    }

    public DocumentTypeModel getDocumentType() {
	return docType;
    }

    public Object putDocumentProperty(String name, Object value) {
	return properties.put(name, value);
    }

    public Object removeDocumentProperty(String name) {
	return properties.remove(name);
    }

    public Object getDocumentProperty(String name) {
	return properties.get(name);
    }

    public int getDocumentPropertyCount() {
	return properties.size();
    }

    public Enumeration getDocumentProperties() {
	return new PropertyEnumeration(properties);
    }

    private static class PropertyEnumeration implements Enumeration {
	private LinearHashtable properties;
	private Enumeration keys;
	private Object[] nameValuePair = new Object[2];

	public PropertyEnumeration(LinearHashtable properties) {
	    this.properties = properties;
	    keys = properties.keys();
	}

	public boolean hasMoreElements() {
	    return keys.hasMoreElements();
	}

	public Object nextElement() {
	    String key = (String) keys.nextElement();
	    nameValuePair[0] = key;
	    nameValuePair[1] = properties.get(key);
	    return nameValuePair;
	}
    }

    public void beginBatchEditing() {
	if (batchEditingNesting == 0)
	    batchEditingNotification(this, BatchEditingEvent.BATCH_STARTED);

	++batchEditingNesting;
    }

    public void endBatchEditing() {
	if (batchEditingNesting > 0) {
	    --batchEditingNesting;

	    if (batchEditingNesting == 0)
		batchEditingNotification(this, 
					 BatchEditingEvent.BATCH_COMPLETED);
	}
    }

    public void addDocumentListener(DocumentListener listener, 
				    boolean hasData) {
	if (root == null)
	    throw new RuntimeException("not initialized");

	if (documentListenerIndex(listener) >= 0 ||
	    findListenerNoData(listener) >= 0)
	    return;

	if (hasData) {
	    DocumentListener[] newListeners = 
		new DocumentListener[listeners.length + 1];
	    if (listeners.length > 0)
		System.arraycopy(listeners, 0, 
				 newListeners, 0, listeners.length);
	    newListeners[listeners.length] = listener;
	    listeners = newListeners;

	    Traversal.Handler handler = new Traversal.HandlerBase() {
		public Object processText(TextModel text) {
		    ((Text) text).listenerAdded();
		    return null;
		}

		public Object enterElement(ElementModel element) {
		    ((Element) element).listenerAdded();
		    return null;
		}
	    };
	    Traversal.traverse(root, handler);
	} else {
	    DocumentListener[] newListenersNoData = 
		new DocumentListener[listenersNoData.length + 1];
	    if (listenersNoData.length > 0)
		System.arraycopy(listenersNoData, 0, newListenersNoData, 0, 
				 listenersNoData.length);
	    newListenersNoData[listenersNoData.length] = listener;
	    listenersNoData = newListenersNoData;
	}
    }

    public void removeDocumentListener(DocumentListener listener) {
	final int listenerIndex = documentListenerIndex(listener);

	if (listenerIndex >= 0) {
	    DocumentListener[] newListeners = 
		new DocumentListener[listeners.length - 1];

	    if (listenerIndex > 0)
		System.arraycopy(listeners, 0, newListeners, 0, listenerIndex);

	    int first = listenerIndex + 1;
	    int tail = listeners.length - first;
	    if (tail > 0) 
		System.arraycopy(listeners, first, 
				 newListeners, listenerIndex, tail);

	    listeners = newListeners;

	    Traversal.Handler handler = new Traversal.HandlerBase() {
		public Object processText(TextModel text) {
		    ((Text) text).listenerRemoved(listenerIndex);
		    return null;
		}

		public Object enterElement(ElementModel element) {
		    ((Element) element).listenerRemoved(listenerIndex);
		    return null;
		}
	    };
	    Traversal.traverse(root, handler);
	} else {
	    int index = findListenerNoData(listener);

	    if (index >= 0) {
		DocumentListener[] newListenersNoData = 
		    new DocumentListener[listenersNoData.length - 1];

		if (index > 0)
		    System.arraycopy(listenersNoData, 0, 
				     newListenersNoData, 0, index);

		int first = index + 1;
		int tail = listenersNoData.length - first;
		if (tail > 0) 
		    System.arraycopy(listenersNoData, first, 
				     newListenersNoData, index, tail);

		listenersNoData = newListenersNoData;
	    }
	}
    }

    public int documentListenerIndex(DocumentListener listener) {
	for (int i = 0; i < listeners.length; ++i) {
	    if (listeners[i] == listener)
		return i;
	}
	return -1;
    }

    public DocumentListener getDocumentListener(int index, boolean hasData) {
	return (hasData? listeners[index] : listenersNoData[index]);
    }

    private int findListenerNoData(DocumentListener listener) {
	for (int i = 0; i < listenersNoData.length; ++i) {
	    if (listenersNoData[i] == listener)
		return i;
	}
	return -1;
    }

    public void notifyDocumentListeners(DocumentEvent event) {
	// Call listeners with data before listeners without data because the
	// listeners without data may depend on the fact that the listeners
	// with data (views) have already updated their views.

	int id = event.getId();
	int i;

	if (id <= TextEvent.LAST) {
	    for (i = 0; i < listeners.length; ++i) 
		listeners[i].textChanged((TextEvent) event, i);

	    for (i = 0; i < listenersNoData.length; ++i) 
		listenersNoData[i].textChanged((TextEvent) event, -1);
	} else if (id <= ElementStructureEvent.LAST) {
	    for (i = 0; i < listeners.length; ++i) 
		listeners[i].elementStructureChanged(
		    (ElementStructureEvent) event, i);

	    for (i = 0; i < listenersNoData.length; ++i) 
		listenersNoData[i].elementStructureChanged(
		    (ElementStructureEvent) event, -1);
	} else if (id <= ElementAttributeEvent.LAST) {
	    for (i = 0; i < listeners.length; ++i) 
		listeners[i].elementAttributeChanged(
		    (ElementAttributeEvent) event, i);

	    for (i = 0; i < listenersNoData.length; ++i) 
		listenersNoData[i].elementAttributeChanged(
		    (ElementAttributeEvent) event, -1);
	} else {
	    switch (id) {
	    case BatchEditingEvent.BATCH_STARTED:
		for (i = 0; i < listeners.length; ++i) 
		    listeners[i].batchEditingStarted(
			(BatchEditingEvent) event, i);

		for (i = 0; i < listenersNoData.length; ++i) 
		    listenersNoData[i].batchEditingStarted(
			(BatchEditingEvent) event, -1);
		break;

	    case BatchEditingEvent.BATCH_COMPLETED:
		for (i = 0; i < listeners.length; ++i) 
		    listeners[i].batchEditingCompleted(
			(BatchEditingEvent) event, i);

		for (i = 0; i < listenersNoData.length; ++i) 
		    listenersNoData[i].batchEditingCompleted(
			(BatchEditingEvent) event, -1);
		break;

	    default:
		for (i = 0; i < listeners.length; ++i) 
		    listeners[i].documentEventHappened(event, i);

		for (i = 0; i < listenersNoData.length; ++i) 
		    listenersNoData[i].documentEventHappened(event, -1);
	    }
	}
    }

    /*package*/ void textNotification(Object source, int id, int offset, 
				      String oldString, String newString) {
	textEvent.initialize(source, id, offset, oldString, newString);
	notifyDocumentListeners(textEvent);
    }

    /*package*/ void elementStructureNotification(Object source, int id, 
						  SubelementModel refChild, 
						  SubelementModel oldChild, 
						  SubelementModel newChild) {
	elementStructureEvent.initialize(source, id, 
					 refChild, oldChild, newChild);
	notifyDocumentListeners(elementStructureEvent);
    }

    /*package*/ void elementAttributeNotification(Object source, int id, 
						  String name, 
						  String oldValue, 
						  String newValue) {
	elementAttributeEvent.initialize(source, id, name, oldValue, newValue);
	notifyDocumentListeners(elementAttributeEvent);
    }

    /*package*/ void batchEditingNotification(Object source, int id) {
	batchEditingEvent.initialize(source, id);
	notifyDocumentListeners(batchEditingEvent);
    }
}

