/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modelimpl;

import java.util.Enumeration;
import fr.pixware.util.LinearHashtable;
import com.xmlmind.xmledit.model.*;

public class Element extends Subelement implements ElementModel {
    private String typeName;
    private LinearHashtable attributes = null;
    private Document document;
    private Subelement first;

    public Element(String typeName, int listenerCount) {
	super(ELEMENT, listenerCount);
	this.typeName = typeName;
    }

    /*package*/ void setDocument(Document document) {
	this.document = document;

	Subelement sub = first;
	while (sub != null) {
	    if (sub instanceof Element)
		((Element) sub).setDocument(document);

	    sub = sub.getNext();
	}
    }

    public DocumentModel getDocument() {
	return document;
    }

    public String getElementTypeName() {
	return typeName;
    }

    public void appendSubelement(SubelementModel sub) {
	insertSubelement(null, sub);
    }

    public void insertSubelement(SubelementModel referenceChild, 
				 SubelementModel sub) {
	Subelement added = (Subelement) sub;
	if (!added.isDetached())
	    throw new IllegalArgumentException("sub already attached");

	Subelement next, previous;
	if (referenceChild == null) {
	    next = null;
	    previous = (Subelement) getLastSubelement();
	} else {
	    if (referenceChild.getParentElement() != this)
		throw new IllegalArgumentException(
		    "referenceChild not a child");

	    next = (Subelement) referenceChild;
	    previous = next.getPrevious();
	}

	if (previous != null)
	    previous.setNext(added);

	if (next != null)
	    next.setPrevious(added);

	added.setAttached(this, previous, next);

	if (first == next)
	    first = added;

	if (document != null) {
	    // A detached element always has a null document.
	    if (added instanceof Element)
		((Element) added).setDocument(document);

	    document.elementStructureNotification(
		this, ElementStructureEvent.CHILD_ADDED, next, null, added);
	}
    }

    public void removeSubelement(SubelementModel child) {
	replaceSubelement(child, null);
    }

    public void replaceSubelement(SubelementModel child, 
				  SubelementModel sub) {
	if (child.getParentElement() != this)
	    throw new IllegalArgumentException("child not a child");

	Subelement removed = (Subelement) child;
	Subelement previous = removed.getPrevious();
	Subelement next = removed.getNext();

	if (sub == null) {
	    removed.setDetached();
	    if ((removed instanceof Element) && 
		((Element) removed).document != null)
		((Element) removed).setDocument(null);

	    if (previous != null)
		previous.setNext(next);

	    if (next != null) 
		next.setPrevious(previous);

	    if (removed == first)
		first = next;

	    if (document != null)
		document.elementStructureNotification(
		    this, ElementStructureEvent.CHILD_REMOVED, 
		    next, removed, null);
	} else {
	    Subelement added = (Subelement) sub;
	    if (!added.isDetached())
		throw new IllegalArgumentException("sub already attached");

	    removed.setDetached();
	    if ((removed instanceof Element) && 
		((Element) removed).document != null)
		((Element) removed).setDocument(null);

	    if (previous != null)
		previous.setNext(added);

	    if (next != null) 
		next.setPrevious(added);

	    added.setAttached(this, previous, next);

	    if (removed == first)
		first = added;

	    if (document != null) {
		// A detached element always has a null document.
		if (added instanceof Element)
		    ((Element) added).setDocument(document);

		document.elementStructureNotification(
		    this, ElementStructureEvent.CHILD_REPLACED, 
		    next, removed, added);
	    }
	}
    }

    public SubelementModel getFirstSubelement() {
	return first;
    }

    public SubelementModel getLastSubelement() {
	SubelementModel last = null;

	SubelementModel child = first;
	while (child != null) {
	    last = child;
	    child = child.getNextSubelement();
	}
	
	return last;
    }

    public int getSubelementCount() {
	int count = 0;

	SubelementModel child = first;
	while (child != null) {
	    ++count;
	    child = child.getNextSubelement();
	}

	return count;
    }

    public String putElementAttribute(String name, String value) {
	if (attributes == null)
	    attributes = new LinearHashtable();

	String oldValue = (String) attributes.put(name, value);

	if (oldValue == null) {
	    if (document != null)
		document.elementAttributeNotification(
		    this, ElementAttributeEvent.ATTRIBUTE_ADDED, 
		    name, null, value);
	} else if (!value.equals(oldValue)) {
	    if (document != null)
		document.elementAttributeNotification(
		    this, ElementAttributeEvent.ATTRIBUTE_CHANGED, 
		    name, oldValue, value);
	}

	return oldValue;
    }

    public String removeElementAttribute(String name) {
	if (attributes == null)
	    return null;

	String oldValue = (String) attributes.remove(name);

	if (oldValue != null && document != null)
	    document.elementAttributeNotification(
		this, ElementAttributeEvent.ATTRIBUTE_REMOVED, 
		name, oldValue, null);

	return oldValue;
    }

    public String getElementAttribute(String name) {
	return (attributes == null)? null : (String) attributes.get(name);
    }

    public int getElementAttributeCount() {
	return (attributes == null)? 0 :attributes.size();
    }

    public Enumeration getElementAttributes() {
	if (attributes == null)
	    attributes = new LinearHashtable();

	return new AttributeEnumeration(attributes);
    }

    private static class AttributeEnumeration implements Enumeration {
	private LinearHashtable attributes;
	private Enumeration keys;
	private String[] nameValuePair = new String[2];

	public AttributeEnumeration(LinearHashtable attributes) {
	    this.attributes = attributes;
	    keys = attributes.keys();
	}

	public boolean hasMoreElements() {
	    return keys.hasMoreElements();
	}

	public Object nextElement() {
	    String key = (String) keys.nextElement();
	    nameValuePair[0] = key;
	    nameValuePair[1] = (String) attributes.get(key);
	    return nameValuePair;
	}
    }
}

