/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modelimpl;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Stack;
import java.util.EmptyStackException;
import java.util.Vector;
import java.util.StringTokenizer;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import com.jclark.xml.parse.*;
import com.jclark.xml.parse.io.*;
import fr.pixware.util.WorkInProgressObserver;
import fr.pixware.util.WorkCanceledException;
import com.xmlmind.xmledit.modeltool.StyleSheetInfo;
import com.xmlmind.xmledit.modeltool.ElementUtil;
import com.xmlmind.xmledit.model.*;

public class XPDocumentLoader extends ApplicationImpl {
    private ParserImpl parser;
    private Hashtable sharedAttributeDefinitions = new Hashtable();
    private WorkInProgressObserver observer = null;
    private boolean guessElementContent = false;

    private DocumentModel document;
    private static final int CHAR_INITIAL_CAPACITY = 1024;
    private char[] chars = new char[CHAR_INITIAL_CAPACITY];
    private int charCount = 0;
    private char[] normalizedChars = new char[CHAR_INITIAL_CAPACITY];
    private Stack stack = new Stack();
    private int notifyObserverCountdown = 0;

    private DocumentTypeModel documentType;
    private boolean insideDocType;
    private int openEntityCount;
    private StringBuffer internalSubset = new StringBuffer();
    private String lookahead;
    private Vector styleSheets = new Vector();

    // -----------------------------------------------------------------------
  
    public XPDocumentLoader() {
	parser = new ParserImpl();
	parser.setApplication(this);
    }
    
    public void load(String fileName, DocumentModel document) 
	throws IOException {
	load(EntityManagerImpl.openFile(fileName), document);
    }

    public void load(URL url, DocumentModel document) 
	throws IOException {
	load(new OpenEntity(url.openStream(), url.toExternalForm(), url), 
	     document);
    }

    public void load(OpenEntity entity, DocumentModel document) 
	throws IOException {
	this.document = document;

	sharedAttributeDefinitions.clear();
	notifyObserverCountdown = 0;

	try {
	    parser.parseDocument(entity);

	    URL baseURL = entity.getBase();
	    if (baseURL != null)
		document.putDocumentProperty(DocumentModel.URL_PROPERTY, 
					     baseURL);

	    int infoCount = styleSheets.size();
	    if (infoCount > 0) {
		StyleSheetInfo[] list = new StyleSheetInfo[infoCount];
		styleSheets.copyInto(list);
		document.putDocumentProperty(StyleSheetInfo.PROPERTY, list);
	    }
	} finally {
	    this.document = null;
	    if  (chars.length > CHAR_INITIAL_CAPACITY)
		chars = new char[CHAR_INITIAL_CAPACITY];
	    charCount = 0;
	    if  (normalizedChars.length > CHAR_INITIAL_CAPACITY)
		normalizedChars = new char[CHAR_INITIAL_CAPACITY];
	    popAll();

	    documentType = null;
	    insideDocType = false;
	    openEntityCount = 0;
	    internalSubset = new StringBuffer();

	    styleSheets = new Vector();

	    if (observer != null)
		observer.workInProgress(this, false);
	}
    }

    public void setObserver(WorkInProgressObserver observer) {
	this.observer = observer;
    }

    public WorkInProgressObserver getObserver() {
	return observer;
    }

    public void setGuessingElementContent(boolean guess) {
	guessElementContent = guess;
    }

    public boolean isGuessingElementContent() {
	return guessElementContent;
    }

    // -----------------------------------------------------------------------

    public void startElement(StartElementEvent event) throws IOException {
	if (observer != null && (notifyObserverCountdown++ % 10) == 0) {
	    if (!observer.workInProgress(this, true))
		throw new WorkCanceledException(this);
	}

	if (charCount > 0) {
	    appendText();
	    charCount = 0;
	}

	ElementModel pushed = document.createElementModel(event.getName());

	int specifiedCount = event.getAttributeSpecifiedCount();
	for (int i = 0; i < specifiedCount; ++i) 
	    pushed.putElementAttribute(event.getAttributeName(i), 
				       event.getAttributeValue(i));

	Stacked top = top();
	if (top == null) {
	    // Pushed is the root element.
	    document.initializeDocument(pushed, documentType);
	} else {
	    top.element.appendSubelement(pushed);
	}

	push(pushed);
    }

    public void endElement(EndElementEvent event) throws IOException {
	Stacked top = top();

	if (charCount > 0) {
	    appendText();
	    charCount = 0;
	} else {
	    // If top is an empty mix, insert a Text in it ---

	    if (top.canContainText && top.element.getSubelementCount() == 0 &&
		ElementUtil.hasMixedContent(top.element, documentType)) 
		top.element.appendSubelement(document.createTextModel(""));
	}

	if (!top.hasPreserveSpace) {
	    if (documentType == null) {
		if (guessElementContent)
		    trimStructure(top.element);
	    } else {
		// If a text container is contained in a structure, the text
		// container is trimmed.

		if (top.canContainText)
		    trimTextContainer(top.element, documentType);
	    }
	}
	    
	pop();
    }

    private static void trimStructure(ElementModel element) {
	// Do not trim if element only contains white space.
	boolean containsElements = false;
	boolean containsDiscardableText = false;

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		TextModel text = (TextModel) sub;
		int charCount = text.getTextLength();

		if (charCount > 0) {
		    char[] chars = text.getTextChars();

		    for (int i = 0; i < charCount; ++i) {
			switch (chars[i]) {
			case ' ': case '\n': case '\r': case '\t':
			    break;
			default:
			    // Really contains text: do not trim.
			    return;
			}
		    }
		}

		containsDiscardableText = true;
	    } else {
		containsElements = true;
	    }

	    sub = sub.getNextSubelement();
	}

	if (containsElements && containsDiscardableText) {
	    sub = element.getFirstSubelement();
	    while (sub != null) {
	       SubelementModel next = sub.getNextSubelement();

	       if (sub.getSubelementType() == SubelementModel.TEXT)
		   element.removeSubelement(sub);

	       sub = next;
	    }
	}
    }

    private static void trimTextContainer(ElementModel element, 
					  DocumentTypeModel documentType) {
	ElementModel parent = element.getParentElement();
	if (parent == null)
	    return;

	ElementTypeModel parentType =
	    documentType.getElementType(parent.getElementTypeName());
	if (parentType != null &&
	    parentType.getElementContentType() == ElementTypeModel.ELEMENT) {
	    SubelementModel sub = element.getFirstSubelement();
	    TextModel text;

	    if (sub != null && 
		sub.getSubelementType() == SubelementModel.TEXT) {
		text = (TextModel) sub;

		// Remember that consecutive white spaces have been
		// compressed. Suffice to remove the first white space if any.
		if (text.getTextLength() > 0 && text.getTextChar(0) == ' ') {
		    text.deleteText(0, 1);

		    if (text.getTextLength() == 0 &&
			element.getSubelementCount() > 1)
			element.removeSubelement(text);
		}
	    }

	    sub = element.getLastSubelement();
	    if (sub != null && 
		sub.getSubelementType() == SubelementModel.TEXT) {
		text = (TextModel) sub;
		int last = text.getTextLength()-1;
		if (last >= 0 && text.getTextChar(last) == ' ') {
		    text.deleteText(last, 1);

		    if (text.getTextLength() == 0 &&
			element.getSubelementCount() > 1)
			element.removeSubelement(text);
		}
	    }
	}
    }

    public void characterData(CharacterDataEvent event) throws IOException {
	if (top().canContainText) {
	    int added = event.getLengthMax();
	    if (charCount + added > chars.length) {
		int newLength = Math.max(charCount + added, 2*chars.length);
		char[] newChars = new char[newLength];
		if (charCount > 0)
		    System.arraycopy(chars, 0, newChars, 0, charCount);
		chars = newChars;
	    }

	    charCount += event.copyChars(chars, charCount);
	}
    }

    private void appendText() {
	// [1.] The parser should normalize line breaks ('\r', "\r\n" converted
        //      to a single '\n').
	//
	// [2.] The application is allowed to have its own white space
	//       handling policy, unless xml:space="preserve" is specified.
        //
	//       XPDocumentLoader converts all space chars to ' ' and 
        //       compresses consecutive white spaces.
        //
        //       If a text container is contained in a structure, the 
	//       text container is trimmed.

	Stacked top = top();

	if (top.hasPreserveSpace) {
	    top.element.appendSubelement(
		document.createTextModel(new String(chars, 0, charCount)));
	} else {
	    if (normalizedChars.length < charCount)
		normalizedChars = 
		    new char[Math.max(charCount, 2*normalizedChars.length)];

	    // Convert white space chars to space and compress consecutive
	    // spaces.

	    int normalCharCount = 0;
	    char prevChar = '?';
	    for (int i = 0; i < charCount; ++i) {
		char c = chars[i];

		switch (c) {
		case '\t': case '\r': case '\n':
		    c = ' ';
		    break;
		}

		if (c == ' ') {
		    if (prevChar != ' ') {
			normalizedChars[normalCharCount++] = c;
			prevChar = c;
		    }
		} else {
		    normalizedChars[normalCharCount++] = c;
		    prevChar = c;
		}
	    }
	    
	    top.element.appendSubelement(document.createTextModel(
		new String(normalizedChars, 0, normalCharCount)));
	}
    }

    private void push(ElementModel pushed) {
	stack.push(
	    new Stacked(pushed, 
			ElementUtil.isTextContainer(pushed, documentType),
			ElementUtil.hasPreserveSpace(pushed, documentType)));
    }

    private Stacked top() {
	return (stack.empty()? null : (Stacked) stack.peek());
    }

    private void pop() {
	stack.pop();
    }

    private void popAll() {
	while (!stack.empty())
	    stack.pop();
    }

    private static class Stacked {
	public ElementModel element;
	public boolean canContainText;
	public boolean hasPreserveSpace;

	public Stacked(ElementModel element, boolean canContainText, 
		       boolean hasPreserveSpace) {
	    this.element = element;
	    this.canContainText = canContainText;
	    this.hasPreserveSpace = hasPreserveSpace;
	}
    }

    // -----------------------------------------------------------------------
  
    public void startDocumentTypeDeclaration(
	                             StartDocumentTypeDeclarationEvent event) 
	throws IOException {
	insideDocType = true;
    }

    public void endDocumentTypeDeclaration(
	                               EndDocumentTypeDeclarationEvent event) 
	throws IOException {
	insideDocType = false;

	DTD dtd = event.getDTD();
	Entity entity = dtd.getEntity(DTD.PARAMETER_ENTITY,
				      DTD.EXTERNAL_SUBSET_NAME);

	String publicId;
	String systemId;
	URL baseURL;
	if (entity != null) {
	    publicId = entity.getPublicId();
	    systemId = entity.getSystemId();
	    baseURL = entity.getBase();
	} else {
	    publicId = null;
	    systemId = null;
	    baseURL = null;
	}

	String subset = internalSubset.toString();
	internalSubset = new StringBuffer();

	documentType = createDocumentType(publicId, systemId, 
					  (subset.length()==0)? null : subset,
					  baseURL, dtd);
    }

    public void startEntityReference(StartEntityReferenceEvent event) 
	throws IOException {
	if (insideDocType) {
	    if (openEntityCount == 0 && !event.getName().equals("#DOCTYPE")) {
		internalSubset.append('%');
		internalSubset.append(event.getName());
		internalSubset.append(";\n");
	    }
	    ++openEntityCount;
	}
    }

    public void endEntityReference(EndEntityReferenceEvent event) 
	throws IOException {
	if (insideDocType)
	    --openEntityCount;
    }
  
    public void processingInstruction(ProcessingInstructionEvent event) 
	throws IOException {
	if (insideDocType && openEntityCount == 0) {
	    internalSubset.append("<?");
	    internalSubset.append(event.getName());
	    internalSubset.append(' ');
	    internalSubset.append(event.getInstruction());
	    internalSubset.append("?>\n");
	}

	if (!event.getName().equals("xml-stylesheet"))
	    return;

	StyleSheetInfo info = new StyleSheetInfo();
	if (!parseStyleSheetInfo(event.getInstruction(), info))
	    return;

	if (info.href != null && info.href.length() > 0 && 
	    info.type != null && info.type.length() > 0) {
	    URL base = event.getLocation().getEntityBase();
	    if (base != null) {
		try {
		    info.url = (new URL(base, info.href)).toExternalForm();
		} catch (MalformedURLException ignored) {
		    info.url = null;
		}
	    }
	    if (info.url != null)
		styleSheets.addElement(info);
	}
    }

    private static boolean parseStyleSheetInfo(String instruction, 
					       StyleSheetInfo info) {
	// Embedded style sheets are not supported.
	info.content = null;

	String[] split = splitStyleSheetInfo(instruction);
	if ((split.length % 3) != 0)
	    return false;

	for (int i = 0; i < split.length; i += 3) {
	    String name = split[i].toLowerCase();
	    String op = split[i+1];
	    String value = split[i+2];

	    if (!op.equals("="))
		return false;

	    if (value.length() < 2)
		return false;
	    char c1 = value.charAt(0);
	    if (c1 != '\'' && c1 != '\"')
		return false;
	    int last = value.length()-1;
	    char c2 = value.charAt(last);
	    if (c2 != c1)
		return false;
	    value = value.substring(1, last);

	    if (name.equals("href"))
		info.href = value;
	    else if (name.equals("type"))
		info.type = value;
	    else if (name.equals("title"))
		info.title = value;
	    else if (name.equals("media"))
		info.media = value;
	    else if (name.equals("charset"))
		info.charset = value;
	    else if (name.equals("alternate")) {
		value = value.toLowerCase();
		if (value.equals("yes"))
		    info.alternate = true;
		else if (value.equals("no"))
		    info.alternate = false;
		else
		    return false;
	    }
	}

	return true;
    }

    private static String[] splitStyleSheetInfo(String instruction) {
	Vector split = new Vector();
	StringBuffer part = new StringBuffer();
	int charCount = instruction.length();
	char quote = 0;

	for (int i = 0; i < charCount; ++i) {
	    char c = instruction.charAt(i);

	    switch (c) {
	    case ' ': case '\n': case '\r': case '\t':
		if (quote != 0) {
		    part.append(c);
		} else {
		    if (part.length() > 0) {
			split.addElement(part.toString());
			part = new StringBuffer();
		    }
		}
		break;

	    case '=':
		if (part.length() > 0) {
		    split.addElement(part.toString());
		    part = new StringBuffer();
		}
		split.addElement("=");
		break;

	    case '\'': case '\"':
		if (quote == 0) {
		    part.append(c);
		    quote = c;
		} else if (quote == c) {
		    part.append(c);
		    split.addElement(part.toString());
		    part = new StringBuffer();
		    quote = 0;
		} else {
		    // The ``other'' quote char.
		    part.append(c);
		}
		break;

	    default:
		part.append(c);
	    }
	}
	if (part.length() > 0) 
	    split.addElement(part.toString());
	
	String[] list = new String[split.size()];
	split.copyInto(list);
	return list;
    }

    public void comment(CommentEvent event) throws IOException {
	if (insideDocType && openEntityCount == 0) {
	    internalSubset.append("<!--");
	    internalSubset.append(event.getComment());
	    internalSubset.append("-->\n");
	}
    }

    public void markupDeclaration(MarkupDeclarationEvent event) 
	throws IOException {
	if (observer != null && (notifyObserverCountdown++ % 10) == 0) {
	    if (!observer.workInProgress(this, true))
		throw new WorkCanceledException(this);
	}

	if (openEntityCount != 0)
	    return;

	String name = event.getName();
	DTD dtd = event.getDTD();
	switch (event.getType()) {
	case MarkupDeclarationEvent.ELEMENT:
	    appendElementDecl(dtd, name);
	    break;
	case MarkupDeclarationEvent.GENERAL_ENTITY:
	    appendEntityDecl(dtd, DTD.GENERAL_ENTITY, name);
	    break;
	case MarkupDeclarationEvent.PARAMETER_ENTITY:
	    appendEntityDecl(dtd, DTD.PARAMETER_ENTITY, name);
	    break;
	case MarkupDeclarationEvent.NOTATION:
	    appendEntityDecl(dtd, DTD.NOTATION, name);
	    break;
	case MarkupDeclarationEvent.ATTRIBUTE:
	    appendAttributeDef(dtd, name, event.getAttributeName());
	    break;
	}
    }

    private void appendEntityDecl(DTD dtd, byte type, String name) {
	switch (type) {
	case DTD.GENERAL_ENTITY:
	    internalSubset.append("<!ENTITY ");
	    break;
	case DTD.PARAMETER_ENTITY:
	    internalSubset.append("<!ENTITY % ");
	    break;
	case DTD.NOTATION:
	    internalSubset.append("<!NOTATION ");
	    break;
	}

	internalSubset.append(name);
	internalSubset.append(' ');

	Entity entity = dtd.getEntity(type, name);
	String text = entity.getReplacementText();
	if (text == null) {
	    appendExternalId(entity);
	    if (type == DTD.GENERAL_ENTITY) {
		name = entity.getNotationName();
		if (name != null) {
		    internalSubset.append(" NDATA ");
		    internalSubset.append(name);
		}
	    }
	} else {
	    internalSubset.append('\"');

	    int length = text.length();
	    for (int i = 0; i < length; ++i) {
		char c = text.charAt(i);

		switch (c) {
		case '&': case '\"': case '%': case '\r':
		    internalSubset.append("&#");
		    internalSubset.append((int) c);
		    internalSubset.append(';');
		    break;
		default:
		    internalSubset.append(c);
		    break;
		}
	    }

	    internalSubset.append('\"');
	}

	internalSubset.append(">\n");
    }      

    private void appendExternalId(Entity entity) {
	String publicId = entity.getPublicId();
	String systemId = entity.getSystemId();

	if (publicId != null) {
	    internalSubset.append("PUBLIC \"");
	    internalSubset.append(publicId);
	    internalSubset.append('\"');
	}

	if (systemId != null) {
	    if (publicId == null)
		internalSubset.append("SYSTEM");
	    internalSubset.append(' ');
	    char quote = (systemId.indexOf('\"') < 0)? '\"' : '\'';
	    internalSubset.append(quote);
	    internalSubset.append(systemId);
	    internalSubset.append(quote);
	}
    }

    private void appendElementDecl(DTD dtd, String elementName) {
	internalSubset.append("<!ELEMENT ");
	internalSubset.append(elementName);
	internalSubset.append(' ');
	internalSubset.append(
	    dtd.getElementType(elementName).getContentSpec());
	internalSubset.append(">\n");
    }

    private void appendAttributeDef(DTD dtd, String elementName, 
				    String attributeName) {
	com.jclark.xml.parse.AttributeDefinition def = 
	    dtd.getElementType(elementName).getAttributeDefinition(
		attributeName);

	internalSubset.append("<!ATTLIST ");
	internalSubset.append(elementName);
	internalSubset.append(' ');
	internalSubset.append(attributeName);

	String type;
	switch (def.getType()) {
	case com.jclark.xml.parse.AttributeDefinition.CDATA:
	    type = "CDATA";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.ID:
	    type = "ID";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.IDREF:
	    type = "IDREF";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.IDREFS:
	    type = "IDREFS";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.ENTITY:
	    type = "ENTITY";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.ENTITIES:
	    type = "ENTITIES";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.NMTOKEN:
	    type = "NMTOKEN";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.NMTOKENS:
	    type = "NMTOKENS";
	    break;
	case com.jclark.xml.parse.AttributeDefinition.NOTATION:
	    type = "NOTATION";
	    break;
	default:
	    type = null;
	}
	if (type != null)
	    internalSubset.append(" " + type);

	Enumeration enumValues = def.allowedValues();
	if (enumValues != null) {
	    internalSubset.append(" (");
	    internalSubset.append((String) enumValues.nextElement());

	    while (enumValues.hasMoreElements()) {
		internalSubset.append('|');
		internalSubset.append((String) enumValues.nextElement());
	    }

	    internalSubset.append(')');
	}

	String defaultValue = def.getDefaultUnnormalizedValue();
	boolean required = def.isRequired();
	if (defaultValue == null) {
	    internalSubset.append(required? " #REQUIRED" : " #IMPLIED");
	} else {
	    if (required)
		internalSubset.append(" #FIXED ");
	    else
		internalSubset.append(' ');

	    char quote = (defaultValue.indexOf('\"') < 0)? '\"' : '\'';
	    internalSubset.append(quote);
	    internalSubset.append(defaultValue);
	    internalSubset.append(quote);
	}

	internalSubset.append(">\n");
    }

    // -----------------------------------------------------------------------
  
    private DocumentType createDocumentType(String publicId, String systemId, 
					    String internalSubset, URL baseURL,
					    DTD dtd) {
	String url = null;
	if (systemId != null) {
	    try {
		url = (new URL(baseURL, systemId)).toString();
	    } catch (MalformedURLException e) {
		url = null;
	    }
	}

	DocumentType docType = 
	    new DocumentType(publicId, systemId, internalSubset, url);
	addNotations(dtd, docType);
	addUnparsedEntities(dtd, docType);
	addElementTypes(dtd, docType);
	initElementContent(dtd, docType);

	return docType;
    }

    private void addNotations(DTD dtd, DocumentType docType) {
	Enumeration notations = dtd.entityNames(DTD.NOTATION);

	while (notations.hasMoreElements()) {
	    String notationName = (String) notations.nextElement();
	    Entity notation = dtd.getEntity(DTD.NOTATION, notationName);

	    String systemId = notation.getSystemId();
	    String url = null;
	    if (systemId != null) {
		try {
		    url = (new URL(notation.getBase(), systemId)).toString();
		} catch (MalformedURLException e) {
		    url = null;
		}
	    }

	    docType.putNotation(
		new Notation(notationName, notation.getPublicId(),
			     systemId, url));
	}
    }

    private void addUnparsedEntities(DTD dtd, DocumentType docType) {
	Enumeration entities = dtd.entityNames(DTD.GENERAL_ENTITY);

	while (entities.hasMoreElements()) {
	    String entityName = (String) entities.nextElement();
	    Entity entity = dtd.getEntity(DTD.GENERAL_ENTITY, entityName);
	    String notationName = entity.getNotationName();

	    if (notationName != null) {
		String systemId = entity.getSystemId();
		String url;
		try {
		    url = (new URL(entity.getBase(), systemId)).toString();
		} catch (MalformedURLException e) {
		    url = null;
		}

		docType.putUnparsedEntity(
		    new UnparsedEntity(entityName, entity.getPublicId(),
				       systemId, notationName, url));
	    }
	}
    }

    private void addElementTypes(DTD dtd, DocumentType docType) {
	Enumeration typeNames = dtd.elementTypeNames();

	while (typeNames.hasMoreElements()) {
	    String typeName = (String) typeNames.nextElement();
	    com.jclark.xml.parse.ElementType elemType = 
		dtd.getElementType(typeName);

	    int contentType;
	    switch (elemType.getContentType()) {
	    case com.jclark.xml.parse.ElementType.ANY_CONTENT:
		contentType = ElementTypeModel.ANY;
		break;
	    case com.jclark.xml.parse.ElementType.EMPTY_CONTENT:
		contentType = ElementTypeModel.EMPTY;
		break;
	    case com.jclark.xml.parse.ElementType.MIXED_CONTENT:
		contentType = ElementTypeModel.MIXED;
		break;
	    case com.jclark.xml.parse.ElementType.ELEMENT_CONTENT:
		contentType = ElementTypeModel.ELEMENT;
		break;
	    default:
		contentType = ElementTypeModel.UNDECLARED;
	    }

            // Note that UNDECLARED is treated as a first class type.

	    com.xmlmind.xmledit.modelimpl.ElementType elemTypeModel = 
		new com.xmlmind.xmledit.modelimpl.ElementType(
		    typeName, contentType);

	    Enumeration attrNames = elemType.attributeNames();
	    while (attrNames.hasMoreElements()) {
		String attrName = (String) attrNames.nextElement();
		com.jclark.xml.parse.AttributeDefinition attrDef = 
		    elemType.getAttributeDefinition(attrName);

		int type = AttributeDefinitionModel.CDATA;
		switch (attrDef.getType()) {
		case com.jclark.xml.parse.AttributeDefinition.ID:
		    type = AttributeDefinitionModel.ID;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.IDREF:
		    type = AttributeDefinitionModel.IDREF;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.IDREFS:
		    type = AttributeDefinitionModel.IDREFS;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.ENTITY:
		    type = AttributeDefinitionModel.ENTITY;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.ENTITIES:
		    type = AttributeDefinitionModel.ENTITIES;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.NMTOKEN:
		    type = AttributeDefinitionModel.NMTOKEN;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.NMTOKENS:
		    type = AttributeDefinitionModel.NMTOKENS;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.NOTATION:
		    type = AttributeDefinitionModel.NOTATION;
		    break;
		case com.jclark.xml.parse.AttributeDefinition.ENUM:
		    type = AttributeDefinitionModel.ENUMERATION;
		    break;
		default:
		    type = AttributeDefinitionModel.CDATA;
		}

		String[] enumValues;
		Enumeration allowedValues = attrDef.allowedValues();
		if (allowedValues == null) {
		    enumValues = null;
		} else {
		    Vector list = new Vector();
		    while (allowedValues.hasMoreElements()) 
			list.addElement(allowedValues.nextElement());

		    enumValues = new String[list.size()];
		    list.copyInto(enumValues);
		}

		int presence;
		String defaultValue = attrDef.getDefaultValue();
		boolean required = attrDef.isRequired();
		if (defaultValue == null)
		    presence = (required? 
				AttributeDefinitionModel.REQUIRED :
				AttributeDefinitionModel.IMPLIED);
		else 
		    presence = (required? AttributeDefinitionModel.FIXED :
				AttributeDefinitionModel.DEFAULTED);

		com.xmlmind.xmledit.modelimpl.AttributeDefinition 
		attrDefModel = getSharedAttributeDefinition(
		    attrName, type, enumValues, presence, defaultValue);

		elemTypeModel.putAttributeDefinition(attrDefModel);
	    }

	    docType.putElementType(elemTypeModel);
	}
    }

    private AttributeDefinition getSharedAttributeDefinition(
	                                  String name, int type, 
					  String[] enumValues, 
					  int presence, String defaultValue) {
	AttributeDefinition key = 
	    new AttributeDefinition(name, type, enumValues, presence, 
				    defaultValue);
	AttributeDefinition value = 
	    (AttributeDefinition) sharedAttributeDefinitions.get(key);
	if (value == null) {
	    sharedAttributeDefinitions.put(key, key);
	    value = key;
	}

	return value;
    }

    private void initElementContent(DTD dtd, DocumentType docType) {
	Enumeration elemTypes = docType.getElementTypes();
	while (elemTypes.hasMoreElements()) {
	    com.xmlmind.xmledit.modelimpl.ElementType elemType = 
		(com.xmlmind.xmledit.modelimpl.ElementType) 
		elemTypes.nextElement();

	    String spec;
	    switch (elemType.getElementContentType()) {
	    case ElementTypeModel.MIXED:
		spec = dtd.getElementType(
		    elemType.getElementTypeName()).getContentSpec();
		elemType.setMixedElementTypes(
		    parseMixedContentSpec(spec, docType));
		break;
	    case ElementTypeModel.ELEMENT:
		spec = dtd.getElementType(
		    elemType.getElementTypeName()).getContentSpec();
		elemType.setElementContent(
		    parseElementContentSpec(spec, docType));
		break;
	    }
	}
    }

    private ElementTypeModel[] parseMixedContentSpec(String spec, 
						     DocumentType docType) {
	StringTokenizer tokens = new StringTokenizer(spec, "(|)*");

	// Skip #PCDATA
	tokens.nextToken();
	if (!tokens.hasMoreTokens()) 
	    return new ElementTypeModel[0];

	Vector list = new Vector();
	while (tokens.hasMoreTokens()) {
	    String token = tokens.nextToken();
	    list.addElement(getElementType(docType, token));
	}

	ElementTypeModel[] mix = new ElementTypeModel[list.size()];
	list.copyInto(mix);
	return mix;
    }

    private ElementContentModel parseElementContentSpec(String spec, 
							DocumentType docType) {
	StringTokenizer tokens = new StringTokenizer(spec, "(|,)?*+", true);
	lookahead = tokens.nextToken();
	return parseElementContentSpec(tokens, docType);
    }

    private ElementContentModel parseElementContentSpec(StringTokenizer tokens,
							DocumentType docType) {
	int listType = ElementContentModel.NONE;
	ElementTypeModel elementType = null;
	ElementContentModel[] particles = null;
	int minOccurences = 1;
	int maxOccurences = 1;

	if (lookahead.charAt(0) == '(') {
	    lookahead = tokens.nextToken();
	    Vector list = new Vector();

	    listType = ElementContentModel.CHOICE;
	    loop: for (;;) {
		switch (lookahead.charAt(0)) {
		case ')':
		    break loop;
		case '|':
		    lookahead = tokens.nextToken();
		    break;
		case ',':
		    listType = ElementContentModel.SEQUENCE;
		    lookahead = tokens.nextToken();
		    break;
		default:
		    list.addElement(parseElementContentSpec(tokens, docType));
		}
	    }

	    // Parse (foo)?*+ as [NONE,foo,n,m] rather than 
	    // [CHOICE,[NONE,foo,1,1],n,m]

	    int count = list.size();
	    if (count == 1) {
		ElementContentModel child = 
		    (ElementContentModel) list.elementAt(0);

		if (child.getMinOccurences() == 1 &&
		    child.getMaxOccurences() == 1) 
		    elementType = child.getElementType();
	    }

	    if (elementType == null) {
		particles = new ElementContentModel[count];
		list.copyInto(particles);
	    } else {
		listType = ElementContentModel.NONE;
	    }
	} else {
	    elementType = getElementType(docType, lookahead);
	}

	if (tokens.hasMoreTokens()) {
	    lookahead = tokens.nextToken();

	    switch (lookahead.charAt(0)) {
	    case '?':
		minOccurences = 0;
		maxOccurences = 1;
		if (tokens.hasMoreTokens())
		    lookahead = tokens.nextToken();
		break;
	    case '*':
		minOccurences = 0;
		maxOccurences = Integer.MAX_VALUE;
		if (tokens.hasMoreTokens())
		    lookahead = tokens.nextToken();
		break;
	    case '+':
		minOccurences = 1;
		maxOccurences = Integer.MAX_VALUE;
		if (tokens.hasMoreTokens())
		    lookahead = tokens.nextToken();
		break;
	    }
	}

	if (listType == ElementContentModel.NONE) {
	    return new ElementContent(elementType, 
				      minOccurences, maxOccurences);
	} else {
	    // Parse (foo?*+) as [NONE,foo,n,m] rather than 
	    // [CHOICE,[NONE,foo,n,m],1,1]

	    if (particles.length == 1 && 
		minOccurences == 1 && maxOccurences == 1)
		return particles[0];
	    else
		return new ElementContent(listType, particles,
					  minOccurences, maxOccurences);
	}
    }

    private ElementTypeModel getElementType(DocumentType docType, 
					    String name) {
	ElementTypeModel elemType = docType.getElementType(name);

	if (elemType == null) {
	    // Found in the content model but not declared.

	    elemType = new com.xmlmind.xmledit.modelimpl.ElementType(
		name, ElementTypeModel.UNDECLARED);
	    docType.putElementType(elemType);
	}

	return elemType;
    }
}




