/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.Vector;
import java.util.Enumeration;
import fr.pixware.util.StringUtil;
import com.xmlmind.xmledit.model.*;

public class DefaultElemStructEditor extends ElementStructureEditor {
    public static final int MAKE_CHOICES     = DocumentUtil.MAKE_CHOICES;
    public static final int USE_FIRST_CHOICE = DocumentUtil.USE_FIRST_CHOICE;

    private boolean lenient;
    private int createContentOptions;
    private DocumentTypeModel docType;
    private ElementStructureChecker checker;

    public DefaultElemStructEditor() {
	super();

	checker = new ElementStructureChecker();
	setLenient(false);
	setCreateContentOptions(MAKE_CHOICES);
	setDocument(null);
    }

    public void setDocument(DocumentModel document) {
	super.setDocument(document);

	checker.setDocument(document);
	if (document == null)
	    docType = null;
	else
	    docType = document.getDocumentType();
    }

    public void setLenient(boolean lenient) {
	this.lenient = lenient;
    }

    public boolean isLenient() {
	return lenient;
    }

    public void setCreateContentOptions(int options) {
	createContentOptions = options;
    }

    public int getCreateContentOptions() {
	return createContentOptions;
    }

    // -----------------------------------------------------------------------

    /**
     * Returns true if selection begin is located after selection end.
     */
    protected boolean doGetCommandInfo(TextModel beginText, int beginOffset,
				       TextModel endText, int endOffset, 
				       CommandInfo info) {
	// No PASTE_BEFORE, PASTE_AFTER.
	// No INSERT_BEFORE, INSERT_AFTER.
	// No INSERT, REPLACE.

	info.commands |= (CommandInfo.COPY|CommandInfo.CUT|CommandInfo.DELETE);

	// Paste ---

	if (docType == null) {
	    if (info.clipboardTag != null) 
		info.commands |= CommandInfo.PASTE;
	} else {
	    if (info.clipboardTag != null) {
		// Paste into text selection is like delete then paste at
		// caret (and caret is supposed to be inside endText whether it
		// is before or after beginText) ---

		ElementModel parent = endText.getParentElement();
		ElementTypeModel parentType = 
		    ElementUtil.getElementType(parent);
		int contentType;
		String[] childTypeNames;

		if (parentType == null ||
		    (contentType = parentType.getElementContentType()) == 
		    ElementTypeModel.UNDECLARED || 
		    contentType == ElementTypeModel.ANY)
		    // ANY means any declared type or #PCDATA.
		    childTypeNames = checker.getDeclaredElementTypeNames();
		else if (contentType == ElementTypeModel.MIXED)
		    childTypeNames = 
			checker.getMixedContentChildTypeNames(parent);
		else
		    // Not a text container even if it currently contains
		    // text: cannot PASTE.
		    childTypeNames = null;

		if (childTypeNames != null &&
		    (info.clipboardTag == PCDATA ||
		     StringUtil.contains(childTypeNames, info.clipboardTag)))
		    info.commands |= CommandInfo.PASTE;
	    }
	}

	// Convert ---
	return convertCommandInfo(beginText, beginOffset, endText, endOffset, 
				  info);
    }

    private boolean convertCommandInfo(TextModel beginText, int beginOffset,
				       TextModel endText, int endOffset, 
				       CommandInfo info) {
	// Make sure that beginText is located before endText:
	// endText may be removed, beginText may not.

	boolean swapped;
	if (!checkTextOrder(beginText, beginOffset, endText, endOffset)) {
	    TextModel swapText = beginText; 
	    beginText = endText; 
	    endText = swapText;

	    int swapOffset = beginOffset; 
	    beginOffset = endOffset; 
	    endOffset = swapOffset;

	    swapped = true;
	} else {
	    swapped = false;
	}

	if (docType == null) {
	    info.commands |= CommandInfo.CONVERT;
	    info.convertArguments = null; /*anything*/
	} else {
	    ElementModel parent = beginText.getParentElement();
	    ElementTypeModel parentType = ElementUtil.getElementType(parent);
	    int contentType;
	    String[] childTypeNames;

	    if (parentType == null ||
		(contentType = parentType.getElementContentType()) == 
		ElementTypeModel.UNDECLARED ||
		contentType == ElementTypeModel.ANY) {
		// ANY means any declared type or #PCDATA.
		// Will not suggest UNDECLARED child type names.
		childTypeNames = checker.getDeclaredTextContainerNames();
	    } else if (contentType == ElementTypeModel.MIXED) {
		Vector list = new Vector();

		ElementTypeModel[] mix = parentType.getMixedElementTypes();
		if (mix != null) {
		    for (int i = 0; i < mix.length; ++i) {
			// Will not suggest UNDECLARED child type names.

			switch (mix[i].getElementContentType()) {
			case ElementTypeModel.MIXED:
			case ElementTypeModel.ANY:
			    list.addElement(mix[i].getElementTypeName());
			    break;
			}
		    }
		}

		childTypeNames = new String[list.size()];
		list.copyInto(childTypeNames);
	    } else {
		// Not a text container even if it currently contains text:
		// cannot CONVERT.
		childTypeNames = null;
	    }

	    if (childTypeNames != null) {
		if (beginText != endText)
		    childTypeNames = 
			StringUtil.prepend(childTypeNames, PCDATA);

		info.convertArguments = childTypeNames;
		if (info.convertArguments.length > 0)
		    info.commands |= CommandInfo.CONVERT;
	    }
	}

	return swapped;
    }

    // -----------------------------------------------------------------------

    protected void doGetCommandInfo(TextModel text, int offset,
				    CommandInfo info) {
	// No COPY, CUT, DELETE.
	// No PASTE_BEFORE, PASTE_AFTER.
	// No INSERT_BEFORE, INSERT_AFTER.
	// No CONVERT, REPLACE.

	if (docType == null) {
	    info.commands |= CommandInfo.INSERT;
	    info.insertArguments = null; /*anything*/

	    if (info.clipboardTag != null) 
		info.commands |= CommandInfo.PASTE;
	} else {
	    ElementModel parent = text.getParentElement();
	    ElementTypeModel parentType = ElementUtil.getElementType(parent);
	    int contentType;
	    String[] childTypeNames;

	    if (parentType == null ||
		(contentType = parentType.getElementContentType()) == 
		ElementTypeModel.UNDECLARED ||
		contentType == ElementTypeModel.ANY)
		// ANY means any declared type or #PCDATA.
		childTypeNames = checker.getDeclaredElementTypeNames();
	    else if (contentType == ElementTypeModel.MIXED)
		childTypeNames = checker.getMixedContentChildTypeNames(parent);
	    else
		// Not a text container even if it currently contains text:
		// cannot INSERT nor PASTE.
		childTypeNames = null;

	    if (childTypeNames != null) {
		info.insertArguments = childTypeNames;
		if (info.insertArguments.length > 0)
		    info.commands |= CommandInfo.INSERT;
		
		if (info.clipboardTag != null &&
		    (info.clipboardTag == PCDATA ||
		     StringUtil.contains(childTypeNames, info.clipboardTag)))
		    info.commands |= CommandInfo.PASTE;
	    }
	}
    }

    // -----------------------------------------------------------------------

    protected void doGetCommandInfo(SubelementModel sub, 
				    CommandInfo info) {
	info.commands |= CommandInfo.COPY;

	// 1) What can we do with sub's parent? ---

	ElementModel parent = sub.getParentElement();

	if (parent != null) {
	    ElementTypeModel elemType = ElementUtil.getElementType(parent);

	    if (elemType == null) {
		undeclaredContentCommandInfo1(parent, sub, info);
	    } else {
		switch (elemType.getElementContentType()) {
		case ElementTypeModel.EMPTY:
		    emptyContentCommandInfo1(parent, sub, info);
		    break;
		case ElementTypeModel.ANY:
		    anyContentCommandInfo1(parent, sub, info);
		    break;
		case ElementTypeModel.MIXED:
		    mixedContentCommandInfo1(parent, sub, info);
		    break;
		case ElementTypeModel.ELEMENT:
		    elementContentCommandInfo1(parent, sub, info);
		    break;
		case ElementTypeModel.UNDECLARED:
		    undeclaredContentCommandInfo1(parent, sub, info);
		    break;
		}
	    }
	}

	// 2) What can we insert or paste *into* sub? ---
	//
	// Works only for some empty elements. 
        //
	// Inserting elements into empty texts is ambiguous: do we insert 
	// before or after the text? Plus, there is a better way to do that.

	if (sub.getSubelementType() == SubelementModel.ELEMENT) {
	    ElementModel element = (ElementModel) sub;

	    if (element.getSubelementCount() == 0) {
		ElementTypeModel elemType=ElementUtil.getElementType(element);

		if (elemType == null) {
		    undeclaredContentCommandInfo2(element, info);
		} else {
		    switch (elemType.getElementContentType()) {
		    case ElementTypeModel.EMPTY:
			emptyContentCommandInfo2(element, info);
			break;
		    case ElementTypeModel.ANY:
			anyContentCommandInfo2(element, info);
			break;
		    case ElementTypeModel.MIXED:
			mixedContentCommandInfo2(element,  info);
			break;
		    case ElementTypeModel.ELEMENT:
			elementContentCommandInfo2(element, info);
			break;
		    case ElementTypeModel.UNDECLARED:
			undeclaredContentCommandInfo2(element, info);
			break;
		    }
		}
	    }
	}
    }

    private void emptyContentCommandInfo1(ElementModel parent, 
					  SubelementModel sub,
					  CommandInfo info) {
	info.commands |= (CommandInfo.DELETE|CommandInfo.CUT);

	// No JOIN.
	// No PASTE_BEFORE, PASTE_AFTER.
	// No INSERT_BEFORE, INSERT_AFTER.
	// No CONVERT, REPLACE.
    }

    private void anyContentCommandInfo1(ElementModel parent, 
					SubelementModel sub,
					CommandInfo info) {
	String[] declaredTypeNames = checker.getDeclaredElementTypeNames();

	info.commands |= (CommandInfo.DELETE|CommandInfo.CUT);

	if (checkCanJoin(sub))
	    info.commands |= CommandInfo.JOIN;
	    
        // ANY means any declared type or #PCDATA.
	if (info.clipboardTag != null &&
	    (info.clipboardTag == PCDATA || 
	     StringUtil.contains(declaredTypeNames, info.clipboardTag)))
	    info.commands |= 
		(CommandInfo.PASTE_BEFORE|CommandInfo.PASTE_AFTER);

	String[] declaredTypeNames2 = 
	    StringUtil.prepend(declaredTypeNames, PCDATA);
	SubelementModel neighbor;

	if (sub.getSubelementType() == SubelementModel.TEXT ||
	    ((neighbor = sub.getPreviousSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT))
	    info.insertBeforeArguments = declaredTypeNames;
	else
	    info.insertBeforeArguments = declaredTypeNames2;

	if (info.insertBeforeArguments.length > 0)
	    info.commands |= CommandInfo.INSERT_BEFORE;

	if (sub.getSubelementType() == SubelementModel.TEXT ||
	    ((neighbor = sub.getNextSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT))
	    info.insertAfterArguments = declaredTypeNames;
	else
	    info.insertAfterArguments = declaredTypeNames2;

	if (info.insertAfterArguments.length > 0)
	    info.commands |= CommandInfo.INSERT_AFTER;

	if (((neighbor = sub.getPreviousSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT) ||
	    ((neighbor = sub.getNextSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT))
	    info.replaceArguments = declaredTypeNames;
	else
	    info.replaceArguments = declaredTypeNames2;

	if (info.replaceArguments.length > 0)
	    info.commands |= CommandInfo.REPLACE;

	// Convert ---
	anyContentConvertInfo1(declaredTypeNames, sub, info);
    }

    private void anyContentConvertInfo1(String[] declaredTypeNames, 
					SubelementModel sub,
					CommandInfo info) {
	if (sub.getSubelementType() == SubelementModel.TEXT) {
	    info.convertArguments = 
		checker.getDeclaredTextContainerNames();
	    if (info.convertArguments.length > 0)
		info.commands |= CommandInfo.CONVERT;
	} else {
	    ElementModel child = (ElementModel) sub;
	    String childTypeName = child.getElementTypeName();

	    boolean quick = 
		((info.needArguments & CommandInfo.CONVERT_ARGUMENTS) == 0);

	    String[] list = new String[declaredTypeNames.length+1];
	    int listSize = 0;

	    if (ElementUtil.isTextContainer(child))
		list[listSize++] = PCDATA;

	    for (int i = 0; i < declaredTypeNames.length; ++i) {
		String typeName = declaredTypeNames[i];

		if (!typeName.equals(childTypeName) &&
		    checker.hasCompatibleContent(child, typeName)) {
		    list[listSize++] = typeName;
		    if (quick)
			break;
		}
	    }

	    if (listSize > 0) {
		info.convertArguments = new String[listSize];
		System.arraycopy(list, 0, info.convertArguments, 0, listSize);
		info.commands |= CommandInfo.CONVERT;
	    }
	}
    }

    private void mixedContentCommandInfo1(ElementModel parent, 
					  SubelementModel sub,
					  CommandInfo info) {
	String[] childTypeNames = 
	    checker.getMixedContentChildTypeNames(parent);
	String[] childTypeNames2 = StringUtil.prepend(childTypeNames, PCDATA);

	info.commands |= (CommandInfo.DELETE|CommandInfo.CUT);

	if (checkCanJoin(sub))
	    info.commands |= CommandInfo.JOIN;
	    
	if (info.clipboardTag != null &&
	    (info.clipboardTag == PCDATA || 
	     StringUtil.contains(childTypeNames, info.clipboardTag)))
	    info.commands |= 
		(CommandInfo.PASTE_BEFORE|CommandInfo.PASTE_AFTER);

	SubelementModel neighbor;

	if (sub.getSubelementType() == SubelementModel.TEXT ||
	    ((neighbor = sub.getPreviousSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT))
	    info.insertBeforeArguments = childTypeNames;
	else
	    info.insertBeforeArguments = childTypeNames2;

	if (info.insertBeforeArguments.length > 0)
	    info.commands |= CommandInfo.INSERT_BEFORE;

	if (sub.getSubelementType() == SubelementModel.TEXT ||
	    ((neighbor = sub.getNextSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT))
	    info.insertAfterArguments = childTypeNames;
	else
	    info.insertAfterArguments = childTypeNames2;

	if (info.insertAfterArguments.length > 0)
	    info.commands |= CommandInfo.INSERT_AFTER;

	// If the sub to be replaced needs to be replaced by text and this sub
	// is preceeded or followed by text, it is simple to delete the sub
	// rather to replace it.

	if (((neighbor = sub.getPreviousSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT)||
	    ((neighbor = sub.getNextSubelement()) != null &&
	     neighbor.getSubelementType() == SubelementModel.TEXT))
	    info.replaceArguments = childTypeNames;
	else
	    info.replaceArguments = childTypeNames2;

	if (info.replaceArguments.length > 0)
	    info.commands |= CommandInfo.REPLACE;

	// Convert ---
	mixedContentConvertInfo1(parent, sub, info);
    }

    private void mixedContentConvertInfo1(ElementModel parent, 
					  SubelementModel sub,
					  CommandInfo info) {
	ElementTypeModel[] mix = 
	    ElementUtil.getElementType(parent).getMixedElementTypes();

	boolean quick = 
	    ((info.needArguments & CommandInfo.CONVERT_ARGUMENTS) == 0);

	String[] list = new String[mix.length+1];
	int listSize = 0;

	if (sub.getSubelementType() == SubelementModel.TEXT) {
	    loop: for (int i = 0; i < mix.length; ++i) {
		ElementTypeModel type = mix[i];

		// Will not suggest UNDECLARED child type names.
		switch (type.getElementContentType()) {
		case ElementTypeModel.MIXED:
		case ElementTypeModel.ANY:
		    list[listSize++] = type.getElementTypeName();
		    if (quick)
			break loop;
		    break;
		}
	    }
	} else {
	    ElementModel child = (ElementModel) sub;
	    ElementTypeModel childType = ElementUtil.getElementType(child);

	    if (ElementUtil.isTextContainer(child))
		list[listSize++] = PCDATA;

	    for (int i = 0; i < mix.length; ++i) {
		ElementTypeModel type = mix[i];

		if (type != childType &&
		    checker.hasCompatibleContent(child, type)) {
		    list[listSize++] = type.getElementTypeName();
		    if (quick)
			break;
		}
	    }
	}

	if (listSize > 0) {
	    info.convertArguments = new String[listSize];
	    System.arraycopy(list, 0, info.convertArguments, 0, listSize);
	    info.commands |= CommandInfo.CONVERT;
	}
    }

    private void elementContentCommandInfo1(ElementModel parent, 
					    SubelementModel sub,
					    CommandInfo info) {
	String[] childTypeNames = 
	    checker.getElementContentChildTypeNames(parent);

	if (lenient ||
	    sub.getSubelementType() == SubelementModel.TEXT) {
	    info.commands |= (CommandInfo.DELETE|CommandInfo.CUT);

	    if (checkCanJoin(sub))
		info.commands |= CommandInfo.JOIN;
	    
	    if (info.clipboardTag != null && 
		info.clipboardTag != PCDATA &&
		StringUtil.contains(childTypeNames, info.clipboardTag))
		info.commands |= 
		    (CommandInfo.PASTE_BEFORE|CommandInfo.PASTE_AFTER);

	    info.commands |= 
		(CommandInfo.INSERT_BEFORE|CommandInfo.INSERT_AFTER|
		 CommandInfo.REPLACE);
	    info.insertBeforeArguments = info.insertAfterArguments = 
		info.replaceArguments = childTypeNames;
	} else {
	    ElementModel child = (ElementModel) sub;

	    if (checker.checkElementContentIfDelete(parent, child))
		info.commands |= (CommandInfo.DELETE|CommandInfo.CUT);

	    if ((info.commands & CommandInfo.DELETE) != 0 &&
		checkCanJoin(sub))
		info.commands |= CommandInfo.JOIN;
	    
	    if (info.clipboardTag != null && 
		info.clipboardTag != PCDATA) {
		if (checker.checkElementContentIfInsertBefore(
		        parent, child, info.clipboardTag))
		    info.commands |= CommandInfo.PASTE_BEFORE;

		if (checker.checkElementContentIfInsertAfter(
		        parent, child, info.clipboardTag))
		    info.commands |= CommandInfo.PASTE_AFTER;
	    }

	    // Insert before ---

	    String[] list = new String[childTypeNames.length];
	    int listSize = 0;
	    boolean quick = 
		((info.needArguments&CommandInfo.INSERT_BEFORE_ARGUMENTS)==0);

	    for (int i = 0; i < childTypeNames.length; ++i) {
		if (checker.checkElementContentIfInsertBefore(
		        parent, child, childTypeNames[i])) {
		    list[listSize++] = childTypeNames[i];
		    if (quick)
			break;
		}
	    }

	    if (listSize > 0) {
		info.insertBeforeArguments = new String[listSize];
		System.arraycopy(list, 0, info.insertBeforeArguments, 0, 
				 listSize);
		info.commands |= CommandInfo.INSERT_BEFORE;
	    }

	    // Insert after ---

	    listSize = 0;
	    quick = 
		((info.needArguments&CommandInfo.INSERT_AFTER_ARGUMENTS)==0);

	    for (int i = 0; i < childTypeNames.length; ++i) {
		if (checker.checkElementContentIfInsertAfter(
		        parent, child, childTypeNames[i])) {
		    list[listSize++] = childTypeNames[i];
		    if (quick)
			break;
		}
	    }

	    if (listSize > 0) {
		info.insertAfterArguments = new String[listSize];
		System.arraycopy(list, 0, info.insertAfterArguments, 0, 
				 listSize);
		info.commands |= CommandInfo.INSERT_AFTER;
	    }

	    // Replace ---

	    listSize = 0;
	    quick = ((info.needArguments&CommandInfo.REPLACE_ARGUMENTS) == 0);

	    for (int i = 0; i < childTypeNames.length; ++i) {
		if (checker.checkElementContentIfReplace(
		        parent, child, childTypeNames[i])) {
		    list[listSize++] = childTypeNames[i];
                    if (quick)
			break;
		}
	    }

	    if (listSize > 0) {
		info.replaceArguments = new String[listSize];
		System.arraycopy(list, 0, info.replaceArguments, 0, listSize);
		info.commands |= CommandInfo.REPLACE;
	    }
	}

	// Convert ---
	elementContentConvertInfo1(parent, childTypeNames, sub, info);
    }

    private void elementContentConvertInfo1(ElementModel parent,
					    String[] childTypeNames,
					    SubelementModel sub,
					    CommandInfo info) {
	if (sub.getSubelementType() == SubelementModel.TEXT)
	    return;

	ElementModel child = (ElementModel) sub;
	String childTypeName = child.getElementTypeName();

	boolean quick = 
	    ((info.needArguments & CommandInfo.CONVERT_ARGUMENTS) == 0);

	String[] list = new String[childTypeNames.length];
	int listSize = 0;

	for (int i = 0; i < childTypeNames.length; ++i) {
	    String typeName = childTypeNames[i];

	    if (!typeName.equals(childTypeName) &&
		checker.checkElementContentIfReplace(parent, child, typeName)&&
		checker.hasCompatibleContent(child, typeName)) {
		list[listSize++] = typeName;
		if (quick)
		    break;
	    }
	}

	if (listSize > 0) {
	    info.convertArguments = new String[listSize];
	    System.arraycopy(list, 0, info.convertArguments, 0, listSize);
	    info.commands |= CommandInfo.CONVERT;
	}
    }

    private void undeclaredContentCommandInfo1(ElementModel parent, 
					       SubelementModel sub,
					       CommandInfo info) {
	// Beware! For this method unlike for other *CommandInfo1 methods
	// there may be no DTD at all.

	if (docType == null) {
	    info.commands |= (CommandInfo.DELETE|CommandInfo.CUT);

	    // Unlike checkCanJoin, canJoin just tests that the
	    // ElementStructureEditor is *able* to join sub to its
	    // previous sibling.
	    if (canJoin(sub))
		info.commands |= CommandInfo.JOIN;
		
	    if (info.clipboardTag != null)
		info.commands |= 
		    (CommandInfo.PASTE_BEFORE|CommandInfo.PASTE_AFTER);

	    // Insert anything. Replace by anything. Convert to anything.
	    info.commands |= 
		(CommandInfo.INSERT_BEFORE|CommandInfo.INSERT_AFTER|
		 CommandInfo.REPLACE|CommandInfo.CONVERT);
	    info.insertBeforeArguments = info.insertAfterArguments = 
		info.replaceArguments = info.convertArguments = null;
	} else {
	    anyContentCommandInfo1(parent, sub, info);
	}
    }

    private void emptyContentCommandInfo2(ElementModel element, 
					  CommandInfo info) {
	// No PASTE.
	// No INSERT.
    }

    private void anyContentCommandInfo2(ElementModel element, 
					CommandInfo info) {
	String[] declaredTypeNames = checker.getDeclaredElementTypeNames();

        // ANY means any declared type or #PCDATA.
	if (info.clipboardTag != null &&
	    (info.clipboardTag == PCDATA ||
	     StringUtil.contains(declaredTypeNames, info.clipboardTag)))
	    info.commands |= CommandInfo.PASTE;

	info.commands |= CommandInfo.INSERT;
	info.insertArguments = StringUtil.prepend(declaredTypeNames, PCDATA);
    }

    private void mixedContentCommandInfo2(ElementModel element, 
					  CommandInfo info) {
	String[] childTypeNames = 
	    checker.getMixedContentChildTypeNames(element);

	if (info.clipboardTag != null &&
	    (info.clipboardTag == PCDATA ||
	     StringUtil.contains(childTypeNames, info.clipboardTag)))
	    info.commands |= CommandInfo.PASTE;

	info.commands |= CommandInfo.INSERT;
	info.insertArguments = 
	    StringUtil.prepend(childTypeNames, PCDATA);
    }

    private void elementContentCommandInfo2(ElementModel element, 
					    CommandInfo info) {
	String[] childTypeNames = 
	    checker.getElementContentChildTypeNames(element);

	if (lenient) {
	    if (info.clipboardTag != null &&
		info.clipboardTag != PCDATA &&
		StringUtil.contains(childTypeNames, info.clipboardTag))
		info.commands |= CommandInfo.PASTE;
		
	    info.commands |= CommandInfo.INSERT;
	    info.insertArguments = childTypeNames;
	} else {
	    if (info.clipboardTag != null &&
		info.clipboardTag != PCDATA &&
		checker.checkElementContentIfInsertBefore(
		    element, null, info.clipboardTag))
		info.commands |= CommandInfo.PASTE;

	    String[] list = new String[childTypeNames.length];
	    int listSize = 0;
	    boolean quick = 
		((info.needArguments & CommandInfo.INSERT_ARGUMENTS) == 0);

	    for (int i = 0; i < childTypeNames.length; ++i) {
		if (checker.checkElementContentIfInsertBefore(
		    element, null, childTypeNames[i])) {
		    list[listSize++] = childTypeNames[i];
		    if (quick)
			break;
		}
	    }

	    if (listSize > 0) {
		info.insertArguments = new String[listSize];
		System.arraycopy(list, 0, info.insertArguments, 0, listSize);
		info.commands |= CommandInfo.INSERT;
	    }
	}
    }

    private void undeclaredContentCommandInfo2(ElementModel element, 
					       CommandInfo info) {
	// Beware! For this method unlike for other *CommandInfo2 methods
	// there may be no DTD at all.

	if (docType == null) {
	    if (info.clipboardTag != null)
		info.commands |= CommandInfo.PASTE;

	    // Insert anything.
	    info.commands |= CommandInfo.INSERT;
	    info.insertArguments = null;
	} else {
	    anyContentCommandInfo2(element, info);
	}
    }

    // -----------------------------------------------------------------------

    public boolean checkCanSplit(ElementModel element, 
				 TextModel text, int offset) {
	if (docType == null)
	    return true;

	ElementModel ancestor = text.getParentElement();
	ElementTypeModel ancestorType = ElementUtil.getElementType(ancestor);
	if (ancestorType == null)
	    return false;
	switch (ancestorType.getElementContentType()) {
	case ElementTypeModel.MIXED:
	case ElementTypeModel.ANY:
	    // ANY means any declared type or #PCDATA.
	case ElementTypeModel.UNDECLARED:
	    // Note that UNDECLARED is treated as a first class type.
	    break;
	default:
	    return false;
	}

	ElementStructureChecker.ElementContentBuffer buffer = 
	    checker.getElementContentBuffer();

	while (ancestor != element) {
	    ElementModel previousAncestor = ancestor;

	    ancestor = ancestor.getParentElement();
	    ancestorType = ElementUtil.getElementType(ancestor);
	    if (ancestorType == null)
		return false;

	    switch (ancestorType.getElementContentType()) {
	    case ElementTypeModel.MIXED:
	    case ElementTypeModel.ANY:
	    case ElementTypeModel.UNDECLARED:
		break;

	    case ElementTypeModel.ELEMENT:
	        {
		    // Check first part ---
		    buffer.clear();

		    SubelementModel sub = ancestor.getFirstSubelement();
		    while (sub != previousAncestor) {
			if (!buffer.append(sub))
			    // Invalid element: contains text.
			    return false;

			sub = sub.getNextSubelement();
		    }
		    buffer.append(previousAncestor);

		    if (!checker.checkElementContent(ancestorType, buffer))
			return false;

		    // Check second part ---
		    buffer.clear();

		    buffer.append(previousAncestor);
		    sub = sub.getNextSubelement();
		    while (sub != null) {
			if (!buffer.append(sub))
			    return false;

			sub = sub.getNextSubelement();
		    }

		    if (!checker.checkElementContent(ancestorType, buffer))
			return false;
	        }
		break;

	    default:
                // EMPTY
		return false;
	    }
	}

	// Check if parent can contain the 2 consecutive parts ---

	ElementModel parent = element.getParentElement();
	ElementTypeModel parentType = ElementUtil.getElementType(parent);
	if (parentType == null)
	    return false;

	switch (parentType.getElementContentType()) {
	case ElementTypeModel.MIXED:
	case ElementTypeModel.ANY:
	case ElementTypeModel.UNDECLARED:
	    return true;

	case ElementTypeModel.ELEMENT:
	    return checker.checkElementContentIfInsertAfter(
		parent, element, element.getElementTypeName());

	default:
	    // EMPTY
	    return false;
	}
    }

    // -----------------------------------------------------------------------

    private boolean checkCanJoin(SubelementModel sub) {
	if (!canJoin(sub))
	    return false;
	ElementModel src = (ElementModel) sub;
	ElementModel dst = (ElementModel) sub.getPreviousSubelement();

	return doCheckCanJoin(src, dst);
    }

    private boolean doCheckCanJoin(ElementModel src, ElementModel dst) {
	// Assumes that src and dst element types are the same.
	ElementTypeModel elemType = ElementUtil.getElementType(dst);
	if (elemType == null)
	    // Not referenced in the DTD.
	    return true;

	switch (elemType.getElementContentType()) {
	case ElementTypeModel.MIXED:
	case ElementTypeModel.ANY:
	case ElementTypeModel.UNDECLARED:
	    return true;

	case ElementTypeModel.ELEMENT:
	    {
		ElementStructureChecker.ElementContentBuffer buffer = 
		    checker.getElementContentBuffer();

		SubelementModel dstSub = dst.getFirstSubelement();
		while (dstSub != null) {
		    if (!buffer.append(dstSub))
			// Invalid element: contains text.
			return false;

		    dstSub = dstSub.getNextSubelement();
		}

		SubelementModel srcSub = src.getFirstSubelement();

		if (srcSub != null && dstSub != null &&
		    srcSub.getSubelementType() == SubelementModel.ELEMENT &&
		    dstSub.getSubelementType() == SubelementModel.ELEMENT &&
		    ((ElementModel) srcSub).getElementTypeName().equals(
			((ElementModel) dstSub).getElementTypeName())) {

		    if (doCheckCanJoin((ElementModel) srcSub, 
				       (ElementModel) dstSub))
			srcSub = srcSub.getNextSubelement();
		    // Otherwise both dstSub and srcSub are to be found in dst.
		}

		while (srcSub != null) {
		    if (!buffer.append(srcSub))
			// Invalid element: contains text.
			return false;

		    srcSub = srcSub.getNextSubelement();
		}

		return checker.checkElementContent(elemType, buffer);
	    }

	default:
	    // EMPTY (use DELETE rather than JOIN)
	    return false;
	}
    }

    // -----------------------------------------------------------------------

    protected ElementModel createConvertedElement(String elementTypeName) {
	return DocumentUtil.createElementModel(doc, elementTypeName, 
                                               /*no content*/ 0x0);
    }

    protected ElementModel createInsertedElement(String elementTypeName) {
	return DocumentUtil.createElementModel(
	    doc, elementTypeName, 
	    (DocumentUtil.CREATE_CONTENT|createContentOptions));
    }

    protected ElementModel createSplitElement(ElementModel element) {
	String elemTypeName = element.getElementTypeName();
	ElementModel split = doc.createElementModel(elemTypeName);

	if (docType != null) {
	    ElementTypeModel elemType = docType.getElementType(elemTypeName);
	    if (elemType != null)
		DocumentUtil.addRequiredAttributes(split, elemType, doc);
	}

	return split;
    }

    protected boolean checkPastedElement(ElementModel element) {
	return ((docType == null)? true : (checker.check(element) == null));
    }
}
