/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.io.*;
import java.util.Enumeration;
import java.util.Hashtable;
import com.xmlmind.xmledit.model.*;

public class DocumentTypeWriter {
    public static void writeDocumentType(DocumentTypeModel documentType, 
					 String fileName, boolean useURL) 
	throws IOException, UnsupportedEncodingException {
	BufferedOutputStream out = 
	    new BufferedOutputStream(new FileOutputStream(fileName));
	    
	try {
	    writeDocumentType(documentType, out, "UTF8", useURL);
	} finally {
	    out.close();
	}
    }

    public static void writeDocumentType(DocumentTypeModel documentType, 
					 OutputStream out, String javaEncoding,
					 boolean useURL)
	throws IOException, UnsupportedEncodingException {
	String xmlEncoding = XMLUtil.javaToIANAEncoding(javaEncoding);
	if (xmlEncoding == null)
	    throw new UnsupportedEncodingException(
		"don't know the XML encoding name for Java encoding '" + 
		javaEncoding + "'");

	Writer writer = new OutputStreamWriter(out, javaEncoding);
	writeTextDeclaration(/*version*/ null, xmlEncoding, writer);
	writeDocumentType(documentType, writer, useURL);
    }

    public static void writeDocumentType(DocumentTypeModel documentType, 
					 Writer writer, boolean useURL) 
	throws IOException {
	Enumeration notations = documentType.getNotations();
	while (notations.hasMoreElements()) 
	    writeNotation((NotationModel) notations.nextElement(), 
			  writer, useURL);
 
	Enumeration unparsedEntities = documentType.getUnparsedEntities();
	while (unparsedEntities.hasMoreElements()) 
	    writeUnparsedEntity(
		(UnparsedEntityModel) unparsedEntities.nextElement(), 
		writer, useURL);

	Hashtable written = new Hashtable();

	Enumeration elementTypes = documentType.getElementTypes();
	while (elementTypes.hasMoreElements()) 
		writeElementTypeTree(
		    (ElementTypeModel) elementTypes.nextElement(), 
		    writer, written);
 
	writer.write('\n');
	writer.flush();
    }

    public static void writeTextDeclaration(String version, String encoding, 
					    Writer out) 
	throws IOException {
	out.write("<?xml");

	if (version != null) {
	    out.write(" version='");
	    out.write(version);
	    out.write('\'');
	}

	out.write(" encoding='");
	out.write(encoding);
	out.write('\'');

	out.write("?>\n");
    }

    public static void writeNotation(NotationModel notation, Writer out, 
				     boolean useURL) 
	throws IOException {
	out.write("<!NOTATION ");
	out.write(notation.getNotationName());

	DocumentWriter.writeExternalId(notation.getNotationPublicId(), 
				       (useURL? notation.getNotationURL() : 
					notation.getNotationSystemId()), 
				       out);

	out.write(">\n");
    }

    public static void writeUnparsedEntity(UnparsedEntityModel entity, 
					   Writer out, boolean useURL) 
	throws IOException {
	out.write("<!ENTITY ");
	out.write(entity.getUnparsedEntityName());

	DocumentWriter.writeExternalId(entity.getUnparsedEntityPublicId(), 
				       (useURL? entity.getUnparsedEntityURL() :
					entity.getUnparsedEntitySystemId()), 
				       out);

	out.write("\nNDATA ");
	out.write(entity.getUnparsedEntityNotationName());
	out.write(">\n");
    }

    private static void writeElementTypeTree(ElementTypeModel elemType, 
					     Writer out, Hashtable written) 
	throws IOException {
	String typeName = elemType.getElementTypeName();

	if (written.get(typeName) == null) {
	    written.put(typeName, typeName);

	    switch (elemType.getElementContentType()) {
	    case ElementTypeModel.EMPTY:
	    case ElementTypeModel.ANY:
	    case ElementTypeModel.UNDECLARED:
		// Note that UNDECLARED is treated as a first class type.
		writeElementType(elemType, out);
		break;

	    case ElementTypeModel.MIXED:
		{
		    ElementTypeModel[] mix = elemType.getMixedElementTypes();
		    for (int i = 0; i < mix.length; ++i) 
			writeElementTypeTree(mix[i], out, written);

		    writeElementType(elemType, out);
		}
		break;

	    case ElementTypeModel.ELEMENT:
		writeElementContentTree(elemType.getElementContent(), 
					out, written);
		writeElementType(elemType, out);
		break;
	    }
	}
    }

    private static void writeElementContentTree(ElementContentModel content, 
						Writer out, Hashtable written) 
	throws IOException {
	int listType = content.getListType();

	if (listType == ElementContentModel.NONE) {
	    writeElementTypeTree(content.getElementType(), out, written);
	} else {
	    ElementContentModel[] list = content.getList();
	    for (int i = 0; i < list.length; ++i) 
		writeElementContentTree(list[i], out, written);
	}
    }

    public static void writeElementType(ElementTypeModel elemType, Writer out) 
	throws IOException {
	int contentType = elemType.getElementContentType();

	if (contentType != ElementTypeModel.UNDECLARED) {
	    out.write("<!ELEMENT ");
	    out.write(elemType.getElementTypeName());

	    out.write(' ');
	    switch (contentType) {
	    case ElementTypeModel.EMPTY:
		out.write("EMPTY");
		break;

	    case ElementTypeModel.ANY:
		out.write("ANY");
		break;

	    case ElementTypeModel.MIXED:
		{
		    ElementTypeModel[] mix = elemType.getMixedElementTypes();

		    out.write("(#PCDATA");
		    if (mix.length == 0) {
			out.write(')');
		    } else {
			for (int i = 0; i < mix.length; ++i) {
			    out.write('|');
			    out.write(mix[i].getElementTypeName());
			}
			out.write(")*");
		    }
		}
		break;

	    case ElementTypeModel.ELEMENT:
		{
		    ElementContentModel content = elemType.getElementContent();
		    boolean parenthesis = 
			(content.getListType() == ElementContentModel.NONE);

		    if (parenthesis)
			out.write('(');
		    writeElementContent(content, out);
		    if (parenthesis)
			out.write(')');
		}
		break;
	    }

	    out.write(">\n");
	}
        
        // Note that UNDECLARED is treated as a first class type.

	if (elemType.getAttributeDefinitionCount() > 0) {
	    out.write("<!ATTLIST ");
	    out.write(elemType.getElementTypeName());
	    out.write('\n');

	    Enumeration attrDefs = elemType.getAttributeDefinitions();
	    while (attrDefs.hasMoreElements()) {
		out.write("  ");
		writeAttributeDefinition(
		    (AttributeDefinitionModel) attrDefs.nextElement(), out);
		out.write('\n');
	    }

	    out.write(">\n");
	}
    }

    private static void writeElementContent(ElementContentModel content, 
					    Writer out) 
	throws IOException {
	int listType = content.getListType();

	if (listType == ElementContentModel.NONE) {
	    out.write(content.getElementType().getElementTypeName());
	} else {
	    char connector = 
		((listType == ElementContentModel.CHOICE)? '|' : ',');
	    ElementContentModel[] list = content.getList();

	    out.write('(');
	    writeElementContent(list[0], out);
	    for (int i = 1; i < list.length; ++i) {
		out.write(connector);
		writeElementContent(list[i], out);
	    }
	    out.write(')');
	}

	int minOccur = content.getMinOccurences();
	int maxOccur = content.getMaxOccurences();
	if (minOccur == 0) {
	    if (maxOccur == 1) 
		out.write('?');
	    else
		out.write('*');
	} else {
	    if (maxOccur != 1) 
		out.write('+');
	}
    }

    private static void writeAttributeDefinition(AttributeDefinitionModel def,
						 Writer out) 
	throws IOException {
	out.write(def.getAttributeName());

	out.write(' ');
	int type = def.getAttributeType();
	switch (type) {
	case AttributeDefinitionModel.CDATA:
	    out.write("CDATA");
	    break;
	case AttributeDefinitionModel.ID:
	    out.write("ID");
	    break;
	case AttributeDefinitionModel.IDREF:
	    out.write("IDREF");
	    break;
	case AttributeDefinitionModel.IDREFS:
	    out.write("IDREFS");
	    break;
	case AttributeDefinitionModel.ENTITY:
	    out.write("ENTITY");
	    break;
	case AttributeDefinitionModel.ENTITIES:
	    out.write("ENTITIES");
	    break;
	case AttributeDefinitionModel.NMTOKEN:
	    out.write("NMTOKEN");
	    break;
	case AttributeDefinitionModel.NMTOKENS:
	    out.write("NMTOKENS");
	    break;
	case AttributeDefinitionModel.NOTATION:
	case AttributeDefinitionModel.ENUMERATION:
	    {
		if (type == AttributeDefinitionModel.NOTATION)
		    out.write("NOTATION ");

		String[] enumValues = def.getAttributeEnumeratedValues();
		out.write('(');
		out.write(enumValues[0]);
		for (int i = 1; i < enumValues.length; ++i) {
		    out.write('|');
		    out.write(enumValues[i]);
		}
		out.write(')');
	    }
	    break;
	}

	switch (def.getAttributePresence()) {
	case AttributeDefinitionModel.REQUIRED:
	    out.write(" #REQUIRED");
	    break;
	case AttributeDefinitionModel.IMPLIED:
	    out.write(" #IMPLIED");
	    break;
	case AttributeDefinitionModel.FIXED:
	    out.write(" #FIXED ");
	    break;
	}

	String defaultValue = def.getAttributeDefaultValue();
	if (defaultValue != null) {
	    out.write(' ');
	    out.write(XMLUtil.qxml(defaultValue));
	}
    }
}
