/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.Hashtable;
import java.util.Enumeration;
import fr.pixware.util.HashtableUtil;
import com.xmlmind.xmledit.model.*;

public class DocumentUtil {
    public static final int CREATE_CONTENT   = 0x1;
    public static final int MAKE_CHOICES     = 0x2;
    public static final int USE_FIRST_CHOICE = 0x4;	

    private static int uniqueId = 0;

    // -----------------------------------------------------------------------

    public static ElementModel createElementModel(DocumentModel doc, 
						  String elemTypeName,
						  int options) {
	DocumentTypeModel docType = doc.getDocumentType();
	ElementTypeModel elemType;
	if (docType == null) 
	    elemType = null;
	else
	    elemType = docType.getElementType(elemTypeName);

	ElementModel placeholder = doc.createElementModel(elemTypeName);

	if (elemType != null) {
	    // Note that UNDECLARED is treated as a first class type.

	    if ((options & CREATE_CONTENT) != 0)
		appendPlaceholders(doc, placeholder, elemType, options);

	    // Doc is passed as an argument because placeholder is not yet 
	    // attached to the document.
	    addRequiredAttributes(placeholder, elemType, doc);
	}

	return placeholder;
    }

    /**
     * @return false if the operation failed because the structure would 
     *         become recursive, true otherwise.
     */
    private static boolean appendPlaceholders(DocumentModel doc, 
					      ElementModel parent,
					      ElementTypeModel elemType, 
					      int options) {
	switch (elemType.getElementContentType()) {
	case ElementTypeModel.ELEMENT:
	    return appendPlaceholders(doc, parent, 
				      elemType.getElementContent(), options);

	case ElementTypeModel.MIXED:
	    parent.appendSubelement(doc.createTextModel(""));
	    /*FALLTHROUGH*/
	default:
	    // EMPTY, ANY, UNDECLARED.
	    return true;
	}
    }

    private static 
    boolean appendPlaceholders(DocumentModel doc, ElementModel parent,
			       ElementContentModel contentModel, int options) {
	boolean done = true;

	if (contentModel.getMinOccurences() > 0) {
	    switch (contentModel.getListType()) {
	    case ElementContentModel.NONE: 
		{
		    ElementTypeModel elemType = contentModel.getElementType();
		    String elemTypeName = elemType.getElementTypeName();
		    if (isRecursiveStructure(parent, elemTypeName)) {
			done = false;
			break;
		    }

		    ElementModel placeholder = 
			doc.createElementModel(elemTypeName);
		    parent.appendSubelement(placeholder);

		    // Adding required attributes *after* placeholder content
		    // has been created is more efficient if building
		    // placeholder content fails.

		    if (!appendPlaceholders(doc, placeholder, elemType, 
					    options)) {
			parent.removeSubelement(placeholder);
			done = false;
			break;
		    }

		    addRequiredAttributes(placeholder, elemType, doc);
		}
		break;

	    case ElementContentModel.CHOICE:
		{
		    ElementContentModel[] list = contentModel.getList();

		    if ((options & MAKE_CHOICES) != 0 || list.length <= 1) {
			done = false;

			if ((options & USE_FIRST_CHOICE) == 0) {
			    // First pass: heuristic: use mixed with shortest
			    // name.

			    int shortestNameIndex = -1;
			    int shortestNameLength = Integer.MAX_VALUE;

			    for (int i = 0; i < list.length; ++i) {
				ElementContentModel item = list[i];
				ElementTypeModel elemType;

				if ((elemType = item.getElementType()) != 
				    null &&
				    elemType.getElementContentType() == 
				    ElementTypeModel.MIXED) {
				    int nameLength = 
					elemType.getElementTypeName().length();

				    if (shortestNameIndex < 0 ||
					nameLength < shortestNameLength) {
					shortestNameIndex = i;
					shortestNameLength = nameLength;
				    }
				}
			    }
			    if (shortestNameIndex >= 0 &&
				appendPlaceholders(doc, parent, 
						   list[shortestNameIndex], 
						   options))
				done = true;

			    // Second pass: try any other mixed.

			    if (!done) { 
				for (int i = 0; i < list.length; ++i) {
				    ElementContentModel item = list[i];
				    ElementTypeModel elemType;

				    if ((elemType = item.getElementType()) != 
					null &&
					elemType.getElementContentType() == 
					ElementTypeModel.MIXED && 
					appendPlaceholders(doc, parent, item, 
							   options)) {
					done = true;
					break;
				    }
				}
			    }
			}

                        // Third pass (or only pass depending on options): try
                        // any other content model.

			if (!done) { 
			    for (int i = 0; i < list.length; ++i) {
				if (appendPlaceholders(doc, parent, list[i], 
						       options)) {
				    done = true;
				    break;
				}
			    }
			}
		    } else {
			done = true;
		    }
		}
		break;

	    case ElementContentModel.SEQUENCE:
		{
		    ElementContentModel[] list = contentModel.getList();
		    int childCount = parent.getSubelementCount();

		    for (int i = 0; i < list.length; ++i) {
			if (!appendPlaceholders(doc, parent, list[i], 
						options)) {
			    done = false;
			    break;
			}
		    }
		    
		    if (!done)
			restoreSubelements(parent, childCount);
		}
		break;
	    }
	}

	return done;
    }

    private static final boolean isRecursiveStructure(ElementModel parent, 
						      String elemTypeName) {
	ElementModel ancestor = parent;
	boolean recursive = false;

	while (ancestor != null) {
	    if (ancestor.getElementTypeName().equals(elemTypeName)) {
		recursive = true;
		break;
	    }
	    ancestor = ancestor.getParentElement();
	}

	return recursive;
    }

    private static final void restoreSubelements(ElementModel parent, 
						 int childCount) {
	int currentCount = parent.getSubelementCount();
	while (currentCount > childCount) {
	    parent.removeSubelement(parent.getLastSubelement());
	    --currentCount;
	}
    }

    /**
     * Works event if element is not yet attached to doc.
     */
    public static void addRequiredAttributes(ElementModel elem, 
					     ElementTypeModel elemType,
					     DocumentModel doc) {
	Enumeration iter = elemType.getAttributeDefinitions();
	while (iter.hasMoreElements()) {
	    AttributeDefinitionModel attrDef = 
		(AttributeDefinitionModel) iter.nextElement();
	    if (attrDef.getAttributePresence() != 
		AttributeDefinitionModel.REQUIRED)
		continue;

	    String value = null;
	    switch (attrDef.getAttributeType()) {
	    case AttributeDefinitionModel.ID:
		value = "___" + Integer.toString(uniqueId++);
		break;
	    case AttributeDefinitionModel.IDREF:
	    case AttributeDefinitionModel.IDREFS:
		value = (new IdFinder()).findId(doc);
		break;
	    case AttributeDefinitionModel.NMTOKEN:
	    case AttributeDefinitionModel.NMTOKENS:
		value = "___";
		break;
	    case AttributeDefinitionModel.ENTITY:
	    case AttributeDefinitionModel.ENTITIES:
	        {
		    Enumeration iter2 = 
			doc.getDocumentType().getUnparsedEntities();

		    if (iter2.hasMoreElements())
			value = ((UnparsedEntityModel) 
				 iter2.nextElement()).getUnparsedEntityName();
		}
		break;
	    case AttributeDefinitionModel.NOTATION:
	    case AttributeDefinitionModel.ENUMERATION:
	        {
		    String[] values = attrDef.getAttributeEnumeratedValues();

		    if (values != null && values.length > 0)
			value = values[0];
		}
		break;
	    case AttributeDefinitionModel.CDATA:
	    default:
		value = "???";
		break;
	    }

	    if (value != null)
		elem.putElementAttribute(attrDef.getAttributeName(), value);
	}
    }

    private static class IdFinder extends Traversal.HandlerBase {
	private DocumentTypeModel docType;

	public String findId(DocumentModel doc) {
	    docType = doc.getDocumentType();

	    return (String) Traversal.traverse(doc.getRootElement(), this);
	}

	public Object enterElement(ElementModel element) { 
	    if (element.getElementAttributeCount() == 0)
		return null;

	    ElementTypeModel elemType = (ElementTypeModel) 
		docType.getElementType(element.getElementTypeName());
	    if (elemType == null) 
		// Note that UNDECLARED is treated as a first class type.
		return null;

	    Enumeration iter = element.getElementAttributes();
	    while (iter.hasMoreElements()) {
		String[] pair = (String[]) iter.nextElement();

		AttributeDefinitionModel attrDef = 
		    elemType.getAttributeDefinition(pair[0]);
		if (attrDef != null &&
		    attrDef.getAttributeType() == AttributeDefinitionModel.ID)
		    return pair[1];
	    }

	    return null;
	}
    }

    // -----------------------------------------------------------------------

    public static String[] getAllElementTypeNames( DocumentModel doc) {
	final Hashtable elementTypeNames = new Hashtable();

	Traversal.traverse(doc.getRootElement(), new Traversal.HandlerBase() {
	    public Object enterElement(ElementModel element) {
		String tag = element.getElementTypeName();
		elementTypeNames.put(tag, tag);

		return null;
	    }
	});

	return HashtableUtil.getAllElements(elementTypeNames);
    }

    public static String[] getAllElementAttributeNames(DocumentModel doc) {
	final Hashtable attributeNames = new Hashtable();

	Traversal.traverse(doc.getRootElement(), new Traversal.HandlerBase() {
	    public Object enterElement(ElementModel element) {
		Enumeration iter = element.getElementAttributes();
		while (iter.hasMoreElements()) {
		    String[] nameValuePair = (String[]) iter.nextElement();
		    attributeNames.put(nameValuePair[0], nameValuePair[0]);
		}

		return null;
	    }
	});

	return HashtableUtil.getAllElements(attributeNames);
    }

    // -----------------------------------------------------------------------

    public static void copyTemplateDocument(DocumentModel template,
					    DocumentModel doc, 
					    int options) {
	ElementModel rootTemplate = template.getRootElement();

	ElementModel root = 
	    createElementModel(template, rootTemplate.getElementTypeName(),
			       (DocumentUtil.CREATE_CONTENT|options));

	Enumeration iter = rootTemplate.getElementAttributes();
	while (iter.hasMoreElements()) {
	    String[] pair = (String[]) iter.nextElement();
	    root.putElementAttribute(pair[0], pair[1]);
	}

	doc.initializeDocument(root, template.getDocumentType());
    
	iter = template.getDocumentProperties();
	while (iter.hasMoreElements()) {
	    Object[] pair = (Object[]) iter.nextElement();
	    doc.putDocumentProperty((String) pair[0], pair[1]);
	}
    }
}
