/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.io.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import com.xmlmind.xmledit.model.*;

public class DocumentWriter {
    public static void writeDocument(DocumentModel document, String fileName, 
				     boolean useURL, int indent, 
				     int maxLineLength) 
	throws IOException, UnsupportedEncodingException {
	BufferedOutputStream out = 
	    new BufferedOutputStream(new FileOutputStream(fileName));
	    
	try {
	    writeDocument(document, out, "UTF8", useURL, indent, 
			  maxLineLength);
	} finally {
	    out.flush();
	    out.close();
	}
    }

    public static void writeDocument(DocumentModel document, OutputStream out,
				     String javaEncoding, boolean useURL, 
				     int indent, int maxLineLength) 
	throws IOException, UnsupportedEncodingException {
	String xmlEncoding = XMLUtil.javaToIANAEncoding(javaEncoding);
	if (xmlEncoding == null)
	    throw new UnsupportedEncodingException(
		"don't know the XML encoding name for Java encoding '" + 
		javaEncoding + "'");

	Writer writer = new OutputStreamWriter(out, javaEncoding);
	writeXMLDeclaration(xmlEncoding, /*standalone*/ null, writer);
	writeDocument(document, writer, useURL, indent, maxLineLength);
    }

    public static void writeDocument(DocumentModel document, Writer writer, 
				     boolean useURL, int indent, 
				     int maxLineLength) 
	throws IOException {
	StyleSheetInfo[] info = (StyleSheetInfo[]) 
	    document.getDocumentProperty("xml-stylesheet");
	if (info != null) {
	    for (int j = 0; j < info.length; ++j)
		writeStyleSheetPI(info[j], writer, useURL);
	}

	ElementModel root = document.getRootElement();
	writeDocumentTypeDeclaration(root, writer, useURL);
	writeElement(root, writer, indent, maxLineLength);

	writer.write('\n');
	writer.flush();
    }

    public static void writeXMLDeclaration(String encoding, String standalone, 
					   Writer out) 
	throws IOException {
	out.write("<?xml version='1.0'");

	if (encoding != null) {
	    out.write(" encoding='");
	    out.write(encoding);
	    out.write('\'');
	}

	if (standalone != null) {
	    out.write(" standalone='");
	    out.write(standalone);
	    out.write('\'');
	}

	out.write("?>\n");
    }

    public static void writeStyleSheetPI(StyleSheetInfo info, Writer out, 
					 boolean useURL) 
	throws IOException {
	if (info.href == null)
	    // Don't know how to write embedded style sheets.
	    return;

	// Note that text contained in PI does not need to be escaped.

	out.write("<?xml-stylesheet type=");
	char quote = (info.type.indexOf('\"') < 0)? '\"' : '\'';
	out.write(quote);
	out.write(info.type);
	out.write(quote);

	out.write("\nhref=");
	String href = useURL? info.url : info.href;
	quote = (href.indexOf('\"') < 0)? '\"' : '\'';
	out.write(quote);
	out.write(href);
	out.write(quote);

	if (info.title != null) {
	    out.write("\ntitle=");
	    quote = (info.title.indexOf('\"') < 0)? '\"' : '\'';
	    out.write(quote);
	    out.write(info.title);
	    out.write(quote);
	}

	if (info.media != null) {
	    out.write("\nmedia=");
	    quote = (info.media.indexOf('\"') < 0)? '\"' : '\'';
	    out.write(quote);
	    out.write(info.media);
	    out.write(quote);
	}

	if (info.charset != null) {
	    out.write("\ncharset=");
	    quote = (info.charset.indexOf('\"') < 0)? '\"' : '\'';
	    out.write(quote);
	    out.write(info.charset);
	    out.write(quote);
	}

	if (info.alternate)
	    out.write("\nalternate=\"yes\"");

	out.write("\n?>\n");
    }
 
    public static void writeDocumentTypeDeclaration(ElementModel rootElement,
						    Writer out, boolean useURL)
	throws IOException {
	DocumentTypeModel docType = 
	    rootElement.getDocument().getDocumentType();
	if (docType != null) 
	    writeDocumentTypeDeclaration(
		rootElement.getElementTypeName(), 
		docType.getDocumentTypePublicId(),
		(useURL? docType.getDocumentTypeURL() : 
		 docType.getDocumentTypeSystemId()),
		docType.getDocumentTypeInternalSubset(), out);
    }

    public static void writeDocumentTypeDeclaration(String rootElementTypeName,
	                                            String publicId, 
						    String systemId, 
						    String internalSubset, 
						    Writer out) 
	throws IOException {
	out.write("<!DOCTYPE ");
	out.write(rootElementTypeName);

	if (systemId != null) 
	    writeExternalId(publicId, systemId, out);

	if (internalSubset != null) {
	    out.write(" [\n");
	    out.write(internalSubset);
	    out.write("\n]");
	}

	out.write(">\n");
    }

    public static void writeExternalId(String publicId, String systemId, 
				       Writer out) throws IOException {
	if (publicId == null) {
	    out.write(" SYSTEM ");
	} else {
	    // publicId cannot contain double-quotes, nor markup.
	    out.write(" PUBLIC \"");
	    out.write(publicId);
	    out.write("\"\n");
	}

	if (systemId != null) {
	    char quote = (systemId.indexOf('\"') < 0)? '\"' : '\'';

	    out.write(quote);
	    // systemId cannot contain markup.
	    out.write(systemId);
	    out.write(quote);
	}
    }

    public static void writeElement(ElementModel element, Writer out, 
				    int indent, int maxLineLength) 
	throws IOException {
	if (indent >= 0) {
	    out.flush();
	    DocumentTypeModel docType = ElementUtil.getDocumentType(element);
	    doWriteElement(element, new LineWriter(out), indent, 
			   maxLineLength, docType, 
			   ElementUtil.hasPreserveSpace(element, docType), 
			   /*paragraphIndent*/ -1);
	} else {
	    writeElement(element, out);
	}
    }

    // -----------------------------------------------------------------------

    public static void writeElement(ElementModel element, Writer out) 
	throws IOException {
	String tagName = element.getElementTypeName();

	out.write('<');
	out.write(tagName);

	Enumeration iter = element.getElementAttributes();
	while (iter.hasMoreElements()) {
	    String[] attribute = (String[]) iter.nextElement();

	    out.write('\n');
	    out.write(attribute[0]);
	    out.write("=");
	    out.write(XMLUtil.qxml(attribute[1]));
	}

	out.write("\n>");

	SubelementModel child = element.getFirstSubelement();
	while (child != null) {
	    switch (child.getSubelementType()) {
	    case SubelementModel.TEXT:
	        {
		    TextModel textChild = (TextModel) child;

		    StringBuffer buffer = new StringBuffer();
		    XMLUtil.escapeXML(textChild.getTextChars(), 0, 
				      textChild.getTextLength(), buffer);
		    out.write(buffer.toString());
		}
		break;

	    case SubelementModel.ELEMENT:
		writeElement((ElementModel) child, out);
		break;
	    }

	    child = child.getNextSubelement();
	}
	
	out.write("</");
	out.write(tagName);
	out.write("\n>");
    }

    // -----------------------------------------------------------------------

    private static void doWriteElement(ElementModel element, LineWriter out, 
				       int indent, int maxLineLength, 
				       DocumentTypeModel docType, 
				       boolean preserveSpace, 
				       int paragraphIndent) 
	throws IOException {
	String spaceAttr = 
	    ElementUtil.getElementAttribute(element, "xml:space", docType);
	if (spaceAttr != null) 
	    preserveSpace = spaceAttr.equals("preserve");

	if (preserveSpace) {
	    doWritePreformatted(element, out, indent, maxLineLength, docType);
	} else {
	    boolean paragraph;

	    if (paragraphIndent >= 0) {
		paragraph = true;
	    } else {
		if (docType == null) {
		    paragraph = false;

		    SubelementModel child = element.getFirstSubelement();
		    while (child != null) {
			if (child.getSubelementType()==SubelementModel.TEXT) {
			    paragraph = true;
			    break;
			}
			child = child.getNextSubelement();
		    }
		} else {
		    paragraph = ElementUtil.isTextContainer(element, docType);
		}
	    }

	    if (paragraph)
		doWriteParagraph(element, out, indent, maxLineLength, docType, 
				 paragraphIndent);
	    else
		doWriteStructure(element, out, indent, maxLineLength, docType);
	}
    }

    private static void doWritePreformatted(ElementModel element, 
					    LineWriter out, 
					    int indent, int maxLineLength,
					    DocumentTypeModel docType) 
	throws IOException {
	int indentation = out.getColumnNumber();

	String tagName = element.getElementTypeName();
	int tagNameLength = tagName.length();

	out.write('<');
	out.write(tagName);

	Enumeration iter = element.getElementAttributes();
	boolean firstAttr = true;
	while (iter.hasMoreElements()) {
	    String[] attribute = (String[]) iter.nextElement();

	    if (firstAttr) {
		firstAttr = false;
		out.write(' ');
	    } else {
		writeIndent(indentation + 1 + tagNameLength + 1, out);
	    }

	    out.write(attribute[0]);
	    out.write('=');
	    out.write(XMLUtil.qxml(attribute[1]));
	}

	if (element.getSubelementCount() == 0) {
	    writeEmptyElementEndTag(element, out, docType);
	} else {
	    out.write('>');

	    SubelementModel child = element.getFirstSubelement();
	    while (child != null) {
		switch (child.getSubelementType()) {
		case SubelementModel.TEXT:
		    {
			TextModel textChild = (TextModel) child;

			StringBuffer buffer = new StringBuffer();
			XMLUtil.escapeXML(textChild.getTextChars(), 0, 
					  textChild.getTextLength(), buffer);
			out.write(buffer.toString());
		    }
		    break;

		case SubelementModel.ELEMENT:
		    doWriteElement((ElementModel) child, out, indent, 
				   maxLineLength, docType, 
                                   /*preserveSpace*/ true, 
				   /*paragraphIndent*/ -1);
		    break;
		}

		child = child.getNextSubelement();
	    }

	    out.write("</");
	    out.write(tagName);
	    out.write('>');
	}
    }

    /**
     * Output '<br />' not '<br/>'. Output '<p></p>' not '<p/>'.
     * 
     * This is used to follow W3C HTML compatibility guidelines 
     * regarding XHTML. For other DTDs, this causes no harm.
     */
    private static void writeEmptyElementEndTag(ElementModel element, 
						LineWriter out, 
						DocumentTypeModel docType) 
	throws IOException {
	String tagName = element.getElementTypeName();
	ElementTypeModel elemType =
	    ((docType == null)? null : docType.getElementType(tagName));

	boolean emptyContent;
	if (elemType == null)
	    emptyContent = true;
	else
	    emptyContent = 
		(elemType.getElementContentType() == ElementTypeModel.EMPTY);

	if (emptyContent) {
	    out.write(" />");
	} else {
	    out.write("></");
	    out.write(tagName);
	    out.write('>');
	}
    }

    private static void doWriteStructure(ElementModel element, LineWriter out,
					 int indent, int maxLineLength, 
					 DocumentTypeModel docType) 
	throws IOException {
	int indentation = out.getColumnNumber();

	String tagName = element.getElementTypeName();
	int tagNameLength = tagName.length();

	out.write('<');
	out.write(tagName);

	Enumeration iter = element.getElementAttributes();
	boolean firstAttr = true;
	while (iter.hasMoreElements()) {
	    String[] attribute = (String[]) iter.nextElement();

	    if (firstAttr) {
		firstAttr = false;
		out.write(' ');
	    } else {
		writeIndent(indentation + 1 + tagNameLength + 1, out);
	    }

	    out.write(attribute[0]);
	    out.write('=');
	    out.write(XMLUtil.qxml(attribute[1]));
	}

	if (element.getSubelementCount() == 0) {
	    writeEmptyElementEndTag(element, out, docType);
	} else {
	    out.write('>');

	    int contentIndentation = indentation + indent;

	    SubelementModel child = element.getFirstSubelement();
	    while (child != null) {
		writeIndent(contentIndentation, out);

		switch (child.getSubelementType()) {
		case SubelementModel.TEXT:
		    // Certainly an invalid structure.
		    writeText((TextModel) child, out, maxLineLength,
			      contentIndentation);
		    break;

		case SubelementModel.ELEMENT:
		    doWriteElement((ElementModel) child, out, indent, 
				   maxLineLength, docType,
				   /*preserveSpace*/ false, 
				   /*paragraphIndent*/ -1);
		    break;
		}

		child = child.getNextSubelement();
		if (child != null)
		    out.write('\n');
	    }

	    writeIndent(indentation, out);
	    out.write("</");
	    out.write(tagName);
	    out.write('>');
	}
    }

    private static void doWriteParagraph(ElementModel element, LineWriter out,
                                         int indent, int maxLineLength, 
					 DocumentTypeModel docType,
					 int paragraphIndent) 
	throws IOException {
	if (paragraphIndent < 0)
	    paragraphIndent = out.getColumnNumber();

	String tagName = element.getElementTypeName();

	out.write('<');
	out.write(tagName);

	Enumeration iter = element.getElementAttributes();
	while (iter.hasMoreElements()) {
	    String[] attribute = (String[]) iter.nextElement();

	    StringBuffer buffer = new StringBuffer();
	    buffer.append(attribute[0]);
	    buffer.append("=\"");
	    XMLUtil.escapeXML(attribute[1], buffer);
	    buffer.append('\"');

	    if (out.getColumnNumber() + 1 + buffer.length() <= maxLineLength) 
		out.write(' ');
	    else 
		writeIndent(paragraphIndent, out);
	    out.write(buffer.toString());
	}

	if (element.getSubelementCount() == 0) {
	    writeEmptyElementEndTag(element, out, docType);
	} else {
	    out.write('>');

	    SubelementModel child = element.getFirstSubelement();
	    while (child != null) {
		switch (child.getSubelementType()) {
		case SubelementModel.TEXT:
		    writeText((TextModel) child, out, maxLineLength,
			      paragraphIndent);
		    break;

		case SubelementModel.ELEMENT:
		    doWriteElement((ElementModel) child, out, indent, 
                                   maxLineLength, docType,
				   /*preserveSpace*/ false, paragraphIndent);
		    break;
		}

		child = child.getNextSubelement();
	    }

	    out.write("</");
	    out.write(tagName);
	    out.write('>');
	}
    }

    private static void writeIndent(int indent, Writer out) 
	throws IOException {
	out.write('\n');
	while (indent > 0) {
	    out.write(' ');
	    --indent;
	}
    }

    private static void writeText(TextModel text, LineWriter out, 
				  int maxLineLength, int paragraphIndent) 
	throws IOException {
	int textLength = text.getTextLength();
	if (textLength == 0)
	    return;

	StringBuffer buffer = new StringBuffer();
	XMLUtil.escapeXML(text.getTextChars(), 0, textLength, buffer);
	String string = buffer.toString();
	int length = string.length();

	StringTokenizer words = new StringTokenizer(string, " \n\r\t");
	boolean firstWord = true;

	while (words.hasMoreTokens()) {
	    String word = words.nextToken();

	    if (!firstWord || XMLUtil.isXMLSpace(string.charAt(0))) {
		if (out.getColumnNumber() + 1 + word.length() <= maxLineLength)
		    out.write(' ');
		else
		    writeIndent(paragraphIndent, out);
	    }
	    out.write(word);

	    if (firstWord) 
		firstWord = false;
	}

	if (XMLUtil.isXMLSpace(string.charAt(length-1))) {
	    SubelementModel sub = text.getNextSubelement();

	    if (sub == null) {
		// Enough space for ' </parent>'?

		String tagName = text.getParentElement().getElementTypeName();
		if (out.getColumnNumber()+4+tagName.length() <= maxLineLength)
		    out.write(' ');
		else
		    writeIndent(paragraphIndent, out);
	    } else {
		if (sub.getSubelementType() == SubelementModel.TEXT) {
		    out.write(' ');
		} else {
		    // Enough space for ' <next' or ' <next></next>'?

		    ElementModel next = (ElementModel) sub;
		    String tagName = next.getElementTypeName();
		    int needed;

		    if (next.getElementAttributeCount() > 0)
			needed = 2 + tagName.length();
		    else
			needed = 6 + 2*tagName.length();

		    if (out.getColumnNumber() + needed <= maxLineLength)
			out.write(' ');
		    else
			writeIndent(paragraphIndent, out);
		}
	    }
	}
    }
}
