/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import fr.pixware.util.StringUtil;
import fr.pixware.util.HashtableUtil;
import com.xmlmind.xmledit.model.*;

public class ElementAttributeChecker implements Traversal.Handler {
    public static final int INVALID_ATTRIBUTE_NAME     = 0;
    public static final int UNDECLARED_ELEMENT         = 1;
    public static final int UNDECLARED_ATTRIBUTE       = 2;
    public static final int EMPTY_LIST                 = 3;
    public static final int NOT_A_NAME                 = 4;
    public static final int NOT_A_NMTOKEN              = 5;
    public static final int NOT_AN_ENTITY              = 6;
    public static final int OUTSIDE_ENUMERATION        = 7;
    public static final int MISSING_REQUIRED_VALUE     = 8;
    public static final int DIFFERENT_FROM_FIXED_VALUE = 9;
    public static final int DUPLICATE_ID               = 10;
    public static final int NO_MATCHING_ID             = 11;

    public static final int XREF_REASON = DUPLICATE_ID;

    private static final String[] messageIds = {
	"invalidAttributeName",
	"undeclaredElement",
	"undeclaredAttribute",
	"emptyList",
	"notAName",
	"notANmtoken",
	"notAnEntity",
	"outsideEnumeration",
	"missingRequiredValue",
	"differentFromFixedValue",
	"duplicateId",
	"noMatchingId"
    };

    public static class Diagnostic extends ElementCheckerDiagnostic {
	public String invalidAttributeName;
	public String invalidAttributeValue;

	public Diagnostic(int invalidReason, ElementModel invalidElement) {
	    this(invalidReason, invalidElement, null, null);
	}

	public Diagnostic(int invalidReason, ElementModel invalidElement,
			  String invalidAttributeName) {
	    this(invalidReason, invalidElement, invalidAttributeName, null);
	}

	public Diagnostic(int invalidReason, ElementModel invalidElement,
			  String invalidAttributeName, 
			  String invalidAttributeValue) {
	    super(invalidReason, invalidElement, null);

	    this.invalidAttributeName = invalidAttributeName;
	    this.invalidAttributeValue = invalidAttributeValue;

	    message = msg(messageIds[invalidReason], 
			  invalidElement.getElementTypeName(), 
			  invalidAttributeName, invalidAttributeValue);
	}
    }

    private DocumentModel doc;
    private DocumentTypeModel docType;
    private Hashtable elemTypeInfo = new Hashtable();

    private Hashtable idToElementMap;
    private Vector diagnostics;
    private int pass;

    private static class ElementTypeInfo {
	public ElementTypeModel elementType;
	public String[] requiredAttributeNames;
    }

    // -----------------------------------------------------------------------

    public ElementAttributeChecker() {
	this(null);
    }

    public ElementAttributeChecker(DocumentModel document) {
	setDocument(document);
    }

    public void setDocument(DocumentModel document) {
	doc = document;
	if (doc == null)
	    docType = null;
	else
	    docType = doc.getDocumentType();

	if (docType == null) {
	    elemTypeInfo.clear();
	} else {
	    Enumeration iter = docType.getElementTypes();
	    while (iter.hasMoreElements()) {
		ElementTypeModel elemType = 
		    (ElementTypeModel) iter.nextElement();
                // Note that UNDECLARED is treated as a first class type.

		ElementTypeInfo info = new ElementTypeInfo();
		info.elementType = elemType;

		int attrDefCount = elemType.getAttributeDefinitionCount();
		if (attrDefCount == 0) {
		    info.requiredAttributeNames = new String[0];
		} else {
		    String[] list = new String[attrDefCount];
		    int listSize = 0;

		    Enumeration iter2 = elemType.getAttributeDefinitions();
		    while (iter2.hasMoreElements()) {
			AttributeDefinitionModel attrDef = 
			    (AttributeDefinitionModel) iter2.nextElement();

			if (attrDef.getAttributePresence() == 
			    AttributeDefinitionModel.REQUIRED)
			    list[listSize++] = attrDef.getAttributeName();
		    }

		    if (listSize == list.length) {
			info.requiredAttributeNames = list;
		    } else {
			info.requiredAttributeNames = new String[listSize];
			System.arraycopy(list, 0, info.requiredAttributeNames,
					 0, listSize);
		    }
		}

		elemTypeInfo.put(elemType.getElementTypeName(), info);
	    }
	}
    }

    public DocumentModel getDocument() {
	return doc;
    }

    public Diagnostic[] check(ElementModel element) {
	Vector diagnostics = new Vector();

	check(element, /*xref*/ true, diagnostics);

	Diagnostic[] result = new Diagnostic[diagnostics.size()];
	diagnostics.copyInto(result);
	return result;
    }

    public void check(ElementModel element, boolean xref, Vector diagnostics) {
	ElementTypeInfo info = 
	    (ElementTypeInfo) elemTypeInfo.get(element.getElementTypeName());
	if (info == null) {
	    diagnostics.addElement(
		new Diagnostic(UNDECLARED_ELEMENT, element));
	    return;
	}

	String[] reqAttrNames = info.requiredAttributeNames;
	if (reqAttrNames.length == 0 &&
	    element.getElementAttributeCount() == 0)
	    return;
	ElementTypeModel elemType = info.elementType;

	// Force idToElementMap to be rebuilt if needed to.
	idToElementMap = null;

	Enumeration iter = element.getElementAttributes();
	while (iter.hasMoreElements()) {
	    String[] pair = (String[]) iter.nextElement();

	    checkAttribute(elemType, element, pair[0], pair[1], xref, 
			   diagnostics);
	}

	// We no longer need idToElementMap.
	idToElementMap = null;

	for (int i = 0; i < reqAttrNames.length; ++i) {
	    if (element.getElementAttribute(reqAttrNames[i]) == null)
		diagnostics.addElement(
		    new Diagnostic(MISSING_REQUIRED_VALUE, element, 
				   reqAttrNames[i]));
	}
    }

    public Diagnostic checkAttribute(ElementModel element, 
				     String name, String value) {
	ElementTypeInfo info = 
	    (ElementTypeInfo) elemTypeInfo.get(element.getElementTypeName());
	if (info == null)
	    return new Diagnostic(UNDECLARED_ELEMENT, element);

	// Force idToElementMap to be rebuilt if needed to.
	idToElementMap = null;

	Vector diagnostics = new Vector();
	checkAttribute(info.elementType, element, name, value, /*xref*/ true, 
		       diagnostics);

	// We no longer need idToElementMap.
	idToElementMap = null;

	if (diagnostics.size() > 0)
	    return (Diagnostic) diagnostics.elementAt(0);
	else
	    return null;
    }

    public void checkAttribute(ElementTypeModel elementType, 
			       ElementModel element, String name, String value,
			       boolean xref, Vector diagnostics) {
	AttributeDefinitionModel attrDef = 
	    elementType.getAttributeDefinition(name);
	if (attrDef == null) {
	    diagnostics.addElement(
		new Diagnostic(UNDECLARED_ATTRIBUTE, element, name));
	    return;
	}

	String id = null;
	String[] idRefs = null;

	// Value is assumed to be normalized.

	switch (attrDef.getAttributeType()) {
	case AttributeDefinitionModel.ID:
	    if (XMLUtil.isName(value)) {
		id = value;
	    } else {
		diagnostics.addElement(
		    new Diagnostic(NOT_A_NAME, element, name, value));
	    }
	    break;
	case AttributeDefinitionModel.IDREF:
	    if (XMLUtil.isName(value)) {
		idRefs = new String[] { value };
	    } else {
		diagnostics.addElement(
		    new Diagnostic(NOT_A_NAME, element, name, value));
	    }
	    break;
	case AttributeDefinitionModel.IDREFS:
	    if (value.length() == 0) {
		diagnostics.addElement(
		    new Diagnostic(EMPTY_LIST, element, name));
	    } else {
		boolean errors = false;

		idRefs = StringUtil.split(value, ' ');
		for (int i = 0; i < idRefs.length; ++i) {
		    if (!XMLUtil.isName(idRefs[i])) {
			diagnostics.addElement(
			    new Diagnostic(NOT_A_NAME, element, 
					   name, idRefs[i]));
			errors = true;
		    }
		}

		if (errors)
		    idRefs = null;
	    }
	    break;
	case AttributeDefinitionModel.ENTITY:
	    if (XMLUtil.isName(value)) {
		if (docType.getUnparsedEntity(value) == null) 
		    diagnostics.addElement(
			new Diagnostic(NOT_AN_ENTITY, element, name, value));
	    } else {
		diagnostics.addElement(
		    new Diagnostic(NOT_A_NAME, element, name, value));
	    }
	    break;
	case AttributeDefinitionModel.ENTITIES:
	    if (value.length() == 0) {
		diagnostics.addElement(
		    new Diagnostic(EMPTY_LIST, element, name));
	    } else {
		String[] list = StringUtil.split(value, ' ');

		for (int i = 0; i < list.length; ++i) {
		    if (XMLUtil.isName(list[i])) {
			if (docType.getUnparsedEntity(list[i]) == null) 
			    diagnostics.addElement(
				new Diagnostic(NOT_AN_ENTITY, element, 
					       name, list[i]));
		    } else {
			diagnostics.addElement(
			    new Diagnostic(NOT_A_NAME, element, 
					   name, list[i]));
		    }
		}
	    }
	    break;
	case AttributeDefinitionModel.NMTOKEN:
	    if (!XMLUtil.isNmtoken(value)) {
		diagnostics.addElement(
		    new Diagnostic(NOT_A_NMTOKEN, element, name, value));
	    }
	    break;
	case AttributeDefinitionModel.NMTOKENS:
	    if (value.length() == 0) {
		diagnostics.addElement(
		    new Diagnostic(EMPTY_LIST, element, name));
	    } else {
		String[] list = StringUtil.split(value, ' ');

		for (int i = 0; i < list.length; ++i) {
		    if (!XMLUtil.isNmtoken(list[i]))
			diagnostics.addElement(
			    new Diagnostic(NOT_A_NMTOKEN, element, 
					   name, list[i]));
		}
	    }
	    break;
	case AttributeDefinitionModel.NOTATION:
	    // The fact that the value should be a declared NOTATION
	    // should be checked once for all at the DTD level.
	case AttributeDefinitionModel.ENUMERATION:
	    {
		String[] list = attrDef.getAttributeEnumeratedValues();
		boolean found = false;

		for (int i = 0; i < list.length; ++i) {
		    if (value.equals(list[i])) {
			found = true;
			break;
		    }
		}

		if (!found) {
		    diagnostics.addElement(
			new Diagnostic(OUTSIDE_ENUMERATION, element, 
				       name, value));
		}
	    }
	    break;

	// CDATA: nothing to ckeck.
	}

	if (attrDef.getAttributePresence() == AttributeDefinitionModel.FIXED &&
	    !value.equals(attrDef.getAttributeDefaultValue())) {
	    diagnostics.addElement(
		new Diagnostic(DIFFERENT_FROM_FIXED_VALUE, element, 
			       name, value));
	}

	if (xref && (id != null || idRefs != null)) {
	    if (idToElementMap == null)
		idToElementMap = getIdToElementMap();
	    
	    if (id == null) {
		for (int i = 0; i < idRefs.length; ++i) {
		    if (idToElementMap.get(idRefs[i]) == null) {
			diagnostics.addElement(
			    new Diagnostic(NO_MATCHING_ID, element, name, 
					   idRefs[i]));
		    }
		}
	    } else {
		ElementModel mapped = (ElementModel) idToElementMap.get(id);
		if (mapped != null && mapped != element)
		    diagnostics.addElement(
			new Diagnostic(DUPLICATE_ID, element, name, id));
	    }
	}
    }

    public Diagnostic[] checkXRef() {
	Vector diagnostics = new Vector();

	checkXRef(diagnostics);

	Diagnostic[] result = new Diagnostic[diagnostics.size()];
	diagnostics.copyInto(result);
	return result;
    }

    public void checkXRef(Vector diagnostics) {
	if (docType != null) {
	    this.diagnostics = diagnostics;
	    idToElementMap = new Hashtable();

	    pass = 1;
	    ElementModel root = doc.getRootElement();
	    Traversal.traverse(root, this);

	    pass = 2;
	    Traversal.traverse(root, this);
	
	    // We no longer need idToElementMap.
	    idToElementMap = null;
	}
    }

    public Hashtable getIdToElementMap() {
	// Don't need diagnostics.
	diagnostics = null;
	idToElementMap = new Hashtable();

	pass = 1;
	Traversal.traverse(doc.getRootElement(), this);

	Hashtable result = idToElementMap;
	idToElementMap = null;
	return result;
    }

    public static Diagnostic checkAttributeName(ElementModel element, 
						String name) {
	if (XMLUtil.isName(name))
	    return null;
	else
	    return new ElementAttributeChecker.Diagnostic(
		INVALID_ATTRIBUTE_NAME, element, name);
    }

    // -----------------------------------------------------------------------

    public Object enterElement(ElementModel element) { 
	if (element.getElementAttributeCount() == 0)
	    // Optimization: most elements have no specified attributes.
	    return null;

	ElementTypeInfo info = 
	    (ElementTypeInfo) elemTypeInfo.get(element.getElementTypeName());
	if (info == null) 
	    return null;
	ElementTypeModel elemType = info.elementType;

	if (pass == 1) {
	    Enumeration iter = element.getElementAttributes();
	    while (iter.hasMoreElements()) {
		String[] pair = (String[]) iter.nextElement();

		AttributeDefinitionModel attrDef = 
		    elemType.getAttributeDefinition(pair[0]);
		if (attrDef != null &&
		    attrDef.getAttributeType()==AttributeDefinitionModel.ID) {
		    if (idToElementMap.put(pair[1], element) != null && 
			diagnostics != null)
			diagnostics.addElement(
			    new Diagnostic(DUPLICATE_ID, element, 
					   pair[0], pair[1]));

		    // For invalid document: ignore other attributes of type
		    // ID.
		    break;
		}
	    }
	} else {
	    Enumeration iter = element.getElementAttributes();
	    while (iter.hasMoreElements()) {
		String[] pair = (String[]) iter.nextElement();

		AttributeDefinitionModel attrDef = 
		    elemType.getAttributeDefinition(pair[0]);
		if (attrDef == null)
		    continue;

		switch (attrDef.getAttributeType()) {
		case AttributeDefinitionModel.IDREF:
		    if (idToElementMap.get(pair[1]) == null) {
			diagnostics.addElement(
			    new Diagnostic(NO_MATCHING_ID, element, 
					   pair[0], pair[1]));
		    }
		    break;
		case AttributeDefinitionModel.IDREFS:
		    {
			String[] idRefs = StringUtil.split(pair[1], ' ');

			for (int i = 0; i < idRefs.length; ++i) {
			    if (idToElementMap.get(idRefs[i]) == null) {
				diagnostics.addElement(
				    new Diagnostic(NO_MATCHING_ID, element, 
						   pair[0], idRefs[i]));
			    }
			}
		    }
		    break;
		}
	    }
	}

	return null; 
    }

    public Object leaveElement(ElementModel element) { return null; }
    public Object processText(TextModel text) { return null; }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmledit.modeltool.ElemAttrCheckerRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id, 
			      Object arg0, Object arg1, Object arg2) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}

	Object[] args;
	if (arg0 == null) {
	    args = new Object[0];
	} else {
	    if (arg1 == null) {
		args = new Object[] { arg0 };
	    } else {
		if (arg2 == null) {
		    args = new Object[] { arg0, arg1 };
		} else {
		    args = new Object[] { arg0, arg1, arg2 };
		}
	    }
	}

	return MessageFormat.format(message, args);
    }
}
