/*
 * Copyright (c) 2001-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.Vector;
import fr.pixware.util.*;
import com.xmlmind.xmledit.model.*;

public class ElementCharSequence implements CharSequence {
    public static final int MARK_NONE = 0;
    public static final int MARK_ALL  = 1;
    public static final int MARK_LAST = 2;

    public static final String DEFAULT_EMPTY_TEXT_CONTENT = "???";
    public static final String DEFAULT_PARAGRAPH_MARK = "\n\n";

    private String emptyTextContent;
    private int emptyTextContentLength;
    private int markParagraphs;
    private String paragraphMark;
    private Entry[] entries;
    private int length;

    private Entry cachedEntry;
    private int cachedEntryFirst, cachedEntryEnd;

    private static final class Entry extends TextRange {
	public int index;

	public Entry() {
	    super();
	}

	public Entry(int index, TextModel text, int offset, int count) {
	    super(text, offset, count);
	    this.index = index;
	}
    }

    // -----------------------------------------------------------------------

    public ElementCharSequence(ElementModel element) {
	this(element, null, null, 
	     DEFAULT_EMPTY_TEXT_CONTENT, MARK_LAST, DEFAULT_PARAGRAPH_MARK);
    }

    public ElementCharSequence(ElementModel element, String emptyTextContent,
			       int markParagraphs, String paragraphMark) {
	this(element, null, null, 
	     emptyTextContent, markParagraphs, paragraphMark);
    }

    public ElementCharSequence(TextOffset begin, TextOffset end) {
	this(null, begin, end, 
	     DEFAULT_EMPTY_TEXT_CONTENT, MARK_LAST, DEFAULT_PARAGRAPH_MARK);
    }

    public ElementCharSequence(TextOffset begin, TextOffset end, 
			       String emptyTextContent,
			       int markParagraphs, String paragraphMark) {
	this(null, begin, end, 
	     emptyTextContent, markParagraphs, paragraphMark);
    }

    public ElementCharSequence(ElementModel element,
			       TextOffset begin, TextOffset end, 
			       String emptyTextContent,
			       int markParagraphs, String paragraphMark) {
	if (emptyTextContent == null || emptyTextContent.length() == 0)
	    emptyTextContent = DEFAULT_EMPTY_TEXT_CONTENT;
	this.emptyTextContent = emptyTextContent;
	emptyTextContentLength = emptyTextContent.length();

	this.markParagraphs = markParagraphs;

	if (paragraphMark == null || paragraphMark.length() == 0)
	    paragraphMark = DEFAULT_PARAGRAPH_MARK;
	this.paragraphMark = paragraphMark;
	final int paragraphMarkLength = paragraphMark.length();

	if (element == null) {
	    if (begin == null) {
		if (end != null)
		    element = end.text.getDocument().getRootElement();
	    } else {
		element = begin.text.getDocument().getRootElement();
	    }
	}
	if (element == null)
	    throw new IllegalArgumentException("unspecified text range");

	final DocumentTypeModel docType = ElementUtil.getDocumentType(element);

	if (begin == null) {
	    TextModel firstText = 
		(TextModel) Traversal.traverse(element, Traversal.textFinder);
	    if (firstText != null)
		begin = new TextOffset(firstText, 0);
	}
	if (begin == null) {
	    // Element contains no text.
	    entries = new Entry[0];
	    length = 0;

	    return;
	}

	if (end == null) {
	    end = new TextOffset();
	    end.text = (TextModel) 
		Traversal.traverseBackward(element, Traversal.textFinder);
	    // Element contains text: at least begin.text.
	    end.offset = end.text.getTextLength();
	}

	if (begin.text == end.text) {
	    length = end.offset - begin.offset;
	    if (length > 0) {
		Entry entry = new Entry(0, begin.text, begin.offset, length);
		entries = new Entry[1];
		entries[0] = entry;
	    } else {
		entries = new Entry[0];
		length = 0;
	    }

	    return;
	}

	final Vector entryList = new Vector();

	int charCount = begin.text.getTextLength();
	if (charCount == 0) {
	    entryList.addElement(new Entry(0, begin.text, 0, 0));
	    length = emptyTextContentLength;

	    markAsParagraph(begin.text, docType, paragraphMarkLength);
	} else {
	    charCount -= begin.offset;
	    if (charCount > 0) {
		entryList.addElement(
		    new Entry(0, begin.text, begin.offset, charCount));
		length = charCount;

		markAsParagraph(begin.text, docType, paragraphMarkLength);
	    } else {
		// Useless empty begin.
		length = 0;
	    }
	}

	final TextModel stop = end.text;
	TextModel found = (TextModel)
	    Traversal.traverseAfter(begin.text, new Traversal.HandlerBase() {
	        public Object processText(TextModel text) {
		    if (text == stop)
			return stop;

		    int count = text.getTextLength();
		    entryList.addElement(new Entry(length, text, 0, count));
		    if (count == 0)
			length += emptyTextContentLength;
		    else
			length += count;

		    markAsParagraph(text, docType, paragraphMarkLength);

		    return null;
		}
	    });

	if (found == null) {
	    // Invalid text range: end.text before begin.text.
	    entries = new Entry[0];
	    length = 0;

	    return;
	}

	charCount = end.text.getTextLength();
	if (charCount == 0) {
	    entryList.addElement(new Entry(length, end.text, 0, 0));
	    length += emptyTextContentLength;

	    // No need to mark this end.text as a paragraph.
	} else {
	    charCount = Math.min(end.offset, charCount);
	    if (charCount > 0) {
		entryList.addElement(new Entry(length, end.text,0,charCount));
		length += charCount;

		// No need to mark this end.text as a paragraph.
	    }
	    // Otherwise useless empty end.
	}

	entries = new Entry[entryList.size()];
	entryList.copyInto(entries);
    }

    private void markAsParagraph(TextModel text, DocumentTypeModel docType,
				 int paragraphMarkLength) {
	switch (markParagraphs) {
	case MARK_ALL:
	    length += paragraphMarkLength;
	    break;

	case MARK_LAST:
	    {
		// Mark the very last #PCDATA of a text container hierarchy or
		// mark last #PCDATA of a text container hierarchy followed by
		// a non text container (ex. <br/>).

		boolean append = false;

		ElementModel textContainer = text.getParentElement();
		if (text == textContainer.getLastSubelement()) {
		    ElementModel parent = textContainer.getParentElement();

		    append = true;
		    while (parent != null && isTextContainer(parent,docType)) {
			if (textContainer != parent.getLastSubelement()) {
			    SubelementModel next = 
				textContainer.getNextSubelement();

			    append = (next.getSubelementType() == 
				      SubelementModel.ELEMENT &&
				      !isTextContainer((ElementModel) next,
						       docType));
			    break;
			}

			textContainer = parent;
			parent = textContainer.getParentElement();
		    }
		} else {
		    SubelementModel next = text.getNextSubelement();

		    append = 
			(next.getSubelementType() == SubelementModel.ELEMENT &&
			 !isTextContainer((ElementModel) next, docType));
		}

		if (append)
		    length += paragraphMarkLength;
	    }
	    break;
	}
    }

    private static final boolean isTextContainer(ElementModel element,
						 DocumentTypeModel docType) {
	if (docType != null) 
	    return ElementUtil.isTextContainer(element, docType);

	// No document type: simple heuristic: if element contains text, it is
	// a text container.

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) 
		return true;
	    sub = sub.getNextSubelement();
	}

	return false;
    }

    // -----------------------------------------------------------------------

    public int length() {
	return length;
    }

    public char charAt(int index) {
	Entry entry = findEntry(index);
	if (entry == null)
	    throw new IndexOutOfBoundsException(Integer.toString(index));

	int offset = entry.offset;
	int count = entry.count;

	int pos = offset + (index - entry.index);
	int end = offset + count;
	if (pos < end)
	    return entry.text.getTextChar(pos);
	
	if (count == 0) {
	    pos -= offset;
	    if (pos < emptyTextContentLength)
		return emptyTextContent.charAt(pos);
	    else
		return paragraphMark.charAt(pos - emptyTextContentLength);
	} else {
	    return paragraphMark.charAt(pos - end);
	}
    }

    public CharSequence subSequence(int start, int end) {
	TextOffset beginOffset = getTextOffset(start);
	TextOffset endOffset = getTextOffset(end);

	return new ElementCharSequence(beginOffset, endOffset, 
				       emptyTextContent,
				       markParagraphs, paragraphMark);
    }

    public String toString() {
	StringBuffer buffer = new StringBuffer();

	int lastEntry = entries.length - 1;
	for (int i = 0; i <= lastEntry; ++i) {
	    Entry entry = entries[i];

	    int endIndex;
	    if (entry.count == 0) {
		buffer.append(emptyTextContent);
		endIndex = entry.index + emptyTextContentLength;
	    } else {
		char[] chars = entry.text.getTextChars();
		buffer.append(chars, entry.offset, entry.count);
		endIndex = entry.index + entry.count;
	    }

	    int nextIndex = (i < lastEntry)? entries[i+1].index : length;
	    if (endIndex < nextIndex)
		buffer.append(paragraphMark);
	}

	return buffer.toString();
    }

    // -----------------------------------------------------------------------

    public TextOffset getTextOffset(int index) {
	Entry entry;
	if (index == length) { 
	    // charAt(length) is illegal. getTextOffset(length) is OK.
	    if (entries.length > 0) 
		entry = entries[entries.length-1];
	    else
		entry = null;
	} else {
	    entry = findEntry(index);
	}
	if (entry == null)
	    throw new IndexOutOfBoundsException(Integer.toString(index));

	int offset = entry.offset;
	int count = entry.count;

	int pos = offset + (index - entry.index);
	int end = offset + count;
	if (pos > end)
	    pos = end;

	return new TextOffset(entry.text, pos);
    }

    private Entry findEntry(int index) {
	if (cachedEntry != null) {
	    if (index >= cachedEntryFirst && index < cachedEntryEnd)
		return cachedEntry;

	    cachedEntry = null;
	    cachedEntryFirst = cachedEntryEnd = -1;
	}

	int begin = 0;
	int end = entries.length - 1;

	while (begin <= end) {
	    int middle = begin + ((end - begin) / 2);
	    Entry middleEntry = entries[middle];
	    int middleEntryFirst = middleEntry.index;
	    int middleEntryEnd = 
		(middle < entries.length-1)? entries[middle+1].index : length;

	    if (index >= middleEntryFirst && index < middleEntryEnd) {
		cachedEntry = middleEntry;
		cachedEntryFirst = middleEntryFirst;
		cachedEntryEnd = middleEntryEnd;

		return middleEntry;
	    } else if (index < middleEntryFirst) {
		end = middle - 1;
	    } else {
		begin = middle + 1;
	    }
	}
	
	return null;
    }
}

