/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.Stack;
import java.util.EmptyStackException;
import java.io.*;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import com.xmlmind.xmledit.model.*;

public class ElementParser extends DefaultHandler {
    public static final String DEFAULT_PARSER_CLASS_NAME = 
        "com.jclark.xml.sax.SAX2Driver";

    private XMLReader parser;

    private StringBuffer buffer = new StringBuffer();
    private Stack stack = new Stack();
    private DocumentModel document;
    private DocumentTypeModel documentType;
    private ElementModel rootElement;

    // -----------------------------------------------------------------------

    public ElementParser()
	throws ClassNotFoundException, 
	       InstantiationException, IllegalAccessException {

	this(DEFAULT_PARSER_CLASS_NAME);
    }

    public ElementParser(String parserClassName) 
	throws ClassNotFoundException, 
	       InstantiationException, IllegalAccessException {

	parser = (XMLReader) Class.forName(parserClassName).newInstance();
	try {
	    parser.setFeature("http://xml.org/sax/features/namespaces", false);
	} catch (SAXException cannotHappen) {}
	parser.setContentHandler(this);
	parser.setEntityResolver(this);
    }

    public ElementModel parse(DocumentModel doc, String s) 
	throws SAXException, IOException {
	ElementModel parsedElement = null;

	document = doc;
	documentType = doc.getDocumentType();
	try {
	    parser.parse(new InputSource(new StringReader(s)));
	    parsedElement = rootElement;
	} finally {
	    popAll();
	    buffer = new StringBuffer();
	    rootElement = null;
	    document = null;
	    documentType = null;
	}

	return parsedElement;
    }

    private void popAll() {
	while (!stack.empty())
	    stack.pop();
    }

    private ElementModel top() {
	return (stack.empty()? null : (ElementModel) stack.peek());
    }

    public InputSource resolveEntity(String publicId, String systemId) 
	throws SAXException {
	// Ignore all external entities.
	return new InputSource(new StringReader("<?xml encoding='UTF-8'?>"));
    }

    public void startElement(String namespaceURI, String localName,
			     String qName, Attributes atts) 
	throws SAXException {
	ElementModel top = top();

	if (buffer.length() > 0) {
	    top.appendSubelement(document.createTextModel(buffer.toString()));
	    buffer = new StringBuffer();
	}

	ElementModel pushed = document.createElementModel(qName);   
	for (int i = 0; i < atts.getLength(); i++) 
	    pushed.putElementAttribute(atts.getQName(i), atts.getValue(i));
	if (top != null)
	    top.appendSubelement(pushed);

	stack.push(pushed);
    }

    public void endElement(String namespaceURI, String localName,
			   String qName) throws SAXException {
	ElementModel top = top();

	if (buffer.length() > 0) {
	    top.appendSubelement(document.createTextModel(buffer.toString()));
	    buffer = new StringBuffer();
	} else {
	    // If top is an empty mix, insert a Text in it.
	    if (top.getSubelementCount() == 0 && 
		ElementUtil.hasMixedContent(top, documentType)) 
		top.appendSubelement(document.createTextModel(""));
	}

	stack.pop();
	if (stack.empty())
	    rootElement = top;
    } 

    public void characters(char ch[], int start, int length) 
	throws SAXException {
	buffer.append(ch, start, length);
    }

    /*************************************************************************

    public static void main(String[] args) throws Exception {
	if (args.length != 2) {
	    System.err.println(
		"usage: java com.xmlmind.xmledit.modeltool.ElementParser" + 
		" doc_xml element_xml");
	    System.exit(1);
	}

	com.xmlmind.xmledit.modelimpl.Document document = 
	    new com.xmlmind.xmledit.modelimpl.Document();
	(new com.xmlmind.xmledit.modelimpl.XPDocumentLoader()).load(args[0], 
								    document);

	String content = fr.pixware.util.FileUtil.loadString(args[1]);

	ElementModel parsedElement = 
	    (new ElementParser()).parse(document, content);

	System.out.println(parsedElement.getElementTypeName());
    }

    *************************************************************************/
}

