/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import fr.pixware.util.*;
import com.xmlmind.namere.*;
import com.xmlmind.xmledit.model.*;

public class ElementStructureChecker {
    public static final int UNDECLARED_ELEMENT      = 0;
    public static final int ELEMENT_NOT_EMPTY       = 1;
    public static final int UNEXPECTED_PCDATA       = 2;
    public static final int UNDECLARED_CHILD        = 3;
    public static final int INVALID_CHILD           = 4;
    public static final int INVALID_ELEMENT_CONTENT = 5;

    private static final String[] messageIds = {
	"undeclaredElement",
	"elementNotEmpty",
	"unexpectedPCDATA",
	"undeclaredChild",
	"invalidChild",
	"invalidElementContent"
    };

    public static class Diagnostic extends ElementCheckerDiagnostic {
	public SubelementModel invalidSubelement;

	public Diagnostic(int invalidReason, ElementModel invalidElement) {
	    this(invalidReason, invalidElement, null);
	}

	public Diagnostic(int invalidReason, ElementModel invalidElement, 
			  SubelementModel invalidSubelement) {
	    super(invalidReason, invalidElement, null);

	    this.invalidSubelement = invalidSubelement;

	    String childTypeName;
	    if (invalidSubelement == null) {
		childTypeName = null;
	    } else {
		if (invalidSubelement.getSubelementType() == 
		    SubelementModel.TEXT) 
		    childTypeName = "#PCDATA";
		else
		    childTypeName = 
			((ElementModel)invalidSubelement).getElementTypeName();
	    }
	    message = msg(messageIds[invalidReason], 
			  invalidElement.getElementTypeName(), 
			  childTypeName);
	}
    }

    public static final class ElementContentBuffer {
	private String[] names = new String[10];
	private int nameCount = 0;

	public void clear() {
	    nameCount = 0;
	}
	
	public boolean append(SubelementModel sub) {
	    if (sub.getSubelementType() == SubelementModel.TEXT)
		return false;
	    return append(((ElementModel) sub).getElementTypeName());
	}
	
	public boolean append(ElementModel element) {
	    return append(element.getElementTypeName());
	}

	public boolean append(String elemTypeName) {
	    if (nameCount + 1 > names.length) {
		String[] names2 = new String[2*names.length];
		System.arraycopy(names, 0, names2, 0, names.length);
		names = names2;
	    }
	    names[nameCount++] = elemTypeName.intern();
	    return true;
	}
    }

    private REFactory reFactory = new REFactoryImpl();
    private DocumentModel doc;
    private DocumentTypeModel docType;
    private Hashtable elemContentMatchers;
    private Hashtable mixedContentChildTypeNames;
    private Hashtable elementContentChildTypeNames;
    private String[] declaredElemTypeNames;
    private String[] declaredTextContainerNames;

    // -----------------------------------------------------------------------

    public ElementStructureChecker() {
	this(null);
    }

    public ElementStructureChecker(DocumentModel document) {
	setDocument(document);
    }

    public void setDocument(DocumentModel document) {
	doc = document;

	docType = null;
	elemContentMatchers = null;
	mixedContentChildTypeNames = null;
	elementContentChildTypeNames = null;
	declaredElemTypeNames = null;
	declaredTextContainerNames = null;

	if (doc != null) {
	    docType = doc.getDocumentType();
	    if (docType == null) 
		return;

	    elemContentMatchers = new Hashtable();
	    mixedContentChildTypeNames = new Hashtable();
	    elementContentChildTypeNames = new Hashtable();

	    Vector elemTypeNames = new Vector();
	    Vector textContainerNames = new Vector();
	    Enumeration iter = docType.getElementTypes();

	    while (iter.hasMoreElements()) {
		ElementTypeModel elemType = 
		    (ElementTypeModel) iter.nextElement();
		String elemTypeName = elemType.getElementTypeName();
		int contentType = elemType.getElementContentType();

		switch (contentType) {
		case ElementTypeModel.MIXED:
		case ElementTypeModel.ANY:
		    textContainerNames.addElement(elemTypeName);
		    /*FALLTHROUGH*/
		default:
		    if (contentType != ElementTypeModel.UNDECLARED)
			elemTypeNames.addElement(elemTypeName);
		}
	    }

	    declaredElemTypeNames = new String[elemTypeNames.size()];
	    elemTypeNames.copyInto(declaredElemTypeNames);

	    declaredTextContainerNames = new String[textContainerNames.size()];
	    textContainerNames.copyInto(declaredTextContainerNames);
	}
    }

    public DocumentModel getDocument() {
	return doc;
    }

    public Diagnostic check(ElementModel element) {
	String elemTypeName = element.getElementTypeName();
	ElementTypeModel elemType;
	if (docType == null ||
	    (elemType = docType.getElementType(elemTypeName)) == null)
	    return new Diagnostic(UNDECLARED_ELEMENT, element);

	switch (elemType.getElementContentType()) {
	case ElementTypeModel.EMPTY:
	    return ((element.getSubelementCount() == 0)? null :
		    new Diagnostic(ELEMENT_NOT_EMPTY, element));

	case ElementTypeModel.ANY:
	    {
		String[] list = getDeclaredElementTypeNames();

		SubelementModel sub = element.getFirstSubelement();
		while (sub != null) {
                    // ANY means any declared type or #PCDATA.

		    if (sub.getSubelementType() != SubelementModel.TEXT &&
			!StringUtil.contains(
			    list, ((ElementModel) sub).getElementTypeName()))
			return new Diagnostic(UNDECLARED_CHILD, element, sub);

		    sub = sub.getNextSubelement();
		}

		return null;
	    }

	case ElementTypeModel.MIXED:
	    {
		String[] list = getMixedContentChildTypeNames(elemTypeName);

		SubelementModel sub = element.getFirstSubelement();
		while (sub != null) {
		    if (sub.getSubelementType() != SubelementModel.TEXT &&
			!StringUtil.contains(
			    list, ((ElementModel) sub).getElementTypeName()))
			return new Diagnostic(INVALID_CHILD, element, sub);

		    sub = sub.getNextSubelement();
		}

		return null;
	    }

	case ElementTypeModel.ELEMENT:
	    {
		ElementContentBuffer elementContent = 
		    new ElementContentBuffer();
		
		SubelementModel sub = element.getFirstSubelement();
		while (sub != null) {
		    if (!elementContent.append(sub))
			return new Diagnostic(UNEXPECTED_PCDATA, element, sub);
			
		    sub = sub.getNextSubelement();
		}

		NameRE re = getElementContentMatcher(elemTypeName);
		if (re.match(elementContent.names, elementContent.nameCount))
		    return null;
		else
		    return new Diagnostic(INVALID_ELEMENT_CONTENT, element);
	    }

	case ElementTypeModel.UNDECLARED:
	default:
	    return new Diagnostic(UNDECLARED_ELEMENT, element);
	}
    }

    private NameRE getElementContentMatcher(String elemTypeName) {
	NameRE re = (NameRE) elemContentMatchers.get(elemTypeName);

	if (re == null) {
	    ElementTypeModel elemType = docType.getElementType(elemTypeName);
	    if (elemType == null)
		return null;
	    ElementContentModel contentModel = elemType.getElementContent();
	    if (contentModel == null)
		return null;

	    Particle particle = contentModelToParticle(contentModel);

	    try {
		re = new NameRE(particle);
	    } catch (NameRE.InvalidParticleException e) {
		docType = null;
		throw new RuntimeException(
		    "internal error: cannot convert the content model of '" + 
		    elemTypeName + "' to a NameRE: " + e.getMessage());
	    }

	    elemContentMatchers.put(elemTypeName, re);
	}

	return re;
    }

    private static final
    Particle contentModelToParticle(ElementContentModel contentModel) {
	int listType = contentModel.getListType();
	int minOccurs = contentModel.getMinOccurences();
	int maxOccurs = contentModel.getMaxOccurences();

	if (listType == ElementContentModel.NONE) {
	    String elemTypeName = 
		contentModel.getElementType().getElementTypeName();
	    return new ElementParticle(elemTypeName, minOccurs, maxOccurs);
	}

	ElementContentModel[] list = contentModel.getList();
	Particle[] particles = new Particle[list.length];

	for (int i = 0; i < list.length; ++i) 
	    particles[i] = contentModelToParticle(list[i]);

	if (listType == ElementContentModel.CHOICE)
	    return new ChoiceParticle(particles, minOccurs, maxOccurs);
	else
	    return new SequenceParticle(particles, minOccurs, maxOccurs);
    }

    // -----------------------------------------------------------------------
    // ElementStructureEditor helpers
    // -----------------------------------------------------------------------

    public String[] getDeclaredElementTypeNames() {
	return declaredElemTypeNames;
    }

    public String[] getDeclaredTextContainerNames() {
	return declaredTextContainerNames;
    }

    public String[] getMixedContentChildTypeNames(ElementModel element) {
	return getMixedContentChildTypeNames(element.getElementTypeName());
    }

    public String[] getMixedContentChildTypeNames(String elemTypeName) {
	if (docType == null)
	    return null;

	String[] list = 
	    (String[]) mixedContentChildTypeNames.get(elemTypeName);
	if (list != null)
	    return list;

	ElementTypeModel elemType = docType.getElementType(elemTypeName);
	if (elemType == null)
	    return null;

	ElementTypeModel[] mix = elemType.getMixedElementTypes();
	if (mix == null)
	    return null;

	list = new String[mix.length];
	for (int i = 0; i < mix.length; ++i)
	    // Even UNDECLARED are added to the list.
	    list[i] = mix[i].getElementTypeName();

	mixedContentChildTypeNames.put(elemTypeName, list);
	return list;
    }

    public String[] getElementContentChildTypeNames(ElementModel element) {
	return getElementContentChildTypeNames(element.getElementTypeName());
    }

    public String[] getElementContentChildTypeNames(String elemTypeName) {
	if (docType == null)
	    return null;

	String[] typeNames = 
	    (String[]) elementContentChildTypeNames.get(elemTypeName);
	if (typeNames != null)
	    return typeNames;

	ElementTypeModel elemType = docType.getElementType(elemTypeName);
	if (elemType == null)
	    return null;

	ElementContentModel contentModel = elemType.getElementContent();
	if (contentModel == null) 
	    return null;

	Vector list = new Vector();
	addElementContentChildTypeNames(contentModel, list);
	typeNames = new String[list.size()];
	list.copyInto(typeNames);

	elementContentChildTypeNames.put(elemTypeName, typeNames);
	return typeNames;
    }

    private static void addElementContentChildTypeNames(
	                                   ElementContentModel contentModel, 
			                   Vector list) {
	switch (contentModel.getListType()) {
	case ElementContentModel.NONE:
	    {
		ElementTypeModel elemType = contentModel.getElementType();
		String elemTypeName = elemType.getElementTypeName();

		// Even UNDECLARED are added to the list.
		if (list.indexOf(elemTypeName) < 0)
		    list.addElement(elemTypeName);
	    }
	    break;

	case ElementContentModel.CHOICE:
	case ElementContentModel.SEQUENCE:
	    {
		ElementContentModel[] seq = contentModel.getList(); 

		for (int i = 0; i < seq.length; ++i)
		    addElementContentChildTypeNames(seq[i], list);
	    }
	    break;
	}
    }

    public boolean checkElementContentIfDelete(ElementModel element,
					       ElementModel child) {
	if (docType == null)
	    return false;

	ElementContentBuffer elementContent = new ElementContentBuffer();

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub != child && !elementContent.append(sub))
		return false;

	    sub = sub.getNextSubelement();
	}

	NameRE re = getElementContentMatcher(element.getElementTypeName());
	if (re == null)
	    // Has no element content.
	    return false;
	else
	    return re.match(elementContent.names, elementContent.nameCount);
    }

    public boolean checkElementContentIfInsertBefore(ElementModel element,
						     ElementModel child,
						     String elemTypeName) {
	if (docType == null)
	    return false;

	ElementContentBuffer elementContent = new ElementContentBuffer();

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub == child)
		elementContent.append(elemTypeName);

	    if (!elementContent.append(sub))
		return false;

	    sub = sub.getNextSubelement();
	}

	if (child == null) 
	    elementContent.append(elemTypeName);

	NameRE re = getElementContentMatcher(element.getElementTypeName());
	if (re == null) 
	    // Has no element content.
	    return false;
	else
	    return re.match(elementContent.names, elementContent.nameCount);
    }

    public boolean checkElementContentIfInsertAfter(ElementModel element,
						    ElementModel child,
						    String elemTypeName) {
	if (docType == null)
	    return false;

	ElementContentBuffer elementContent = new ElementContentBuffer();

	if (child == null) 
	    elementContent.append(elemTypeName);

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (!elementContent.append(sub))
		return false;
	
	    if (sub == child)
		elementContent.append(elemTypeName);

	    sub = sub.getNextSubelement();
	}

	NameRE re = getElementContentMatcher(element.getElementTypeName());
	if (re == null)
	    // Has no element content.
	    return false;
	else
	    return re.match(elementContent.names, elementContent.nameCount);
    }

    public boolean checkElementContentIfReplace(ElementModel element,
						ElementModel child,
						String elemTypeName) {
	if (docType == null)
	    return false;

	ElementContentBuffer elementContent = new ElementContentBuffer();

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub == child) {
		elementContent.append(elemTypeName);
	    } else {
		if (!elementContent.append(sub))
		    return false;
	    }

	    sub = sub.getNextSubelement();
	}

	NameRE re = getElementContentMatcher(element.getElementTypeName());
	if (re == null) 
	    // Has no element content.
	    return false;
	else
	    return re.match(elementContent.names, elementContent.nameCount);
    }

    public boolean hasCompatibleContent(ElementModel element, 
					String dstTypeName) {
	ElementTypeModel dstType;
	if (docType == null ||
	    (dstType = docType.getElementType(dstTypeName)) == null)
	    return true;
	else
	    return hasCompatibleContent(element, dstType);
    }

    public boolean hasCompatibleContent(ElementModel element, 
					ElementTypeModel dstType) {
	// Note that element attributes compatibility with dstType is not
	// checked.

	String dstTypeName = dstType.getElementTypeName();

	switch (dstType.getElementContentType()) {
	case ElementTypeModel.EMPTY:
	    return (element.getSubelementCount() == 0);

	case ElementTypeModel.ANY:
	case ElementTypeModel.UNDECLARED:
	    {
		String[] list = getDeclaredElementTypeNames();

		SubelementModel sub = element.getFirstSubelement();
		while (sub != null) {
                    // ANY means any declared type or #PCDATA.

		    if (sub.getSubelementType() != SubelementModel.TEXT &&
			!StringUtil.contains(
			    list, ((ElementModel) sub).getElementTypeName()))
			return false;

		    sub = sub.getNextSubelement();
		}

		return true;
	    }

	case ElementTypeModel.MIXED:
	    {
		String[] list = getMixedContentChildTypeNames(dstTypeName);

		SubelementModel sub = element.getFirstSubelement();
		while (sub != null) {
		    if (sub.getSubelementType() != SubelementModel.TEXT &&
			!StringUtil.contains(
			    list, ((ElementModel) sub).getElementTypeName()))
			return false;

		    sub = sub.getNextSubelement();
		}

		return true;
	    }

	case ElementTypeModel.ELEMENT:
	    {
		ElementContentBuffer elementContent = 
		    new ElementContentBuffer();

		SubelementModel sub = element.getFirstSubelement();
		while (sub != null) {
		    if (!elementContent.append(sub))
			return false;

		    sub = sub.getNextSubelement();
		}

		NameRE re = getElementContentMatcher(dstTypeName);
		return re.match(elementContent.names, 
				elementContent.nameCount);
	    }

	default:
	    return false;
	}
    }

    public ElementContentBuffer getElementContentBuffer() {
	return new ElementContentBuffer();
    }

    public boolean checkElementContent(ElementTypeModel elementType,
				       ElementContentBuffer buffer) {
	if (docType == null)
	    return false;

	NameRE re = getElementContentMatcher(elementType.getElementTypeName());
	if (re == null)
	    // Has no element content.
	    return false;
	else
	    return re.match(buffer.names, buffer.nameCount);
    }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmledit.modeltool.ElemStructCheckerRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id, Object arg0, Object arg1) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}

	Object[] args;
	if (arg0 == null) {
	    args = new Object[0];
	} else {
	    if (arg1 == null) {
		args = new Object[] { arg0 };
	    } else {
		args = new Object[] { arg0, arg1 };
	    }
	}

	return MessageFormat.format(message, args);
    }
}

