/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;
import com.xmlmind.xmledit.model.*;

public abstract class ElementStructureEditor {
    public static class TextCollector extends Traversal.HandlerBase {
	public TextModel beginText;
	public int beginOffset;
	public TextModel endText;
	public int endOffset;
	public TextRangeList rangeList;

	public Object processText(TextModel text) {
	    // Empty ranges are added.

	    int textLength = text.getTextLength();
	    if (text == endText) {
		if (endOffset < 0)
		    endOffset = 0;
		else if (endOffset > textLength)
		    endOffset = textLength;

		if (text == beginText) {
		    if (beginOffset < 0)
			beginOffset = 0;
		    else if (beginOffset > textLength)
			beginOffset = textLength;

		    if (beginOffset > endOffset) {
			int tmp = beginOffset;
			beginOffset = endOffset;
			endOffset = tmp;
		    }

		    rangeList.append(text, beginOffset, endOffset-beginOffset);
		} else {
		    rangeList.append(text, 0, endOffset);
		}

		return text;
	    } else {
		if (text == beginText) {
		    if (beginOffset < 0)
			beginOffset = 0;
		    else if (beginOffset > textLength)
			beginOffset = textLength;

		    rangeList.append(text,beginOffset, textLength-beginOffset);
		} else {
		    rangeList.append(text, 0, textLength);
		}

		return null;
	    }
	}
    }

    // -----------------------------------------------------------------------

    public static final String PCDATA = "#PCDATA";

    protected DocumentModel doc;

    private ElementParser elementParser = null;

    // Scratch variables.
    private TextTransferEvent textTransferEvent = new TextTransferEvent();
    protected Traversal.TextMatcher textMatcher = new Traversal.TextMatcher();
    private Traversal.TextGrabber textGrabber = new Traversal.TextGrabber();
    private TextCollector textCollector = new TextCollector();
    private TextRangeList rangeList = new TextRangeList();
    private Vector editedElements = new Vector();
    private Vector nextEditedElements = new Vector();

    private static final int DELETE_CHARS                         = 0;
    private static final int DELETE_INNER_TEXT_CONTAINERS         = 1;
    private static final int DELETE_ALL_TEXT_CONTAINERS_BUT_FIRST = 2;

    // -----------------------------------------------------------------------

    protected ElementStructureEditor() {
	try {
	    elementParser = new ElementParser();
	} catch (Exception ignored) {
            //ignored.printStackTrace();
	}
    }

    public void setDocument(DocumentModel document) {
	doc = document;
    }

    public DocumentModel getDocument() {
	return doc;
    }

    // -----------------------------------------------------------------------

    /**
     * Returns true if beginText is located after endText.
     */
    public boolean getCommandInfo(TextModel beginText, int beginOffset,
				  TextModel endText, int endOffset, 
				  String clipboardContent, CommandInfo info) {
	info.clear();
	if (beginText == null || endText == null ||
	    (beginText == endText && beginOffset == endOffset))
	    return false;

	setClipboardTag(clipboardContent, info);
	return doGetCommandInfo(beginText, beginOffset, endText, endOffset, 
				info);
    }

    public boolean checkTextOrder(TextModel beginText, int beginOffset,
				  TextModel endText, int endOffset) {
	if (beginText == endText) {
	    if (beginOffset > endOffset)
		return false;
	} else {
	    textMatcher.searchedText = endText;
	    if (Traversal.traverseAfter(beginText, textMatcher) == null)
		return false;
	}

	return true;
    }

    private static final void setClipboardTag(String clipboardContent, 
					      CommandInfo info) {
	if (clipboardContent == null) {
	    info.clipboardTag = null;
	} else {
	    // Note that the clipboard may contain a well-formed but invalid
	    // element!

	    info.clipboardTag = parseElementTypeName(clipboardContent);
	    if (info.clipboardTag == null)
		info.clipboardTag = PCDATA;
	}
    }

    public static final String parseElementTypeName(String s) {
	if (!s.startsWith("<?xml "))
	    // Not an XML document.
	    return null;

	int length = s.length();
	int pos = 7;

	while (pos < length) {
	    pos = s.indexOf('<', pos);
	    if (pos < 0)
		return null;

	    if (pos+1 < length && isNameStartChar(s.charAt(pos+1))) {
		int end;
		loop: for (end = pos+2; end < length; ++end) {
		    switch (s.charAt(end)) {
		    case '>':
		    case ' ': case '\t': case '\r': case '\n':
			break loop;
		    }
		}

		String name = s.substring(pos+1, end);
		if (XMLUtil.isName(name))
		    return name;

		pos = end;
	    } else {
		++pos;
	    }
	}

	return null;
    }

    private static final boolean isNameStartChar(char c) {
	switch (c) {
	case '_': case ':':
	    break;
	default:
	    if (!Character.isLetter(c))
		return false;
	}
	return true;
    }

    public void get(TextModel beginText, int beginOffset,
		    TextModel endText, int endOffset, 
		    TextRangeList rangeList) {
	rangeList.removeAll();

	if (beginText == endText) {
	    if (beginOffset > endOffset) {
		int tmp = beginOffset;
		beginOffset = endOffset;
		endOffset = tmp;
	    }

	    rangeList.append(beginText, beginOffset, endOffset - beginOffset);
	    return;
	}

	// Try from beginText to endText ---

	textCollector.beginText = beginText;
	textCollector.beginOffset = beginOffset;
	textCollector.endText = endText;
	textCollector.endOffset = endOffset;
	textCollector.rangeList = rangeList;
	if (Traversal.traverseFrom(beginText, textCollector) != null) 
	    return;
	
	// EndText not after beginText. Try from endText to beginText ---

	rangeList.removeAll();

	textCollector.beginText = endText;
	textCollector.beginOffset = endOffset;
	textCollector.endText = beginText;
	textCollector.endOffset = beginOffset;
	if (Traversal.traverseFrom(endText, textCollector) != null) 
	    return;

	// BeginText/endText not a valid range ---

	rangeList.removeAll();
    }

    public String copy(TextModel beginText, int beginOffset,
		       TextModel endText, int endOffset) {
	String text = doCopy(beginText, beginOffset, endText, endOffset);
	rangeList.removeAll();
	return text;
    }

    private String doCopy(TextModel beginText, int beginOffset,
			  TextModel endText, int endOffset) {
	rangeList.removeAll();
	get(beginText, beginOffset, endText, endOffset, rangeList);

	TextRange[] ranges = rangeList.list;
	int rangeCount = rangeList.size;

	StringBuffer buffer = new StringBuffer();
	for (int i = 0; i < rangeCount; ++i) {
	    TextRange range = ranges[i];
	    if (range.count > 0)
		buffer.append(range.text.getTextChars(), 
			      range.offset, range.count);
	}

	if (buffer.length() == 0) {
	    return null;
	} else {
	    String text = buffer.toString();
	    return text;
	}
    }

    public String cut(TextModel beginText, int beginOffset,
		      TextModel endText, int endOffset) {
	return cut(beginText, beginOffset, endText, endOffset, false);
    }

    public String cut(TextModel beginText, int beginOffset,
		      TextModel endText, int endOffset, boolean deleteChars) {
	String text = doCopy(beginText, beginOffset, endText, endOffset);
	if (rangeList.size > 0) {
	    doDelete(rangeList, (deleteChars? DELETE_CHARS :
				 DELETE_INNER_TEXT_CONTAINERS));
	    rangeList.removeAll();
	}
	return text;
    }

    private void doDelete(TextRangeList rangeList, int mode) {
	doc.beginBatchEditing();

	TextRange[] ranges = rangeList.list;
	int rangeCount = rangeList.size;

	if (mode == DELETE_CHARS) {
	    for (int i = 0; i < rangeCount; ++i) {
		TextRange range = ranges[i];
		if (range.count > 0) 
		    range.text.deleteText(range.offset, range.count);
	    }
	} else {

	    SubelementModel dontRemove1, dontRemove2;
	    switch (mode) {
	    case DELETE_INNER_TEXT_CONTAINERS:
		dontRemove1 = ranges[0].text;
		dontRemove2 = ranges[rangeCount-1].text;
		break;
	    case DELETE_ALL_TEXT_CONTAINERS_BUT_FIRST:
		dontRemove1 = ranges[0].text;
		dontRemove2 = null;
		break;
	    default:
		dontRemove1 = null;
		dontRemove2 = null;
	    }

	    for (int i = 0; i < rangeCount; ++i) {
		TextRange range = ranges[i];

		TextModel text = range.text;
		if (range.count > 0) 
		    text.deleteText(range.offset, range.count);

		if (text.getTextLength() > 0)
		    continue;

		// Empty text in a text container ---

		ElementModel parent = text.getParentElement();

		if (text != dontRemove1 && text != dontRemove2) {
		    parent.removeSubelement(text);
		    if (!editedElements.contains(parent))
			editedElements.addElement(parent);
		}
	    }

	    ElementModel parent;
	    int editElementCount;
	    while ((editElementCount = editedElements.size()) > 0) {

		for (int i = 0; i < editElementCount; ++i) {
		    ElementModel element = 
			(ElementModel) editedElements.elementAt(i);

		    ElementUtil.normalizeTextContainer(element, doc);

		    if (!isEmptyTextContainer(element))
			continue;

		    parent = element.getParentElement();
		    if (parent == null || !ElementUtil.isTextContainer(parent))
			// Empty text container contained in a non-text
			// container parent.
			continue;

		    // An empty text container contained in a text container
		    // parent ---

		    if ((dontRemove1 == null || 
			 !isAncestorOf(element, dontRemove1)) && 
			(dontRemove2 == null || 
			 !isAncestorOf(element, dontRemove2))) {
			parent.removeSubelement(element);
			if (!nextEditedElements.contains(parent))
			    nextEditedElements.addElement(parent);
		    }
		}

		editedElements.removeAllElements();

		Vector swap = editedElements;
		editedElements = nextEditedElements;
		nextEditedElements = swap;
	    }
	}

	doc.endBatchEditing();
    }

    private static final boolean isAncestorOf(ElementModel element, 
					      SubelementModel sub) {
	while (sub != null) {
	    if (sub == element)
		return true;
	    sub = sub.getParentElement();
	}
	return false;
    }

    private static boolean isEmptyTextContainer(ElementModel element) {
	// Note that an empty text container with attributes is considered as
	// being really empty.

	switch (element.getSubelementCount()) {
	case 0:
	    return true;
	case 1:
	    {
		SubelementModel sub = element.getFirstSubelement();
		return (sub.getSubelementType() == SubelementModel.TEXT &&
			((TextModel) sub).getTextLength() == 0);
	    } 
	default:
	    return false;
	}
    }

    public void delete(TextModel beginText, int beginOffset,
		       TextModel endText, int endOffset) {
	delete(beginText, beginOffset, endText, endOffset, false);
    }

    public void delete(TextModel beginText, int beginOffset,
		       TextModel endText, int endOffset, boolean deleteChars) {
	rangeList.removeAll();
	get(beginText, beginOffset, endText, endOffset, rangeList);
	if (rangeList.size > 0) {
	    doDelete(rangeList, (deleteChars? DELETE_CHARS : 
				 DELETE_INNER_TEXT_CONTAINERS));
	    rangeList.removeAll();
	}
    }

    public void convert(TextModel beginText, int beginOffset,
			TextModel endText, int endOffset, 
			String elementTypeName) {
	convert(beginText, beginOffset, endText, endOffset, elementTypeName, 
		false);
    }

    public void convert(TextModel beginText, int beginOffset,
			TextModel endText, int endOffset, 
			String elementTypeName, boolean deleteChars) {
	if (!XMLUtil.isName(elementTypeName) && 
	    !elementTypeName.equals(PCDATA))
	    return;

	doc.beginBatchEditing();

	if (beginText == endText) {
	    if (beginOffset > endOffset) {
		int swapOffset = beginOffset; 
		beginOffset = endOffset; 
		endOffset = swapOffset;
	    }
	} else {
	    // Make sure that beginText is located before endText:
	    // endText may be removed, beginText may not.

	    textMatcher.searchedText = endText;
	    if (Traversal.traverseAfter(beginText, textMatcher) == null) {
		TextModel swapText = beginText; 
		beginText = endText; 
		endText = swapText;

		int swapOffset = beginOffset; 
		beginOffset = endOffset; 
		endOffset = swapOffset;
	    }
	}

	String paste = doCopy(beginText, beginOffset, endText, endOffset);
	if (rangeList.size > 0) {
	    doDelete(rangeList, (deleteChars? DELETE_CHARS :
				 DELETE_ALL_TEXT_CONTAINERS_BUT_FIRST));
	    rangeList.removeAll();
	}
	if (paste == null) {
	    doc.endBatchEditing();
	    return;
	}
	
	ElementModel parent = beginText.getParentElement();

	if (elementTypeName.equals(PCDATA)) {
	    // Convert to ``plain'' #PCDATA contained in parent.
	    beginText.insertText(beginOffset, paste);
	} else {
	    ElementModel newChild = createConvertedElement(elementTypeName);
	    newChild.appendSubelement(doc.createTextModel(paste));

	    insertElement(beginText, beginOffset, newChild);

	    if (beginText.getTextLength() == 0 && !deleteChars) 
		// BeginText has been entirely converted to something else:
		// remove useless empty beginText.
		parent.removeSubelement(beginText);
	}

	doc.endBatchEditing();
    }

    // -----------------------------------------------------------------------

    public void getCommandInfo(TextModel text, int offset,
			       String clipboardContent, CommandInfo info) {
	info.clear();
	if (text == null)
	    return;

	setClipboardTag(clipboardContent, info);
	doGetCommandInfo(text, offset, info);
    }

    public void paste(TextModel text, int offset, String clipboardContent) {
	if (clipboardContent == null || clipboardContent.length() == 0)
	    return;

	ElementModel newChild = parseElement(clipboardContent);
	if (newChild == null) {
	    // Plain string.
	    text.insertText(offset, filterText(clipboardContent));
	} else {
	    if (checkPastedElement(newChild)) {
		doc.beginBatchEditing();
		insertElement(text, offset, newChild);
		doc.endBatchEditing();
	    }
	}
    }

    private ElementModel parseElement(String s) {
	if (!s.startsWith("<?xml "))
	    // Not an XML document.
	    return null;

	ElementModel parsedElement = null;

	try {
	    if (elementParser != null)
		parsedElement = elementParser.parse(doc, s);
	} catch (Exception ignored) {
	    //ignored.printStackTrace();
	}

	return parsedElement;
    }

    private static final String filterText(String text) {
	// Char ::= #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | 
        //          [#x10000-#x10FFFF]

	char[] chars = text.toCharArray();
	int charCount = 0;

	for (int i = 0; i < chars.length; ++i) {
	    char c = chars[i];

	    switch (c) {
	    case 0x9:
	    case 0xA:
	    case 0xD:
		chars[charCount++] = c;
		break;
	    default:
		if ((c >= 0x20 && c <= 0xD7FF) ||
		    (c >= 0xE000 && c <= 0xFFFD))
		    chars[charCount++] = c;
	    }
	}

	return new String(chars, 0, charCount);
    }

    public ElementModel insert(TextModel text, int offset, 
			       String elementTypeName) {
	if (!XMLUtil.isName(elementTypeName))
	    return null;

	ElementModel newChild = createInsertedElement(elementTypeName);

	doc.beginBatchEditing();
	insertElement(text, offset, newChild);
	doc.endBatchEditing();

	return newChild;
    }

    private void insertElement(TextModel text, int offset, 
			       ElementModel newChild) {
	ElementModel parent = text.getParentElement();

	int textLength = text.getTextLength();
	if (offset < 0)
	    offset = 0;
	else if (offset > textLength)
	    offset = textLength;

	if (offset == 0) {
	    parent.insertSubelement(text, newChild);
	} else if (offset == textLength) {
	    parent.insertSubelement(text.getNextSubelement(), newChild);
	} else {
	    // Insert a text container in the middle of text ---

	    char[] chars = text.getTextChars();
	    TextModel split = doc.createTextModel(
	       new String(chars, offset, textLength-offset));

	    SubelementModel next = text.getNextSubelement();
	    parent.insertSubelement(next, newChild);
	    parent.insertSubelement(next, split);

	    textTransferEvent.initialize(this, text, offset, 
					 split, 0, textLength);
	    doc.notifyDocumentListeners(textTransferEvent);

	    text.deleteText(offset, textLength-offset);
	}
    }

    // -----------------------------------------------------------------------

    public void getCommandInfo(SubelementModel sub,
			       String clipboardContent, CommandInfo info) {
	info.clear();
	if (sub == null)
	    return;

	setClipboardTag(clipboardContent, info);
	doGetCommandInfo(sub, info);
    }

    public String copy(SubelementModel sub) {
	return SubelementUtil.toString(sub);
    }

    public String cut(SubelementModel sub) {
	String cut = SubelementUtil.toString(sub);

	delete(sub);

	return cut;
    }

    public void delete(SubelementModel sub) {
	ElementModel parent = sub.getParentElement();
	if (parent == null)
	    return;

	if (ElementUtil.isTextContainer(parent)) {
	    doc.beginBatchEditing();
	    parent.removeSubelement(sub);
	    ElementUtil.normalizeTextContainer(parent, doc);
	    doc.endBatchEditing();
	} else {
	    parent.removeSubelement(sub);
	}
    }

    public void paste(SubelementModel sub, String clipboardContent) {
	if (clipboardContent == null || clipboardContent.length() == 0)
	    return;

	if (sub.getSubelementType() == SubelementModel.TEXT) {
	    TextModel text = (TextModel) sub;
	    paste(text, text.getTextLength(), clipboardContent);
	    return;
	}

	ElementModel parent = (ElementModel) sub;

	ElementModel newChild = parseElement(clipboardContent);
	if (newChild == null) {
	    // Plain string.
	    doc.beginBatchEditing();
	    parent.appendSubelement(
		doc.createTextModel(filterText(clipboardContent)));
	    ElementUtil.mergeContiguousTexts(parent);
	    doc.endBatchEditing();
	} else {
	    if (checkPastedElement(newChild))
		parent.appendSubelement(newChild);
	}
    }

    public void pasteBefore(SubelementModel sub, String clipboardContent) {
	if (clipboardContent == null || clipboardContent.length() == 0)
	    return;

	ElementModel parent = sub.getParentElement();
	if (parent == null)
	    return;

	ElementModel newChild = parseElement(clipboardContent);
	if (newChild == null) {
	    // Plain string.
	    doc.beginBatchEditing();
	    parent.insertSubelement(
		sub, doc.createTextModel(filterText(clipboardContent)));
	    ElementUtil.mergeContiguousTexts(parent);
	    doc.endBatchEditing();
	} else {
	    if (checkPastedElement(newChild))
		parent.insertSubelement(sub, newChild);
	}
    }

    public void pasteAfter(SubelementModel sub, String clipboardContent) {
	if (clipboardContent == null || clipboardContent.length() == 0)
	    return;

	ElementModel parent = sub.getParentElement();
	if (parent == null)
	    return;

	ElementModel newChild = parseElement(clipboardContent);
	if (newChild == null) {
	    // Plain string.
	    doc.beginBatchEditing();
	    parent.insertSubelement(
		sub.getNextSubelement(), 
		doc.createTextModel(filterText(clipboardContent)));
	    ElementUtil.mergeContiguousTexts(parent);
	    doc.endBatchEditing();
	} else {
	    if (checkPastedElement(newChild))
		parent.insertSubelement(sub.getNextSubelement(), newChild);
	}
    }

    public SubelementModel insert(SubelementModel sub, 
				  String elementTypeName) {
	if (!XMLUtil.isName(elementTypeName) && 
	    !elementTypeName.equals(PCDATA))
	    return null;

	if (sub.getSubelementType() == SubelementModel.TEXT) {
	    if (elementTypeName.equals(PCDATA)) {
		return null;
	    } else {
		TextModel text = (TextModel) sub;
		return insert(text, text.getTextLength(), elementTypeName);
	    }
	} else {
	    ElementModel parent = (ElementModel) sub;

	    if (elementTypeName.equals(PCDATA)) {
		TextModel newText = doc.createTextModel("");

		doc.beginBatchEditing();
		parent.appendSubelement(newText);
		ElementUtil.mergeContiguousTexts(parent);
		doc.endBatchEditing();

		// newText may have been removed by mergeContiguousTexts
		// in which case this insertion op is a no-op.
		return (newText.getDocument() == null)? null : newText;
	    } else {
		ElementModel newChild = createInsertedElement(elementTypeName);
		parent.appendSubelement(newChild);
		return newChild;
	    }
	}
    }

    public SubelementModel insertBefore(SubelementModel sub, 
					String elementTypeName) {
	if (!XMLUtil.isName(elementTypeName) && 
	    !elementTypeName.equals(PCDATA))
	    return null;

	ElementModel parent = sub.getParentElement();
	if (parent == null)
	    return null;

	if (elementTypeName.equals(PCDATA)) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		// Inserting an empty text before a text is a no-op.
		return null;
	    } else {
		TextModel newText = doc.createTextModel("");

		doc.beginBatchEditing();
		parent.insertSubelement(sub, newText);
		ElementUtil.mergeContiguousTexts(parent);
		doc.endBatchEditing();

		return (newText.getDocument() == null)? null : newText;
	    }
	} else {
	    ElementModel newChild = createInsertedElement(elementTypeName);
	    parent.insertSubelement(sub, newChild);
	    return newChild;
	}
    }

    public SubelementModel insertAfter(SubelementModel sub, 
				       String elementTypeName) {
	if (!XMLUtil.isName(elementTypeName) && 
	    !elementTypeName.equals(PCDATA))
	    return null;

	ElementModel parent = sub.getParentElement();
	if (parent == null)
	    return null;

	if (elementTypeName.equals(PCDATA)) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		// Inserting an empty text after a text is a no-op.
		return null;
	    } else {
		TextModel newText = doc.createTextModel("");

		doc.beginBatchEditing();
		parent.insertSubelement(sub.getNextSubelement(), newText);
		ElementUtil.mergeContiguousTexts(parent);
		doc.endBatchEditing();

		return (newText.getDocument() == null)? null : newText;
	    }
	} else {
	    ElementModel newChild = createInsertedElement(elementTypeName);
	    parent.insertSubelement(sub.getNextSubelement(), newChild);
	    return newChild;
	}
    }

    public void convert(SubelementModel sub, String elementTypeName) {
	if (!XMLUtil.isName(elementTypeName) && 
	    !elementTypeName.equals(PCDATA))
	    return;

	ElementModel parent = sub.getParentElement();
	if (parent == null)
	    return;

	if (sub.getSubelementType() == SubelementModel.TEXT) {
	    TextModel text = (TextModel) sub;

	    if (!elementTypeName.equals(PCDATA)) {
		ElementModel newChild = 
		    createConvertedElement(elementTypeName);
		newChild.appendSubelement(doc.createTextModel(text.getText()));

		parent.replaceSubelement(text, newChild);
	    }
	} else {
	    ElementModel element = (ElementModel) sub;

	    if (elementTypeName.equals(PCDATA)) {
		textGrabber.grabbed = new StringBuffer();
		Traversal.traverse(element, textGrabber);

		TextModel newChild = 
		    doc.createTextModel(textGrabber.grabbed.toString());

		doc.beginBatchEditing();
		parent.replaceSubelement(element, newChild);
		ElementUtil.mergeContiguousTexts(parent);
		doc.endBatchEditing();

	    } else if (!elementTypeName.equals(element.getElementTypeName())) {
		ElementModel newChild = 
		    createConvertedElement(elementTypeName);

		doc.beginBatchEditing();
		parent.replaceSubelement(element, newChild);
		copySubelements(element, newChild, doc);
		doc.endBatchEditing();
	    }
	}
    }

    private static void copySubelements(ElementModel src, ElementModel dst,
					DocumentModel doc) {
	// Note that src attributes are not copied to dst.

	SubelementModel sub = src.getFirstSubelement();
	while (sub != null) {
	    dst.appendSubelement(SubelementUtil.copy(sub, doc));

	    sub = sub.getNextSubelement();
	}
    }

    public SubelementModel replace(SubelementModel sub, 
				   String elementTypeName) {
	if (!XMLUtil.isName(elementTypeName) && 
	    !elementTypeName.equals(PCDATA))
	    return null;

	ElementModel parent = sub.getParentElement();
	if (parent == null)
	    return null;

	SubelementModel newChild;

	if (elementTypeName.equals(PCDATA)) {
	    newChild = doc.createTextModel("");

	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		// Replace a text by an empty text.
		parent.replaceSubelement(sub, newChild);
	    } else {
		doc.beginBatchEditing();
		parent.replaceSubelement(sub, newChild);
		ElementUtil.mergeContiguousTexts(parent);
		doc.endBatchEditing();
	    }
	} else {
	    newChild = createInsertedElement(elementTypeName);
	    parent.replaceSubelement(sub, newChild);
	}

	return newChild;
    }

    // -----------------------------------------------------------------------

    public boolean canSplit(ElementModel element, TextModel text, int offset) {
	if (!checkSplitPosition(element, text, offset))
	    return false;

	return checkCanSplit(element, text, offset);
    }

    private boolean checkSplitPosition(ElementModel element, 
				       TextModel text, int offset) {
	if (text == null || element == null || 
	    element.getParentElement() == null)
	    return false;
	
	ElementModel ancestor = text.getParentElement();
	while (ancestor != null) {
	    if (ancestor == element)
		return true;
	    ancestor = ancestor.getParentElement();
	}

	return false;
    }

    public void split(ElementModel element, TextModel text, int offset) {
	if (!checkSplitPosition(element, text, offset))
	    return;

	doc.beginBatchEditing();

	ElementModel ancestor = text.getParentElement();
	ElementModel newElement = createSplitElement(ancestor);

	int textLength = text.getTextLength();
	if (offset < 0)
	    offset = 0;
	else if (offset > textLength)
	    offset = textLength;

	char[] chars = text.getTextChars();
	TextModel split = doc.createTextModel(
	   new String(chars, offset, textLength-offset));

	newElement.appendSubelement(split);
	transferSubelements(ancestor, text, newElement, doc);

	while (ancestor != element) {
	    ElementModel previousAncestor = ancestor;
	    ancestor = ancestor.getParentElement();

	    ElementModel nextNewElement = createSplitElement(ancestor);
	    nextNewElement.appendSubelement(newElement);
	    transferSubelements(ancestor, previousAncestor, 
				nextNewElement, doc);

	    newElement = nextNewElement;
	}

	element.getParentElement().insertSubelement(
	    element.getNextSubelement(), newElement);

	textTransferEvent.initialize(this, text, offset, 
				     split, 0, textLength);
	doc.notifyDocumentListeners(textTransferEvent);

	text.deleteText(offset, textLength-offset);

	doc.endBatchEditing();
    }

    private static void transferSubelements(ElementModel src, 
					    SubelementModel child,
					    ElementModel dst,
					    DocumentModel doc) {
	SubelementModel sub = child.getNextSubelement();
	while (sub != null) {
	    SubelementModel moved = sub;
	    sub = moved.getNextSubelement();
	    src.removeSubelement(moved);

	    dst.appendSubelement(SubelementUtil.copy(moved, doc));
	}
    }

    // -----------------------------------------------------------------------

    public void join(SubelementModel sub) {
	if (!canJoin(sub))
	    return;
	ElementModel src = (ElementModel) sub;
	ElementModel dst = (ElementModel) sub.getPreviousSubelement();

	doc.beginBatchEditing();

	SubelementModel joinSub = null;
	boolean textAdded = false;

	SubelementModel srcSub = src.getFirstSubelement();
	while (srcSub != null) {
	    SubelementModel newSub = SubelementUtil.copy(srcSub, doc);

	    if (joinSub == null)
		joinSub = newSub;

	    if (newSub.getSubelementType() == SubelementModel.TEXT)
		textAdded = true;

	    dst.appendSubelement(newSub);

	    srcSub = srcSub.getNextSubelement();
	}

	if (joinSub != null)
	    join(joinSub);

	if (textAdded)
	    ElementUtil.mergeContiguousTexts(dst);

	src.getParentElement().removeSubelement(src);

	doc.endBatchEditing();
    }

    protected boolean canJoin(SubelementModel sub) {
	SubelementModel prevSub;
	if (sub.getSubelementType() != SubelementModel.ELEMENT ||
	    (prevSub = sub.getPreviousSubelement()) == null ||
	    prevSub.getSubelementType() != SubelementModel.ELEMENT)
	    return false;

	ElementModel src = (ElementModel) sub;
	ElementModel dst = (ElementModel) prevSub;
	if (!src.getElementTypeName().equals(dst.getElementTypeName()))
	    return false;

	return true;
    }

    // -----------------------------------------------------------------------

    protected abstract boolean doGetCommandInfo(TextModel beginText, 
						int beginOffset,
						TextModel endText, 
						int endOffset, 
						CommandInfo info);

    protected abstract void doGetCommandInfo(TextModel text, int offset,
					     CommandInfo info);

    protected abstract void doGetCommandInfo(SubelementModel sub,
					     CommandInfo info);

    protected abstract boolean checkCanSplit(ElementModel element, 
					     TextModel text, int offset);

    protected ElementModel createConvertedElement(String elementTypeName) {
	return doc.createElementModel(elementTypeName);
    }

    protected ElementModel createInsertedElement(String elementTypeName) {
	return doc.createElementModel(elementTypeName);
    }

    protected ElementModel createSplitElement(ElementModel element) {
	return doc.createElementModel(element.getElementTypeName());
    }

    protected boolean checkPastedElement(ElementModel element) {
	return true;
    }
}
