/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import com.xmlmind.xmledit.model.*;

public class ElementUtil {    
    public static final DocumentTypeModel getDocumentType(
	                                              ElementModel element) {
	DocumentModel doc = element.getDocument();
	return (doc == null)? null : doc.getDocumentType();
    }

    public static final ElementTypeModel getElementType(ElementModel element) {
	return getElementType(element, getDocumentType(element));
    }

    public static final ElementTypeModel getElementType(
	                                        ElementModel element,
						DocumentTypeModel docType) {
	return ((docType == null)? null : 
		docType.getElementType(element.getElementTypeName()));
    }

    public static final AttributeDefinitionModel getElementAttributeDefinition(
	                                               ElementModel element, 
						       String name) {
	return getElementAttributeDefinition(element, name, 
					     getDocumentType(element));
    }

    public static final AttributeDefinitionModel getElementAttributeDefinition(
	                                           ElementModel element, 
						   String name, 
						   DocumentTypeModel docType) {
	ElementTypeModel elemType = getElementType(element, docType);
	return ((elemType == null)? 
		null : elemType.getAttributeDefinition(name));
    }

    // -----------------------------------------------------------------------

    public static final String getElementAttribute(ElementModel element, 
						   String name) {
	return getElementAttribute(element, name, getDocumentType(element));
    }

    public static final String getElementAttribute(ElementModel element, 
						   String name, 
						   DocumentTypeModel docType) {
	String value = element.getElementAttribute(name);

	if (value == null) {
	    AttributeDefinitionModel attrDef = 
		getElementAttributeDefinition(element, name, docType);

	    if (attrDef != null) {
		switch (attrDef.getAttributePresence()) {
		case AttributeDefinitionModel.DEFAULTED:
		case AttributeDefinitionModel.FIXED:
		    value = attrDef.getAttributeDefaultValue();
		    break;
		}
	    }
	}

	return value;
    }

    public static final String lookupElementAttribute(ElementModel element, 
						      String name) {
	return lookupElementAttribute(element, name, getDocumentType(element));
    }

    public static final String lookupElementAttribute(
	                                          ElementModel element, 
						  String name, 
						  DocumentTypeModel docType) {
	String value = getElementAttribute(element, name, docType);

	if (value == null) {
	    element = element.getParentElement();
	    if (element != null)
		value = lookupElementAttribute(element, name, docType);
	}

	return value;
    }

    public static final int getIntAttribute(ElementModel element, String name, 
					    int defaultValue) {
	return getIntAttribute(element, name, defaultValue, 
			       getDocumentType(element));
    }

    public static final int getIntAttribute(ElementModel element, 
					    String name, 
					    int defaultValue, 
					    DocumentTypeModel docType) {
	int result = defaultValue;

	String value = getElementAttribute(element, name, docType);
	if (value != null) {
	    try {
		result = Integer.parseInt(value);
	    } catch (NumberFormatException ignored) {}
	}

	return result;
    }

    public static final String getLanguage(ElementModel element) {
	return getLanguage(element, getDocumentType(element));
    }

    public static final String getLanguage(ElementModel element, 
					   DocumentTypeModel docType) {
	return lookupElementAttribute(element, "xml:lang", docType);
    }

    public static final boolean hasPreserveSpace(ElementModel element) {
	return hasPreserveSpace(element, getDocumentType(element));
    }

    public static final boolean hasPreserveSpace(ElementModel element, 
						 DocumentTypeModel docType) {
	String value = lookupElementAttribute(element, "xml:space", docType);
	return (value != null && value.equals("preserve"));
    }

    // -----------------------------------------------------------------------

    public static final boolean isTextContainer(ElementModel element) {
	return isTextContainer(element, getDocumentType(element));
    }

    public static final boolean isTextContainer(ElementModel element, 
						DocumentTypeModel docType) {
	ElementTypeModel elemType = getElementType(element, docType);

	if (elemType == null) {
	    // No DTD or undeclared.
	    return true;
	} else {
	    switch (elemType.getElementContentType()) {
	    case ElementTypeModel.MIXED:
	    case ElementTypeModel.ANY:
		// ANY means any declared type or #PCDATA.
	    case ElementTypeModel.UNDECLARED:
		return true;
	    default:
		return false;
	    }
	}
    }

    public static final boolean hasMixedContent(ElementModel element) {
	return hasMixedContent(element, getDocumentType(element));
    }

    public static final boolean hasMixedContent(ElementModel element, 
						DocumentTypeModel docType) {
	ElementTypeModel elemType = getElementType(element, docType);
	return (elemType != null && 
		elemType.getElementContentType() == ElementTypeModel.MIXED);
    }

    // -----------------------------------------------------------------------

    public static final void normalizeTextContainer(ElementModel element) {
	normalizeTextContainer(element, element.getDocument());
    }

    public static final void normalizeTextContainer(ElementModel element,
						    DocumentModel doc) {
	// May be called even with element not yet attached to doc.

	if (element.getSubelementCount() == 0) {
	    if (ElementUtil.hasMixedContent(element))
		element.appendSubelement(doc.createTextModel(""));
	} else {
	    mergeContiguousTexts(element);
	}
    }

    public static final void mergeContiguousTexts(ElementModel element) {
	// Note that there is no need to notify text transfers if element is
	// not attached to the document.
	DocumentModel doc = element.getDocument();
	TextTransferEvent textTransferEvent = null;

	SubelementModel sub = element.getFirstSubelement();

	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.TEXT) {
		TextModel text = (TextModel) sub;
		int endOffset = text.getTextLength();

		SubelementModel next = sub.getNextSubelement();
		while (next != null &&
		       next.getSubelementType() == SubelementModel.TEXT) {

		    TextModel nextText = (TextModel) next;
		    int charCount = nextText.getTextLength();
		    if (charCount > 0)
			text.appendText(nextText.getText());

		    if (doc != null) {
			// Properly transfer TextLocations from nextText to
			// text. (do this even if nextText was empty.)

			if (textTransferEvent == null)
			    textTransferEvent = new TextTransferEvent();

			textTransferEvent.initialize(
			    ElementUtil.class, nextText, 0, text, endOffset, 
			    charCount);
			doc.notifyDocumentListeners(textTransferEvent);
			endOffset += charCount;
		    }

		    SubelementModel removing = next;
		    next = removing.getNextSubelement();
		    element.removeSubelement(removing);
		}
	    } 

	    sub = sub.getNextSubelement();
	}
    }
}
