/*
 * Copyright (c) 2001-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.Vector;
import fr.pixware.util.*;
import com.xmlmind.xmledit.model.*;

public class SearchReplaceText {
    public static final int MARK_NONE     = 0x1;
    public static final int MARK_ALL      = 0x2;
    public static final int MARK_LAST     = 0x4;
    public static final int IGNORE_CASE   = 0x8;
    public static final int WHOLE_WORD    = 0x10;
    public static final int BACKWARD      = 0x20;
    public static final int REPLACE_CHARS = 0x40;

    public static final String DEFAULT_EMPTY_TEXT_CONTENT = 
        ElementCharSequence.DEFAULT_EMPTY_TEXT_CONTENT;
    public static final String DEFAULT_PARAGRAPH_MARK = 
        ElementCharSequence.DEFAULT_PARAGRAPH_MARK;

    private ElementModel searchedElement;
    private TextOffset searchBegin;
    private TextOffset searchEnd;
    private ElementStructureEditor structureEditor;
    private String searched;
    private RE searchedRE;
    private String replacement;
    private boolean ignoreCase;
    private boolean wholeWord;
    private boolean backward;
    private boolean replaceChars;
    private int markParagraphs;
    private String emptyTextContent;
    private String paragraphMark;

    private ElementCharSequence chars;
    private int searchFrom;
    private String replacementText;
    private TextOffset foundBegin;
    private TextOffset foundEnd;
    private int replacementCount;

    // -----------------------------------------------------------------------

    public SearchReplaceText(ElementModel element, 
			     String searched, String replacement,
			     int options) {
	this(element, null, null, searched, null, replacement, options,
	     DEFAULT_EMPTY_TEXT_CONTENT, DEFAULT_PARAGRAPH_MARK);
    }
    public SearchReplaceText(ElementModel element, 
			     String searched, String replacement,
			     int options, String emptyTextContent,
			     String paragraphMark) {
	this(element, null, null, searched, null, replacement, options,
	     emptyTextContent, paragraphMark);
    }
    public SearchReplaceText(TextOffset begin, TextOffset end, 
			     String searched, String replacement,
			     int options) {
	this(null, begin, end, searched, null, replacement, options,
	     DEFAULT_EMPTY_TEXT_CONTENT, DEFAULT_PARAGRAPH_MARK);
    }
    public SearchReplaceText(TextOffset begin, TextOffset end, 
			     String searched, String replacement,
			     int options, String emptyTextContent,
			     String paragraphMark) {
	this(null, begin, end, searched, null, replacement, options,
	     emptyTextContent, paragraphMark);
    }
    public SearchReplaceText(ElementModel element, 
			     TextOffset begin, TextOffset end, 
			     String searched, String replacement,
			     int options) {
	this(element, begin, end, searched, null, replacement, options,
	     DEFAULT_EMPTY_TEXT_CONTENT, DEFAULT_PARAGRAPH_MARK);
    }
    public SearchReplaceText(ElementModel element, 
			     TextOffset begin, TextOffset end, 
			     String searched, String replacement,
			     int options, String emptyTextContent,
			     String paragraphMark) {
	this(element, begin, end, searched, null, replacement, options,
	     emptyTextContent, paragraphMark);
    }

    // -----------------------------------------------------------------------

    public SearchReplaceText(ElementModel element, 
			     RE searchedRE, String replacement,
			     int options) {
	this(element, null, null, null, searchedRE, replacement, options,
	     DEFAULT_EMPTY_TEXT_CONTENT, DEFAULT_PARAGRAPH_MARK);
    }
    public SearchReplaceText(ElementModel element, 
			     RE searchedRE, String replacement,
			     int options, String emptyTextContent,
			     String paragraphMark) {
	this(element, null, null, null, searchedRE, replacement, options,
	     emptyTextContent, paragraphMark);
    }
    public SearchReplaceText(TextOffset begin, TextOffset end, 
			     RE searchedRE, String replacement,
			     int options) {
	this(null, begin, end, null, searchedRE, replacement, options,
	     DEFAULT_EMPTY_TEXT_CONTENT, DEFAULT_PARAGRAPH_MARK);
    }
    public SearchReplaceText(TextOffset begin, TextOffset end, 
			     RE searchedRE, String replacement,
			     int options, String emptyTextContent,
			     String paragraphMark) {
	this(null, begin, end, null, searchedRE, replacement, options,
	     emptyTextContent, paragraphMark);
    }
    public SearchReplaceText(ElementModel element, 
			     TextOffset begin, TextOffset end, 
			     RE searchedRE, String replacement,
			     int options) {
	this(element, begin, end, null, searchedRE, replacement, options,
	     DEFAULT_EMPTY_TEXT_CONTENT, DEFAULT_PARAGRAPH_MARK);
    }
    public SearchReplaceText(ElementModel element, 
			     TextOffset begin, TextOffset end, 
			     RE searchedRE, String replacement,
			     int options, String emptyTextContent,
			     String paragraphMark) {
	this(element, begin, end, null, searchedRE, replacement, options,
	     emptyTextContent, paragraphMark);
    }

    // -----------------------------------------------------------------------

    private SearchReplaceText(ElementModel element, 
			      TextOffset begin, TextOffset end, 
			      String searched, RE searchedRE, 
			      String replacement,
			      int options, String emptyTextContent,
			      String paragraphMark) {
	if (element == null) {
	    if (begin == null) {
		if (end != null)
		    element = end.text.getDocument().getRootElement();
	    } else {
		element = begin.text.getDocument().getRootElement();
	    }
	}
	if (element == null)
	    throw new IllegalArgumentException("unspecified text range");
	searchedElement = element;
	structureEditor = new DefaultElemStructEditor();
	structureEditor.setDocument(element.getDocument());

	searchBegin = begin;
	searchEnd = end;

	if (searchedRE == null) {
	    if (searched == null || searched.length() == 0)
		throw new IllegalArgumentException("nothing to search");
	    this.searched = searched;
	} else {
	    this.searchedRE = searchedRE;
	}

	// A null replacement means: we don't want to replace found text.
	this.replacement = replacement;

	ignoreCase = ((options & IGNORE_CASE) != 0);
	wholeWord = ((options & WHOLE_WORD) != 0);
	backward = ((options & BACKWARD) != 0);
	replaceChars = ((options & REPLACE_CHARS) != 0);

	if (emptyTextContent == null || emptyTextContent.length() == 0)
	    emptyTextContent = DEFAULT_EMPTY_TEXT_CONTENT;
	this.emptyTextContent = emptyTextContent;

	if ((options & MARK_NONE) != 0)
	    markParagraphs = ElementCharSequence.MARK_NONE;
	else if ((options & MARK_ALL) != 0)
	    markParagraphs = ElementCharSequence.MARK_ALL;
	else
	    markParagraphs = ElementCharSequence.MARK_LAST;

	if (paragraphMark == null || paragraphMark.length() == 0)
	    paragraphMark = DEFAULT_PARAGRAPH_MARK;
	this.paragraphMark = paragraphMark;

	chars = new ElementCharSequence(element, begin, end, emptyTextContent,
					markParagraphs, paragraphMark);
//System.err.println(chars.toString());

	if (backward)
	    searchFrom = chars.length() - 1;
	else
	    searchFrom = 0;
	foundBegin = foundEnd = null;
	replacementText = null;
	replacementCount = 0;
    }

    // -----------------------------------------------------------------------

    public TextOffset[] search() {
	if (searchFrom < 0) {
	    return null;
	}

	int foundIndex;
	String found;

	for (;;) {
	    if (searchedRE != null) {
		REMatch[] match;
		if (backward)
		    match = getMatchBackward(searchedRE, chars, searchFrom);
		else
		    match = searchedRE.getMatch(chars, searchFrom);

		if (match == null) {
		    foundIndex = -1;
		    found = null;
		    replacementText = null;
		} else {
		    foundIndex = match[0].getStartIndex();
		    found = match[0].getText();
		    if (replacement == null)
			replacementText = null;
		    else
			replacementText = substituteMatchVars(replacement, 
							      match);
		}
	    } else {
		if (backward)
		    foundIndex = 
			CharSequenceUtil.lastIndexOf(chars, searched, 
						     searchFrom, ignoreCase);
		else
		    foundIndex = 
			CharSequenceUtil.indexOf(chars, searched, searchFrom,
						 ignoreCase);
		if (foundIndex < 0) {
		    found = null;
		    replacementText = null;
		} else {
		    found = searched;
		    replacementText = replacement;
		}
	    }

	    // Prepare next iteration.
	    if (found == null) {
		searchFrom = -1;
		break;
	    }
	    int foundEnd = foundIndex + found.length();
	    if (backward)
		searchFrom = foundIndex-1;
	    else
		searchFrom = foundEnd;
		
	    if (!wholeWord)
		break;

	    if ((foundIndex == 0 || 
		 Character.isWhitespace(chars.charAt(foundIndex-1))) &&
		(foundEnd == chars.length() ||
		 Character.isWhitespace(chars.charAt(foundEnd))))
		// Found a word.
		break;

	    if (backward)
		--searchFrom;
	    else
		++searchFrom;
	}

	if (found == null) {
	    foundBegin = foundEnd = null;
	    return null;
	}

	foundBegin = chars.getTextOffset(foundIndex);
	foundEnd = chars.getTextOffset(foundIndex + found.length());
	if (foundEnd.offset == 0 && foundBegin.text != foundEnd.text) {
	    TextModel beforeEnd = 
		(TextModel) Traversal.traverseBefore(foundEnd.text, 
						     Traversal.textFinder);
	    if (beforeEnd != null) {
		foundEnd.text = beforeEnd;
		foundEnd.offset = beforeEnd.getTextLength();
	    }
	}

	return new TextOffset[] { foundBegin, foundEnd };
    }

    private static final REMatch[] getMatchBackward(RE re, CharSequence chars,
						    int from) {
	// Really, really inefficient!!!

	for (int i = from-1; i >= 0; --i) {
	    REMatch[] match = re.getMatch(chars, i);
	    if (match != null && match[0].getStartIndex() == i)
		return match;
	}

	return null;
    }

    private static final String substituteMatchVars(String template,
						    REMatch[] match) {
	if (template.indexOf('$') < 0)
	    return template;

	String substituted = template;

	substituted = StringUtil.replaceAll(substituted, "$$", "\0");

	for (int i = 0; i < match.length; ++i)
	    substituted = StringUtil.replaceAll(substituted, "$" + i, 
						match[i].getText());

	return substituted.replace('\0', '$');
    }

    public TextRange replace() {
	if (replacementText == null)
	    return null;

	DocumentModel document = structureEditor.getDocument();
	document.beginBatchEditing();

	// foundBegin.text is guaranteed no to be deleted even if 
	// the text container is normalized.
	// foundEnd.text may be deleted, unless the REPLACE_CHARS option
	// is selected.
	structureEditor.delete(foundBegin.text, foundBegin.offset,
			       foundEnd.text, foundEnd.offset, replaceChars);
	foundBegin.text.insertText(foundBegin.offset, replacementText);

	document.endBatchEditing();
	++replacementCount;

	TextRange replacedRange = 
	    new TextRange(foundBegin.text, foundBegin.offset,
			  replacementText.length());
	
	TextOffset begin, end;

	if (backward) {
	    begin = searchBegin;
	    end = foundBegin;
	} else {
	    foundBegin.offset += replacedRange.count;
	    begin = foundBegin;
	    // Note that if foundEnd.text is deleted, searchEnd (which may be
	    // equal to foundEnd.text) may no longer be part of the document.
	    end = searchEnd;
	}
	foundBegin = foundEnd = null;

	chars = new ElementCharSequence(searchedElement, begin, end, 
					emptyTextContent, 
					markParagraphs, paragraphMark);

	if (backward)
	    searchFrom = chars.length() - 1;
	else
	    searchFrom = 0;

	return replacedRange;
    }

    public TextRange[] replaceAll() {
	// At least one replacement.
	if (replacementText == null)
	    return null;

	Vector replacedRangeList = new Vector();

	DocumentModel document = structureEditor.getDocument();
	document.beginBatchEditing();

	do {
	    TextRange replacedRange = replace();
	    replacedRangeList.addElement(replacedRange);
	} while (search() != null);

	document.endBatchEditing();

	TextRange[] replacedRanges = new TextRange[replacedRangeList.size()];
	replacedRangeList.copyInto(replacedRanges);
	return replacedRanges;
    }

    public int getReplacementCount() {
	return replacementCount;
    }
}
