/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.Stack;
import java.util.EmptyStackException;
import com.xmlmind.xmledit.model.*;

public class SubelementMark implements DocumentListener {
    private DocumentModel doc = null;
    private SubelementModel marked = null;
    private Stack stack = new Stack();

    private SubelementMarkListener[] listeners = new SubelementMarkListener[0];
    private SubelementMarkEvent subelementMarkEvent=new SubelementMarkEvent();

    // -----------------------------------------------------------------------

    public SubelementMark() {
	this(null);
    }

    public SubelementMark(DocumentModel document) {
	setDocument(document);
    }

    public void setDocument(DocumentModel document) {
	if (doc != null) {
	    setSubelement(null);
	    doc.removeDocumentListener(this);
	    doc = null;
	}

	if (document != null) {
	    doc = document;
	    document.addDocumentListener(this, false);
	}
    }

    public DocumentModel getDocument() {
	return doc;
    }

    public boolean setSubelement(SubelementModel sub) {
	if (!moveTo(sub))
	    return false;

	popAll();
	return true;
    }

    private boolean moveTo(SubelementModel sub) {
	if (sub == marked)
	    return false;

	SubelementModel prevMarked = marked;
	marked = sub;

	notifySubelementMarkListeners(prevMarked, sub);
	
	return true;
    }

    private void popAll() {
	while (!stack.empty())
	    stack.pop();
    }

    public SubelementModel getSubelement() {
	return marked;
    }

    public boolean moveToParent() {
	if (marked == null)
	    return false;

	ElementModel parent = marked.getParentElement();
	if (parent == null)
	    return false;

	stack.push(marked);

	return moveTo(parent);
    }

    public boolean moveToChild() {
	if (stack.empty()) {
	    if (marked == null ||
		marked.getSubelementType() == SubelementModel.TEXT)
		return false;

	    SubelementModel firstChild = 
		((ElementModel) marked).getFirstSubelement();
	    if (firstChild == null)
		return false;
	    
	    return moveTo(firstChild);
	} else {
	    SubelementModel child = (SubelementModel) stack.peek();
	    stack.pop();
	    
	    return moveTo(child);
	}
    }

    public void addSubelementMarkListener(SubelementMarkListener listener) {
	if (indexOfSubelementMarkListener(listener) >= 0)
	    return;

	SubelementMarkListener[] newListeners = 
	    new SubelementMarkListener[listeners.length + 1];

	if (listeners.length > 0)
	    System.arraycopy(listeners, 0, newListeners, 0, listeners.length);
	newListeners[listeners.length] = listener;

	listeners = newListeners;
    }

    public void removeSubelementMarkListener(SubelementMarkListener listener) {
	int listenerIndex = indexOfSubelementMarkListener(listener);
	if (listenerIndex < 0)
	    return;

	SubelementMarkListener[] newListeners = 
	    new SubelementMarkListener[listeners.length - 1];

	if (listenerIndex > 0)
	    System.arraycopy(listeners, 0, newListeners, 0, listenerIndex);

	int first = listenerIndex + 1;
	int tail = listeners.length - first;
	if (tail > 0) 
	    System.arraycopy(listeners, first, 
			     newListeners, listenerIndex, tail);

	listeners = newListeners;
    }

    private 
    int indexOfSubelementMarkListener(SubelementMarkListener listener) {
	for (int i = 0; i < listeners.length; ++i) {
	    if (listeners[i] == listener)
		return i;
	}
	return -1;
    }

    private void notifySubelementMarkListeners(SubelementModel prevSub, 
					       SubelementModel sub) {
	subelementMarkEvent.initialize(this, prevSub, sub);

	for (int i = 0; i < listeners.length; ++i)
	    listeners[i].subelementMarkMoved(subelementMarkEvent);
    }

    // -----------------------------------------------------------------------

    public void elementStructureChanged(ElementStructureEvent event, 
					int listenerIndex) {
	switch (event.getId()) {
	case ElementStructureEvent.CHILD_REPLACED:
	case ElementStructureEvent.CHILD_REMOVED:
	    if (event.getOldChild() == marked)
		setSubelement(null);
	    break;
	}
    }

    public void elementAttributeChanged(ElementAttributeEvent event, 
					int listenerIndex) {}
    public void textChanged(TextEvent event, int listenerIndex) {}
    public void batchEditingStarted(BatchEditingEvent event, 
				    int listenerIndex) {}
    public void batchEditingCompleted(BatchEditingEvent event, 
				      int listenerIndex) {}
    public void documentEventHappened(DocumentEvent event,
				      int listenerIndex) {}
}
