/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import java.util.Enumeration;
import java.util.Hashtable;
import com.xmlmind.xmledit.model.*;

public class TextLocations implements DocumentListener {
    private DocumentModel doc = null;
    private Hashtable locations = new Hashtable();
    private TextLocation[] locationList = null;

    private TextLocationListener[] listeners = new TextLocationListener[0];
    private TextLocationEvent textLocationEvent = new TextLocationEvent();

    private boolean batchEditing = false;
    private Hashtable eventBatch = new Hashtable();

    // Scratch variables.
    private Traversal.TextMatcher textMatcher = new Traversal.TextMatcher();

    // -----------------------------------------------------------------------

    public TextLocations() {
	this(null);
    }

    public TextLocations(DocumentModel document) {
	setDocument(document);
    }

    public void setDocument(DocumentModel document) {
	if (doc != null) {
	    removeAll();
	    doc.removeDocumentListener(this);
	    doc = null;
	}

	if (document != null) {
	    doc = document;
	    document.addDocumentListener(this, false);
	}
    }

    public DocumentModel getDocument() {
	return doc;
    }

    public TextLocation add(Object id) {
	TextLocation location = (TextLocation) locations.get(id);
	if (location == null) {
	    location = new TextLocation(this, id);
	    locations.put(id, location);

	    // Clear cached list.
	    locationList = null;

	    notifyTextLocationListeners(location, 
					TextLocationEvent.LOCATION_ADDED,
					null, 0);
	}
	return location;
    }

    public boolean remove(TextLocation location) {
	return remove(location.getId());
    }

    public boolean remove(Object id) {
	TextLocation removed = (TextLocation) locations.remove(id);
	if (removed == null) 
	    return false;
	
	// Clear cached list.
	locationList = null;

	TextModel prevText = removed.getText();
	int prevOffset = removed.getOffset();

	removed.unsetManager();
	// Will not notify because it is no longer managed.
	removed.moveTo(null, 0);

	notifyTextLocationListeners(removed, 
				    TextLocationEvent.LOCATION_REMOVED,
				    prevText, prevOffset);
	return true;
    }

    public void removeAll() {
	TextLocation[] list = getAll();
	for (int i = 0; i < list.length; ++i) 
	    remove(list[i]);
    }

    public TextLocation get(Object id) {
	return (TextLocation) locations.get(id);
    }

    public int size() {
	return (locationList == null)? locations.size() : locationList.length;
    }

    public TextLocation[] getAll() {
	if (locationList == null) {
	    locationList = new TextLocation[locations.size()];

	    int i = 0;
	    Enumeration iter = locations.elements();
	    while (iter.hasMoreElements()) 
		locationList[i++] = (TextLocation) iter.nextElement();
	}
	
	return locationList;
    }

    public Enumeration locations() {
	return locations.elements();
    }

    public void addTextLocationListener(TextLocationListener listener) {
	if (indexOfTextLocationListener(listener) >= 0)
	    return;

	TextLocationListener[] newListeners = 
	    new TextLocationListener[listeners.length + 1];

	if (listeners.length > 0)
	    System.arraycopy(listeners, 0, newListeners, 0, listeners.length);
	newListeners[listeners.length] = listener;

	listeners = newListeners;
    }

    public void removeTextLocationListener(TextLocationListener listener) {
	int listenerIndex = indexOfTextLocationListener(listener);
	if (listenerIndex < 0)
	    return;

	TextLocationListener[] newListeners = 
	    new TextLocationListener[listeners.length - 1];

	if (listenerIndex > 0)
	    System.arraycopy(listeners, 0, newListeners, 0, listenerIndex);

	int first = listenerIndex + 1;
	int tail = listeners.length - first;
	if (tail > 0) 
	    System.arraycopy(listeners, first, 
			     newListeners, listenerIndex, tail);

	listeners = newListeners;
    }

    private int indexOfTextLocationListener(TextLocationListener listener) {
	for (int i = 0; i < listeners.length; ++i) {
	    if (listeners[i] == listener)
		return i;
	}
	return -1;
    }

    // -----------------------------------------------------------------------

    public void batchEditingStarted(BatchEditingEvent event, 
				    int listenerIndex) {
	batchEditing = true;
    }

    public void batchEditingCompleted(BatchEditingEvent event, 
				      int listenerIndex) {
	batchEditing = false;

	int eventCount = eventBatch.size();
	if (eventCount > 0) {
	    TextLocationEvent[] batch = new TextLocationEvent[eventCount];
	    int i = 0;
	    Enumeration iter = eventBatch.elements();
	    while (iter.hasMoreElements())
		batch[i++] = (TextLocationEvent) iter.nextElement();

	    for (int j = TextLocationEvent.LOCATION_REMOVED; 
		 j >= TextLocationEvent.LOCATION_ADDED;
		 --j) {
		for (i = 0; i < eventCount; ++i) {
		    TextLocationEvent e = batch[i];
		    if (e.getId() == j)
			notifyTextLocationListeners(e);
		}
	    }

	    eventBatch.clear();
	}
    }

    public void textChanged(TextEvent event, int listenerIndex) {
	TextModel source = event.getTextSource();
	int eventOffset = event.getOffset();

	TextLocation[] list = getAll();
	for (int i = 0; i < list.length; ++i) {
	    TextLocation location = list[i];
	    int offset;

	    if (location.getText() != source ||
		(offset = location.getOffset()) < eventOffset)
		continue;

	    int inserted = 0;
	    int deleted = 0;

	    switch (event.getId()) {
	    case TextEvent.TEXT_INSERTED:
		inserted = event.getNewString().length();
		location.moveTo(source, offset + inserted);
		break;

	    case TextEvent.TEXT_REPLACED:
		inserted = event.getNewString().length();
		/*FALLTHROUGH*/

	    case TextEvent.TEXT_DELETED:
		deleted = event.getOldString().length();
		if (eventOffset + deleted <= offset) {
		    location.moveTo(source, offset - deleted + inserted);
		} else {
		    if (location.isCursor()) 
			location.moveTo(source, eventOffset + inserted);
		    else
			remove(location);
		}
		break;
	    }
	}
    }

    public void elementStructureChanged(ElementStructureEvent event, 
					int listenerIndex) {
	TextLocation[] list = getAll();

	switch (event.getId()) {
	case ElementStructureEvent.CHILD_ADDED:
	    {
		TextModel newText = (TextModel) 
		    Traversal.traverse(event.getNewChild(), 
				       Traversal.textFinder);

		if (newText != null) {
		    for (int i = 0; i < list.length; ++i) {
			TextLocation location = list[i];

			if (location.isCursor() && location.getText() == null)
			    location.moveTo(newText, 0);
		    }
		}
	    }
	    break;

	case ElementStructureEvent.CHILD_REPLACED:
	case ElementStructureEvent.CHILD_REMOVED:
	    {
		SubelementModel oldChild = event.getOldChild();

		for (int i = 0; i < list.length; ++i) {
		    TextLocation location = list[i];
		    TextModel text = location.getText();

		    if (text != null) { 
			textMatcher.searchedText = text;
			if (Traversal.traverse(oldChild,textMatcher) != null) {
			    // oldChild is or contains text.

			    if (location.isCursor()) 
				moveCursor(location, event);
			    else
				remove(location);
			}
		    }
		}
	    }
	    break;
	}
    }

    private void moveCursor(TextLocation cursor, ElementStructureEvent event) {
	ElementModel source = event.getElementSource();
	TextModel newText;
	int newOffset;

	// Search BEFORE deleted node.

	SubelementModel from = event.getReferenceChild();
	if (from == null) {
	    // Was last node. 
	    from = source.getLastSubelement();

	    if (from == null)
		// Source no longer has subelements.
		newText = (TextModel) 
		    Traversal.traverseBefore(source, Traversal.textFinder);
	    else
		newText = (TextModel) 
		    Traversal.traverseBackwardFrom(from,Traversal.textFinder);
	} else {
	    newText = (TextModel) 
		Traversal.traverseBefore(from, Traversal.textFinder);
	}

	if (newText == null) {
	    // Search AFTER deleted node. There's nothing BEFORE deleted node
	    // that's why the fact that search begins at source is not a
	    // problem.

	    newText = (TextModel) 
		Traversal.traverseFrom(source, Traversal.textFinder);
	    newOffset = 0;
	} else {
	    newOffset = newText.getTextLength();
	}

	cursor.moveTo(newText, newOffset);
    }

    public void documentEventHappened(DocumentEvent event, int listenerIndex) {
	if (event instanceof TextTransferEvent) {
	    // The TextTransferEvent is sent after the text has been
	    // transfered to the destination and before the source is detached
	    // from the document.

	    TextTransferEvent e = (TextTransferEvent) event;

	    TextLocation[] list = getAll();
	    for (int i = 0; i < list.length; ++i) {
		TextLocation location = list[i];
		if (!location.isCursor())
		    continue;

		if (location.getText() == e.srcText) {
		    location.moveTo(
			e.dstText, 
			e.dstOffset + (location.getOffset() - e.srcOffset));
		} else if (location.getText() == e.dstText) {
		    int endOffset = e.dstText.getTextLength();

		    if (e.dstOffset + e.charCount == endOffset &&
			location.getOffset() == endOffset)
			location.moveTo(e.dstText, e.dstOffset);
		}
	    }
	}
    }

    public void elementAttributeChanged(ElementAttributeEvent event, 
					int listenerIndex) {}

    // -----------------------------------------------------------------------

    /*package*/ void notifyTextLocationListeners(TextLocation source, int id,
						 TextModel previousText,
						 int previousOffset) {
	textLocationEvent.initialize(source, id, previousText, previousOffset);

	notifyTextLocationListeners(textLocationEvent);
    }

    private void notifyTextLocationListeners(TextLocationEvent event) {
	int i;

	switch (event.getId()) {
	case TextLocationEvent.LOCATION_ADDED:
	    // Doesn't happen during a batch.
	    for (i = 0; i < listeners.length; ++i)
		listeners[i].textLocationAdded(event);
	    break;

	case TextLocationEvent.LOCATION_MOVED:
	    if (batchEditing) {
		TextLocation source = event.getSource();
		Object key = source.getId();

		TextLocationEvent savedEvent = 
		    (TextLocationEvent) eventBatch.get(key);

		if (savedEvent == null) {
		    savedEvent = new TextLocationEvent(
			source, TextLocationEvent.LOCATION_MOVED,
			event.getPreviousText(), event.getPreviousOffset());
		    eventBatch.put(key, savedEvent);
		}
		// Otherwise, nothing to do: the source in the savedEvent
		// changes but not the original text and offset.

	    } else {
		for (i = 0; i < listeners.length; ++i)
		    listeners[i].textLocationMoved(event);
	    }
	    break;

	case TextLocationEvent.LOCATION_REMOVED:
	    if (batchEditing) {
		TextLocation source = event.getSource();
		Object key = source.getId();

		TextLocationEvent savedEvent = 
		    (TextLocationEvent) eventBatch.get(key);

		if (savedEvent == null) {
		    savedEvent = new TextLocationEvent(
			source, TextLocationEvent.LOCATION_REMOVED,
			event.getPreviousText(), event.getPreviousOffset());
		    eventBatch.put(key, savedEvent);
		} else {
		    // Transform the MOVED event into a REMOVED event
		    // but keep the original text and offset.

		    savedEvent.initialize(
			source, TextLocationEvent.LOCATION_REMOVED,
			savedEvent.getPreviousText(), 
			savedEvent.getPreviousOffset());
		}

	    } else {
		for (i = 0; i < listeners.length; ++i)
		    listeners[i].textLocationRemoved(event);
	    }
	    break;
	}
    }
}

