/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import com.xmlmind.xmledit.model.*;

public class Traversal {
    public static final Object LEAVE_ELEMENT = new Object();

    public interface Handler {
	Object processText(TextModel text);
	Object enterElement(ElementModel element);
	Object leaveElement(ElementModel element);
    }

    public static class HandlerBase implements Handler {
	public Object processText(TextModel text) {
	    return null;
	}

	public Object enterElement(ElementModel element) {
	    return null;
	}

	public Object leaveElement(ElementModel element) {
	    return null;
	}
    }

    // -----------------------------------------------------------------------

    public static class TextFinder extends HandlerBase {
	public Object processText(TextModel text) {
	    return text;
	}
    }
    public static final TextFinder textFinder = new TextFinder();

    public static class TextMatcher extends HandlerBase {
	public TextModel searchedText;

	public Object processText(TextModel text) {
	    if (text == searchedText)
		return text;
	    else
		return null;
	}
    }

    public static class TextGrabber extends HandlerBase {
	public StringBuffer grabbed = new StringBuffer();

	public Object processText(TextModel text) {
	    grabbed.append(text.getText());
	    return null;
	}
    }

    // -----------------------------------------------------------------------

    public static Object traverse(SubelementModel node, Handler handler) {
	switch (node.getSubelementType()) {
	case SubelementModel.TEXT:
	    return handler.processText((TextModel) node);

	case SubelementModel.ELEMENT:
	    {
		ElementModel element = (ElementModel) node;
		Object result;

		result = handler.enterElement(element);
		if (result != null)
		    return (result == LEAVE_ELEMENT)? null : result;

		SubelementModel child = element.getFirstSubelement();
		while (child != null) {
		    result = traverse(child, handler);
		    if (result != null)
			return result;

		    child = child.getNextSubelement();
		}

		result = handler.leaveElement(element);
		if (result != null)
		    return result;
	    }
	    return null;

	default:
	    return null;
	}
    }

    // -----------------------------------------------------------------------

    public static Object traverseBackward(SubelementModel node, 
					  Handler handler) {
	switch (node.getSubelementType()) {
	case SubelementModel.TEXT:
	    return handler.processText((TextModel) node);

	case SubelementModel.ELEMENT:
	    {
		ElementModel element = (ElementModel) node;
		Object result;

		result = handler.enterElement(element);
		if (result != null)
		    return (result == LEAVE_ELEMENT)? null : result;

		SubelementModel child = element.getLastSubelement();
		while (child != null) {
		    result = traverseBackward(child, handler);
		    if (result != null)
			return result;

		    child = child.getPreviousSubelement();
		}

		result = handler.leaveElement(element);
		if (result != null)
		    return result;
	    }
	    return null;

	default:
	    return null;
	}
    }

    // -----------------------------------------------------------------------

    public static Object traverseAfter(SubelementModel node, Handler handler) {
	SubelementModel sibling = node.getNextSubelement();
	while (sibling != null) {
	    Object result = traverse(sibling, handler);
	    if (result != null)
		return result;

	    sibling = sibling.getNextSubelement();
	}

	for (;;) {
	    ElementModel parent = node.getParentElement();
	    if (parent == null)
		return null;

	    SubelementModel from = parent.getNextSubelement();
	    if (from != null)
		return traverseFrom(from, handler);

	    node = parent;
	}
    }

    public static Object traverseFrom(SubelementModel node, Handler handler) {
	SubelementModel sibling = node;
	while (sibling != null) {
	    Object result = traverse(sibling, handler);
	    if (result != null)
		return result;

	    sibling = sibling.getNextSubelement();
	}

	for (;;) {
	    ElementModel parent = node.getParentElement();
	    if (parent == null)
		return null;

	    SubelementModel from = parent.getNextSubelement();
	    if (from != null)
		return traverseFrom(from, handler);

	    node = parent;
	}
    }

    // -----------------------------------------------------------------------

    public static Object traverseBefore(SubelementModel node, 
					Handler handler) {
	SubelementModel sibling = node.getPreviousSubelement();
	while (sibling != null) {
	    Object result = traverseBackward(sibling, handler);
	    if (result != null)
		return result;

	    sibling = sibling.getPreviousSubelement();
	}

	for (;;) {
	    ElementModel parent = node.getParentElement();
	    if (parent == null)
		return null;

	    SubelementModel from = parent.getPreviousSubelement();
	    if (from != null)
		return traverseBackwardFrom(from, handler);

	    node = parent;
	}
    }

    public static Object traverseBackwardFrom(SubelementModel node, 
					      Handler handler) {
	SubelementModel sibling = node;
	while (sibling != null) {
	    Object result = traverseBackward(sibling, handler);
	    if (result != null)
		return result;

	    sibling = sibling.getPreviousSubelement();
	}

	for (;;) {
	    ElementModel parent = node.getParentElement();
	    if (parent == null)
		return null;

	    SubelementModel from = parent.getPreviousSubelement();
	    if (from != null)
		return traverseBackwardFrom(from, handler);

	    node = parent;
	}
    }
}

