/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.modeltool;

import com.xmlmind.xmledit.model.*;

public class UndoManager implements DocumentListener {
    public static final int DEFAULT_UNDO_MAX_COUNT = 10;

    public static abstract class Action {
	protected String description = null;

	public void setDescription(String description) {
	    this.description = description;
	}

	public String getDescription() {
	    return description;
	}

	public abstract DocumentEvent[] getDocumentEvents();
    }

    private DocumentModel doc = null;
    private int limit = DEFAULT_UNDO_MAX_COUNT;
    private UndoActionStack undoStack = 
        new UndoActionStack(DEFAULT_UNDO_MAX_COUNT);
    private UndoActionStack redoStack = 
        new UndoActionStack(DEFAULT_UNDO_MAX_COUNT);

    private UndoManagerListener[] listeners = new UndoManagerListener[0];
    private UndoManagerEvent undoManagerEvent = new UndoManagerEvent(this);

    private UndoAction targetAction = null;
    private UndoActionStack targetStack = null;

    private static final class UndoAction extends Action {
	private DocumentEvent[] events;

	public UndoAction(DocumentEvent event) {
	    events = new DocumentEvent[] { event.duplicate() };
	}

	public UndoAction() {
	    events = new DocumentEvent[0];
	}

	public DocumentEvent[] getDocumentEvents() {
	    return events;
	}

	public void add(DocumentEvent event) {
	    int count = events.length;
	    DocumentEvent[] newEvents = new DocumentEvent[count+1];
	    System.arraycopy(events, 0, newEvents, 0, count);
	    events = newEvents;

	    events[count] = event.duplicate();
	}

	public boolean mergeTextEvent(TextEvent event) {
	    int eventId = event.getId();
	    TextModel text = event.getTextSource();

	    DocumentEvent lastEvent = null;
	    if (events.length == 0 || 
		(lastEvent = events[events.length-1]).getId() != eventId ||
		lastEvent.getSource() != text)
		return false;

	    switch (eventId) {
	    case TextEvent.TEXT_INSERTED:
	        {
		    String srcNewString = event.getNewString();
		    if (srcNewString.length() > 1)
			return false;

		    TextEvent dst = (TextEvent) lastEvent;
		    int dstOffset = dst.getOffset();
		    String dstNewString = dst.getNewString();
		    
		    if (dstOffset+dstNewString.length() == event.getOffset()) {
			dst.initialize(text, eventId, dstOffset, null, 
				       dstNewString + srcNewString);
			return true;
		    }
	        }
		break;

	    case TextEvent.TEXT_DELETED:
	        {
		    String srcOldString = event.getOldString();
		    int srcOldStringLength = srcOldString.length();
		    if (srcOldStringLength > 1)
			return false;

		    TextEvent dst = (TextEvent) lastEvent;
		    int dstOffset = dst.getOffset();
		    String dstOldString = dst.getOldString();
		    int srcOffset = event.getOffset();
		    
		    if (srcOffset + srcOldStringLength == dstOffset) {
			// Deleting chars with the Backspace key.
			dst.initialize(text, eventId, srcOffset, 
				       srcOldString + dstOldString, null);
			return true;
		    } else if (dstOffset == srcOffset) {
			// Deleting chars with the Del key.
			dst.initialize(text, eventId, dstOffset, 
				       dstOldString + srcOldString, null);
			return true;
		    }
	        }
		break;
	    }

	    return false;
	}

	public SubelementModel perform(DocumentModel doc) {
	    doc.beginBatchEditing();
	    
	    boolean normalizeParent = false;
	    for (int i = events.length-1; i >= 0; --i) {
		DocumentEvent event = events[i];

		switch (event.getId()) {
		case TextEvent.TEXT_INSERTED:
		    {
			TextEvent e = (TextEvent) event;

			e.getTextSource().deleteText(
			    e.getOffset(), e.getNewString().length());
		    }
		    break;
		case TextEvent.TEXT_DELETED:
		    {
			TextEvent e = (TextEvent) event;

			e.getTextSource().insertText(
			    e.getOffset(), e.getOldString());
		    }
		    break;
		case TextEvent.TEXT_REPLACED:
		    {
			TextEvent e = (TextEvent) event;

			e.getTextSource().replaceText(
			    e.getOffset(), e.getNewString().length(),
			    e.getOldString());
		    }
		    break;
		case ElementAttributeEvent.ATTRIBUTE_ADDED:
		    {
			ElementAttributeEvent e = (ElementAttributeEvent)event;

			e.getElementSource().removeElementAttribute(
			    e.getName());
		    }
		    break;
		case ElementAttributeEvent.ATTRIBUTE_REMOVED:
		case ElementAttributeEvent.ATTRIBUTE_CHANGED:
		    {
			ElementAttributeEvent e = (ElementAttributeEvent)event;

			e.getElementSource().putElementAttribute(
			    e.getName(), e.getOldValue());
		    }
		    break;
		case ElementStructureEvent.CHILD_ADDED:
		    {
			ElementStructureEvent e = (ElementStructureEvent)event;
			ElementModel parent = e.getElementSource();

			parent.removeSubelement(e.getNewChild());
			normalizeParent = true;
		    }
		    break;
		case ElementStructureEvent.CHILD_REMOVED:
		    {
			ElementStructureEvent e = (ElementStructureEvent)event;
			ElementModel parent = e.getElementSource();

			parent.insertSubelement(e.getReferenceChild(),
						e.getOldChild());
			normalizeParent = true;
		    }
		    break;
		case ElementStructureEvent.CHILD_REPLACED:
		    {
			ElementStructureEvent e = (ElementStructureEvent)event;
			ElementModel parent = e.getElementSource();

			parent.replaceSubelement(e.getNewChild(), 
						 e.getOldChild());
			normalizeParent = true;
		    }
		    break;
		}
	    }

	    if (normalizeParent) {
		for (int i = events.length-1; i >= 0; --i) {
		    DocumentEvent event = events[i];

		    switch (event.getId()) {
		    case ElementStructureEvent.CHILD_ADDED:
			{
			    ElementStructureEvent e = 
				(ElementStructureEvent) event;

			    ElementUtil.normalizeTextContainer(
				e.getElementSource(), doc);
			}
			break;
		    case ElementStructureEvent.CHILD_REMOVED:
		    case ElementStructureEvent.CHILD_REPLACED:
			{
			    ElementStructureEvent e = 
				(ElementStructureEvent) event;

			    ElementUtil.mergeContiguousTexts(
				e.getElementSource());
			}
			break;
		    }
		}
	    }

	    SubelementModel modified = null;
	    loop: for (int i = events.length-1; i >= 0; --i) {
		DocumentEvent event = events[i];

		switch (event.getId()) {
		case TextEvent.TEXT_INSERTED:
		case TextEvent.TEXT_DELETED:
		case TextEvent.TEXT_REPLACED:
		    {
			TextEvent e = (TextEvent) event;
			TextModel text = e.getTextSource();

			if (text.getDocument() != null) {
			    modified = text;
			    break loop;
			}
		    }
		    break;
		case ElementAttributeEvent.ATTRIBUTE_ADDED:
		case ElementAttributeEvent.ATTRIBUTE_REMOVED:
		case ElementAttributeEvent.ATTRIBUTE_CHANGED:
		    {
			ElementAttributeEvent e = (ElementAttributeEvent)event;
			ElementModel element = e.getElementSource();

			if (element.getDocument() != null) {
			    modified = element;
			    break loop;
			}
		    }
		    break;
		case ElementStructureEvent.CHILD_ADDED:
		case ElementStructureEvent.CHILD_REMOVED:
		case ElementStructureEvent.CHILD_REPLACED:
		    {
			ElementStructureEvent e = (ElementStructureEvent)event;
			ElementModel element = e.getElementSource();

			if (element.getDocument() != null) {
			    modified = element;
			    break loop;
			}
		    }
		    break;
		}
	    }

	    doc.endBatchEditing();

	    return modified;
	}

	public String toString() {
	    StringBuffer buffer = new StringBuffer();

	    buffer.append(getClass().getName());
	    buffer.append('[');
	    if (description == null) {
		buffer.append(description);
	    } else {
		buffer.append("description=\"");
		buffer.append(description);
		buffer.append('\"');
	    }
	    for (int i = 0; i < events.length; ++i) {
		buffer.append(",\nevents[");
		buffer.append(i);
		buffer.append("]=");
		buffer.append(events[i]);
	    }
	    buffer.append("\n]");
	    
	    return buffer.toString();
	}
    }

    private static final class UndoActionStack {
	private UndoAction[] actions;
	private int count = 0;

	public UndoActionStack(int capacity) {
	    actions = new UndoAction[capacity];
	}

	public UndoAction top() {
	    return (count > 0)? actions[count-1] : null;
	}

	public UndoAction pop() {
	    if (count > 0) {
		UndoAction top = actions[count-1];
		// Help gc.
		actions[--count] = null;

		return top;
	    } else {
		return null;
	    }
	}

	public void popAll() {
	    if (count > 0) {
		// Help gc.
		for (int i = 0; i < count; ++i)
		    actions[i] = null;

		count = 0;
	    }
	}

	public void push(UndoAction action) {
	    if (count == actions.length)
		System.arraycopy(actions, 1, actions, 0, --count);

	    actions[count++] = action;
	}

	public boolean pushTextEvent(TextEvent event) {
	    if (count == 0 || !actions[count-1].mergeTextEvent(event)) {
		push(new UndoAction(event));
		return true;
	    } else {
		return false;
	    }
	}
    }

    // -----------------------------------------------------------------------

    public UndoManager() {
	this(null);
    }

    public UndoManager(DocumentModel document) {
	setDocument(document);
    }

    public void setDocument(DocumentModel document) {
	if (doc != null) {
	    reset();
	    doc.removeDocumentListener(this);
	    doc = null;
	}

	if (document != null) {
	    doc = document;
	    document.addDocumentListener(this, false);
	}
    }

    public DocumentModel getDocument() {
	return doc;
    }

    public void setUndoMaxCount(int count) {
	if (count < 1)
	    count = 1;
	limit = count;

	reset();
    }

    public int getUndoMaxCount() {
	return limit;
    }

    public void reset() {
	undoStack = new UndoActionStack(DEFAULT_UNDO_MAX_COUNT);
	redoStack = new UndoActionStack(DEFAULT_UNDO_MAX_COUNT);

	notifyUndoManagerListeners();
    }

    public Action getTopUndoAction() {
	return undoStack.top();
    }

    public Action getTopRedoAction() {
	return redoStack.top();
    }

    public boolean canUndo() {
	return (undoStack.top() != null);
    }

    public boolean canRedo() {
	return (redoStack.top() != null);
    }

    public SubelementModel undo() {
	UndoAction action = undoStack.pop();
	if (action == null)
	    return null;

	targetStack = redoStack;
	SubelementModel modified = action.perform(doc);

	UndoAction inverseAction = redoStack.top();
	inverseAction.setDescription(action.getDescription());

	notifyUndoManagerListeners();
	return modified;
    }

    public SubelementModel redo() {
	UndoAction action = redoStack.pop();
	if (action == null)
	    return null;

	targetStack = undoStack;
	SubelementModel modified = action.perform(doc);

	UndoAction inverseAction = undoStack.top();
	inverseAction.setDescription(action.getDescription());

	notifyUndoManagerListeners();
	return modified;
    }

    // -----------------------------------------------------------------------

    public void addUndoManagerListener(UndoManagerListener listener) {
	if (indexOfUndoManagerListener(listener) >= 0)
	    return;

	UndoManagerListener[] newListeners = 
	    new UndoManagerListener[listeners.length + 1];

	if (listeners.length > 0)
	    System.arraycopy(listeners, 0, newListeners, 0, listeners.length);
	newListeners[listeners.length] = listener;

	listeners = newListeners;
    }

    public void removeUndoManagerListener(UndoManagerListener listener) {
	int listenerIndex = indexOfUndoManagerListener(listener);
	if (listenerIndex < 0)
	    return;

	UndoManagerListener[] newListeners = 
	    new UndoManagerListener[listeners.length - 1];

	if (listenerIndex > 0)
	    System.arraycopy(listeners, 0, newListeners, 0, listenerIndex);

	int first = listenerIndex + 1;
	int tail = listeners.length - first;
	if (tail > 0) 
	    System.arraycopy(listeners, first, 
			     newListeners, listenerIndex, tail);

	listeners = newListeners;
    }

    private int indexOfUndoManagerListener(UndoManagerListener listener) {
	for (int i = 0; i < listeners.length; ++i) {
	    if (listeners[i] == listener)
		return i;
	}
	return -1;
    }

    private void notifyUndoManagerListeners() {
	for (int i = 0; i < listeners.length; ++i)
	    listeners[i].undoManagerChange(undoManagerEvent);
    }

    // -----------------------------------------------------------------------

    public void batchEditingStarted(BatchEditingEvent event, 
				    int listenerIndex) {
	targetAction = new UndoAction();
    }

    public void batchEditingCompleted(BatchEditingEvent event, 
				      int listenerIndex) {
	if (targetStack == null) {
	    // Normal editing (i.e. not undoing or redoing).
	    undoStack.push(targetAction);
	    redoStack.popAll();

	    notifyUndoManagerListeners();
	} else {
	    targetStack.push(targetAction);
	    targetStack = null;
	}
	targetAction = null;
    }

    public void elementAttributeChanged(ElementAttributeEvent event, 
					int listenerIndex) {
	pushEvent(event);
    }

    private void pushEvent(DocumentEvent event) {
	if (targetAction == null) {
	    if (targetStack == null) {
		// Normal editing (i.e. not undoing or redoing).
		undoStack.push(new UndoAction(event));
		redoStack.popAll();

		notifyUndoManagerListeners();
	    } else {
		targetStack.push(new UndoAction(event));
		targetStack = null;
	    }
	} else {
	    // Batch editing.
	    targetAction.add(event);
	}
    }

    public void elementStructureChanged(ElementStructureEvent event, 
					int listenerIndex) {
	pushEvent(event);
    }

    public void textChanged(TextEvent event, int listenerIndex) {
	if (targetAction == null) {
	    if (targetStack == null) {
		// Normal editing (i.e. not undoing or redoing).
		boolean pushed = undoStack.pushTextEvent(event);
		redoStack.popAll();

		// If the TextEvent has not been pushed, no need to check for
		// a redo stack change: it has not changed because the redo
		// stack has already been emptied by a preceeding TextEvent.
		if (pushed)
		    notifyUndoManagerListeners();
	    } else {
		targetStack.pushTextEvent(event);
		targetStack = null;
	    }
	} else {
	    // Batch editing.
	    targetAction.add(event);
	}
    }

    public void documentEventHappened(DocumentEvent event,
				      int listenerIndex) {}
}
