/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledgadget;

import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Image;
import com.xmlmind.gadget.Gadget;

public abstract class Box extends Gadget implements Styled {
    public static final Style DEFAULT_STYLE = new Style();

    protected Style style;

    // -----------------------------------------------------------------------

    protected Box() {
	this(DEFAULT_STYLE);
    }

    protected Box(Style style) {
	this.style = style;
    }

    public void setStyle(Style style) {
	this.style = style;
    }

    public Style getStyle() {
	return style;
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	int decorWidth = 
	    2*style.borderWidth + style.paddingLeft + style.paddingRight;
	int decorHeight = 
	    2*style.borderWidth + style.paddingTop + style.paddingBottom;

	if (constrainedWidth > 0) {
	    constrainedWidth -= decorWidth;
	    if (constrainedWidth <= 0)
		constrainedWidth = 1;
	}
	if (constrainedHeight > 0) {
	    constrainedHeight -= decorHeight;
	    if (constrainedHeight <= 0)
		constrainedHeight = 1;
	}

	layoutContent(constrainedWidth, constrainedHeight, recursive);

	width += decorWidth;
	height += decorHeight;
    }

    protected abstract void layoutContent(int constrainedWidth, 
					  int constrainedHeight,
					  boolean recursive);

    public void paint(Graphics g) {
	Box.drawBox(g, style, width, height);

	g.translate(style.borderWidth + style.paddingLeft,
		    style.borderWidth + style.paddingTop);
	paintContent(g);
    }

    protected abstract void paintContent(Graphics g);

    // -----------------------------------------------------------------------

    public static void drawBox(Graphics g, Style style, 
			       int width, int height) {
	// This function is able to draw very tall boxes (> 65000px) despite
	// the fact Java 1.1 Graphics.fillRect, .drawLine, etc seem to give up
	// when the dimensions involved exceed 65000px.
	// In order to speed up the code, very wide boxes are not supported
	// the same way. (It is unlikely to create very wide boxes.)

	Rectangle clip = null;

	if (style.backgroundColor != null) {
	    g.setColor(style.backgroundColor);
	    clip = g.getClipBounds();
	    g.fillRect(clip.x, clip.y, clip.width, clip.height);
	}

	if (style.borderStyle == Style.BORDER_STYLE_NONE || 
	    style.borderWidth <= 0)
	    return;

	boolean topSide = (style.borderTopColor != null);
	boolean rightSide = (style.borderRightColor != null);
	boolean bottomSide = (style.borderBottomColor != null);
	boolean leftSide = (style.borderLeftColor != null);
	if (!topSide && !rightSide && !bottomSide && !leftSide)
	    return;

	if (clip == null)
	    clip = g.getClipBounds();
	
	int thick1 = 0, thick2 = 0, thick3 = 0;
	Color topColor1 = null, topColor3 = null;
	Color rightColor1 = null, rightColor3 = null;
	Color bottomColor1 = null, bottomColor3 = null;
	Color leftColor1 = null, leftColor3 = null;

	switch (style.borderStyle) {
	case Style.BORDER_STYLE_DOUBLE:
	    thick1 = thick2 = style.borderWidth / 3;
	    thick3 = style.borderWidth - 2*thick1;
	    if (topSide)
		topColor1 = topColor3 = style.borderTopColor;
	    if (rightSide)
		rightColor1 = rightColor3 = style.borderRightColor;
	    if (bottomSide)
		bottomColor1 = bottomColor3 = style.borderBottomColor;
	    if (leftSide)
		leftColor1 = leftColor3 = style.borderLeftColor;
	    break;
	case Style.BORDER_STYLE_OUTSET:
	    thick1 = style.borderWidth;
	    if (topSide)
		topColor1 = style.borderTopColor.brighter();
	    if (leftSide)
		leftColor1 = style.borderLeftColor.brighter();
	    if (rightSide)
		rightColor1 = style.borderRightColor.darker();
	    if (bottomSide)
		bottomColor1 = style.borderBottomColor.darker();
	    break;
	case Style.BORDER_STYLE_INSET:
	    thick1 = style.borderWidth;
	    if (topSide)
		topColor1 = style.borderTopColor.darker();
	    if (leftSide)
		leftColor1 = style.borderLeftColor.darker();
	    if (rightSide)
		rightColor1 = style.borderRightColor.brighter();
	    if (bottomSide)
		bottomColor1 = style.borderBottomColor.brighter();
	    break;
	case Style.BORDER_STYLE_GROOVE:
	    thick1 = style.borderWidth / 2;
	    thick2 = 0;
	    thick3 = style.borderWidth - thick1;
	    if (topSide) {
		topColor1 = style.borderTopColor.darker();
		topColor3 = style.borderTopColor.brighter();
	    }
	    if (leftSide) {
		leftColor1 = style.borderLeftColor.darker();
		leftColor3 = style.borderLeftColor.brighter();
	    }
	    if (rightSide) {
		rightColor1 = style.borderRightColor.brighter();
		rightColor3 = style.borderRightColor.darker();
	    }
	    if (bottomSide) {
		bottomColor1 = style.borderBottomColor.brighter();
		bottomColor3 = style.borderBottomColor.darker();
	    }
	    break;
	case Style.BORDER_STYLE_RIDGE:
	    thick1 = style.borderWidth / 2;
	    thick2 = 0;
	    thick3 = style.borderWidth - thick1;
	    if (topSide) {
		topColor1 = style.borderTopColor.brighter();
		topColor3 = style.borderTopColor.darker();
	    }
	    if (leftSide) {
		leftColor1 = style.borderLeftColor.brighter();
		leftColor3 = style.borderLeftColor.darker();
	    }
	    if (rightSide) {
		rightColor1 = style.borderRightColor.darker();
		rightColor3 = style.borderRightColor.brighter();
	    }
	    if (bottomSide) {
		bottomColor1 = style.borderBottomColor.darker();
		bottomColor3 = style.borderBottomColor.brighter();
	    }
	    break;
	case Style.BORDER_STYLE_SOLID:
	case Style.BORDER_STYLE_DOTTED:
	case Style.BORDER_STYLE_DASHED:
	default:
	    thick1 = style.borderWidth;
	    if (topSide)
		topColor1 = style.borderTopColor;
	    if (rightSide)
		rightColor1 = style.borderRightColor;
	    if (bottomSide)
		bottomColor1 = style.borderBottomColor;
	    if (leftSide)
		leftColor1 = style.borderLeftColor;
	    break;
	}

	int clipY1 = clip.y;
	int clipY2 = clipY1 + clip.height - 1;
	int x1, y1, x2, y2;

	if (topSide && clipY1 < style.borderWidth) {
	    x1 = 0;
	    y2 = y1 = 0;
	    x2 = width - 1;

	    g.setColor(topColor1);
	    for (int i = 0; i < thick1; ++i) {
		g.drawLine(x1, y1, x2, y2);
		if (leftSide)
		    ++x1;
		++y1;
		if (rightSide)
		    --x2;
		y2 = y1;
	    }

	    if (thick3 > 0) {
		if (leftSide)
		    x1 += thick2;
		y1 += thick2;
		if (rightSide)
		    x2 -= thick2;
		y2 = y1;

		g.setColor(topColor3);
		for (int i = 0; i < thick3; ++i) {
		    g.drawLine(x1, y1, x2, y2);
		    if (leftSide)
			++x1;
		    ++y1;
		    if (rightSide)
			--x2;
		    y2 = y1;
		}
	    }
	}

	if (rightSide && clip.x + clip.width > width - style.borderWidth) {
	    x2 = x1 = width - 1;
	    y1 = 0;
	    y2 = height - 1;

	    g.setColor(rightColor1);
	    for (int i = 0; i < thick1; ++i) {
		g.drawLine(x1, ((y1 < clipY1)? clipY1 : y1), 
			   x2, ((y2 > clipY2)? clipY2 : y2));
		--x1;
		if (topSide)
		    ++y1;
		x2 = x1;
		if (bottomSide)
		    --y2;
	    }

	    if (thick3 > 0) {
		x1 -= thick2;
		if (topSide)
		    y1 += thick2;
		x2 = x1;
		if (bottomSide)
		    y2 -= thick2;

		g.setColor(rightColor3);
		for (int i = 0; i < thick3; ++i) {
		    g.drawLine(x1, ((y1 < clipY1)? clipY1 : y1), 
			       x2, ((y2 > clipY2)? clipY2 : y2));
		    --x1;
		    if (topSide)
			++y1;
		    x2 = x1;
		    if (bottomSide)
			--y2;
		}
	    }
	}

	if (bottomSide && clipY1 + clip.height > height - style.borderWidth) {
	    x1 = 0;
	    y2 = y1 = height - 1;
	    x2 = width - 1;

	    g.setColor(bottomColor1);
	    for (int i = 0; i < thick1; ++i) {
		g.drawLine(x1, y1, x2, y2);
		if (leftSide)
		    ++x1;
		--y1;
		if (rightSide)
		    --x2;
		y2 = y1;
	    }

	    if (thick3 > 0) {
		if (leftSide)
		    x1 += thick2;
		y1 -= thick2;
		if (rightSide)
		    x2 -= thick2;
		y2 = y1;

		g.setColor(bottomColor3);
		for (int i = 0; i < thick3; ++i) {
		    g.drawLine(x1, y1, x2, y2);
		    if (leftSide)
			++x1;
		    --y1;
		    if (rightSide)
			--x2;
		    y2 = y1;
		}
	    }
	}

	if (leftSide && clip.x < style.borderWidth) {
	    x2 = x1 = 0;
	    y1 = 0;
	    y2 = height - 1;

	    g.setColor(leftColor1);
	    for (int i = 0; i < thick1; ++i) {
		g.drawLine(x1, ((y1 < clipY1)? clipY1 : y1), 
			   x2, ((y2 > clipY2)? clipY2 : y2));
		++x1;
		if (topSide)
		    ++y1;
		x2 = x1;
		if (bottomSide)
		    --y2;
	    }

	    if (thick3 > 0) {
		x1 += thick2;
		if (topSide)
		    y1 += thick2;
		x2 = x1;
		if (bottomSide)
		    y2 -= thick2;

		g.setColor(leftColor3);
		for (int i = 0; i < thick3; ++i) {
		    g.drawLine(x1, ((y1 < clipY1)? clipY1 : y1), 
			       x2, ((y2 > clipY2)? clipY2 : y2));
		    ++x1;
		    if (topSide)
			++y1;
		    x2 = x1;
		    if (bottomSide)
			--y2;
		}
	    }
	}
    }

    public static void getContentSizeConstraints(Gadget gadget, Style style,
						 Dimension constraints) {
	gadget.parent.getSizeConstraints(gadget, constraints);
	if (constraints.width > 0) {
	    constraints.width -= 
		2*style.borderWidth + style.paddingLeft + style.paddingRight;
	    if (constraints.width <= 0)
		constraints.width = 1;
	}
	if (constraints.height > 0) {
	    constraints.height -= 
		2*style.borderWidth + style.paddingTop + style.paddingBottom;
	    if (constraints.height <= 0)
		constraints.height = 1;
	}
    }

    // -----------------------------------------------------------------------

    public static void drawEmptyContent(Graphics g, Style style,
					int width, int height) {
	drawPattern(g, style, Icon.PATTERN1, width, height);
    }

    private static void drawPattern(Graphics g, Style style, int patternId,
				    int width, int height) {
	Icon.Info patternInfo = Icon.list[patternId];
	// Transparent background is required otherwise subelement selection
	// is hidden by the pattern.
	Image pattern = Icon.get(patternInfo.icon, style.color, null);
	int patternW = patternInfo.width;
	int patternH = patternInfo.height;

	for (int y = 0; y < height; y += patternH) {
	    int tileH = Math.min(patternH, height - y);

	    for (int x = 0; x < width; x += patternW) {
		int tileW = Math.min(patternW, width - x);

		g.drawImage(pattern, x, y, x+tileW, y+tileH, 
			    0, 0, tileW, tileH, null);
	    }
	}
    }

    public static void drawEmptyTextContent(Graphics g, Style style, 
					    int width, int height) {
	drawPattern(g, style, Icon.PATTERN2, width, height);
    }

    public static void drawEmptyTextContent2(Graphics g, Style style, 
					     int width, int height) {
	FontMetrics metrics = g.getFontMetrics(style.font);
	int ascent = metrics.getAscent();
	int descent = metrics.getDescent();
	int lineHeight = (int) (0.5 + metrics.getHeight()*style.lineHeight);
	// Simplified form of (lineHeight - ascent - descent)/2 + ascent.
	int lineY = (lineHeight + ascent - descent)/2;

	g.setColor(style.color);

	int x1, y1, x2, y2;

	x1 = 0;
	y1 = lineY;
	x2 = 0;
	y2 = lineY + descent - 1;
	g.drawLine(x1, y1, x2, y2);

	x1 = width - 1;
	y1 = y2;
	g.drawLine(x2, y2, x1, y1);

	x2 = x1;
	y2 = lineY;
	g.drawLine(x1, y1, x2, y2);
    }

    public static Dimension getEmptyContentSize(Gadget gadget, Style style) {
	Dimension size = new Dimension();
	getEmptyContentSize(gadget, style, size);
	return size;
    }

    public static void getEmptyContentSize(Gadget gadget, 
					   Style style, Dimension size) {
	FontMetrics metrics = gadget.getFontMetrics(style.font);

	size.width = metrics.charWidth('M');
	size.height = (int) (0.5 + metrics.getHeight()*style.lineHeight);

	if (size.width < 4)
	    size.width = 4;
	if (size.height < 4)
	    size.height = 4;
    }
}


