/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledgadget;

import java.awt.Graphics;
import java.awt.Dimension;
import com.xmlmind.gadget.*;

public class LabeledItem extends GadgetContainer implements Styled {
    protected Style style;
    protected short startLabelOffset = -1;
    protected short endLabelOffset = -1;

    // -----------------------------------------------------------------------

    public LabeledItem() {
	this(Box.DEFAULT_STYLE);
    }

    public LabeledItem(Style style) {
	this.style = style;
    }

    public void setStyle(Style style) {
	this.style = style;
    }

    public Style getStyle() {
	return style;
    }
    
    public void setStartLabel(Gadget gadget, int offset) {
	// Remove previous start label gadget if any.
	if (startLabelOffset >= 0) 
	    removeChild(0);

	if (gadget != null) {
	    insertChild(0, gadget);
	    if (offset < 0)
		offset = 0;
	    else if (offset > Short.MAX_VALUE)
		offset = Short.MAX_VALUE;
	    startLabelOffset = (short) offset;
	}
    }

    public Gadget getStartLabel() {
	return (startLabelOffset < 0)? null : children[0];
    }

    public int getStartLabelOffset() {
	return startLabelOffset;
    }

    public void setItem(Gadget gadget) {
	Gadget prevItem = getItem();
	if (prevItem != null)
	    removeChild(prevItem);

	if (gadget != null)
	    insertChild(((startLabelOffset < 0)? 0 : 1), gadget);
    }

    public Gadget getItem() {
	if (startLabelOffset < 0) {
	    if (endLabelOffset < 0) 
		// No start label, no end label.
		return (children.length > 0)? children[0] : null;
	    else 
		// No start label, an end label.
		return (children.length > 1)? children[0] : null;
	} else {
	    if (endLabelOffset < 0)
		// A start label, no end label.
		return (children.length > 1)? children[1] : null;
	    else
		// A start label, an end label.
		return (children.length > 2)? children[1] : null;
	}
    }

    public void setEndLabel(Gadget gadget, int offset) {
	// Remove previous end label gadget if any.
	if (endLabelOffset >= 0)
	    removeChild(children.length-1);

	if (gadget != null) {
	    addChild(gadget);
	    if (offset < 0)
		offset = 0;
	    else if (offset > Short.MAX_VALUE)
		offset = Short.MAX_VALUE;
	    endLabelOffset = (short) offset;
	}
    }

    public Gadget getEndLabel() {
	return (endLabelOffset < 0)? null : children[children.length-1];
    }

    public int getEndLabelOffset() {
	return endLabelOffset;
    }

    public void removeChild(int index) {
	if (index == 0 && startLabelOffset >= 0)
	    startLabelOffset = -1;
	else if (index == children.length-1 && endLabelOffset >= 0)
	    endLabelOffset = -1;

	super.removeChild(index);
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	Gadget startLabel = getStartLabel();
	Gadget item = getItem();
	Gadget endLabel = getEndLabel();

	for (int i = 0; i < children.length; ++i) {
	    Gadget child = children[i];

	    if (recursive) {
		if (child == item) {
		    Style childStyle = ((Styled) child).getStyle();
		    int constrainedChildWidth;

		    if (constrainedWidth > 0) {
			constrainedChildWidth = constrainedWidth;
			if (childStyle.marginLeft > 0)
			    constrainedChildWidth -= childStyle.marginLeft;
			if (childStyle.marginRight > 0)
			    constrainedChildWidth -= childStyle.marginRight;
			if (constrainedChildWidth <= 0)
			    constrainedChildWidth = 1;
		    } else {
			constrainedChildWidth = -1;
		    }

		    child.layout(constrainedChildWidth, -1, true);
		} else {
		    // Note that the labels are supposed to have an intrinsic
		    // size.
		    child.layout(-1, -1, true);
		}
	    }

	    child.x = child.y = 0;
	}

	if (item == null) {
	    // This should not occur very often: a LabeledItem is not intended
	    // to be used without an item.

	    item = new Box() {
		protected void layoutContent(int constrainedWidth, 
					     int constrainedHeight,
					     boolean recursive) {
		    width = height = 1;
		}
		protected void paintContent(Graphics g) {}
	    };
	    Style style = new Style();
	    style.marginLeft = style.marginRight = -1;
	    ((Box) item).setStyle(style);
	    item.layout(-1, -1, false);
	} 

	Style itemStyle = ((Styled) item).getStyle();

	width = item.width;
	if (itemStyle.marginLeft > 0)
	    width += itemStyle.marginLeft;
	if (itemStyle.marginRight > 0)
	    width += itemStyle.marginRight;

	if (constrainedWidth > 0 && width < constrainedWidth)
	    width = constrainedWidth;
	if (width == 0)
	    width = 1;

	if (itemStyle.marginLeft >= 0) {
	    // Left aligned.
	    item.x = itemStyle.marginLeft;
	} else {
	    if (itemStyle.marginRight >= 0)
		// Right aligned.
		item.x = width - (item.width + itemStyle.marginRight);
	    else
		// Centered.
		item.x = (width - item.width) / 2;
	}
	// item.y already set to 0.

	if (startLabel != null) {
	    startLabel.x = item.x - (startLabel.width + startLabelOffset);
	    // startLabel.y already set to 0.

	    if ((item instanceof TextLines) &&
		(startLabel instanceof TextLines)) {
		int itemBaseLine = ((TextLines) item).getFirstBaseLine();
		int startLabelBaseLine = 
		    ((TextLines) startLabel).getFirstBaseLine();

		if (itemBaseLine >= 0 && startLabelBaseLine >= 0) {
		    if (itemBaseLine > startLabelBaseLine)
			startLabel.y = itemBaseLine - startLabelBaseLine;
		    else
			item.y = startLabelBaseLine - itemBaseLine;
		}
	    }
	}

	if (endLabel != null) {
	    endLabel.x = item.x + item.width + endLabelOffset;
	    // endLabel.y already set to 0.

	    boolean noBaseLine;
	    if ((item instanceof TextLines) &&
		(endLabel instanceof TextLines)) {
		int itemLastBaseLine = 
		    item.y + ((TextLines) item).getLastBaseLine();
		int endLabelBaseLine = 
		    ((TextLines) endLabel).getFirstBaseLine();

		if (itemLastBaseLine >= 0 && endLabelBaseLine >= 0) {
		    if (itemLastBaseLine > endLabelBaseLine) {
			endLabel.y = itemLastBaseLine - endLabelBaseLine;
		    } else {
			item.y += (endLabelBaseLine - itemLastBaseLine);
			if (startLabel != null)
			    startLabel.y += 
				(endLabelBaseLine - itemLastBaseLine);
		    }
		    noBaseLine = false;
		} else {
		    noBaseLine = true;
		}
	    } else {
		noBaseLine = true;
	    }

	    if (noBaseLine) {
		int itemBottom = item.y+item.height;

		if (itemBottom > endLabel.height) {
		    endLabel.y = itemBottom - endLabel.height;
		} else {
		    item.y += (endLabel.height - itemBottom);
		    if (startLabel != null)
			startLabel.y += (endLabel.height - itemBottom);
		}
	    }
	}

	height = item.y + item.height;
	if (startLabel != null && startLabel.y+startLabel.height > height)
	    height = startLabel.y+startLabel.height;
	if (endLabel != null && endLabel.y+endLabel.height > height)
	    height = endLabel.y+endLabel.height;
	if (height == 0)
	    height = 1;
    }

    public void getSizeConstraints(Gadget child, Dimension constraints) {
	if (child == getItem()) {
	    parent.getSizeConstraints(this, constraints);
	    constraints.height = -1;

	    if (constraints.width > 0) {
		Style childStyle = ((Styled) child).getStyle();
		if (childStyle.marginLeft > 0)
		    constraints.width -= childStyle.marginLeft;
		if (childStyle.marginRight > 0)
		    constraints.width -= childStyle.marginRight;
		if (constraints.width <= 0)
		    constraints.width = 1;
	    }
	} else {
	    // Note that the labels are supposed to have an intrinsic size.
	    constraints.width = -1;
	    constraints.height = -1;
	}
    }

    public void paintRelayout(Gadget child, int prevWidth, int prevHeight,
			      int damagedSide, int damagedLength) {
	Style childStyle;

	if (damagedSide == BOTTOM_SIDE &&
	    child == getItem() &&
	    (childStyle = ((Styled) child).getStyle()).marginLeft >= 0 &&
	    getEndLabel() == null) {
	    Gadget startLabel = getStartLabel();

	    int oldHeight = height;
	    height = child.y + child.height;
	    if (startLabel != null && startLabel.y+startLabel.height > height)
		height = startLabel.y+startLabel.height;
	    if (height == 0)
		height = 1;

	    int fromY = child.y + child.height - damagedLength;
	    if (parent != null)
		parent.paintRelayout(this, width, oldHeight, 
				     BOTTOM_SIDE, height-fromY);
//System.err.println("parent.paintRelayout BOTTOM_SIDE");
	} else {
	    super.paintRelayout(child, prevWidth, prevHeight,
				damagedSide, damagedLength);
//System.err.println("super.paintRelayout");
	}
    }
}
