/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledgadget;

import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.gadget.GadgetContainer;
import com.xmlmind.xmledit.model.TextAccess;

public class Paragraph extends TextLineContainerBox {
    private static final class GadgetList {
	public Gadget[] list;
	public int count;

	public GadgetList(Gadget[] prevList, int prevCount) {
	    list = new Gadget[prevList.length];
	    System.arraycopy(prevList, 0, list, 0, prevCount);
	    count = prevCount;
	}

	public void add(Gadget gadget) {
	    if (count == list.length) {
		Gadget[] newList = new Gadget[(count == 0)? 2 : 2*count];
		System.arraycopy(list, 0, newList, 0, count);
		list = newList;
	    }
	    list[count++] = gadget;
	}

	public Gadget[] toArray() {
	    Gadget[] array = new Gadget[count];
	    System.arraycopy(list, 0, array, 0, count);
	    return array;
	}
    }

    private static final class LineList {
	public int[] list;
	public int count;

	public LineList(int[] list, int count) {
	    this.list = list;
	    this.count = count;
	    if (count > 0)
		add(-1);
	}

	public void check(int offset) {
	    if (count == 0)
		add(offset);
	    else if (list[count-1] < 0)
		list[count-1] = offset;
	}

	public void add(int offset) {
	    if (count == list.length) {
		int[] newList = new int[(count == 0)? 2 : 2*count];
		System.arraycopy(list, 0, newList, 0, count);
		list = newList;
	    }
	    list[count++] = offset;
	}

	public int[] toArray() {
	    int length = (count > 0 && list[count-1] < 0)? count-1 : count;
	    int[] array = new int[length];
	    System.arraycopy(list, 0, array, 0, length);
	    return array;
	}
    }

    private static final char[] EM = { 'M' };

    private int[] lines = new int[0];

    // -----------------------------------------------------------------------
    
    public Paragraph() {
	super();
    }

    public Paragraph(Style style) {
	super(style);
    }

    public ParagraphText getFirstParagraphText(TextAccess text) {
	int index = indexOfFirstParagraphText(text);
	return (index < 0)? null : (ParagraphText) children[index];
    }

    public int indexOfFirstParagraphText(TextAccess text) {
	for (int i = 0; i < children.length; ++i) {
	    Gadget child = children[i];

	    if ((child instanceof ParagraphText) && 
		((ParagraphText) child).getTextAccess() == text)
		return i;
	}
	return -1;
    }

    public void getParagraphTexts(TextAccess text, 
				  com.xmlmind.xmledit.GadgetList list) {
	int first = -1;
	int last = -1;

	for (int i = 0; i < children.length; ++i) {
	    Gadget child = children[i];

	    if ((child instanceof ParagraphText) && 
		((ParagraphText) child).getTextAccess() == text) {
		if (first < 0)
		    first = i;
		last = i;
	    } else {
		if (first >= 0)
		    break;
	    }
	}

	if (first >= 0)
	    list.appendGadgets(children, first, last - first + 1);
    }

    // -----------------------------------------------------------------------

    protected void layoutContent(int constrainedWidth, int constrainedHeight,
				 boolean recursive) {
	layoutContentFromLine(0, constrainedWidth, recursive);
    }

    private int layoutContentFromLine(int fromLine, int constrainedWidth, 
				      boolean recursive) {
	// Split text into lines ---

	if (style.whiteSpace == Style.WHITE_SPACE_NORMAL)
	    wrapWords(fromLine, constrainedWidth, recursive);
	else
	    wrapLines(fromLine, constrainedWidth, recursive);

	// Move gadgets and compute the dimension of the content ---

	FontMetrics metrics = getFontMetrics(style.font);
	int[] lineWidthes;
	if (style.textAlign == Style.TEXT_ALIGN_LEFT) 
	    lineWidthes = null;
	else
	    lineWidthes = new int[lines.length];

	width = height = 0;
	int recomputedHeight= 0;
	if (fromLine > 0) {
	    recomputeSize(fromLine, lineWidthes);
	    recomputedHeight = height;
	}

	for (int i = fromLine; i < lines.length; ++i) {
	    int lineFirst = lines[i];
	    int lineEnd = (i+1 >= lines.length)? children.length : lines[i+1];
	    int lineWidth = 0;
	    int lineTop = Integer.MAX_VALUE;
	    int lineBottom = Integer.MIN_VALUE;
	    int lineHeight = 0;

	    if (lineWidthes == null) { // Left aligned.
		if (style.textIndent > 0) 
		    lineWidth = ((i == 0)? style.textIndent : 0);
		else if (style.textIndent < 0)
		    lineWidth = ((i == 0)? 0 : -style.textIndent);
	    }

	    for (int j = lineFirst; j < lineEnd; ++j) {
		Gadget child = children[j];
		Style childStyle = ((Styled) child).getStyle();

		child.x = lineWidth;
		lineWidth += child.width;

		if (!(child instanceof ParagraphText)) {
		    if (childStyle.marginLeft > 0) {
			child.x += childStyle.marginLeft;
			lineWidth += childStyle.marginLeft;
		    }
		    if (childStyle.marginRight > 0)
			lineWidth += childStyle.marginRight;
		}

		boolean relativeToLine = false;
		switch (childStyle.verticalAlign) {
		case Style.VERTICAL_ALIGN_SUPER:
		case Style.VERTICAL_ALIGN_SUB:
		    {
			int shift = (childStyle.verticalAlign == 
				     Style.VERTICAL_ALIGN_SUPER)? 
			    -metrics.getDescent() : metrics.getDescent();
			int baseLine;

			if ((child instanceof TextLines) &&
			    (baseLine = 
			     ((TextLines) child).getFirstBaseLine()) >= 0)
			    child.y = shift - baseLine;
			else
			    child.y = shift - child.height;
		    }
		    break;
		case Style.VERTICAL_ALIGN_TEXT_TOP:
		    child.y = -metrics.getAscent();
		    break;
		case Style.VERTICAL_ALIGN_TEXT_BOTTOM:
		    child.y = metrics.getDescent() - child.height;
		    break;
		case Style.VERTICAL_ALIGN_MIDDLE:
		    child.y = -(metrics.getAscent()-metrics.getDescent())/2 - 
			child.height/2;
		    break;
		case Style.VERTICAL_ALIGN_BASELINE:
		    {
			int baseLine;

			if ((child instanceof TextLines) &&
			    (baseLine = 
			     ((TextLines) child).getFirstBaseLine()) >= 0)
			    child.y = -baseLine;
			else
			    child.y = -child.height;
		    }
		    break;
		default:
		    relativeToLine = true;
		}
		if (relativeToLine) {
		    if (child.height > lineHeight)
			lineHeight = child.height;
		} else {
		    if (child.y < lineTop)
			lineTop = child.y;
		    if (child.y + child.height > lineBottom)
			lineBottom = child.y + child.height;
		}
	    }

	    if (lineWidth > width)
		width = lineWidth;
	    if (lineWidthes != null)
		lineWidthes[i] = lineWidth;

	    if (lineTop == Integer.MAX_VALUE) {
		lineTop = -lineHeight;
		lineBottom = 0;
	    }
	    if (lineHeight > (lineBottom - lineTop))
		lineTop -= (lineHeight - (lineBottom - lineTop))/2;
	    else
		lineHeight = (lineBottom - lineTop);

	    for (int j = lineFirst; j < lineEnd; ++j) {
		Gadget child = children[j];
		Style childStyle = ((Styled) child).getStyle();

		switch (childStyle.verticalAlign) {
		case Style.VERTICAL_ALIGN_TOP:
		    child.y = height;
		    break;
		case Style.VERTICAL_ALIGN_BOTTOM:
		    child.y = height + lineHeight - child.height;
		    break;
		default:
		    child.y = height + (child.y - lineTop);
		}
	    }

	    if (i == 0)
		firstBaseLine = -lineTop;
	    if (i == lines.length-1)
		lastBaseLine = height - lineTop;

	    height += lineHeight;
	}

	if (constrainedWidth > 0 && width < constrainedWidth)
	    width = constrainedWidth;
	if (width < 3)
	    width = 3;
	if (height < 3)
	    height = 3;

	if (lineWidthes == null) 
	    return (height - recomputedHeight);

	// Center or right justify lines ---

	// First ``unindent'' lines which have not been touched.
	int x0 = style.borderWidth + style.paddingLeft;
	for (int i = 0; i < fromLine; ++i) {
	    int lineFirst = lines[i];
	    int lineEnd = ((i+1 >= lines.length)? 
			   children.length : lines[i+1]);

	    int shift = -children[lineFirst].x + x0;
	    for (int j = lineFirst; j < lineEnd; ++j) 
		children[j].x += shift;
	}

	for (int i = 0; i < lines.length; ++i) {
	    int lineFirst = lines[i];
	    int lineEnd = ((i+1 >= lines.length)? 
			   children.length : lines[i+1]);

	    int shift = 0;
	    switch (style.textAlign) {
	    case Style.TEXT_ALIGN_RIGHT:
		shift = width - lineWidthes[i];
		break;
	    case Style.TEXT_ALIGN_CENTER:
		shift = (width - lineWidthes[i])/2;
		break;
	    }
	    for (int j = lineFirst; j < lineEnd; ++j) 
		children[j].x += shift;
	}

	return height;
    }

    private void recomputeSize(int fromLine, int[] lineWidthes) {
	for (int i = 0; i < fromLine; ++i) {
	    int lineFirst = lines[i];
	    int lineEnd = ((i+1 >= lines.length)? 
			   children.length : lines[i+1]);
	    int lineTop = Integer.MAX_VALUE;
	    int lineBottom = Integer.MIN_VALUE;
	    Gadget child;

	    for (int j = lineFirst; j < lineEnd; ++j) {
		child = children[j];

		if (child.y < lineTop)
		    lineTop = child.y;
		if (child.y + child.height > lineBottom)
		    lineBottom = child.y + child.height;
	    }

	    child = children[lineEnd-1];
	    int lineWidth = (child.x + child.width) - children[lineFirst].x;
	    if (!(child instanceof ParagraphText)) {
		Style childStyle = ((Styled) child).getStyle();

		if (childStyle.marginRight > 0)
		    lineWidth += childStyle.marginRight;
	    }

	    if (lineWidth > width)
		width = lineWidth;
	    if (lineWidthes != null)
		lineWidthes[i] = lineWidth;
	    
	    // firstBaseLine, lastBaseLine not changed.

	    height += (lineBottom - lineTop);
	}
    }

    private void wrapLines(int fromLine, int constrainedWidth, 
			   boolean recursive) {
	int fromChild = (lines.length > 0)? lines[fromLine] : 0;
	GadgetList childList = new GadgetList(children, fromChild);
	LineList lineList = new LineList(lines, fromLine);

	for (int i = fromChild; i < children.length; ++i) {
	    Gadget child = children[i];

	    if (child instanceof ParagraphText) {
		TextAccess model = ((ParagraphText) child).getTextAccess();
		int firstText = i;
		int lastText = i;

		while (lastText+1 < children.length &&
		       (children[lastText+1] instanceof ParagraphText) &&
		       ((ParagraphText)
			children[lastText+1]).getTextAccess() == model)
		    ++lastText;

		doWrapLines(firstText, lastText, childList, lineList);
		
		i = lastText;
	    } else {
		lineList.check(childList.count);

		if (recursive) {
		    int constrainedChildWidth = constrainedWidth;

		    if (constrainedWidth > 0) {
			if (style.textAlign == Style.TEXT_ALIGN_LEFT)
			   constrainedChildWidth -= Math.abs(style.textIndent);

			Style childStyle = ((Styled) child).getStyle();
			if (childStyle.marginLeft > 0)
			    constrainedChildWidth -= childStyle.marginLeft;
			if (childStyle.marginRight > 0)
			    constrainedChildWidth -= childStyle.marginRight;

			if (constrainedChildWidth <= 0)
			    constrainedChildWidth = 1;
		    }

		    child.layout(constrainedChildWidth, -1, true);
		}

		childList.add(child);		
	    }
	}

	lines = lineList.toArray();
	children = childList.toArray();
    }

    private void doWrapLines(int firstText, int lastText, 
			     GadgetList childList, LineList lineList) {
	ParagraphText text = (ParagraphText) children[firstText];
	TextAccess model = text.getTextAccess();
	int textLength = model.getTextLength();
	char[] textChars = model.getTextChars();
	int lineBegin = text.getRangeOffset();

	for (int i = lineBegin; i < textLength; ++i) {
	    if (textChars[i] == '\n') {
		lineList.check(childList.count);
		lineList.add(-1);

		if (firstText <= lastText) 
		    text = (ParagraphText) children[firstText++];
		else
		    text = ((ParagraphText) children[lastText]).copy();
		text.setRange(lineBegin, i+1 - lineBegin);
		text.layout(-1, -1, false);
		childList.add(text);

		lineBegin = i+1;
	    }
	}

	if (lineBegin < textLength || textLength == 0) {
	    lineList.check(childList.count);

	    if (firstText <= lastText) 
		text = (ParagraphText) children[firstText++];
	    else
		text = ((ParagraphText) children[lastText]).copy();
	    text.setRange(lineBegin, textLength - lineBegin);
	    text.layout(-1, -1, false);
	    childList.add(text);
	}
    }

    private void wrapWords(int fromLine, int constrainedWidth, 
			   boolean recursive) {
	if (children.length == 0) {
	    lines = new int[0];
	    return;
	}

	int fromChild = (lines.length > 0)? lines[fromLine] : 0;
	GadgetList childList = new GadgetList(children, fromChild);
	LineList lineList = new LineList(lines, fromLine);

	int remainLineWidth, maxLineWidth;
	if (constrainedWidth > 0) {
	    remainLineWidth = maxLineWidth = constrainedWidth;

	    if (style.textAlign == Style.TEXT_ALIGN_LEFT) {
		if (style.textIndent > 0) {
		    remainLineWidth = constrainedWidth - style.textIndent;
		    maxLineWidth = constrainedWidth;
		} else if (style.textIndent < 0) {
		    remainLineWidth = constrainedWidth;
		    maxLineWidth = constrainedWidth + style.textIndent;
		}
		if (fromLine > 0)
		    remainLineWidth = maxLineWidth;
	    }
	} else {
	    remainLineWidth = maxLineWidth = Integer.MAX_VALUE;
	}

	int lineBeginChild = fromChild;
	int lineFirstChar = 
	    ((children[fromChild] instanceof ParagraphText)?
	     ((ParagraphText) children[fromChild]).getRangeOffset() : -1);
	int lineWidth = 0;
	int wordBeginChild = -1;
	int wordFirstChar = -1;
	int wordWidth = 0;
	int reuse = fromChild;

	for (int i = fromChild; i < children.length; ++i) {
	    Gadget child = children[i];
	    Style childStyle = ((Styled) child).getStyle();

	    if (child instanceof ParagraphText) {
		ParagraphText para = (ParagraphText) child;
		TextAccess model = para.getTextAccess();
		int textLength = model.getTextLength();
		char[] textChars = model.getTextChars();
		FontMetrics metrics = getFontMetrics(childStyle.font);
		int[] charWidths = metrics.getWidths();

		if (textLength == 0) {
		    textLength = 1;
		    textChars = EM;
		}
		for (int j = para.getRangeOffset(); j < textLength; ++j) {
		    char c = textChars[j];
		    int w = (c < 256)? charWidths[c] : metrics.charWidth(c);

		    switch (c) {
		    case ' ': case '\n': case '\r': case '\t':
			if (j+1 >= textLength) {
			    // A white space at the end of a text is always
			    // displayed.

			    if (c == '\n')
				w = Icon.list[Icon.LINE_BREAK].width;

			    if (wordBeginChild < 0) {
				wordBeginChild = i;
				wordFirstChar = j;
			    }
			    wordWidth += w;

			    if (lineWidth + wordWidth > remainLineWidth &&
				lineWidth > 0) {
				reuse = putLine(reuse, 
						lineBeginChild, lineFirstChar, 
						wordBeginChild, wordFirstChar,
						childList, lineList);
				lineList.add(-1);
				remainLineWidth = maxLineWidth;

				lineBeginChild = wordBeginChild;
				lineFirstChar = wordFirstChar;
				lineWidth = 0;
			    } else {
				wordBeginChild = -1;
				wordFirstChar = -1;
				lineWidth += wordWidth;
				wordWidth = 0;
			    }
			    
			    if (c == '\n') {
				while (i+1 < children.length &&
				       (children[i+1] 
					instanceof ParagraphText) &&
				       ((ParagraphText)
					children[i+1]).getTextAccess()==model)
				    ++i;

				int nextLineBeginChild = i+1;
				int nextLineFirstChar = 
				    ((i+1 < children.length && 
				     (children[i+1] instanceof ParagraphText))?
				     0 : -1);

				reuse = putLine(
				    reuse, lineBeginChild, lineFirstChar, 
				    nextLineBeginChild, nextLineFirstChar, 
				    childList, lineList);
				lineList.add(-1);
				remainLineWidth = maxLineWidth;

				lineBeginChild = nextLineBeginChild;
				lineFirstChar = nextLineFirstChar;
				lineWidth = 0;

				wordBeginChild = -1;
				wordFirstChar = -1;
				wordWidth = 0;
			    }
			} else {
			    // A white space that causes word wrap to happen
			    // is not displayed.

			    if (wordBeginChild >= 0) {
				wordBeginChild = -1;
				wordFirstChar = -1;
				lineWidth += wordWidth;
				wordWidth = 0;
			    }

			    if (c == '\n' || lineWidth + w > remainLineWidth) {
				reuse = putLine(reuse, 
						lineBeginChild, lineFirstChar, 
						i, j+1, childList, lineList);
				lineList.add(-1);
				remainLineWidth = maxLineWidth;

				lineBeginChild = i;
				lineFirstChar = j+1;
				lineWidth = 0;
			    } else {
				lineWidth += w;
			    }
			}
			break;

		    default:
			if (wordBeginChild < 0) {
			    wordBeginChild = i;
			    wordFirstChar = j;
			}
			wordWidth += w;

			if (lineWidth + wordWidth > remainLineWidth && 
			    lineWidth > 0) {
			    reuse = putLine(reuse, 
					    lineBeginChild, lineFirstChar, 
					    wordBeginChild, wordFirstChar,
					    childList, lineList);
			    lineList.add(-1);
			    remainLineWidth = maxLineWidth;

			    lineBeginChild = wordBeginChild;
			    lineFirstChar = wordFirstChar;
			    lineWidth = 0;
			}
		    }
		}

		while (i+1 < children.length &&
		       (children[i+1] instanceof ParagraphText) &&
		       ((ParagraphText)children[i+1]).getTextAccess() == model)
		    ++i;
	    } else {
		if (recursive) {
		    int constrainedChildWidth = constrainedWidth;

		    if (constrainedWidth > 0) {
			if (style.textAlign == Style.TEXT_ALIGN_LEFT)
			   constrainedChildWidth -= Math.abs(style.textIndent);

			if (childStyle.marginLeft > 0)
			    constrainedChildWidth -= childStyle.marginLeft;
			if (childStyle.marginRight > 0)
			    constrainedChildWidth -= childStyle.marginRight;

			if (constrainedChildWidth <= 0)
			    constrainedChildWidth = 1;
		    }

		    child.layout(constrainedChildWidth, -1, true);
		}

		// Non text gadgets behaves like a non-space char.
		if (wordBeginChild < 0) {
		    wordBeginChild = i;
		    wordFirstChar = -1;
		}
		wordWidth += child.width;
		if (childStyle.marginLeft > 0)
		    wordWidth += childStyle.marginLeft;
		if (childStyle.marginRight > 0)
		    wordWidth += childStyle.marginRight;

		if (lineWidth + wordWidth > remainLineWidth && lineWidth > 0) {
		    reuse = putLine(reuse, lineBeginChild, lineFirstChar, 
				    wordBeginChild, wordFirstChar,
				    childList, lineList);
		    lineList.add(-1);
		    remainLineWidth = maxLineWidth;

		    lineBeginChild = wordBeginChild;
		    lineFirstChar = wordFirstChar;
		    lineWidth = 0;
		}
	    }
	}

	if (lineWidth + wordWidth > 0)
	    reuse = putLine(reuse, lineBeginChild, lineFirstChar, 
			    children.length, -1, childList, lineList);

	lines = lineList.toArray();
	children = childList.toArray();
    }

    private int putLine(int reuse, int beginChild, int beginChar, 
			int endChild, int endChar,
			GadgetList childList, LineList lineList) {
	lineList.check(childList.count);

	int lastChild = ((beginChild == endChild || endChar > 0)? 
			 endChild : endChild-1);
	for (int i = beginChild; i <= lastChild; ++i) {
	    Gadget child = children[i];

	    if (child instanceof ParagraphText) {
		ParagraphText text = (ParagraphText) child;
		TextAccess model = text.getTextAccess();

		ParagraphText text2 = null;
		for (int j = reuse; j < children.length; ++j) {
		    Gadget gadget = children[j];

		    if ((gadget instanceof ParagraphText) &&
			((ParagraphText) gadget).getTextAccess() == model) {
			text2 = (ParagraphText) gadget;
			reuse = j+1;
			break;
		    }
		}
		if (text2 == null)
		    text2 = text.copy();

		int rangeFirst = (i == beginChild)? beginChar : 0;
		int rangeEnd = (i==endChild)? endChar : model.getTextLength();
		text2.setRange(rangeFirst, rangeEnd - rangeFirst);
		text2.layout(-1, -1, false);
		childList.add(text2);

		while (i+1 < endChild &&
		       (children[i+1] instanceof ParagraphText) &&
		       ((ParagraphText) children[i+1]).getTextAccess() == 
		       model)
		    ++i;
	    } else {
		childList.add(child);
	    }
	}

	return reuse;
    }

    // -----------------------------------------------------------------------

    public void getSizeConstraints(Gadget child, Dimension constraints) {
	Box.getContentSizeConstraints(this, style, constraints);
	constraints.height = -1;

	if (constraints.width > 0) {
	    if (style.textAlign == Style.TEXT_ALIGN_LEFT)
		constraints.width -= Math.abs(style.textIndent);

	    Style childStyle = ((Styled) child).getStyle();
	    if (childStyle.marginLeft > 0)
		constraints.width -= childStyle.marginLeft;
	    if (childStyle.marginRight > 0)
		constraints.width -= childStyle.marginRight;

	    if (constraints.width <= 0)
		constraints.width = 1;
	}
    }

    public void paintRelayout(Gadget child, int prevWidth, int prevHeight,
			      int damagedSide, int damagedLength) {
	refreshFromLine(lineIndexOfChild(indexOfChild(child)));
    }

    private int lineIndexOfChild(int childIndex) {
	int lineIndex = 0;
	for (int i = 1; i < lines.length; ++i) {
	    if (lines[i] > childIndex) 
		break;
	    ++lineIndex;
	}
	return lineIndex;
    }

    private void refreshFromLine(int fromLine) {
//System.err.print("refreshFromLine " + fromLine + " ==> ");

	int oldWidth = width;
	int oldHeight = height;
	int oldFirstBaseLine = firstBaseLine;
	int oldLastBaseLine = lastBaseLine;

	int x0 = style.borderWidth + style.paddingLeft;
	int y0 = style.borderWidth + style.paddingTop;
	int decorWidth = x0 + style.borderWidth + style.paddingRight;
	int decorBottom = style.borderWidth + style.paddingBottom;

	Dimension constraints = new Dimension();
	parent.getSizeConstraints(this, constraints);
	constraints.height = -1;
	if (constraints.width > 0) {
	    constraints.width -= decorWidth;
	    if (constraints.width <= 0)
		constraints.width = 1;
	}

	int damagedHeight = layoutContentFromLine(fromLine, constraints.width,
						  false);

	int fromChild = (lines.length > 0)? lines[fromLine] : 0;
	for (int i = fromChild; i < children.length; ++i) {
	    Gadget child = children[i];
	    child.x += x0;
	    child.y += y0;
	}

	width += decorWidth;
	height += y0 + decorBottom;
	damagedHeight += decorBottom;

	if (fromLine == 0)
	    firstBaseLine += y0;
	lastBaseLine += y0;

	if (width == oldWidth && 
	    height == oldHeight &&
	    firstBaseLine == oldFirstBaseLine && 
	    lastBaseLine == oldLastBaseLine) {
	    repaint(0, 0, width, height);
//System.err.println("repaint");
	} else {
	    if (parent != null) {
		if (width == oldWidth) {
		    parent.paintRelayout(this, oldWidth, oldHeight, 
					 BOTTOM_SIDE, damagedHeight);
//System.err.println("paintRelayout BOTTOM_SIDE");
		} else {
		    parent.paintRelayout(this, oldWidth, oldHeight, 
					 NO_SIDE, -1);
//System.err.println("paintRelayout NO_SIDE oldW=" + oldWidth + ",w=" + width);
		}
	    }
	}
    }

    public void charInserted(char c, TextAccess text, int offset) {
	if (style.whiteSpace == Style.WHITE_SPACE_NORMAL) {
	    switch (c) {
	    case ' ': case '\r': case '\t': case '\n': 
		// Inserting a space in the first word of a line means
		// refreshing from previous line (if any).
		int fromLine = 
		    lineIndexOfChild(childIndexOfTextOffset(text, offset));
		if (fromLine > 0)
		    --fromLine;
		refreshFromLine(fromLine);
		break;
	    default:
		refeshTextLine(text, offset, /*deleted*/ false);
	    }
	} else {
	    switch (c) {
	    case '\n': 
		refreshFromLine(
		    lineIndexOfChild(childIndexOfTextOffset(text, offset)));
		break;
	    default:
		// Appending a char after the newline icon ending a text means
		// adding a new line.
		if (offset == text.getTextLength()-1 &&
		    offset > 0 && text.getTextChar(offset-1) == '\n')
		    refreshFromLine(lineIndexOfChild(
			childIndexOfTextOffset(text, offset)));
		else
		    refeshTextLine(text, offset, /*deleted*/ false);
	    }
	}
    }

    public void charDeleted(char c, TextAccess text, int offset) {
	if (style.whiteSpace == Style.WHITE_SPACE_NORMAL) {
	    // Deleting a char in the first word of a line means refreshing
	    // from previous line (if any).
	    int fromLine = 
		lineIndexOfChild(childIndexOfTextOffset(text, offset));
	    if (fromLine > 0)
		--fromLine;
	    refreshFromLine(fromLine);
	} else {
	    switch (c) {
	    case '\n': 
		refreshFromLine(
		    lineIndexOfChild(childIndexOfTextOffset(text, offset)));
		break;
	    default:
		// Deleting last char of last line means suppressing this line.
		if (offset == text.getTextLength() &&
		    offset > 0 && text.getTextChar(offset-1) == '\n')
		    refreshFromLine(lineIndexOfChild(
			childIndexOfTextOffset(text, offset-1)));
		else
		    refeshTextLine(text, offset, /*deleted*/ true);
	    }
	}
    }

    private void refeshTextLine(TextAccess text, int offset, boolean deleted) {
//System.err.print("refeshTextLine "+text+","+offset+","+deleted+" ==> ");

	int childIndex = childIndexOfTextOffset(text, offset);
	if (childIndex < 0) {
	    // Not found (???).
//System.err.println("not found");
	    return;
	}

	int lineNumber = -1;
	int lastLine = lines.length - 1;
	for (int i = 0; i <= lastLine; ++i) {
	    if (lines[i] == childIndex) {
		lineNumber = i;
		break;
	    }
	}
	if (lineNumber < 0 || 
	    (lineNumber == lastLine && childIndex != children.length-1) ||
	    (lineNumber != lastLine && lines[lineNumber+1] != childIndex+1)) {
	    // Not alone on its line.
	    refreshFromLine(lineIndexOfChild(childIndex));
//System.err.println("not alone on its line");
	    return;
	}

	ParagraphText para = (ParagraphText) children[childIndex];
	int prevParaWidth = para.width;
	if (deleted) 
	    para.setRange(para.getRangeOffset(), para.getRangeLength()-1);
	else
	    para.setRange(para.getRangeOffset(), para.getRangeLength()+1);
	para.layout(-1, -1, false);

	int contentX = style.borderWidth + style.paddingLeft;
	int contentWidth = width - 
	    (contentX + style.paddingRight + style.borderWidth);
	if (style.textAlign == Style.TEXT_ALIGN_LEFT) {
	    if (style.textIndent > 0 && lineNumber == 0)
		contentWidth -= style.textIndent;
	    else if (style.textIndent < 0 && lineNumber > 0)
		contentWidth += style.textIndent;
	} 
	if (contentWidth <= 0)
	    contentWidth = 1;

	if (prevParaWidth < contentWidth && para.width <= contentWidth) {
	    // The new width cannot have changed the overall width of the
	    // container. Adjust para.x and repaint.

	    if (deleted) {
		for (int i = childIndex+1; i < children.length; ++i) {
		    Gadget child = children[i];
		    ParagraphText para2;

		    if (!(child instanceof ParagraphText) ||
			(para2=(ParagraphText)child).getTextAccess() != text)
			break;

		    para2.setRange(para2.getRangeOffset()-1,
				   para2.getRangeLength());
		}
	    } else {
		for (int i = childIndex+1; i < children.length; ++i) {
		    Gadget child = children[i];
		    ParagraphText para2;

		    if (!(child instanceof ParagraphText) ||
			(para2=(ParagraphText)child).getTextAccess() != text)
			break;

		    para2.setRange(para2.getRangeOffset()+1,
				   para2.getRangeLength());
		}
	    }

	    switch (style.textAlign) {
	    case Style.TEXT_ALIGN_RIGHT:
		para.x -= (para.width - prevParaWidth);
		break;
	    case Style.TEXT_ALIGN_CENTER:
		para.x -= (para.width - prevParaWidth)/2;
		break;
	    default:
		// Left aligned: no need to adjust para.x.  
		break;
	    }
	    repaint(contentX, para.y, contentWidth, para.height);
//System.err.println("repaint");
	} else {
	    refreshFromLine(lineIndexOfChild(childIndex));
//System.err.println("may change the overall width of the paragraph");
	}
    }

    private int childIndexOfTextOffset(TextAccess text, int offset) {
	int childIndex = -1;
	for (int i = 0; i < children.length; ++i) {
	    Gadget child = children[i];
	    ParagraphText para2;

	    if ((child instanceof ParagraphText) && 
		(para2 = (ParagraphText) child).getTextAccess() == text) {
		childIndex = i;
		if (para2.containsChar(offset))
		    break;
		// Use last ParagraphText when appending a char to the text.
	    }
	}
	return childIndex;
    }
}
