/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledgadget;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Dimension;
import com.xmlmind.awtutil.Bitmap;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.gadget.GadgetContainer;
import com.xmlmind.xmledit.model.TextAccess;

public class ParagraphText extends Gadget 
                           implements Styled, TextLines, Cloneable {
    public static final class Access implements TextAccess {
	private char[] chars;

	public Access(String text) {
	    this(text.toCharArray());
	}

	public Access(char[] chars) {
	    this.chars = chars;
	}

	public String getText() {
	    return new String(chars);
	}

	public int getTextLength() {
	    return chars.length;
	}

	public char[] getTextChars() {
	    return chars;
	}

	public char getTextChar(int offset) {
	    return chars[offset];
	}
    }

    // -----------------------------------------------------------------------

    protected Style style;
    protected TextAccess text;
    protected int rangeOffset, rangeLength;

    public ParagraphText copy() {
	ParagraphText clone;
	try {
	    clone = (ParagraphText) clone();
	} catch (CloneNotSupportedException cannotHappen) {
	    cannotHappen.printStackTrace();
	    clone = null;
	}

	//Style style2 = this.style.duplicate();
	//style2.backgroundColor = Color.orange;
	//clone.style = style2;

	return clone;
    }

    public ParagraphText(String text) {
	this(Box.DEFAULT_STYLE, text);
    }

    public ParagraphText(Style style, String text) {
	this(style, new Access(text));
    }

    public ParagraphText(Style style, TextAccess text) {
	setStyle(style);
	setTextAccess(text);
    }

    /**
     * Can only be used at ParagraphText creation time (i.e. its sole purpose 
     * is to make the construction of a ParagraphText more convenient).
     */
    public void setStyle(Style style) {
	this.style = style;
    }

    /**
     * Can only be used at ParagraphText creation time (i.e. its sole purpose 
     * is to make the construction of a ParagraphText more convenient).
     */
    public void setTextAccess(TextAccess text) {
	this.text = text;
	rangeOffset = 0;
	rangeLength = text.getTextLength();
    }

    public Style getStyle() {
	return style;
    }

    public TextAccess getTextAccess() {
	return text;
    }

    public int getRangeOffset() {
	return rangeOffset;
    }

    public int getRangeLength() {
	return rangeLength;
    }

    public void setRange(int offset, int length) {
	if (offset < 0 || length < 0 || offset + length > text.getTextLength())
	    throw new IllegalArgumentException("invalid range");

	rangeOffset = offset;
	rangeLength = length;
    }

    public boolean containsChar(int offset) {
	int rangeEnd = rangeOffset+rangeLength;
	return ((offset >= rangeOffset && offset < rangeEnd) ||
		(offset == rangeEnd && rangeEnd >= text.getTextLength()));
    }

    public void refresh() {
	int oldWidth = width;
	int oldHeight = height;

	relayout(true);

	// A ParagraphText is almost always contained in a Paragraph which
	// needs to be refreshed even if the size of the ParagraphText has not
	// changed (ex. replacing an 'a' with an ' ' using a monospaced font).

	if (parent != null)
	    parent.paintRelayout(this, oldWidth, oldHeight, 
				 GadgetContainer.NO_SIDE, -1);
    }

    public void layout(int constrainedWidth, int constrainedHeight,
		       boolean recursive) {
	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();
	int firstChar = rangeOffset;
	int charCount = rangeLength;

	// The text accessed may have changed and we may have a displayed
	// range temporarily out of sync.
	if (firstChar + charCount > textLength)
	    charCount = textLength - firstChar;

	if (charCount <= 0) {
	    Dimension size = Box.getEmptyContentSize(this, style);
	    width = size.width;
	    height = size.height;
	} else {
	    FontMetrics metrics = getFontMetrics(style.font);
	    height = (int) (0.5 + metrics.getHeight()*style.lineHeight);

	    // Do not draw the space char where the word wrap occurs.
	    int last = firstChar + charCount - 1;
	    switch (textChars[last]) {
	    case '\n': 
		width = metrics.charsWidth(textChars, firstChar, charCount-1);
		if (last == textLength - 1) 
		    width += Icon.list[Icon.LINE_BREAK].width;
		else if (width == 0)
		    width = 1;
		break;
	    case ' ': case '\r': case '\t':
		if (last != textLength - 1) {
		    width = metrics.charsWidth(textChars, firstChar, 
					       charCount-1);
		    if (width == 0)
			width = 1;
		    break;
		}
		/*FALLTHROUGH*/
	    default:
		width = metrics.charsWidth(textChars, firstChar, charCount);
	    }
	}
    }

    public void paint(Graphics g) {
	if (style.backgroundColor != null) {
	    g.setColor(style.backgroundColor);
	    g.fillRect(0, 0, width, height);
	}

	paintText(g);
    }

    public void paintText(Graphics g) {
	char[] textChars = text.getTextChars();
	int textLength = text.getTextLength();
	int firstChar = rangeOffset;
	int charCount = rangeLength;

	// The text accessed may have changed and we may have a displayed
	// range temporarily out of sync.
	if (firstChar + charCount > textLength)
	    charCount = textLength - firstChar;

	if (charCount <= 0) {
	    // ParagraphText cannot have a Box: so use full width and height.
	    Box.drawEmptyTextContent(g, style, width, height);
	    return;
	}

	FontMetrics metrics = g.getFontMetrics(style.font);
	int ascent = metrics.getAscent();
	int descent = metrics.getDescent();
	int lineHeight = (int) (0.5 + metrics.getHeight()*style.lineHeight);
	// Simplified form of (lineHeight - ascent - descent)/2 + ascent.
	int lineY = (lineHeight + ascent - descent)/2;

	g.setColor(style.color);
	g.setFont(style.font);

	// Do not draw the space char where the word wrap occurs.
	int last = firstChar + charCount - 1;

	switch (textChars[last]) {
	case '\n': 
	    --charCount;
	    g.drawChars(textChars, firstChar, charCount, 0, lineY);
	    if (last == textLength - 1) 
		g.drawImage(
		    Icon.get(Icon.LINE_BREAK, style.color, null),
		    width - Icon.list[Icon.LINE_BREAK].width, 
		    (lineHeight - Icon.list[Icon.LINE_BREAK].height)/2,
		    null);
	    break;
	case ' ': case '\r': case '\t':
	    if (last != textLength - 1)
		--charCount;
	    /*FALLTHROUGH*/
	default:
	    g.drawChars(textChars, firstChar, charCount, 0, lineY);
	}

	switch (style.textDecoration) {
	case Style.TEXT_DECORATION_UNDERLINE:
	    {
		int decorY = lineY + descent - 1;
		int lineW = 
		    metrics.charsWidth(textChars, firstChar, charCount);
		g.drawLine(0, decorY, lineW-1, decorY);
	    }
	    break;
	case Style.TEXT_DECORATION_OVERLINE:
	    {
		int decorY = lineY - ascent + 1;
		int lineW = 
		    metrics.charsWidth(textChars, firstChar, charCount);
		g.drawLine(0, decorY, lineW-1, decorY);
	    }
	    break;
	case Style.TEXT_DECORATION_LINE_THROUGH:
	    {
		// Simplication of (descent - (ascent+descent)/2)
		int decorY = lineY + (descent - ascent)/2;
		int lineW = 
		    metrics.charsWidth(textChars, firstChar, charCount);
		g.drawLine(0, decorY, lineW-1, decorY);
	    }
	    break;
	}
    }

    public int getFirstBaseLine() {
	FontMetrics metrics = getFontMetrics(style.font);
	int lineHeight = (int) (0.5 + metrics.getHeight()*style.lineHeight);
	int lineY = 
	    (lineHeight + metrics.getAscent() - metrics.getDescent())/2;

	return lineY;
    }

    public int getLastBaseLine() {
	return getFirstBaseLine();
    }
}


