/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledgadget;

import java.awt.Color;
import java.awt.Font;

public class Style implements Cloneable {
    public static final int BORDER_STYLE_NONE   = 0;
    public static final int BORDER_STYLE_DOTTED = 1;
    public static final int BORDER_STYLE_DASHED = 2;
    public static final int BORDER_STYLE_SOLID  = 3;
    public static final int BORDER_STYLE_DOUBLE = 4;
    public static final int BORDER_STYLE_GROOVE = 5;
    public static final int BORDER_STYLE_RIDGE  = 6;
    public static final int BORDER_STYLE_INSET  = 7;
    public static final int BORDER_STYLE_OUTSET = 8;

    public static final int VERTICAL_ALIGN_BASELINE    = 0;
    public static final int VERTICAL_ALIGN_SUB         = 1;
    public static final int VERTICAL_ALIGN_SUPER       = 2;
    public static final int VERTICAL_ALIGN_TOP         = 3;
    public static final int VERTICAL_ALIGN_TEXT_TOP    = 4;
    public static final int VERTICAL_ALIGN_MIDDLE      = 5;
    public static final int VERTICAL_ALIGN_BOTTOM      = 6;
    public static final int VERTICAL_ALIGN_TEXT_BOTTOM = 7;

    public static final int TEXT_ALIGN_LEFT   = 0;
    public static final int TEXT_ALIGN_RIGHT  = 1;
    public static final int TEXT_ALIGN_CENTER = 2;

    public static final int TEXT_DECORATION_NONE         = 0;
    public static final int TEXT_DECORATION_UNDERLINE    = 1;
    public static final int TEXT_DECORATION_OVERLINE     = 2;
    public static final int TEXT_DECORATION_LINE_THROUGH = 3;

    public static final int WHITE_SPACE_NORMAL = 0;
    public static final int WHITE_SPACE_PRE    = 1;
    public static final int WHITE_SPACE_NOWRAP = 2;

    public static final Font DEFAULT_FONT = 
        new Font("SansSerif", Font.PLAIN, 12);

    // marginXXX < 0 means AUTO.
    //
    // backgroundColor == null means transparent. Other colors cannot be 
    // transparent.
    //
    // Here lineHeight is a specification: lineHeight*fontSize gives the
    // distance in pixels between the baselines of 2 consecutive lines. Normal
    // lineHeight is between 1.0 to 1.2.

    public int marginTop, marginRight, marginBottom, marginLeft; 
    public int paddingTop, paddingRight, paddingBottom, paddingLeft; 
    public int borderStyle; 
    public int borderWidth; 
    public Color borderTopColor; 
    public Color borderRightColor; 
    public Color borderBottomColor; 
    public Color borderLeftColor; 
    public double lineHeight;
    public int verticalAlign;
    public Color color; 
    public Color backgroundColor;
    public Font font;
    public int textIndent;
    public int textAlign;
    public int textDecoration;
    public int whiteSpace;

    // -----------------------------------------------------------------------

    public Style() {
	reset();
    }

    public void reset() {
	color = Color.black; 
	font = DEFAULT_FONT;
	lineHeight = 1.0;
	textIndent = 0;
	textAlign = TEXT_ALIGN_LEFT;
	whiteSpace = WHITE_SPACE_NORMAL;

	resetNonInherited();
    }

    public void resetNonInherited() {
	verticalAlign = VERTICAL_ALIGN_BASELINE;
	textDecoration = TEXT_DECORATION_NONE;

	resetBox();
    }

    public void resetBox() {
	marginTop = marginRight = marginBottom = marginLeft = 0; 
	paddingTop = paddingRight = paddingBottom = paddingLeft = 0; 
	borderStyle = BORDER_STYLE_NONE; 
	borderWidth = 0; 
	borderTopColor = color;
	borderRightColor = color;
	borderBottomColor = color;
	borderLeftColor = color;
	backgroundColor = null;
    }

    public Style duplicate() {
	try {
	    return (Style) clone();
	} catch (CloneNotSupportedException e) {
	    e.printStackTrace();
	    return null;
	}
    }

    // -----------------------------------------------------------------------

    public boolean equals(Object other) {
	if (other == null || !(other instanceof Style))
	    return false;
	
	Style s = (Style) other;

	return (marginTop == s.marginTop && 
		marginRight == s.marginRight &&
		marginBottom == s.marginBottom && 
		marginLeft == s.marginLeft &&

		paddingTop == s.paddingTop && 
		paddingRight == s.paddingRight &&
		paddingBottom == s.paddingBottom && 
		paddingLeft == s.paddingLeft &&

		borderWidth == s.borderWidth && 
		((borderTopColor == null && s.borderTopColor == null) ||
		 (borderTopColor != null &&
		  borderTopColor.equals(s.borderTopColor))) &&
		((borderRightColor == null && s.borderRightColor == null) ||
		 (borderRightColor != null &&
		  borderRightColor.equals(s.borderRightColor))) &&
		((borderBottomColor == null && s.borderBottomColor == null) ||
		 (borderBottomColor != null &&
		  borderBottomColor.equals(s.borderBottomColor))) &&
		((borderLeftColor == null && s.borderLeftColor == null) ||
		 (borderLeftColor != null &&
		  borderLeftColor.equals(s.borderLeftColor))) &&
		borderStyle == s.borderStyle &&

		lineHeight == s.lineHeight &&
	    
	        verticalAlign == s.verticalAlign &&

		color.equals(s.color) &&
		((backgroundColor == null && s.backgroundColor == null) ||
		 (backgroundColor != null &&
		  backgroundColor.equals(s.backgroundColor))) &&

	        font.equals(s.font) &&

		textIndent == s.textIndent &&
		textAlign == s.textAlign &&
		textDecoration == s.textDecoration &&

                whiteSpace == s.whiteSpace);
    }

    public int hashCode() {
	int code = 0;

	// Margins are small pixel values. It should be sufficient to just
	// hash the low-order byte.
	code = hashByte((byte) marginTop, code);
	code = hashByte((byte) marginRight, code);
	code = hashByte((byte) marginBottom, code);
	code = hashByte((byte) marginLeft, code);

	code = hashByte((byte) paddingTop, code);
	code = hashByte((byte) paddingRight, code);
	code = hashByte((byte) paddingBottom, code);
	code = hashByte((byte) paddingLeft, code);

	code = hashByte((byte) borderStyle, code);
	code = hashByte((byte) borderWidth, code);
	code = hashInt(((borderTopColor == null)? 
			0 : borderTopColor.hashCode()),
		       code);
	code = hashInt(((borderRightColor == null)? 
			0 : borderRightColor.hashCode()), 
		       code);
	code = hashInt(((borderBottomColor == null)? 
			0 : borderBottomColor.hashCode()), 
		       code);
	code = hashInt(((borderLeftColor == null)? 
			0 : borderLeftColor.hashCode()),
		       code);

	code = hashByte((byte) (lineHeight - 1.0), code);

	code = hashByte((byte) verticalAlign, code);

	code = hashInt(color.hashCode(), code);
	code = hashInt(((backgroundColor == null)? 
			0 : backgroundColor.hashCode()),
		       code);

	code = hashInt(font.hashCode(), code);

	code = hashByte((byte) textIndent, code);
	code = hashByte((byte) textAlign, code);
	code = hashByte((byte) textDecoration, code);

	code = hashByte((byte) whiteSpace, code);

	return code;
    }

    /* -----------------------------------------------------------------------
       ElfHash 

       The published hash algorithm used in the UNIX ELF format for object
       files. Accepts a pointer to a string to be hashed and returns an
       unsigned long.

       unsigned long ElfHash ( const unsigned char *name )
       {
	   unsigned long   h = 0, g;
	   while ( *name )
	   {
	       h = ( h << 4 ) + *name++;
	       if ( g = h & 0xF0000000 )
		   h ^= g >> 24;
	       h &= ~g;
	   }
	   return h;
       }
    ----------------------------------------------------------------------- */

    public static final int hashByte(byte value, int code) {
	int highBits;

	code = (code << 4) + value;
	highBits = (code & 0xF0000000);
	if (highBits != 0)
	    code ^= (highBits >>> 24);
	code &= ~highBits;

	return code;
    }

    public static final int hashInt(int value, int code) {
	int highBits;

	code = (code << 4) + (value >>> 24);
	highBits = (code & 0xF0000000);
	if (highBits != 0)
	    code ^= (highBits >>> 24);
	code &= ~highBits;

	code = (code << 4) + ((value >>> 16) & 0xFF);
	highBits = (code & 0xF0000000);
	if (highBits != 0)
	    code ^= (highBits >>> 24);
	code &= ~highBits;

	code = (code << 4) + ((value >>> 8) & 0xFF);
	highBits = (code & 0xF0000000);
	if (highBits != 0)
	    code ^= (highBits >>> 24);
	code &= ~highBits;

	code = (code << 4) + (value & 0xFF);
	highBits = (code & 0xF0000000);
	if (highBits != 0)
	    code ^= (highBits >>> 24);
	code &= ~highBits;

	return code;
    }
}


