/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledgadget;

import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Rectangle;
import com.xmlmind.gadget.*;

public class TableBody extends ContainerBox {
    public static class Cell {
	public Gadget gadget = null;
	public int rowSpan = 1;
	public int columnSpan = 1;
	public Style style = Box.DEFAULT_STYLE;
	public int minContentWidth = 0;
    }
    public static class CellInfo extends Cell {
	public int rowIndex = -1;
	public int cellIndex = -1;
    }
    private static class CellImpl extends Cell {
	public int column = -1;
	public int x = -1;
	public int y = -1;
	public int width = -1;
	public int height = -1;
	public int gadgetBaseLine = -1;
	public int rowBaseLine = -1;

	public CellImpl(Cell spec) {
	    gadget = spec.gadget;
	    rowSpan = spec.rowSpan;
	    columnSpan = spec.columnSpan;
	    style = spec.style;
	    minContentWidth = spec.minContentWidth;
	}
    }
    private static class GridItem {
	public CellImpl cell;
	public int width;
	public int height;

	public GridItem() {
	    this(null);
	}

	public GridItem(CellImpl cell) {
	    this.cell = cell;
	    width = -1;
	    height = -1;
	}
    }

    private static final Cell[] noCells = new Cell[0];

    private int horizontalSpacing = 0;
    private int verticalSpacing = 0;
    private CellImpl[][] rows = new CellImpl[0][];

    private int[] columnConstrainedWidth = new int[0];

    CellImpl cachedCell = null;
    int cachedCellRowIndex = -1;
    int cachedCellConstrainedWidth = -1;

    // Scratch variables
    CellInfo cellInfo = new CellInfo();

    // -----------------------------------------------------------------------

    public TableBody() {
	super();
    }

    public TableBody(Style style) {
	super(style);
    }

    public void setHorizontalCellSpacing(int spacing) {
	horizontalSpacing = spacing;
    }
    
    public int getHorizontalCellSpacing() {
	return horizontalSpacing;
    }

    public void setVerticalCellSpacing(int spacing) {
	verticalSpacing = spacing;
    }
    
    public int getVerticalCellSpacing() {
	return verticalSpacing;
    }

    public void addRow() {
	insertRow(rows.length, noCells);
    }

    public void addRow(Cell[] cells) {
	insertRow(rows.length, cells);
    }

    public void insertRow(int rowIndex) {
	insertRow(rowIndex, noCells);
    }

    public void insertRow(int rowIndex, Cell[] cells) {
	CellImpl[][] newRows = new CellImpl[rows.length+1][];

	if (rowIndex > 0)
	    System.arraycopy(rows, 0, newRows, 0, rowIndex);

	int tail = rows.length - rowIndex;
	if (tail > 0) 
	    System.arraycopy(rows, rowIndex, newRows, rowIndex+1, tail);

	CellImpl[] newRow = new CellImpl[cells.length];
	for (int i = 0; i < cells.length; ++i)
	    newRow[i] = new CellImpl(cells[i]);

	newRows[rowIndex] = newRow;
	rows = newRows;

	// Update the children array.
	int newRowFirst = rowToChildIndex(rowIndex);
	for (int i = 0; i < cells.length; ++i)
	    super.insertChild(newRowFirst+i, cells[i].gadget);
    }

    private int rowToChildIndex(int rowIndex) {
	int index = 0;

	for (int i = 0; i < rowIndex; ++i)
	    index += rows[i].length;

	return index;
    }

    public void removeRow(int rowIndex) {
	CellImpl[] oldRow = rows[rowIndex];
	int oldRowFirst = rowToChildIndex(rowIndex);

	CellImpl[][] newRows = new CellImpl[rows.length-1][];

	if (rowIndex > 0)
	    System.arraycopy(rows, 0, newRows, 0, rowIndex);

	int first = rowIndex + 1;
	int tail = rows.length - first;
	if (tail > 0) 
	    System.arraycopy(rows, first, newRows, rowIndex, tail);

	rows = newRows;

	// Update the children array.
	for (int i = oldRow.length-1; i >= 0; --i)
	    super.removeChild(oldRowFirst+i);

	clearCachedCell();
    }

    private void clearCachedCell() {
	cachedCell = null;
	cachedCellRowIndex = -1;
	cachedCellConstrainedWidth = -1;
    }

    public Cell[] getRow(int rowIndex) {
	return rows[rowIndex];
    }

    public int getRowCount() {
	return rows.length;
    }

    public int getCellCount(int rowIndex) {
	return rows[rowIndex].length;
    }

    public void insertChild(int childIndex, Gadget gadget) {
	Cell cell = new Cell();
	cell.gadget = gadget;

	insertCell(childIndex, cell);
    }

    public void insertCell(int childIndex, Cell cell) {
	if (childIndex >= children.length) {
	    if (rows.length == 0) {
		insertRow(0, new Cell[] { cell });
	    } else {
		int last = rows.length-1;
		insertCell(last, rows[last].length, cell);
	    }
	} else {
	    cellInfo.gadget = children[childIndex];
	    findCell(cellInfo);

	    insertCell(cellInfo.rowIndex, cellInfo.cellIndex, cell);
	}
    }

    public void insertCell(int rowIndex, int cellIndex, Cell cell) {
	CellImpl[] row = rows[rowIndex];
	CellImpl[] newRow = new CellImpl[row.length+1];

	if (cellIndex > 0)
	    System.arraycopy(row, 0, newRow, 0, cellIndex);

	int tail = row.length - cellIndex;
	if (tail > 0) 
	    System.arraycopy(row, cellIndex, newRow, cellIndex+1, tail);

	newRow[cellIndex] = new CellImpl(cell);
	rows[rowIndex] = newRow;

	// Update the children array.
	super.insertChild(cellToChildIndex(rowIndex, cellIndex), cell.gadget);
    }

    private int cellToChildIndex(int rowIndex, int cellIndex) {
	return rowToChildIndex(rowIndex) + cellIndex;
    }

    public void removeChild(Gadget gadget) {
	cellInfo.gadget = gadget;
	if (findCell(cellInfo))
	    removeCell(cellInfo.rowIndex, cellInfo.cellIndex);
    }

    public void removeChild(int childIndex) {
	cellInfo.gadget = children[childIndex];
	findCell(cellInfo);

	removeCell(cellInfo.rowIndex, cellInfo.cellIndex);
    }

    public void removeCell(int rowIndex, int cellIndex) {
	CellImpl[] row = rows[rowIndex];
	CellImpl[] newRow = new CellImpl[row.length-1];

	if (cellIndex > 0)
	    System.arraycopy(row, 0, newRow, 0, cellIndex);

	int first = cellIndex + 1;
	int tail = row.length - first;
	if (tail > 0) 
	    System.arraycopy(row, first, newRow, cellIndex, tail);

	rows[rowIndex] = newRow;

	// Update the children array.
	super.removeChild(cellToChildIndex(rowIndex, cellIndex));

	clearCachedCell();
    }

    public void replaceChild(Gadget oldGadget, Gadget newGadget) {
	cellInfo.gadget = oldGadget;
	findCell(cellInfo);

	cellInfo.gadget = newGadget;
	replaceCell(cellInfo.rowIndex, cellInfo.cellIndex, cellInfo);
    }

    public void replaceChild(int childIndex, Gadget newGadget) {
	cellInfo.gadget = children[childIndex];
	findCell(cellInfo);

	cellInfo.gadget = newGadget;
	replaceCell(cellInfo.rowIndex, cellInfo.cellIndex, cellInfo);
    }

    public void replaceCell(int rowIndex, int cellIndex, Cell newCell) {
	CellImpl oldCell = rows[rowIndex][cellIndex];
	rows[rowIndex][cellIndex] = new CellImpl(newCell);

	// Update the children array.
	super.replaceChild(cellToChildIndex(rowIndex, cellIndex), 
			   newCell.gadget);

	clearCachedCell();
    }

    public CellInfo getCell(int rowIndex, int cellIndex) {
	CellInfo info = new CellInfo();
	info.rowIndex = rowIndex;
	info.cellIndex = cellIndex;
	getCell(info);

	return info;
    }

    public void getCell(CellInfo info) {
	CellImpl cell = rows[info.rowIndex][info.cellIndex];

	info.gadget = cell.gadget;
	info.rowSpan = cell.rowSpan;
	info.columnSpan = cell.columnSpan;
	info.style = cell.style;
	info.minContentWidth = cell.minContentWidth;
    }

    public CellInfo findCell(Gadget gadget) {
	CellInfo info = new CellInfo();
	info.gadget = gadget;

	return findCell(info)? info : null;
    }

    public boolean findCell(CellInfo info) {
	Gadget gadget = info.gadget;

	for (int i = 0; i < rows.length; ++i) {
	    CellImpl[] row = rows[i];

	    for (int j = 0; j < row.length; ++j) {
		CellImpl cell = row[j];

		if (cell.gadget == gadget) {
		    info.rowIndex = i;
		    info.cellIndex = j;
		    info.rowSpan = cell.rowSpan;
		    info.columnSpan = cell.columnSpan;
		    info.style = cell.style;
		    info.minContentWidth = cell.minContentWidth;
		    return true;
		}
	    }
	}

	return false;
    }

    // -----------------------------------------------------------------------
    // Layout
    // -----------------------------------------------------------------------

    protected void layoutContent(int constrainedWidth, int constrainedHeight,
				 boolean recursive) {
	int columnCount = estimateColumnCount();
	GridItem[][] grid = new GridItem[rows.length][columnCount];
	columnCount = placeCells(grid);

	if (recursive) {
	    clearCachedCell();
	    computeColumnConstrainedWidth(constrainedWidth, grid, columnCount);
	    relayoutChildren();
	}

	horizontalLayout(grid, columnCount);
	verticalLayout(grid, columnCount);
    }

    private int estimateColumnCount() {
	int maxColumnCount = 0;
	int spannedColumnCount = 0;

	for (int i = 0; i < rows.length; ++i) {
	    CellImpl[] row = rows[i];

	    int columnCount = 0;
	    for (int j = 0; j < row.length; ++j) {
		CellImpl cell = row[j];

		if (cell.rowSpan > 1) 
		    spannedColumnCount += cell.columnSpan;
		else
		    columnCount += cell.columnSpan;
	    }

	    if (columnCount > maxColumnCount)
		maxColumnCount = columnCount;
	}

	return (maxColumnCount + spannedColumnCount);
    }

    private int placeCells(GridItem[][] grid) {
	int maxColumnCount = 0;

	for (int i = 0; i < rows.length; ++i) {
	    CellImpl[] row = rows[i];
	    GridItem[] gridRow = grid[i];

	    int columnCount = 0;
	    for (int j = 0; j < row.length; ++j) {
		CellImpl cell = row[j];

		// Skip reserved row-spanned cell.
		while(gridRow[columnCount] != null)
		    ++columnCount;

		if (cell.rowSpan > 1) {
		    // Mark grid cells in the following rows as being reserved.
		    int endRow = Math.min(rows.length, i+cell.rowSpan);

		    for (int r = i+1; r < endRow; ++r) {
			GridItem[] gridRow2 = grid[r];

			for (int c = 0; c < cell.columnSpan; ++c)
			    gridRow2[columnCount+c] = new GridItem();
		    }
		}

		cell.column = columnCount;
		gridRow[columnCount++] = new GridItem(cell);
		for (int c = 1; c < cell.columnSpan; ++c)
		    gridRow[columnCount++] = new GridItem();
	    }

	    if (columnCount > maxColumnCount)
		maxColumnCount = columnCount;
	}

	for (int i = 0; i < rows.length; ++i) {
	    GridItem[] gridRow = grid[i];

	    for (int j = 0; j < maxColumnCount; ++j) {
		if (gridRow[j] == null)
		    gridRow[j] = new GridItem();
	    }
	}

	return maxColumnCount;
    }

    private void computeColumnConstrainedWidth(int constrainedWidth, 
					       GridItem[][] grid, 
					       int columnCount) {
	int[] colWidth = new int[columnCount];

	for (int j = 0; j < columnCount; ++j) {
	    int greatestWidth = -1;

	    for (int i = 0; i < rows.length; ++i) {
		GridItem item = grid[i][j];

		if (item.cell != null) {
		    CellImpl cell = item.cell;

		    if (cell.minContentWidth > 0) {
			Style cellStyle = cell.style;

			int totalWidth = 
			    (cell.minContentWidth + 2*cellStyle.borderWidth +
			     cellStyle.paddingLeft + cellStyle.paddingRight);

			item.width = 
			    computeWidthPerColumn(totalWidth, cell.columnSpan);
		    } else {
			item.width = -1;
		    }
		    
		    for (int c = 1; c < cell.columnSpan; ++c) {
			GridItem spanned = grid[i][j+c];
			spanned.width = item.width;
		    }
		}

		if (item.width > greatestWidth)
		    greatestWidth = item.width;
		item.width = 0;
	    }

	    colWidth[j] = greatestWidth;
	}

	if (constrainedWidth > 0) {
	    int usedWidth = 0;
	    int freeColumns = 0;

	    for (int j = 0; j < columnCount; ++j) {
		if (colWidth[j] > 0) 
		    usedWidth += horizontalSpacing + colWidth[j];
		else
		    ++freeColumns;
	    }

	    int columnBaseWidth = -1;
	    if (constrainedWidth > usedWidth && freeColumns > 0)
		columnBaseWidth = 
		    computeWidthPerColumn(constrainedWidth - usedWidth, 
					  freeColumns);
	    if (columnBaseWidth <= 0)
		columnBaseWidth = 1;

	    for (int j = 0; j < columnCount; ++j) {
		if (colWidth[j] <= 0) 
		    colWidth[j] = columnBaseWidth;
	    }
	}

	columnConstrainedWidth = colWidth;
    }

    private int computeWidthPerColumn(int totalWidth, int columnCount) {
	int widthPerColumn = totalWidth;

	if (columnCount > 1) {
	    widthPerColumn -= (columnCount - 1)*horizontalSpacing;
	    widthPerColumn /= columnCount;
	}

	if (widthPerColumn <= 0)
	    widthPerColumn = 1;

	return widthPerColumn;
    }

    private void relayoutChildren() {
	for (int i = 0; i < rows.length; ++i) {
	    CellImpl[] row = rows[i];

	    for (int j = 0; j < row.length; ++j) {
		CellImpl cell = row[j];

		cell.gadget.layout(computeChildConstrainedWidth(cell), -1, 
				   true);
	    }
	}
    }

    private int computeChildConstrainedWidth(CellImpl cell) {
	int cellCol = cell.column;
	
	int constrainedWidth = columnConstrainedWidth[cellCol];
	if (constrainedWidth <= 0)
	    return -1;
	for (int c = 1; c < cell.columnSpan; ++c) {
	    int incr = columnConstrainedWidth[cellCol+c];
	    if (incr <= 0)
		return -1;
	    constrainedWidth += horizontalSpacing + incr;
	}

	Style cellStyle = cell.style;
	constrainedWidth -=
	    (2*cellStyle.borderWidth +
	     cellStyle.paddingLeft + cellStyle.paddingRight);

	if (constrainedWidth <= 0)
	    constrainedWidth = 1;

	return constrainedWidth;
    }

    private void horizontalLayout(GridItem[][] grid, int columnCount) {
	// First pass: compute column width ---

	int[] columnWidth = new int[columnCount];

	for (int j = 0; j < columnCount; ++j) {
	    int maxWidth = 0;

	    for (int i = 0; i < rows.length; ++i) {
		GridItem item = grid[i][j];

		if (item.cell != null) {
		    CellImpl cell = item.cell;
		    Style cellStyle = cell.style;
		    int contentWidth = cell.gadget.width;

		    if (cell.minContentWidth > 0) 
			contentWidth = 
			    Math.max(contentWidth, cell.minContentWidth);

		    int totalWidth = 
			(contentWidth + 2*cellStyle.borderWidth +
			 cellStyle.paddingLeft + cellStyle.paddingRight);

		    item.width = 
			computeWidthPerColumn(totalWidth, cell.columnSpan);

		    for (int c = 1; c < cell.columnSpan; ++c) 
			grid[i][j+c].width = item.width;
		}

		if (item.width > maxWidth)
		    maxWidth = item.width;
	    }

	    columnWidth[j] = maxWidth;
	}

	// Second pass: layout cells and gadgets horizontally ---

	int cellX = 0;

	for (int j = 0; j < columnCount; ++j) {
	    if (j > 0)
		cellX += horizontalSpacing;

	    for (int i = 0; i < rows.length; ++i) {
		GridItem item = grid[i][j];

		if (item.cell != null) {
		    CellImpl cell = item.cell;

		    cell.x = cellX;
		    cell.width = columnWidth[j];
		    for (int c = 1; c < cell.columnSpan; ++c) 
			cell.width += (horizontalSpacing + columnWidth[j+c]);

		    Style cellStyle = cell.style;

		    switch (cellStyle.textAlign) {
		    case Style.TEXT_ALIGN_RIGHT:
			cell.gadget.x = 
			    (cell.x + cell.width -
			     cellStyle.borderWidth - cellStyle.paddingRight -
			     cell.gadget.width);
			break;
		    case Style.TEXT_ALIGN_CENTER:
			cell.gadget.x = 
			    (cell.x + 
			     cellStyle.borderWidth + cellStyle.paddingLeft) +
			    ((cell.width - 2*cellStyle.borderWidth - 
			      cellStyle.paddingLeft - cellStyle.paddingRight) -
			     cell.gadget.width)/2;
			break;
		    default:
			cell.gadget.x = 
			    (cell.x + 
			     cellStyle.borderWidth + cellStyle.paddingLeft);
		    }
		}
	    }

	    cellX += columnWidth[j];
	}

	width = cellX;
	if (width < 1)
	    width = 1;
    }

    private void verticalLayout(GridItem[][] grid, int columnCount) {
	// First pass: compute row height ---

	int[] rowHeight = new int[rows.length];

	for (int i = 0; i < rows.length; ++i) {
	    GridItem[] gridRow = grid[i];
	    int rowBaseLine = 0;

	    for (int j = 0; j < columnCount; ++j) {
		GridItem item = gridRow[j];

		if (item.cell != null) {
		    CellImpl cell = item.cell;
		    Style cellStyle = cell.style;

		    int totalHeight = 
			(cell.gadget.height + 2*cellStyle.borderWidth +
			 cellStyle.paddingTop + cellStyle.paddingBottom);

		    item.height = 
			computeHeightPerRow(totalHeight, cell.rowSpan);

		    int endRow = Math.min(rows.length, i+cell.rowSpan);
		    for (int r = i+1; r < endRow; ++r)
			grid[r][j].height = item.height;

		    switch (cellStyle.verticalAlign) {
		    case Style.VERTICAL_ALIGN_TOP:
		    case Style.VERTICAL_ALIGN_BOTTOM:
		    case Style.VERTICAL_ALIGN_MIDDLE:
			break;
		    default:
			if (cell.gadget instanceof TextLines) 
			    cell.gadgetBaseLine = 
				((TextLines) cell.gadget).getFirstBaseLine();
			else
			    cell.gadgetBaseLine = -1;
			if (cell.gadgetBaseLine < 0)
			    cell.gadgetBaseLine = cell.gadget.height;

			cell.rowBaseLine = 
			    (cellStyle.borderWidth + cellStyle.paddingTop +
			     cell.gadgetBaseLine);

			if (cell.rowBaseLine > rowBaseLine)
			    rowBaseLine = cell.rowBaseLine;
		    }
		}
	    }

	    int maxHeight = 0;

	    for (int j = 0; j < columnCount; ++j) {
		GridItem item = gridRow[j];
		CellImpl cell = item.cell;

		if (cell != null && cell.rowBaseLine > 0) {
		    item.height += (rowBaseLine - cell.rowBaseLine);
		    cell.rowBaseLine = rowBaseLine;
		}

		if (item.height > maxHeight)
		    maxHeight = item.height;
	    }

	    rowHeight[i] = maxHeight;
	}

	// Second pass: layout cells and gadgets vertically ---

	int cellY = 0;

	for (int i = 0; i < rows.length; ++i) {
	    GridItem[] gridRow = grid[i];

	    if (i > 0)
		cellY += verticalSpacing;

	    for (int j = 0; j < columnCount; ++j) {
		GridItem item = gridRow[j];

		if (item.cell != null) {
		    CellImpl cell = item.cell;

		    cell.y = cellY;
		    cell.height = rowHeight[i];
		    int endRow = Math.min(rows.length, i+cell.rowSpan);
		    for (int r = i+1; r < endRow; ++r) 
			cell.height += (verticalSpacing + rowHeight[r]);

		    Style cellStyle = cell.style;

		    switch (cellStyle.verticalAlign) {
		    case Style.VERTICAL_ALIGN_TOP:
			cell.gadget.y = 
			    (cell.y + 
			     cellStyle.borderWidth + cellStyle.paddingTop);
			break;
		    case Style.VERTICAL_ALIGN_BOTTOM:
			cell.gadget.y = 
			    (cell.y + cell.height - cellStyle.borderWidth -
			     cellStyle.paddingBottom - cell.gadget.height);
			break;
		    case Style.VERTICAL_ALIGN_MIDDLE:
			cell.gadget.y = 
			    (cell.y + 
			     cellStyle.borderWidth + cellStyle.paddingTop) +
			    ((cell.height - 2*cellStyle.borderWidth -
			      cellStyle.paddingTop - cellStyle.paddingBottom) -
			     cell.gadget.height)/2;
			break;
		    default:
			cell.gadget.y = 
			    (cell.y + cell.rowBaseLine - cell.gadgetBaseLine);
		    }
		}
	    }

	    cellY += rowHeight[i];
	}

	height = cellY;
	if (height < 1)
	    height = 1;
    }

    private int computeHeightPerRow(int totalHeight, int rowCount) {
	int heightPerRow = totalHeight;

	if (rowCount > 1) {
	    heightPerRow -= (rowCount - 1)*verticalSpacing;
	    heightPerRow /= rowCount;
	}

	if (heightPerRow <= 0)
	    heightPerRow = 1;

	return heightPerRow;
    }

    // -----------------------------------------------------------------------

    public void getSizeConstraints(Gadget child, Dimension constraints) {
	constraints.width = constraints.height = -1;

	CellImpl cell = fetchCell(child);
	if (cell == null)
	    return;

	if (cachedCellConstrainedWidth > 0) {
	    constraints.width = cachedCellConstrainedWidth;
	} else {
	    // Recompute cachedCellConstrainedWidth.
	    // (Skip if cell has not yet been layout.)

	    if (cell.column >= 0) 
		constraints.width = cachedCellConstrainedWidth = 
		    computeChildConstrainedWidth(cell);
	}
    }

    private CellImpl fetchCell(Gadget gadget) {
	if (cachedCell != null && cachedCell.gadget == gadget) 
	    return cachedCell;

	for (int i = 0; i < rows.length; ++i) {
	    CellImpl[] row = rows[i];

	    for (int j = 0; j < row.length; ++j) {
		CellImpl cell = row[j];

		if (cell.gadget == gadget) {
		    cachedCell = cell;
		    cachedCellRowIndex = i;
		    cachedCellConstrainedWidth = -1;
		    return cell;
		}
	    }
	}

	return null;
    }

    public void paintRelayout(Gadget child, int prevWidth, int prevHeight,
			      int damagedSide, int damagedLength) {
	CellImpl childCell = fetchCell(child);
	if (childCell == null)
	    return;

	if (childCell.column < 0) {
	    // Cannot do much if cell has not yet been layout.
	    super.paintRelayout(child, prevWidth, prevHeight,
				damagedSide, damagedLength);
//System.err.println("super.paintRelayout (cell has not yet been layout)");
	    return;
	}
	
	int tableLeft = style.borderWidth + style.paddingLeft;
	int tableTop = style.borderWidth + style.paddingTop;

	boolean sameWidth = (child.width == prevWidth);
	Style childCellStyle = childCell.style;
	int contentLeft = 
	    childCellStyle.borderWidth + childCellStyle.paddingLeft;
	int contentWidth = 
	    (childCell.width - contentLeft - 
	     childCellStyle.borderWidth - childCellStyle.paddingRight);

	if (sameWidth ||
	    (prevWidth < contentWidth && child.width <= contentWidth)) {
	    // The new width cannot have changed the overall width of the
	    // column.

	    if (!sameWidth) {
		switch (childCellStyle.textAlign) {
		case Style.TEXT_ALIGN_RIGHT:
		    child.x = 
			(tableLeft + childCell.x + childCell.width -
			 childCellStyle.borderWidth - 
			 childCellStyle.paddingRight - child.width);
		    break;
		case Style.TEXT_ALIGN_CENTER:
		    child.x = tableLeft + childCell.x + contentLeft + 
			(contentWidth - child.width)/2;
		    break;
		}
	    }

	    int contentTop = 
		childCellStyle.borderWidth + childCellStyle.paddingTop;
	    int contentHeight = 
		(childCell.height - contentTop - 
		 childCellStyle.borderWidth - childCellStyle.paddingBottom);

	    int prevBaseLine = childCell.gadgetBaseLine;
	    switch (childCellStyle.verticalAlign) {
	    case Style.VERTICAL_ALIGN_TOP:
	    case Style.VERTICAL_ALIGN_BOTTOM:
	    case Style.VERTICAL_ALIGN_MIDDLE:
		break;
	    default:
		if (child instanceof TextLines) 
		    childCell.gadgetBaseLine = 
			((TextLines) child).getFirstBaseLine();
		else
		    childCell.gadgetBaseLine = -1;
		if (childCell.gadgetBaseLine < 0)
		    childCell.gadgetBaseLine = child.height;
	    }

	    if (childCell.gadgetBaseLine < 0 || 
		(childCell.gadgetBaseLine == prevBaseLine &&
		 prevHeight-prevBaseLine < 
		 contentHeight-childCell.rowBaseLine && 
		 child.height-childCell.gadgetBaseLine <= 
		 contentHeight-childCell.rowBaseLine)) {

		// The new height/baseLine cannot have changed the overall
		// height/baseLine of the row.

		switch (childCellStyle.verticalAlign) {
		case Style.VERTICAL_ALIGN_BOTTOM:
		    child.y = 
			(tableTop + childCell.y + childCell.height - 
			 childCellStyle.borderWidth - 
			 childCellStyle.paddingBottom - child.height);
		    break;
		case Style.VERTICAL_ALIGN_MIDDLE:
		    child.y = (tableTop + childCell.y + contentTop + 
			       (contentHeight - child.height)/2);
		    break;
		}

		repaint(tableLeft + childCell.x + contentLeft, 
			tableTop + childCell.y + contentTop, 
			contentWidth, contentHeight);
//System.err.println("repaint");
		return;
	    }

	    // Optimized vertical layout (unless a cell spans several rows).
	    if (verticalRelayoutFromRow(cachedCellRowIndex, childCell.y, 
					childCell.height)) {
//System.err.println("verticalRelayoutFromRow");
		return;
	    }
	}

	super.paintRelayout(child, prevWidth, prevHeight,
			    damagedSide, damagedLength);
//System.err.println("super.paintRelayout");
    }

    private boolean verticalRelayoutFromRow(int rowIndex, 
					    int rowY, int rowHeight) {
	CellImpl[] fromRow = rows[rowIndex];
	int rowBaseLine = 0;

	for (int j = 0; j < fromRow.length; ++j) {
	    CellImpl cell = fromRow[j];
	    if (cell.rowSpan > 1)
		return false;

	    Style cellStyle = cell.style;

	    cell.height = 
		(cell.gadget.height + 2*cellStyle.borderWidth +
		 cellStyle.paddingTop + cellStyle.paddingBottom);

	    switch (cellStyle.verticalAlign) {
	    case Style.VERTICAL_ALIGN_TOP:
	    case Style.VERTICAL_ALIGN_BOTTOM:
	    case Style.VERTICAL_ALIGN_MIDDLE:
		break;
	    default:
		// cell.gadgetBaseLine is already up to date.

		cell.rowBaseLine = 
		    (cellStyle.borderWidth + cellStyle.paddingTop +
		     cell.gadgetBaseLine);

		if (cell.rowBaseLine > rowBaseLine)
		    rowBaseLine = cell.rowBaseLine;
	    }
	}

	int prevRowHeight = rowHeight;
	rowHeight = 0;

	for (int j = 0; j < fromRow.length; ++j) {
	    CellImpl cell = fromRow[j];

	    if (cell.rowBaseLine > 0) {
		cell.height += (rowBaseLine - cell.rowBaseLine);
		cell.rowBaseLine = rowBaseLine;
	    }

	    if (cell.height > rowHeight)
		rowHeight = cell.height;
	}

	int tableTop = style.borderWidth + style.paddingTop;

	for (int j = 0; j < fromRow.length; ++j) {
	    CellImpl cell = fromRow[j];

	    // The rowY has not changed.
	    cell.height = rowHeight;

	    Style cellStyle = cell.style;

	    switch (cellStyle.verticalAlign) {
	    case Style.VERTICAL_ALIGN_TOP:
		cell.gadget.y = 
		    (tableTop + cell.y + 
		     cellStyle.borderWidth + cellStyle.paddingTop);
		break;
	    case Style.VERTICAL_ALIGN_BOTTOM:
		cell.gadget.y = 
		    (tableTop + cell.y + cell.height - cellStyle.borderWidth -
		     cellStyle.paddingBottom - cell.gadget.height);
		break;
	    case Style.VERTICAL_ALIGN_MIDDLE:
		cell.gadget.y = 
		    (tableTop + cell.y + 
		     cellStyle.borderWidth + cellStyle.paddingTop) +
		    ((cell.height - 2*cellStyle.borderWidth -
		      cellStyle.paddingTop - cellStyle.paddingBottom) -
		     cell.gadget.height)/2;
		break;
	    default:
		cell.gadget.y = (tableTop + cell.y + cell.rowBaseLine - 
				 cell.gadgetBaseLine);
	    }
	}

	int delta = rowHeight - prevRowHeight;
	for (int i = rowIndex+1; i < rows.length; ++i) {
	    CellImpl[] cells = rows[i];

	    for (int j = 0; j < cells.length; ++j) {
		CellImpl cell = cells[j];

		cell.y += delta;
		cell.gadget.y += delta;
	    }
	}
	height += delta;

	parent.paintRelayout(this, width, height - delta,
			     BOTTOM_SIDE, height - (tableTop + rowY));
	return true;
    }

    protected void paintContent(Graphics g) {
	Rectangle clip = g.getClipBounds();
	int contentTop = style.borderWidth + style.paddingTop;
	int contentLeft = style.borderWidth + style.paddingLeft;

	for (int i = 0; i < rows.length; ++i) {
	    CellImpl[] row = rows[i];

	    for (int j = 0; j < row.length; ++j) {
		CellImpl cell = row[j];
		if (cell.column < 0) 
		    // Skip if cell has not yet been layout.
		    continue;

		if (clip == null || 
		    // Cell intersects clip.
		    (cell.x < clip.x+clip.width && 
		     cell.y < clip.y+clip.height &&
		     cell.x+cell.width > clip.x && 
		     cell.y+cell.height > clip.y)) {

		    Graphics g2 = 
			g.create(contentLeft+cell.x, contentTop+cell.y, 
				 cell.width, cell.height);
		    try {
			Box.drawBox(g2, cell.style, 
				    cell.width, cell.height);
		    } finally {
			g2.dispose();
		    }
		}
	    }
	}

	super.paintContent(g);
    }

    protected void printContent(Graphics g) {
	paintContent(g);
    }
}


