/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledgadget;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import com.xmlmind.gadget.GadgetContainer;

public class TextLabel extends TextLineBox {
    private char[] textChars;

    // -----------------------------------------------------------------------

    public TextLabel() {
	this(Box.DEFAULT_STYLE, "");
    }

    public TextLabel(Style style) {
	this(style, "");
    }

    public TextLabel(String text) {
	this(Box.DEFAULT_STYLE, text);
    }

    public TextLabel(Style style, String text) {
	super(style);
	setText(text);
    }

    public void setText(String text) {
	this.textChars = text.toCharArray();
    }

    public String getText() {
	return new String(textChars, 0, textChars.length);
    }

    protected void layoutContent(int constrainedWidth, int constrainedHeight,
				 boolean recursive) {
	FontMetrics metrics = getFontMetrics(style.font);
	boolean firstLine = true;
	int lineBegin = 0;
	int maxLineWidth = 0;
	int lineWidth;
	int lineCount = 1;

	int textLength = textChars.length;
	for (int i = 0; i < textLength; ++i) {
	    char c = textChars[i];

	    if (c == '\n') {
		if (i > lineBegin) {
		    lineWidth = metrics.charsWidth(textChars, 
						   lineBegin, i-lineBegin);
		    if (style.textAlign == Style.TEXT_ALIGN_LEFT) {
			if (style.textIndent > 0) 
			    lineWidth += (firstLine? style.textIndent : 0);
			else if (style.textIndent < 0)
			    lineWidth += (firstLine? 0 : -style.textIndent);
		    }
		    if (lineWidth > maxLineWidth)
			maxLineWidth = lineWidth;
		}

		firstLine = false;
		lineBegin = i+1;

		++lineCount;
	    }
	}

	// Do not forget last line if any ---

	if (textLength > lineBegin) {
	    lineWidth = metrics.charsWidth(textChars, 
					   lineBegin, textLength-lineBegin);
	    if (style.textAlign == Style.TEXT_ALIGN_LEFT) {
		if (style.textIndent > 0) 
		    lineWidth += (firstLine? style.textIndent : 0);
		else if (style.textIndent < 0)
		    lineWidth += (firstLine? 0 : -style.textIndent);
	    }
	    if (lineWidth > maxLineWidth)
		maxLineWidth = lineWidth;
	}

	width = maxLineWidth;
	height = lineCount * 
	    ((int) (0.5 + metrics.getHeight()*style.lineHeight));
    }

    protected void paintContent(Graphics g) {
	g.setColor(style.color);
	g.setFont(style.font);

	FontMetrics metrics = g.getFontMetrics(style.font);
	int ascent = metrics.getAscent();
	int descent = metrics.getDescent();
	int contentWidth = width - 
	    (2*style.borderWidth + style.paddingLeft + style.paddingRight);
	boolean firstLine = true;
	int lineBegin = 0;
	int lineX; 
	int lineHeight = (int) (0.5 + metrics.getHeight()*style.lineHeight);
	// Simplified form of (lineHeight - ascent - descent)/2 + ascent.
	int lineY = (lineHeight + ascent - descent)/2;

	int textLength = textChars.length;
	for (int i = 0; i < textLength; ++i) {
	    char c = textChars[i];

	    if (c == '\n') {
		if (i > lineBegin) {
		    switch (style.textAlign) {
		    case Style.TEXT_ALIGN_RIGHT:
			lineX = contentWidth - 
			    metrics.charsWidth(textChars, 
					       lineBegin, i-lineBegin);
			break;
		    case Style.TEXT_ALIGN_CENTER:
			lineX = (contentWidth - 
		            metrics.charsWidth(textChars, 
					       lineBegin, i-lineBegin)) / 2;
			break;
		    default:
			if (style.textIndent > 0) 
			    lineX = firstLine? style.textIndent : 0;
			else if (style.textIndent < 0)
			    lineX = firstLine? 0 : -style.textIndent;
			else
			    lineX = 0;
		    }
		    g.drawChars(textChars, lineBegin, i-lineBegin, 
				lineX, lineY);

		    switch (style.textDecoration) {
		    case Style.TEXT_DECORATION_UNDERLINE:
		        {
			    int decorY = lineY + descent - 1;
			    int lineW = metrics.charsWidth(
				textChars, lineBegin, i-lineBegin);
			    g.drawLine(lineX, decorY, lineX+lineW-1, decorY);
		        }
			break;
		    case Style.TEXT_DECORATION_OVERLINE:
		        {
			    int decorY = lineY - ascent + 1;
			    int lineW = metrics.charsWidth(
				textChars, lineBegin, i-lineBegin);
			    g.drawLine(lineX, decorY, lineX+lineW-1, decorY);
		        }
			break;
		    case Style.TEXT_DECORATION_LINE_THROUGH:
		        {
			    // Simplication of (descent - (ascent+descent)/2)
			    int decorY = lineY + (descent - ascent)/2;
			    int lineW = metrics.charsWidth(
				textChars, lineBegin, i-lineBegin);
			    g.drawLine(lineX, decorY, lineX+lineW-1, decorY);
		        }
			break;
		    }
		}

		firstLine = false;
		lineBegin = i+1;
		lineY += lineHeight;
	    }
	}

	// Paint last line if any ---

	if (textLength > lineBegin) {
	    switch (style.textAlign) {
	    case Style.TEXT_ALIGN_RIGHT:
		lineX = contentWidth - 
		    metrics.charsWidth(textChars, 
				       lineBegin, textLength-lineBegin);
		break;
	    case Style.TEXT_ALIGN_CENTER:
		lineX = (contentWidth - 
		    metrics.charsWidth(textChars, 
				       lineBegin, textLength-lineBegin)) / 2;
		break;
	    default:
		if (style.textIndent > 0) 
		    lineX = firstLine? style.textIndent : 0;
		else if (style.textIndent < 0)
		    lineX = firstLine? 0 : -style.textIndent;
		else
		    lineX = 0;
	    }
	    g.drawChars(textChars, lineBegin, textLength-lineBegin, 
			lineX, lineY);

	    switch (style.textDecoration) {
	    case Style.TEXT_DECORATION_UNDERLINE:
		{
		    int decorY = lineY + descent - 1;
		    int lineW = metrics.charsWidth(
			textChars, lineBegin, textLength-lineBegin);
		    g.drawLine(lineX, decorY, lineX+lineW-1, decorY);
		}
		break;
	    case Style.TEXT_DECORATION_OVERLINE:
		{
		    int decorY = lineY - ascent + 1;
		    int lineW = metrics.charsWidth(
			textChars, lineBegin, textLength-lineBegin);
		    g.drawLine(lineX, decorY, lineX+lineW-1, decorY);
		}
		break;
	    case Style.TEXT_DECORATION_LINE_THROUGH:
		{
		    int decorY = lineY + (descent - ascent)/2;
		    int lineW = metrics.charsWidth(
			textChars, lineBegin, textLength-lineBegin);
		    g.drawLine(lineX, decorY, lineX+lineW-1, decorY);
		}
		break;
	    }
	}
    }

    public int getFirstBaseLine() {
	FontMetrics metrics = getFontMetrics(style.font);
	int lineHeight = (int) (0.5 + metrics.getHeight()*style.lineHeight);
	int lineY = 
	    (lineHeight + metrics.getAscent() - metrics.getDescent())/2;

	return (style.borderWidth + style.paddingTop + lineY);
    }

    public int getLastBaseLine() {
	int lineCount = 1;
	int textLength = textChars.length;
	for (int i = 0; i < textLength; ++i) {
	    if (textChars[i] == '\n')
		++lineCount;
	}

	FontMetrics metrics = getFontMetrics(style.font);
	int lineHeight = (int) (0.5 + metrics.getHeight()*style.lineHeight);
	int lineY = 
	    (lineHeight + metrics.getAscent() - metrics.getDescent())/2;

	return (style.borderWidth + style.paddingTop + lineY + 
		(lineCount-1)*lineHeight);
    }
}

