/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.xmlmind.gadget.*;
import com.xmlmind.xmledit.styledgadget.*;
import fr.pixware.util.StringUtil;

public class TestTable {
    public static void main(String[] args) throws Exception {
	JFrame frame = new JFrame();
   	frame.addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        System.exit(0);
   	    }
   	});

	Style style = new Style();
	style.backgroundColor = Color.lightGray;
	style.borderStyle = Style.BORDER_STYLE_OUTSET; 
	style.borderWidth = 2; 
	style.borderTopColor = style.borderRightColor = 
	    style.borderBottomColor = style.borderLeftColor = Color.lightGray;
	style.paddingLeft = style.paddingRight =
	    style.paddingTop = style.paddingBottom = 2;

	TableBody table = new TableBody(style);
	table.setHorizontalCellSpacing(2);
	table.setVerticalCellSpacing(2);

	fillTable(table, args);

	JGadgetPanel panel = new JGadgetPanel();
	final RootGadget root = panel.getRoot();
	root.addChild(table);
	root.refresh();

	final JPanel workArea = (JPanel) frame.getContentPane();
	workArea.add(panel, BorderLayout.CENTER);

	workArea.addComponentListener(new ComponentAdapter() {
	    private Dimension dim = new Dimension();

	    public void componentResized(ComponentEvent e) {
		workArea.getSize(dim);
		root.layout(dim.width, dim.height);
		root.repaint();
	    }
	});
	
	frame.setSize(400,400);
	frame.show();
    }

    /**
     * usage: row --- row --- ... --- row
     *
     * row = cellStyle cellContent ... cellStyle cellContent 
     * 
     * cellStyle = borderWidth,paddingSize,textAlign,verticalAlign | -
     *
     * cellContent = text,t|p|i,fontSize,columnSpan,rowSpan,minContentWidth
     *
     * Inside text, '+' is replaced by '\n'.
     */
    private static void fillTable(TableBody table, String[] args) 
	throws Exception {
	int rowIndex = 0;
	int cellIndex = 0;
	Style cellStyle = null;

	for (int i = 0; i < args.length; ++i) {
	    String arg = args[i];

	    if (arg.equals("---")) {
		++rowIndex;
		cellIndex = 0;
		cellStyle = null;
	    } else {
		if (cellStyle == null) {
		    cellStyle = parseCellStyle(arg);
		} else {
		    TableBody.Cell cell = parseCellContent(arg);
		    cell.style = cellStyle;
		    cellStyle = null;

		    if (cellIndex == 0)
			table.addRow();
		    table.insertCell(rowIndex, cellIndex, cell);
		    ++cellIndex;
		}
	    }
	}
    }

    private static Style parseCellStyle(String arg) throws Exception {
	if (arg.equals("-"))
	    return com.xmlmind.xmledit.styledgadget.Box.DEFAULT_STYLE;

	Style style = new Style();
	style.backgroundColor = Color.white;
	style.borderTopColor = style.borderRightColor = 
	    style.borderBottomColor = style.borderLeftColor = Color.lightGray;
	style.borderStyle = Style.BORDER_STYLE_INSET; 
	//style.borderStyle = Style.BORDER_STYLE_SOLID;
	//style.borderSides = Style.BORDER_SIDE_BOTTOM;

	String[] split = StringUtil.split(arg, ',');
	switch (split.length) {
	case 4:
	    if (split[3].length() > 0) {
		switch (split[3].charAt(0)) {
		case 't':
		    style.verticalAlign = Style.VERTICAL_ALIGN_TOP;
		    break;
		case 'm':
		    style.verticalAlign = Style.VERTICAL_ALIGN_MIDDLE;
		    break;
		case 'b':
		    style.verticalAlign = Style.VERTICAL_ALIGN_BOTTOM;
		    break;
		default:
		    style.verticalAlign = Style.VERTICAL_ALIGN_BASELINE;
		}
	    }
	    /*FALLTHROUGH*/

	case 3:
	    if (split[2].length() > 0) {
		switch (split[2].charAt(0)) {
		case 'r':
		    style.textAlign = Style.TEXT_ALIGN_RIGHT;
		    break;
		case 'c':
		    style.textAlign = Style.TEXT_ALIGN_CENTER;
		    break;
		default:
		    style.textAlign = Style.TEXT_ALIGN_LEFT;
		}
	    }
	    /*FALLTHROUGH*/

	case 2:
	    style.paddingTop = Integer.parseInt(split[1]);
	    if (style.paddingTop < 0)
		style.paddingTop = 0;
	    style.paddingLeft = style.paddingRight = style.paddingBottom = 
		style.paddingTop;
	    /*FALLTHROUGH*/

	case 1:
	    style.borderWidth = Integer.parseInt(split[0]);
	    if (style.borderWidth < 0)
		style.borderWidth = 0;
	    break;

	default:
	    throw new IllegalArgumentException("'" + arg + 
					       "', invalid cell style");
	}

	return style;
    }

    private static TableBody.Cell parseCellContent(String arg) 
	throws Exception {
	TableBody.Cell cell = new TableBody.Cell();

	char gadgetType = 't';
	int fontSize = 12;
	String text;

	String[] split = StringUtil.split(arg, ',');
	switch (split.length) {
	case 6:
	    cell.minContentWidth = Integer.parseInt(split[5]);
	    /*FALLTHROUGH*/

	case 5:
	    cell.rowSpan = Integer.parseInt(split[4]);
	    if (cell.rowSpan < 1)
		cell.rowSpan = 1;
	    /*FALLTHROUGH*/

	case 4:
	    cell.columnSpan = Integer.parseInt(split[3]);
	    if (cell.columnSpan < 1)
		cell.columnSpan = 1;
	    /*FALLTHROUGH*/

	case 3:
	    fontSize = Integer.parseInt(split[2]);
	    if (fontSize < 8)
		fontSize = 8;
	    /*FALLTHROUGH*/

	case 2:
	    if (split[1].length() > 0)
		gadgetType = split[1].charAt(0);
	    /*FALLTHROUGH*/

	case 1:
	    text = split[0].replace('+', '\n');
	    break;

	default:
	    throw new IllegalArgumentException("'" + arg + 
					       "', invalid cell content");
	}

	Style style = new Style();
	style.font = new Font("SansSerif", Font.PLAIN, fontSize);

	switch (gadgetType) {
	case 'p':
	    {
		Paragraph paragraph = new Paragraph(style);
		paragraph.addChild(new ParagraphText(style, text));
		cell.gadget = paragraph;
	    }
	    break;
	case 'i':
	    cell.gadget = new ImageLabel(style);
	    break;
	default:
	    style.color = Color.blue;
	    cell.gadget = new TextLabel(style, text);
	}

	return cell;
    }
}
