/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledview;

import java.util.Hashtable;
import java.util.Enumeration;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.xmledit.*;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.*;
import com.xmlmind.xmledit.stylesheet.*;
import com.xmlmind.xmledit.styledgadget.*;

public class CounterManager implements DocumentListener, Traversal.Handler {
    private static final class Counter {
	public ElementModel element = null;
	public int value;
	public boolean hasChildCounters;
	public CounterVP view;
	private boolean refreshed = false;

	public Counter(ElementModel element) {
	    this.element = element;
	    value = -1;
	    hasChildCounters = false;
	    view = null;
	    refreshed = false;
	}

	public int hashCode() {
	    return System.identityHashCode(element);
	}

	public boolean equals(Object other) {
	    if (other != null && (other instanceof Counter)) 
		return (element == ((Counter) other).element);
	    else
		return false;
	}
    }

    private DocumentModel doc = null;
    private Hashtable elementToCounterMap = new Hashtable();

    private boolean batchEditing = false;
    private boolean anyUpdatedHasChildren = false;
    private int updatedCount = 0;
    private Hashtable typeToCounterMap = new Hashtable();

    // Scratch variables.
    private Counter probe = new Counter(null);

    // -----------------------------------------------------------------------

    public CounterManager() {
	this(null);
    }

    public CounterManager(DocumentModel document) {
	setDocument(document);
    }

    public void setDocument(DocumentModel document) {
	if (doc != null) {
	    doc.removeDocumentListener(this);
	    doc = null;
	    elementToCounterMap.clear();

	    // Clean-up just in case.
	    batchEditing = false;
	    anyUpdatedHasChildren = false;
	    updatedCount = 0;
	    typeToCounterMap.clear();
	}

	if (document != null) {
	    doc = document;
	    document.addDocumentListener(this, /*hasData*/ false);
	}
    }

    public DocumentModel getDocument() {
	return doc;
    }

    public 
    StyledElementViewPart counterContentToViewPart(StyleValue contentItem,
						   Style vpStyle, 
						   StyledElementView view, 
						   int role) {
	String spec = contentItem.string;
	String separator;
	int style;
	if (contentItem.type == StyleValue.COUNTER) {
	    style = contentItem.list[0].type;
	    separator = ".";
	} else {
	    spec = contentItem.string;
	    separator = contentItem.list[0].string;
	    style = contentItem.list[1].type;
	}

	CounterVP gadget = new CounterVP(vpStyle, spec, style, separator, 
					 view, role);

	probe.element = view.getElement();
	Counter counter = (Counter) elementToCounterMap.get(probe);

	// Warning: if several counters are added to the same element, only
	// the last added one is refreshed.
	//
	// At the same time, this implementation flaw has a useful
	// side-effect: when a CounterVP is replaced (ex. change attribute
	// 'numeration' from 'arabic' to 'lowerroman'), we automatically get
	// rid of the old CounterVP.
	counter.view = gadget;

	return gadget;
    }

    private void getCounters(ElementModel element, int[] counters) {
	if (!batchEditing) {
	    anyUpdatedHasChildren = false;
	    updatedCount = 0;
	}

	int depth = counters.length;

	ElementModel ancestor = element;
	while (ancestor != null && depth > 0) {
	    probe.element = ancestor;
	    Counter counter = (Counter) elementToCounterMap.get(probe);

	    --depth;
	    if (counter == null || counter.value <= 0)
		counters[depth] = addCounter(ancestor);
	    else
		counters[depth] = counter.value;

	    ancestor = ancestor.getParentElement();
	}

	while (depth > 0)
	    counters[--depth] = -1;

	if (!batchEditing && updatedCount > 0) 
	    refreshDescendantViews();
    }

    private void refreshDescendantViews() {
	boolean pass2 = (anyUpdatedHasChildren && 
			 updatedCount < elementToCounterMap.size());

	Enumeration iter = elementToCounterMap.elements();
	while (iter.hasMoreElements()) {
	    Counter counter = (Counter) iter.nextElement();

	    if (counter.refreshed) {
		counter.refreshed = false;
	    } else {
		if (pass2 && 
		    counter.view != null && counter.view.isHierarchical()) {
		    counter.view.refreshCounter();
		}
	    }
	}

	anyUpdatedHasChildren = false;
	updatedCount = 0;
    }

    private int addCounter(ElementModel element) {
	Counter counter;

	ElementModel parent = element.getParentElement();
	if (parent == null) {
	    probe.element = element;
	    counter = (Counter) elementToCounterMap.get(probe);
	    if (counter == null) {
		counter = new Counter(element);
		elementToCounterMap.put(counter, counter);
	    }
	    counter.value = 1;
	    return 1;
	}

	probe.element = parent;
	counter = (Counter) elementToCounterMap.get(probe);
	if (counter == null) {
	    counter = new Counter(parent);
	    elementToCounterMap.put(counter, counter);
	}
	counter.hasChildCounters = true;

	String elementType = element.getElementTypeName();
	Counter prevCounter = null;

	SubelementModel sub = parent.getFirstSubelement();
	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.ELEMENT) {
		ElementModel child = (ElementModel) sub;
		String childType = child.getElementTypeName();

		if (childType.equals(elementType)) {
		    probe.element = child;
		    counter = (Counter) elementToCounterMap.get(probe);
		    int prevValue;

		    if (counter == null) {
			counter = new Counter(child);
			elementToCounterMap.put(counter, counter);
			prevValue = -1;
		    } else {
			prevValue = counter.value;
		    }

		    if (prevCounter == null)
			counter.value = 1;
		    else
			counter.value = prevCounter.value + 1;

		    prevCounter = counter;

		    if (prevValue > 0 && counter.value != prevValue)
			counterUpdated(counter);
		}
	    }

	    sub = sub.getNextSubelement();
	}

	probe.element = element;
	return ((Counter) elementToCounterMap.get(probe)).value;
    }

    private void counterUpdated(Counter counter) {
	++updatedCount;
	if (counter.hasChildCounters)
	    anyUpdatedHasChildren = true;

	if (counter.view != null) {
	    // Is view still attached to the gadget tree?
	    if (counter.view.getRoot() == null) {
		counter.view = null;
	    } else {
		counter.view.refreshCounter();
		counter.refreshed = true;
	    }
	}
    }

    // -----------------------------------------------------------------------

    public void textChanged(TextEvent event, int listenerIndex) {}
    public void elementAttributeChanged(ElementAttributeEvent event, 
					int listenerIndex) {}
    public void documentEventHappened(DocumentEvent event,
				      int listenerIndex) {}

    public void batchEditingStarted(BatchEditingEvent event, 
				    int listenerIndex) {
	batchEditing = true;
	anyUpdatedHasChildren = false;
	updatedCount = 0;
    }

    public void batchEditingCompleted(BatchEditingEvent event, 
				      int listenerIndex) {
	batchEditing = false;
	if (updatedCount > 0)
	    refreshDescendantViews();
    }

    public void elementStructureChanged(ElementStructureEvent event, 
					int listenerIndex) {
	if (!batchEditing) {
	    anyUpdatedHasChildren = false;
	    updatedCount = 0;
	}

	SubelementModel sub = event.getOldChild();
	if (sub != null && sub.getSubelementType() == SubelementModel.ELEMENT)
	    Traversal.traverse(sub, this);

	ElementModel element = event.getElementSource();
	probe.element = element;
	Counter counter = (Counter) elementToCounterMap.get(probe);
	if (counter != null && counter.hasChildCounters) {
	    sub = event.getNewChild();

	    if (sub != null && 
		sub.getSubelementType() == SubelementModel.ELEMENT) {
		probe.element = (ElementModel) sub;
		counter = (Counter) elementToCounterMap.get(probe);
		if (counter == null || counter.value <= 0)
		    countNewChild(element, (ElementModel) sub);
	    }

	    recountChildren(element);
	}

	if (!batchEditing && updatedCount > 0)
	    refreshDescendantViews();
    }

    public Object processText(TextModel text) { return null; }
    public Object leaveElement(ElementModel element) { return null; }

    public Object enterElement(ElementModel element) {
	probe.element = element;
	elementToCounterMap.remove(probe);
	
	return null;
    }

    private void countNewChild(ElementModel element, ElementModel newChild) {
	String newChildType = newChild.getElementTypeName();
	boolean addCounter = false;
	Counter counter;

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.ELEMENT) {
		ElementModel child = (ElementModel) sub;

		if (child.getElementTypeName().equals(newChildType)) {
		    probe.element = child;
		    counter = (Counter) elementToCounterMap.get(probe);
		    if (counter != null && counter.value > 0) {
			addCounter = true;
			break;
		    }
		}
	    }

	    sub = sub.getNextSubelement();
	}

	if (addCounter) {
	    probe.element = newChild;
	    counter = (Counter) elementToCounterMap.get(probe);
	    if (counter == null) {
		counter = new Counter(newChild);
		elementToCounterMap.put(counter, counter);
	    }
	    // Force update in recountChildren().
	    counter.value = Integer.MAX_VALUE;
	}
    }

    private void recountChildren(ElementModel element) {
	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub.getSubelementType() == SubelementModel.ELEMENT) {
		ElementModel child = (ElementModel) sub;

		probe.element = child;
		Counter counter = (Counter) elementToCounterMap.get(probe);
		int prevValue;

		if (counter != null && (prevValue = counter.value) > 0) {
		    String childType = child.getElementTypeName();

		    counter.value = 1;
		    Counter prevCounter = 
			(Counter) typeToCounterMap.put(childType, counter);
		    if (prevCounter != null)
			counter.value = prevCounter.value + 1;

		    if (counter.value != prevValue)
			counterUpdated(counter);
		}
	    }

	    sub = sub.getNextSubelement();
	}

	typeToCounterMap.clear();
    }

    // -----------------------------------------------------------------------

    private class CounterVP extends TextLabel 
	                    implements StyledElementViewPart {
	private StyledElementView view;
	private byte role;

	private int[] counterValues;
	private int[] nextCounterValues;
	private boolean[] counterMasks;
	private int counterStyle;
	private String countersSeparator;

	public CounterVP(Style style, String counterName, int counterStyle,
			 String countersSeparator, 
			 StyledElementView view, int role) {
	    super(style);

	    this.view = view;
	    this.role = (byte) role;

	    int specLength = counterName.length();
	    counterValues = new int[specLength];
	    nextCounterValues = new int[specLength];
	    counterMasks = new boolean[specLength];

	    for (int i = specLength-1; i >= 0; --i) {
		counterMasks[i] = (counterName.charAt(i) != '-');
		counterValues[i] = -1;
		nextCounterValues[i] = 0;
	    }

	    this.counterStyle = counterStyle;
	    if (countersSeparator == null)
		countersSeparator = ".";
	    this.countersSeparator = countersSeparator;

	    refreshCounter();
	}

	public SubelementView getView() {
	    return view;
	}

	public ElementView getElementView() {
	    return view;
	}

	public StyledElementView getStyledElementView() {
	    return view;
	}

	public int getRole() {
	    return role;
	}

	public boolean isHierarchical() {
	    return (counterValues.length > 1);
	}

	public void refreshCounter() {
	    CounterManager.this.getCounters(view.getElement(), 
					    nextCounterValues);

	    boolean update = false;
	    for (int i = counterValues.length-1; i >= 0; --i) {
		if (nextCounterValues[i] != counterValues[i]) {
		    update = true;
		    break;
		}
	    }
	    if (!update)
		return;

	    int[] swap = counterValues;
	    counterValues = nextCounterValues;
	    nextCounterValues = swap;

	    StringBuffer label = new StringBuffer();

	    for (int i = 0; i < counterValues.length; ++i) {
		int value;

		if (!counterMasks[i] ||
		    (value = counterValues[i]) <= 0)
		    continue;

		if (i > 0)
		    label.append(countersSeparator);

		switch (counterStyle) {
		case StyleValue.LOWER_ALPHA:
		    CounterManager.toAlpha(value, /*lower*/ true, label);
		    break;
		case StyleValue.UPPER_ALPHA:
		    CounterManager.toAlpha(value, /*lower*/ false, label);
		    break;
		case StyleValue.LOWER_ROMAN:
		    CounterManager.toRoman(value, /*lower*/ true, label);
		    break;
		case StyleValue.UPPER_ROMAN:
		    CounterManager.toRoman(value, /*lower*/ false, label);
		    break;
		case StyleValue.DECIMAL:
		default:
		    label.append(Integer.toString(value));
		}
	    }

	    setText(label.toString());
	    refresh();
	}
    }
    
    // -----------------------------------------------------------------------

    private static void toAlpha(int n, boolean lowerCase, 
				StringBuffer alpha) {
	int begin = alpha.length();

	char zeroLetter = lowerCase? '`' : '@';

	while (n > 0) {
	    char letter = (char) (zeroLetter + (n % 27));
	    if (letter == zeroLetter)
		letter = '0';
	    alpha.insert(begin, letter);

	    n /= 27;
	}
    }

    // -----------------------------------------------------------------------

    private static final int[] ROMAN_SPECIAL_NUMBERS = { 
	1000,  900,  500,  400,  100,   90,  
	50,   40,   10,    9,    5,    4,    1 
    };
    private static final String[] ROMAN_UPPER_CASE_LETTERS = { 
	"M",  "CM",  "D",  "CD", "C",  "XC",
	"L",  "XL",  "X",  "IX", "V",  "IV", "I" 
    };       
    private static final String[] ROMAN_LOWER_CASE_LETTERS = { 
	"m",  "cm",  "d",  "cd", "c",  "xc",
	"l",  "xl",  "x",  "ix", "v",  "iv", "i" 
    };       

    private static void toRoman(int n, boolean lowerCase, 
				StringBuffer roman) {
	String[] letters = 
	    lowerCase? ROMAN_LOWER_CASE_LETTERS : ROMAN_UPPER_CASE_LETTERS;

	for (int i = 0; i < ROMAN_SPECIAL_NUMBERS.length; ++i) {
	    while (n >= ROMAN_SPECIAL_NUMBERS[i]) {
		roman.append(letters[i]);
		n -= ROMAN_SPECIAL_NUMBERS[i];
	    }
	}
    }
}

