/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledview;

import com.xmlmind.gadget.Gadget;
import com.xmlmind.gadget.GadgetContainer;
import com.xmlmind.xmledit.GadgetListImpl;
import com.xmlmind.xmledit.SubelementView;
import com.xmlmind.xmledit.model.ElementModel;
import com.xmlmind.xmledit.model.SubelementModel;
import com.xmlmind.xmledit.stylesheet.StyleValue;
import com.xmlmind.xmledit.styledgadget.Style;
import com.xmlmind.xmledit.styledgadget.Styled;

public class ParagraphView extends ContainerView {
    private Gadget viewGadget;

    public ParagraphView(ElementModel element, Style style) {
	super(element, style);
    }

    public void initView(ViewParameters param,
			 ViewParameters paramBefore,
			 Style styleBefore,
			 ViewParameters paramAfter, 
			 Style styleAfter,
			 StyledViewFactory viewFactory) {

	boolean isCustom = (param.content != null);
	if (isCustom)
	    flags |= CUSTOM_VIEW;
	if (param.display == StyleValue.INLINE_BLOCK)
	    flags |= INLINE_VIEW;

	ParagraphVP paragraph = 
	    new ParagraphVP(style, this, 
			    (isCustom? StyledElementViewPart.ROLE_CUSTOM_VIEW :
			     StyledElementViewPart.ROLE_STANDARD_VIEW));

	if (paramBefore != null && paramBefore.content != null &&
	    paramBefore.display == StyleValue.INLINE) {
	    addContent(paramBefore.content, 
		       StyledElementViewPart.ROLE_START_LABEL, paragraph);
	    paramBefore.content = null;
	}

	if (isCustom) {
	    addContent(param.content, StyledElementViewPart.ROLE_CUSTOM_VIEW,
		       paragraph);
	    param.content = null;
	}

	if (paramAfter != null && paramAfter.content != null &&
	    paramAfter.display == StyleValue.INLINE) {
	    addContent(paramAfter.content, 
		       StyledElementViewPart.ROLE_END_LABEL, paragraph);
	    paramAfter.content = null;
	}

	if (isCustom && paragraph.getChildCount() == 1) {
	    // Optimize the case where the paragraph only contains a single
	    // custom component.

	    Gadget child = paragraph.getChild(0);
	    ((StyledElementViewPart) child).setStyle(style);

	    if (child instanceof ParagraphTextVP) {
		// Convert back to a TextLabelVP! Never use a ParagraphTextVP
		// outside a ParagraphVP.

		ParagraphTextVP paragraphText = (ParagraphTextVP) child;

		child = new TextLabelVP(paragraphText.getStyle(), 
				       paragraphText.getTextAccess().getText(),
				       this, 
				       StyledElementViewPart.ROLE_CUSTOM_VIEW);
	    }

	    viewGadget = child;
	} else {
	    viewGadget = paragraph;
	}

	super.initView(param, paramBefore, styleBefore, paramAfter, styleAfter,
		       viewFactory);
    }

    private void addContent(StyledElementViewPart[] content, int role,
			    GadgetContainer container) {
	for (int i = 0; i < content.length; ++i) {
	    StyledElementViewPart part = content[i];

	    if (part instanceof TextLabelVP) {
		// Convert TextLabelVPs that does not support word wrap to
		// ParagraphTextVPs that support word wrap when added to
		// ParagraphVPs.
	        //
	        // Note that CounterManager.CounterVP is a subclass of
	        // TextLabel and not of TextLabelVP.

		TextLabelVP textLabel = (TextLabelVP) part;

		part = new ParagraphTextVP(textLabel.getStyle(), 
					   textLabel.getText(),
					   this, role);
	    }

	    container.addChild((Gadget) part);
	}
    }

    protected Gadget getViewGadget() {
	return viewGadget;
    }

    public Gadget addChildViewGadgets(SubelementView childView, 
				      int viewIndex) {
	// Find the gadget container and the index where to insert the child
	// view gadgets ---

	GadgetListImpl list = new GadgetListImpl();
	SubelementModel sub = childView.getModel();
	SubelementView subView = null;
	GadgetContainer container = null;
	int index = -1;

	for (;;) {
	    SubelementModel prevSub = sub.getPreviousSubelement();

	    if (prevSub == null) {
		sub = sub.getParentElement();

		if (sub == this.element) {
		    // First descendant, insert after start labels, before
		    // other descendant or before end labels ---

		    container = (GadgetContainer) viewGadget;
		    index = -1;

		    boolean found = false;
		    while (!found) {
			Gadget[] children = container.getChildren();

			index = children.length;
			found = true;

			loop: for (int i = 0; i < children.length; ++i) {
			    Gadget child = children[i];

			    StyledElementViewPart vp;
			    if (!(child instanceof StyledElementViewPart) ||
				(vp = ((StyledElementViewPart) 
				       child)).getView() != this) {
				index = i;
				break loop;
			    }

			    switch (vp.getRole()) {
			    case StyledElementViewPart.ROLE_START_LABEL:
				break;
			    case StyledElementViewPart.ROLE_STANDARD_VIEW:
				if (child instanceof GadgetContainer) {
				    // Anonymous paragraph.
				    container = (GadgetContainer) child;
				    found = false;
				} else {
				    // What is this???
				    index = i;
				}
				break loop;
			    case StyledElementViewPart.ROLE_CUSTOM_VIEW:
				// Inserting inside a custom view should never
				// happen.
			    case StyledElementViewPart.ROLE_END_LABEL:
				index = i;
				break loop;
			    }
			}
		    }

		    // Found container and index.
		    break;
		} else {
		    subView = (SubelementView) 
			sub.getDocumentListenerData(viewIndex);

		    if (subView != null) {
			// First descendant of a descendant group view,
			// insert after start labels if any, otherwise give up
			// and try after previous sibling ---
			
			list.removeAll();
			((GroupView) subView).getStartLabelGadgets(list);

			if (list.size > 0) {
			    Gadget lastGadget = list.list[list.size-1];
			    container = lastGadget.parent;
			    index = container.indexOfChild(lastGadget) + 1;

			    // Found container and index.
			    break;
			}
		    }
		}
		
	    } else {
		sub = prevSub;

		subView = 
		    (SubelementView) sub.getDocumentListenerData(viewIndex);
		if (subView != null) {
		    // Insert after the gadgets of the sibling ---

		    list.removeAll();
		    subView.getGadgets(list, viewIndex);

		    if (list.size > 0) {
			Gadget lastGadget = list.list[list.size-1];
			container = lastGadget.parent;
			index = container.indexOfChild(lastGadget) + 1;

			// Found container and index.
			break;
		    }
		}
	    }
	}

	// Get the gadgets to be inserted --- 
	
	list.removeAll();
	childView.getGadgets(list, viewIndex);
	if (list.size == 0)
	    // Nothing to insert, no container to refresh.
	    return null;

	// The found container is either:
	//     * A Paragraph being the viewGadget.
	//     * A Division being the viewGadget.
	//     * An ``anonymous'' Paragraph contained in the Division 
	//       viewGadget.
	//
	// Add the child view gadgets in this container ---

	if (container instanceof ParagraphVP) {
	    if (childView instanceof StyledElementView &&
		!((StyledElementView) childView).isInlineView()) {
		GadgetContainer refreshed;
		ParagraphVP paragraph = (ParagraphVP) container;

		if (paragraph == viewGadget) {
		    // Wrap paragraph into a Division.

		    Style paragraphStyle = paragraph.getStyle();
		    DivisionVP div = 
			new DivisionVP(
			    paragraphStyle, this, 
			    StyledElementViewPart.ROLE_STANDARD_VIEW);

		    paragraphStyle = paragraphStyle.duplicate();
		    paragraphStyle.resetBox();
		    paragraph.setStyle(
			getViewFactory(viewIndex).shareStyle(paragraphStyle));

		    refreshed = paragraph.parent;
		    if (refreshed != null)
			refreshed.replaceChild(paragraph, div);
		    // Otherwise, happens when the top ParagraphView is not
		    // yet attached to the DocumentView.

		    div.addChild(paragraph);
		    container = div;

		    viewGadget = div;
		} else {
		    // ``Anonymous'' ParagraphVP contained in a DivisionVP.
		    refreshed = container = paragraph.parent;
		}

		// Split paragraph at index.

		int pos = container.indexOfChild(paragraph);
		int paragraphChildCount = paragraph.getChildCount();
		if (paragraphChildCount == 0) {
		    // Useless empty paragraph.
		    container.removeChild(pos);
		    insertGadgets(container, pos, list);
		} else {
		    if (index == 0) {
			insertGadgets(container, pos, list);
		    } else if (index == paragraphChildCount) {
			insertGadgets(container, pos+1, list);
		    } else {
			ParagraphVP paragraph2 = 
			    new ParagraphVP(
				paragraph.getStyle(), this, 
                                StyledElementViewPart.ROLE_STANDARD_VIEW);

			Gadget[] moved = paragraph.getChildren();
			for (int i = index; i < paragraphChildCount; ++i) {
			    paragraph.removeChild(moved[i]);
			    paragraph2.addChild(moved[i]);
			}

			container.insertChild(pos+1, paragraph2);

			insertGadgets(container, pos+1, list);
		    }
		}

		return refreshed;
	    } else {
		// Any inlined view.
		insertGadgets(container, index, list);
		return container;
	    }
	} else {
	    // Container is a Division.

	    if (childView instanceof StyledElementView &&
		!((StyledElementView) childView).isInlineView()) {
		insertGadgets(container, index, list);
	    } else {
		// Wrap inlined gadgets into an ``anonymous'' paragraph.

		Style paragraphStyle = 
		    ((Styled) container).getStyle().duplicate();
		paragraphStyle.resetBox();
		paragraphStyle = 
		    getViewFactory(viewIndex).shareStyle(paragraphStyle);

		ParagraphVP paragraph = 
		    new ParagraphVP(
			paragraphStyle, this, 
			StyledElementViewPart.ROLE_STANDARD_VIEW);

		insertGadgets(paragraph, 0, list);
		container.insertChild(index, paragraph);

		sweepAnonymousParagraphs();
	    }

	    return container;
	}
    }

    protected static final void insertGadgets(GadgetContainer container,
					      int index, GadgetListImpl list) {
	for (int i = list.size-1; i >= 0; --i)
	    container.insertChild(index, list.list[i]);
    }

    public Gadget removeChildViewGadgets(SubelementView childView, 
					 int viewIndex) {
	GadgetListImpl list = new GadgetListImpl();
	childView.getGadgets(list, viewIndex);
	for (int i = 0; i < list.size; ++i) {
	    Gadget gadget = list.list[i];
	    gadget.parent.removeChild(gadget);
	}

	if (viewGadget instanceof ParagraphVP) 
	    return viewGadget;

	// Simplify gadget hierarchy whenever possible.

	sweepAnonymousParagraphs();

	// If the viewGadget gadget is a DivisionVP containing a single
	// ``anonymous'' ParagraphVP, let this ParagraphVP be the new
	// viewGadget.

	DivisionVP div = (DivisionVP) viewGadget;
	if (div.getChildCount() != 1)
	    return div; 

	Gadget child = div.getChild(0);
	if (child instanceof ParagraphVP &&
	    ((ParagraphVP) child).getView() == this) {
	    ParagraphVP paragraph = (ParagraphVP) child;

	    paragraph.setStyle(div.getStyle());

	    // Should work fine with a LabeledItem too.
	    GadgetContainer parent = div.parent;
	    parent.replaceChild(div, paragraph);

	    viewGadget = paragraph;

	    return parent;
	} else {
	    return div;
	}
    }

    private void sweepAnonymousParagraphs() {
	GadgetContainer viewContainerGadget = (GadgetContainer) viewGadget;
	Gadget[] children = viewContainerGadget.getChildren();
	ParagraphVP prevParagraph = null;

	for (int i = 0; i < children.length; ++i) {
	    Gadget child = children[i];

	    if (child instanceof ParagraphVP &&
		((ParagraphVP) child).getView() == this) {
		ParagraphVP paragraph = (ParagraphVP) child;
		Gadget[] moved = paragraph.getChildren();

		if (moved.length == 0) {
		    // Empty anonymous paragraph: get rid of it.
		    viewContainerGadget.removeChild(paragraph);
		} else {
		    if (prevParagraph == null) {
			prevParagraph = paragraph;
		    } else {
			// Adding several ParagraphTextVP corresponding to the
			// same TextAccess is not a problem.
			for (int j = 0; j < moved.length; ++j) {
			    paragraph.removeChild(moved[j]);
			    prevParagraph.addChild(moved[j]);
			}
			viewContainerGadget.removeChild(paragraph);
		    }
		}
	    } else {
		prevParagraph = null;
	    }
	}
    }
}
