/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledview;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.Enumeration;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.File;
import java.awt.*;
import fr.pixware.util.FileUtil;
import fr.pixware.util.StringUtil;
import com.xmlmind.awtutil.ImageResource;
import com.xmlmind.cssparser.ParseUtil;
import com.xmlmind.cssparser.ValueComponent;
import com.xmlmind.gadget.Gadget;
import com.xmlmind.xmledit.*;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.*;
import com.xmlmind.xmledit.stylesheet.*;
import com.xmlmind.xmledit.styledgadget.*;

public class StyledViewFactory extends DefaultViewFactory {
    public static final int MIN_FONT_SIZE = 6;
    public static final int MAX_FONT_SIZE = 72;
    public static String[] FONT_FAMILIES;
    static {
	java.awt.GraphicsEnvironment ge = 
	    java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment();
	FONT_FAMILIES = ge.getAvailableFontFamilyNames();
    }

    private Style defaultStyle = new Style();
    private String serifFamily = "Serif";
    private String sansSerifFamily = "SansSerif";
    private String monospaceFamily = "Monospaced";
    private double screenResolution;
    private StyleSheet styleSheet = null;

    private Hashtable sharedStyles = new Hashtable();
    private Hashtable dependencies = new Hashtable();
    private StyleValueParser styleValueParser = new StyleValueParser();
    private ExpressionEvaluator evaluator = null;
    private CounterManager counterManager = null;

    private static class FontInfo {
	public String family;
	public int style;
	public int size;

	public FontInfo(String family, int style, int size) {
	    this.family = family;
	    this.style = style;
	    this.size = size;
	}

	public boolean equals(Object other) {
	    if (other == null || !(other instanceof FontInfo))
		return false;

	    FontInfo o = (FontInfo) other;
	    return (family.equals(o.family) && style == o.style && 
		    size == o.size);
	}

	public int hashCode() {
	    return (family.hashCode() ^ style ^ size);
	}
    }
    private Hashtable fontCache = new Hashtable();

    // Scratch variables.
    private StyleSpec[] specs = new StyleSpec[] { 
	new StyleSpec(), new StyleSpec(), new StyleSpec()
    };
    private FontInfo fontProbe = new FontInfo(null, 0, 0);

    // Private constants.
    private static final StyleSpec emptySpec = new StyleSpec();
    private static Field[] styleSpecFields;
    static {
	Field[] fields = StyleSpec.class.getFields();
	Field[] instanceFields = new Field[fields.length];

	int count = 0;
	for (int i = 0; i < fields.length; ++i) {
	    Field field = fields[i];

	    if (!Modifier.isStatic(field.getModifiers()))
		instanceFields[count++] = field;
	}

	styleSpecFields = new Field[count];
	System.arraycopy(instanceFields, 0, styleSpecFields, 0, count);
    }

    private static final StyleValue noValue = new StyleValue();

    // More specific must preceed less specific (e.g. "en_US" preceeds "en").
    private static final String[] knownQuotes = {
	"fr", "\u00AB", "\u00BB",
	"de", ",,", "``",
	"en", "``", "''"
    };

    // -----------------------------------------------------------------------

    public StyledViewFactory(DocumentView documentView) {
	super(documentView);

	setScreenResolution(-1);
    }

    public void setDocument(DocumentModel document) {
	if (counterManager != null) {
	    counterManager.setDocument(null);
	    counterManager = null;
	}

	super.setDocument(document);

	// counterManager will be lazily recreated if needed to.
    }

    public void setStyleSheet(StyleSheet sheet) {
	DocumentModel doc = docView.getDocument();
	TextModel pointText = null;
	int pointOffset = 0;
	TextModel markText = null;
	int markOffset = 0;
	SubelementModel selectedSub = null;

	if (doc != null) {
	    Caret caret = docView.getCaret();
	    TextLocation point = caret.getPoint();
	    pointText = point.getText();
	    pointOffset = point.getOffset();

	    TextLocation mark = caret.getMark();
	    if (mark != null) {
		markText = mark.getText();
		markOffset = mark.getOffset();
	    }

	    SubelementSelection subelemSelection = 
		docView.getSubelementSelection();
	    selectedSub = subelemSelection.getSubelement();

	    docView.setDocument(null);
	}

	styleSheet = sheet; // May be null.
	sharedStyles.clear();
	updateDependencies();

	if (doc != null) {
	    docView.setDocument(doc);
	    docView.refresh();

	    Caret caret = docView.getCaret();
	    TextLocation point = caret.getPoint();
	    if (pointText != null) {
		point.moveTo(pointText, pointOffset);
		pointText = null;
		pointOffset = 0;
	    }

	    if (markText != null) {
		TextLocation mark = caret.mark();
		mark.moveTo(markText, markOffset);
		markText = null;
		markOffset = 0;
	    }

	    SubelementSelection subelemSelection = 
		docView.getSubelementSelection();
	    if (selectedSub != null) {
		subelemSelection.setSubelement(selectedSub);
		selectedSub = null;
	    }

	    // When a mark is moved, docView automatically scrolls to make it
	    // visible.
	}
    }

    private void updateDependencies() {
	dependencies.clear();
	if (styleSheet == null)
	    return;

	Rule[] rules = styleSheet.getRules();
	for (int i = 0; i < rules.length; ++i) {
	    Rule rule = rules[i];
	    String elementName;
	    String dependency;

	    // Analyze selector ---

	    Selector.Matcher[] matchers = rule.getSelector().getMatchers();

	    for (int j = 0; j < matchers.length; ++j) {
		Selector.Matcher matcher = matchers[j];
		Selector.Test[] tests = matcher.tests;

		// Selector.Tests are sorted in order to have
		// TEST_ELEMENT_TYPE_NAME_IS as the first one.
		if (tests.length > 0 &&
		    tests[0].type == Selector.TEST_ELEMENT_TYPE_NAME_IS)
		    elementName = tests[0].name;
		else
		    elementName = "*";
		    
		for (int k = 0; k < tests.length; ++k) {
		    Selector.Test test = tests[k];

		    switch (test.type) {
		    case Selector.TEST_HAS_ATTRIBUTE:
		    case Selector.TEST_ATTRIBUTE_EQUALS:
		    case Selector.TEST_ATTRIBUTE_CONTAINS:
		    case Selector.TEST_ATTRIBUTE_DASH_MATCH:
			dependency = elementName + ' ' + test.name;
			dependencies.put(dependency, dependency);
			break;
		    case Selector.TEST_LANGUAGE_IS:
			dependency = elementName + " xml:lang";
			dependencies.put(dependency, dependency);
			break;
		    }
		}
	    }

	    // Analyze style spec ---

	    if (matchers.length > 0 &&
		matchers[0].tests.length > 0 &&
		matchers[0].tests[0].type==Selector.TEST_ELEMENT_TYPE_NAME_IS)
		elementName = matchers[0].tests[0].name;
	    else
		elementName = "*";

	    StyleSpec spec = rule.getStyleSpec();
	    boolean anyAttribute = false;

	    if (spec.display != null) {
		if (spec.display.type == StyleValue.FUNCTION && 
		    "eval".equals(spec.display.string)) {
		    // Dynamically evaluated display.
		    anyAttribute = true;
		} else {
		    switch (spec.display.type) {
		    case StyleValue.TABLE_COLUMN:
		    case StyleValue.TABLE_COLUMN_GROUP:
		    case StyleValue.TREE:
			anyAttribute = true;
			break;
		    }
		}
	    }

	    if (anyAttribute) {
		dependency = elementName + " *";
		dependencies.put(dependency, dependency);
	    } else {
		for (int j = 0; j < styleSpecFields.length; ++j) {
		    Field field = styleSpecFields[j];

		    StyleValue val;
		    try {
			val = (StyleValue) field.get(spec);
		    } catch (IllegalAccessException cannotHappen) {
			val = null;
			cannotHappen.printStackTrace();
		    }

		    if (val != null)
			addAttrDependencies(val, elementName, dependencies);
		}
	    }
	}

//Enumeration iter = dependencies.keys();
//while (iter.hasMoreElements()) System.err.println(iter.nextElement());
    }

    private static final void addAttrDependencies(StyleValue val,
						  String elementName,
						  Hashtable dependencies) {
	switch (val.type) {
	case StyleValue.ATTRIBUTE:
	    {
		String dependency = elementName + ' ' + val.string;
		dependencies.put(dependency, dependency);
	    }
	    break;
	case StyleValue.LIST:
	case StyleValue.COUNTER:
	case StyleValue.COUNTERS:
	case StyleValue.FUNCTION:
	    {
		StyleValue[] list = val.list;

		for (int i = 0; i < list.length; ++i)
		    addAttrDependencies(list[i], elementName, dependencies);
	    }
	    break;
	}
    }

    public StyleSheet getStyleSheet() {
	return styleSheet;
    }

    public void setDefaultStyle(Style style) {
	defaultStyle = style;
    }

    public Style getDefaultStyle() {
	return defaultStyle;
    }

    public void setSerifFontFamily(String name) {
	if (name == null || !StringUtil.contains(FONT_FAMILIES, name))
	    name = "Serif";
	serifFamily = name;
    }

    public String getSerifFontFamily() {
	return serifFamily;
    }

    public void setSansSerifFontFamily(String name) {
	if (name == null || !StringUtil.contains(FONT_FAMILIES, name))
	    name = "SansSerif";
	sansSerifFamily = name;
    }

    public String getSansSerifFontFamily() {
	return sansSerifFamily;
    }

    public void setMonospaceFontFamily(String name) {
	if (name == null || !StringUtil.contains(FONT_FAMILIES, name))
	    name = "Monospaced";
	monospaceFamily = name;
    }

    public String getMonospaceFontFamily() {
	return monospaceFamily;
    }

    public void setScreenResolution(double res) {
	if (res <= 0)
	    res = docView.getPanel().getToolkit().getScreenResolution();

	screenResolution = res;
    }

    public double getScreenResolution() {
	return screenResolution;
    }

    public void setExpressionEvaluator(ExpressionEvaluator evaluator) {
	this.evaluator = evaluator;
    }

    public ExpressionEvaluator getExpressionEvaluator() {
	return evaluator;
    }

    /**
     * The caller must pass a copy of the style, if this style is to be
     * modified later.
     */
    public Style shareStyle(Style style) {
	Style shared = (Style) sharedStyles.get(style);
	if (shared == null) {
	    sharedStyles.put(style, style);
	    shared = style;
	}
	return shared;
    }

    // -----------------------------------------------------------------------

    public boolean isElementViewAttributeDependant(ElementModel element, 
						   String attributeName,
						   ElementView parentView) {
	if (styleSheet == null)
	    return super.isElementViewAttributeDependant(
		element, attributeName, parentView);

	String elementName = element.getElementTypeName();

	if (parentView == null) {
	    ElementModel parent = element.getParentElement();

	    if (parent == null) {
		// Root element.
		return (dependencies.containsKey(
		                          elementName + ' ' + attributeName) ||
			dependencies.containsKey("* " + attributeName) || 
			dependencies.containsKey(elementName + " *"));
	    } else {
		// Ancestor display is NONE or ancestor has replaced content.
		return false;
	    }
	}

	return (dependencies.containsKey(elementName + ' ' + attributeName) ||
		dependencies.containsKey("* " + attributeName) || 
		dependencies.containsKey(elementName + " *") ||
		(parentView instanceof TreeView));
    }

    public TextView createTextView(TextModel text, ElementView parentView) {
	if (styleSheet == null)
	    return super.createTextView(text, parentView);

	if (parentView == null)
	    // Ancestor display is NONE or ancestor has replaced content.
	    return null;

	if (parentView instanceof TreeView)
	    return super.createTextView(text, parentView);

	StyledElementView styledParentView = (StyledElementView) parentView;
	if (styledParentView.isCustomView())
	    return null;
	
	return new ParagraphTextView(styledParentView.getStyle(), text);
    }

    public ElementView createElementView(ElementModel element, 
					 ElementView parentView) {
	if (styleSheet == null)
	    return super.createElementView(element, parentView);

	ElementModel parent = element.getParentElement();
	StyledElementView styledParentView;
	Style parentStyle;
	if (parent == null) {
	    styledParentView = null;
	    parentStyle = null;
	} else {
	    if (parentView == null)
		// Ancestor display is NONE or ancestor has replaced content.
		return null;

	    if (parentView instanceof TreeView)
		return super.createElementView(element, parentView);

	    styledParentView = (StyledElementView) parentView;
	    if (styledParentView.isCustomView())
		return null;

	    parentStyle = styledParentView.getStyle();
	}

	StyleSpec spec;
	Style style;
	ViewParameters param;

	StyleSpec specBefore = null;
	Style styleBefore = null;
	ViewParameters paramBefore = null;

	StyleSpec specAfter = null;
	Style styleAfter = null;
	ViewParameters paramAfter = null;

	for (int j = 0; j < 3; ++j)
	    specs[j].clear();
	int found = styleSheet.findStyleSpec(element, specs);

	spec = (((found & StyleSheet.ELEMENT_MASK) == 0)? 
		emptySpec : specs[StyleSheet.ELEMENT]);
	param = new ViewParameters();
	style = specToStyle(spec, parentStyle, element, param);
	style = shareStyle(style);

	if ((found & StyleSheet.BEFORE_ELEMENT_MASK) != 0) {
	    specBefore = specs[StyleSheet.BEFORE_ELEMENT];

	    paramBefore = new ViewParameters();
	    // Pseudo-elements inherits the style of the element they are
	    // attached to.
	    styleBefore = specToStyle(specBefore, style, element, paramBefore);
	    styleBefore = shareStyle(styleBefore);
	}

	if ((found & StyleSheet.AFTER_ELEMENT_MASK) != 0) {
	    specAfter = specs[StyleSheet.AFTER_ELEMENT];

	    paramAfter = new ViewParameters();
	    styleAfter = specToStyle(specAfter, style, element, paramAfter);
	    styleAfter = shareStyle(styleAfter);
	}

	if (parent == null) {
	    // The root element cannot be represented by certain views,
	    // cannot have before or after pseudo-elements, etc.

	    switch (param.display) {
	    case StyleValue.TREE:
	    case StyleValue.NONE:
	    case StyleValue.TABLE_COLUMN_GROUP:
	    case StyleValue.TABLE_COLUMN:
		param.display = StyleValue.TREE;
		break;

	    case StyleValue.TABLE:
	    case StyleValue.INLINE_TABLE:
	    case StyleValue.TABLE_HEADER_GROUP:
	    case StyleValue.TABLE_FOOTER_GROUP:
	    case StyleValue.TABLE_ROW_GROUP:
	    case StyleValue.TABLE_ROW:
		param.display = StyleValue.TABLE;
		break;

	    case StyleValue.BLOCK:
	    case StyleValue.INLINE_BLOCK:
	    case StyleValue.MARKER:
	    case StyleValue.TABLE_CAPTION:
	    case StyleValue.TABLE_CELL:
	    case StyleValue.INLINE:
	    default:
		param.display = StyleValue.BLOCK;
	    }

	    specBefore = specAfter = null;
	    styleBefore = styleAfter = null;
	    paramBefore = paramAfter = null;
	}

	StyledElementView view;
	switch (param.display) {
	case StyleValue.NONE:
	    // Nothing to do.
	    return null;

	case StyleValue.TREE:
	    view = new StyledTreeView(element, style);
	    break;

	case StyleValue.TABLE:
	case StyleValue.INLINE_TABLE:
	    view = new TableView(element, style);
	    break;

	case StyleValue.TABLE_COLUMN_GROUP:
	    view = new ColumnGroupView(element, style);
	    break;

	case StyleValue.TABLE_COLUMN:
	    view = new ColumnView(element, style);
	    break;

	case StyleValue.TABLE_ROW_GROUP:
	case StyleValue.TABLE_HEADER_GROUP:
	case StyleValue.TABLE_FOOTER_GROUP:
	    view = new RowGroupView(element, style);
	    break;

	case StyleValue.TABLE_ROW:
	    view = new RowView(element, style);
	    break;

	case StyleValue.TABLE_CELL:
	    view = new CellView(element, style);
	    break;

	case StyleValue.TABLE_CAPTION:
	    view = new CaptionView(element, style);
	    break;

	case StyleValue.BLOCK:
	case StyleValue.INLINE_BLOCK:
	    view = new ParagraphView(element, style);
	    break;

	case StyleValue.INLINE:
	case StyleValue.MARKER: // A user mistake.
	default:
	    view = new InlineGroupView(element, style);
	}

	if (spec.content != null)
	    param.content = 
		contentToViewParts(spec.content, style, view, 
				   StyledElementViewPart.ROLE_CUSTOM_VIEW);

	if (specBefore != null) {
	    if (specBefore.content != null && 
		paramBefore.display != StyleValue.NONE) 
		paramBefore.content = 
		    contentToViewParts(specBefore.content, styleBefore, view, 
				       StyledElementViewPart.ROLE_START_LABEL);

	    if (paramBefore.content == null) {
		paramBefore = null;
		styleBefore = null;
	    }
	}

	if (specAfter != null) {
	    if (specAfter.content != null&& 
		paramAfter.display != StyleValue.NONE)
		paramAfter.content = 
		    contentToViewParts(specAfter.content, styleAfter, view, 
				       StyledElementViewPart.ROLE_END_LABEL);

	    if (paramAfter.content == null) {
		paramAfter = null;
		styleAfter = null;
	    }
	}

	view.initView(param, paramBefore, styleBefore, paramAfter, styleAfter,
		      this);

	return view;
    }

    // -----------------------------------------------------------------------

    private Style specToStyle(StyleSpec spec, Style parentStyle, 
			      ElementModel element, ViewParameters param) {
	Style style;
	if (parentStyle == null) {
	    parentStyle = defaultStyle;
	    style = defaultStyle.duplicate();
	} else {
	    style = parentStyle.duplicate();
	    style.resetNonInherited();
	}

	StyleValue val;
	Font parentFont = parentStyle.font;
	String fontFamily = parentFont.getFamily();
	int fontStyle = parentFont.getStyle();
	int fontSize = parentFont.getSize();
	FontMetrics metrics = null;

	if (spec.fontFamily != null) {
	    val = eval(spec.fontFamily, element);

	    switch (val.type) {
	    case StyleValue.SERIF:
		fontFamily = serifFamily;
		break;
	    case StyleValue.SANS_SERIF:
		fontFamily = sansSerifFamily;
		break;
	    case StyleValue.MONOSPACE:
		fontFamily = monospaceFamily;
		break;
	    }
	}

	if (spec.fontStyle != null) {
	    val = eval(spec.fontStyle, element);

	    switch (val.type) {
	    case StyleValue.NORMAL:
		fontStyle &= ~Font.ITALIC;
		break;
	    case StyleValue.ITALIC:
	    case StyleValue.OBLIQUE:
		fontStyle |= Font.ITALIC;
		break;
	    }
	}

	if (spec.fontWeight != null) {
	    val = eval(spec.fontWeight, element);

	    switch (val.type) {
	    case StyleValue.NORMAL:
		fontStyle &= ~Font.BOLD;
		break;
	    case StyleValue.BOLD:
		fontStyle |= Font.BOLD;
		break;
	    }
	}

	if (spec.fontSize != null) {
	    val = eval(spec.fontSize, element);

	    switch (val.type) {
	    case StyleValue.MEDIUM:
		fontSize = defaultStyle.font.getSize();
		break;
	    case StyleValue.SMALL:
		fontSize = defaultStyle.font.getSize() - 2;
		break;
	    case StyleValue.LARGE:
		fontSize = defaultStyle.font.getSize() + 4;
		break;
	    case StyleValue.X_SMALL:
		fontSize = defaultStyle.font.getSize() - 4;
		break;
	    case StyleValue.X_LARGE:
		fontSize = defaultStyle.font.getSize() + 8;
		break;
	    case StyleValue.XX_SMALL:
		fontSize = defaultStyle.font.getSize() - 6;
		break;
	    case StyleValue.XX_LARGE:
		fontSize = defaultStyle.font.getSize() + 12;
		break;

	    case StyleValue.LARGER:
		fontSize += 2;
		break;
	    case StyleValue.SMALLER:
		fontSize -= 2;
		break;

	    case StyleValue.PERCENTAGE:
		fontSize = (int) Math.rint((fontSize * val.number)/100);
		break;

	    case StyleValue.LENGTH_IN:
		fontSize = (int) Math.rint(val.number * 72.0);
		break;
	    case StyleValue.LENGTH_CM:
		fontSize = (int) Math.rint((val.number/2.54) * 72.0);
		break;
	    case StyleValue.LENGTH_MM:
		fontSize = (int) Math.rint((val.number/25.4) * 72.0);
		break;
	    case StyleValue.LENGTH_PT:
		fontSize = (int) Math.rint(val.number);
		break;
	    case StyleValue.LENGTH_PC:
		fontSize = (int) Math.rint(val.number * 12.0);
		break;
	    case StyleValue.LENGTH_EM:
		fontSize = (int) Math.rint(val.number * fontSize);
		break;
	    case StyleValue.LENGTH_EX:
		// 'x' height is approximately half the ascent.
		metrics = docView.getFontMetrics(parentFont);
		fontSize = (int) 
		    Math.rint((((val.number * metrics.getAscent())/2) / 
			       screenResolution) * 72.0);
		break;
	    case StyleValue.LENGTH_PX:
	    case StyleValue.NUMBER:
		fontSize = 
		    (int) Math.rint((val.number/screenResolution) * 72.0);
		break;
	    }
	}
	if (fontSize < MIN_FONT_SIZE)
	    fontSize = MIN_FONT_SIZE;
	else if (fontSize > MAX_FONT_SIZE)
	    fontSize = MAX_FONT_SIZE;

	if (fontSize != parentFont.getSize() ||
	    fontStyle != parentFont.getStyle() ||
	    !fontFamily.toLowerCase().equals(
		parentFont.getFamily().toLowerCase())) 
	    style.font = getFont(fontFamily, fontStyle, fontSize);

	metrics = docView.getFontMetrics(style.font);

	if (spec.marginTop != null) {
	    val = eval(spec.marginTop, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.marginTop = parentStyle.marginTop;
		break;
	    case StyleValue.AUTO:
		style.marginTop = -1;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.marginTop = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.marginRight != null) {
	    val = eval(spec.marginRight, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.marginRight = parentStyle.marginRight;
		break;
	    case StyleValue.AUTO:
		style.marginRight = -1;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.marginRight = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.marginBottom != null) {
	    val = eval(spec.marginBottom, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.marginBottom = parentStyle.marginBottom;
		break;
	    case StyleValue.AUTO:
		style.marginBottom = -1;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.marginBottom = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.marginLeft != null) {
	    val = eval(spec.marginLeft, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.marginLeft = parentStyle.marginLeft;
		break;
	    case StyleValue.AUTO:
		style.marginLeft = -1;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.marginLeft = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.paddingTop != null) {
	    val = eval(spec.paddingTop, element);

	    if (val.type == StyleValue.INHERIT) {
		style.paddingTop = parentStyle.paddingTop;
	    } else {
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.paddingTop = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.paddingRight != null) {
	    val = eval(spec.paddingRight, element);

	    if (val.type == StyleValue.INHERIT) {
		style.paddingRight = parentStyle.paddingRight;
	    } else {
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.paddingRight = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.paddingBottom != null) {
	    val = eval(spec.paddingBottom, element);

	    if (val.type == StyleValue.INHERIT) {
		style.paddingBottom = parentStyle.paddingBottom;
	    } else {
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.paddingBottom = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.paddingLeft != null) {
	    val = eval(spec.paddingLeft, element);

	    if (val.type == StyleValue.INHERIT) {
		style.paddingLeft = parentStyle.paddingLeft;
	    } else {
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.paddingLeft = 
			lengthToPixels(val.type, val.number, 
				       metrics, screenResolution);
	    }
	}

	if (spec.borderStyle != null) {
	    val = eval(spec.borderStyle, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.borderStyle = parentStyle.borderStyle;
		break;
	    case StyleValue.NONE:
		style.borderStyle = Style.BORDER_STYLE_NONE;
		break;
	    case StyleValue.DOTTED:
		style.borderStyle = Style.BORDER_STYLE_DOTTED;
		break;
	    case StyleValue.DASHED:
		style.borderStyle = Style.BORDER_STYLE_DASHED;
		break;
	    case StyleValue.SOLID:
		style.borderStyle = Style.BORDER_STYLE_SOLID;
		break;
	    case StyleValue.DOUBLE:
		style.borderStyle = Style.BORDER_STYLE_DOUBLE;
		break;
	    case StyleValue.GROOVE:
		style.borderStyle = Style.BORDER_STYLE_GROOVE;
		break;
	    case StyleValue.RIDGE:
		style.borderStyle = Style.BORDER_STYLE_RIDGE;
		break;
	    case StyleValue.INSET:
		style.borderStyle = Style.BORDER_STYLE_INSET;
		break;
	    case StyleValue.OUTSET:
		style.borderStyle = Style.BORDER_STYLE_OUTSET;
		break;
	    }
	}

	if (spec.borderWidth != null) {
	    val = eval(spec.borderWidth, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.borderWidth = parentStyle.borderWidth;
		break;
	    case StyleValue.THIN:
		style.borderWidth = 1;
		break;
	    case StyleValue.MEDIUM:
		style.borderWidth = 2;
		break;
	    case StyleValue.THICK:
		style.borderWidth = 4;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    style.borderWidth = 
			lengthToPixels(val.type, val.number,
				       metrics, screenResolution);
	    }
	}

	Color namedColor;

	if (spec.borderTopColor != null) {
	    val = eval(spec.borderTopColor, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.borderTopColor = parentStyle.borderTopColor;
		break;
	    case StyleValue.TRANSPARENT:
		style.borderTopColor = null;
		break;
	    case StyleValue.COLOR:
		style.borderTopColor = val.color;
		break;
	    case StyleValue.IDENTIFIER:
		namedColor = val.colorValue();
		if (namedColor != null)
		    style.borderTopColor = namedColor;
		break;
	    }
	}

	if (spec.borderRightColor != null) {
	    val = eval(spec.borderRightColor, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.borderRightColor = parentStyle.borderRightColor;
		break;
	    case StyleValue.TRANSPARENT:
		style.borderRightColor = null;
		break;
	    case StyleValue.COLOR:
		style.borderRightColor = val.color;
		break;
	    case StyleValue.IDENTIFIER:
		namedColor = val.colorValue();
		if (namedColor != null)
		    style.borderRightColor = namedColor;
		break;
	    }
	}

	if (spec.borderBottomColor != null) {
	    val = eval(spec.borderBottomColor, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.borderBottomColor = parentStyle.borderBottomColor;
		break;
	    case StyleValue.TRANSPARENT:
		style.borderBottomColor = null;
		break;
	    case StyleValue.COLOR:
		style.borderBottomColor = val.color;
		break;
	    case StyleValue.IDENTIFIER:
		namedColor = val.colorValue();
		if (namedColor != null)
		    style.borderBottomColor = namedColor;
		break;
	    }
	}

	if (spec.borderLeftColor != null) {
	    val = eval(spec.borderLeftColor, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.borderLeftColor = parentStyle.borderLeftColor;
		break;
	    case StyleValue.TRANSPARENT:
		style.borderLeftColor = null;
		break;
	    case StyleValue.COLOR:
		style.borderLeftColor = val.color;
		break;
	    case StyleValue.IDENTIFIER:
		namedColor = val.colorValue();
		if (namedColor != null)
		    style.borderLeftColor = namedColor;
		break;
	    }
	}

	if (spec.backgroundColor != null) {
	    val = eval(spec.backgroundColor, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.backgroundColor = parentStyle.backgroundColor;
		break;
	    case StyleValue.TRANSPARENT:
		style.backgroundColor = null;
		break;
	    case StyleValue.COLOR:
		style.backgroundColor = val.color;
		break;
	    case StyleValue.IDENTIFIER:
		namedColor = val.colorValue();
		if (namedColor != null)
		    style.backgroundColor = namedColor;
		break;
	    }
	}

	if (spec.color != null) {
	    val = eval(spec.color, element);

	    switch (val.type) {
	    case StyleValue.COLOR:
		style.color = val.color;
		break;
	    case StyleValue.IDENTIFIER:
		namedColor = val.colorValue();
		if (namedColor != null)
		    style.color = namedColor;
		break;
	    }
	}

	if (spec.textDecoration != null) {
	    val = eval(spec.textDecoration, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.textDecoration = parentStyle.textDecoration;
		break;
	    case StyleValue.NONE:
		style.textDecoration = Style.TEXT_DECORATION_NONE;
		break;
	    case StyleValue.UNDERLINE:
		style.textDecoration = Style.TEXT_DECORATION_UNDERLINE;
		break;
	    case StyleValue.OVERLINE:
		style.textDecoration = Style.TEXT_DECORATION_OVERLINE;
		break;
	    case StyleValue.LINE_THROUGH:
		style.textDecoration = Style.TEXT_DECORATION_LINE_THROUGH;
		break;
	    }
	}

	if (spec.textAlign != null) {
	    val = eval(spec.textAlign, element);

	    switch (val.type) {
	    case StyleValue.LEFT:
		style.textAlign = Style.TEXT_ALIGN_LEFT;
		break;
	    case StyleValue.RIGHT:
		style.textAlign = Style.TEXT_ALIGN_RIGHT;
		break;
	    case StyleValue.CENTER:
		style.textAlign = Style.TEXT_ALIGN_CENTER;
		break;
	    }
	}

	if (spec.textIndent != null) {
	    val = eval(spec.textIndent, element);

	    if (((val.type >= StyleValue.LENGTH_PX && 
		  val.type <= StyleValue.LENGTH_EX) || 
		 val.type == StyleValue.NUMBER))
		style.textIndent = lengthToPixels(val.type, val.number,
						  metrics, screenResolution);
	}

	if (spec.verticalAlign != null) {
	    val = eval(spec.verticalAlign, element);

	    switch (val.type) {
	    case StyleValue.INHERIT:
		style.verticalAlign = parentStyle.verticalAlign;
		break;
	    case StyleValue.BASELINE:
		style.verticalAlign = Style.VERTICAL_ALIGN_BASELINE;
		break;
	    case StyleValue.SUB:
		style.verticalAlign = Style.VERTICAL_ALIGN_SUB;
		break;
	    case StyleValue.SUPER:
		style.verticalAlign = Style.VERTICAL_ALIGN_SUPER;
		break;
	    case StyleValue.TOP:
		style.verticalAlign = Style.VERTICAL_ALIGN_TOP;
		break;
	    case StyleValue.TEXT_TOP:
		style.verticalAlign = Style.VERTICAL_ALIGN_TEXT_TOP;
		break;
	    case StyleValue.MIDDLE:
		style.verticalAlign = Style.VERTICAL_ALIGN_MIDDLE;
		break;
	    case StyleValue.BOTTOM:
		style.verticalAlign = Style.VERTICAL_ALIGN_BOTTOM;
		break;
	    case StyleValue.TEXT_BOTTOM:
		style.verticalAlign = Style.VERTICAL_ALIGN_TEXT_BOTTOM;
		break;
	    }
	}

	if (spec.lineHeight != null) {
	    val = eval(spec.lineHeight, element);

	    switch (val.type) {
	    case StyleValue.NORMAL:
		style.lineHeight = 1.0;
		break;
	    case StyleValue.NUMBER:
		if (val.number > 0)
		    style.lineHeight = val.number;
		break;
	    }
	}

	if (spec.whiteSpace != null) {
	    val = eval(spec.whiteSpace, element);

	    switch (val.type) {
	    case StyleValue.NORMAL:
		style.whiteSpace = Style.WHITE_SPACE_NORMAL;
		break;
	    case StyleValue.PRE:
		style.whiteSpace = Style.WHITE_SPACE_PRE;
		break;
	    case StyleValue.NOWRAP:
		style.whiteSpace = Style.WHITE_SPACE_NOWRAP;
		break;
	    }
	}

	// View parameters ---

	if (spec.display != null) {
	    val = eval(spec.display, element);

	    switch (val.type) {
	    case StyleValue.BLOCK:
	    case StyleValue.INLINE:
	    case StyleValue.INLINE_BLOCK:
	    case StyleValue.INLINE_TABLE:
	    case StyleValue.MARKER:
	    case StyleValue.NONE:
	    case StyleValue.TABLE:
	    case StyleValue.TABLE_CAPTION:
	    case StyleValue.TABLE_CELL:
	    case StyleValue.TABLE_ROW:
	    case StyleValue.TABLE_HEADER_GROUP:
	    case StyleValue.TABLE_FOOTER_GROUP:
	    case StyleValue.TABLE_ROW_GROUP:
	    case StyleValue.TABLE_COLUMN:
	    case StyleValue.TABLE_COLUMN_GROUP:
	    case StyleValue.TREE:
		param.display = val.type;
		break;
	    }
	}

	if (spec.markerOffset != null) {
	    val = eval(spec.markerOffset, element);

	    switch (val.type) {
	    case StyleValue.AUTO:
		param.markerOffset = ViewParameters.DEFAULT_MARKER_OFFSET;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number >= 0)
		    param.markerOffset = 
			lengthToPixels(val.type, val.number,
				       metrics, screenResolution);
	    }
	}

	if (spec.width != null) {
	    val = eval(spec.width, element);

	    switch (val.type) {
	    case StyleValue.AUTO:
		param.width = -1;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number > 0)
		    param.width = lengthToPixels(val.type, val.number,
						 metrics, screenResolution);
	    }
	}

	if (spec.height != null) {
	    val = eval(spec.height, element);

	    switch (val.type) {
	    case StyleValue.AUTO:
		param.height = -1;
		break;
	    default:
		if (((val.type >= StyleValue.LENGTH_PX && 
		      val.type <= StyleValue.LENGTH_EX) || 
		     val.type == StyleValue.NUMBER) &&
		    val.number > 0)
		    param.height = lengthToPixels(val.type, val.number,
						  metrics, screenResolution);
	    }
	}

	if (spec.captionSide != null) {
	    val = eval(spec.captionSide, element);

	    switch (val.type) {
	    case StyleValue.TOP:
	    case StyleValue.BOTTOM:
		param.captionSide = val.type;
		break;
	    }
	}

	if (spec.borderHorizontalSpacing != null) {
	    val = eval(spec.borderHorizontalSpacing, element);

	    if (((val.type >= StyleValue.LENGTH_PX && 
		  val.type <= StyleValue.LENGTH_EX) || 
		 val.type == StyleValue.NUMBER) &&
		val.number >= 0)
		param.borderHorizontalSpacing = 
		    lengthToPixels(val.type, val.number,
				   metrics, screenResolution);
	}

	if (spec.borderVerticalSpacing != null) {
	    val = eval(spec.borderVerticalSpacing, element);

	    if (((val.type >= StyleValue.LENGTH_PX && 
		  val.type <= StyleValue.LENGTH_EX) || 
		 val.type == StyleValue.NUMBER) &&
		val.number >= 0)
		param.borderVerticalSpacing = 
		    lengthToPixels(val.type, val.number,
				   metrics, screenResolution);
	}

	if (spec.rowSpan != null) {
	    val = eval(spec.rowSpan, element);

	    if (val.type == StyleValue.NUMBER &&
		(int) val.number > 0)
		param.rowSpan = (int) val.number;
	}

	if (spec.columnSpan != null) {
	    val = eval(spec.columnSpan, element);

	    if (val.type == StyleValue.NUMBER &&
		(int) val.number > 0)
		param.columnSpan = (int) val.number;
	}

	return style;
    }

    public static int lengthToPixels(int unit, double value, 
				     FontMetrics metrics, 
				     double screenResolution) {
	double pixels;

	switch (unit) {
	case StyleValue.LENGTH_IN:
	    pixels = value * screenResolution;
	    break;
	case StyleValue.LENGTH_CM:
	    pixels = (value/2.54) * screenResolution;
	    break;
	case StyleValue.LENGTH_MM:
	    pixels = (value/25.4) * screenResolution;
	    break;
	case StyleValue.LENGTH_PT:
	    // 1in = 72pt.
	    pixels = (value/72.0) * screenResolution;
	    break;
	case StyleValue.LENGTH_PC:
	    // 1pc = 12pt ==> 1in = 6pc.
	    pixels = (value/6.0) * screenResolution;
	    break;
	case StyleValue.LENGTH_EM:
	    pixels = value * metrics.getHeight();
	    break;
	case StyleValue.LENGTH_EX:
	    // 'x' height is approximately half the ascent.
	    pixels = (value * metrics.getAscent())/2;
	    break;
	case StyleValue.LENGTH_PX:
	case StyleValue.NUMBER:
	default:
	    pixels = value;
	}

	return (int) Math.rint(pixels);
    }

    /**
     * Creating a Font seems to be a pretty heavyweight operation for JDK1.3.
     */
    private Font getFont(String family, int style, int size) {
	fontProbe.family = family;
	fontProbe.style = style;
	fontProbe.size = size;
	Font font = (Font) fontCache.get(fontProbe);
	if (font == null) {
	    font = new Font(family, style, size);
	    fontCache.put(new FontInfo(family, style, size), font);
	}

	return font;
    }

    // -----------------------------------------------------------------------

    private StyleValue eval(StyleValue val, ElementModel element) {
	if (val.type == StyleValue.FUNCTION && "eval".equals(val.string)) {
	    String result;
	    try {
		result = eval(val.list, element);
	    } catch (Exception ignored) {
		result = null;
	    }
	    if (result == null)
		return noValue;

	    StyleValue parsed = styleValueParser.parse(result, 
						       styleSheet.getURL());
	    return (parsed == null)? noValue : parsed;
	} else {
	    return val;
	}
    }

    private String eval(StyleValue[] args, ElementModel element) 
	throws Exception {
	StringBuffer expr = new StringBuffer();

	for (int i = 0; i < args.length; ++i) 
	    toString(args[i], element, expr);

	// Even without an evaluator, being able to specify
	// eval(attr(width), "px") is useful.

	return ((evaluator == null)? expr.toString() : 
		evaluator.evaluateExpression(expr.toString(), element));
    }

    private void toString(StyleValue val, ElementModel element, 
			  StringBuffer buffer) {
	switch (val.type) {
	case StyleValue.NUMBER:
	    buffer.append(val.number);
	    break;
	case StyleValue.STRING:
	case StyleValue.IDENTIFIER:
	    buffer.append(val.string);
	    break;
	case StyleValue.LIST:
	    {
		StyleValue[] list = val.list;

		for (int i = 0; i < list.length; ++i) {
		    if (i > 0)
			buffer.append(' ');
		    toString(list[i], element, buffer);
		}
	    }
	    break;
	case StyleValue.LENGTH_PX:
	    buffer.append(val.number);
	    buffer.append("px");
	    break;
	case StyleValue.LENGTH_IN:
	    buffer.append(val.number);
	    buffer.append("in");
	    break;
	case StyleValue.LENGTH_CM:
	    buffer.append(val.number);
	    buffer.append("cm");
	    break;
	case StyleValue.LENGTH_MM:
	    buffer.append(val.number);
	    buffer.append("mm");
	    break;
	case StyleValue.LENGTH_PT:
	    buffer.append(val.number);
	    buffer.append("pt");
	    break;
	case StyleValue.LENGTH_PC:
	    buffer.append(val.number);
	    buffer.append("pc");
	    break;
	case StyleValue.LENGTH_EM:
	    buffer.append(val.number);
	    buffer.append("em");
	    break;
	case StyleValue.LENGTH_EX:
	    buffer.append(val.number);
	    buffer.append("ex");
	    break;
	case StyleValue.ANGLE_DEG:
	    buffer.append(val.number);
	    buffer.append("deg");
	    break;
	case StyleValue.ANGLE_GRAD:
	    buffer.append(val.number);
	    buffer.append("grad");
	    break;
	case StyleValue.ANGLE_RAD:
	    buffer.append(val.number);
	    buffer.append("rad");
	    break;
	case StyleValue.TIME_MS:
	    buffer.append(val.number);
	    buffer.append("ms");
	    break;
	case StyleValue.TIME_S:
	    buffer.append(val.number);
	    buffer.append("s");
	    break;
	case StyleValue.FREQUENCY_HZ:
	    buffer.append(val.number);
	    buffer.append("hz");
	    break;
	case StyleValue.FREQUENCY_KHZ:
	    buffer.append(val.number);
	    buffer.append("khz");
	    break;
	case StyleValue.PERCENTAGE:
	    buffer.append(val.number);
	    buffer.append("%");
	    break;
	case StyleValue.DIMENSION:
	    buffer.append(val.number);
	    buffer.append(val.string);
	    break;
	case StyleValue.URI:
	    // The StyleSheetLoader always returns fully resolved URLs.
	    buffer.append(val.string);
	    break;
	case StyleValue.ATTRIBUTE:
	    {
		String attrValue = 
		    ElementUtil.getElementAttribute(element, val.string);

		if (attrValue != null)
		    buffer.append(attrValue);
	    }
	    break;
	case StyleValue.COLOR:
	    {
		Color color = val.color;

		buffer.append("rgb(");
		buffer.append(color.getRed());
		buffer.append(',');
		buffer.append(color.getGreen());
		buffer.append(',');
		buffer.append(color.getBlue());
		buffer.append(')');
	    }
	    break;
	case StyleValue.COUNTER:
	    buffer.append("counter(");
	    buffer.append(val.string);
	    buffer.append(',');
	    toString(val.list[0], element, buffer);
	    buffer.append(')');
	    break;
	case StyleValue.COUNTERS:
	    buffer.append("counters(");
	    buffer.append(val.string);
	    buffer.append(',');
	    toString(val.list[0], element, buffer);
	    buffer.append(',');
	    toString(val.list[1], element, buffer);
	    buffer.append(')');
	    break;
	case StyleValue.FUNCTION:
	    if ("eval".equals(val.string)) {
		try {
		    buffer.append(eval(val.list, element));
		} catch (Exception ignored) {}
	    } else {
		buffer.append(val.string);
		buffer.append('(');

		StyleValue[] args = val.list;

		for (int i = 0; i < args.length; ++i) {
		    if (i > 0)
			buffer.append(',');
		    toString(args[i], element, buffer);
		}

		buffer.append(')');
	    }
	    break;
	default:
	    { 
		String keyword = StyleValue.getKeyword(val.type);
		if (keyword != null)
		    buffer.append(keyword);
	    }
	    break;
	}
    }

    // -----------------------------------------------------------------------

    private StyledElementViewPart[] contentToViewParts(StyleValue val, 
						       Style contentStyle,
						       StyledElementView view,
						       int role) {
	// The box is around the whole generated content, not around each
	// of its components.
	Style vpStyle = contentStyle.duplicate();
	vpStyle.resetBox();
	vpStyle = shareStyle(vpStyle);

	ElementModel element = view.getElement();

	StyleValue[] list;
	if (val.type == StyleValue.LIST)
	    list = val.list;
	else
	    list = new StyleValue[] { val };

	StyledElementViewPart[] partList = 
	    new StyledElementViewPart[list.length];
	int partCount = 0;
	StringBuffer buffer = new StringBuffer();

	for (int i = 0; i < list.length; ++i) {
	    StyleValue item = eval(list[i], element);

	    switch (item.type) {
	    case StyleValue.NO_OPEN_QUOTE:
	    case StyleValue.NO_CLOSE_QUOTE:
		// Nothing to do.
		break;

	    case StyleValue.STRING:
		buffer.append(item.string);
		break;
	    case StyleValue.OPEN_QUOTE:
		buffer.append(nativeQuote(element, /*closing*/ false));
		break;
	    case StyleValue.CLOSE_QUOTE:
		buffer.append(nativeQuote(element, /*closing*/ true));
		break;
	    case StyleValue.ATTRIBUTE:
		{
		    String value = ElementUtil.getElementAttribute(
			element, item.string);
		    if (value != null)
			buffer.append(value);
		}
		break;

	    case StyleValue.URI:
	    case StyleValue.DISC:
	    case StyleValue.CIRCLE:
	    case StyleValue.SQUARE:
		{
		   Image image = null;

		   switch (item.type) {
		   case StyleValue.URI:
		       {
			    try {
				image = Toolkit.getDefaultToolkit().getImage(
				    new URL(item.string));
			    } catch (MalformedURLException ignored) {}

			   if (image != null) 
			       image = ImageResource.loadImage(image);

			   if (image == null) 
			       image = ImageLabel.DEFAULT_IMAGE;
		       }
		       break;
		   case StyleValue.DISC:
		       image = Icon.get(Icon.DISC, vpStyle.color, 
					vpStyle.backgroundColor);
		       break;
		   case StyleValue.CIRCLE:
		       image = Icon.get(Icon.CIRCLE, vpStyle.color, 
					vpStyle.backgroundColor);
		       break;
		   case StyleValue.SQUARE:
		       image = Icon.get(Icon.SQUARE, vpStyle.color, 
					vpStyle.backgroundColor);
		       break;
		   }

		   if (image != null) {
		       if (buffer.length() > 0) {
			   partList[partCount++] = 
			       new TextLabelVP(vpStyle, buffer.toString(), 
					       view, role);
			   buffer = new StringBuffer();
		       }
		       partList[partCount++] = 
			   new ImageLabelVP(vpStyle, image, view, role);
		   }
		}
		break;

	    case StyleValue.COUNTER:
	    case StyleValue.COUNTERS:
		{
		    StyledElementViewPart part = 
			counterContentToViewPart(item, vpStyle, view, role);

		    if (part != null) {
			if (buffer.length() > 0) {
			    partList[partCount++] = 
				new TextLabelVP(vpStyle, buffer.toString(),
						view, role);
			    buffer = new StringBuffer();
			}
			partList[partCount++] = part;
		    }
		}
		break;

	    case StyleValue.FUNCTION:
		{
		    StyledElementViewPart part = null;

		    if ("icon".equals(item.string))
			part = iconContentToViewPart(item.list, vpStyle, 
						     view, role);
		    else if ("image".equals(item.string))
			part = imageContentToViewPart(item.list, vpStyle, 
						      view, role);
		    else if ("component".equals(item.string))
			part = componentContentToViewPart(item.list, vpStyle, 
							  view, role);
		    else if ("rows".equals(item.string))
			part = rowsContentToViewPart(item.list, vpStyle, 
						     view, role);
		    else if ("row".equals(item.string))
			part = rowContentToViewPart(item.list, vpStyle, 
						    view, role);
		    else if ("cell".equals(item.string))
			part = cellContentToViewPart(item.list, vpStyle, 
						     view, role);
		    else if ("paragraph".equals(item.string))
			part = paragraphContentToViewPart(item.list, vpStyle, 
							  view, role);

		    if (part != null) {
			if (buffer.length() > 0) {
			    partList[partCount++] = 
				new TextLabelVP(vpStyle, buffer.toString(),
						view, role);
			    buffer = new StringBuffer();
			}
			partList[partCount++] = part;
		    }
		}
		break;
	    }
	}
	if (buffer.length() > 0) 
	    partList[partCount++] = 
		new TextLabelVP(vpStyle, buffer.toString(), view, role);

	StyledElementViewPart[] result;

	if (partCount > 0) {
	    if (partCount == partList.length) {
		result = partList;
	    } else {
		StyledElementViewPart[] parts = 
		    new StyledElementViewPart[partCount];
		System.arraycopy(partList, 0, parts, 0, partCount);
		result = parts;
	    }
	} else {
	    result = null;
	}

	return result;
    }

    public static final String nativeQuote(ElementModel element, 
					   boolean closing) {
	String language = ElementUtil.getLanguage(element); // "xml:lang"
	if (language == null) 
	    language = ElementUtil.lookupElementAttribute(element, "lang");

	if (language != null) {
	    for (int i = 0; i < knownQuotes.length; i += 3) {
		if (language.startsWith(knownQuotes[i])) 
		    return (closing? knownQuotes[i+2] : knownQuotes[i+1]);
	    }
	}

	return "\"";
    }

    // -----------------------------------------------------------------------

    private
    StyledElementViewPart counterContentToViewPart(StyleValue contentItem,
						   Style vpStyle, 
						   StyledElementView view, 
						   int role) {
	if (counterManager == null) 
	    counterManager = new CounterManager(doc);
	return counterManager.counterContentToViewPart(contentItem, vpStyle, 
						       view, role);
    }

    // -----------------------------------------------------------------------

    private StyledElementViewPart iconContentToViewPart(StyleValue[] args,
							Style vpStyle,
							StyledElementView view,
							int role) {
	String iconName = null;
	if (args.length != 1 || (iconName = args[0].stringValue()) == null)
	    return null;

	Image image = Icon.get(iconName, vpStyle.color, 
			       vpStyle.backgroundColor);

	return new ImageLabelVP(vpStyle, image, view, role);
    }

    // -----------------------------------------------------------------------

    private StyledElementViewPart imageContentToViewPart(StyleValue[] args,
							Style vpStyle,
							StyledElementView view,
							int role) {
	Image image = ImageLabel.DEFAULT_IMAGE;
	int widthSpec = 0; // 0=auto, >0=size, <0=maxSize
	int heightSpec = 0;
	int scaling = Image.SCALE_DEFAULT;

	ElementModel element = view.getElement();

	if (args.length >= 1)
	    image = valueToImage(args[0], element, docView,
				 ImageLabel.DEFAULT_IMAGE);

	int width = image.getWidth(null);
	int height = image.getHeight(null);

	if (args.length >= 2) {
	    FontMetrics metrics = docView.getFontMetrics(vpStyle.font);
	    widthSpec = valueToPixels(args[1], element, 
				      metrics, screenResolution, 0);
	    
	    if (args.length >= 3) {
		heightSpec = valueToPixels(args[2], element, 
					   metrics, screenResolution, 0);

		if (args.length >= 4) {
		    StyleValue arg = args[3];
		    switch (arg.type) {
		    case StyleValue.STRING:
		    case StyleValue.IDENTIFIER:
			if (arg.string.equals("smooth"))
			    scaling = Image.SCALE_SMOOTH;
			break;
		    }
		}
	    }
	}

	int scaledWidth, scaledHeight;
	if (widthSpec < 0 && heightSpec < 0) {
	    widthSpec = -widthSpec;
	    heightSpec = -heightSpec;
	    if (width <= widthSpec) {
		if (height <= heightSpec) {
		    scaledWidth = width;
		    scaledHeight = height;
		} else {
		    scaledHeight = heightSpec;
		    scaledWidth = ((width * scaledHeight) / height);
		}
	    } else {
		if (height <= heightSpec) {
		    scaledWidth = widthSpec;
		    scaledHeight = ((height * scaledWidth) / width);
		} else {
		    double xFactor = (double) widthSpec / (double) width;
		    double yFactor = (double) heightSpec / (double) height;
		    double factor = Math.min(xFactor, yFactor);
		    scaledWidth = (int) Math.rint(width * factor);
		    scaledHeight = (int) Math.rint(height * factor);
		}
	    }
	} else {
	    if (widthSpec < 0) {
		widthSpec = -widthSpec;
		if (width <= widthSpec) {
		    scaledWidth = width;
		    scaledHeight = height;
		} else {
		    scaledWidth = widthSpec;
		    scaledHeight = ((height * scaledWidth) / width);
		}
	    } else if (heightSpec < 0) {
		heightSpec = -heightSpec;
		if (height <= heightSpec) {
		    scaledHeight = height;
		    scaledWidth = width;
		} else {
		    scaledHeight = heightSpec;
		    scaledWidth = ((width * scaledHeight) / height);
		}
	    } else {
		if (widthSpec > 0)
		    scaledWidth = widthSpec;
		else
		    scaledWidth = width;

		if (heightSpec > 0)
		    scaledHeight = heightSpec;
		else
		    scaledHeight = height;
	    }
	}

	if (scaledWidth != width || scaledHeight != height) {
	    image = image.getScaledInstance(scaledWidth, scaledHeight, 
					    scaling);
	    image = ImageResource.loadImage(image);
	    if (image == null)
		image = ImageLabel.DEFAULT_IMAGE;
	}

	return new ImageLabelVP(vpStyle, image, view, role);
    }

    private static Image valueToImage(StyleValue value, 
				      ElementModel element, 
				      DocumentView docView, 
				      Image fallbackImage) {
	if (value.type == StyleValue.ATTRIBUTE) {
	    String attrValue = 
		ElementUtil.getElementAttribute(element, value.string);

	    if (attrValue != null) {
		value = new StyleValue();
		value.type = StyleValue.STRING;
		value.string = attrValue;
	    }
	}

	Toolkit toolkit = docView.getToolkit();
	Image image = null;

	switch (value.type) {
	case StyleValue.URI:
	    {
		try {
		    image = toolkit.createImage(new URL(value.string));
		} catch (MalformedURLException ignored) {}
	    }
	    break;
	case StyleValue.STRING:
	    {
		URL baseURL = 
		    (URL) element.getDocument().getDocumentProperty(
			DocumentModel.URL_PROPERTY);

		if (baseURL != null) {
		    // We are not sure that value.string is an URL.
		    // [1.] May be it's really an URL.

		    try {
			image = toolkit.createImage(
			    new URL(baseURL, value.string));
		    } catch (MalformedURLException ignored) {}

		    // [2.] May be it's a relative file path with the
		    // wrong separator char.

		    if (image == null) {
			try {
			    if (File.separatorChar != '/') {
				String path = value.string.replace(
				    File.separatorChar, '/');
				image = 
				   toolkit.createImage(new URL(baseURL, path));
			    }
			} catch (MalformedURLException ignored) {}
		    }

		    // [3.] May be it's an absolute file path.

		    if (image == null) {
			image = toolkit.createImage(
			    FileUtil.fileToURL(value.string));
		    }
		}		
	    }
	    break;
	}

	if (image != null) 
	    image = ImageResource.loadImage(image);

	if (image == null)
	    image = fallbackImage;

	return image;
    }

    private static int valueToPixels(StyleValue value, 
				     ElementModel element, 
				     FontMetrics metrics, 
				     double screenResolution,
				     int fallbackSize) {
	if (value.type == StyleValue.ATTRIBUTE) {
	    String attrValue = 
		ElementUtil.getElementAttribute(element, value.string);

	    if (attrValue != null) {
		ValueComponent comp = new ValueComponent();
		comp.numericValue = ParseUtil.lexemeToNumber(attrValue);

		if (!Double.isNaN(comp.numericValue)) {
		    comp.type = ValueComponent.NUMBER;
		    comp.unit = ParseUtil.lexemeToUnit(attrValue);

		    value = StyleValue.fromValueComponent(comp);
		}
	    }
	}

	int valueType = value.type;
	if (valueType == StyleValue.NUMBER)
	    valueType = StyleValue.LENGTH_PX;

	if (valueType >= StyleValue.LENGTH_PX && 
	    valueType <= StyleValue.LENGTH_EX)
	    return lengthToPixels(valueType, value.number,
				  metrics, screenResolution);
	else
	    return fallbackSize;
    }

    // -----------------------------------------------------------------------

    private 
    StyledElementViewPart componentContentToViewPart(StyleValue[] args,
						     Style vpStyle,
						     StyledElementView view, 
						     int role) {
	if (args.length < 1 || args[0].type != StyleValue.STRING)
	    return null;
	String className = args[0].string;

	boolean[] stretch = new boolean[] { false, false };
	StyleValue[] params = new StyleValue[args.length-1];
	if (params.length > 0)
	    System.arraycopy(args, 1, params, 0, params.length);

	Component component;
	try {

	    ComponentFactory factory = 
		(ComponentFactory) Class.forName(className).newInstance();
	    component = factory.createComponent(view.getElement(), vpStyle, 
						params, this, stretch);

	} catch (Exception e) {
	    component = null;
	}
	if (component == null)
	    return null;

	ComponentBoxVP componentBox = 
	    new ComponentBoxVP(vpStyle, component, view, role);
	componentBox.setXStretchable(stretch[0]);
	componentBox.setYStretchable(stretch[1]);
	return componentBox;
    }

    // -----------------------------------------------------------------------

    private StyledElementViewPart rowsContentToViewPart(StyleValue[] args,
							Style vpStyle,
							StyledElementView view,
							int role) {
	int rowCount = 0;
	for (int i = 0; i < args.length; ++i) {
	    StyleValue arg = args[i];
	    if (arg.type != StyleValue.FUNCTION || !"row".equals(arg.string))
		break;
	
	    ++rowCount;
	}
	if (rowCount == 0)
	    return null;

	StyleSpec cellStyleSpec = null;
	if (rowCount < args.length)
	    cellStyleSpec = valueListToCellStyleSpec(args, rowCount);

	TableBodyVP table = new TableBodyVP(vpStyle, view, role);

	for (int i = 0; i < rowCount; ++i) {
	    TableBody.Cell[] row = 
		contentToRow(args[i].list, cellStyleSpec, vpStyle, view, role);
	    if (row == null)
		return null;
	
	    table.addRow(row);
	}

	return table;
    }

    private StyleSpec valueListToCellStyleSpec(StyleValue[] list, int offset) {
	StyleSpec styleSpec = new StyleSpec();
	int fieldCount = 0;

	int listSize = 2*(list.length/2);
	for (int i = offset; i < listSize; i += 2) {
	    StyleValue key = list[i];
	    if (key.type != StyleValue.IDENTIFIER)
		break;

	    try {

		Field field = 
		    StyleSpec.class.getField(propToFieldName(key.string));

		if (!Modifier.isStatic(field.getModifiers())) {
		    field.set(styleSpec, list[i+1]);
		    ++fieldCount;
		}

	    } catch (Exception ignored) {
		ignored.printStackTrace();
	    }
	}

	return (fieldCount == 0)? null : styleSpec;
    }

    private static final String propToFieldName(String prop) {
	StringBuffer buffer = new StringBuffer();
	boolean capitalize = false;

	int charCount = prop.length();
	for (int i = 0; i < charCount; ++i) {
	    char c = prop.charAt(i);
	    
	    if (c == '-') {
		capitalize = true;
	    } else {
		if (capitalize) {
		    buffer.append(Character.toUpperCase(c));
		    capitalize = false;
		} else {
		    buffer.append(c);
		}
	    }
	}

	return buffer.toString();
    }

    private StyledElementViewPart rowContentToViewPart(StyleValue[] args,
						       Style vpStyle,
						       StyledElementView view,
						       int role) {
	TableBody.Cell[] row = contentToRow(args, null, vpStyle, view, role);
	if (row == null)
	    return null;

	TableBodyVP table = new TableBodyVP(vpStyle, view, role);
	table.addRow(row);

	return table;
    }

    private TableBody.Cell[] contentToRow(StyleValue[] args, 
					  StyleSpec cellStyleSpec,
					  Style vpStyle,
					  StyledElementView view, 
					  int role) {
	int cellCount = 0;
	for (int i = 0; i < args.length; ++i) {
	    StyleValue arg = args[i];
	    if (arg.type != StyleValue.FUNCTION || !"cell".equals(arg.string))
		break;

	    ++cellCount;
	}
	if (cellCount == 0)
	    return null;

	StyleSpec cellStyleSpec2 = null;
	if (cellCount < args.length)
	    cellStyleSpec2 = valueListToCellStyleSpec(args, cellCount);
	if (cellStyleSpec2 != null) {
	    if (cellStyleSpec == null) {
		cellStyleSpec = cellStyleSpec2;
	    } else {
		cellStyleSpec = cellStyleSpec.duplicate();
		cellStyleSpec.put(cellStyleSpec2);
	    }
	}

	TableBody.Cell[] row = new TableBody.Cell[cellCount];

	for (int i = 0; i < cellCount; ++i) {
	    row[i] = contentToCell(args[i].list, cellStyleSpec, vpStyle, 
				   view, role);
	    if (row[i] == null)
		return null;
	}

	return row;
    }

    private StyledElementViewPart cellContentToViewPart(StyleValue[] args,
							Style vpStyle,
							StyledElementView view,
							int role) {
	TableBody.Cell cell = contentToCell(args, null, vpStyle, view, role);
	if (cell == null)
	    return null;

	TableBodyVP table = new TableBodyVP(vpStyle, view, role);
	table.addRow(new TableBody.Cell[] { cell });

	return table;
    }

    private TableBody.Cell contentToCell(StyleValue[] args, 
					 StyleSpec cellStyleSpec,
					 Style vpStyle,
					 StyledElementView view, 
					 int role) {
	if (args.length < 1)
	    return null;

	StyleSpec cellStyleSpec2 = null;
	if (args.length > 1) 
	    cellStyleSpec2 = valueListToCellStyleSpec(args, 1);
	if (cellStyleSpec2 != null) {
	    if (cellStyleSpec == null) {
		cellStyleSpec = cellStyleSpec2;
	    } else {
		cellStyleSpec = cellStyleSpec.duplicate();
		cellStyleSpec.put(cellStyleSpec2);
	    }
	}

	ElementModel element = view.getElement();
	ViewParameters param = new ViewParameters();
	Style cellStyle = vpStyle;
	if (cellStyleSpec != null)
	    cellStyle = specToStyle(cellStyleSpec, vpStyle, element, param);

	Style noBoxStyle = cellStyle.duplicate();
	noBoxStyle.resetBox();

	String text = null;
	ParagraphVP paragraph = null;
	switch (args[0].type) {
	case StyleValue.STRING:
	    text = args[0].string;
	    break;
	case StyleValue.ATTRIBUTE:
	    text = ElementUtil.getElementAttribute(element, args[0].string);
	    break;
	case StyleValue.FUNCTION:
	    if ("paragraph".equals(args[0].string))
		paragraph = (ParagraphVP) 
		    paragraphContentToViewPart(args[0].list, 
					       shareStyle(noBoxStyle), 
					       view, role);
	    break;
	}
	if (text == null && paragraph == null)
	    return null;

	TableBody.Cell cell = new TableBody.Cell();

	if (paragraph != null)
	    cell.gadget = paragraph;
	else
	    cell.gadget = 
		new TextLabelVP(shareStyle(noBoxStyle), text, view, role);
	cell.style = shareStyle(cellStyle);
	cell.columnSpan = param.columnSpan;
	cell.rowSpan = param.rowSpan;
	cell.minContentWidth = param.width;

	return cell;
    }

    // -----------------------------------------------------------------------

    private 
    StyledElementViewPart paragraphContentToViewPart(StyleValue[] args,
						     Style vpStyle,
						     StyledElementView view,
						     int role) {
	StyleValue argList = new StyleValue();
	argList.type = StyleValue.LIST;
	argList.list = args;

	StyledElementViewPart[] parts = 
	    contentToViewParts(argList, vpStyle, view, role);
	if (parts == null)
	    return null;

	ParagraphVP paragraph = new ParagraphVP(vpStyle, view, role);

	for (int i = 0; i < parts.length; ++i) {
	    Gadget child;

	    // Convert TextLabelVPs that does not support word wrap to
	    // ParagraphTextVPs that support word wrap when added to
	    // ParagraphVPs.
	    //
	    // Note that CounterManager.CounterVP is a subclass of
	    // TextLabel and not of TextLabelVP.

	    if (parts[i] instanceof TextLabelVP) {
		TextLabelVP textLabel = (TextLabelVP) parts[i];

		child = new ParagraphTextVP(textLabel.getStyle(), 
					    textLabel.getText(),
					    view, role);
	    } else {
		child = (Gadget) parts[i];
	    }

	    paragraph.addChild(child);
	}

	return paragraph;
    }
}

