/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.styledview;

import com.xmlmind.gadget.Gadget;
import com.xmlmind.gadget.GadgetContainer;
import com.xmlmind.xmledit.GadgetListImpl;
import com.xmlmind.xmledit.SubelementView;
import com.xmlmind.xmledit.model.SubelementModel;
import com.xmlmind.xmledit.model.ElementModel;
import com.xmlmind.xmledit.stylesheet.StyleValue;
import com.xmlmind.xmledit.styledgadget.Style;
import com.xmlmind.xmledit.styledgadget.TableBody;
import com.xmlmind.xmledit.styledgadget.ParagraphText;
import com.xmlmind.xmledit.styledgadget.Box;

public class TableView extends ContainerView {
    protected static final byte BOTTOM_CAPTION = (byte) (INLINE_VIEW << 1);

    private Gadget viewGadget;
    private int startLabelRowCount;

    // -----------------------------------------------------------------------

    public TableView(ElementModel element, Style style) {
	super(element, style);
    }

    public void initView(ViewParameters param,
			 ViewParameters paramBefore,
			 Style styleBefore,
			 ViewParameters paramAfter, 
			 Style styleAfter,
			 StyledViewFactory viewFactory) {

	boolean isCustom = (param.content != null);
	if (isCustom)
	    flags |= CUSTOM_VIEW;
	if (param.display == StyleValue.INLINE_TABLE)
	    flags |= INLINE_VIEW;

	TableBodyVP table = 
	    new TableBodyVP(style, this, 
			    (isCustom? StyledElementViewPart.ROLE_CUSTOM_VIEW :
			     StyledElementViewPart.ROLE_STANDARD_VIEW));

	table.setHorizontalCellSpacing(param.borderHorizontalSpacing);
	table.setVerticalCellSpacing(param.borderVerticalSpacing);
	if (param.captionSide == StyleValue.BOTTOM)
	    flags |= BOTTOM_CAPTION;

	startLabelRowCount = 0;
	if (paramBefore != null && paramBefore.content != null) {
	    switch (paramBefore.display) {
	    case StyleValue.TABLE_ROW:
	    case StyleValue.TABLE_ROW_GROUP:
		startLabelRowCount = 
		    addRows(
			contentToGadget(paramBefore.content, styleBefore,
					StyledElementViewPart.ROLE_START_LABEL,
					/*noWrap*/ false, viewFactory),
			table, viewFactory);
		paramBefore.content = null;
		break;
	    }
	}

	if (isCustom) {
	    addRows(
		contentToGadget(param.content, style, 
				StyledElementViewPart.ROLE_CUSTOM_VIEW,
                                /*noWrap*/ false, viewFactory),
		table, viewFactory);
	    param.content = null;
	}

	if (paramAfter != null && paramAfter.content != null) {
	    switch (paramAfter.display) {
	    case StyleValue.TABLE_ROW:
	    case StyleValue.TABLE_ROW_GROUP:
		addRows(
		    contentToGadget(paramAfter.content, styleAfter,
				    StyledElementViewPart.ROLE_END_LABEL,
				    /*noWrap*/ false, viewFactory), 
		    table, viewFactory);
		paramAfter.content = null;
		break;
	    }
	}

	if (isCustom && table.getChildCount() == 1) {
	    // Optimize the case where the table only contains a single
	    // custom component.

	    Gadget child = table.getChild(0);
	    ((StyledElementViewPart) child).setStyle(style);

	    viewGadget = child;
	} else {
	    viewGadget = table;
	}

	super.initView(param, paramBefore, styleBefore, paramAfter, styleAfter,
		       viewFactory);
    }

    private static final int addRows(Gadget gadget, TableBody table,
				     StyledViewFactory viewFactory) {
	int rowCount;

	if (gadget instanceof TableBody) {
	    TableBody rows = (TableBody) gadget;
	    rowCount = rows.getRowCount();

	    for (int i = 0; i < rowCount; ++i)
		table.addRow(rows.getRow(i));
	} else {
	    table.addRow(new TableBody.Cell[] { 
		gadgetToCell(gadget, 1, 1, 0, viewFactory) 
	    });

	    rowCount = 1;
	}

	return rowCount;
    }

    private static final TableBody.Cell gadgetToCell(
					    Gadget gadget, 
					    int columnSpan, 
					    int rowSpan, 
					    int width, 
					    StyledViewFactory viewFactory) {
	TableBody.Cell cell = new TableBody.Cell();
	cell.gadget = gadget;	
	cell.columnSpan = columnSpan;
	cell.rowSpan = rowSpan;
	cell.minContentWidth = width;

	// The box, if any, is now around the cell and no longer around the
	// gadget.

	StyledElementViewPart part = (StyledElementViewPart) gadget;
	cell.style = part.getStyle();

	Style noBoxStyle = cell.style.duplicate();
	noBoxStyle.resetBox();
	part.setStyle(viewFactory.shareStyle(noBoxStyle));

	return cell;
    }

    protected Gadget getViewGadget() {
	return viewGadget;
    }

    public Gadget addChildViewGadgets(SubelementView childView, 
				      int viewIndex) {
	if ((childView instanceof CaptionView ||
	     childView instanceof ColumnView ||
	     childView instanceof ColumnGroupView) &&
	    childView.getModel().getParentElement() == element)
	    return addMetaInfo(childView, viewIndex);
	else
	    return addCells(childView, viewIndex);
    }

    public Gadget removeChildViewGadgets(SubelementView childView, 
					 int viewIndex) {
	if (childView instanceof CaptionView ||
	    childView instanceof ColumnView ||
	    childView instanceof ColumnGroupView)
	    return removeMetaInfo(childView, viewIndex);
	else
	    return removeCells(childView, viewIndex);
    }

    // -----------------------------------------------------------------------

    private Gadget addMetaInfo(SubelementView childView, int viewIndex) {
	GadgetContainer viewContainerGadget;
	GadgetContainer refreshed;

	if (viewGadget instanceof TableBodyVP) {
	    // Wrap table inside a Division.
	    TableBodyVP table = (TableBodyVP) viewGadget;

	    // Move marginTop, marginBottom values from table to div ---

	    Style anonymousStyle = new Style();
	    Style tableStyle = table.getStyle();
	    anonymousStyle.marginTop = tableStyle.marginTop;
	    anonymousStyle.marginBottom = tableStyle.marginBottom;

	    StyledViewFactory viewFactory = getViewFactory(viewIndex);

	    Style noMarginStyle = tableStyle.duplicate();
	    noMarginStyle.marginTop = noMarginStyle.marginBottom = 0;
	    table.setStyle(viewFactory.shareStyle(noMarginStyle));

	    DivisionVP div = 
		new DivisionVP(viewFactory.shareStyle(anonymousStyle), this, 
			       StyledElementViewPart.ROLE_STANDARD_VIEW);

	    refreshed = table.parent;
	    if (refreshed != null)
		refreshed.replaceChild(table, div);
	    // Otherwise, happens when the top TableView is not yet attached
	    // to the DocumentView.

	    div.addChild(table);
	    viewContainerGadget = div;

	    viewGadget = div;
	} else {
	    refreshed = viewContainerGadget = (GadgetContainer) viewGadget;
	}

	// Insert the meta-info gadget in div ---

	int insertIndex = 0;
	SubelementModel childModel = childView.getModel();
	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    if (sub == childModel)
		break;
	    ++insertIndex;
	    sub = sub.getNextSubelement();
	}
	int endIndex = viewContainerGadget.getChildCount();
	if (insertIndex > endIndex)
	    insertIndex = endIndex;

	Gadget gadget;
	if (childView instanceof CaptionView) {
	    gadget = ((CaptionView) childView).getGadget();

	    if ((flags & BOTTOM_CAPTION) != 0)
		insertIndex = endIndex;
	} else {
	    // A ColumnGroupView or a ColumnView is a StyledTreeView.
	    gadget = (Gadget) childView;
	}

	viewContainerGadget.insertChild(insertIndex, gadget);
	    
	return refreshed;
    }

    private Gadget removeMetaInfo(SubelementView childView, int viewIndex) {
	Gadget gadget;
	if (childView instanceof CaptionView)
	    gadget = ((CaptionView) childView).getGadget();
	else
	    // A ColumnGroupView or a ColumnView is a StyledTreeView.
	    gadget = (Gadget) childView;

	if (!(viewGadget instanceof DivisionVP) || gadget.parent != viewGadget)
	    return removeCells(childView, viewIndex);

	DivisionVP div = (DivisionVP) viewGadget;
	div.removeChild(gadget);

	if (div.getChildCount() != 1)
	    return div;

	// If the viewGadget gadget is a DivisionVP containing the table, 
	// let the table be the viewGadget, once again.

	TableBodyVP table = (TableBodyVP) div.getChild(0);

	Style marginStyle = table.getStyle().duplicate();
	Style divStyle = div.getStyle();
	marginStyle.marginTop = divStyle.marginTop;
	marginStyle.marginBottom = divStyle.marginBottom;
	table.setStyle(getViewFactory(viewIndex).shareStyle(marginStyle));

	// Should work fine with a LabeledItem too.
	GadgetContainer parent = div.parent;
	parent.replaceChild(div, table);

	viewGadget = table;

	return parent;
    }

    // -----------------------------------------------------------------------

    private Gadget addCells(SubelementView childView, int viewIndex) {
	GadgetListImpl list = new GadgetListImpl();
	int[] location = new int[2];
	int rowIndex = startLabelRowCount;
	int cellIndex = -1;

	SubelementModel sub = element.getFirstSubelement();
	while (sub != null) {
	    SubelementView subView = 
		(SubelementView) sub.getDocumentListenerData(viewIndex);

	    if (subView != null && 
		!(subView instanceof CaptionView ||
		  subView instanceof ColumnView ||
		  subView instanceof ColumnGroupView)) {
		if (findInRows(childView, subView, rowIndex, location, 
			       viewIndex, list)) {
		    rowIndex = location[0];
		    cellIndex = location[1];
		    break;
		}

		rowIndex = location[0];
	    }
	    
	    sub = sub.getNextSubelement();
	}

	TableBodyVP table = getTable();
	StyledViewFactory viewFactory = getViewFactory(viewIndex);

	if (cellIndex < 0)
	    insertRows(table, rowIndex, childView, viewIndex, viewFactory, 
		       list);
	else
	    insertCells(table, rowIndex, cellIndex, childView, viewIndex, 
			viewFactory, list);

	return table;
    }

    private static boolean findInRows(SubelementView searchedView, 
				      SubelementView fromView, int rowIndex,  
				      int[] location, int viewIndex, 
				      GadgetListImpl list) {
	if (fromView == searchedView) {
	    location[0] = rowIndex;
	    location[1] = -1;
	    return true;
	}

	if (fromView instanceof GroupView) {
	    if (fromView instanceof RowView) {
		if (findInRow(searchedView, (RowView) fromView, rowIndex, 
			      location, viewIndex, list))
		    return true;
		++rowIndex;
	    } else {
		GroupView groupView = (GroupView) fromView;

		list.removeAll();
		groupView.getStartLabelGadgets(list);
		rowIndex += list.size;

		list.removeAll();
		groupView.getCustomViewGadgets(list);
		if (list.size > 0) {
		    rowIndex += list.size;
		} else {
		    SubelementModel sub = 
			 groupView.getElement().getFirstSubelement();
		    while (sub != null) {
			SubelementView subView = (SubelementView) 
			    sub.getDocumentListenerData(viewIndex);

			if (subView != null) {
			    if (findInRows(searchedView, subView, rowIndex, 
					   location, viewIndex, list))
				return true;
			    rowIndex = location[0];
			}

			sub = sub.getNextSubelement();
		    }
		}
		
		list.removeAll();
		groupView.getEndLabelGadgets(list);
		rowIndex += list.size;
	    }
	} else {
	    if (fromView instanceof CellView) {
		// Little optimization.
		++rowIndex;
	    } else {
		list.removeAll();
		fromView.getGadgets(list, viewIndex);
		rowIndex += list.size;
	    }
	}

	location[0] = rowIndex;
	location[1] = -1;
	return false;
    }

    private static boolean findInRow(SubelementView searchedView, 
				     RowView rowView, int rowIndex,  
				     int[] location, int viewIndex, 
				     GadgetListImpl list) {
	if (rowView == searchedView) {
	    location[0] = rowIndex;
	    location[1] = -1;
	    return true;
	}

	if (rowView.getCustomViewGadget() == null) {
	    boolean found = false;
	    int cellIndex = (rowView.getStartLabelGadget() == null)? 0 : 1;

	    SubelementModel sub = rowView.getElement().getFirstSubelement();
	    while (sub != null) {
		SubelementView subView = 
		    (SubelementView) sub.getDocumentListenerData(viewIndex);

		if (subView != null) {
		    if (findInCells(searchedView, subView, rowIndex, cellIndex,
				    location, viewIndex, list)) {
			if (cellIndex > 0)
			    return true;
			else
			    // Found but continue in order to count the cells
			    // in this RowView
			    found = true;
		    }

		    cellIndex = location[1];
		}

		sub = sub.getNextSubelement();
	    }

	    if (rowView.getEndLabelGadget() != null)
		++cellIndex;

	    if (found) {
		location[0] = rowIndex;
		if (cellIndex > 0)
		    // The row has cells: it does not need to be recreated.
		    location[1] = 0;
		else
		    // The row needs to be recreated.
		    location[1] = -1;
		return true;
	    }
	}

	location[0] = rowIndex+1;
	location[1] = -1;
	return false;
    }

    private static boolean findInCells(SubelementView searchedView, 
				       SubelementView fromView, 
				       int rowIndex, int cellIndex,  
				       int[] location, int viewIndex, 
				       GadgetListImpl list) {
	if (fromView == searchedView) {
	    location[0] = rowIndex;
	    location[1] = cellIndex;
	    return true;
	}

	if (fromView instanceof GroupView) {
	    GroupView groupView = (GroupView) fromView;

	    list.removeAll();
	    groupView.getStartLabelGadgets(list);
	    cellIndex += list.size;

	    list.removeAll();
	    groupView.getCustomViewGadgets(list);
	    if (list.size > 0) {
		cellIndex += list.size;
	    } else {
		SubelementModel sub = 
		     groupView.getElement().getFirstSubelement();
		while (sub != null) {
		    SubelementView subView = (SubelementView) 
			sub.getDocumentListenerData(viewIndex);

		    if (subView != null) {
			if (findInCells(searchedView, subView, rowIndex, 
					cellIndex, location, viewIndex, list))
			    return true;
			cellIndex = location[1];
		    }

		    sub = sub.getNextSubelement();
		}
	    }

	    list.removeAll();
	    groupView.getEndLabelGadgets(list);
	    cellIndex += list.size;
	} else {
	    if (fromView instanceof CellView) {
		// Little optimization.
		++cellIndex;
	    } else {
		list.removeAll();
		fromView.getGadgets(list, viewIndex);
		cellIndex += list.size;
	    }
	}

	location[0] = rowIndex;
	location[1] = cellIndex;
	return false;
    }

    private static int insertRows(TableBodyVP table, int rowIndex, 
				  SubelementView childView, int viewIndex, 
				  StyledViewFactory viewFactory,
				  GadgetListImpl list) {
	if (childView instanceof GroupView) {
	    if (childView instanceof RowView) {
		insertRow(table, rowIndex, (RowView) childView, viewIndex,
			  viewFactory, list);
		++rowIndex;
	    } else {
		GroupView groupView = (GroupView) childView;

		list.removeAll();
		groupView.getStartLabelGadgets(list);
		for (int i = 0; i < list.size; ++i) {
		    table.insertRow(rowIndex);
		    insertCell(table, rowIndex, 0, list.list[i], viewFactory);
		    ++rowIndex;
		}

		list.removeAll();
		groupView.getCustomViewGadgets(list);
		if (list.size > 0) {
		    for (int i = 0; i < list.size; ++i) {
			table.insertRow(rowIndex);
			insertCell(table, rowIndex, 0, list.list[i], 
				   viewFactory);
			++rowIndex;
		    }
		} else {
		    SubelementModel sub = 
			 groupView.getElement().getFirstSubelement();
		    while (sub != null) {
			SubelementView subView = (SubelementView) 
			    sub.getDocumentListenerData(viewIndex);

			if (subView != null)
			    rowIndex = 
				insertRows(table, rowIndex, subView, 
					   viewIndex, viewFactory, list);

			sub = sub.getNextSubelement();
		    }
		}
		
		list.removeAll();
		groupView.getEndLabelGadgets(list);
		for (int i = 0; i < list.size; ++i) {
		    table.insertRow(rowIndex);
		    insertCell(table, rowIndex, 0, list.list[i], viewFactory);
		    ++rowIndex;
		}
	    }
	} else {
	    if (childView instanceof CellView) {
		table.insertRow(rowIndex);
		insertCell(table, rowIndex, 0, (CellView) childView, 
			   viewFactory);
		++rowIndex;
	    } else {
		list.removeAll();
		childView.getGadgets(list, viewIndex);
		for (int i = 0; i < list.size; ++i) {
		    table.insertRow(rowIndex);
		    insertCell(table, rowIndex, 0, list.list[i], viewFactory);
		    ++rowIndex;
		}
	    }
	}

	return rowIndex;
    }

    private static void insertRow(TableBodyVP table, int rowIndex, 
				  RowView rowView, int viewIndex, 
				  StyledViewFactory viewFactory,
				  GadgetListImpl list) {
	table.insertRow(rowIndex);

	int cellIndex = 0;

	Gadget gadget = rowView.getStartLabelGadget();
	if (gadget != null) {
	    insertCell(table, rowIndex, cellIndex, gadget, viewFactory);
	    ++cellIndex;
	}

	gadget = rowView.getCustomViewGadget();
	if (gadget != null) {
	    insertCell(table, rowIndex, cellIndex, gadget, viewFactory);
	    ++cellIndex;
	} else {
	    SubelementModel sub = rowView.getElement().getFirstSubelement();
	    while (sub != null) {
		SubelementView subView = 
		    (SubelementView) sub.getDocumentListenerData(viewIndex);

		if (subView != null)
		    cellIndex = 
			insertCells(table, rowIndex, cellIndex,
				    subView, viewIndex, viewFactory, list);

		sub = sub.getNextSubelement();
	    }
	}

	gadget = rowView.getEndLabelGadget();
	if (gadget != null) {
	    insertCell(table, rowIndex, cellIndex, gadget, viewFactory);
	    ++cellIndex;
	}
    }

    private static int insertCells(TableBodyVP table, 
				   int rowIndex, int cellIndex,
				   SubelementView childView, int viewIndex, 
				   StyledViewFactory viewFactory,
				   GadgetListImpl list) {
	if (childView instanceof GroupView) {
	    GroupView groupView = (GroupView) childView;

	    list.removeAll();
	    groupView.getStartLabelGadgets(list);
	    for (int i = 0; i < list.size; ++i) {
		insertCell(table, rowIndex, cellIndex, list.list[i], 
			   viewFactory);
		++cellIndex;
	    }

	    list.removeAll();
	    groupView.getCustomViewGadgets(list);
	    if (list.size > 0) {
		for (int i = 0; i < list.size; ++i) {
		    insertCell(table, rowIndex, cellIndex, list.list[i], 
			       viewFactory);
		    ++cellIndex;
		}
	    } else {
		SubelementModel sub = 
		     groupView.getElement().getFirstSubelement();
		while (sub != null) {
		    SubelementView subView = (SubelementView) 
			sub.getDocumentListenerData(viewIndex);

		    if (subView != null)
			cellIndex = 
			    insertCells(table, rowIndex, cellIndex, subView, 
					viewIndex, viewFactory, list);

		    sub = sub.getNextSubelement();
		}
	    }

	    list.removeAll();
	    groupView.getEndLabelGadgets(list);
	    for (int i = 0; i < list.size; ++i) {
		insertCell(table, rowIndex, cellIndex, list.list[i], 
			   viewFactory);
		++cellIndex;
	    }
	} else {
	    if (childView instanceof CellView) {
		insertCell(table, rowIndex, cellIndex, (CellView) childView,
			   viewFactory);
		++cellIndex;
	    } else {
		list.removeAll();
		childView.getGadgets(list, viewIndex);
		for (int i = 0; i < list.size; ++i) {
		    insertCell(table, rowIndex, cellIndex, list.list[i],
			       viewFactory);
		    ++cellIndex;
		}
	    }
	}

	return cellIndex;
    }

    private static void insertCell(TableBodyVP table, int rowIndex, 
				   int cellIndex, CellView cellView, 
				   StyledViewFactory viewFactory) {
	TableBody.Cell cell = 
	    gadgetToCell(cellView.getGadget(), cellView.getColumnSpan(),
			 cellView.getRowSpan(), cellView.getMinContentWidth(),
			 viewFactory);

	table.insertCell(rowIndex, cellIndex, cell);
    }

    private static void insertCell(TableBodyVP table, int rowIndex, 
				   int cellIndex, Gadget gadget, 
				   StyledViewFactory viewFactory) {
	if (gadget instanceof ParagraphText) {
	    ParagraphVP paragraph = 
		new ParagraphVP(Box.DEFAULT_STYLE, 
				table.getStyledElementView(), 
				StyledElementViewPart.ROLE_STANDARD_VIEW);
	    paragraph.addChild(gadget);
	    gadget = paragraph;
	}

	TableBody.Cell cell = gadgetToCell(gadget, 1, 1, 0, viewFactory);
	table.insertCell(rowIndex, cellIndex, cell);
    }

    private Gadget removeCells(SubelementView childView, int viewIndex) {
	GadgetListImpl list = new GadgetListImpl();
	childView.getGadgets(list, viewIndex);
	for (int i = 0; i < list.size; ++i) {
	    Gadget gadget = list.list[i];
	    gadget.parent.removeChild(gadget);
	}

	TableBodyVP table = getTable();

	sweepEmptyRows(table);

	return table;
    }

    private TableBodyVP getTable() {
	if (viewGadget instanceof TableBodyVP) {
	    return (TableBodyVP) viewGadget;
	} else {
	    DivisionVP div = (DivisionVP) viewGadget;

	    Gadget[] children = div.getChildren();
	    for (int i = 0; i < children.length; ++i) {
		Gadget child = children[i];
		TableBodyVP table;

		if (child instanceof TableBodyVP &&
		    (table = (TableBodyVP) child).getElementView() == this)
		    return table;
	    }

	    // Should never happen, this view always contains a table even if
	    // it is empty.
	    return null;
	}
    }

    private void sweepEmptyRows(TableBodyVP table) {
	for (int i = table.getRowCount()-1; i >= 0; --i) {
	    TableBody.Cell[] row = table.getRow(i);

	    // Remove ``anonymous cells'' ---

	    for (int j = row.length-1; j >= 0; --j) {
		Gadget gadget = row[j].gadget;

		if (gadget instanceof StyledElementViewPart &&
		    ((StyledElementViewPart)gadget).getElementView() == this &&
		    gadget instanceof GadgetContainer &&
		    ((GadgetContainer) gadget).getChildCount() == 0)
		    table.removeCell(i, j);
	    }

	    // Remove empty row ---

	    if (table.getCellCount(i) == 0) 
		table.removeRow(i);
	}

	// Note that the table, even empty, is never swept.
    }
}
