/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.stylesheet;

import java.util.StringTokenizer;
import fr.pixware.util.QuickSort;
import fr.pixware.util.StringUtil;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.ElementUtil;

public class Selector {
    public static final int SUBJECT_CURRENT     = 0;
    public static final int SUBJECT_PARENT      = 1;
    public static final int SUBJECT_PREDECESSOR = 2;
    public static final int SUBJECT_ANCESTOR    = 3;

    public static final int TEST_ELEMENT_TYPE_NAME_IS = 0;
    public static final int TEST_IS_FIRST_CHILD       = 1;
    public static final int TEST_HAS_ATTRIBUTE        = 2;
    public static final int TEST_ATTRIBUTE_EQUALS     = 3;
    public static final int TEST_ATTRIBUTE_CONTAINS   = 4;
    public static final int TEST_ATTRIBUTE_DASH_MATCH = 5;
    public static final int TEST_LANGUAGE_IS          = 6;

    public static final int PSEUDO_ELEMENT_NONE   = 0;
    public static final int PSEUDO_ELEMENT_BEFORE = 1;
    public static final int PSEUDO_ELEMENT_AFTER  = 2;

    // .foo <=> {TEST_ATTRIBUTE_CONTAINS, "class", "foo"}
    // #bar <=> {TEST_ATTRIBUTE_EQUALS, "id", "bar"}
    public static class Test {
	public int type;
	public String name;
	public String value;
    }

    public static class Matcher {
	public int subject;
	public Test[] tests;
    }

    public static class TestCompare implements QuickSort.Compare {
	public int compare(Object o1, Object o2) {
	    Test t1 = (Test) o1;
	    Test t2 = (Test) o2;
	    
	    if (t1.type == t2.type) {
		switch (t1.type) {
		case TEST_ELEMENT_TYPE_NAME_IS:
		case TEST_HAS_ATTRIBUTE:
		    return (t1.name.compareTo(t2.name));
		case TEST_ATTRIBUTE_EQUALS:
		case TEST_ATTRIBUTE_CONTAINS:
		case TEST_ATTRIBUTE_DASH_MATCH:
		    {
			int delta = (t1.name.compareTo(t2.name));
			if (delta == 0) 
			    return (t1.value.compareTo(t2.value));
			else
			    return delta;
		    }
		case TEST_LANGUAGE_IS:
		    return (t1.value.compareTo(t2.value));
		case TEST_IS_FIRST_CHILD:
		default:
		    return 0;
		}
	    } else {
		return t1.type - t2.type;
	    }
	}
    }
    public static TestCompare testCompare = new TestCompare();

    private Matcher[] matchers;
    private int pseudoElement;
    private int specificity = -1;

    // -----------------------------------------------------------------------

    public Selector(Matcher[] matchers, int pseudoElement) {	
	this.matchers = matchers; 
	this.pseudoElement = pseudoElement;

	sortTests();
	computeSpecificity();
    }

    private void sortTests() {
	for (int i = 0; i < matchers.length; ++i) {
	    Matcher matcher = matchers[i];

	    QuickSort.sort(matcher.tests, testCompare);
	}
    }

    private void computeSpecificity() {
	int a = 0;
	int b = 0;
	int c = 0;

	for (int i = 0; i < matchers.length; ++i) {
	    Matcher matcher = matchers[i];
	    int testCount = matcher.tests.length;

	    for (int j = 0; j < testCount; ++j) {
		Test test = matcher.tests[j];

		switch (test.type) {
		case TEST_ELEMENT_TYPE_NAME_IS:
		    ++c;
		    break;
		case TEST_HAS_ATTRIBUTE:
		case TEST_ATTRIBUTE_CONTAINS:
		case TEST_ATTRIBUTE_DASH_MATCH:
		case TEST_LANGUAGE_IS:
		case TEST_IS_FIRST_CHILD:
		    ++b;
		    break;
		case TEST_ATTRIBUTE_EQUALS:
		    if (test.name.equals("id") || test.name.equals("ID"))
			++a;
		    else
			++b;
		    break;
		}
	    }
	}

	specificity = a*512*512 + b*512 + c;
    }

    public Matcher[] getMatchers() {
	return matchers;
    }

    public int getPseudoElement() {
	return pseudoElement;
    }

    public int getSpecificity() {
	return specificity;
    }

    public boolean match(ElementModel element) {
	ElementModel subject = element;

	for (int i = 0; i < matchers.length; ++i) {
	    Matcher matcher = matchers[i];

	    switch (matcher.subject) {
	    case SUBJECT_CURRENT:
		if (!test(subject, matcher.tests))
		    return false;
		break;

	    case SUBJECT_PARENT:
		subject = subject.getParentElement();
		if (subject == null || !test(subject, matcher.tests))
		    return false;
		break;

	    case SUBJECT_PREDECESSOR:
	        {
		    SubelementModel sub = subject.getPreviousSubelement();
		    if (sub == null || 
			sub.getSubelementType() == SubelementModel.TEXT)
			return false;

		    subject = (ElementModel) sub;
		    if (!test(subject, matcher.tests))
			return false;
	        }
		break;

	    case SUBJECT_ANCESTOR:
		for (;;) {
		    subject = subject.getParentElement();
		    if (subject == null)
			return false;

		    if (test(subject, matcher.tests))
			break;
		}
		break;
	    }
	}

	return true;
    }

    private static final boolean test(ElementModel element, Test[] tests) {
	int testCount = tests.length;
	for (int j = 0; j < testCount; ++j) {
	    Test test = tests[j];

	    switch (test.type) {
	    case TEST_ELEMENT_TYPE_NAME_IS:
		if (!element.getElementTypeName().equals(test.name))
		    return false;
		break;
	    case TEST_IS_FIRST_CHILD:
		if (element.getPreviousSubelement() != null)
		    return false;
		break;
	    case TEST_HAS_ATTRIBUTE:
		if (element.getElementAttribute(test.name) == null)
		    return false;
		break;
	    case TEST_ATTRIBUTE_EQUALS:
	        {
		    String attrVal = element.getElementAttribute(test.name);
		    if (attrVal == null || !attrVal.equals(test.value))
			return false;
	        }
		break;
	    case TEST_ATTRIBUTE_CONTAINS:
	        {
		    String attrVal = element.getElementAttribute(test.name);
		    if (attrVal == null || attrVal.indexOf(test.value) < 0)
			return false;

		    StringTokenizer words = 
			new StringTokenizer(attrVal, " \n\r\t");
		    boolean found = false;
		    while (words.hasMoreTokens()) {
			if (words.nextToken().equals(test.value)) {
			    found = true;
			    break;
			}
		    }
		    if (!found)
			return false;
	        }
		break;
	    case TEST_ATTRIBUTE_DASH_MATCH:
	        {
		    String attrVal = element.getElementAttribute(test.name);
		    if (attrVal == null || attrVal.indexOf(test.value) < 0)
			return false;

		    String[] words = StringUtil.split(attrVal, '-');
		    boolean found = false;
		    for (int i = 0; i < words.length; ++i) {
			if (words[i].equals(test.value)) {
			    found = true;
			    break;
			}
		    }
		    if (!found)
			return false;
	        }
		break;
	    case TEST_LANGUAGE_IS:
	        {
		    String lang = ElementUtil.getLanguage(element);
		    if (lang == null || !lang.equals(test.value))
			return false;
		}
		break;
	    }
	}

	return true;
    }

    public boolean equals(Object other) {
	if (other == null || !(other instanceof Selector))
	    return false;

	Selector s = (Selector) other;
	if (s.matchers.length != matchers.length ||
	    s.pseudoElement != pseudoElement)
	    return false;

	for (int i = 0; i < matchers.length; ++i) {
	    Matcher matcher = matchers[i];
	    Matcher matcher2 = s.matchers[i];
	    int testCount;

	    if (matcher2.subject != matcher.subject ||
		matcher2.tests.length != (testCount = matcher.tests.length))
		return false;

	    for (int j = 0; j < testCount; ++j) {
		Test test = matcher.tests[j];
		Test test2 = matcher2.tests[j];

		if (test.type != test2.type ||
		    ((test.name == null && test2.name != null) ||
		     (test.name != null && !test.name.equals(test2.name)))  ||
		    ((test.value == null && test2.value != null) ||
		     (test.value != null && !test.value.equals(test2.value))))
		    return false;
	    }
	}

	return true;
    }

    public int hashCode() {
	return pseudoElement*512*512*512 + specificity;
    }
}


