/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.stylesheet;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.net.URL;
import fr.pixware.util.QuickSort;
import com.xmlmind.xmledit.model.*;

public class StyleSheet {
    public static final int ELEMENT        = 0;
    public static final int BEFORE_ELEMENT = 1;
    public static final int AFTER_ELEMENT  = 2;

    public static final int ELEMENT_MASK        = (1 << ELEMENT);
    public static final int BEFORE_ELEMENT_MASK = (1 << BEFORE_ELEMENT);
    public static final int AFTER_ELEMENT_MASK  = (1 << AFTER_ELEMENT);

    private URL url;
    private String[] defaultMedia;
    private String[] otherMedia;

    private int nextRuleId = 0;
    private Hashtable ruleSet = new Hashtable();
    private Rule[] rules = null;
    private Hashtable elementTypeNameToRules = null;

    public static class RuleCompare implements QuickSort.Compare {
	public int compare(Object o1, Object o2) {
	    Rule r1 = (Rule) o1;
	    Rule r2 = (Rule) o2;
	    int s1 = r1.getSelector().getSpecificity();
	    int s2 = r2.getSelector().getSpecificity();

	    if (s1 == s2)
		return (r1.getId() - r2.getId());
	    else
		return (s1 - s2);
	}
    }
    public static RuleCompare ruleCompare = new RuleCompare();

    // -----------------------------------------------------------------------

    public void setURL(URL url) {
	this.url = url;
    }

    public URL getURL() {
	return url;
    }

    public void setDefaultMedia(String[] media) {
	defaultMedia = media;
    }

    public String[] getDefaultMedia() {
	return defaultMedia;
    }

    public void setOtherMedia(String[] media) {
	otherMedia = media;
    }

    public String[] getOtherMedia() {
	return otherMedia;
    }

    public void addRule(Selector selector, StyleSpec styleSpec) {
	if (ruleSet == null)
	    addRules();

	Rule rule = (Rule) ruleSet.get(selector);
	if (rule == null) {
	    rule = new Rule(nextRuleId++, selector, styleSpec);
	    ruleSet.put(selector, rule);
	} else {
	    // Same selector, add/replace more style properties.
	    rule.getStyleSpec().put(styleSpec);
	}
    }

    private void addRules() {
	for (int i = 0; i < rules.length; ++i) {
	    Rule rule = rules[i];
	    ruleSet.put(rule.getSelector(), rule);
	}
	
	rules = null;
	elementTypeNameToRules = null;
    }

    public Rule[] getRules() {
	if (rules == null) 
	    sortRules();

	return rules;
    }

    private void sortRules() {
	// Build rules ---

	rules = new Rule[ruleSet.size()];
	Enumeration iter = ruleSet.elements();
	int i = 0;
	while (iter.hasMoreElements()) 
	    rules[i++] = (Rule) iter.nextElement();

	QuickSort.sort(rules, ruleCompare);

	ruleSet = null;

	// Build elementTypeNameToRules ---

	Hashtable map = new Hashtable();

	for (i = 0; i < rules.length; ++i) {
	    Rule rule = rules[i];
	    Selector selector = rule.getSelector();
	    // Selector.Tests are sorted in order to have
	    // TEST_ELEMENT_TYPE_NAME_IS as the first one.
	    Selector.Matcher matcher = (selector.getMatchers())[0];
	    String key;

	    if (matcher.tests.length > 0 &&
		matcher.tests[0].type == Selector.TEST_ELEMENT_TYPE_NAME_IS)
		key = matcher.tests[0].name;
	    else
		key = "*";

	    Vector[] lists = (Vector[]) map.get(key);
	    if (lists == null) {
		lists = new Vector[3];
		for (int j = 0; j < lists.length; ++j)
		    lists[j] = new Vector();
		map.put(key, lists);
	    }

	    switch (selector.getPseudoElement()) {
	    case Selector.PSEUDO_ELEMENT_NONE:
		lists[ELEMENT].addElement(rule);
		break;
	    case Selector.PSEUDO_ELEMENT_BEFORE:
		lists[BEFORE_ELEMENT].addElement(rule);
		break;
	    case Selector.PSEUDO_ELEMENT_AFTER:
		lists[AFTER_ELEMENT].addElement(rule);
		break;
	    }
	}

	// Convert map's vectors to elementTypeNameToRules'arrays.

	Vector[] anyLists = (Vector[]) map.get("*");
	if (anyLists != null)
	    map.remove("*");

	elementTypeNameToRules = new Hashtable();

	iter = map.keys();
	while (iter.hasMoreElements()) {
	    String key = (String) iter.nextElement();
	    Vector[] lists = (Vector[]) map.get(key);

	    Rule[][] ruleLists = new Rule[lists.length][];

	    for (int j = 0; j < lists.length; ++j) {
		if (anyLists != null) 
		    addElements(lists[j], anyLists[j]);
	    
		Rule[] ruleList = new Rule[lists[j].size()];
		lists[j].copyInto(ruleList);
		QuickSort.sort(ruleList, ruleCompare);

		ruleLists[j] = ruleList;
	    }

	    elementTypeNameToRules.put(key, ruleLists);
	}

	if (anyLists != null) {
	    Rule[][] ruleLists = new Rule[anyLists.length][];

	    for (int j = 0; j < anyLists.length; ++j) {
		Rule[] ruleList = new Rule[anyLists[j].size()];
		anyLists[j].copyInto(ruleList);
		// Already sorted.

		ruleLists[j] = ruleList;
	    }

	    elementTypeNameToRules.put("*", ruleLists);
	}
    }

    private static final void addElements(Vector list, Vector elements) {
	int count = elements.size();
	for (int i = 0; i < count; ++i)
	    list.addElement(elements.elementAt(i));
    }

    public Rule[][] findRules(String elementTypeName) {
	if (rules == null) 
	    sortRules();

	Rule[][] ruleLists = 
	    (Rule[][]) elementTypeNameToRules.get(elementTypeName);
	if (ruleLists == null)
	    ruleLists = (Rule[][]) elementTypeNameToRules.get("*");

	return ruleLists;
    }

    public int findStyleSpec(ElementModel element, StyleSpec[] specs) {
	Rule[][] ruleLists = findRules(element.getElementTypeName());
	if (ruleLists == null)
	    return 0x0;

	int found = 0x0;

	for (int j = 0; j < 3; ++j) {
	    StyleSpec spec = specs[j];
	    if (spec == null) 
		continue;

	    Rule[] ruleList = ruleLists[j];
	    int mask = (1 << j);

	    for (int i = 0; i < ruleList.length; ++i) {
		Rule rule = ruleList[i];

		if (rule.getSelector().match(element)) {
		    spec.put(rule.getStyleSpec());
		    found |= mask;
		}
	    }

	    switch (j) {
	    case BEFORE_ELEMENT:
	    case AFTER_ELEMENT:
		if ((found & mask) != 0) {
		    // Without a generated content, an after or before
		    // pseudo-element is meaningless.
		    if (spec.content == null)
			found &= ~mask;
		}
		break;
	    }
	}

	return found;
    }
}


