/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.stylesheet;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.awt.Color;
import fr.pixware.util.FileUtil;
import fr.pixware.util.StringUtil;
import com.xmlmind.cssparser.*;

public class StyleSheetLoader extends HandlerBase {
    public static final int SYNTAX_ERROR               = 0;
    public static final int UNSUPPORTED_PROPERTY       = 1;
    public static final int UNSUPPORTED_PROPERTY_VALUE = 2;

    public static class Warning {
	public URL url;
	public int lineNumber;
	public int columnNumber;
	public int type;
	public String detail; // Parser message or property name.

	public Warning(URL url, int lineNumber, int columnNumber, int type, 
		       String detail) {
	    this.url = url;
	    this.lineNumber = lineNumber;
	    this.columnNumber = columnNumber;
	    this.type = type;
	    this.detail = detail;
	}
    }

    // -----------------------------------------------------------------------

    public static final String[] SCREEN = { "screen" };
    public static final String[] PRINT  = { "print" };

    private Parser parser;
    private String[] handledMedia = SCREEN;

    private Vector warnings = new Vector();
    private StyleSheet styleSheet;
    private Vector selectors;
    private StyleSpec spec;
    private URL baseURL;

    // -----------------------------------------------------------------------

    private static Hashtable propertyHandlers = new Hashtable();
    static {
	Method[] methods = StyleSheetLoader.class.getDeclaredMethods();

	for (int i = 0; i < methods.length; ++i) {
	    Method method = methods[i];

	    String methodName = method.getName();
	    int charCount = methodName.length();
	    if (charCount <= 6 || methodName.indexOf("handle") != 0)
		continue;

	    StringBuffer propName = new StringBuffer();

	    propName.append(Character.toLowerCase(methodName.charAt(6)));

	    for (int j = 7; j < charCount; ++j) {
		char c = methodName.charAt(j);

		if (Character.isUpperCase(c)) {
		    propName.append('-');
		    propName.append(Character.toLowerCase(c));
		} else {
		    propName.append(c);
		}
	    }

	    propertyHandlers.put(propName.toString(), method);
	}
    }

    // -----------------------------------------------------------------------

    public StyleSheetLoader() {
	parser = new Parser();
	parser.setHandler(this);
    }

    public void setHandledMedia(String[] handledMedia) {
	if (handledMedia == null || handledMedia.length == 0)
	    handledMedia = SCREEN;

	this.handledMedia = new String[handledMedia.length]; 
	for (int i = 0; i < handledMedia.length; ++i)
	    this.handledMedia[i] = handledMedia[i].toLowerCase();
    }

    public String[] getHandledMedia() {
	return handledMedia;
    }

    public void load(String fileName, StyleSheet styleSheet) 
	throws MalformedURLException, IOException {
	load(fileName, null, styleSheet);
    }

    public void load(String fileName, String[] media, StyleSheet styleSheet) 
	throws IOException {
	load(FileUtil.fileToURL(fileName), media, styleSheet);
    }

    public void load(URL url, String[] media, StyleSheet styleSheet) 
	throws IOException {
	load(ParseUtil.loadSource(url), url, media, styleSheet);
    }

    public void load(String source, URL url, String[] media, 
		     StyleSheet styleSheet) 
	throws IOException {
	styleSheet.setURL(url);
	styleSheet.setDefaultMedia(media);

	this.styleSheet = styleSheet;
	warnings.removeAllElements();

	try {
	    parser.parse(source, url, media);
	} finally {
	    baseURL = null;
	    selectors = null;
	    spec = null;
	    styleSheet = null;
	}
    }

    public Warning[] getWarnings() {
	Warning[] list = new Warning[warnings.size()];
	warnings.copyInto(list);

	return list;
    }

    // -----------------------------------------------------------------------

    public boolean beginStyleSheet(URL baseURL, String importSpec, 
				   String[] media) {
	addOtherMedia(styleSheet, media);
	if (!containsHandledMedia(media))
	    return false;

	if (importSpec == null) {
	    this.baseURL = baseURL;
	} else {
	    try {
		this.baseURL = new URL(baseURL, importSpec);
	    } catch (MalformedURLException e) {
		// In this case, endStyleSheet() is directly called by the
		// parser.
		this.baseURL = baseURL;
	    }
	}
	return true;
    }

    public void endStyleSheet(URL baseURL, String importSpec, String[] media) {
	this.baseURL = baseURL;
    }

    public boolean beginMedia(String[] media) {
	addOtherMedia(styleSheet, media);
	if (!containsHandledMedia(media))
	    return false;

	return true;
    }

    private static final void addOtherMedia(StyleSheet styleSheet, 
					    String[] media) {
	if (media != null && media.length > 0) {
	    String[] otherMedia = styleSheet.getOtherMedia();

	    for (int i = 0; i < media.length; ++i) {
		String medium = media[i];

		if (otherMedia == null)
		    otherMedia = new String[] { medium };
		else if (!StringUtil.contains(otherMedia, medium))
		    otherMedia = StringUtil.append(otherMedia, medium);
	    }

	    styleSheet.setOtherMedia(otherMedia);
	}
    }

    public boolean beginPage(String name, String pseudoPage) {
	return false;
    }

    public boolean beginFontFace() {
	return false;
    }

    public void beginRule() {
	selectors = new Vector();
	spec = new StyleSpec();
    }

    public void selector(SelectorComponent[] components) {
	Vector matchers = new Vector();
	int pseudoElement = Selector.PSEUDO_ELEMENT_NONE;

	Selector.Matcher matcher = new Selector.Matcher();
	matcher.subject = Selector.SUBJECT_CURRENT;
	Vector tests = new Vector();
	Selector.Test test;

	int last = components.length - 1;
	for (int i = last; i >= 0; --i) {
	    SelectorComponent component = components[i];

	    if (component.elementName.charAt(0) != '*') {
		test = new Selector.Test();
		test.type = Selector.TEST_ELEMENT_TYPE_NAME_IS;
		test.name = component.elementName;
		tests.addElement(test);
	    }
		
	    for (int j = 0; j < component.attributeSelectors.length; ++j) {
		AttributeSelector attrSel = component.attributeSelectors[j];
		test = new Selector.Test();

		if (attrSel.name == null) {
		    if (attrSel.matcher == null || 
			attrSel.matcher.length() == 0 ||
			attrSel.value == null)
			// Ignore bad selector.
			return;

		    switch (attrSel.matcher.charAt(0)) {
		    case '#':
			test.type = Selector.TEST_ATTRIBUTE_EQUALS;
			test.name = "id";
			break;
		    case '.':
			test.type = Selector.TEST_ATTRIBUTE_CONTAINS;
			test.name = "class";
			break;
		    default:
			// Ignore bad selector.
			return;
		    }
		    test.value = attrSel.value;
		} else {
		    test.name = attrSel.name;

		    if (attrSel.matcher == null || 
			attrSel.matcher.length() == 0 ||
			attrSel.value == null) {
			test.type = Selector.TEST_HAS_ATTRIBUTE;
		    } else {
			switch (attrSel.matcher.charAt(0)) {
			case '=':
			    test.type = Selector.TEST_ATTRIBUTE_EQUALS;
			    break;
			case '~':
			    test.type = Selector.TEST_ATTRIBUTE_CONTAINS;
			    break;
			case '|':
			    test.type = Selector.TEST_ATTRIBUTE_DASH_MATCH;
			    break;
			default:
			    // Ignore bad selector.
			    return;
			}
			test.value = attrSel.value;
		    }
		}

		tests.addElement(test);
	    }

	    for (int k = 0; k < component.pseudos.length; ++k) {
		String pseudo = component.pseudos[k];

		if (pseudo.equals("first-child")) {
		    test = new Selector.Test();
		    test.type = Selector.TEST_IS_FIRST_CHILD;
		    tests.addElement(test);
		} else if (pseudo.equals("before")) {
		    if (i != last)
			// Ignore bad selector.
			return;
		    pseudoElement = Selector.PSEUDO_ELEMENT_BEFORE;
		} else if (pseudo.equals("after")) {
		    if (i != last)
			// Ignore bad selector.
			return;
		    pseudoElement = Selector.PSEUDO_ELEMENT_AFTER;
		} else if (pseudo.indexOf("lang(") == 0) {
			int parent = pseudo.length() - 1;

			if (parent >= 6 && pseudo.charAt(parent) == ')') {
			    test = new Selector.Test();
			    test.type = Selector.TEST_LANGUAGE_IS;
			    test.value = pseudo.substring(5, parent);
			    tests.addElement(test);
			}
		} else {
		    // Ignore bad selector.
		    return;
		}
	    }

	    matcher.tests = new Selector.Test[tests.size()];
	    tests.copyInto(matcher.tests);
	    matchers.addElement(matcher);

	    matcher = new Selector.Matcher();
	    tests = new Vector();
	    switch (component.combinator) {
	    case '>':
		matcher.subject = Selector.SUBJECT_PARENT;
		break;
	    case '+':
		matcher.subject = Selector.SUBJECT_PREDECESSOR;
		break;
	    default:
		matcher.subject = Selector.SUBJECT_ANCESTOR;
	    }
	}

	Selector.Matcher[] matcherList = new Selector.Matcher[matchers.size()];
	matchers.copyInto(matcherList);
	selectors.addElement(new Selector(matcherList, pseudoElement));
    }

    public void property(String name, ValueComponent[] components, 
			 boolean important) {
	Method method = (Method) propertyHandlers.get(name);
	if (method == null) {
	    unsupportedProperty(name);
	    return;
	}

	try {
	    method.invoke(this, new Object[] { components });
	}  catch (InvocationTargetException e1) {
	    Throwable e = e1.getTargetException();
	    e.printStackTrace();
	    throw wrapException(e);
	} catch (Exception e2) {
	    throw wrapException(e2);
	}
    }

    private static RuntimeException wrapException(Throwable e) {
	String fullMessage = e.getClass().getName();

	String message = e.getMessage();
	if (message != null)
	    fullMessage += ": " + message;

	return new RuntimeException(fullMessage);
    }

    public void endRule() {
	int count = selectors.size();
	if (count > 0) {
	    styleSheet.addRule((Selector) selectors.elementAt(0), spec);
	    for (int i = 1; i < count; ++i) {
		// duplicate() because spec is orphaned to styleSheet which
		// will modify it if several rules are to be merged in the
		// first one.
		StyleSpec spec2 = spec.duplicate();

		styleSheet.addRule((Selector) selectors.elementAt(i), spec2);
	    }
	}

	selectors = null;
	spec = null;
    }

    public void error(URL url, int lineNumber, int columnNumber, 
		      String message) {
	warnings.addElement(
	    new Warning(url, lineNumber, columnNumber, SYNTAX_ERROR, message));
    }


    private void unsupportedProperty(String propertyName) {
	warnings.addElement(
	    new Warning(parser.getURL(), 
			parser.getLineNumber(), parser.getColumnNumber(), 
			UNSUPPORTED_PROPERTY, propertyName));
    }

    private void unsupportedPropertyValue(String propertyName) {
	warnings.addElement(
	    new Warning(parser.getURL(), 
			parser.getLineNumber(), parser.getColumnNumber(), 
			UNSUPPORTED_PROPERTY_VALUE, propertyName));
    }

    // -----------------------------------------------------------------------

    private boolean containsHandledMedia(String[] media) {
	if (media == null || media.length == 0 ||
	    StringUtil.contains(media, "all") ||
	    StringUtil.contains(handledMedia, "all"))
	    return true;

	for (int i = 0; i < handledMedia.length; ++i) {
	    if (StringUtil.contains(media, handledMedia[i]))
		return true;
	}

	return false;
    }

    // -----------------------------------------------------------------------

    private void handleMarginTop(ValueComponent[] components) {
	spec.marginTop = getMargin(components[0]);
    }

    private void handleMarginRight(ValueComponent[] components) {
	spec.marginRight = getMargin(components[0]);
    }

    private void handleMarginBottom(ValueComponent[] components) {
	spec.marginBottom = getMargin(components[0]);
    }

    private void handleMarginLeft(ValueComponent[] components) {
	spec.marginLeft = getMargin(components[0]);
    }

    private void handleMargin(ValueComponent[] components) {
	switch (components.length) {
	case 1:
	    spec.marginTop = getMargin(components[0]);
	    spec.marginRight = spec.marginTop;
	    spec.marginBottom = spec.marginTop;
	    spec.marginLeft = spec.marginTop;
	    break;
	case 2:
	    spec.marginTop = getMargin(components[0]);
	    spec.marginRight = getMargin(components[1]);
	    spec.marginBottom = spec.marginTop;
	    spec.marginLeft = spec.marginRight;
	    break;
	case 3:
	    spec.marginTop = getMargin(components[0]);
	    spec.marginRight = getMargin(components[1]);
	    spec.marginBottom = getMargin(components[2]);
	    spec.marginLeft = spec.marginRight;
	    break;
	case 4:
	    spec.marginTop = getMargin(components[0]);
	    spec.marginRight = getMargin(components[1]);
	    spec.marginBottom = getMargin(components[2]);
	    spec.marginLeft = getMargin(components[3]);
	    break;
	default:
	    unsupportedPropertyValue("margin");
	}

	if (spec.marginTop == null || spec.marginRight == null ||
	    spec.marginBottom == null || spec.marginLeft == null) 
	    spec.marginTop = spec.marginRight = 
		spec.marginBottom = spec.marginLeft = null;
    }

    private StyleValue getMargin(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.AUTO:
	    case StyleValue.INHERIT:
		break;
	    default:
		val = val.asPositiveLength();
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("margin");
	return val;
    }

    private void handlePaddingTop(ValueComponent[] components) {
	spec.paddingTop = getPadding(components[0]);
    }

    private void handlePaddingRight(ValueComponent[] components) {
	spec.paddingRight = getPadding(components[0]);
    }

    private void handlePaddingBottom(ValueComponent[] components) {
	spec.paddingBottom = getPadding(components[0]);
    }

    private void handlePaddingLeft(ValueComponent[] components) {
	spec.paddingLeft = getPadding(components[0]);
    }

    private void handlePadding(ValueComponent[] components) {
	switch (components.length) {
	case 1:
	    spec.paddingTop = getPadding(components[0]);
	    spec.paddingRight = spec.paddingTop;
	    spec.paddingBottom = spec.paddingTop;
	    spec.paddingLeft = spec.paddingTop;
	    break;
	case 2:
	    spec.paddingTop = getPadding(components[0]);
	    spec.paddingRight = getPadding(components[1]);
	    spec.paddingBottom = spec.paddingTop;
	    spec.paddingLeft = spec.paddingRight;
	    break;
	case 3:
	    spec.paddingTop = getPadding(components[0]);
	    spec.paddingRight = getPadding(components[1]);
	    spec.paddingBottom = getPadding(components[2]);
	    spec.paddingLeft = spec.paddingRight;
	    break;
	case 4:
	    spec.paddingTop = getPadding(components[0]);
	    spec.paddingRight = getPadding(components[1]);
	    spec.paddingBottom = getPadding(components[2]);
	    spec.paddingLeft = getPadding(components[3]);
	    break;
	default:
	    unsupportedPropertyValue("padding");
	}

	if (spec.paddingTop == null || spec.paddingRight == null ||
	    spec.paddingBottom == null || spec.paddingLeft == null) 
	    spec.paddingTop = spec.paddingRight = 
		spec.paddingBottom = spec.paddingLeft = null;
    }

    private StyleValue getPadding(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
		break;
	    default:
		val = val.asPositiveLength();
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("padding");
	return val;
    }

    private void handleBorderStyle(ValueComponent[] components) {
	spec.borderStyle = getBorderStyle(components[0]);
    }

    private StyleValue getBorderStyle(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.NONE:
	    case StyleValue.DOTTED:
	    case StyleValue.DASHED:
	    case StyleValue.SOLID:
	    case StyleValue.DOUBLE:
	    case StyleValue.GROOVE:
	    case StyleValue.RIDGE:
	    case StyleValue.INSET:
	    case StyleValue.OUTSET:
		break;
	    default:
		val = null;
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("border-style");
	return val;
    }

    private void handleBorderWidth(ValueComponent[] components) {
	spec.borderWidth = getBorderWidth(components[0]);
    }
    
    private StyleValue getBorderWidth(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.THIN:
	    case StyleValue.MEDIUM:
	    case StyleValue.THICK:
		break;
	    default:
		val = val.asPositiveLength();
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("border-width");
	return val;
    }

    private void handleBorderTopColor(ValueComponent[] components) {
	spec.borderTopColor = getColor(components[0], true);
    }

    private void handleBorderRightColor(ValueComponent[] components) {
	spec.borderRightColor = getColor(components[0], true);
    }

    private void handleBorderBottomColor(ValueComponent[] components) {
	spec.borderBottomColor = getColor(components[0], true);
    }

    private void handleBorderLeftColor(ValueComponent[] components) {
	spec.borderLeftColor = getColor(components[0], true);
    }

    private void handleBorderColor(ValueComponent[] components) {
	switch (components.length) {
	case 1:
	    spec.borderTopColor = getColor(components[0], true);
	    spec.borderRightColor = spec.borderTopColor;
	    spec.borderBottomColor = spec.borderTopColor;
	    spec.borderLeftColor = spec.borderTopColor;
	    break;
	case 2:
	    spec.borderTopColor = getColor(components[0], true);
	    spec.borderRightColor = getColor(components[1], true);
	    spec.borderBottomColor = spec.borderTopColor;
	    spec.borderLeftColor = spec.borderRightColor;
	    break;
	case 3:
	    spec.borderTopColor = getColor(components[0], true);
	    spec.borderRightColor = getColor(components[1], true);
	    spec.borderBottomColor = getColor(components[2], true);
	    spec.borderLeftColor = spec.borderRightColor;
	    break;
	case 4:
	    spec.borderTopColor = getColor(components[0], true);
	    spec.borderRightColor = getColor(components[1], true);
	    spec.borderBottomColor = getColor(components[2], true);
	    spec.borderLeftColor = getColor(components[3], true);
	    break;
	default:
	    unsupportedPropertyValue("border-color");
	}

	if (spec.borderTopColor == null || spec.borderRightColor == null ||
	    spec.borderBottomColor == null || spec.borderLeftColor == null) 
	    spec.borderTopColor = spec.borderRightColor = 
		spec.borderBottomColor = spec.borderLeftColor = null;
    }

    private StyleValue getColor(ValueComponent component, 
				boolean allowTransparent) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
		break;
	    case StyleValue.TRANSPARENT:
		if (!allowTransparent)
		    val = null;
		break;
	    case StyleValue.COLOR:
	    case StyleValue.IDENTIFIER:
		val = val.asColor();
		break;
	    default:
		val = null;
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("color");
	return val;
    }

    private void handleBorder(ValueComponent[] components) {
	switch (components.length) {
	case 1:
	    spec.borderWidth = getBorderWidth(components[0]);

	    if (spec.borderWidth != null && 
		spec.borderWidth.type == StyleValue.INHERIT) {
		spec.borderStyle = 
		    spec.borderTopColor = spec.borderRightColor = 
		    spec.borderBottomColor = spec.borderLeftColor = 
		    spec.borderWidth;
	    }
	    break;
	case 2:
	    spec.borderWidth = getBorderWidth(components[0]);
	    spec.borderStyle = getBorderStyle(components[1]);

	    if (spec.borderWidth == null || spec.borderStyle == null) {
		spec.borderWidth = spec.borderStyle = null;
	    }
	    break;
	case 3:
	    spec.borderWidth = getBorderWidth(components[0]);
	    spec.borderStyle = getBorderStyle(components[1]);
	    spec.borderTopColor = spec.borderRightColor = 
		spec.borderBottomColor = spec.borderLeftColor = 
		getColor(components[2], true);

	    if (spec.borderWidth == null || spec.borderStyle == null ||
		spec.borderTopColor == null) {
		spec.borderWidth = spec.borderStyle = 
		    spec.borderTopColor = spec.borderRightColor = 
		    spec.borderBottomColor = spec.borderLeftColor = null;
	    }
	    break;
	default:
	    unsupportedPropertyValue("border");
	}
    }

    private void handleBackgroundColor(ValueComponent[] components) {
	spec.backgroundColor = getColor(components[0], true);
    }

    private void handleColor(ValueComponent[] components) {
	spec.color = getColor(components[0], false);
    }

    private void handleFontFamily(ValueComponent[] components) {
	spec.fontFamily = getFontFamily(components, 0);
    }

    private StyleValue getFontFamily(ValueComponent[] components, int offset) {
	StyleValue val = null;

	// Example: font-family: Baskerville, "Heisi Mincho W3", Symbol, serif
	for (int i = offset; i < components.length; ++i) {
	    val = StyleValue.fromValueComponent(components[i]);

	    if (val != null) {
		switch (val.type) {
		case StyleValue.FUNCTION:
		    val.resolveURL(baseURL);
		    break;
		case StyleValue.INHERIT:
		case StyleValue.SERIF:
		case StyleValue.SANS_SERIF:
		case StyleValue.MONOSPACE:
		    break;
		default:
		    val = null;
		}
	    }

	    if (val != null)
		break;
	}

	if (val == null)
	    unsupportedPropertyValue("font-family");
	return val;
    }

    private void handleFontStyle(ValueComponent[] components) {
	spec.fontStyle = getFontStyle(components[0]);
    }

    private StyleValue getFontStyle(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.NORMAL:
	    case StyleValue.ITALIC:
	    case StyleValue.OBLIQUE:
		break;
	    default:
		val = null;
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("font-style");
	return val;
    }

    private void handleFontWeight(ValueComponent[] components) {
	spec.fontWeight = getFontWeight(components[0]);
    }

    private StyleValue getFontWeight(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.NORMAL:
	    case StyleValue.BOLD:
		break;
	    default:
		val = null;
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("font-weight");
	return val;
    }

    private void handleFontSize(ValueComponent[] components) {
	spec.fontSize = getFontSize(components[0]);
    }

    private StyleValue getFontSize(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.MEDIUM:
	    case StyleValue.SMALL:
	    case StyleValue.LARGE:
	    case StyleValue.X_SMALL:
	    case StyleValue.X_LARGE:
	    case StyleValue.XX_SMALL:
	    case StyleValue.XX_LARGE:
	    case StyleValue.LARGER:
	    case StyleValue.SMALLER:
		break;
	    case StyleValue.PERCENTAGE:
		if (val.number <= 0)
		    val = null;
		break;
	    default:
		val = val.asStrictlyPositiveLength();
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("font-size");
	return val;
    }

    private void handleFont(ValueComponent[] components) {
	ValueComponent component = components[0];

	switch (components.length) {
	case 1:
	    {
		StyleValue val = StyleValue.fromValueComponent(component);
		if (val != null && val.type == StyleValue.INHERIT) {
		    spec.fontStyle = spec.fontWeight = 
			spec.fontSize = spec.fontFamily = val;
		}
	    }
	    break;
	case 2:
	    spec.fontSize = getFontSize(component);
	    spec.fontFamily = getFontFamily(components, 1);

	    if (spec.fontSize == null || spec.fontFamily == null) {
		spec.fontSize = spec.fontFamily = null;
	    }
	    break;
	case 3:
	    spec.fontStyle = getFontStyle(component);
	    spec.fontSize = getFontSize(components[1]);
	    spec.fontFamily = getFontFamily(components, 2);

	    if (spec.fontStyle == null || spec.fontSize == null ||
		spec.fontFamily == null) {
		spec.fontStyle = 
		    spec.fontSize = spec.fontFamily = null;
	    }
	    break;
	case 4:
	    spec.fontStyle = getFontStyle(component);
	    spec.fontWeight = getFontWeight(components[1]);
	    spec.fontSize = getFontSize(components[2]);
	    spec.fontFamily = getFontFamily(components, 3);

	    if (spec.fontStyle == null || spec.fontWeight == null ||
		spec.fontSize == null || spec.fontFamily == null) {
		spec.fontStyle = spec.fontWeight = 
		    spec.fontSize = spec.fontFamily = null;
	    }
	    break;
	default:
	    unsupportedPropertyValue("font");
	}
    }

    private void handleTextDecoration(ValueComponent[] components) {
	spec.textDecoration = StyleValue.fromValueComponent(components[0]);

	if (spec.textDecoration != null) {
	    switch (spec.textDecoration.type) {
	    case StyleValue.FUNCTION:
		spec.textDecoration.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.NONE:
	    case StyleValue.UNDERLINE:
	    case StyleValue.OVERLINE:
	    case StyleValue.LINE_THROUGH:
		break;
	    default:
		spec.textDecoration = null;
	    }
	}

	if (spec.textDecoration == null)
	    unsupportedPropertyValue("text-decoration");
    }

    private void handleTextAlign(ValueComponent[] components) {
	spec.textAlign = StyleValue.fromValueComponent(components[0]);

	if (spec.textAlign != null) {
	    switch (spec.textAlign.type) {
	    case StyleValue.FUNCTION:
		spec.textAlign.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.LEFT:
	    case StyleValue.RIGHT:
	    case StyleValue.CENTER:
		break;
	    default:
		spec.textAlign = null;
	    }
	}

	if (spec.textAlign == null)
	    unsupportedPropertyValue("text-align");
    }

    private void handleTextIndent(ValueComponent[] components) {
	spec.textIndent = StyleValue.fromValueComponent(components[0]);

	if (spec.textIndent != null) {
	    switch (spec.textIndent.type) {
	    case StyleValue.FUNCTION:
		spec.textIndent.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
		break;
	    default:
		spec.textIndent = spec.textIndent.asLength();
	    }
	}

	if (spec.textIndent == null)
	    unsupportedPropertyValue("text-indent");
    }

    private void handleVerticalAlign(ValueComponent[] components) {
	spec.verticalAlign = StyleValue.fromValueComponent(components[0]);

	if (spec.verticalAlign != null) {
	    switch (spec.verticalAlign.type) {
	    case StyleValue.FUNCTION:
		spec.verticalAlign.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.BASELINE:
	    case StyleValue.SUB:
	    case StyleValue.SUPER:
	    case StyleValue.TOP:
	    case StyleValue.TEXT_TOP:
	    case StyleValue.MIDDLE:
	    case StyleValue.BOTTOM:
	    case StyleValue.TEXT_BOTTOM:
		break;
	    default:
		spec.verticalAlign = null;
	    }
	}

	if (spec.verticalAlign == null)
	    unsupportedPropertyValue("vertical-align");
    }

    private void handleLineHeight(ValueComponent[] components) {
	spec.lineHeight = StyleValue.fromValueComponent(components[0]);

	if (spec.lineHeight != null) {
	    switch (spec.lineHeight.type) {
	    case StyleValue.FUNCTION:
		spec.lineHeight.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.NORMAL:
		break;
	    case StyleValue.NUMBER:
		if (spec.lineHeight.number <= 0)
		    spec.lineHeight = null;
		break;
	    default:
		spec.lineHeight = null;
	    }
	}

	if (spec.lineHeight == null)
	    unsupportedPropertyValue("line-height");
    }

    private void handleWhiteSpace(ValueComponent[] components) {
	spec.whiteSpace = StyleValue.fromValueComponent(components[0]);

	if (spec.whiteSpace != null) {
	    switch (spec.whiteSpace.type) {
	    case StyleValue.FUNCTION:
		spec.whiteSpace.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.NORMAL:
	    case StyleValue.PRE:
	    case StyleValue.NOWRAP:
		break;
	    default:
		spec.whiteSpace = null;
	    }
	}

	if (spec.whiteSpace == null)
	    unsupportedPropertyValue("white-space");
    }

    private void handleDisplay(ValueComponent[] components) {
	spec.display = StyleValue.fromValueComponent(components[0]);

	if (spec.display != null) {
	    switch (spec.display.type) {
	    case StyleValue.FUNCTION:
		spec.display.resolveURL(baseURL);
		break;
	    case StyleValue.INHERIT:
	    case StyleValue.NONE:
	    case StyleValue.INLINE:
	    case StyleValue.BLOCK:
	    case StyleValue.MARKER:
	    case StyleValue.TABLE:
	    case StyleValue.INLINE_TABLE:
	    case StyleValue.TABLE_COLUMN_GROUP:
	    case StyleValue.TABLE_COLUMN:
	    case StyleValue.TABLE_HEADER_GROUP:
	    case StyleValue.TABLE_FOOTER_GROUP:
	    case StyleValue.TABLE_ROW_GROUP:
	    case StyleValue.TABLE_ROW:
	    case StyleValue.TABLE_CELL:
	    case StyleValue.TABLE_CAPTION:
	    case StyleValue.INLINE_BLOCK:
	    case StyleValue.TREE:
		break;
	    default:
		spec.display = null;
	    }
	}

	if (spec.display == null)
	    unsupportedPropertyValue("display");
    }

    private void handleContent(ValueComponent[] components) {
	spec.content = StyleValue.fromValueComponents(components);

	if (spec.content != null) {
	    StyleValue[] list;
	    if (spec.content.type == StyleValue.LIST)
		list = spec.content.list;
	    else
		list = new StyleValue[] { spec.content };

	    for (int i = 0; i < list.length; ++i) {
		StyleValue item = list[i];

		switch (item.type) {
		case StyleValue.FUNCTION:
		case StyleValue.URI:
		    item.resolveURL(baseURL);
		    break;
		case StyleValue.STRING:
		case StyleValue.ATTRIBUTE:
		case StyleValue.COUNTER:
		case StyleValue.COUNTERS:
		case StyleValue.OPEN_QUOTE:
		case StyleValue.CLOSE_QUOTE:
		case StyleValue.NO_OPEN_QUOTE:
		case StyleValue.NO_CLOSE_QUOTE:
		case StyleValue.DISC:
		case StyleValue.CIRCLE:
		case StyleValue.SQUARE:
		    break;
		default:
		    spec.content = null;
		}

		if (spec.content == null)
		    break;
	    }
	}

	if (spec.content == null)
	    unsupportedPropertyValue("content");
    }

    private void handleMarkerOffset(ValueComponent[] components) {
	spec.markerOffset = StyleValue.fromValueComponent(components[0]);

	if (spec.markerOffset != null) {
	    switch (spec.markerOffset.type) {
	    case StyleValue.FUNCTION:
		spec.markerOffset.resolveURL(baseURL);
		break;
	    case StyleValue.AUTO:
		break;
	    default:
		spec.markerOffset = spec.markerOffset.asPositiveLength();
	    }
	}

	if (spec.markerOffset == null)
	    unsupportedPropertyValue("marker-offset");
    }

    private void handleWidth(ValueComponent[] components) {
	spec.width = getSize(components[0]);
    }

    private void handleHeight(ValueComponent[] components) {
	spec.height = getSize(components[0]);
    }

    private StyleValue getSize(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.AUTO:
		break;
	    default:
		val = val.asStrictlyPositiveLength();
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("width/height");
	return val;
    }

    private void handleCaptionSide(ValueComponent[] components) {
	spec.captionSide = StyleValue.fromValueComponent(components[0]);

	if (spec.captionSide != null) {
	    switch (spec.captionSide.type) {
	    case StyleValue.FUNCTION:
		spec.captionSide.resolveURL(baseURL);
		break;
	    case StyleValue.TOP:
	    case StyleValue.BOTTOM:
		break;
	    }
	}

	if (spec.captionSide == null)
	    unsupportedPropertyValue("caption-side");
    }

    private void handleBorderSpacing(ValueComponent[] components) {
	switch (components.length) {
	case 1:
	    spec.borderHorizontalSpacing = 
		spec.borderVerticalSpacing = getBorderSpacing(components[0]);
	    break;
	case 2:
	    spec.borderHorizontalSpacing = getBorderSpacing(components[0]);
	    spec.borderVerticalSpacing = getBorderSpacing(components[1]);

	    if (spec.borderHorizontalSpacing == null ||
		spec.borderVerticalSpacing == null) {
		spec.borderHorizontalSpacing = 
		    spec.borderVerticalSpacing = null;
	    }
	    break;
	default:
	    unsupportedPropertyValue("border-spacing");
	}
    }

    private StyleValue getBorderSpacing(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    default:
		val = val.asPositiveLength();
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("border-spacing");
	return val;
    }

    private void handleRowSpan(ValueComponent[] components) {
	spec.rowSpan = getSpan(components[0]);
    }

    private void handleColumnSpan(ValueComponent[] components) {
	spec.columnSpan = getSpan(components[0]);
    }

    private StyleValue getSpan(ValueComponent component) {
	StyleValue val = StyleValue.fromValueComponent(component);
	
	if (val != null) {
	    switch (val.type) {
	    case StyleValue.FUNCTION:
		val.resolveURL(baseURL);
		break;
	    case StyleValue.NUMBER:
		if ((int) val.number <= 0)
		    val = null;
		break;
	    default:
		val = null;
	    }
	}

	if (val == null)
	    unsupportedPropertyValue("row/column-span");
	return val;
    }
}
