/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.stylesheet;

import java.io.*;
import java.awt.Color;

public class StyleSheetWriter {
    public static void writeStyleSheet(StyleSheet styleSheet, String fileName) 
	throws IOException {
	PrintWriter out = new PrintWriter(new FileWriter(fileName));
	    
	try {
	    writeStyleSheet(styleSheet, out);
	} finally {
	    out.flush();
	    out.close();
	}
    }

    public static void writeStyleSheet(StyleSheet styleSheet, PrintWriter out)
	throws IOException {
	Rule[] rules = styleSheet.getRules();
	for (int i = 0; i < rules.length; ++i) {
	    writeRule(rules[i], out);
	    out.println();
	}
    }

    public static void writeRule(Rule rule, PrintWriter out) 
	throws IOException {
	writeSelector(rule.getSelector(), out);
	out.print(' ');
	writeStyleSpec(rule.getStyleSpec(), out);
    }

    public static void writeSelector(Selector selector, PrintWriter out) 
	throws IOException {
	Selector.Matcher[] matchers = selector.getMatchers();

	for (int i = matchers.length - 1; i >= 0; --i) {
	    Selector.Matcher matcher = matchers[i];
	    int testCount = matcher.tests.length;

	    if (testCount == 0) {
		out.print('*');
	    } else {
		for (int j = 0; j < testCount; ++j) {
		    Selector.Test test = matcher.tests[j];

		    if (j == 0 && 
			test.type != Selector.TEST_ELEMENT_TYPE_NAME_IS)
			out.print('*');

		    switch (test.type) {
		    case Selector.TEST_ELEMENT_TYPE_NAME_IS:
			out.print(escapeIdent(test.name));
			break;
		    case Selector.TEST_IS_FIRST_CHILD:
			out.print(":first-child");
			break;
		    case Selector.TEST_HAS_ATTRIBUTE:
			out.print("[" + escapeIdent(test.name) + ']');
			break;
		    case Selector.TEST_ATTRIBUTE_EQUALS:
			out.print("[" + escapeIdent(test.name) + '=' + 
				  quote(test.value) + ']');
			break;
		    case Selector.TEST_ATTRIBUTE_CONTAINS:
			out.print("[" + escapeIdent(test.name) + "~=" + 
				  quote(test.value) + ']');
			break;
		    case Selector.TEST_ATTRIBUTE_DASH_MATCH:
			out.print("[" + escapeIdent(test.name) + "|=" + 
				  quote(test.value) + ']');
			break;
		    case Selector.TEST_LANGUAGE_IS:
			out.print(":lang(" + escapeIdent(test.value) + ')');
			break;
		    }
		}
	    }

	    switch (matcher.subject) {
	    case Selector.SUBJECT_PARENT:
		out.print(" > ");
		break;
	    case Selector.SUBJECT_PREDECESSOR:
		out.print(" + ");
		break;
	    case Selector.SUBJECT_ANCESTOR:
		out.print(' ');
		break;
	    }
	}

	switch (selector.getPseudoElement()) {
	case Selector.PSEUDO_ELEMENT_BEFORE:
	    out.print(":before");
	    break;
	case Selector.PSEUDO_ELEMENT_AFTER:
	    out.print(":after");
	    break;
	}
    }

    public static void writeStyleSpec(StyleSpec spec, PrintWriter out) 
	throws IOException {
	out.println('{');

	writeProperty("margin-top", spec.marginTop, out);
	writeProperty("margin-right", spec.marginRight, out);
	writeProperty("margin-bottom", spec.marginBottom, out);
	writeProperty("margin-left", spec.marginLeft, out);

	writeProperty("padding-top", spec.paddingTop, out);
	writeProperty("padding-right", spec.paddingRight, out);
	writeProperty("padding-bottom", spec.paddingBottom, out);
	writeProperty("padding-left", spec.paddingLeft, out);

	writeProperty("border-style", spec.borderStyle, out);
	writeProperty("border-width", spec.borderWidth, out);
	writeProperty("border-top-color", spec.borderTopColor, out);
	writeProperty("border-right-color", spec.borderRightColor, out);
	writeProperty("border-bottom-color", spec.borderBottomColor, out);
	writeProperty("border-left-color", spec.borderLeftColor, out);

	writeProperty("background-color", spec.backgroundColor, out);
	writeProperty("color", spec.color, out);

	writeProperty("font-family", spec.fontFamily, out);
	writeProperty("font-style", spec.fontStyle, out);
	writeProperty("font-weight", spec.fontWeight, out);
	writeProperty("font-size", spec.fontSize, out);

	writeProperty("text-decoration", spec.textDecoration, out);

	writeProperty("text-align", spec.textAlign, out);
	writeProperty("text-indent", spec.textIndent, out);

	writeProperty("vertical-align", spec.verticalAlign, out);

	writeProperty("line-height", spec.lineHeight, out);

	writeProperty("white-space", spec.whiteSpace, out);

	writeProperty("display", spec.display, out);

	writeProperty("content", spec.content, out);

	writeProperty("marker-offset", spec.markerOffset, out);

	writeProperty("width", spec.width, out);
	writeProperty("height", spec.height, out);

	writeProperty("caption-side", spec.captionSide, out);

	writeProperty("border-spacing", spec.borderHorizontalSpacing, 
		      spec.borderVerticalSpacing, out);

	writeProperty("row-span", spec.rowSpan, out);
	writeProperty("column-span", spec.columnSpan, out);

	out.println('}');
    }

    public static void writeProperty(String name, StyleValue value, 
				     PrintWriter out) 
	throws IOException {
	if (value != null) {
	    out.print("    " + name + ": ");
	    writeStyleValue(value, out);
	    out.println(';');
	}
    }

    public static void writeProperty(String name, StyleValue value1, 
				     StyleValue value2, PrintWriter out) 
	throws IOException {
	if (value1 != null || value2 != null) {
	    out.print("    " + name + ":");

	    if (value1 != null) {
		out.print(' ');
		writeStyleValue(value1, out);
	    }

	    if (value2 != null) {
		out.print(' ');
		writeStyleValue(value2, out);
	    }

	    out.println(';');
	}
    }

    public static void writeStyleValue(StyleValue val, PrintWriter out) 
	throws IOException {
	switch (val.type) {
	case StyleValue.NUMBER:
	    out.print(val.number);
	    break;
	case StyleValue.STRING:
	    out.print(quote(val.string));
	    break;
	case StyleValue.IDENTIFIER:
	    out.print(escapeIdent(val.string));
	    break;
	case StyleValue.LIST:
	    {
		StyleValue[] list = val.list;

		for (int i = 0; i < list.length; ++i) {
		    if (i > 0)
			out.print(' ');
		    writeStyleValue(list[i], out);
		}
	    }
	    break;
	case StyleValue.LENGTH_PX:
	    out.print(val.number);
	    out.print("px");
	    break;
	case StyleValue.LENGTH_IN:
	    out.print(val.number);
	    out.print("in");
	    break;
	case StyleValue.LENGTH_CM:
	    out.print(val.number);
	    out.print("cm");
	    break;
	case StyleValue.LENGTH_MM:
	    out.print(val.number);
	    out.print("mm");
	    break;
	case StyleValue.LENGTH_PT:
	    out.print(val.number);
	    out.print("pt");
	    break;
	case StyleValue.LENGTH_PC:
	    out.print(val.number);
	    out.print("pc");
	    break;
	case StyleValue.LENGTH_EM:
	    out.print(val.number);
	    out.print("em");
	    break;
	case StyleValue.LENGTH_EX:
	    out.print(val.number);
	    out.print("ex");
	    break;
	case StyleValue.ANGLE_DEG:
	    out.print(val.number);
	    out.print("deg");
	    break;
	case StyleValue.ANGLE_GRAD:
	    out.print(val.number);
	    out.print("grad");
	    break;
	case StyleValue.ANGLE_RAD:
	    out.print(val.number);
	    out.print("rad");
	    break;
	case StyleValue.TIME_MS:
	    out.print(val.number);
	    out.print("ms");
	    break;
	case StyleValue.TIME_S:
	    out.print(val.number);
	    out.print("s");
	    break;
	case StyleValue.FREQUENCY_HZ:
	    out.print(val.number);
	    out.print("hz");
	    break;
	case StyleValue.FREQUENCY_KHZ:
	    out.print(val.number);
	    out.print("khz");
	    break;
	case StyleValue.PERCENTAGE:
	    out.print(val.number);
	    out.print("%");
	    break;
	case StyleValue.DIMENSION:
	    out.print(val.number);
	    out.print(escapeIdent(val.string));
	    break;
	case StyleValue.URI:
	    out.print("url(");
	    out.print(quote(val.string));
	    out.print(')');
	    break;
	case StyleValue.ATTRIBUTE:
	    out.print("attr(");
	    out.print(escapeIdent(val.string));
	    out.print(')');
	    break;
	case StyleValue.COLOR:
	    {
		Color color = val.color;

		out.print("rgb(");
		out.print(color.getRed());
		out.print(',');
		out.print(color.getGreen());
		out.print(',');
		out.print(color.getBlue());
		out.print(')');
	    }
	    break;
	case StyleValue.COUNTER:
	    out.print("counter(");
	    out.print(escapeIdent(val.string));
	    out.print(", ");
	    writeStyleValue(val.list[0], out);
	    out.print(')');
	    break;
	case StyleValue.COUNTERS:
	    out.print("counters(");
	    out.print(escapeIdent(val.string));
	    out.print(", ");
	    writeStyleValue(val.list[0], out);
	    out.print(", ");
	    writeStyleValue(val.list[1], out);
	    out.print(')');
	    break;
	case StyleValue.FUNCTION:
	    {
		out.print(escapeIdent(val.string));
		out.print('(');

		StyleValue[] args = val.list;

		for (int i = 0; i < args.length; ++i) {
		    if (i > 0)
			out.print(", ");
		    writeStyleValue(args[i], out);
		}

		out.print(')');
	    }
	    break;
	default:
	    { 
		String keyword = StyleValue.getKeyword(val.type);
		if (keyword != null)
		    // No need to escape keywords.
		    out.print(keyword);
	    }
	    break;
	}
    }

    public static String quote(String s) {
	StringBuffer buffer = new StringBuffer();
	buffer.append('\"');

	int count = s.length();
	for (int i = 0; i < count; ++i) {
	    char c = s.charAt(i);

	    switch (c) {
	    case '\"':
		buffer.append("\\\"");
		break;
	    default:
		if ((c >= '#' && c <= '~') ||
		    c == ' ' || c == '\t'  || c == '!' || 
		    c >= '\200') {
		    buffer.append(c);
		} else {
		    buffer.append('\\');
		    String hex = Integer.toHexString(c).toUpperCase();
		    buffer.append(hex);
		    if (hex.length() < 6)
			buffer.append(' ');
		}
	    }
	}

	buffer.append('\"');
	return buffer.toString();
    }

    public static String escapeIdent(String s) {
	StringBuffer buffer = new StringBuffer();

	int count = s.length();
	for (int i = 0; i < count; ++i) {
	    char c = s.charAt(i);

	    if (!(c == '-' ||
		  (c >= '0' && c <= '9') ||
		  (c >= 'A' && c <= 'Z') ||
		  (c >= 'a' && c <= 'z') ||
		  c >= '\200')) {
		buffer.append('\\');
	    }
	    buffer.append(c);
	}

	return buffer.toString();
    }
}
