/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmledit.stylesheet;

import java.net.URL;
import java.net.MalformedURLException;
import java.util.Hashtable;
import java.awt.Color;
import com.xmlmind.cssparser.ValueComponent;
import com.xmlmind.cssparser.ParseUtil;

public class StyleValue {
    public static final int UNKNOWN_TYPE = -1; // N/A

    public static final int NUMBER     = 0; // number
    public static final int STRING     = 1; // string
    public static final int IDENTIFIER = 2; // string
    public static final int LIST       = 3; // list

    public static final int LENGTH_PX = 4; // number
    public static final int LENGTH_IN = 5;
    public static final int LENGTH_CM = 6;
    public static final int LENGTH_MM = 7;
    public static final int LENGTH_PT = 8;
    public static final int LENGTH_PC = 9;
    public static final int LENGTH_EM = 10;
    public static final int LENGTH_EX = 11;

    public static final int ANGLE_DEG  = 12; // number
    public static final int ANGLE_GRAD = 13;
    public static final int ANGLE_RAD  = 14;

    public static final int TIME_MS = 15; // number
    public static final int TIME_S  = 16;

    public static final int FREQUENCY_HZ  = 17; // number
    public static final int FREQUENCY_KHZ = 18;

    public static final int PERCENTAGE = 19; // number
    public static final int DIMENSION  = 20; // number,string=unit

    public static final int URI       = 21; // string
    public static final int ATTRIBUTE = 22; // string
    public static final int COUNTER   = 23; // string=name,list[0]=numbering
    public static final int COUNTERS  = 24; 
                             // string=name,list[0]=separator,list[1]=numbering
    public static final int COLOR     = 25; // color
    public static final int FUNCTION  = 26; // string=name,list=arguments

    public static final int INHERIT = 27; // type=enumerated value
    public static final int AUTO    = 28;
    public static final int NORMAL  = 29;
    public static final int NONE    = 30;
    // border-style
    public static final int DOTTED  = 31;
    public static final int DASHED  = 32;
    public static final int SOLID   = 33;
    public static final int DOUBLE  = 34;
    public static final int GROOVE  = 35;
    public static final int RIDGE   = 36;
    public static final int INSET   = 37;
    public static final int OUTSET  = 38;
    // border-width
    public static final int THIN   = 39;
    public static final int MEDIUM = 40;
    public static final int THICK  = 41;
    // background-color, border-color
    public static final int TRANSPARENT = 42;
    // font-family
    public static final int SERIF      = 43;
    public static final int SANS_SERIF = 44;
    public static final int MONOSPACE  = 45;
    // font-style
    public static final int ITALIC  = 46;
    public static final int OBLIQUE = 47;
    // font-weight
    public static final int BOLD = 48;
    // font-size
    public static final int SMALL    = 49;
    public static final int LARGE    = 50;
    public static final int X_SMALL  = 51;
    public static final int X_LARGE  = 52;
    public static final int XX_SMALL = 53;
    public static final int XX_LARGE = 54;
    public static final int LARGER   = 55;
    public static final int SMALLER  = 56;
    // text-decoration
    public static final int UNDERLINE    = 57;
    public static final int OVERLINE     = 58;
    public static final int LINE_THROUGH = 59;
    // text-align
    public static final int LEFT    = 60;
    public static final int RIGHT   = 61;
    public static final int CENTER  = 62;
    // vertical-align
    public static final int BASELINE    = 63;
    public static final int TOP         = 64;
    public static final int MIDDLE      = 65;
    public static final int BOTTOM      = 66;
    public static final int SUB         = 67;
    public static final int SUPER       = 68;
    public static final int TEXT_TOP    = 69;
    public static final int TEXT_BOTTOM = 70;
    // white-space
    public static final int PRE    = 71;
    public static final int NOWRAP = 72;
    // display
    public static final int INLINE             = 73;
    public static final int BLOCK              = 74;
    public static final int MARKER             = 75;
    public static final int TABLE              = 76;
    public static final int INLINE_TABLE       = 77;
    public static final int TABLE_COLUMN_GROUP = 78;
    public static final int TABLE_COLUMN       = 79;
    public static final int TABLE_ROW_GROUP    = 80;
    public static final int TABLE_HEADER_GROUP = 81;
    public static final int TABLE_FOOTER_GROUP = 82;
    public static final int TABLE_ROW          = 83;
    public static final int TABLE_CELL         = 84;
    public static final int TABLE_CAPTION      = 85;
    public static final int INLINE_BLOCK       = 86;
    public static final int TREE               = 87;
    // content items
    public static final int OPEN_QUOTE      = 88;
    public static final int CLOSE_QUOTE     = 89;
    public static final int NO_OPEN_QUOTE   = 90;
    public static final int NO_CLOSE_QUOTE  = 91;
    public static final int DISC            = 92;
    public static final int CIRCLE          = 93;
    public static final int SQUARE          = 94;
    // counter styles
    public static final int DECIMAL     = 95;
    public static final int LOWER_ALPHA = 96;
    public static final int UPPER_ALPHA = 97;
    public static final int LOWER_ROMAN = 98;
    public static final int UPPER_ROMAN = 99;

    public int type = UNKNOWN_TYPE;
    public double number = 0.0;
    public String string = null; 
    public StyleValue[] list = null;
    public Color color = null;

    // -----------------------------------------------------------------------

    private static final class KeywordValue {
	public String keyword;
	public int value;

	public KeywordValue(String keyword, int value) {
	    this.keyword = keyword;
	    this.value = value;
	}

	public int hashCode() {
	    return value;
	}

	public boolean equals(Object other) {
	    if (other == null || !(other instanceof KeywordValue))
		return false;
	    return (((KeywordValue) other).value == value);
	}
    }

    private static final KeywordValue[] keywordValues = {
	new KeywordValue("inherit", INHERIT),
	new KeywordValue("auto", AUTO),
	new KeywordValue("normal", NORMAL),
	new KeywordValue("none", NONE),
	new KeywordValue("dotted", DOTTED),
	new KeywordValue("dashed", DASHED),
	new KeywordValue("solid", SOLID),
	new KeywordValue("double", DOUBLE),
	new KeywordValue("groove", GROOVE),
	new KeywordValue("ridge", RIDGE),
	new KeywordValue("inset", INSET),
	new KeywordValue("outset", OUTSET),
	new KeywordValue("thin", THIN),
	new KeywordValue("medium", MEDIUM),
	new KeywordValue("thick", THICK),
	new KeywordValue("transparent", TRANSPARENT),
	new KeywordValue("serif", SERIF),
	new KeywordValue("sans-serif", SANS_SERIF),
	new KeywordValue("monospace", MONOSPACE),
	new KeywordValue("italic", ITALIC),
	new KeywordValue("oblique", OBLIQUE),
	new KeywordValue("bold", BOLD),
	new KeywordValue("small", SMALL),
	new KeywordValue("large", LARGE),
	new KeywordValue("x-small", X_SMALL),
	new KeywordValue("x-large", X_LARGE),
	new KeywordValue("xx-small", XX_SMALL),
	new KeywordValue("xx-large", XX_LARGE),
	new KeywordValue("larger", LARGER),
	new KeywordValue("smaller", SMALLER),
	new KeywordValue("underline", UNDERLINE),
	new KeywordValue("overline", OVERLINE),
	new KeywordValue("line-through", LINE_THROUGH),
	new KeywordValue("left", LEFT),
	new KeywordValue("right", RIGHT),
	new KeywordValue("center", CENTER),
	new KeywordValue("baseline", BASELINE),
	new KeywordValue("top", TOP),
	new KeywordValue("middle", MIDDLE),
	new KeywordValue("bottom", BOTTOM),
	new KeywordValue("sub", SUB),
	new KeywordValue("super", SUPER),
	new KeywordValue("text-top", TEXT_TOP),
	new KeywordValue("text-bottom", TEXT_BOTTOM),
	new KeywordValue("pre", PRE),
	new KeywordValue("nowrap", NOWRAP),
	new KeywordValue("inline", INLINE),
	new KeywordValue("block", BLOCK),
	new KeywordValue("marker", MARKER),
	new KeywordValue("table", TABLE),
	new KeywordValue("inline-table", INLINE_TABLE),
	new KeywordValue("table-column-group", TABLE_COLUMN_GROUP),
	new KeywordValue("table-column", TABLE_COLUMN),
	new KeywordValue("table-row-group", TABLE_ROW_GROUP),
	new KeywordValue("table-header-group", TABLE_HEADER_GROUP),
	new KeywordValue("table-footer-group", TABLE_FOOTER_GROUP),
	new KeywordValue("table-row", TABLE_ROW),
	new KeywordValue("table-cell", TABLE_CELL),
	new KeywordValue("table-caption", TABLE_CAPTION),
	new KeywordValue("inline-block", INLINE_BLOCK),
	new KeywordValue("tree", TREE),
	new KeywordValue("open-quote", OPEN_QUOTE),
	new KeywordValue("close-quote", CLOSE_QUOTE),
	new KeywordValue("no-open-quote", NO_OPEN_QUOTE),
	new KeywordValue("no-close-quote", NO_CLOSE_QUOTE),
	new KeywordValue("disc", DISC),
	new KeywordValue("circle", CIRCLE),
	new KeywordValue("square", SQUARE),
	new KeywordValue("decimal", DECIMAL),
	new KeywordValue("lower-alpha", LOWER_ALPHA),
	new KeywordValue("upper-alpha", UPPER_ALPHA),
	new KeywordValue("lower-latin", LOWER_ALPHA),
	new KeywordValue("upper-latin", UPPER_ALPHA),
	new KeywordValue("lower-roman", LOWER_ROMAN),
	new KeywordValue("upper-roman", UPPER_ROMAN)
    };

    private static Hashtable keywordToValue = new Hashtable();
    private static Hashtable valueToKeyword = new Hashtable();
    static {
	for (int i = 0; i < keywordValues.length; ++i) {
	    KeywordValue entry = keywordValues[i];
	    keywordToValue.put(entry.keyword, entry);
	    if (!valueToKeyword.contains(entry))
		valueToKeyword.put(entry, entry);
	}
    }

    // -----------------------------------------------------------------------

    public static StyleValue fromValueComponents(ValueComponent[] components) {
	switch (components.length) {
	case 0:
	    return null;
	case 1:
	    return fromValueComponent(components[0]);
	default:
	    {
		StyleValue val = new StyleValue();
		val.type = LIST;
		val.list = new StyleValue[components.length];

		for (int i = 0; i < components.length; ++i) {
		    // Don't care about the operator (= syntactic sugar).

		    val.list[i] = fromValueComponent(components[i]);
		    if (val.list[i] == null)
			return null;
		}

		return val;
	    }
	}
    }

    public static StyleValue fromValueComponent(ValueComponent comp) {
	StyleValue val = new StyleValue();

	switch (comp.type) {
	case ValueComponent.NUMBER:
	    val.type = NUMBER;
	    val.number = comp.numericValue;
	    break;
	case ValueComponent.STRING:
	    val.type = STRING;
	    val.string = comp.stringValue;
	    break;
	case ValueComponent.IDENTIFIER:
	    val.type = getKeywordValue(comp.stringValue);
	    if (val.type < 0) {
		val.type = IDENTIFIER;
		val.string = comp.stringValue;
	    }
	    break;

	case ValueComponent.LENGTH:
	    val.number = comp.numericValue;
	    if ("em".equals(comp.unit))
		val.type = LENGTH_EM;
	    else if ("ex".equals(comp.unit))
		val.type = LENGTH_EX;
	    else if ("pt".equals(comp.unit))
		val.type = LENGTH_PT;
	    else if ("px".equals(comp.unit))
		val.type = LENGTH_PX;
	    else if ("in".equals(comp.unit))
		val.type = LENGTH_IN;
	    else if ("cm".equals(comp.unit))
		val.type = LENGTH_CM;
	    else if ("mm".equals(comp.unit))
		val.type = LENGTH_MM;
	    else if ("pc".equals(comp.unit))
		val.type = LENGTH_PC;
	    else
		return null;
	    break;
	case ValueComponent.ANGLE:
	    val.number = comp.numericValue;
	    if ("deg".equals(comp.unit))
		val.type = ANGLE_DEG;
	    else if ("rad".equals(comp.unit))
		val.type = ANGLE_RAD;
	    else if ("grad".equals(comp.unit))
		val.type = ANGLE_GRAD;
	    else
		return null;
	    break;
	case ValueComponent.TIME:
	    val.number = comp.numericValue;
	    if ("ms".equals(comp.unit))
		val.type = TIME_MS;
	    else if ("s".equals(comp.unit))
		val.type = TIME_S;
	    else
		return null;
	    break;
	case ValueComponent.FREQUENCY:
	    val.number = comp.numericValue;
	    if ("hz".equals(comp.unit))
		val.type = FREQUENCY_HZ;
	    else if ("mhz".equals(comp.unit))
		val.type = FREQUENCY_KHZ;
	    else
		return null;
	    break;
	case ValueComponent.PERCENTAGE:
	    val.type = PERCENTAGE;
	    val.number = comp.numericValue;
	    break;
	case ValueComponent.DIMENSION:
	    val.type = DIMENSION;
	    val.number = comp.numericValue;
	    val.string = comp.unit;
	    break;

	case ValueComponent.URI:
	    val.type = URI;
	    val.string = comp.stringValue;
	    break;

	case ValueComponent.FUNCTION:
	    {
		String name = comp.functionName;
		ValueComponent[] args = comp.functionArguments;

		if (name.equals("attr")) {

		    if (args.length != 1 || 
			(args[0].type != ValueComponent.IDENTIFIER &&
			 args[0].type != ValueComponent.STRING))
			return null;

		    val.type = ATTRIBUTE;
		    val.string = args[0].stringValue;

		} else if (name.equals("counter")) {

		    val.list = new StyleValue[1];
		    StyleValue numbering = new StyleValue();
		    val.list[0] = numbering;

		    numbering.type = DECIMAL;
		    switch (args.length) {
		    case 2:
			if (args[1].type == ValueComponent.IDENTIFIER) {
			    numbering.type = 
				getKeywordValue(args[1].stringValue);
			    switch (numbering.type) {
			    case DECIMAL:
			    case LOWER_ALPHA:
			    case UPPER_ALPHA:
			    case LOWER_ROMAN:
			    case UPPER_ROMAN:
				break;
			    default:
				return null;
			    }
			} else {
			    return null;
			}
			/*FALLTHROUGH*/
		    case 1:
			if (args[0].type == ValueComponent.IDENTIFIER) {
			    val.type = COUNTER;
			    val.string = args[0].stringValue;
			} else {
			    return null;
			}
			break;
		    default:
			return null;
		    }

		} else if (name.equals("counters")) {

		    val.list = new StyleValue[2];
		    StyleValue separator = new StyleValue();
		    val.list[0] = separator;
		    StyleValue numbering = new StyleValue();
		    val.list[1] = numbering;

		    numbering.type = DECIMAL;
		    switch (args.length) {
		    case 3:
			if (args[2].type == ValueComponent.IDENTIFIER) {
			    numbering.type = 
				getKeywordValue(args[2].stringValue);
			    switch (numbering.type) {
			    case DECIMAL:
			    case LOWER_ALPHA:
			    case UPPER_ALPHA:
			    case LOWER_ROMAN:
			    case UPPER_ROMAN:
				break;
			    default:
				return null;
			    }
			} else {
			    return null;
			}
			/*FALLTHROUGH*/
		    case 2:
			if (args[0].type == ValueComponent.IDENTIFIER &&
			    args[1].type == ValueComponent.STRING) {
			    val.type = COUNTERS;
			    val.string = args[0].stringValue;

			    separator.type = STRING;
			    separator.string = args[1].stringValue;
			} else {
			    return null;
			}
			break;
		    default:
			return null;
		    }

		} else if (name.equals("rgb")) {
		    int[] rgb = ParseUtil.functionArgumentsToRGB(args);
		    if (rgb == null) 
			return null;

		    val.type = COLOR;
		    val.color = new Color(rgb[0], rgb[1], rgb[2]);

		} else {
		    val.type = FUNCTION;
		    val.string = name;

		    int count = args.length;
		    val.list = new StyleValue[count];
		    for (int i = 0; i < count; ++i) {
			val.list[i] = fromValueComponent(args[i]);
			if (val.list[i] == null)
			    return null;
		    }
		}
	    }
	    break;

	case ValueComponent.HEX_COLOR:
	    val.type = COLOR;
	    val.color = 
	       new Color(comp.hexColor[0], comp.hexColor[1], comp.hexColor[2]);
	    break;

	default:
	    return null;
	}

	return val;
    }

    private static int getKeywordValue(String keyword) {
	KeywordValue entry = (KeywordValue) keywordToValue.get(keyword);
	if (entry == null)
	    return -1;
	else
	    return entry.value;
    }

    public static String getKeyword(int value) {
	KeywordValue probe = new KeywordValue(null, value);

	KeywordValue entry = (KeywordValue) valueToKeyword.get(probe);
	if (entry == null)
	    return null;
	else
	    return entry.keyword;
    }

    public StyleValue asLength() {
	switch (type) {
	case NUMBER:
	    type = StyleValue.LENGTH_PX;
	case LENGTH_PX:
	case LENGTH_IN:
	case LENGTH_CM:
	case LENGTH_MM:
	case LENGTH_PT:
	case LENGTH_PC:
	case LENGTH_EM:
	case LENGTH_EX:
	    return this;
	default:
	    return null;
	}
    }

    public StyleValue asPositiveLength() {
	switch (type) {
	case NUMBER:
	    type = StyleValue.LENGTH_PX;
	case LENGTH_PX:
	case LENGTH_IN:
	case LENGTH_CM:
	case LENGTH_MM:
	case LENGTH_PT:
	case LENGTH_PC:
	case LENGTH_EM:
	case LENGTH_EX:
	    return ((number >= 0)? this : null);
	default:
	    return null;
	}
    }

    public StyleValue asStrictlyPositiveLength() {
	switch (type) {
	case NUMBER:
	    type = StyleValue.LENGTH_PX;
	case LENGTH_PX:
	case LENGTH_IN:
	case LENGTH_CM:
	case LENGTH_MM:
	case LENGTH_PT:
	case LENGTH_PC:
	case LENGTH_EM:
	case LENGTH_EX:
	    return ((number > 0)? this : null);
	default:
	    return null;
	}
    }

    public StyleValue asColor() {
	switch (type) {
	case COLOR:
	    return this;
	case IDENTIFIER:
	    {
		int[] rgb = ParseUtil.nameToRGB(string);
		if (rgb == null)
		    return null;

		type = COLOR;
		string = null;
		color = new Color(rgb[0], rgb[1], rgb[2]);
		return this;
	    }
	default:
	    return null;
	}
    }

    public void resolveURL(URL baseURL) {
	switch (type) {
	case StyleValue.URI:
	    {
		String url = string;

		if (baseURL != null) {
		    try {
			url = new URL(baseURL, string).toExternalForm();
		    } catch (MalformedURLException ignored) {}
		}

		string = url;
	    }
	    break;

	case StyleValue.FUNCTION:
	case StyleValue.LIST:
	    {
		for (int i = 0; i < list.length; ++i) 
		    list[i].resolveURL(baseURL);
	    }
	    break;
	}
    }

    public String stringValue() {
	switch (type) {
	case StyleValue.STRING:
	case StyleValue.IDENTIFIER:
	    return string;

	default:
	    return getKeyword(type);
	}
    }

    public Color colorValue() {
	switch (type) {
	case StyleValue.COLOR:
	    return color;

	case StyleValue.IDENTIFIER:
	    {
		int[] rgb = ParseUtil.nameToRGB(string);
		if (rgb == null)
		    return null;
		else
		    return new Color(rgb[0], rgb[1], rgb[2]);
	    }

	default:
	    return null;
	}
    }
}
