/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
import java.awt.*;
import java.awt.event.*;
import java.awt.print.*;
import javax.swing.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.MenuListener;
import javax.swing.event.MenuEvent;
import fr.pixware.util.*;
import com.xmlmind.awtutil.ImageResource;
import com.xmlmind.awtutil.AWTUtil;
import com.xmlmind.swingutil.*;
import com.xmlmind.gadget.*;
import com.xmlmind.structformat.*;
import com.xmlmind.xmledit.*;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.*;
import com.xmlmind.xmledit.modelimpl.*;
import com.xmlmind.xmledit.stylesheet.*;
import com.xmlmind.xmledit.styledgadget.Style;
import com.xmlmind.xmledit.styledview.*;

public class App implements WorkInProgressObserver, PrintInProgressObserver,
                            DocumentListener, ContextChangeListener, 
                            UndoManagerListener,
                            ActionListener, ItemListener, MenuListener {

    public static final String TITLE = "XMLmind XML Editor M1.3p3";

    // -----------------------------------------------------------------------
    // Internal state
    // -----------------------------------------------------------------------

    private Preferences pref;
    private StructuredFormat[] structuredFormats = new StructuredFormat[0];

    private JFrame frame;
    private Hashtable menus = null;

    private JViewport viewport;
    private boolean docPaneTracksViewportWidth = true;

    private JStyledDocumentPane docPane;
    private JLabel clipboardTag;
    private SubelementPath subelementPath;

    private JLabel invalid;
    private ImageIcon[] invalidIcons = new ImageIcon[4];
    private DocumentChecker docChecker = null;

    private JLabel progress;
    private ImageIcon[] progressIcons = new ImageIcon[5];
    private Timer progressTimer = null;
    private int progressIcon = 1;

    private AsyncCommand asyncCommand = null;
    private boolean cancelAsyncCommand = false;

    private JLabel message;
    private Timer messageTimeout = null;

    private String docFileName = null;
    private boolean needSaveAs = false;
    private boolean needSave = false;

    private StyleSheetInfo[] styleSheets = null;
    private int styleSheetIndex = -1;

    private ContextCommandInfo commandInfo = new ContextCommandInfo();

    private UndoManager undoManager;
    private String undoDescription;
    private UndoHelper undoHelper;

    private String lastConvertTag = null;

    private SearchTextDialog.Info searchInfo = null;
    private SearchReplaceText searchReplaceText = null;

    private CreateDialog createDialog = null;
    private TagChooserDialog tagChooserDialog = null;
    private SearchTextDialog searchTextDialog = null;
    private ReplaceTextDialog replaceTextDialog = null;
    private AttributesTool attributesTool = null;
    private ValidityDiagnosticsTool validityDiagnosticsTool = null;
    private SpellCheckerDialog spellCheckerDialog = null;
    private OptionsDialog optionsDialog = null;
    private StructuredFormatsDialog structuredFormatsDialog = null;

    private static final String FORMAT_PROPERTY = 
        "com.xmlmind.xmleditapp.App:StructuredFormat";

    private static final String[] editMenuItemIconNames = {
	"cut.gif", "copy.gif", 
	"pastebefore.gif", "paste.gif", "pasteafter.gif", 
	"delete.gif", "replace.gif", 
	"insbefore.gif", "insert.gif", "insafter.gif", 
	"split.gif", "join.gif", 
	"convert.gif"
    };
    private static final String[] editMenuItemNames = {
	"edit.cut", "edit.copy", 
	"edit.pasteBefore", "edit.paste", "edit.pasteAfter", 
	"edit.delete", "edit.replace", 
	"edit.insertBefore", "edit.insert", "edit.insertAfter", 
	"edit.split", "edit.join", 
	"edit.convert"
    };
    private static final int[] editMenuItemCommands = {
	ContextCommandInfo.CUT, ContextCommandInfo.COPY, 
	ContextCommandInfo.PASTE_BEFORE, ContextCommandInfo.PASTE, 
	ContextCommandInfo.PASTE_AFTER, 
	ContextCommandInfo.DELETE, ContextCommandInfo.REPLACE, 
	ContextCommandInfo.INSERT_BEFORE, ContextCommandInfo.INSERT, 
	ContextCommandInfo.INSERT_AFTER, 
	ContextCommandInfo.SPLIT, ContextCommandInfo.JOIN, 
	ContextCommandInfo.CONVERT
    };
    private static final String[] insertFileMenuItemNames = {
	"edit.insertFile.pasteFileBefore", "edit.insertFile.pasteFile", 
	"edit.insertFile.pasteFileAfter", 
    };
    private static final int[] insertFileMenuItemCommands = {
	ContextCommandInfo.PASTE_BEFORE, ContextCommandInfo.PASTE, 
	ContextCommandInfo.PASTE_AFTER
    };

    private static final Class[] noParamTypes = new Class[0];
    private static final Object[] noArgs = new Object[0];

    private static ResourceBundle resourceBundle = null;
    static {
	// JDK1.3.1 bug? this should be automatically done by Java.
	String language = System.getProperty("user.language");
        if (language != null)
	    Locale.setDefault(new Locale(language, ""));

	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.AppRes");
	} catch (MissingResourceException ignored) {}
    }

    // -----------------------------------------------------------------------
    // Initialization
    // -----------------------------------------------------------------------

    public static void main(String[] args) {
	if (args.length > 1 || 
	    (args.length == 1 && !(new File(args[0])).isFile())) {
	    System.out.println(msg("usage"));
	    System.exit(1);
	}
	final String openFileName = (args.length == 0)? null : args[0];

	final App app = new App();

	SwingUtilities.invokeLater(new Runnable() {
	    public void run() {
		if (openFileName == null)
		    app.close();
		else
		    app.open(openFileName);
	    }
	});
    }

    public App() {
	// Load user preferences ---

	pref = Preferences.load();

	// Create main window ---

	frame = new JFrame(TITLE);
	frame.setIconImage(ImageResource.get(getClass(), "appicon.gif"));
   	frame.addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
		App.this.quit();
   	    }
	});
	initGlassPane(frame.getGlassPane());

        // workArea is a JPanel with a BorderLayout.
	Container workArea = frame.getContentPane(); 

	docPane = new JStyledDocumentPane();
	//docPane.setBorder(new javax.swing.border.EmptyBorder(10,10,10,10));
	docPane.addContextChangeListener(this);
	// docPane needs to manage Ctr+Tab and Ctrl+Shift+Tab too.
	FocusManager.setCurrentManager(new JDocumentPane.FocusManager());
	applyMakeChoice();

	registerKeyboardActions();
	setDefaultStyle(pref);

	JScrollPane scroller = new JScrollPane(docPane);
	workArea.add(scroller, BorderLayout.CENTER);

	viewport = scroller.getViewport();
	configureViewport(viewport);

	JMenuBar bar = new JMenuBar();
	frame.setJMenuBar(bar);
	addActions(bar);

	initUndoSupport();

	updateFileMenu();
	enableMenus();

	JToolBar toolBar = new JToolBar();
	workArea.add(toolBar, BorderLayout.NORTH);
	initToolBar(toolBar);

	JPanel bottomBars = new JPanel(new GridBagLayout());
	workArea.add(bottomBars, BorderLayout.SOUTH);

	GridBagConstraints constraints = new GridBagConstraints();
	constraints.insets.top = 1;
	constraints.insets.bottom = 1;
	constraints.insets.left = 1;
	constraints.insets.right = 1;
	constraints.weightx = 1;
	constraints.fill = GridBagConstraints.HORIZONTAL;
	constraints.gridx = 0;
	constraints.gridy = 0;

	JPanel contextBar = new JPanel(new GridBagLayout());
	bottomBars.add(contextBar, constraints);
	initContextBar(contextBar);
	++constraints.gridy;

	JPanel statusBar = new JPanel(new GridBagLayout());
	bottomBars.add(statusBar, constraints);
	initStatusBar(statusBar);

	frame.setBounds(pref.geometry.x, pref.geometry.y, 
			pref.geometry.width, pref.geometry.height);
	frame.show();

	// Add structured formats found in the user preferences file.

	// XMLFormat must always be registered as the first format.
	XMLFormat xmlFormat = new XMLFormat();
	xmlFormat.setGuessingElementContent(pref.guessElementContent);
	xmlFormat.setIndentation(pref.indentIfDocType, pref.indentIfNoDocType, 
				 pref.indentation, pref.maxLineLength);
	addStructuredFormat(xmlFormat);

	addNonXMLFormats();
    }

    private void initGlassPane(final Component glassPane) {
	glassPane.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

	glassPane.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		Dimension corner = invalid.getSize();
		Dimension appWin = glassPane.getSize();

		if (e.getX() >= corner.width && 
		    e.getY() >= appWin.height - corner.height)
		    cancelAsyncCommand = true;
	    }
	});
	glassPane.addMouseMotionListener(new MouseMotionAdapter() {});
	glassPane.addKeyListener(new KeyAdapter() {});
    }

    private void registerKeyboardActions() {
	docPane.registerKeyboardAction(
	    this, "insertTextAfter", 
	    KeyStroke.getKeyStroke(KeyEvent.VK_INSERT, 0), 
	    JComponent.WHEN_FOCUSED);

	docPane.registerKeyboardAction(
	    this, "insertTextBefore", 
	    KeyStroke.getKeyStroke(KeyEvent.VK_INSERT, KeyEvent.SHIFT_MASK), 
	    JComponent.WHEN_FOCUSED);

	docPane.registerKeyboardAction(
	    this, "insertLikeSelectedAfter", 
	    KeyStroke.getKeyStroke(KeyEvent.VK_INSERT, AWTUtil.KEY_MODIFIER), 
	    JComponent.WHEN_FOCUSED);

	docPane.registerKeyboardAction(
	    this, "insertLikeSelectedBefore", 
	    KeyStroke.getKeyStroke(KeyEvent.VK_INSERT, 
				   AWTUtil.KEY_MODIFIER|KeyEvent.SHIFT_MASK), 
	    JComponent.WHEN_FOCUSED);

	docPane.registerKeyboardAction(
	    this, "splitSelected", 
	    KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), 
	    JComponent.WHEN_FOCUSED);
    }

    private void configureViewport(final JViewport viewport) {
	viewport.addComponentListener(new ComponentAdapter() {
	    private Dimension dim = new Dimension();

	    public void componentResized(ComponentEvent e) {
		if (docPaneTracksViewportWidth) {
		    viewport.getSize(dim);
		    RootGadget root = docPane.getRoot();
		    root.layout(dim.width, -1);
		    root.repaint();
		}
	    }
	});

	//viewport.setScrollMode(JViewport.BACKINGSTORE_SCROLL_MODE);
	//JDK1.3: Better algorithm that doesn't use a backing store.
	// (the default for the JDK1.3.1).
	viewport.setScrollMode(JViewport.BLIT_SCROLL_MODE);

	viewport.addChangeListener(new ChangeListener() {
	    public void stateChanged(ChangeEvent e) {
		Caret caret = docPane.getCaret();

		if (caret != null && caret.getMark() == null)
		    docPane.ensurePointIsVisible();
	    }
        });
    }

    private void initToolBar(JToolBar toolBar) {
	toolBar.setFloatable(false);
	// Prevent the toolbar from capturing the keyboard focus.
	toolBar.setRequestFocusEnabled(false);

	addToolBarAction(toolBar, 
			 "file.newFromTemplate", "new.gif").setEnabled(true);
	addToolBarAction(toolBar, "file.open", "open.gif").setEnabled(true);
	addToolBarAction(toolBar, "file.save", "save.gif");

	toolBar.addSeparator();

	addToolBarAction(toolBar, "edit.undo", "undo.gif");
	addToolBarAction(toolBar, "edit.redo", "redo.gif");

	for (int i = 0; i < editMenuItemNames.length; ++i) 
	    addToolBarAction(toolBar, editMenuItemNames[i],
			     editMenuItemIconNames[i]);

	toolBar.addSeparator();

	addToolBarAction(toolBar, "select.selectParent", "selparent.gif");
	addToolBarAction(toolBar, "select.selectChild", "selchild.gif");
	addToolBarAction(toolBar, "select.selectPreceeding", "selprev.gif");
	addToolBarAction(toolBar, "select.selectFollowing", "selnext.gif");
    }

    private Action addToolBarAction(JToolBar toolBar, String name, 
				    String iconName) {
	Action action = MenuUtil.getAction(menus, name);

	JButton button = toolBar.add(action);
	button.setRequestFocusEnabled(false);
	button.setText("");
	button.setIcon(
	    new ImageIcon(ImageResource.get(getClass(), iconName)));
	button.setMargin(new Insets(1, 1, 1, 1));

	String tooltip = formatToTooltip(name);

	JMenuItem menuItem = MenuUtil.getMenuItem(menus, name);
	String accelerator = MenuUtil.getAcceleratorText(menuItem);
	if (accelerator != null) {
	    StringBuffer buffer = new StringBuffer();

	    buffer.append(tooltip);
	    buffer.append(" [");
	    buffer.append(accelerator);
	    buffer.append(']');

	    tooltip = buffer.toString();
	}

	button.setToolTipText(tooltip);

	action.setEnabled(false);

	return action;
    }

    private String formatToTooltip(String name) {
	String message = msg(name, "", "", "", "", "");
	int length = message.length();
	StringBuffer buffer = new StringBuffer();

	for (int i = 0; i < length; ++i) {
	    char c = message.charAt(i);

	    if (Character.isLetterOrDigit(c) || Character.isSpaceChar(c))
		buffer.append(c);
	}

	return buffer.toString().trim();
    }

    private void initContextBar(JPanel contextBar) {
	Image clipboardIcon = ImageResource.get(getClass(), "clipboard.gif");
	Image selectedIcon = ImageResource.get(getClass(), "selected.gif");
	Color bg = contextBar.getBackground();

	GridBagConstraints constraints = new GridBagConstraints();
	constraints.insets.top = 1;
	constraints.insets.bottom = 1;
	constraints.insets.left = 1;
	constraints.insets.right = 1;
	constraints.gridx = 0;
	constraints.gridy = 0;

	JLabel icon = new JLabel(new ImageIcon(clipboardIcon));
	constraints.fill = GridBagConstraints.VERTICAL;
	contextBar.add(icon, constraints);
	++constraints.gridx;

	clipboardTag = new JLabel("xxxxxxxxxx");
	clipboardTag.setBorder(new ThinBorder(bg, /*raised*/ false));
	Dimension minSize = clipboardTag.getPreferredSize();
	clipboardTag.setMinimumSize(minSize);
	clipboardTag.setPreferredSize(clipboardTag.getPreferredSize());
	clipboardTag.setText("");
	constraints.fill = GridBagConstraints.HORIZONTAL;
	contextBar.add(clipboardTag, constraints);
	++constraints.gridx;

	icon = new JLabel(new ImageIcon(selectedIcon));
	constraints.fill = GridBagConstraints.VERTICAL;
	contextBar.add(icon, constraints);
	++constraints.gridx;

	subelementPath = new SubelementPath();
	subelementPath.setFont(clipboardTag.getFont());
	subelementPath.setForeground(clipboardTag.getForeground());
	bg = clipboardTag.getBackground();
	subelementPath.setBackground(bg);
	subelementPath.setSeparatorColor(bg.darker().darker());
	subelementPath.setSeparatorRelief(SubelementPath.RELIEF_RAISED);
	subelementPath.setBorderRelief(SubelementPath.RELIEF_SUNKEN);
	constraints.weightx = 1;
	constraints.fill = GridBagConstraints.HORIZONTAL;
	contextBar.add(subelementPath, constraints);
	subelementPath.addItemListener(this);
	subelementPath.addActionListener(this);
    }

    private void initStatusBar(JPanel statusBar) {
	for (int i = 0; i < progressIcons.length; ++i) {
	    Image image = ImageResource.get(getClass(), 
					    "progress" + i + ".gif");
	    progressIcons[i] = new ImageIcon(image);
	}
	for (int j = 0; j < invalidIcons.length; ++j) {
	    Image image = ImageResource.get(getClass(), 
					    "invalid" + j + ".gif");
	    invalidIcons[j] = new ImageIcon(image);
	}

	Color bg = statusBar.getBackground();

	GridBagConstraints constraints = new GridBagConstraints();
	constraints.insets.top = 1;
	constraints.insets.bottom = 1;
	constraints.insets.left = 1;
	constraints.insets.right = 1;
	constraints.fill = GridBagConstraints.VERTICAL;
	constraints.gridx = 0;
	constraints.gridy = 0;

	invalid = new JLabel(invalidIcons[0]);
	invalid.setBorder(new ThinBorder(bg, /*raised*/ false));
	invalid.setToolTipText("");
	statusBar.add(invalid, constraints);
	++constraints.gridx;

	// Not a real button for esthetics reasons.
	invalid.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		checkValidity();
	    }
	});

	progress = new JLabel(progressIcons[0]);
	progress.setBorder(new ThinBorder(bg, /*raised*/ false));
	statusBar.add(progress, constraints);
	++constraints.gridx;

	message = new JLabel("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
	message.setBorder(new ThinBorder(bg, /*raised*/ false));
	Dimension minSize = message.getPreferredSize();
	message.setMinimumSize(minSize);
	message.setPreferredSize(message.getPreferredSize());
	message.setText("");
	constraints.weightx = 1;
	constraints.fill = GridBagConstraints.HORIZONTAL;
	statusBar.add(message, constraints);
    }

    private void addNonXMLFormats() {
	for (int k = 0; k < pref.structuredFormats.length; ++k) {
	    // An entry is: className ?arg? ... ?arg?
	    String entry = pref.structuredFormats[k];
	    if (entry == null)
		break;

	    String[] args = Preferences.splitArguments(entry);
	    if (args.length == 0)
		// Empty entry.
		return;

	    StructuredFormat format = null;

	    try {
		format = addStructuredFormat(args[0]);
	    } catch (Exception e) {
		String reason = e.getClass().getName();
		if (e.getMessage() != null)
		    reason += ": " + e.getMessage();

		System.err.println(msg("cannotLoadFormat", args[0], reason));
	    }

	    if (format != null && args.length > 1) {
		String[] params = new String[args.length-1];
		System.arraycopy(args, 1, params, 0, params.length);

		try {
		    format.setParameters(params);
		} catch (Exception e) {
		    String reason = e.getMessage();
		    if (reason == null)
			reason = e.getClass().getName();

		    System.err.println(
			msg("cannotConfigureFormat", args[0], reason));
		}
	    }
	}
    }

    private void initUndoSupport() {
	undoManager = new UndoManager();
	undoManager.setUndoMaxCount(pref.maxUndo);
	undoManager.addUndoManagerListener(this);

	undoDescription = null;

	undoHelper = new UndoHelper();

	// Undo ---

	undoHelper.undoAction = MenuUtil.getAction(menus, "edit.undo");
	undoHelper.undoMenuItem = MenuUtil.getMenuItem(menus, "edit.undo");

	String pattern;
	try {
	    pattern = resourceBundle.getString("edit.undo");
	} catch (Exception e) {
	    pattern = "edit.undo";
	}
	undoHelper.undoFormat = new MessageFormat(pattern);

	undoHelper.undoMessage = undoHelper.formatUndoMessage("");
	undoHelper.undoTypingDescription = msg("undoTypingDescription");
	undoHelper.undoTypingMessage = 
	    undoHelper.formatUndoMessage(undoHelper.undoTypingDescription);
	undoHelper.undoAttributeChangeDescription = 
	    msg("undoAttributeChangeDescription");
	undoHelper.undoAttributeChangeMessage = 
	    undoHelper.formatUndoMessage(
		undoHelper.undoAttributeChangeDescription);

	// Redo ---

	undoHelper.redoAction = MenuUtil.getAction(menus, "edit.redo");
	undoHelper.redoMenuItem = MenuUtil.getMenuItem(menus, "edit.redo");

	try {
	    pattern = resourceBundle.getString("edit.redo");
	} catch (Exception e) {
	    pattern = "edit.redo";
	}
	undoHelper.redoFormat = new MessageFormat(pattern);

	undoHelper.redoMessage = undoHelper.formatRedoMessage("");
    }

    // -----------------------------------------------------------------------
    // File menu
    // -----------------------------------------------------------------------

    // --------------------------------------
    // New
    // --------------------------------------

    public void newFromTemplate() {
	if (createDialog == null)
	    createDialog = new CreateDialog(frame);

	Vector templateNames = Preferences.listToVector(pref.templateNames);
 	String templateName = createDialog.getTemplateName(templateNames);

	// User may have deleted useless template names.
	Preferences.updateFromVector(pref.templateNames, templateNames);
	    
	if (templateName == null)
	    return;

	URL url;
	try {
	    url = new URL(templateName);
	} catch (MalformedURLException cannotHappen) {
	    cannotHappen.printStackTrace();
	    url = null;
	}
	if (url != null)
	    create(url, /*useAsIs*/ true);
    }

    public void openAsTemplate() {
	String template = pref.openedFileNames[0];
	if (template == null)
	    template = getUntitledFileName("xml");

	String fileName = FileChooserUtil.getOpenFile(frame, null, template, 
                                                      /*filtered*/ false);
	if (fileName == null)
	    return;

	URL url = FileUtil.fileToURL(fileName);
	if (url != null)
	    create(url, /*useAsIs*/ false);
    }

    public void create(URL url, boolean useAsIs) {
	if (!checkNeedSave())
	    return;

	asyncCommand = new CreateCommand(url, useAsIs);
	asyncCommand.start();
    }

    private class CreateCommand extends OpenCommand {
	private boolean useAsIs;

	private String urlName;
	private String templateName;

	public CreateCommand(URL url, boolean useAsIs) {
	    super(url);
	    this.useAsIs = useAsIs;
	}

	protected boolean init() {
	    waitCursor(true, true);
	    urlName = url.toExternalForm();
	    templateName = urlBaseName(urlName);
	    setMessage(msg("loadingTemplate", templateName) + 
		       msg("clickToCancel"), false);

	    return true;
	}

	protected void finish() {
	    if (doc != null) {
		syncSetMessage(msg("loadingTemplate", templateName));

		setDocument(doc, /*asTemplate*/ !useAsIs);

		docFileName = getUntitledFileName(urlExtension(urlName));
		needSaveAs = true;
		needSave = false;
		updateTitle();
		enableMenus();
	    }

	    setMessage("", false);
	    waitCursor(false, true);
	    
	    if (openError != null) {
		if (useAsIs)
		    pref.removeTemplateName(urlName);

		JOptionPane.showMessageDialog(
		    frame, 
		    msg("cannotOpenFile", urlName, openError), 
		    msg("error"), JOptionPane.ERROR_MESSAGE);
	    } else if (openCanceled) {
		setMessage(msg("loadingTemplateCanceled"), true);
	    } else {
		if (useAsIs)
		    pref.addTemplateName(urlName);
	    }
	}
    }

    private static String urlBaseName(String urlName) {
	int slash = urlName.lastIndexOf('/');
	String name;
	if (slash < 0)
	    name = urlName;
	else
	    name = urlName.substring(slash+1);
	return name;
    }

    private static String urlExtension(String urlName) {
	int slash = urlName.lastIndexOf('/');
	if (slash < 0) 
	    slash = 0;
	else 
	    ++slash;

	int dot = urlName.lastIndexOf('.');
	if (dot <= slash)
            // '.profile' has no extension!
	    return "";
	else
	    return urlName.substring(dot+1);
    }

    private String getUntitledFileName(String extension) {
	return (new File(msg("untitled") + '.' + extension)).getAbsolutePath();
    }

    // --------------------------------------
    // Open
    // --------------------------------------

    public void open() {
	String template = pref.openedFileNames[0];
	if (template == null)
	    template = getUntitledFileName("xml");

	String fileName = FileChooserUtil.getOpenFile(frame, null, template, 
                                                      /*filtered*/ false);
	if (fileName == null)
	    return;

	if (!checkNeedSave())
	    return;

	open(fileName);
    }

    public void open(String fileName) {
	asyncCommand = new OpenCommand(fileName);
	asyncCommand.start();
    }

    private class OpenCommand extends AsyncCommand {
	protected String fileName;
	protected File file;
	protected URL url;

	protected String shortName;
	protected boolean openCanceled;
	protected String openError;
	protected DocumentModel doc;

	public OpenCommand(String fileName) {
	    this.fileName = fileName;
	    file = new File(fileName);
	    url = FileUtil.fileToURL(file);
	}

	protected OpenCommand(URL url) {
	    fileName = null;
	    file = null;
	    this.url = url;
	}

	protected boolean init() {
	    if (url == null)
		return false;

	    waitCursor(true, true);
	    shortName = FileChooserUtil.truncateFileName(fileName);
	    setMessage(msg("loadingFile", shortName) + msg("clickToCancel"), 
		       false);

	    return true;
	}

	protected void exec() {
	    try {
		doc = loadDocument(url);
	    } catch (WorkCanceledException e) {
		openCanceled = true;
	    } catch (StructuredFormatException e) {
		openError = msg("documentError", e.getLocation(), 
				new Integer(e.getLineNumber()), 
				new Integer(e.getColumnNumber()), 
				e.getMessage()); 
	    } catch (FileNotFoundException e) {
		openError = msg("fileNotFound", e.getMessage());
	    } catch (IOException e) {
		openError = e.getMessage();
		if (openError == null)
		    openError = e.getClass().getName();
	    }
	}

	protected void finish() {
	    if (doc != null) {
		syncSetMessage(msg("loadingFile", shortName));

		setDocument(doc, /*asTemplate*/ false);

		needSaveAs = !file.canWrite();
		needSave = false;
		docFileName = fileName;
		updateTitle();
		pref.addOpenedFileName(fileName);
		updateFileMenu();
		enableMenus();
	    }

	    setMessage("", false);
	    waitCursor(false, true);
	    
	    if (openError != null) {
		pref.removeOpenedFileName(fileName);
		updateFileMenu();

		JOptionPane.showMessageDialog(
		    frame, 
		    msg("cannotOpenFile", url.toExternalForm(), openError), 
		    msg("error"), JOptionPane.ERROR_MESSAGE);
	    } else if (openCanceled) {
		setMessage(msg("loadingFileCanceled"), true);
	    } else {
		if (needSaveAs)
		    setMessage(msg("fileIsReadOnly", shortName), true);
	    }
	}
    }

    // -----------------------------------------------------------------------

    private DocumentModel loadDocument(URL url) 
	throws IOException, StructuredFormatException {
	URLConnection connection = url.openConnection();
	connection.connect();

	StructuredFormat structuredFormat = null;
	String contentType = connection.getContentType();
	if (contentType != null)
	    structuredFormat = findStructuredFormatFromType(contentType);
	if (structuredFormat == null)
	    structuredFormat = 
		findStructuredFormatFromExtension(url.toExternalForm());
	if (structuredFormat == null)
	    // Fallback: XMLFormat (always first format)
	    structuredFormat = structuredFormats[0];

	String contentEncoding = connection.getContentEncoding();
	String javaEncoding = null;
	if (contentEncoding != null)
	    javaEncoding = XMLUtil.ianaToJavaEncoding(contentEncoding);
	if (javaEncoding == null)
	    javaEncoding = null;

	Document doc = new Document();
	InputStream in = connection.getInputStream();
	try {
	    structuredFormat.loadDocument(in, url, javaEncoding, doc);
	} finally {
	    in.close();
	}

	// Save structuredFormat as a document property in order to
	// use it again when saving a modified document.
	doc.putDocumentProperty(FORMAT_PROPERTY, structuredFormat);

	return doc;
    }

    private StructuredFormat findStructuredFormatFromType(String type) {
	// Search for most specific to less specific (XMLFormat at index 0).
	for (int i = structuredFormats.length-1; i >= 0; --i) {
	    StructuredFormat format = structuredFormats[i];

	    String mimeType = format.getMimeType();
	    if (mimeType == null)
		return format;
	}
	return null;
    }

    private StructuredFormat findStructuredFormatFromExtension(String path) {
	// Search for most specific to less specific (XMLFormat at index 0).
	for (int i = structuredFormats.length-1; i >= 0; --i) {
	    StructuredFormat format = structuredFormats[i];

	    String[] extensions = format.getFileNameExtensions();
	    if (extensions == null)
		// Any extension is OK.
		return format;

	    for (int j = 0; j < extensions.length; ++j) {
		if (hasExtension(path, extensions[j]))
		    return format;
	    }
	}
	return null;
    }

    private static final boolean hasExtension(String path, String extension) {
	int pos = path.lastIndexOf('.');
	return (pos >= 0 && 
		path.regionMatches(/*ignoreCase*/ true, pos+1, 
				   extension, 0, extension.length()));
    }

    // -----------------------------------------------------------------------

    private void setDocument(DocumentModel doc, boolean asTemplate) {
	if (docPane.getDocument() != null) {
	    if (tagChooserDialog != null)
		tagChooserDialog.setTagHistory(null);

	    if (attributesTool != null)
		attributesTool.setDocument(null);

	    doCheckValidity(null);

	    undoManager.setDocument(null);

	    lastConvertTag = null;

	    docPane.setDocument(null);
	    refreshDocPane();
	}

	if (doc != null) {
	    if (tagChooserDialog != null && doc.getDocumentType() == null)
		tagChooserDialog.setTagHistory(
		    DocumentUtil.getAllElementTypeNames(doc));

	    if (asTemplate) {
		Document doc2 = new Document();
		DocumentUtil.copyTemplateDocument(doc, doc2, 
						  makeChoiceOptions());
		doc = doc2;
	    }

	    if (!installStyleSheets(doc) || pref.ignoreStyleSheets) 
		doSetStyle(-1);
	    else
		doSetStyle(0);

	    if (attributesTool != null)
		// Install doc asap to make the attributesTool ready for the
		// initial context change.
		attributesTool.setDocument(doc);

	    docPane.setDocument(doc);
	    refreshDocPane();

	    doc.addDocumentListener(this, /*hasData*/ false);

	    undoManager.setDocument(doc);

	    doCheckValidity(doc);
	}
    }

    private void refreshDocPane() {
	docPane.refresh();
	viewport.setViewPosition(new Point(0, 0));
	viewport.repaint();
    }

    private boolean installStyleSheets(DocumentModel doc) {
	StyleSheetInfo[] prop = 
	    (StyleSheetInfo[]) doc.getDocumentProperty("xml-stylesheet");
	StyleSheetInfo[] list = null;
	int count = 0;
	int firstNonAlternate = -1;

	if (prop != null) {
	    list = new StyleSheetInfo[prop.length];

	    for (int i = 0; i < prop.length; ++i) {
		StyleSheetInfo info = prop[i];

		if (info.type.equals("text/css") &&
		    (info.media == null || info.media.equals("screen"))) {
		    if (!info.alternate && firstNonAlternate < 0)
			firstNonAlternate = count;
		    list[count++] = info;
		}
	    }
	}

	if (count == 0) {
	    uninstallStyleSheets();
	    return false;
	} 

	if (count > 10)
	    count = 10;
	if (count == list.length) {
	    styleSheets = list;
	} else {
	    styleSheets = new StyleSheetInfo[count];
	    System.arraycopy(list, 0, styleSheets, 0, count);
	}
	styleSheetIndex = -1;

	if (firstNonAlternate < 0)
	    firstNonAlternate = 0;
	if (firstNonAlternate != 0) {
	    StyleSheetInfo swap = list[firstNonAlternate];
	    if (firstNonAlternate < styleSheets.length)
		styleSheets[firstNonAlternate] = styleSheets[0];
	    styleSheets[0] = swap;
	}
	updateStyleMenu();

	return true;
    }

    private void uninstallStyleSheets() {
	styleSheets = null;
	styleSheetIndex = -1;
	updateStyleMenu();
    }

    private void updateStyleMenu() {
	JMenu menu = MenuUtil.getMenu(menus, "style");
	// [0] noStyle, [1] -
	for (int i = menu.getItemCount() - 1; i >= 1; --i) 
	    menu.remove(i);

	boolean hasStyleSheets = (styleSheets != null);
	if (hasStyleSheets) {
	    menu.addSeparator();

	    for (int i = 0; i < styleSheets.length; ++i) {
		StyleSheetInfo info = styleSheets[i];
		String title = info.title;
		if (title == null) {
		    title = info.url;
		    int pos = title.lastIndexOf('/');
		    if (pos >= 0 && pos < title.length()-1) 
			title = title.substring(pos+1);
		    pos = title.lastIndexOf('.');
		    if (pos > 0)
			title = title.substring(0, pos);
		}
		
		JMenuItem item = menu.add(new AppAction("setStyle" + i));
		item.setText(title);
	    }
	}
    }

    // --------------------------------------
    // Close
    // --------------------------------------

    public void close() {
	if (!checkNeedSave())
	    return;

	setDocument(null, false);

	docFileName = null;
	needSaveAs = false;
	needSave = false;
	updateTitle();
	enableMenus();
	uninstallStyleSheets();
	doSetStyle(-1);
	setMessage("", false);
    }

    // --------------------------------------
    // Save
    // --------------------------------------

    public void save() {
	if (needSave) {
	    if (needSaveAs) {
		saveAs();
	    } else {
		backup(docFileName);
		save(docFileName);
	    }
	} else {
	    setMessage(msg("noChangesToSave"), true);
	}
    }

    private boolean backup(String fileName) {
	File file = new File(fileName);

	if (!file.exists()) 
	    return false;

	File backupFile = new File(fileName + '~');

	if (backupFile.exists())
	    backupFile.delete();
	return file.renameTo(backupFile);
    }

    public void save(String fileName) {
	asyncCommand = new SaveCommand(fileName, /*saveAs*/ false);
	asyncCommand.start();
    }

    private class SaveCommand extends AsyncCommand {
	private String fileName;
	private boolean saveAs;

	private boolean saveCanceled;
	private String saveError;

	public SaveCommand(String fileName, boolean saveAs) {
	    this.fileName = fileName;
	    this.saveAs = saveAs;
	}

	protected boolean init() {
	    waitCursor(true, true);
	    String shortName = FileChooserUtil.truncateFileName(fileName);
	    setMessage(msg("savingFile", shortName), false);
	    return true;
	}

	protected void exec() {
	    try {
		saveDocument(fileName);
	    } catch (WorkCanceledException e) {
		saveCanceled = true;
	    } catch (Exception e) {
		saveError = e.getMessage();
		if (saveError == null)
		    saveError = e.getClass().getName();
	    }
	}

	protected void finish() {
	    setMessage("", false);
	    waitCursor(false, true);

	    if (saveError != null) {
		JOptionPane.showMessageDialog(
		    frame, msg("cannotSaveFile", fileName, saveError), 
		    msg("error"), JOptionPane.ERROR_MESSAGE);
	    } else if (saveCanceled) {
		setMessage(msg("savingFileCanceled"), true);
	    } else {
		doCheckValidity(docPane.getDocument());

		needSave = false;
		if (saveAs) {
		    needSaveAs = false;
		    docFileName = fileName;
		    pref.addOpenedFileName(fileName);
		    updateFileMenu();
		}
		updateTitle();
		MenuUtil.getAction(menus, "file.save").setEnabled(false);
	    }
	}
    }

    private void saveDocument(String fileName) 
	throws IOException, StructuredFormatException {
	DocumentModel doc = docPane.getDocument();
	StructuredFormat structuredFormat = (StructuredFormat) 
	    doc.getDocumentProperty(FORMAT_PROPERTY);

	String encoding;
	if (structuredFormat == structuredFormats[0])
	    // XMLFormat is always the first format.
	    encoding = pref.encoding;
	else
	    encoding = null;

	OutputStream out = new FileOutputStream(fileName);
	try {
	    structuredFormat.saveDocument(
		doc, out, FileUtil.fileToURL(fileName), encoding);
	} finally {
	    out.close();
	}
    }

    public void saveAs() {
	String fileName = 
	    FileChooserUtil.getSaveFile(frame, null, docFileName, 
                                        /*filtered*/ false);
	if (fileName != null) 
	    saveAs(fileName);
    }

    public void saveAs(String fileName) {
	asyncCommand = new SaveCommand(fileName, /*saveAs*/ true);
	asyncCommand.start();
    }

    // --------------------------------------
    // Print
    // --------------------------------------

    public void pageSetup() {
	PrinterJob printJob = PrinterJob.getPrinterJob();

	pref.pageFormat = printJob.pageDialog(pref.pageFormat);
    }

    public void print() {
	PrinterJob printJob = PrinterJob.getPrinterJob();

	GadgetPrinter printer = new GadgetPrinter();

	PageFormat pageFormat = printJob.validatePage(pref.pageFormat);
	printJob.setPrintable(printer, pageFormat);

	if (!printJob.printDialog())
	    return;

	asyncCommand = new PrintCommand(printJob, printer, pageFormat);
	asyncCommand.start();
    }

    private class PrintCommand extends AsyncCommand {
	private PrinterJob printJob;
	private GadgetPrinter printer;
	private PageFormat pageFormat;

	private String printingMessage;
	private StyleSheet savedStyleSheet;
	private boolean printCanceled;
	private String printError;

	public PrintCommand(PrinterJob printJob, GadgetPrinter printer,
			    PageFormat pageFormat) {
	    this.printJob = printJob;
	    this.printer = printer;
	    this.pageFormat = pageFormat;
	}

	protected boolean init() {
	    waitCursor(true, true);

	    SubelementModel selection = 
		docPane.getSubelementSelection().getSubelement();
	    if (selection == null) {
		String shortName = 
		    FileChooserUtil.truncateFileName(docFileName);
		printingMessage = msg("printingFile", shortName);
	    } else {
		printingMessage = msg("printingSelectedElement");
	    }
	    syncSetMessage(printingMessage);

	    savedStyleSheet = setPrintStyleSheet();

	    printer.setObserver(App.this);
	    printer.setScreenResolution(pref.screenResolutionWhenPrinting);

	    Font font = docPane.getStyledViewFactory().getDefaultStyle().font;

	    boolean hasHeader = false;
	    if (pref.headerBegin != null) {
		printer.setHeaderBegin(
		    substituteHeaderVariables(pref.headerBegin));
		hasHeader = true;
	    }
	    if (pref.headerMiddle != null) {
		printer.setHeaderMiddle(
		    substituteHeaderVariables(pref.headerMiddle));
		hasHeader = true;
	    }
	    if (pref.headerEnd != null) {
		printer.setHeaderEnd(
		    substituteHeaderVariables(pref.headerEnd));
		hasHeader = true;
	    }
	    if (hasHeader) {
		printer.setHeaderFont(font);
		printer.setHeaderColor(pref.headerColor);
		printer.setHeaderUnderlined(pref.underlineHeader);
	    }

	    hasHeader = false;
	    if (pref.footerBegin != null) {
		printer.setFooterBegin(
		    substituteHeaderVariables(pref.footerBegin));
		hasHeader = true;
	    }
	    if (pref.footerMiddle != null) {
		printer.setFooterMiddle(
		    substituteHeaderVariables(pref.footerMiddle));
		hasHeader = true;
	    }
	    if (pref.footerEnd != null) {
		printer.setFooterEnd(
		    substituteHeaderVariables(pref.footerEnd));
		hasHeader = true;
	    }
	    if (hasHeader) {
		printer.setFooterFont(font);
		printer.setFooterColor(pref.footerColor);
		printer.setFooterOverlined(pref.overlineFooter);
	    }

	    DocumentView docView = docPane.getDocumentView();
	    Gadget printedGadget = docView;
	    if (selection != null) {
		SubelementView view = docView.getSubelementView(selection);
		if (view != null) {
		    Gadget gadget = docView.getGadgetContainingView(view);
		    if (gadget != null)
			printedGadget = gadget;
		}
	    }

	    docPaneTracksViewportWidth = false;
	    printer.begin(printedGadget, pageFormat);

	    return true;
	}

	protected void exec() {
	    try {
		printJob.print();
	    } catch (PrintCanceledException e) {
		printCanceled = true;
	    } catch (PrinterException e) {
		printError = e.getMessage();
		if (printError == null)
		    printError = e.getClass().getName();
	    }
	}

	protected void finish() {
	    syncSetMessage(printingMessage);

	    docPaneTracksViewportWidth = true;
	    printer.end();

	    if (savedStyleSheet != null)
		docPane.getStyledViewFactory().setStyleSheet(savedStyleSheet);

	    setMessage("", false);
	    waitCursor(false, true);

	    if (printError != null) {
		JOptionPane.showMessageDialog(
		    frame, msg("cannotPrint", printError), 
		    msg("error"), JOptionPane.ERROR_MESSAGE);
	    } else if (printCanceled) {
		setMessage(msg("printingCanceled"), true);
	    }
	}
    }

    private StyleSheet setPrintStyleSheet() {
	StyledViewFactory factory = docPane.getStyledViewFactory();
	StyleSheet savedStyleSheet = factory.getStyleSheet();
	String[] otherMedia;

	if (savedStyleSheet != null && 
	    (otherMedia = savedStyleSheet.getOtherMedia()) != null &&
	    StringUtil.contains(otherMedia, "print")) {
	    StyleSheetInfo info = styleSheets[styleSheetIndex];
	    StyleSheet printStyleSheet = 
		loadStyleSheet(info.content, info.url, /*printing*/ true);

	    if (printStyleSheet == null)
		// '@media print {}' contains severe errors.
		savedStyleSheet = null;
	    else
		factory.setStyleSheet(printStyleSheet);
	} else {
	    savedStyleSheet = null;
	}

	return savedStyleSheet;
    }

    private String substituteHeaderVariables(String template) {
	String text = StringUtil.replaceAll(template, "%F", docFileName);
	text = StringUtil.replaceAll(text, "%P", msg("pageIndex"));
	// %I, %C handled by GadgetPrinter.
	return text;
    }

    public boolean printInProgress(GadgetPrinter printer, 
				   int pageIndex, int pageCount) {
	Thread.yield();

	if (pageIndex < pageCount) {
	    asyncSetMessage(msg("printingPage", new Integer(1+pageIndex), 
				new Integer(pageCount)) + 
			    msg("clickToCancel"));

	    if (cancelAsyncCommand)
		return false;
	}

	return true;
    }

    // --------------------------------------
    // Quit
    // --------------------------------------

    public void quit() {
	if (!checkNeedSave())
	    return;

	updatePreferences();
	pref.save();

	System.exit(0);
    }

    private void updatePreferences() {
	if (spellCheckerDialog != null)
	    spellCheckerDialog.getPreferences(pref);

	int count = pref.structuredFormats.length;
	for (int i = 0; i < count; ++i) 
	    pref.structuredFormats[i] = null;

	count = Math.min(1+pref.structuredFormats.length,
			 structuredFormats.length);

	for (int i = 1; i < count; ++i) {
	    StructuredFormat format = structuredFormats[i];
	    StringBuffer entry = new StringBuffer();

	    entry.append(format.getClass().getName());

	    String[] params = format.getParameters();
	    if (params != null) {
		for (int l = 0; l < params.length; ++l) {
		    entry.append(' ');
		    entry.append(Preferences.quoteArgument(params[l]));
		}
	    }

	    pref.structuredFormats[i-1] = entry.toString();
	}

	pref.geometry = frame.getBounds();
    }

    // --------------------------------------
    // Recently opened files
    // --------------------------------------

    public void reopen0() { reopen(0); }
    public void reopen1() { reopen(1); }
    public void reopen2() { reopen(2); }
    public void reopen3() { reopen(3); }
    public void reopen4() { reopen(4); }

    public void reopen(int index) {
	String fileName = pref.openedFileNames[index];
	if (fileName == null) 
	    return;

	if (!(new File(fileName)).exists()) {
	    frame.getToolkit().beep();
	    pref.removeOpenedFileName(fileName);
	    updateFileMenu();
	    return;
	}

	if (!checkNeedSave())
	    return;

	open(fileName);
    }

    // -----------------------------------------------------------------------
    // Select menu
    // -----------------------------------------------------------------------

    public void selectParent() {
	docPane.selectSubelement(JDocumentPane.PARENT_OR_TEXT);
    }

    public void selectChild() {
	docPane.selectSubelement(JDocumentPane.CHILD_OR_NONE);
    }

    public void selectPreceeding() {
	docPane.selectSubelement(JDocumentPane.PREVIOUS_SIBLING);
    }

    public void selectFollowing() {
	docPane.selectSubelement(JDocumentPane.NEXT_SIBLING);
    }

    public void selectAll() {
	docPane.selectAll();
    }

    // -----------------------------------------------------------------------
    // Edit menu
    // -----------------------------------------------------------------------

    public void undo() {
	if (undoManager.canUndo()) {
	    // Description already attached to undo action.
	    undoDescription = null;
	    undoAction();
	} else {
	    frame.getToolkit().beep();
	}
    }

    private void undoAction() {
	docPane.beginAction();

	docPane.cancelSelection();
	SubelementModel modified = undoManager.undo();
	if (modified != null)
	    docPane.movePointInto(modified, /*selectIfNoText*/ true);

	docPane.endAction();
    }

    public void redo() {
	if (undoManager.canRedo()) {
	    // Description already attached to undo action.
	    undoDescription = null;
	    redoAction();
	} else {
	    frame.getToolkit().beep();
	}
    }

    private void redoAction() {
	docPane.beginAction();

	docPane.cancelSelection();
	SubelementModel modified = undoManager.redo();
	if (modified != null)
	    docPane.movePointInto(modified, /*selectIfNoText*/ true);

	docPane.endAction();
    }

    public void cut() {
	if ((commandInfo.commands & ContextCommandInfo.CUT) != 0) {
	    undoDescription = commandInfoToLabel(commandInfo, 
						 ContextCommandInfo.CUT);
	    docPane.cut();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void copy() {
	if ((commandInfo.commands & ContextCommandInfo.COPY) != 0) {
	    // Nothing to be undone.
	    undoDescription = null;
	    docPane.copy();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void paste() {
	if ((commandInfo.commands & ContextCommandInfo.PASTE) != 0) {
	    undoDescription = commandInfoToLabel(commandInfo, 
						 ContextCommandInfo.PASTE);
	    docPane.paste();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void pasteBefore() {
	if ((commandInfo.commands & ContextCommandInfo.PASTE_BEFORE) != 0) {
	    undoDescription = commandInfoToLabel(
		commandInfo, ContextCommandInfo.PASTE_BEFORE);
	    docPane.pasteBefore();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void pasteAfter() {
	if ((commandInfo.commands & ContextCommandInfo.PASTE_AFTER) != 0) {
	    undoDescription = commandInfoToLabel(
		commandInfo, ContextCommandInfo.PASTE_AFTER);
	    docPane.pasteAfter();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void delete() {
	if ((commandInfo.commands & ContextCommandInfo.DELETE) != 0) {
	    undoDescription = commandInfoToLabel(commandInfo, 
						 ContextCommandInfo.DELETE);
	    docPane.delete();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void replace() {
	commandInfo.needArguments = ContextCommandInfo.REPLACE_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);

	if ((commandInfo.commands & ContextCommandInfo.REPLACE) != 0) {
	    String description = commandInfoToLabel(
		commandInfo, ContextCommandInfo.REPLACE);
	    String tag = getTagChooserDialog().selectTag(
		description, commandInfo.replaceArguments, hasDocType());

	    if (tag != null) {
		undoDescription = description;
		docPane.replace(tag);
	    }
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void insert() {
	commandInfo.needArguments = ContextCommandInfo.INSERT_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);

	if ((commandInfo.commands & ContextCommandInfo.INSERT) != 0) {
	    String description = 
		commandInfoToLabel(commandInfo, ContextCommandInfo.INSERT);
	    String tag = getTagChooserDialog().selectTag(
		description, commandInfo.insertArguments, hasDocType());

	    if (tag != null) {
		undoDescription = description;
		docPane.insert(tag);
	    }
	} else {
	    frame.getToolkit().beep();
	}
    }

    private boolean hasDocType() {
	DocumentModel doc = docPane.getDocument();
	return (doc != null && doc.getDocumentType() != null);
    }

    public void insertBefore() {
	commandInfo.needArguments = ContextCommandInfo.INSERT_BEFORE_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);

	if ((commandInfo.commands & ContextCommandInfo.INSERT_BEFORE) != 0) {
	    String description = commandInfoToLabel(
		commandInfo, ContextCommandInfo.INSERT_BEFORE);
	    String tag = getTagChooserDialog().selectTag(
		description, commandInfo.insertBeforeArguments, hasDocType());

	    if (tag != null) {
		undoDescription = description;
		docPane.insertBefore(tag);
	    }
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void insertAfter() {
	commandInfo.needArguments = ContextCommandInfo.INSERT_AFTER_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);

	if ((commandInfo.commands & ContextCommandInfo.INSERT_AFTER) != 0) {
	    String description = commandInfoToLabel(
		commandInfo, ContextCommandInfo.INSERT_AFTER);
	    String tag = getTagChooserDialog().selectTag(
		description, commandInfo.insertAfterArguments, hasDocType());

	    if (tag != null) {
		undoDescription = description;
		docPane.insertAfter(tag);
	    }
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void split() {
	if ((commandInfo.commands & ContextCommandInfo.SPLIT) != 0) {
	    undoDescription = commandInfoToLabel(commandInfo, 
						 ContextCommandInfo.SPLIT);
	    docPane.split();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void join() {
	if ((commandInfo.commands & ContextCommandInfo.JOIN) != 0) {
	    undoDescription = commandInfoToLabel(commandInfo, 
						 ContextCommandInfo.JOIN);
	    docPane.join();
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void convert() {
	commandInfo.needArguments = ContextCommandInfo.CONVERT_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);

	if ((commandInfo.commands & ContextCommandInfo.CONVERT) != 0) {
	    String description = commandInfoToLabel(
		commandInfo, ContextCommandInfo.CONVERT);
	    String tag = getTagChooserDialog().selectTag(
		description, commandInfo.convertArguments, hasDocType());

	    if (tag != null) {
		lastConvertTag = tag;

		undoDescription = description;
		docPane.convert(tag);
	    }
	} else {
	    frame.getToolkit().beep();
	}
    }

    public void applyLastConversion() {
	if (lastConvertTag == null) {
	    frame.getToolkit().beep();
	    return;
	}

	commandInfo.needArguments = ContextCommandInfo.CONVERT_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);

	if ((commandInfo.commands & ContextCommandInfo.CONVERT) != 0 &&
	    (!hasDocType() || StringUtil.contains(commandInfo.convertArguments,
						  lastConvertTag))) {
	    undoDescription = commandInfoToLabel(
		commandInfo, ContextCommandInfo.CONVERT);;
	    docPane.convert(lastConvertTag);
	} else {
	    frame.getToolkit().beep();
	}
    }

    // --------------------------------------
    // Insert file
    // --------------------------------------

    public void pasteFileBefore() {
	undoDescription = commandUndoDescription("edit.pasteBefore");
	doPasteFile(JDocumentPane.PASTE_BEFORE);
    }

    public void pasteFile() {
	undoDescription = commandUndoDescription("edit.paste");
	doPasteFile(JDocumentPane.PASTE_INTO);
    }

    public void pasteFileAfter() {
	undoDescription = commandUndoDescription("edit.pasteAfter");
	doPasteFile(JDocumentPane.PASTE_AFTER);
    }

    private String commandUndoDescription(String id) {
	String message = msg(id);

	if (message.endsWith("..."))
	    message = message.substring(0, message.length()-3);

	return message;
    }

    private void doPasteFile(int where) {
	String template = pref.openedFileNames[0];
	if (template == null)
	    template = getUntitledFileName("xml");

	String fileName = FileChooserUtil.getOpenFile(frame, null, template, 
                                                      /*filtered*/ false);
	if (fileName == null) {
	    undoDescription = null;
	    return;
	}

	asyncCommand = new PasteFileCommand(fileName, where);
	asyncCommand.start();
    }

    private class PasteFileCommand extends AsyncCommand {
	private String fileName;
	private int where;

	private boolean loadCanceled;
	private String loadError;

	public PasteFileCommand(String fileName, int where) {
	    this.fileName = fileName;
	    this.where = where;
	}

	protected boolean init() {
	    waitCursor(true, true);

	    String shortName = FileChooserUtil.truncateFileName(fileName);
	    setMessage(msg("loadingFile", shortName), false);

	    return true;
	}

	protected void exec() {
	    try {
		if (!pasteAtSelected(new File(fileName), where))
		    frame.getToolkit().beep();
	    } catch (WorkCanceledException e) {
		loadCanceled = true;
	    } catch (IOException e) {
		if (e instanceof FileNotFoundException)
		    loadError = msg("fileNotFound", e.getMessage());
		else
		    loadError = e.getMessage();
		if (loadError == null)
		    loadError = e.getClass().getName();
	    }
	}

	protected void finish() {
	    undoDescription = null;

	    setMessage("", false);
	    waitCursor(false, true);

	    if (loadError != null) {
		JOptionPane.showMessageDialog(
		    frame, msg("cannotLoadFile", fileName, loadError), 
		    msg("error"), JOptionPane.ERROR_MESSAGE);
	    } else if (loadCanceled) {
		setMessage(msg("loadingFileCanceled"), true);
	    }
	}
    }

    private boolean pasteAtSelected(File file, int where) throws IOException {
	return pasteAtSelected(FileUtil.fileToURL(file), where);
    }

    private boolean pasteAtSelected(URL url, int where) throws IOException {
	URLConnection connection = url.openConnection();
	connection.connect();

	String contentEncoding = connection.getContentEncoding();
	String javaEncoding = null;
	if (contentEncoding != null)
	    javaEncoding = XMLUtil.ianaToJavaEncoding(contentEncoding);
	if (javaEncoding == null)
	    javaEncoding = FileUtil.defaultEncoding();

	InputStream in = connection.getInputStream();
	byte[] bytes = null;

	try {
	    bytes = FileUtil.loadBytes(in);
	} finally {
	    in.close();
	}

	com.xmlmind.xmledit.modelimpl.Document document = 
	    new com.xmlmind.xmledit.modelimpl.Document();

	com.jclark.xml.parse.OpenEntity entity = 
	    new com.jclark.xml.parse.OpenEntity(
		new ByteArrayInputStream(bytes), url.toExternalForm(), url);

	com.xmlmind.xmledit.modelimpl.XPDocumentLoader loader = 
	    new com.xmlmind.xmledit.modelimpl.XPDocumentLoader();
	loader.setObserver(this);

	try {
	    loader.load(entity, document);
	} catch (IOException ignored) {
	    document = null;
	}

	String content;
	if (document == null)
	    content = new String(bytes, 0, bytes.length, javaEncoding);
	else
	    content = SubelementUtil.toString(document.getRootElement());

	return docPane.pasteAtSelected(content, where);
    }

    // -----------------------------------------------------------------------
    // Search menu
    // -----------------------------------------------------------------------

    public void search() {
	SearchTextDialog.Info info = getSearchReplaceInfo(false, false);
	if (info == null)
	    return;

	searchInfo = info;
	doFind(info, info.backward);
    }

    private SearchTextDialog.Info getSearchReplaceInfo(boolean replace,
						       boolean backward) {
	if (searchTextDialog == null)
	    searchTextDialog = new SearchTextDialog(frame);

	SearchTextDialog.Info in = new SearchTextDialog.Info();
	if (replace)
	    in.replacement = "";
	in.ignoreCase = pref.ignoreCase;
	in.wholeWord = pref.wholeWord;
	in.regularExpression = pref.regularExpression;
	in.backward = backward;

	return searchTextDialog.getInfo(in);
    }

    public void searchBackward() {
	SearchTextDialog.Info info = getSearchReplaceInfo(false, true);
	if (info == null)
	    return;

	searchInfo = info;
	doFind(info, info.backward);
    }

    public void findNext() {
	if (checkSearchInfo())
	    doFind(searchInfo, false);
    }

    private boolean checkSearchInfo() {
	if (searchInfo == null) {
	    frame.getToolkit().beep();
	    setMessage(msg("noSearchedText"), true);
	    return false;
	} else {
	    return true;
	}
    }

    public void findPrevious() {
	if (checkSearchInfo())
	    doFind(searchInfo, true);
    }

    private void doFind(SearchTextDialog.Info info, boolean backward) {
	SearchReplaceText searchReplaceText = 
	    createSearchReplaceText(info, null, backward);
	if (searchReplaceText == null)
	    return;

	TextOffset[] span = searchReplaceText.search();
	if (span == null) {
	    setMessage(msg("searchedTextNotFound", shorten(info.searched, 20)),
		       true);
	    return;
	}

	setMessage("", true);
	docPane.selectText(span[0], span[1]);
    }

    private static final String shorten(String s, int max) {
	if (s.length() > max)
	    s = s.substring(0, max-3) + "...";
	return s;
    }

    private 
    SearchReplaceText createSearchReplaceText(SearchTextDialog.Info info, 
					      String replacement,
					      boolean backward) {
	Caret caret = docPane.getCaret();
	if (caret == null)
	    // No document.
	    return null;

	ElementModel inside;
	TextOffset from;

	if (replacement != null &&
	    commandInfo.context == ContextCommandInfo.SUBELEMENT_SELECTION) {
	    SubelementModel sub = commandInfo.subelement;
	    if (sub.getSubelementType() == SubelementModel.ELEMENT)
		inside = (ElementModel) sub;
	    else
		inside = sub.getParentElement();

	    from = null;
	} else {
	    inside = null; // whole document.

	    TextLocation textLocation;
	    if (commandInfo.context == ContextCommandInfo.TEXT_SELECTION) {
		// If search started at the beginning of the selection,
		// it would be stuck at the first found word (we here we
		// need to highlight text rather than to select it.)

		if ((backward && !commandInfo.markAfterPoint) ||
		    (!backward && commandInfo.markAfterPoint))
		    textLocation = caret.getMark();
		else
		    textLocation = caret.getPoint();
	    } else {
		textLocation = caret.getPoint();
	    }

	    from = new TextOffset();
	    from.text = textLocation.getText();
	    if (from.text == null) {
		frame.getToolkit().beep();
		setMessage(msg("documentContainsNoText"), true);
		return null;
	    }
	    from.offset = textLocation.getOffset();
	}

	int options;
	switch (pref.markParagraphs) {
	case Preferences.MARK_NONE:
	    options = SearchReplaceText.MARK_NONE;
	    break;
	case Preferences.MARK_ALL:
	    options = SearchReplaceText.MARK_ALL;
	    break;
	default:
	    options = SearchReplaceText.MARK_LAST;
	}
	if (info.ignoreCase)
	    options |= SearchReplaceText.IGNORE_CASE;
	if (info.wholeWord)
	    options |= SearchReplaceText.WHOLE_WORD;
	if (backward)
	    options |= SearchReplaceText.BACKWARD;

	SearchReplaceText searchReplaceText;
	if (backward) {
	    if (info.regularExpression)
		searchReplaceText = 
		    new SearchReplaceText(inside, null, from, info.searchedRE, 
					  replacement, options);
	    else
		searchReplaceText = 
		    new SearchReplaceText(inside, null, from, info.searched, 
					  replacement, options);
	} else {
	    if (info.regularExpression)
		searchReplaceText = 
		    new SearchReplaceText(inside, from, null, info.searchedRE, 
					  replacement, options);
	    else
		searchReplaceText = 
		    new SearchReplaceText(inside, from, null, info.searched, 
					  replacement, options);
	}

	return searchReplaceText;
    }

    public void searchReplace() {
	SearchTextDialog.Info info = getSearchReplaceInfo(true, false);
	if (info == null)
	    return;

	doReplace(info);
    }

    public void searchReplaceBackward() {
	SearchTextDialog.Info info = getSearchReplaceInfo(true, true);
	if (info == null)
	    return;

	doReplace(info);
    }

    private void doReplace(SearchTextDialog.Info info) {
	searchReplaceText = createSearchReplaceText(info, info.replacement, 
						    info.backward);
	if (searchReplaceText == null)
	    return;

	if (!searchText()) {
	    setMessage(msg("searchedTextNotFound", shorten(info.searched, 20)),
		       true);
	    searchReplaceText = null;
	    return;
	}

	if (replaceTextDialog == null)
	    replaceTextDialog = new ReplaceTextDialog(frame, this);
        replaceTextDialog.showDialog();
        // Modal means blocked here until Cancel is pressed or search/replace
        // is completed.

	int count = searchReplaceText.getReplacementCount();
	searchReplaceText = null;
	setMessage(msg("replacedText", new Integer(count),
		       shorten(info.searched, 20)), true);
    }

    private boolean searchText() {
	TextOffset[] span = searchReplaceText.search();
	if (span == null)
	    return false;

	docPane.selectText(span[0], span[1]);
	return true;
    }

    private void replaceText() {
	undoDescription = msg("searchReplaceText");

	docPane.beginAction();
	TextRange range = searchReplaceText.replace();
	docPane.endAction();

	undoDescription = null;

	if (!searchText()) {
	    docPane.selectText(range);
	    replaceTextDialog.hide();
	    replaceTextDialog.dispose();
	}
    }

    private void skipText() {
	if (!searchText()) {
	    replaceTextDialog.hide();
	    replaceTextDialog.dispose();
	}
    }

    private void replaceAllText() {
	undoDescription = msg("searchReplaceAllText");

	docPane.beginAction();
	TextRange[] ranges = searchReplaceText.replaceAll();
	docPane.endAction();

	undoDescription = null;

	docPane.selectText(ranges[ranges.length-1]);
        replaceTextDialog.hide();
        replaceTextDialog.dispose();
    }

    // -----------------------------------------------------------------------
    // Style menu
    // -----------------------------------------------------------------------

    public void noStyle() {
	setStyle(-1);
    }

    public void setStyle0() { setStyle(0); }
    public void setStyle1() { setStyle(1); }
    public void setStyle2() { setStyle(2); }
    public void setStyle3() { setStyle(3); }
    public void setStyle4() { setStyle(4); }
    public void setStyle5() { setStyle(5); }
    public void setStyle6() { setStyle(6); }
    public void setStyle7() { setStyle(7); }
    public void setStyle8() { setStyle(8); }
    public void setStyle9() { setStyle(9); }

    private void setStyle(int index) {
	waitCursor(true, false);
	try {
	    doSetStyle(index);
	} finally {
	    waitCursor(false, false);
	}
    }

    private void doSetStyle(int index) {
	StyleSheet styleSheet = null;

	if (index >= 0) {
	    StyleSheetInfo info = styleSheets[index];

	    styleSheet = loadStyleSheet(info.content, info.url, 
                                        /*printing*/ false);
	    if (styleSheet == null) {
		JMenu menu = MenuUtil.getMenu(menus, "style");
		JMenuItem item = menu.getItem(2+index);
		item.setEnabled(false);
	    }
	}

	StyledViewFactory factory = docPane.getStyledViewFactory();
	if (styleSheet != null || factory.getStyleSheet() != null)
	    factory.setStyleSheet(styleSheet);
	styleSheetIndex = index;
	    
	boolean hasStyle = (styleSheet != null);
	MenuUtil.getAction(menus, "style.noStyle").setEnabled(hasStyle);
    }

    private StyleSheet loadStyleSheet(String content, String url, 
				      boolean printing) {
	StyleSheetLoader loader = new StyleSheetLoader();
	if (printing)
	    loader.setHandledMedia(StyleSheetLoader.PRINT);

	StyleSheet styleSheet = new StyleSheet();
	String error = null;

	try {
	    if (content == null)
		loader.load(new URL(url), null, styleSheet);
	    else
		loader.load(content, new URL(url), null, styleSheet);

	    //StyleSheetWriter.writeStyleSheet(styleSheet, "__STYLESHEET");
	} catch (Exception e) {
	    styleSheet = null;
	    error = e.getMessage();
	    if (error == null)
		error = e.getClass().getName();
	}

	if (styleSheet == null) {
	    JOptionPane.showMessageDialog(
		frame, 
		msg("cannotLoadFile", url, error),
		msg("error"), JOptionPane.ERROR_MESSAGE);
	} else {
	    StyleSheetLoader.Warning[] warnings = loader.getWarnings();

	    if (warnings.length > 0) {
		StringBuffer messages = new StringBuffer();

		for (int j = 0; j < warnings.length; ++j) {
		    StyleSheetLoader.Warning warn = warnings[j];
		    String msgId;

		    switch (warn.type) {
		    case StyleSheetLoader.UNSUPPORTED_PROPERTY:
			msgId = "styleSheetSyntaxError";
			break;
		    case StyleSheetLoader.UNSUPPORTED_PROPERTY_VALUE:
			msgId = "styleSheetUnsupportedPropertyValue";
			break;
		    case StyleSheetLoader.SYNTAX_ERROR:
		    default:
			msgId = "styleSheetSyntaxError";
		    }

		    messages.append(
			msg(msgId, urlBaseName(warn.url.toExternalForm()), 
			    new Integer(warn.lineNumber), 
			    new Integer(warn.columnNumber), 
			    warn.detail));
		    messages.append("\n\n");
		}

		JTextArea textArea = 
		    new JTextArea(messages.toString(), 10, 40);
		textArea.setEditable(false);
		textArea.setLineWrap(true);
		textArea.setWrapStyleWord(true);
		JOptionPane.showMessageDialog(
		    frame, 
		    new Object[] { 
			msg("styleSheetHasErrors", url),
			new JScrollPane(textArea)
		    }, 
		    msg("warning"), JOptionPane.WARNING_MESSAGE);
	    }
	}

	return styleSheet;
    }

    // -----------------------------------------------------------------------
    // Tools menu
    // -----------------------------------------------------------------------

    public void editAttributes() {
	if (attributesTool == null) {
	    attributesTool = new AttributesTool(frame);
	    attributesTool.setDocument(docPane.getDocument());
	} 

	commandInfo.needArguments = ContextCommandInfo.NO_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);
	// Do not use getContext() which is faster but erases all the possible
	// commands.

	attributesTool.setElement(getSelectedElement(commandInfo));

	attributesTool.showTool();
    }

    private static ElementModel getSelectedElement(ContextCommandInfo context){
	ElementModel selectedElement = null;

	switch (context.context) {
	case ContextCommandInfo.SUBELEMENT_SELECTION:
	    if (context.subelement.getSubelementType() == 
		SubelementModel.ELEMENT)
		selectedElement = (ElementModel) context.subelement;
	    break;
	case ContextCommandInfo.POINT_IN_ELEMENT:
	    selectedElement = context.pointText.getParentElement();
	    break;
	}

	return selectedElement;
    }

    public void checkValidity() {
	DocumentModel doc = docPane.getDocument();
	if (doc == null || doc.getDocumentType() == null) {
	    frame.getToolkit().beep();
	    return;
	}

	doCheckValidity(doc);

	if (validityDiagnosticsTool.getDiagnostics() == null) {
	    setMessage(msg("documentIsValid"), true);
	} else {
	    if (!validityDiagnosticsTool.isVisible())
		validityDiagnosticsTool.showDiagnostics();
	}
    }

    private void doCheckValidity(DocumentModel doc) {
	DefaultElemStructEditor structEditor = (DefaultElemStructEditor) 
	    docPane.getDocumentView().getElementStructureEditor();

	if (doc == null || doc.getDocumentType() == null) {
	    docChecker = null;
	    structEditor.setLenient(false);
	    if (validityDiagnosticsTool != null)
		validityDiagnosticsTool.setDiagnostics(null);
	    invalid.setIcon(invalidIcons[0]);
	    invalid.setToolTipText("");
	    return;
	}

	if (docChecker == null)
	    docChecker = new DocumentChecker(doc);
	ElementCheckerDiagnostic[] diagnostics = docChecker.check();

	int invalidity = 0;
	for (int i = 0; i < diagnostics.length; ++i) {
	    ElementCheckerDiagnostic d = diagnostics[i];
	    
	    int severity;
	    if (d instanceof ElementStructureChecker.Diagnostic) {
		severity = 3;
	    } else {
		if (d.invalidReason >= ElementAttributeChecker.XREF_REASON)
		    severity = 1;
		else
		    severity = 2;
	    }

	    if (severity > invalidity)
		invalidity = severity;
	}

	invalid.setIcon(invalidIcons[invalidity]);
	invalid.setToolTipText(
	    (invalidity == 0)? "" : msg("invalidReason." + invalidity));

	if (validityDiagnosticsTool == null)
	    validityDiagnosticsTool = 
		new ValidityDiagnosticsTool(frame, docPane);
	validityDiagnosticsTool.setDiagnostics(diagnostics);

	boolean wasLenient = structEditor.isLenient();
	if (wasLenient && invalidity < 3) {
	    structEditor.setLenient(false);

	    JOptionPane.showMessageDialog(
		frame, msg("usingNormalMode"), 
		msg("info"), JOptionPane.INFORMATION_MESSAGE);
	} else if (!wasLenient && invalidity >= 3) {
	    structEditor.setLenient(true);

	    JOptionPane.showMessageDialog(
		frame, msg("usingRepairMode"), 
		msg("warning"), JOptionPane.WARNING_MESSAGE);
	}
    }

    public void checkSpelling() {
	waitCursor(true, false);
	syncSetMessage(msg("checkingSpellingErrors"));
	boolean checked = false;

	try {
	    if (spellCheckerDialog == null) {
		spellCheckerDialog = new SpellCheckerDialog(frame, docPane);
		spellCheckerDialog.setPreferences(pref);
	    }

	    checked = spellCheckerDialog.checkSpelling();
	} finally {
	    if (checked)
		setMessage(msg("noSpellingErrors"), true);
	    else
		setMessage("", false);
	    waitCursor(false, false);
	}
    }

    public void refreshSelected() {
	ElementModel refreshed = docPane.refreshSelected();
	if (refreshed == null)
	    frame.getToolkit().beep();
	else
	    setMessage(msg("refreshed", refreshed.getElementTypeName()), true);
    }

    // -----------------------------------------------------------------------
    // Options menu
    // -----------------------------------------------------------------------

    public void options() {
	if (optionsDialog == null)
	    optionsDialog = new OptionsDialog(frame);

	Options newOptions = optionsDialog.getOptions(pref);
	if (newOptions == null)
	    return;

	boolean newDefaultStyle = 
	    (newOptions.defaultFontSize != pref.defaultFontSize || 
	     newOptions.defaultFontIsSerif != pref.defaultFontIsSerif || 
	     !newOptions.defaultBackground.equals(pref.defaultBackground) || 
	     !newOptions.defaultForeground.equals(pref.defaultForeground) ||
	     !newOptions.serifFontFamily.equals(pref.serifFontFamily) || 
	     !newOptions.sansSerifFontFamily.equals(pref.sansSerifFontFamily)||
	     !newOptions.monospaceFontFamily.equals(pref.monospaceFontFamily));
	pref.setOptions(newOptions);

	XMLFormat xmlFormat = (XMLFormat) structuredFormats[0];
	xmlFormat.setGuessingElementContent(pref.guessElementContent);
	xmlFormat.setIndentation(pref.indentIfDocType, pref.indentIfNoDocType, 
				 pref.indentation, pref.maxLineLength);

	undoManager.setUndoMaxCount(pref.maxUndo);
	applyMakeChoice();

	if (spellCheckerDialog != null)
	    spellCheckerDialog.setPreferences(pref);

	if (newDefaultStyle) {
	    setDefaultStyle(pref);
	    applyDefaultStyle();
	}
    }

    private void applyMakeChoice() {
	DefaultElemStructEditor structEditor = (DefaultElemStructEditor) 
	    docPane.getDocumentView().getElementStructureEditor();
	structEditor.setCreateContentOptions(makeChoiceOptions());
    }

    private int makeChoiceOptions() {
	int options;

	switch (pref.makeChoice) {
	case Options.NO_CHOICE:
	    options = 0x0;
	    break;
	case Options.FIRST_CHOICE:
	    options = (DefaultElemStructEditor.MAKE_CHOICES|
		       DefaultElemStructEditor.USE_FIRST_CHOICE);
	    break;
	case Options.SMART_CHOICE:
	default:
	    options = DefaultElemStructEditor.MAKE_CHOICES;
	}

	return options;
    }

    private void setDefaultStyle(Options options) {
	Style defaultStyle = new Style();
	defaultStyle.paddingTop = defaultStyle.paddingRight = 
	    defaultStyle.paddingBottom = defaultStyle.paddingLeft = 5; 

	StyledViewFactory styledViewFactory = docPane.getStyledViewFactory();

	styledViewFactory.setSerifFontFamily(options.serifFontFamily);
	styledViewFactory.setSansSerifFontFamily(options.sansSerifFontFamily);
	styledViewFactory.setMonospaceFontFamily(options.monospaceFontFamily);

	defaultStyle.font = 
	    new Font((options.defaultFontIsSerif? 
		      options.serifFontFamily : options.sansSerifFontFamily), 
		     Font.PLAIN, options.defaultFontSize);
	defaultStyle.backgroundColor = options.defaultBackground;
	defaultStyle.color = options.defaultForeground; 
	styledViewFactory.setDefaultStyle(defaultStyle);
    }

    private void applyDefaultStyle() {
	StyledViewFactory styledViewFactory = docPane.getStyledViewFactory();
	StyleSheet styleSheet = styledViewFactory.getStyleSheet();
	if (styleSheet != null) {
	    waitCursor(true, false);
	    try {
		styledViewFactory.setStyleSheet(styleSheet);
	    } finally {
		waitCursor(false, false);
	    }
	}
    }

    public void structuredFormats() {
	if (structuredFormatsDialog == null)
	    structuredFormatsDialog = new StructuredFormatsDialog(frame);

	structuredFormatsDialog.edit(this);
    }

    // -----------------------------------------------------------------------
    // Set of StructuredFormats API
    // -----------------------------------------------------------------------

    public StructuredFormat[] getStructuredFormats() {
	// First format: XMLFormat not editable.
	StructuredFormat[] formats = 
	    new StructuredFormat[structuredFormats.length-1];
	System.arraycopy(structuredFormats, 1, formats, 0, formats.length);

	return formats;
    }

    public StructuredFormat addStructuredFormat(String className) 
	throws ClassNotFoundException, InstantiationException, 
	       IllegalAccessException {
	int index = indexOfStructuredFormat(className);
	if (index >= 0)
	    return structuredFormats[index];

	StructuredFormat format = (StructuredFormat) 
	    Class.forName(className).newInstance();
	addStructuredFormat(format);

	return format;
    }

    private int indexOfStructuredFormat(String className) {
	for (int i = 0; i < structuredFormats.length; ++i) {
	    if (structuredFormats[i].getClass().getName().equals(className))
		return i;
	}
	return -1;
    }

    private void addStructuredFormat(StructuredFormat format) {
	format.setWorkInProgressObserver(this);

	int registered = structuredFormats.length;
	StructuredFormat[] formats = new StructuredFormat[registered+1];
	System.arraycopy(structuredFormats, 0, formats, 0, registered);
	formats[registered] = format;

	structuredFormats = formats;
    }

    public StructuredFormat removeStructuredFormat(String className) {
	int index = indexOfStructuredFormat(className);
	if (index <= 0)
	    // Cannot remove XMLFormat at index 0.
	    return null;

	StructuredFormat removed = structuredFormats[index];
	StructuredFormat[] formats = 
	    new StructuredFormat[structuredFormats.length-1];

	if (index > 0)
	    System.arraycopy(structuredFormats, 0, formats, 0, index);

	int first = index + 1;
	int tail = structuredFormats.length - first;
	if (tail > 0) 
	    System.arraycopy(structuredFormats, first, formats, index, tail);

	structuredFormats = formats;

	return removed;
    }

    // -----------------------------------------------------------------------
    // DocumentListener
    // -----------------------------------------------------------------------

    public void textChanged(com.xmlmind.xmledit.model.TextEvent event, 
			    int listenerIndex) {
	if (!needSave)
	    setNeedSave();
    }

    private void setNeedSave() {
	needSave = true;
	updateTitle();
	MenuUtil.getAction(menus, "file.save").setEnabled(true);
    }

    public void elementAttributeChanged(ElementAttributeEvent event, 
					int listenerIndex) {
	if (!needSave)
	    setNeedSave();
    }

    public void elementStructureChanged(ElementStructureEvent event, 
					int listenerIndex) {
	if (!needSave)
	    setNeedSave();
    }

    public void batchEditingStarted(BatchEditingEvent event, 
				    int listenerIndex) {}
    public void batchEditingCompleted(BatchEditingEvent event, 
				      int listenerIndex) {}
    public void documentEventHappened(DocumentEvent event,
				      int listenerIndex) {}

    // -----------------------------------------------------------------------
    // ContextChangeListener
    // -----------------------------------------------------------------------

    public void contextChanged(ContextChangeEvent event) {
	commandInfo.needArguments = ContextCommandInfo.NO_ARGUMENTS;
	docPane.getContextCommandInfo(commandInfo);

	for (int i = 0; i < editMenuItemNames.length; ++i) {
	    int command = editMenuItemCommands[i];
	    Action action = MenuUtil.getAction(menus, editMenuItemNames[i]);

	    action.setEnabled((commandInfo.commands & command) != 0);
	}

	MenuUtil.getAction(menus, "edit.applyLastConversion").setEnabled(
	    lastConvertTag != null && 
	    (commandInfo.commands & ContextCommandInfo.CONVERT) != 0);

	if (commandInfo.clipboardTag == null)
	    clipboardTag.setText("");
	else
	    clipboardTag.setText(commandInfo.clipboardTag);
	    
	SubelementModel sub = null;
	boolean enableSelectParent = false;
	boolean enableSelectChild = false;
	boolean enableSelectPreceeding = false;
	boolean enableSelectFollowing = false;

	switch (commandInfo.context) {
	case ContextCommandInfo.SUBELEMENT_SELECTION:
	    sub = commandInfo.subelement;
	    enableSelectParent = (sub.getParentElement() != null);
	    enableSelectChild = true; // At worst, clears subelement selection.
	    enableSelectPreceeding = (sub.getPreviousSubelement() != null);
	    enableSelectFollowing = (sub.getNextSubelement() != null);
	    break;

	case ContextCommandInfo.POINT_IN_ELEMENT:
	    sub = commandInfo.pointText.getParentElement();
	    enableSelectParent = true; // The #PCDATA containing the caret.
	    enableSelectPreceeding = (sub.getPreviousSubelement() != null);
	    enableSelectFollowing = (sub.getNextSubelement() != null);
	    break;

	case ContextCommandInfo.TEXT_SELECTION:
	    if (commandInfo.markAfterPoint)
		sub = commandInfo.pointText;
	    else
		sub = commandInfo.markText;
	    break;

	case ContextCommandInfo.POINT:
	    sub = commandInfo.pointText;
	    break;
	}

	MenuUtil.getAction(menus, "select.selectParent").setEnabled(
	    enableSelectParent);
	MenuUtil.getAction(menus, "select.selectChild").setEnabled(
	    enableSelectChild);
	MenuUtil.getAction(menus, "select.selectPreceeding").setEnabled(
	    enableSelectPreceeding);
	MenuUtil.getAction(menus, "select.selectFollowing").setEnabled(
	    enableSelectFollowing);

	subelementPath.setSubelement(sub);

	if (attributesTool != null && attributesTool.isVisible())
	    attributesTool.setElement(getSelectedElement(commandInfo));
    }

    // -----------------------------------------------------------------------
    // ItemListener
    // -----------------------------------------------------------------------

    public void itemStateChanged(ItemEvent e) {
	if (e.getStateChange() == ItemEvent.SELECTED)
	    docPane.selectSubelement((SubelementModel) e.getItem());
    }

    // -----------------------------------------------------------------------
    // UndoManagerListener
    // -----------------------------------------------------------------------

    public void undoManagerChange(UndoManagerEvent e) {
	UndoManager.Action action;
	String description;
	String itemLabel;

	action = undoManager.getTopUndoAction();
	if (action == null) {
	    itemLabel = undoHelper.undoMessage;
	} else {
	    if (undoDescription == null) {
		description = action.getDescription();
	    } else {
		action.setDescription(undoDescription);
		description = undoDescription;
	    }

	    if (description == null) {
		DocumentEvent[] actionEvents = action.getDocumentEvents();

		if (actionEvents.length == 1) {
		    DocumentEvent actionEvent = actionEvents[0];

		    if (actionEvent instanceof 
			    com.xmlmind.xmledit.model.TextEvent) {
			action.setDescription(
			    undoHelper.undoTypingDescription);
			itemLabel = undoHelper.undoTypingMessage;
		    } else if (actionEvent instanceof ElementAttributeEvent) {
			action.setDescription(
			    undoHelper.undoAttributeChangeDescription);
			itemLabel = undoHelper.undoAttributeChangeMessage;
		    } else {
			itemLabel = undoHelper.undoMessage;
		    }
		} else {
		    itemLabel = undoHelper.undoMessage;
		}
	    } else {
		itemLabel = undoHelper.formatUndoMessage(description);
	    }
	}
	undoHelper.undoAction.setEnabled(action != null);
	undoHelper.undoMenuItem.setText(itemLabel);

	undoDescription = null;

	action = undoManager.getTopRedoAction();
	if (action == null || (description = action.getDescription()) == null) 
	    itemLabel = undoHelper.redoMessage;
	else
	    itemLabel = undoHelper.formatRedoMessage(description);
	undoHelper.redoAction.setEnabled(action != null);
	undoHelper.redoMenuItem.setText(itemLabel);
    }

    // -----------------------------------------------------------------------
    // ActionListener
    // -----------------------------------------------------------------------

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	boolean done = true;
	
	if (command.equals("insertTextAfter")) {
	    undoDescription = commandUndoDescription("edit.insertAfter");
	    done = docPane.insertAtSelected(/*text*/ true, /*before*/ false);
	} else if (command.equals("insertTextBefore")) {
	    undoDescription = commandUndoDescription("edit.insertBefore");
	    done = docPane.insertAtSelected(/*text*/ true, /*before*/ true);
	} else if (command.equals("insertLikeSelectedAfter")) {
	    undoDescription = commandUndoDescription("edit.insertAfter");
	    done = docPane.insertAtSelected(/*text*/ false, /*before*/ false);
	} else if (command.equals("insertLikeSelectedBefore")) {
	    undoDescription = commandUndoDescription("edit.insertBefore");
	    done = docPane.insertAtSelected(/*text*/ false, /*before*/ true);
	} else if (command.equals("splitSelected")) {
	    undoDescription = commandUndoDescription("edit.split");
	    done = docPane.splitSelected();
	} else if (command.equals("replaceText")) {
	    replaceText();
	} else if (command.equals("skipText")) {
	    skipText();
	} else if (command.equals("replaceAllText")) {
	    replaceAllText();
	}

	undoDescription = null;
	if (!done)
	    frame.getToolkit().beep();
    }

    // -----------------------------------------------------------------------
    // Helpers
    // -----------------------------------------------------------------------

    private void updateTitle() {
	StringBuffer title = new StringBuffer();

	if (docFileName == null) {
	    title.append(TITLE);
	} else {
	    title.append(docFileName);

	    if (needSave) {
		title.append(" (");
		title.append(msg("modified"));
		title.append(')');
	    }
	}

	frame.setTitle(title.toString());
    }

    private void updateFileMenu() {
	JMenu menu = MenuUtil.getMenu(menus, "file");

	// [0] newFromTemplate, [1] openAsTemplate, [2]  -, 
	// [3] open, [4] save, [5] saveAs, [6] close, [7]  -, 
	// [8] pageSetup, [9] print, [10] -, [11] quit, [12] -
	for (int i = menu.getItemCount() - 1; i >= 12; --i) 
	    menu.remove(i);

	if (pref.openedFileNames[0] != null) {
	    menu.addSeparator();

	    for (int i = 0; i < pref.openedFileNames.length; ++i) {
		String fileName = pref.openedFileNames[i];
		if (fileName == null)
		    break;
		
		JMenuItem item = menu.add(new AppAction("reopen" + i));
		item.setText(FileChooserUtil.truncateFileName(fileName));
	    }
	}
    }

    private void enableMenus() {
	DocumentModel doc = docPane.getDocument();
	boolean docLoaded = (doc != null);
	boolean hasDocType = (docLoaded && doc.getDocumentType() != null);

	MenuUtil.getAction(menus, "file.close").setEnabled(docLoaded);
	MenuUtil.getAction(menus, "file.save").setEnabled(false);
	MenuUtil.getAction(menus, "file.saveAs").setEnabled(docLoaded);
	MenuUtil.getAction(menus, "file.print").setEnabled(docLoaded);

	MenuUtil.getAction(menus, "select.selectAll").setEnabled(docLoaded);

	undoDescription = null;
	undoManagerChange(null);
	MenuUtil.getAction(
	    menus, "edit.applyLastConversion").setEnabled(false);
	MenuUtil.getAction(
	    menus, "edit.insertFile.pasteFileBefore").setEnabled(docLoaded);
	MenuUtil.getAction(
	    menus, "edit.insertFile.pasteFile").setEnabled(docLoaded);
	MenuUtil.getAction(
	    menus, "edit.insertFile.pasteFileAfter").setEnabled(docLoaded);

	MenuUtil.getAction(menus, "search.search").setEnabled(docLoaded);
	MenuUtil.getAction(
	    menus, "search.searchBackward").setEnabled(docLoaded);
	MenuUtil.getAction(menus, "search.findNext").setEnabled(docLoaded);
	MenuUtil.getAction(menus, "search.findPrevious").setEnabled(docLoaded);
	MenuUtil.getAction(
	    menus, "search.searchReplace").setEnabled(docLoaded);
	MenuUtil.getAction(
	    menus, "search.searchReplaceBackward").setEnabled(docLoaded);

	MenuUtil.getAction(menus,"tools.checkValidity").setEnabled(hasDocType);
	MenuUtil.getAction(menus, "tools.checkSpelling").setEnabled(docLoaded);
	MenuUtil.getAction(
	    menus, "tools.refreshSelected").setEnabled(docLoaded);
    }

    private boolean checkNeedSave() {
	if (needSave) {
	    int option = JOptionPane.showConfirmDialog(
		frame, msg("saveDocument.message", docFileName), 
		msg("saveDocument.title"), JOptionPane.YES_NO_CANCEL_OPTION);

	    switch (option) {
	    case JOptionPane.YES_OPTION:
		save();
		/*FALLTHROUGH*/
	    case JOptionPane.NO_OPTION:
		return true;
	    default:
		return false;
	    }
	} else {
	    return true;
	}
    }
    
    private void waitCursor(boolean on, boolean spinning) {
	Component glassPane = frame.getGlassPane();

	if (on && !glassPane.isVisible()) {
	    cancelAsyncCommand = false;
	    glassPane.setVisible(true);
	    // Windows JDK1.3 focus problem workaround.
	    //glassPane.requestFocus();

	    if (spinning && progressTimer == null) {
		progressIcon = 1;
		progressTimer = 
		    new Timer(250 /*ms*/, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    progress.setIcon(progressIcons[progressIcon]);
			    ++progressIcon;
			    if (progressIcon == progressIcons.length)
				progressIcon = 1;
			}
		    });
		progressTimer.setInitialDelay(0);
		progressTimer.setRepeats(true);
		progressTimer.setCoalesce(false);
		progressTimer.start();
	    }
	} else if (!on && glassPane.isVisible()) {
	    cancelAsyncCommand = false;
	    glassPane.setVisible(false);
	    //docPane.requestFocus();

	    if (progressTimer != null) {
		progressTimer.stop();
		progressTimer = null;
		progress.setIcon(progressIcons[0]);
	    }
	}
    }

    private void syncSetMessage(final String text) {
	setMessage(text, /*autoErase*/ false);
	message.paintImmediately(0, 0, 1000, 1000);
    }

    private void asyncSetMessage(final String text) {
	SwingUtilities.invokeLater(new Runnable() {
	    public void run() {
		setMessage(text, /*autoErase*/ false);
	    }
	});
    }

    private void setMessage(String text, boolean autoErase) {
	if (messageTimeout != null) {
	    messageTimeout.stop();
	    messageTimeout = null;
	}

	message.setText(text);
	if (text.length() == 0)
	    return;

	if (autoErase) {
	    messageTimeout = 
		new Timer(10000 /*ms*/, new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			message.setText("");
		    }
		});
	    messageTimeout.setRepeats(false);
	    messageTimeout.start();
	}
    }

    public boolean workInProgress(Object worker, boolean working) {
	Thread.yield();

	if (working && cancelAsyncCommand)
	    return false;
	else
	    return true;
    }

    private TagChooserDialog getTagChooserDialog() {
	if (tagChooserDialog == null) {
	    tagChooserDialog = new TagChooserDialog(frame);
	    
	    DocumentModel doc = docPane.getDocument();
	    if (doc != null && doc.getDocumentType() == null)
		tagChooserDialog.setTagHistory(
		    DocumentUtil.getAllElementTypeNames(doc));
	}

	return tagChooserDialog;
    }

    // -----------------------------------------------------------------------
    // Menus
    // -----------------------------------------------------------------------

    private void addActions(JMenuBar bar) {
	MenuUtil.MenuSpec[] menuSpecs = {
	   new MenuUtil.MenuSpec("file",
		new MenuUtil.MenuItemSpec[] {
		    new MenuUtil.MenuItemSpec(new AppAction("newFromTemplate"),
					      KeyEvent.VK_N),
		    new MenuUtil.MenuItemSpec(new AppAction("openAsTemplate")),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuItemSpec(new AppAction("open"), 
					      KeyEvent.VK_O),
		    new MenuUtil.MenuItemSpec(new AppAction("save"), 
					      KeyEvent.VK_S),
		    new MenuUtil.MenuItemSpec(new AppAction("saveAs")),
		    new MenuUtil.MenuItemSpec(new AppAction("close")),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuItemSpec(
			new AppAction("pageSetup"),
			KeyEvent.VK_P,
			AWTUtil.KEY_MODIFIER|Event.SHIFT_MASK),
		    new MenuUtil.MenuItemSpec(new AppAction("print"),
					      KeyEvent.VK_P),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuItemSpec(new AppAction("quit"), 
					      KeyEvent.VK_Q)
		}),

	    new MenuUtil.MenuSpec("select",
		new MenuUtil.MenuItemSpec[] {
		    new MenuUtil.MenuItemSpec(new AppAction("selectParent"),
					      KeyEvent.VK_UP),
		    new MenuUtil.MenuItemSpec(new AppAction("selectChild"),
					      KeyEvent.VK_DOWN),
		    new MenuUtil.MenuItemSpec(
			new AppAction("selectPreceeding"),
			KeyEvent.VK_UP, 
			AWTUtil.KEY_MODIFIER|Event.SHIFT_MASK),
		    new MenuUtil.MenuItemSpec(
			new AppAction("selectFollowing"),
			KeyEvent.VK_DOWN, 
			AWTUtil.KEY_MODIFIER|Event.SHIFT_MASK),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuItemSpec(new AppAction("selectAll"))
		}),

	    new MenuUtil.MenuSpec("edit",
		new MenuUtil.MenuItemSpec[] {
		    new MenuUtil.MenuItemSpec(new AppAction("undo"), 
					      KeyEvent.VK_Z),
		    new MenuUtil.MenuItemSpec(new AppAction("redo"), 
					      KeyEvent.VK_Y),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuItemSpec(new AppAction("cut"), 
					      KeyEvent.VK_X),
		    new MenuUtil.MenuItemSpec(new AppAction("copy"), 
					      KeyEvent.VK_C),
		    new MenuUtil.MenuItemSpec(new AppAction("pasteBefore"), 
					      KeyEvent.VK_U),
		    new MenuUtil.MenuItemSpec(new AppAction("paste"), 
					      KeyEvent.VK_V),
		    new MenuUtil.MenuItemSpec(new AppAction("pasteAfter"), 
					      KeyEvent.VK_W),
		    new MenuUtil.MenuItemSpec(new AppAction("delete"), 
					      KeyEvent.VK_K),
		    new MenuUtil.MenuItemSpec(new AppAction("replace"), 
					      KeyEvent.VK_R),
		    new MenuUtil.MenuItemSpec(new AppAction("insertBefore"), 
					      KeyEvent.VK_H),
		    new MenuUtil.MenuItemSpec(new AppAction("insert"), 
					      KeyEvent.VK_I),
		    new MenuUtil.MenuItemSpec(new AppAction("insertAfter"), 
					      KeyEvent.VK_J),
		    new MenuUtil.MenuItemSpec(new AppAction("split"), 
					      KeyEvent.VK_ENTER),
		    new MenuUtil.MenuItemSpec(new AppAction("join"), 
					      KeyEvent.VK_BACK_SPACE),
		    new MenuUtil.MenuItemSpec(new AppAction("convert"), 
					      KeyEvent.VK_T),
		    new MenuUtil.MenuItemSpec(
			new AppAction("applyLastConversion"), 
			KeyEvent.VK_A),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuSpec("insertFile",
		        new MenuUtil.MenuItemSpec[] {
			    new MenuUtil.MenuItemSpec(
				new AppAction("pasteFileBefore")),
			    new MenuUtil.MenuItemSpec(
				new AppAction("pasteFile")),
			    new MenuUtil.MenuItemSpec(
				new AppAction("pasteFileAfter"))
		        })
		}),

	    new MenuUtil.MenuSpec("search",
		new MenuUtil.MenuItemSpec[] {
		    new MenuUtil.MenuItemSpec(new AppAction("search"), 
					      KeyEvent.VK_F),
		    new MenuUtil.MenuItemSpec(
			new AppAction("searchBackward"),
			KeyEvent.VK_F, 
			AWTUtil.KEY_MODIFIER|Event.SHIFT_MASK),
		    new MenuUtil.MenuItemSpec(new AppAction("findNext"), 
					      KeyEvent.VK_G),
		    new MenuUtil.MenuItemSpec(
			new AppAction("findPrevious"), 
			KeyEvent.VK_G, 
			AWTUtil.KEY_MODIFIER|Event.SHIFT_MASK),
		    new MenuUtil.MenuItemSpec(new AppAction("searchReplace"), 
					      KeyEvent.VK_M),
		    new MenuUtil.MenuItemSpec(
			new AppAction("searchReplaceBackward"), 
			KeyEvent.VK_M,
			AWTUtil.KEY_MODIFIER|Event.SHIFT_MASK)
		}),

	    new MenuUtil.MenuSpec("style",
		new MenuUtil.MenuItemSpec[] {
		    new MenuUtil.MenuItemSpec(new AppAction("noStyle"))
		}),

	    new MenuUtil.MenuSpec("tools",
		new MenuUtil.MenuItemSpec[] {
		    new MenuUtil.MenuItemSpec(new AppAction("editAttributes"),
					      KeyEvent.VK_E),
		    new MenuUtil.MenuItemSpec(new AppAction("checkValidity")),
		    new MenuUtil.MenuItemSpec(
			new AppAction("checkSpelling"), 
			KeyEvent.VK_S,
			AWTUtil.KEY_MODIFIER|Event.SHIFT_MASK),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuItemSpec(new AppAction("refreshSelected"),
					      KeyEvent.VK_L)
		}),

	    new MenuUtil.MenuSpec("options",
		new MenuUtil.MenuItemSpec[] {
		    new MenuUtil.MenuItemSpec(new AppAction("options")),
		    new MenuUtil.MenuItemSpec(),
		    new MenuUtil.MenuItemSpec(
			new AppAction("structuredFormats"))
		})
	};

	menus = MenuUtil.addMenus(bar, menuSpecs, resourceBundle);

	JMenu editMenu = MenuUtil.getMenu(menus, "edit");
	editMenu.addMenuListener(this);
    }

    public void menuSelected(MenuEvent e) {
	for (int i = 0; i < editMenuItemNames.length; ++i) {
	    JMenuItem item = MenuUtil.getMenuItem(menus, editMenuItemNames[i]);

	    int command = editMenuItemCommands[i];
	    String label = commandInfoToLabel(commandInfo, command);

	    switch (command) {
	    case ContextCommandInfo.REPLACE:
	    case ContextCommandInfo.INSERT_BEFORE: 
	    case ContextCommandInfo.INSERT: 
	    case ContextCommandInfo.INSERT_AFTER:
	    case ContextCommandInfo.CONVERT: 
		label += "...";
		break;
	    }

	    item.setText(label);
	}

	for (int i = 0; i < insertFileMenuItemNames.length; ++i) {
	    JMenuItem item = MenuUtil.getMenuItem(menus, 
						  insertFileMenuItemNames[i]);

	    int  command = insertFileMenuItemCommands[i];
	    String label = commandInfoToLabel(commandInfo, command);

	    item.setText(label + "...");
	}
    }
    public void menuDeselected(MenuEvent e) {}
    public void menuCanceled(MenuEvent e) {}

    private String commandInfoToLabel(ContextCommandInfo commandInfo, 
				      int command) {
	String label = null;
	String arg0, arg1;

	switch (commandInfo.context) {
	case ContextCommandInfo.TEXT_SELECTION:
	    switch (command) {
	    case ContextCommandInfo.CUT:
		label = msg("edit.cut.sel");
		break;
	    case ContextCommandInfo.COPY:
		label = msg("edit.copy.sel");
		break;
	    case ContextCommandInfo.PASTE:
		label = msg("edit.paste.sel");
		break;
	    case ContextCommandInfo.PASTE_BEFORE:
		label = msg("edit.pasteBefore.sel");
		break;
	    case ContextCommandInfo.PASTE_AFTER:
		label = msg("edit.pasteAfter.sel");
		break;
	    case ContextCommandInfo.DELETE:
		label = msg("edit.delete.sel");
		break;
	    case ContextCommandInfo.REPLACE:
		label = msg("edit.replace.sel");
		break;
	    case ContextCommandInfo.INSERT:
		label = msg("edit.insert.sel");
		break;
	    case ContextCommandInfo.INSERT_BEFORE:
		label = msg("edit.insertBefore.sel");
		break;
	    case ContextCommandInfo.INSERT_AFTER:
		label = msg("edit.insertAfter.sel");
		break;
	    case ContextCommandInfo.SPLIT:
		label = msg("edit.split.sel");
		break;
	    case ContextCommandInfo.JOIN:
		label = msg("edit.join.sel");
		break;
	    case ContextCommandInfo.CONVERT:
		label = msg("edit.convert.sel");
		break;
	    }
	    break;
	case ContextCommandInfo.SUBELEMENT_SELECTION:
	    if (commandInfo.subelement.getSubelementType() == 
		SubelementModel.TEXT) {
		arg0 = "#PCDATA";
		arg1 = "";
	    } else {
		arg0 = "<" +
		  ((ElementModel)commandInfo.subelement).getElementTypeName() +
		  ">";
		arg1 = msg("element");
	    }
	    switch (command) {
	    case ContextCommandInfo.CUT:
		label = msg("edit.cut.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.COPY:
		label = msg("edit.copy.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.PASTE:
		label = msg("edit.paste.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.PASTE_BEFORE:
		label = msg("edit.pasteBefore.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.PASTE_AFTER:
		label = msg("edit.pasteAfter.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.DELETE:
		label = msg("edit.delete.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.REPLACE:
		label = msg("edit.replace.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.INSERT:
		label = msg("edit.insert.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.INSERT_BEFORE:
		label = msg("edit.insertBefore.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.INSERT_AFTER:
		label = msg("edit.insertAfter.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.SPLIT:
		label = msg("edit.split.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.JOIN:
		label = msg("edit.join.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.CONVERT:
		label = msg("edit.convert.sub", arg0, arg1);
		break;
	    }
	    break;
	case ContextCommandInfo.POINT_IN_ELEMENT:
	    arg0 = "<" +
		commandInfo.pointText.getParentElement().getElementTypeName() +
		">";
	    arg1 = msg("element");
	    switch (command) {
	    case ContextCommandInfo.CUT:
		label = msg("edit.cut.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.COPY:
		label = msg("edit.copy.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.PASTE:
		label = msg("edit.paste.point");
		break;
	    case ContextCommandInfo.PASTE_BEFORE:
		label = msg("edit.pasteBefore.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.PASTE_AFTER:
		label = msg("edit.pasteAfter.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.DELETE:
		label = msg("edit.delete.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.REPLACE:
		label = msg("edit.replace.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.INSERT:
		label = msg("edit.insert.point");
		break;
	    case ContextCommandInfo.INSERT_BEFORE:
		label = msg("edit.insertBefore.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.INSERT_AFTER:
		label = msg("edit.insertAfter.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.SPLIT:
		label = msg("edit.split.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.JOIN:
		label = msg("edit.join.sub", arg0, arg1);
		break;
	    case ContextCommandInfo.CONVERT:
		label = msg("edit.convert.sub", arg0, arg1);
		break;
	    }
	    break;
	case ContextCommandInfo.POINT:
	    switch (command) {
	    case ContextCommandInfo.CUT:
		label = msg("edit.cut.point");
		break;
	    case ContextCommandInfo.COPY:
		label = msg("edit.copy.point");
		break;
	    case ContextCommandInfo.PASTE:
		label = msg("edit.paste.point");
		break;
	    case ContextCommandInfo.PASTE_BEFORE:
		label = msg("edit.pasteBefore.point");
		break;
	    case ContextCommandInfo.PASTE_AFTER:
		label = msg("edit.pasteAfter.point");
		break;
	    case ContextCommandInfo.DELETE:
		label = msg("edit.delete.point");
		break;
	    case ContextCommandInfo.REPLACE:
		label = msg("edit.replace.point");
		break;
	    case ContextCommandInfo.INSERT:
		label = msg("edit.insert.point");
		break;
	    case ContextCommandInfo.INSERT_BEFORE:
		label = msg("edit.insertBefore.point");
		break;
	    case ContextCommandInfo.INSERT_AFTER:
		label = msg("edit.insertAfter.point");
		break;
	    case ContextCommandInfo.SPLIT:
		label = msg("edit.split.point");
		break;
	    case ContextCommandInfo.JOIN:
		label = msg("edit.join.point");
		break;
	    case ContextCommandInfo.CONVERT:
		label = msg("edit.convert.point");
		break;
	    }
	    break;
	default:
	    switch (command) {
	    case ContextCommandInfo.CUT:
		label = msg("edit.cut");
		break;
	    case ContextCommandInfo.COPY:
		label = msg("edit.copy");
		break;
	    case ContextCommandInfo.PASTE:
		label = msg("edit.paste");
		break;
	    case ContextCommandInfo.PASTE_BEFORE:
		label = msg("edit.pasteBefore");
		break;
	    case ContextCommandInfo.PASTE_AFTER:
		label = msg("edit.pasteAfter");
		break;
	    case ContextCommandInfo.DELETE:
		label = msg("edit.delete");
		break;
	    case ContextCommandInfo.REPLACE:
		label = msg("edit.replace");
		break;
	    case ContextCommandInfo.INSERT:
		label = msg("edit.insert");
		break;
	    case ContextCommandInfo.INSERT_BEFORE:
		label = msg("edit.insertBefore");
		break;
	    case ContextCommandInfo.INSERT_AFTER:
		label = msg("edit.insertAfter");
		break;
	    case ContextCommandInfo.SPLIT:
		label = msg("edit.split");
		break;
	    case ContextCommandInfo.JOIN:
		label = msg("edit.join");
		break;
	    case ContextCommandInfo.CONVERT:
		label = msg("edit.convert");
		break;
	    }
	}

	return label;
    }

    private class AppAction extends AbstractAction {
	public AppAction(String name) {
	    super(name);
	}

	public void actionPerformed(ActionEvent e) {
	    String command = (String) getValue(Action.NAME);
	    try {
		App.class.getMethod(command, noParamTypes).invoke(App.this, 
								  noArgs);
	    } catch (Exception ex) {
		Throwable ex2;

		if (ex instanceof InvocationTargetException)
		    ex2 = ((InvocationTargetException)ex).getTargetException();
		else
		    ex2 = ex;

		String message = ex2.getMessage();
		if (message == null)
		    message = ex2.toString();
		else
		    message = ex2.toString() + ": " + message;
		System.err.println(msg("internalError", message));
		ex2.printStackTrace();
	    }
	}
    }

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    public static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    public static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    public static String msg(String id, Object arg0, Object arg1) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0, arg1 });
    }

    public static String msg(String id, Object arg0, Object arg1, 
			     Object arg2) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, 
				    new Object[] { arg0, arg1, arg2 });
    }

    public static String msg(String id, Object arg0, Object arg1, 
			     Object arg2, Object arg3) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, 
				    new Object[] { arg0, arg1, arg2, arg3 });
    }

    public static String msg(String id, Object arg0, Object arg1, 
			     Object arg2, Object arg3, Object arg4) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, 
				    new Object[] { arg0,arg1,arg2,arg3,arg4 });
    }
}
