/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.io.File;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import fr.pixware.util.StringUtil;
import com.xmlmind.awtutil.MultiLineLabel;
import com.xmlmind.xmledit.styledview.StyledViewFactory;

public class OptionsDialog extends JDialog implements ActionListener {
    private Frame frame;
    private JTabbedPane tabs;

    private JCheckBox guessElementContent;
    private JCheckBox ignoreStyleSheets;

    private JComboBox encoding;
    private JCheckBox indentIfDocType;
    private JCheckBox indentIfNoDocType;
    private JTextField indentation;
    private JTextField maxLineLength;

    private JTextField screenResolutionWhenPrinting;
    private JTextField headerBegin;
    private JTextField headerMiddle;
    private JTextField headerEnd;
    private JButton headerColor;
    private JCheckBox underlineHeader;
    private JTextField footerBegin;
    private JTextField footerMiddle;
    private JTextField footerEnd;
    private JButton footerColor;
    private JCheckBox overlineFooter;

    private JTextField maxUndo;
    private ButtonGroup makeChoiceGroup;
    private JRadioButton makeNoChoice;
    private JRadioButton makeFirstChoice;
    private JRadioButton makeSmartChoice;

    private JCheckBox ignoreCase;
    private JCheckBox wholeWord;
    private JCheckBox regularExpression;
    private ButtonGroup markParagraphsGroup;
    private JRadioButton markNone;
    private JRadioButton markAll;
    private JRadioButton markLast;

    private JComboBox serifFontFamily;
    private JComboBox sansSerifFontFamily;
    private JComboBox monospaceFontFamily;
    private ButtonGroup defaultFontGroup;
    private JRadioButton defaultFontIsSerif;
    private JRadioButton defaultFontIsSans;
    private JComboBox defaultFontSize;
    private JButton defaultBackground;
    private JButton defaultForeground;

    private Options result = null;

    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.OptionsDialogRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    // -----------------------------------------------------------------------

    public OptionsDialog(Frame frame) {
	super(frame, msg("title"), /*modal*/ true); 

	this.frame = frame;

   	addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        cancel();
   	    }
   	});

        // workArea is a JPanel with a BorderLayout.
	JPanel workArea = (JPanel) getContentPane();
	workArea.setLayout(new GridBagLayout());

	// Should be implemented by all LAFs.
	workArea.registerKeyboardAction(
	    this, "cancel", KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), 
	    JPanel.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
        constraints.gridx = 0;
        constraints.gridy = 0;

	// Tabs ---

	tabs = new JTabbedPane();
	tabs.addTab(msg("load"), loadTab());
	tabs.addTab(msg("save"), saveTab());
	tabs.addTab(msg("print"), printTab());
	tabs.addTab(msg("edit"), editTab());
	tabs.addTab(msg("search"), searchTab());
	tabs.addTab(msg("style"), styleTab());

	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
	workArea.add(tabs, constraints);
	++constraints.gridy;

	// Buttons ---

	JPanel buttonFrame = new JPanel(new GridLayout(1, 2, 10, 0));
	constraints.fill = GridBagConstraints.NONE;
	constraints.weightx = 0;
	constraints.weighty = 0;
        constraints.insets.bottom = 5;
	workArea.add(buttonFrame, constraints);

        JButton button = new JButton(msg("ok"));
	button.setActionCommand("ok");
	button.addActionListener(this);
        buttonFrame.add(button);
	getRootPane().setDefaultButton(button);
          
        button = new JButton(msg("cancel"));
        button.setActionCommand("cancel");
        button.addActionListener(this);
        buttonFrame.add(button);

        pack();
	setResizable(true);
    }

    private JPanel loadTab() {
	JPanel loadTab = new JPanel(new GridBagLayout());

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	constraints.anchor = GridBagConstraints.WEST;

	guessElementContent = new JCheckBox();
        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.insets.right = 0;
	loadTab.add(guessElementContent, constraints);

	MultiLineLabel label = new MultiLineLabel(msg("guessElementContent"));
        constraints.gridx = 1;
        constraints.gridy = 0;
        constraints.insets.right = 5;
	loadTab.add(label, constraints);

	ignoreStyleSheets = new JCheckBox();
        constraints.gridx = 0;
        constraints.gridy = 1;
        constraints.insets.right = 0;
	loadTab.add(ignoreStyleSheets, constraints);

	label = new MultiLineLabel(msg("ignoreStyleSheets"));
        constraints.gridx = 1;
        constraints.gridy = 1;
        constraints.insets.right = 5;
	loadTab.add(label, constraints);

	return loadTab;
    }

    private JPanel saveTab() {
	JPanel saveTab = new JPanel(new GridBagLayout());

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	constraints.anchor = GridBagConstraints.WEST;

	MultiLineLabel label = new MultiLineLabel(msg("encoding"));
        constraints.gridx = 0;
        constraints.gridy = 0;
	saveTab.add(label, constraints);

	Font labelFont = label.getFont();

	encoding = new JComboBox(Options.IANA_ENCODINGS);
	encoding.setEditable(false);
	encoding.setFont(labelFont);
        constraints.gridx = 1;
        constraints.gridy = 0;
	saveTab.add(encoding, constraints);

	label = new MultiLineLabel(msg("indent"));
        constraints.gridx = 0;
        constraints.gridy = 1;
	saveTab.add(label, constraints);

	indentIfDocType = new JCheckBox(msg("indentIfDocType"));
	indentIfDocType.setFont(labelFont);
        constraints.gridx = 1;
        constraints.gridy = 1;
	saveTab.add(indentIfDocType, constraints);

	indentIfNoDocType = new JCheckBox(msg("indentIfNoDocType"));
	indentIfNoDocType.setFont(labelFont);
        constraints.gridx = 2;
        constraints.gridy = 1;
	saveTab.add(indentIfNoDocType, constraints);

	label = new MultiLineLabel(msg("indentation"));
        constraints.gridx = 0;
        constraints.gridy = 2;
	saveTab.add(label, constraints);

	indentation = new JTextField(15);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 2;
	saveTab.add(indentation, constraints);
	indentation.setActionCommand("ok");
	indentation.addActionListener(this);

	label = new MultiLineLabel(msg("maxLineLength"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 3;
	saveTab.add(label, constraints);

	maxLineLength = new JTextField(15);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 3;
	saveTab.add(maxLineLength, constraints);
	maxLineLength.setActionCommand("ok");
	maxLineLength.addActionListener(this);

	return saveTab;
    }

    private JPanel printTab() {
	JPanel printTab = new JPanel(new GridBagLayout());

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	constraints.anchor = GridBagConstraints.WEST;

	MultiLineLabel label = 
	    new MultiLineLabel(msg("screenResolutionWhenPrinting"));
        constraints.gridx = 0;
        constraints.gridy = 0;
	printTab.add(label, constraints);

	screenResolutionWhenPrinting = new JTextField(15);
        constraints.gridx = 1;
        constraints.gridy = 0;
	printTab.add(screenResolutionWhenPrinting, constraints);
	screenResolutionWhenPrinting.setActionCommand("ok");
	screenResolutionWhenPrinting.addActionListener(this);

	JPanel headerFrame = new JPanel(new GridBagLayout());
	headerFrame.setBorder(new TitledBorder(msg("header")));
        constraints.gridx = 0;
	constraints.gridwidth = 2;
        constraints.gridy = 1;
	printTab.add(headerFrame, constraints);

	JPanel footerFrame = new JPanel(new GridBagLayout());
	footerFrame.setBorder(new TitledBorder(msg("footer")));
        constraints.gridy = 2;
	printTab.add(footerFrame, constraints);

	// Header ---

	label = new MultiLineLabel(msg("headerBegin"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 0;
	headerFrame.add(label, constraints);

	headerBegin = new JTextField(25);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 0;
	headerFrame.add(headerBegin, constraints);
	headerBegin.setActionCommand("ok");
	headerBegin.addActionListener(this);

	label = new MultiLineLabel(msg("headerMiddle"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 1;
	headerFrame.add(label, constraints);

	headerMiddle = new JTextField(25);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 1;
	headerFrame.add(headerMiddle, constraints);
	headerMiddle.setActionCommand("ok");
	headerMiddle.addActionListener(this);

	label = new MultiLineLabel(msg("headerEnd"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 2;
	headerFrame.add(label, constraints);

	headerEnd = new JTextField(25);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 2;
	headerFrame.add(headerEnd, constraints);
	headerEnd.setActionCommand("ok");
	headerEnd.addActionListener(this);

	label = new MultiLineLabel(msg("headerColor"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 3;
	headerFrame.add(label, constraints);

	Font labelFont = label.getFont();

	headerColor = new JButton();
	headerColor.setPreferredSize(new Dimension(40, 20));
        constraints.gridx = 1;
        constraints.gridy = 3;
	headerFrame.add(headerColor, constraints);
	headerColor.setActionCommand("selectColor");
	headerColor.addActionListener(this);

	underlineHeader = new JCheckBox(msg("underlineHeader"));
	underlineHeader.setFont(labelFont);
        constraints.gridx = 2;
        constraints.gridy = 3;
	constraints.anchor = GridBagConstraints.EAST;
	headerFrame.add(underlineHeader, constraints);

	// Footer ---

	label = new MultiLineLabel(msg("footerBegin"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 0;
	constraints.anchor = GridBagConstraints.WEST;
	footerFrame.add(label, constraints);

	footerBegin = new JTextField(25);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 0;
	footerFrame.add(footerBegin, constraints);
	footerBegin.setActionCommand("ok");
	footerBegin.addActionListener(this);

	label = new MultiLineLabel(msg("footerMiddle"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 1;
	footerFrame.add(label, constraints);

	footerMiddle = new JTextField(25);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 1;
	footerFrame.add(footerMiddle, constraints);
	footerMiddle.setActionCommand("ok");
	footerMiddle.addActionListener(this);

	label = new MultiLineLabel(msg("footerEnd"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 2;
	footerFrame.add(label, constraints);

	footerEnd = new JTextField(25);
        constraints.gridx = 1;
	constraints.gridwidth = 2;
        constraints.gridy = 2;
	footerFrame.add(footerEnd, constraints);
	footerEnd.setActionCommand("ok");
	footerEnd.addActionListener(this);

	label = new MultiLineLabel(msg("footerColor"));
        constraints.gridx = 0;
	constraints.gridwidth = 1;
        constraints.gridy = 3;
	footerFrame.add(label, constraints);

	footerColor = new JButton();
	footerColor.setPreferredSize(new Dimension(40, 20));
        constraints.gridx = 1;
        constraints.gridy = 3;
	footerFrame.add(footerColor, constraints);
	footerColor.setActionCommand("selectColor");
	footerColor.addActionListener(this);

	overlineFooter = new JCheckBox(msg("overlineFooter"));
	overlineFooter.setFont(labelFont);
        constraints.gridx = 2;
        constraints.gridy = 3;
	constraints.anchor = GridBagConstraints.EAST;
	footerFrame.add(overlineFooter, constraints);

	label = new MultiLineLabel(msg("helpAboutVariables"));
	label.setFont(new Font(labelFont.getFamily(), Font.ITALIC,
			       labelFont.getSize()));
	label.setForeground(
	    ((TitledBorder) headerFrame.getBorder()).getTitleColor());
        constraints.gridx = 0;
	constraints.gridwidth = 2;
        constraints.gridy = 3;
	constraints.anchor = GridBagConstraints.CENTER;
	printTab.add(label, constraints);

	return printTab;
    }

    private JPanel editTab() {
	JPanel editTab = new JPanel(new GridBagLayout());

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	constraints.anchor = GridBagConstraints.WEST;

	MultiLineLabel label = new MultiLineLabel(msg("maxUndo"));
        constraints.gridx = 0;
        constraints.gridy = 0;
	editTab.add(label, constraints);

	maxUndo = new JTextField(15);
        constraints.gridx = 1;
        constraints.gridy = 0;
	editTab.add(maxUndo, constraints);
	maxUndo.setActionCommand("ok");
	maxUndo.addActionListener(this);

	label = new MultiLineLabel(msg("makeChoice"));
        constraints.gridx = 0;
        constraints.gridwidth = 2;
        constraints.gridy = 1;
	editTab.add(label, constraints);

	Font labelFont = label.getFont();

	makeNoChoice = new JRadioButton(msg("makeNoChoice"));
	makeNoChoice.setFont(labelFont);
        constraints.gridx = 0;
        constraints.gridy = 2;
        constraints.insets.top = 0;
	editTab.add(makeNoChoice, constraints);

	makeFirstChoice = new JRadioButton(msg("makeFirstChoice"));
	makeFirstChoice.setFont(labelFont);
        constraints.gridx = 0;
        constraints.gridy = 3;
	editTab.add(makeFirstChoice, constraints);

	makeSmartChoice = new JRadioButton(msg("makeSmartChoice"));
	makeSmartChoice.setFont(labelFont);
        constraints.gridx = 0;
        constraints.gridy = 4;
	editTab.add(makeSmartChoice, constraints);

	makeChoiceGroup = new ButtonGroup();
	makeChoiceGroup.add(makeNoChoice);
	makeChoiceGroup.add(makeFirstChoice);
	makeChoiceGroup.add(makeSmartChoice);

	return editTab;
    }

    private JPanel searchTab() {
	JPanel searchTab = new JPanel(new GridBagLayout());

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	constraints.anchor = GridBagConstraints.WEST;

	// Text boundaries ---

	JPanel markParagraphsFrame = new JPanel(new GridBagLayout());
	markParagraphsFrame.setBorder(new TitledBorder(msg("markParagraphs")));
        constraints.gridx = 0;
        constraints.gridy = 0;
	searchTab.add(markParagraphsFrame, constraints);

	markLast = new JRadioButton(" ");
	markLast.setHorizontalTextPosition(JRadioButton.LEFT);
        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.insets.left = 5;
        constraints.insets.right = 0;
	markParagraphsFrame.add(markLast, constraints);

	MultiLineLabel label = new MultiLineLabel(msg("markLast"));
	Font labelFont = label.getFont();
        constraints.gridx = 1;
        constraints.gridy = 0;
        constraints.insets.left = 0;
        constraints.insets.right = 5;
	markParagraphsFrame.add(label, constraints);

	markNone = new JRadioButton(" ");
	markNone.setHorizontalTextPosition(JRadioButton.LEFT);
        constraints.gridx = 0;
        constraints.gridy = 1;
        constraints.insets.left = 5;
        constraints.insets.right = 0;
	markParagraphsFrame.add(markNone, constraints);

	label = new MultiLineLabel(msg("markNone"));
        constraints.gridx = 1;
        constraints.gridy = 1;
        constraints.insets.left = 0;
        constraints.insets.right = 5;
	markParagraphsFrame.add(label, constraints);

	markAll = new JRadioButton(" ");
	markAll.setHorizontalTextPosition(JRadioButton.LEFT);
        constraints.gridx = 0;
        constraints.gridy = 2;
        constraints.insets.left = 5;
        constraints.insets.right = 0;
	markParagraphsFrame.add(markAll, constraints);

	label = new MultiLineLabel(msg("markAll"));
        constraints.gridx = 1;
        constraints.gridy = 2;
        constraints.insets.left = 0;
        constraints.insets.right = 5;
	markParagraphsFrame.add(label, constraints);

	markParagraphsGroup = new ButtonGroup();
	markParagraphsGroup.add(markLast);
	markParagraphsGroup.add(markNone);
	markParagraphsGroup.add(markAll);

	// Search options ---

	ignoreCase = new JCheckBox(msg("ignoreCase"));
	ignoreCase.setFont(labelFont);
        constraints.gridx = 0;
        constraints.gridy = 1;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	searchTab.add(ignoreCase, constraints);

	wholeWord = new JCheckBox(msg("wholeWord"));
	wholeWord.setFont(labelFont);
        constraints.gridx = 0;
        constraints.gridy = 2;
	searchTab.add(wholeWord, constraints);

	regularExpression = new JCheckBox(msg("regularExpression"));
	regularExpression.setFont(labelFont);
        constraints.gridx = 0;
        constraints.gridy = 3;
	searchTab.add(regularExpression, constraints);

	return searchTab;
    }

    private JPanel styleTab() {
	JPanel styleTab = new JPanel(new GridBagLayout());

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	constraints.anchor = GridBagConstraints.WEST;

	MultiLineLabel label = new MultiLineLabel(msg("serifFontFamily"));
        constraints.gridx = 0;
        constraints.gridwidth = 1;
        constraints.gridy = 0;
	styleTab.add(label, constraints);

	Font labelFont = label.getFont();

	serifFontFamily = new JComboBox(StyledViewFactory.FONT_FAMILIES);
	serifFontFamily.setEditable(false);
	serifFontFamily.setFont(labelFont);
        constraints.gridx = 1;
        constraints.gridwidth = 2;
        constraints.gridy = 0;
	styleTab.add(serifFontFamily, constraints);
	
	label = new MultiLineLabel(msg("sansSerifFontFamily"));
        constraints.gridx = 0;
        constraints.gridwidth = 1;
        constraints.gridy = 1;
	styleTab.add(label, constraints);

	sansSerifFontFamily = new JComboBox(StyledViewFactory.FONT_FAMILIES);
	sansSerifFontFamily.setEditable(false);
	sansSerifFontFamily.setFont(labelFont);
        constraints.gridx = 1;
        constraints.gridwidth = 2;
        constraints.gridy = 1;
	styleTab.add(sansSerifFontFamily, constraints);

	label = new MultiLineLabel(msg("monospaceFontFamily"));
        constraints.gridx = 0;
        constraints.gridwidth = 1;
        constraints.gridy = 2;
	styleTab.add(label, constraints);

	monospaceFontFamily = new JComboBox(StyledViewFactory.FONT_FAMILIES);
	monospaceFontFamily.setEditable(false);
	monospaceFontFamily.setFont(labelFont);
        constraints.gridx = 1;
        constraints.gridwidth = 2;
        constraints.gridy = 2;
	styleTab.add(monospaceFontFamily, constraints);

	label = new MultiLineLabel(msg("defaultFontFamily"));
        constraints.gridx = 0;
        constraints.gridwidth = 1;
        constraints.gridy = 3;
	styleTab.add(label, constraints);

	defaultFontIsSerif = new JRadioButton(msg("serif"));
	defaultFontIsSerif.setFont(labelFont);
        constraints.gridx = 1;
        constraints.gridy = 3;
	styleTab.add(defaultFontIsSerif, constraints);

	defaultFontIsSans = new JRadioButton(msg("sansSerif"));
	defaultFontIsSans.setFont(labelFont);
        constraints.gridx = 2;
        constraints.gridy = 3;
	styleTab.add(defaultFontIsSans, constraints);

	defaultFontGroup = new ButtonGroup();
	defaultFontGroup.add(defaultFontIsSerif);
	defaultFontGroup.add(defaultFontIsSans);

	label = new MultiLineLabel(msg("defaultFontSize"));
        constraints.gridx = 0;
        constraints.gridwidth = 1;
        constraints.gridy = 4;
	styleTab.add(label, constraints);

	defaultFontSize = new JComboBox();
	for (int i = 10; i <= 20; ++i)
	    defaultFontSize.addItem(new Integer(i));
	defaultFontSize.setEditable(false);
	defaultFontSize.setFont(labelFont);
        constraints.gridx = 1;
        constraints.gridwidth = 2;
        constraints.gridy = 4;
	styleTab.add(defaultFontSize, constraints);

	label = new MultiLineLabel(msg("defaultBackground"));
        constraints.gridx = 0;
        constraints.gridwidth = 1;
        constraints.gridy = 5;
	styleTab.add(label, constraints);

	defaultBackground = new JButton();
	defaultBackground.setPreferredSize(new Dimension(40, 20));
        constraints.gridx = 1;
        constraints.gridwidth = 2;
        constraints.gridy = 5;
	styleTab.add(defaultBackground, constraints);
	defaultBackground.setActionCommand("selectColor");
	defaultBackground.addActionListener(this);

	label = new MultiLineLabel(msg("defaultForeground"));
        constraints.gridx = 0;
        constraints.gridwidth = 1;
        constraints.gridy = 6;
	styleTab.add(label, constraints);

	defaultForeground = new JButton();
	defaultForeground.setPreferredSize(new Dimension(40, 20));
        constraints.gridx = 1;
        constraints.gridwidth = 2;
        constraints.gridy = 6;
	styleTab.add(defaultForeground, constraints);
	defaultForeground.setActionCommand("selectColor");
	defaultForeground.addActionListener(this);

	return styleTab;
    }

    public Options getOptions(Options in) {
	guessElementContent.setSelected(in.guessElementContent);
	ignoreStyleSheets.setSelected(in.ignoreStyleSheets);

	int encodingIndex = 
	    StringUtil.indexOf(Options.JAVA_ENCODINGS, in.encoding);
	if (encodingIndex < 0)
	    encodingIndex = 0;
	encoding.setSelectedItem(Options.IANA_ENCODINGS[encodingIndex]);
	indentIfDocType.setSelected(in.indentIfDocType);
	indentIfNoDocType.setSelected(in.indentIfNoDocType);
	indentation.setText(Integer.toString(in.indentation));
	maxLineLength.setText(Integer.toString(in.maxLineLength));

	screenResolutionWhenPrinting.setText(
	    Double.toString(in.screenResolutionWhenPrinting));
	headerBegin.setText((in.headerBegin == null)? "" : in.headerBegin);
	headerMiddle.setText((in.headerMiddle == null)? "" : in.headerMiddle);
	headerEnd.setText((in.headerEnd == null)? "" : in.headerEnd);
	headerColor.setBackground(in.headerColor);
	underlineHeader.setSelected(in.underlineHeader);
	footerBegin.setText((in.footerBegin == null)? "" : in.footerBegin);
	footerMiddle.setText((in.footerMiddle == null)? "" : in.footerMiddle);
	footerEnd.setText((in.footerEnd == null)? "" : in.footerEnd);
	footerColor.setBackground(in.footerColor);
	overlineFooter.setSelected(in.overlineFooter);

	maxUndo.setText(Integer.toString(in.maxUndo));
	makeNoChoice.setSelected(in.makeChoice == Options.NO_CHOICE);
	makeFirstChoice.setSelected(in.makeChoice == Options.FIRST_CHOICE);
	makeSmartChoice.setSelected(in.makeChoice == Options.SMART_CHOICE);

	ignoreCase.setSelected(in.ignoreCase);
	wholeWord.setSelected(in.wholeWord);
	regularExpression.setSelected(in.regularExpression);
	markNone.setSelected(in.markParagraphs == Options.MARK_NONE);
	markAll.setSelected(in.markParagraphs == Options.MARK_ALL);
	markLast.setSelected(in.markParagraphs == Options.MARK_LAST);
	
	serifFontFamily.setSelectedItem(in.serifFontFamily);
	sansSerifFontFamily.setSelectedItem(in.sansSerifFontFamily);
	monospaceFontFamily.setSelectedItem(in.monospaceFontFamily);
	defaultFontIsSerif.setSelected(in.defaultFontIsSerif);
	defaultFontIsSans.setSelected(!in.defaultFontIsSerif);
	defaultFontSize.setSelectedItem(new Integer(in.defaultFontSize));
	defaultBackground.setBackground(in.defaultBackground);
	defaultForeground.setBackground(in.defaultForeground);

	setLocationRelativeTo(frame);
	pack();
        show();
        // Modal means blocked here until OK or Cancel are pressed.

        return result;
    }

    // -----------------------------------------------------------------------

    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();

        if (command.equals("selectColor"))
            selectColor((Component) e.getSource());
        else if (command.equals("ok"))
            ok();
        else if (command.equals("cancel"))
            cancel();
    }

    private void selectColor(Component component) {
	String title = "";
	if (component == defaultBackground)
	    title = "defaultBackground";
	else if (component == defaultBackground)
	    title = "defaultForeground";
	else if (component == headerColor)
	    title = "headerColor";
	else if (component == footerColor)
	    title = "footerColor";

	String localizedTitle = msg(title);
	if (localizedTitle.endsWith(":"))
	    localizedTitle = 
		localizedTitle.substring(0, localizedTitle.length()-1);

	Color color = JColorChooser.showDialog(frame, localizedTitle, 
					       component.getBackground());
	if (color != null)
	    component.setBackground(color);
    }

    private void ok() {
	result = new Options();

	// Tab #0 ---

	result.guessElementContent = guessElementContent.isSelected();
	result.ignoreStyleSheets = ignoreStyleSheets.isSelected();
	
	// Tab #1 ---

	int encodingIndex = 
	    StringUtil.indexOf(Options.IANA_ENCODINGS, 
			       (String) encoding.getSelectedItem());
	if (encodingIndex < 0)
	    encodingIndex = 0;
	result.encoding = Options.JAVA_ENCODINGS[encodingIndex];
	result.indentIfDocType = indentIfDocType.isSelected();
	result.indentIfNoDocType = indentIfNoDocType.isSelected();
	result.indentation = 
	    (int) checkNumericField(indentation, 1, 0, 10, -1);
	if (result.indentation == -1)
	    return;
	result.maxLineLength = 
	   (int) checkNumericField(maxLineLength, 1, 1, Integer.MAX_VALUE, -1);
	if (result.maxLineLength == -1)
	    return;

	// Tab #2 ---

	result.screenResolutionWhenPrinting = 
	    checkNumericField(screenResolutionWhenPrinting, 2, 50, 150, -1);
	if (result.screenResolutionWhenPrinting == -1)
	    return;
	result.headerBegin = headerBegin.getText().trim();
	if (result.headerBegin.length() == 0)
	    result.headerBegin = null;
	result.headerMiddle = headerMiddle.getText().trim();
	if (result.headerMiddle.length() == 0)
	    result.headerMiddle = null;
	result.headerEnd = headerEnd.getText().trim();
	if (result.headerEnd.length() == 0)
	    result.headerEnd = null;
	result.headerColor = headerColor.getBackground();
	result.underlineHeader = underlineHeader.isSelected();
	result.footerBegin = footerBegin.getText().trim();
	if (result.footerBegin.length() == 0)
	    result.footerBegin = null;
	result.footerMiddle = footerMiddle.getText().trim();
	if (result.footerMiddle.length() == 0)
	    result.footerMiddle = null;
	result.footerEnd = footerEnd.getText().trim();
	if (result.footerEnd.length() == 0)
	    result.footerEnd = null;
	result.footerColor = footerColor.getBackground();
	result.overlineFooter = overlineFooter.isSelected();

	// Tab #3 ---

	result.maxUndo = (int) checkNumericField(maxUndo, 3, 1, 100, -1);
	if (result.maxUndo == -1)
	    return;

	if (makeNoChoice.isSelected())
	    result.makeChoice = Options.NO_CHOICE;
	else if (makeFirstChoice.isSelected())
	    result.makeChoice = Options.FIRST_CHOICE;
	else
	    result.makeChoice = Options.SMART_CHOICE;

	// Tab #4 ---

	result.ignoreCase = ignoreCase.isSelected();
	result.wholeWord = wholeWord.isSelected();
	result.regularExpression = regularExpression.isSelected();
	if (markNone.isSelected())
	    result.markParagraphs = Options.MARK_NONE;
	else if (markAll.isSelected())
	    result.markParagraphs = Options.MARK_ALL;
	else if (markLast.isSelected())
	    result.markParagraphs = Options.MARK_LAST;

	// Tab #5 ---

	result.serifFontFamily = 
	    (String) serifFontFamily.getSelectedItem();
	result.sansSerifFontFamily = 
	    (String) sansSerifFontFamily.getSelectedItem();
	result.monospaceFontFamily = 
	    (String) monospaceFontFamily.getSelectedItem();
	result.defaultFontIsSerif = defaultFontIsSerif.isSelected();
	result.defaultFontSize = 
	    ((Integer) defaultFontSize.getSelectedItem()).intValue();
	result.defaultBackground = defaultBackground.getBackground();
	result.defaultForeground = defaultForeground.getBackground();

	hide();
	dispose();
    }

    private double checkNumericField(JTextField field, int tab,
				     double min, double max, double invalid) {
	String s = field.getText().trim();
	double number;

	if (s.length() == 0) {
	    number = invalid;
	} else {
	    try {
		number = Double.valueOf(s).doubleValue();
	    } catch (NumberFormatException e) {
		number = invalid;
	    }

	    if (number != invalid) {
		if (number < min || number > max)
		    number = invalid;
	    }
	}

	if (number == invalid) {
	    tabs.setSelectedIndex(tab);
	    field.getToolkit().beep();
	    field.requestFocus();
	    field.selectAll();
	}

	return number;
    }

    private void cancel() {
	result = null;

	hide();
	dispose();
    }
}

