/*
 * Copyright (c) 2000-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.io.*;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Properties;
import java.util.Vector;
import java.util.StringTokenizer;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.print.Paper;
import java.awt.print.PageFormat;
import fr.pixware.util.PlatformUtil;
import fr.pixware.util.FileUtil;
import fr.pixware.util.StringUtil;
import com.xmlmind.xmledit.styledview.StyledViewFactory;

public class Preferences extends Options {
    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.PreferencesRes");
	} catch (MissingResourceException ignored) {}
    }

    public static String msg(String id, Object arg0, Object arg1) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0, arg1 });
    }

    // -----------------------------------------------------------------------
    // Persistent state
    // -----------------------------------------------------------------------

    private static final Rectangle DEFAULT_GEOMETRY = 
        new Rectangle(0, 0, 640, 480);

    private static final Paper newPaper(double paperW, double paperH,
					double x, double y, 
					double width, double height) {
	if (paperW < 0 || paperH < 0 ||
	    x < 0 || x >= paperW ||
	    y < 0 || y >= paperH ||
	    width <= 0 || width > paperW-x ||
	    height <= 0 || height > paperH-y)
	    return null;

	Paper paper = new Paper();
	paper.setSize(paperW, paperH);
	paper.setImageableArea(x, y, width, height);

	return paper;
    }

    private static final Paper DEFAULT_PAPER = 
        newPaper(595, 842, 28, 30, 538, 781); // A4 ``small''

    // -----------------------------------------------------------------------

    public Rectangle geometry = DEFAULT_GEOMETRY;
    public String[] openedFileNames = new String[5];
    public String[] templateNames = new String[20];
    public String[] structuredFormats = new String[10];
    public PageFormat pageFormat = new PageFormat();
    public String spellCheckerLanguage = "en";
    public int spellCheckerMarkParagraphs = MARK_ALL;

    // -----------------------------------------------------------------------

    public Preferences() {
	pageFormat.setPaper(DEFAULT_PAPER);
    }

    public void setOptions(Options options) {
	guessElementContent = options.guessElementContent;
	ignoreStyleSheets = options.ignoreStyleSheets;

	encoding = options.encoding;
	indentIfDocType = options.indentIfDocType;
	indentIfNoDocType = options.indentIfNoDocType;
	indentation = options.indentation;
	maxLineLength = options.maxLineLength;

	screenResolutionWhenPrinting = options.screenResolutionWhenPrinting;
	headerBegin = options.headerBegin;
	headerMiddle = options.headerMiddle;
	headerEnd = options.headerEnd;
	headerColor = options.headerColor;
	underlineHeader = options.underlineHeader;
	footerBegin = options.footerBegin;
	footerMiddle = options.footerMiddle;
	footerEnd = options.footerEnd;
	footerColor = options.footerColor;
	overlineFooter = options.overlineFooter;

	maxUndo = options.maxUndo;
	makeChoice = options.makeChoice;

	ignoreCase = options.ignoreCase;
	wholeWord = options.wholeWord;
	regularExpression = options.regularExpression;
	markParagraphs = options.markParagraphs;

	serifFontFamily = options.serifFontFamily;
	sansSerifFontFamily = options.sansSerifFontFamily;
	monospaceFontFamily = options.monospaceFontFamily;
	defaultFontIsSerif = options.defaultFontIsSerif;
	defaultFontSize = options.defaultFontSize;
	defaultBackground = options.defaultBackground;
	defaultForeground = options.defaultForeground;
    }

    public void addTemplateName(String templateName) {
	templateNames = addToList(templateNames, templateName);
    }

    public void removeTemplateName(String templateName) {
	templateNames = removeFromList(templateNames, templateName);
    }

    public void addOpenedFileName(String fileName) {
	String entry;
	try {
	    entry = (new File(fileName)).getCanonicalPath();
	} catch (IOException e) {
	    entry = fileName;
	}
	openedFileNames = addToList(openedFileNames, entry);
    }

    public void removeOpenedFileName(String fileName) {
	openedFileNames = removeFromList(openedFileNames, fileName);
    }

    public static Preferences load() {
	String rcFileName = PlatformUtil.rcFileName("xxe");
	if (!(new File(rcFileName)).isFile()) {
	    Preferences pref = new Preferences();
	    pref.firstTimeInit();
	    return pref;
	}

	Properties prop = new Properties();

	try {
	    InputStream in = new FileInputStream(rcFileName);
	    prop.load(in);
	    in.close();
	} catch (Exception e) {
	    System.err.println(
		msg("cannotLoadPreferences", rcFileName, e.getMessage()));
			       
	    return new Preferences();
	}

	Preferences pref = null;

	try {
	    pref = parseProperties(prop);
	} catch (Exception e) {
	    System.err.println(
		msg("invalidPreferences", rcFileName, e.getMessage()));

	    return new Preferences();
	}

	return pref;
    }

    private void firstTimeInit() {
	// All the bundled non-XML formats are by default loaded into XXE.
	structuredFormats[0] = "com.xmlmind.javadocformat.Format";
	structuredFormats[1] = "com.xmlmind.aptformat.APTFormat";

	File templateDir = findDistribSubDirectory("template");
	if (templateDir != null) {
	    File[] files = templateDir.listFiles();
	    int j = 0;

	    for (int i = 0; i < files.length; ++i) {
		File file = files[i];
		if (!file.isFile() || !file.canRead())
		    continue;

		String path = file.getPath();
		if (j < templateNames.length &&
		    (path.endsWith(".xml") ||
		     path.endsWith(".xhtml") ||
		     path.endsWith(".html") ||
		     path.endsWith(".sdocb") ||
		     path.endsWith(".docb")))
		    templateNames[j++] = FileUtil.fileToURLName(path);
	    }
	}

	File docsDir = findDistribSubDirectory("docs");
	if (docsDir != null) {
	    File[] files = docsDir.listFiles();
	    int j = 0;

	    for (int i = 0; i < files.length; ++i) {
		File file = files[i];
		if (!file.isFile() || !file.canRead())
		    continue;

		String path = file.getPath();
		if (j < openedFileNames.length &&
		    (path.endsWith(".xml") ||
		     path.endsWith(".xhtml") ||
		     path.endsWith(".sdocb") ||
		     path.endsWith(".docb")))
		    openedFileNames[j++] = file.getAbsolutePath();
	    }
	}
    }

    private static Preferences parseProperties(Properties prop) 
	throws Exception {
	Preferences pref = new Preferences();
	String s;
	int i;

	s = prop.getProperty("guessElementContent");
	if (s != null)
	    pref.guessElementContent = stringToBoolean(s, true);
	s = prop.getProperty("ignoreStyleSheets");
	if (s != null)
	    pref.ignoreStyleSheets = stringToBoolean(s, false);

	s = prop.getProperty("encodeInISO8859_1");
	if (s != null) {
	    if (stringToBoolean(s, false))
		pref.encoding = "ISO8859_1";
	    else
		pref.encoding = "UTF8";
	} else {
	    pref.encoding = prop.getProperty("encoding");
	    if (pref.encoding == null ||
		!StringUtil.contains(JAVA_ENCODINGS, pref.encoding)) 
		// Unknown encoding.
		pref.encoding = "UTF8";
	}
	s = prop.getProperty("indentIfDocType");
	if (s != null)
	    pref.indentIfDocType = stringToBoolean(s, true);
	s = prop.getProperty("indentIfNoDocType");
	if (s != null)
	    pref.indentIfNoDocType = stringToBoolean(s, false);
	s = prop.getProperty("indentation");
	if (s != null)
	    pref.indentation = stringToInt(s, Integer.MIN_VALUE, 100, 2);
	s = prop.getProperty("maxLineLength");
	if (s != null)
	    pref.maxLineLength = stringToInt(s, 1, Integer.MAX_VALUE, 78);

	s = prop.getProperty("screenResolutionWhenPrinting");
	if (s != null)
	    pref.screenResolutionWhenPrinting = 
		stringToDouble(s, 50, 150, 100);
	pref.headerBegin = nonEmptyString(prop.getProperty("headerBegin"));
	pref.headerMiddle = nonEmptyString(prop.getProperty("headerMiddle"));
	pref.headerEnd = nonEmptyString(prop.getProperty("headerEnd"));
	s = prop.getProperty("headerColor");
	if (s != null)
	    pref.headerColor = stringToColor(s, Color.gray);
	s = prop.getProperty("underlineHeader");
	if (s != null)
	    pref.underlineHeader = stringToBoolean(s, true);
	pref.footerBegin = nonEmptyString(prop.getProperty("footerBegin"));
	pref.footerMiddle = nonEmptyString(prop.getProperty("footerMiddle"));
	pref.footerEnd = nonEmptyString(prop.getProperty("footerEnd"));
	s = prop.getProperty("footerColor");
	if (s != null)
	    pref.footerColor = stringToColor(s, Color.gray);
	s = prop.getProperty("overlineFooter");
	if (s != null)
	    pref.overlineFooter = stringToBoolean(s, true);

	s = prop.getProperty("maxUndo");
	if (s != null)
	    pref.maxUndo = stringToInt(s, 1, 100, 10);
	s = prop.getProperty("makeChoice");
	if (s != null)
	    pref.makeChoice = 
		stringToInt(s, NO_CHOICE, SMART_CHOICE, SMART_CHOICE);

	s = prop.getProperty("ignoreCase");
	if (s != null)
	    pref.ignoreCase = stringToBoolean(s, true);
	s = prop.getProperty("wholeWord");
	if (s != null)
	    pref.wholeWord = stringToBoolean(s, false);
	s = prop.getProperty("regularExpression");
	if (s != null)
	    pref.regularExpression = stringToBoolean(s, false);
	s = prop.getProperty("markParagraphs");
	if (s != null)
	    pref.markParagraphs = 
		stringToInt(s, MARK_NONE, MARK_LAST, MARK_ALL);

	s = prop.getProperty("serifFontFamily");
	if (s != null)
	    pref.serifFontFamily = stringToFontFamily(s, "Serif");
	s = prop.getProperty("sansSerifFontFamily");
	if (s != null)
	    pref.sansSerifFontFamily = stringToFontFamily(s, "SansSerif");
	s = prop.getProperty("monospaceFontFamily");
	if (s != null)
	    pref.monospaceFontFamily = stringToFontFamily(s, "Monospaced");
	s = prop.getProperty("defaultFontIsSerif");
	if (s != null)
	    pref.defaultFontIsSerif = stringToBoolean(s, false);
	s = prop.getProperty("defaultFontSize");
	if (s != null)
	    pref.defaultFontSize = stringToInt(s, 10, 20, 12);
	s = prop.getProperty("defaultBackground");
	if (s != null)
	    pref.defaultBackground = stringToColor(s, Color.white);
	s = prop.getProperty("defaultForeground");
	if (s != null)
	    pref.defaultForeground = stringToColor(s, Color.black);

	pref.geometry = loadRectangle(prop, "geometry");
	if (pref.geometry == null ||
	    pref.geometry.x < 0 || pref.geometry.y < 0)
	    pref.geometry = DEFAULT_GEOMETRY;

	loadList(prop, "openedFileNames", pref.openedFileNames);

	loadList(prop, "templateNames", pref.templateNames);

	loadList(prop, "structuredFormats", pref.structuredFormats);

	Paper paper = loadPaper(prop, "paper");
	if (paper == null)
	    paper = DEFAULT_PAPER;
	pref.pageFormat.setPaper(paper);

	s = prop.getProperty("spellCheckerLanguage");
	if (s != null)
	    pref.spellCheckerLanguage = s;

	s = prop.getProperty("spellCheckerMarkParagraphs");
	if (s != null)
	    pref.spellCheckerMarkParagraphs = 
		stringToInt(s, MARK_NONE, MARK_LAST, MARK_ALL);

	return pref;
    }

    public void save() {
	String rcFileName = PlatformUtil.rcFileName("xxe");
	Properties prop = new Properties();

	prop.put("guessElementContent", booleanToString(guessElementContent));
	prop.put("ignoreStyleSheets", booleanToString(ignoreStyleSheets));

	prop.put("encoding", encoding);
	prop.put("indentIfDocType", booleanToString(indentIfDocType));
	prop.put("indentIfNoDocType", booleanToString(indentIfNoDocType));
	prop.put("indentation", Integer.toString(indentation));
	prop.put("maxLineLength", Integer.toString(maxLineLength));

	prop.put("screenResolutionWhenPrinting", 
		 Double.toString(screenResolutionWhenPrinting));
	prop.put("headerBegin", (headerBegin == null)? "" : headerBegin);
	prop.put("headerMiddle", (headerMiddle == null)? "" : headerMiddle);
	prop.put("headerEnd", (headerEnd == null)? "" : headerEnd);
	prop.put("headerColor", colorToString(headerColor));
	prop.put("underlineHeader", booleanToString(underlineHeader));
	prop.put("footerBegin", (footerBegin == null)? "" : footerBegin);
	prop.put("footerMiddle", (footerMiddle == null)? "" : footerMiddle);
	prop.put("footerEnd", (footerEnd == null)? "" : footerEnd);
	prop.put("footerColor", colorToString(footerColor));
	prop.put("overlineFooter", booleanToString(overlineFooter));

	prop.put("maxUndo", Integer.toString(maxUndo));
	prop.put("makeChoice", Integer.toString(makeChoice));

	prop.put("ignoreCase", booleanToString(ignoreCase));
	prop.put("wholeWord", booleanToString(wholeWord));
	prop.put("regularExpression", booleanToString(regularExpression));
	prop.put("markParagraphs", Integer.toString(markParagraphs));

	prop.put("serifFontFamily", serifFontFamily);
	prop.put("sansSerifFontFamily", sansSerifFontFamily);
	prop.put("monospaceFontFamily", monospaceFontFamily);
	prop.put("defaultFontIsSerif", booleanToString(defaultFontIsSerif));
	prop.put("defaultFontSize", Integer.toString(defaultFontSize));
	prop.put("defaultBackground", colorToString(defaultBackground));
	prop.put("defaultForeground", colorToString(defaultForeground));

	saveRectangle(geometry, prop, "geometry");

	saveList(openedFileNames, prop, "openedFileNames");

	saveList(templateNames, prop, "templateNames");

	saveList(structuredFormats, prop, "structuredFormats");

	savePaper(pageFormat.getPaper(), prop, "paper");

	prop.put("spellCheckerLanguage", spellCheckerLanguage);
	prop.put("spellCheckerMarkParagraphs", 
		 Integer.toString(spellCheckerMarkParagraphs));

	try {
	    OutputStream out = new FileOutputStream(rcFileName);

	    // JDK1.3: save() has been deprecated.
	    prop.store(out, "xxe");

	    out.close();
	} catch (Exception e) {
	    System.err.println(
		msg("cannotSavePreferences", rcFileName, e.getMessage()));
	}
    }

    // -----------------------------------------------------------------------

    public static Vector listToVector(String[] list) {
	Vector items = new Vector();

	for (int i = 0; i < list.length; ++i) {
	    if (list[i] == null)
		break;

	    items.addElement(list[i]);
	}

	return items;
    }

    public static void updateFromVector(String[] list, Vector items) {
	int count = items.size();
	for (int i = 0; i < list.length; ++i) {
	    if (i < count)
		list[i] = (String) items.elementAt(i);
	    else
		list[i] = null;
	}
    }

    private static String[] addToList(String[] list, String entry) {
	int length = list.length;
	String[] newList = new String[length];

	int j = 0;
	newList[j++] = entry;
	for (int i = 0; i < length; ++i) {
	    if (list[i] == null) 
		break;

	    if (!list[i].equals(entry) && j < length)
		newList[j++] = list[i];
	}
	    
	return newList;
    }

    private static String[] removeFromList(String[] list, String entry) {
	int length = list.length;
	String[] newList = new String[length];

	int j = 0;
	for (int i = 0; i < length; ++i) {
	    if (list[i] == null) 
		break;

	    if (!list[i].equals(entry))
		newList[j++] = list[i];
	}
	    
	return newList;
    }

    private static void emptyList(String[] list) {
	for (int i = 0; i < list.length; ++i)
	    list[i] = null;
    }

    private static void loadList(Properties prop, String name, String[] list) {
	emptyList(list);

	int j = 0;
	for (int i = 0; i < list.length; ++i) {
	    String property = prop.getProperty(name + "." + i);
	    if (property != null) 
		list[j++] = property;
	}
    }

    private static void saveList(String[] list, Properties prop, String name) {
	int j = 0;
	for (int i = 0; i < list.length; ++i) {
	    if (list[i] != null) {
		prop.put(name + "." + j, list[i]);
		++j;
	    }
	}
    }

    private static Rectangle loadRectangle(Properties prop, String name) {
	String value = prop.getProperty(name);
	if (value == null)
	    return null;

	StringTokenizer tokens = new StringTokenizer(value);
	if (tokens.countTokens() != 4)
	    return null;

	int x = -1, y = -1, width = -1, height = -1;

	try {
	    x = Integer.parseInt(tokens.nextToken());
	    y = Integer.parseInt(tokens.nextToken());
	    width = Integer.parseInt(tokens.nextToken());
	    height = Integer.parseInt(tokens.nextToken());
	} catch (NumberFormatException ignored) {}

	if (width <= 0 || height <= 0)
	    return null;

	return new Rectangle(x, y, width, height);
    }

    private static void saveRectangle(Rectangle r, 
				      Properties prop, String name) {
	StringBuffer value = new StringBuffer();

	value.append(Integer.toString(r.x));
	value.append(' ');
	value.append(Integer.toString(r.y));
	value.append(' ');
	value.append(Integer.toString(r.width));
	value.append(' ');
	value.append(Integer.toString(r.height));
	
	prop.put(name, value.toString());
    }

    private static Paper loadPaper(Properties prop, String name) {
	String value = prop.getProperty(name);
	if (value == null)
	    return null;

	StringTokenizer tokens = new StringTokenizer(value);
	if (tokens.countTokens() != 6)
	    return null;

	double paperW = -1, paperH = -1;
	double x = -1, y = -1, width = -1, height = -1;

	try {
	    paperW = Double.valueOf(tokens.nextToken()).doubleValue();
	    paperH = Double.valueOf(tokens.nextToken()).doubleValue();
	    x = Double.valueOf(tokens.nextToken()).doubleValue();
	    y = Double.valueOf(tokens.nextToken()).doubleValue();
	    width = Double.valueOf(tokens.nextToken()).doubleValue();
	    height = Double.valueOf(tokens.nextToken()).doubleValue();
	} catch (NumberFormatException ignored) {}

	return newPaper(paperW, paperH, x, y, width, height);
    }

    private static void savePaper(Paper p, Properties prop, String name) {
	StringBuffer value = new StringBuffer();

	value.append(Double.toString(p.getWidth()));
	value.append(' ');
	value.append(Double.toString(p.getHeight()));
	value.append(' ');
	value.append(Double.toString(p.getImageableX()));
	value.append(' ');
	value.append(Double.toString(p.getImageableY()));
	value.append(' ');
	value.append(Double.toString(p.getImageableWidth()));
	value.append(' ');
	value.append(Double.toString(p.getImageableHeight()));
	
	prop.put(name, value.toString());
    }

    // -----------------------------------------------------------------------

    private static final String nonEmptyString(String s) {
	if (s != null && (s = s.trim()).length() == 0)
	    s = null;
	return s;
    }

    private static final boolean stringToBoolean(String s, boolean fallback) {
	if ("true".equalsIgnoreCase(s))
	    return true;
	else if ("false".equalsIgnoreCase(s))
	    return false;
	else
	    return fallback;
    }

    private static final String booleanToString(boolean b) {
	return (b? "true" : "false");
    }

    private static final int stringToInt(String s, int min, int max,
					 int fallback) {
	int i = 0;

	try {
	    i = Integer.parseInt(s);
	} catch (NumberFormatException e) {
	    return fallback;
	}

	if (i < min || i > max)
	    return fallback;
	else
	    return i;
    }

    private static final double stringToDouble(String s, double min, 
					       double max, double fallback) {
	double d = 0;

	try {
	    d = Double.valueOf(s).doubleValue();
	} catch (NumberFormatException e) {
	    return fallback;
	}

	if (d < min || d > max)
	    return fallback;
	else
	    return d;
    }

    private static final Color stringToColor(String s, Color fallback) {
	int r = -1;
	int g = -1;
	int b = -1;

	if (s.length() == 7 && s.charAt(0) == '#') {
	    try {
		r = Integer.parseInt(s.substring(1, 3), 16);
		g = Integer.parseInt(s.substring(3, 5), 16);
		b = Integer.parseInt(s.substring(5, 7), 16);
	    } catch (NumberFormatException e) {}
	}

	if (r < 0 || r > 255 ||
	    g < 0 || g > 255 ||
	    b < 0 || b > 255)
	    return fallback;
	else
	    return new Color(r, g, b);
    }

    private static final String colorToString(Color color) {
	StringBuffer buffer = new StringBuffer();

	buffer.append('#');

	String hex = Integer.toString(color.getRed(), 16);
	if (hex.length() == 1)
	    buffer.append('0');
	buffer.append(hex);

	hex = Integer.toString(color.getGreen(), 16);
	if (hex.length() == 1)
	    buffer.append('0');
	buffer.append(hex);

	hex = Integer.toString(color.getBlue(), 16);
	if (hex.length() == 1)
	    buffer.append('0');
	buffer.append(hex);

	return buffer.toString();
    }

    private static final String stringToFontFamily(String s, String fallback) {
	if (StringUtil.contains(StyledViewFactory.FONT_FAMILIES, s))
	    return s;
	else
	    return fallback;
    }

    // -----------------------------------------------------------------------

    public static File findDistribSubDirectory(String subDirName) {
	String classPath = System.getProperty("java.class.path");
	if (classPath == null)
	    return null;

	String[] paths = StringUtil.split(classPath, File.pathSeparatorChar);
	for (int i = 0; i < paths.length; ++i) {
	    String path = paths[i].trim();
	    if (path.length() == 0)
		continue;

	    if (path.indexOf("xxe.jar") >= 0) {
		File jarFile = new File(path);
		if (!jarFile.isFile())
		    continue;

		File dictDir = new File(jarFile.getParent(), subDirName);
		if (dictDir.isDirectory()) {
		    try {
			return dictDir.getCanonicalFile();
		    } catch (IOException cannotHappen) {
			cannotHappen.printStackTrace();
			return null;
		    }
		}
	    } else if (path.indexOf("class") >= 0) {
		// This code is not used in the production environment.
		File classDir = new File(path);
		if (!classDir.isDirectory())
		    continue;

		File dictDir = 
		    new File(classDir.getParent(), 
			     "distrib" + File.separatorChar + subDirName);
		if (dictDir.isDirectory()) {
		    try {
			return dictDir.getCanonicalFile();
		    } catch (IOException cannotHappen) {
			cannotHappen.printStackTrace();
			return null;
		    }
		}
	    }
	}

	return null;
    }

    // -----------------------------------------------------------------------

    public static String[] splitArguments(String string) {
	int length = string.length();
	Vector list = new Vector();

	char quote = 0;
	StringBuffer arg = null;
	for (int i = 0; i < length; ++i) {
	    char c = string.charAt(i);

	    if (Character.isWhitespace(c)) {
		if (quote != 0) {
		    arg.append(c);
		} else {
		    if (arg != null) {
			list.addElement(arg.toString());
			arg = null;
		    }
		}
	    } else {
		if (arg == null) {
		    arg = new StringBuffer();

		    switch (c) {
		    case '\"': case '\'':
			quote = c;
			break;
		    default:
			arg.append(c);
		    }
		} else {
		    if (c == quote) {
			int last = arg.length()-1;

			if (last > 0 && arg.charAt(last) == '\\') {
			    arg.setCharAt(last, quote);
			} else {
			    list.addElement(arg.toString());
			    arg = null;
			    quote = 0;
			}
		    } else {
			arg.append(c);
		    }
		}
	    }
	}
	if (arg != null)
	    list.addElement(arg.toString());

	String[] args = new String[list.size()];
	list.copyInto(args);
	return args;
    }

    public static String quoteArgument(String arg) {
	int length = arg.length();

	boolean needQuotes;
	if (length == 0) {
	    needQuotes = true;
	} else {
	    switch (arg.charAt(0)) {
	    case '\"': case '\'':
		needQuotes = true;
		break;
	    default:
	        {
		    needQuotes = false;
		    for (int i = 0; i < length; ++i) {
			if (Character.isWhitespace(arg.charAt(i))) {
			    needQuotes = true;
			    break;
			}
		    }
		}
	    }
	}
	if (!needQuotes)
	    return arg;

	StringBuffer quoted = new StringBuffer();
	quoted.append('\"');
	for (int i = 0; i < length; ++i) {
	    char c = arg.charAt(i);

	    if (c == '\"') 
		quoted.append("\\\"");
	    else
		quoted.append(c);
	}
	quoted.append('\"');

	return quoted.toString();
    }

    // Tests:
    //
    // java com.xmlmind.xmleditapp.Preferences " '' x s'r '123' 'y \'z\' w' "
    //
    // java com.xmlmind.xmleditapp.Preferences ' "" x s"r "123" "y \"z\" w" '

    public static void main(String[] args) {
	for (int i = 0; i < args.length; ++i) {
	    String[] split = splitArguments(args[i]);

	    System.out.print('{');
	    System.out.print(args[i]);
	    System.out.println('}');
	    for (int j = 0; j < split.length; ++j) {
		System.out.print('\t');
		System.out.println(quoteArgument(split[j]));
	    }
	    System.out.println();
	}
	System.out.flush();
    }
}
