/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import fr.pixware.util.REFactoryImpl;
import fr.pixware.util.RESyntaxException;
import fr.pixware.util.RE;

public class SearchTextDialog extends JDialog 
                              implements ActionListener, ItemListener {
    private static class SearchInfo {
	public String searched;
	public boolean ignoreCase;
	public boolean wholeWord;
	public boolean regularExpression;

	public SearchInfo() {}

	public SearchInfo(String searched, boolean ignoreCase, 
			  boolean wholeWord, boolean regularExpression) {
	    this.searched = searched;
	    this.ignoreCase = ignoreCase;
	    this.wholeWord = wholeWord;
	    this.regularExpression = regularExpression;
	}

	public boolean equals(Object other) {
	    if (other == null || !(other instanceof SearchInfo))
		return false;
	    return searched.equals(((SearchInfo) other).searched);
	}
    }
    public static class Info extends SearchInfo {
	public String replacement;
	public boolean backward;
	public RE searchedRE;

	public Info() {
	    super();
	}
    }

    private Frame frame;
    private JComboBox searched;
    private JCheckBox ignoreCase;
    private JCheckBox wholeWord;
    private JCheckBox regularExpression;
    private JComboBox replacement;
    private ButtonGroup directionGroup;
    private JRadioButton directionUp;
    private JRadioButton directionDown;

    private Vector searchedList = new Vector();
    private Vector replacementList = new Vector();

    private Info result;

    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.SearchTextDialogRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    // -----------------------------------------------------------------------

    public SearchTextDialog(Frame frame) {
	super(frame, msg("title"), /*modal*/ true); 

	this.frame = frame;

   	addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        cancel();
   	    }
   	});

        // workArea is a JPanel with a BorderLayout.
	JPanel workArea = (JPanel) getContentPane();
	workArea.setLayout(new GridBagLayout());

	// Should be implemented by all LAFs.
	workArea.registerKeyboardAction(
	    this, "cancel", KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), 
	    JPanel.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
	constraints.anchor = GridBagConstraints.WEST;

	Font font = new Font("Dialog", Font.PLAIN, 12);

	// Search ---

	JPanel searchFrame = new JPanel(new GridBagLayout());
	searchFrame.setBorder(new TitledBorder(msg("search")));
        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.fill = GridBagConstraints.HORIZONTAL;
        constraints.weightx = 1;
	workArea.add(searchFrame, constraints);

        GridBagConstraints constraints2 = new GridBagConstraints();
        constraints2.insets.top = 5;
        constraints2.insets.bottom = 5;
        constraints2.insets.left = 5;
        constraints2.insets.right = 5;

	searched = new JComboBox();
	searched.setFont(font);
	searched.setEditable(true);
	Dimension dim = searched.getPreferredSize();
	FontMetrics metrics = searched.getFontMetrics(searched.getFont());
	dim.width = metrics.stringWidth("mmmmmmmmmmmmmmmmmmmmmmmmmmmmmm");
	searched.setPreferredSize(dim);
        constraints2.gridx = 0;
	constraints2.gridwidth = 3;
        constraints2.gridy = 0;
        constraints2.fill = GridBagConstraints.HORIZONTAL;
        constraints2.weightx = 1;
	searchFrame.add(searched, constraints2);
	searched.addItemListener(this);
	// Let the combobox do its job as an action listener before
	// invoking ok.
	ActionListener actionListener = new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
	        SwingUtilities.invokeLater(new Runnable() {
		    public void run() { ok(); }
		});
	    }
	};
	searched.getEditor().addActionListener(actionListener);

	// Replace ---

	JPanel replaceFrame = new JPanel(new GridBagLayout());
	replaceFrame.setBorder(new TitledBorder(msg("replace")));
        constraints.gridx = 0;
        constraints.gridy = 1;
        constraints.insets.top = 0;
	workArea.add(replaceFrame, constraints);

	replacement = new JComboBox();
	replacement.setFont(font);
	replacement.setEditable(true);
	replacement.setPreferredSize(dim);
        constraints2.gridx = 0;
	constraints2.gridwidth = 1;
        constraints2.gridy = 0;
        constraints2.fill = GridBagConstraints.HORIZONTAL;
        constraints2.weightx = 1;
	replaceFrame.add(replacement, constraints2);
	replacement.getEditor().addActionListener(actionListener);

	// Search options ---

	JPanel optionsFrame = new JPanel(new GridBagLayout());
        constraints.gridx = 0;
        constraints.gridy = 2;
        constraints.fill = GridBagConstraints.NONE;
        constraints.weightx = 0;
	workArea.add(optionsFrame, constraints);

	ignoreCase = new JCheckBox(msg("ignoreCase"));
	ignoreCase.setFont(font);
        constraints2.gridx = 0;
        constraints2.gridy = 0;
        constraints2.insets.top = 0;
        constraints2.insets.bottom = 0;
	optionsFrame.add(ignoreCase, constraints2);

	wholeWord = new JCheckBox(msg("wholeWord"));
	wholeWord.setFont(font);
        constraints2.gridx = 1;
        constraints2.gridy = 0;
	optionsFrame.add(wholeWord, constraints2);

	regularExpression = new JCheckBox(msg("regularExpression"));
	regularExpression.setFont(font);
        constraints2.gridx = 2;
        constraints2.gridy = 0;
	optionsFrame.add(regularExpression, constraints2);

	// Direction ---

	JPanel directionFrame = new JPanel(new GridBagLayout());
        constraints.gridx = 0;
        constraints.gridy = 3;
	workArea.add(directionFrame, constraints);

	JLabel label = new JLabel(msg("direction"));
        constraints2.gridx = 0;
        constraints2.gridy = 0;
	directionFrame.add(label, constraints2);

	directionUp = new JRadioButton(msg("up"));
	directionUp.setFont(font);
        constraints2.gridx = 1;
        constraints2.gridy = 0;
	directionFrame.add(directionUp, constraints2);

	directionDown = new JRadioButton(msg("down"));
	directionDown.setFont(font);
        constraints2.gridx = 2;
        constraints2.gridy = 0;
	directionFrame.add(directionDown, constraints2);

	directionGroup = new ButtonGroup();
	directionGroup.add(directionUp);
	directionGroup.add(directionDown);

	// Buttons ---

	JPanel buttons = new JPanel(new GridLayout(1, 2, 10, 0));
        constraints.gridx = 0;
        constraints.gridy = 4;
	constraints.anchor = GridBagConstraints.CENTER;
	workArea.add(buttons, constraints);

        JButton button = new JButton(msg("ok"));
	button.setActionCommand("ok");
	button.addActionListener(this);
        buttons.add(button);
	getRootPane().setDefaultButton(button);
          
        button = new JButton(msg("cancel"));
        button.setActionCommand("cancel");
        button.addActionListener(this);
        buttons.add(button);

        pack();
	setResizable(true);
    }

    public Info getInfo(Info in) {
	searched.setSelectedItem((in.searched == null)? "" : in.searched);

	ignoreCase.setSelected(in.ignoreCase);
	wholeWord.setSelected(in.wholeWord);
	regularExpression.setSelected(in.regularExpression);

	boolean replace = (in.replacement != null);
	comboBoxEditor(replacement).setBackground(
	  replace? comboBoxEditor(searched).getBackground() : getBackground());
	replacement.setSelectedItem(replace? in.replacement : "");
	replacement.setEnabled(replace);

	directionUp.setSelected(in.backward);
	directionDown.setSelected(!in.backward);

	setLocationRelativeTo(frame);
	pack();

	comboBoxEditor(searched).requestFocus();

        show();
        // Modal means blocked here until OK or Cancel are pressed.

        return result;
    }

    private static final Component comboBoxEditor(JComboBox comboBox) {
	return comboBox.getEditor().getEditorComponent();
    }

    // -----------------------------------------------------------------------

    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();

        if (command.equals("ok"))
            ok();
        else if (command.equals("cancel"))
            cancel();
    }
    
    public void itemStateChanged(ItemEvent e) {
	if (e.getStateChange() != ItemEvent.SELECTED)
	    return;

	int index = searched.getSelectedIndex();
	if (index < 0 || index >= searchedList.size())
	    return;

	SearchInfo info = (SearchInfo) searchedList.elementAt(index);
	ignoreCase.setSelected(info.ignoreCase);
	wholeWord.setSelected(info.wholeWord);
	regularExpression.setSelected(info.regularExpression);
    }

    private void ok() {
	Info info = new Info();

	info.searched = (String) searched.getSelectedItem();
	// Do no trim!
	if (info.searched == null || info.searched.length() == 0) {
	    getToolkit().beep();
	    searched.requestFocus();
	    return;
	}

	info.ignoreCase = ignoreCase.isSelected();
	info.wholeWord = wholeWord.isSelected();
	info.regularExpression = regularExpression.isSelected();

	if (info.regularExpression) {
	    int options = REFactoryImpl.MULTI_LINE;
	    if (info.ignoreCase)
		options |= REFactoryImpl.CASE_INSENSITIVE;

	    try {

		info.searchedRE = 
		    (new REFactoryImpl()).createRE(info.searched, options);

	    } catch (RESyntaxException e) {
		getToolkit().beep();
		JOptionPane.showMessageDialog(
		    frame, msg("invalidRE", info.searched),
		    msg("error"), JOptionPane.ERROR_MESSAGE);
		searched.requestFocus();
		return;
	    }
	}

	if (replacement.isEnabled()) {
	    // An empty string is OK.
	    info.replacement = (String) replacement.getSelectedItem();
	    if (info.replacement == null)
		info.replacement = "";
	} else {
	    info.replacement = null;
	}

	info.backward = directionUp.isSelected();

	remember(info);
	result = info;

	hide();
	dispose();
    }

    private void remember(Info info) {
	// Remove previous/last entry. Add as first entry ---

	SearchInfo searchInfo = 
	    new SearchInfo(info.searched, info.ignoreCase,
			   info.wholeWord, info.regularExpression);

	int index = searchedList.indexOf(searchInfo);
	if (index >= 0)
	    searchedList.removeElementAt(index);

	int count = searchedList.size();
	if (count > 10)
	    searchedList.removeElementAt(count-1);

	searchedList.insertElementAt(searchInfo, 0);

	// Update combobox ---

	searched.removeAllItems();
	count = searchedList.size();
	for (int i = 0; i < count; ++i) {
	    searchInfo = (SearchInfo) searchedList.elementAt(i);
	    searched.addItem(searchInfo.searched);
	}

	if (info.replacement != null && info.replacement.length() > 0) {
	    // Remove previous/last entry. Add as first entry ---

	    index = replacementList.indexOf(info.replacement);
	    if (index >= 0)
		replacementList.removeElementAt(index);

	    count = replacementList.size();
	    if (count > 10)
		replacementList.removeElementAt(count-1);

	    replacementList.insertElementAt(info.replacement, 0);

	    // Update combobox ---

	    replacement.removeAllItems();
	    count = replacementList.size();
	    for (int i = 0; i < count; ++i)
		replacement.addItem(replacementList.elementAt(i));
	}
    }

    private void cancel() {
	result = null;

	hide();
	dispose();
    }
}

