/*
 * Copyright (c) 2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.io.IOException;
import java.io.File;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.awt.*;
import javax.swing.*;
import javax.swing.text.Document;
import java.awt.event.*;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.DocumentEvent;
import fr.pixware.util.PlatformUtil;
import fr.pixware.util.StringUtil;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.xmledit.modeltool.*;
import com.xmlmind.xmledit.*;
import azcheck.engine.SpellChecker;
import azcheck.engine.Suggestions;
import azcheck.engine.DictionaryManagerItem;
import azcheck.util.SpellException;

public class SpellCheckerDialog extends JDialog
                                implements ActionListener, 
                                           ListSelectionListener,
                                           DocumentListener {

    private JDocumentPane docPane;
    private SpellChecker spellChecker;
    private String[] languages;

    private int markParagraphs = -1;

    private Frame frame;
    private JTextField word;
    private JList suggestions;
    private JComboBox language;
    private JButton replace;
    private JButton replaceAll;
    private JButton ignore;
    private JButton ignoreAll;
    private JButton learn;
    private JButton stop;
    private JTextField status;

    private boolean positioned;

    private boolean replaceEnabled;
    private int checkResult;
    private ElementModel checkedElement;
    private CheckableCharSequence charInput;
    private Hashtable ignoredWords = new Hashtable();
    private Hashtable replacedWords = new Hashtable();

    private static final String[] NO_SUGGESTIONS = new String[0];

    // -----------------------------------------------------------------------
    // Localized messages
    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.SpellCheckDialogRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    public static String msg(String id, Object arg0, Object arg1) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0, arg1 });
    }

    // -----------------------------------------------------------------------
    // Initialization
    // -----------------------------------------------------------------------

    public SpellCheckerDialog(Frame frame, JDocumentPane docPane) {
	super(frame, msg("title"), /*modal*/ true); 
	
	this.frame = frame;

   	addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        stop();
   	    }
   	});

	this.docPane = docPane;

        // workArea is a JPanel with a BorderLayout.
	JPanel workArea = (JPanel) getContentPane();
	workArea.setLayout(new GridBagLayout());

	// Should be implemented by all LAFs.
	workArea.registerKeyboardAction(
	    this, "stop", KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), 
	    JPanel.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.bottom = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;

	// Words area ---

	JPanel left = new JPanel(new GridBagLayout());
        constraints.gridx = 0;
        constraints.gridy = 0;
	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
	workArea.add(left, constraints);

        GridBagConstraints constraints2 = new GridBagConstraints();
        constraints2.gridx = 0;
        constraints2.gridy = 0;
	constraints2.gridwidth = 2;

	JLabel label = new JLabel(msg("word"));
	constraints2.fill = GridBagConstraints.HORIZONTAL;
	constraints2.anchor = GridBagConstraints.WEST;
	left.add(label, constraints2);
	++constraints2.gridy;

	word = new JTextField(20);
	word.setActionCommand("replace");
	word.addActionListener(this);
	word.getDocument().addDocumentListener(this);
	constraints2.fill = GridBagConstraints.HORIZONTAL;
	constraints2.weightx = 1;
        constraints2.insets.top = 2;
	left.add(word, constraints2);
	++constraints2.gridy;

	label = new JLabel(msg("suggestions"));
	constraints2.fill = GridBagConstraints.HORIZONTAL;
	constraints2.weightx = 0;
	constraints2.anchor = GridBagConstraints.WEST;
        constraints2.insets.top = 10;
	left.add(label, constraints2);
	++constraints2.gridy;

	suggestions = new JList();
	suggestions.setVisibleRowCount(10);
	suggestions.addListSelectionListener(this);
	suggestions.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		if (e.getClickCount() == 2) {
		    replace();
		}
	    }
	});

	JScrollPane scroller = new JScrollPane(suggestions);
	constraints2.fill = GridBagConstraints.BOTH;
	constraints2.weightx = 1;
	constraints2.weighty = 1;
        constraints2.insets.top = 2;
	left.add(scroller, constraints2);
	++constraints2.gridy;

	label = new JLabel(msg("language"));
	constraints2.fill = GridBagConstraints.NONE;
	constraints2.weightx = 0;
	constraints2.weighty = 0;
	constraints2.gridwidth = 1;
        constraints2.insets.top = 10;
	left.add(label, constraints2);

	language = new JComboBox();
	language.setEditable(false);
	constraints2.gridx = 1;
	constraints2.fill = GridBagConstraints.HORIZONTAL;
	constraints2.weightx = 1;
        constraints2.insets.top = 10;
        constraints2.insets.left = 5;
	left.add(language, constraints2);

	// Buttons ---

	JPanel right = new JPanel(new GridBagLayout());
        constraints.gridx = 1;
        constraints.gridy = 0;
	constraints.fill = GridBagConstraints.VERTICAL;
	constraints.weightx = 0;
	constraints.weighty = 1;
	workArea.add(right, constraints);

        constraints2 = new GridBagConstraints();
        constraints2.gridx = 0;
        constraints2.gridy = 0;
	constraints2.anchor = GridBagConstraints.NORTH;
        constraints2.insets.bottom = 5;
	constraints2.fill = GridBagConstraints.HORIZONTAL;

	replace = new JButton(msg("replace"));
	replace.setActionCommand("replace");
	replace.addActionListener(this);
	right.add(replace, constraints2);
	++constraints2.gridy;

	replaceAll = new JButton(msg("replaceAll"));
	replaceAll.setActionCommand("replaceAll");
	replaceAll.addActionListener(this);
	right.add(replaceAll, constraints2);
	++constraints2.gridy;

	ignore = new JButton(msg("ignore"));
	ignore.setActionCommand("ignore");
	ignore.addActionListener(this);
	right.add(ignore, constraints2);
	++constraints2.gridy;

	ignoreAll = new JButton(msg("ignoreAll"));
	ignoreAll.setActionCommand("ignoreAll");
	ignoreAll.addActionListener(this);
	right.add(ignoreAll, constraints2);
	++constraints2.gridy;

	learn = new JButton(msg("learn"));
	learn.setActionCommand("learn");
	learn.addActionListener(this);
	right.add(learn, constraints2);
	++constraints2.gridy;

	stop = new JButton(msg("stop"));
	stop.setActionCommand("stop");
	stop.addActionListener(this);
	constraints2.weighty = 1;
	constraints2.anchor = GridBagConstraints.SOUTH;
        constraints2.insets.top = 5;
        constraints2.insets.bottom = 0;
	right.add(stop, constraints2);

	// Status area ---

	status = new JTextField(25);
	status.setEditable(false);
        constraints.gridx = 0;
        constraints.gridy = 1;
        constraints.gridwidth = 2;
	constraints.fill = GridBagConstraints.HORIZONTAL;
	constraints.weightx = 1;
	constraints.weighty = 0;
	workArea.add(status, constraints);

	pack();
	setResizable(true);

	initialize();
    }

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	
	if (command.equals("replace")) 
	    replace();
	else if (command.equals("replaceAll")) 
	    replaceAll();
	else if (command.equals("ignore")) 
	    ignore();
	else if (command.equals("ignoreAll")) 
	    ignoreAll();
	else if (command.equals("learn")) 
	    learn();
	else if (command.equals("selectLanguage")) 
	    selectLanguage();
	else if (command.equals("stop")) 
	    stop();
    }

    public void valueChanged(ListSelectionEvent e) {
	String selection  = (String) suggestions.getSelectedValue();
	if (selection != null) {
	    word.setText(selection);
	    if (!replaceEnabled)
		enableReplace(true);
	}
    }

    public void insertUpdate(DocumentEvent e) {
	if (!replaceEnabled)
	    enableReplace(true);
    }

    public void removeUpdate(DocumentEvent e) {
	if (!replaceEnabled)
	    enableReplace(true);
    }

    public void changedUpdate(DocumentEvent e) {
	if (!replaceEnabled)
	    enableReplace(true);
    }

    private void enableReplace(boolean enabled) {
	replaceEnabled = enabled;
	replace.setEnabled(enabled);
	replaceAll.setEnabled(enabled &&
			      (checkResult == SpellChecker.ERR_UNKNOWN_WORD ||
			       checkResult == SpellChecker.ERR_WRONG_CAP));
    }

    private void initialize() {
	try {
	    spellChecker = createSpellChecker();
	} catch (IOException ignored) {
	    ignored.printStackTrace();
	    spellChecker = null;
	}

	DictionaryManagerItem[] items = null;
	if (spellChecker != null) {
	    items = spellChecker.listLanguages();
	    if (items.length == 0)
		spellChecker = null;
	}

	if (spellChecker != null) {
	    languages = new String[items.length];
	    for (int i = 0; i < languages.length; ++i) {
		DictionaryManagerItem item = items[i];

		//System.err.println(item.name);
		languages[i] = item.name;
		language.addItem(item.local);
	    }

	    language.setActionCommand("selectLanguage");
	    language.addActionListener(this);
	}
    }

    private static SpellChecker createSpellChecker() 
	throws IOException {
	File dictDir = findDictionariesDirectory();
	if (dictDir == null)
	    return null;

	SpellChecker spellChecker = 
	    new SpellChecker(dictDir.getAbsolutePath());

	setPersonalDictionaryPath(spellChecker);

	return spellChecker;
    }

    private static File findDictionariesDirectory() {
	return Preferences.findDistribSubDirectory("dict");
    }

    private static void setPersonalDictionaryPath(SpellChecker spellChecker) 
	throws IOException {
	String homeDirName = PlatformUtil.homeDir();
	if (homeDirName == null) {
	    homeDirName = PlatformUtil.tmpDir();
	    if (homeDirName == null)
		homeDirName = (new File(".")).getCanonicalPath();
	}

	String spellBaseName;
	if ("windows".equals(PlatformUtil.platform()))
	    spellBaseName = "xxespell";
	else
	    spellBaseName = ".xxespell";

	File spellDir = 
	    new File(homeDirName + File.separatorChar + spellBaseName);
	if (!spellDir.isDirectory() && !spellDir.mkdir())
	    spellDir = new File(homeDirName);

	spellChecker.setPersonalDictionaryPath(
	    spellDir.getAbsolutePath() + File.separatorChar + "dict_%L%.txt");
    }

    // -----------------------------------------------------------------------
    // Persistent user's preferences
    // -----------------------------------------------------------------------

    public void setPreferences(Preferences pref) {
	if (spellChecker == null)
	    return;

	if (markParagraphs < 0) {
	    // First time setPreferences is called.

	    String initialLanguage = pref.spellCheckerLanguage;
	    int index = StringUtil.indexOf(languages, initialLanguage);

	    if (index < 0) {
		initialLanguage = "en";
		index = StringUtil.indexOf(languages, initialLanguage);

		if (index < 0) {
		    initialLanguage = languages[0];
		    index = 0;
		}
	    }

	    try {
		spellChecker.setSelectedLanguage(initialLanguage);
	    } catch (Exception ignored) {
		ignored.printStackTrace();
		spellChecker = null;
		return;
	    }

	    language.setSelectedIndex(index);
	}

	// For now, stick to defaults.

	/*
	System.err.println(
	    "Checking case=" + spellChecker.getCheckCase() +
	    ", digits=" + spellChecker.getCheckDigits() +
	    ", duplicate=" + spellChecker.getCheckDuplicate() +
	    ", mixedCase=" + spellChecker.getCheckMixedCase() +
	    ", punctuation=" + spellChecker.getCheckPunctuation() +
	    ", URL=" + spellChecker.getCheckURL());
	*/

	spellChecker.setCheckCase(true);
	spellChecker.setCheckDigits(false);
	spellChecker.setCheckDuplicate(true);
	spellChecker.setCheckMixedCase(true);
	spellChecker.setCheckPunctuation(true);
	spellChecker.setCheckURL(false);

	// For now, this option cannot be edited using the Options dialog box.
	// Its default value is MARK_ALL.

	switch (pref.spellCheckerMarkParagraphs) {
	case Preferences.MARK_NONE:
	    markParagraphs = CheckableCharSequence.MARK_NONE;
	    break;
	case Preferences.MARK_LAST:
	    markParagraphs = CheckableCharSequence.MARK_LAST;
	    break;
	default:
	    markParagraphs = CheckableCharSequence.MARK_ALL;
	}
    }

    public void getPreferences(Preferences pref) {
	if (spellChecker == null)
	    return;

	pref.spellCheckerLanguage = languages[language.getSelectedIndex()];

	try {
	    spellChecker.savePersonalDictionaries();
	} catch (Exception e) {
	    System.err.println(msg("cannotSavePersonalDictionaries", 
				   e.getMessage()));
	}
    }

    // -----------------------------------------------------------------------
    // Checking spelling errors
    // -----------------------------------------------------------------------

    public boolean checkSpelling() {
	if (spellChecker == null) {
	    JOptionPane.showMessageDialog(
		frame, msg("noDictionaries"),
		msg("error"), JOptionPane.ERROR_MESSAGE);
	    return false;
	}

	charInput = createCharSequence();
	if (charInput == null)
	    // No text.
	    return true;

	spellChecker.setInput(charInput);
	try {
	    checkResult = spellChecker.checkNext();
	} catch (SpellException e) {
	    spellCheckerError(e);
	    return false;
	}

	if (checkResult == SpellChecker.ERR_NONE) {
	    setDone(true, "");
	    return true;
	} else {
	    SwingUtilities.invokeLater(new Runnable() {
		public void run() {
		    showDialog();
		}
	    });
	    return false;
	}
    }
    private CheckableCharSequence createCharSequence() {
	Caret caret = docPane.getCaret();
	if (caret == null)
	    // No document.
	    return null;

	ContextCommandInfo commandInfo = new ContextCommandInfo();
	docPane.getContext(commandInfo);

	TextOffset from;

	if (commandInfo.context == ContextCommandInfo.SUBELEMENT_SELECTION) {
	    SubelementModel sub = commandInfo.subelement;
	    if (sub.getSubelementType() == SubelementModel.ELEMENT)
		checkedElement = (ElementModel) sub;
	    else
		checkedElement = sub.getParentElement();

	    from = null;
	} else {
	    checkedElement = null; // whole document.

	    TextLocation textLocation;
	    if (commandInfo.context == ContextCommandInfo.TEXT_SELECTION) {
		if (commandInfo.markAfterPoint)
		    textLocation = caret.getMark();
		else
		    textLocation = caret.getPoint();
	    } else {
		textLocation = caret.getPoint();
	    }

	    from = new TextOffset();
	    from.text = textLocation.getText();
	    if (from.text == null)
		// Document contains no text.
		return null;
	    from.offset = textLocation.getOffset();

	    // Do not begin checking in the middle of a word ---
	    char[] chars = from.text.getTextChars();

	    if (from.offset < chars.length && 
		!XMLUtil.isXMLSpace(chars[from.offset])) {
		while (from.offset > 0 && 
		       !XMLUtil.isXMLSpace(chars[from.offset-1]))
		    --from.offset;
	    }
	}

	if (markParagraphs < 0)
	    markParagraphs = CheckableCharSequence.MARK_LAST;
	return new CheckableCharSequence(checkedElement, from, null, 
					 /*emptyTextContent*/ null,
					 markParagraphs,
					 /*paragraphMark*/ null);
    }

    private void showDialog() {
	if (!positioned) {
	    positioned = true;
	    Tool.setRelativeLocation(this, frame, 
                                     /*top*/ false, /*right*/ true);
	}
	pack();

	setDone(false, "");
	try {
	    showResult();
	} catch (SpellException e) {
	    spellCheckerError(e);
	    return;
	}

        show();
        // Modal means blocked here until stop is pressed.
    }

    private void showResult() throws SpellException {
	for (;;) {
	    switch (checkResult) {
	    case SpellChecker.ERR_UNKNOWN_WORD:
	    case SpellChecker.ERR_WRONG_CAP:
		{
		    String badWord = spellChecker.getWord();

		    if (ignoredWords.containsKey(badWord)) {
			checkResult = spellChecker.checkNext();
			continue;
		    } else {
			String replacement = 
			    (String) replacedWords.get(badWord);
			if (replacement != null) {
			    doReplace(replacement);
			    checkResult = spellChecker.checkNext();
			    continue;
			}
		    }
		}
		/*FALLTHROUGH*/

	    case SpellChecker.ERR_DUPLICATE:
	    case SpellChecker.ERR_PUNCTUATION:
		{
		    String badWord = spellChecker.getWord();
		    word.setText(badWord);
		    word.requestFocus();
		    enableReplace(false);

		    int pos = spellChecker.getPosition();
		    TextOffset offset1 = charInput.getTextOffset(pos);
		    TextOffset offset2 = 
			charInput.getTextOffset(pos + badWord.length());
		    docPane.selectText(offset1, offset2);

		    boolean enableIgnoreAll = false;
		    boolean enableLearn = false;
		    int showSuggestions = 0;

		    switch (checkResult) {
		    case SpellChecker.ERR_UNKNOWN_WORD:
			status.setText(msg("unknownWord"));
			enableIgnoreAll = true;
			enableLearn = true;
			showSuggestions = 2;
			break;
	            case SpellChecker.ERR_WRONG_CAP:
			status.setText(msg("wrongCap"));
			enableIgnoreAll = true;
			showSuggestions = 1;
			break;
		    case SpellChecker.ERR_DUPLICATE:
			status.setText(msg("duplicatedWords"));
			break;
		    case SpellChecker.ERR_PUNCTUATION:
			status.setText(msg("badPunctuation"));
			break;
		    }

		    ignoreAll.setEnabled(enableIgnoreAll);
		    learn.setEnabled(enableLearn);

		    switch (showSuggestions) {
		    case 0:
			suggestions.setListData(NO_SUGGESTIONS);
			break;
		    case 1:
			suggestions.setListData(new String[] {
			    spellChecker.getSuggestions().getSuggestion(0)
                        });
			break;
		    default:
		        {
			    String[] list = 
				spellChecker.getSuggestions().toArray();
			    suggestions.setListData(list);
		        }
		    }
		    suggestions.ensureIndexIsVisible(0);
		}
		return;

	    case SpellChecker.ERR_NONE:
		setDone(true, msg("done"));
		return;

	    default:
		checkResult = spellChecker.checkNext();
	    }
	}
    }

    private String doReplace(String replacement) {
	String badWord = spellChecker.getWord();
	if (badWord.equals(replacement))
	    return null;

	int pos = spellChecker.getPosition();
	TextOffset offset1 = charInput.getTextOffset(pos);
	TextOffset offset2 = charInput.getTextOffset(pos + badWord.length());

	ElementStructureEditor structureEditor = 
	    docPane.getDocumentView().getElementStructureEditor();
	DocumentModel document = docPane.getDocument();

	document.beginBatchEditing();

	structureEditor.delete(offset1.text, offset1.offset,
			       offset2.text, offset2.offset);
	if (replacement.length() > 0)
	    offset1.text.insertText(offset1.offset, replacement);

	document.endBatchEditing();

	offset1.offset += replacement.length();
	charInput = new CheckableCharSequence(checkedElement, offset1, null, 
					      /*emptyTextContent*/ null,
					      markParagraphs,
					      /*paragraphMark*/ null);
	spellChecker.setInput(charInput);

	return badWord;
    }

    private void replace() {
	String replacement = word.getText(); // Do not trim + may be empty.

	try {
	    String badWord = doReplace(replacement);

	    if (badWord != null &&
		checkResult == SpellChecker.ERR_UNKNOWN_WORD) {
		learnSuggestion(badWord, replacement);
	    }
	    // Otherwise not replaced, because same word.

	    checkResult = spellChecker.checkNext();
	    showResult();
	} catch (SpellException e) {
	    spellCheckerError(e);
	}
    }

    private void learnSuggestion(String badWord, String replacement) 
	throws SpellException {
	ListModel list = suggestions.getModel();
	int count = Math.min(3, list.getSize());
	boolean learn = true;

	for (int i = 0; i < count; ++i) {
	    if (replacement.equals((String) list.getElementAt(i))) {
		learn = false;
		break;
	    }
	}
	if (learn)
	    spellChecker.learnSuggestion(badWord, replacement, 
					 SpellChecker.TMP_DICT);
    }

    private void replaceAll() {
	String replacement = word.getText(); // Do not trim + may be empty.

	try {
	    String badWord = doReplace(replacement);

	    if (badWord != null &&
		(checkResult == SpellChecker.ERR_UNKNOWN_WORD ||
		 checkResult == SpellChecker.ERR_WRONG_CAP)) {
		replacedWords.put(badWord, replacement);

		if (checkResult == SpellChecker.ERR_UNKNOWN_WORD)
		    learnSuggestion(badWord, replacement);
	    }
	    // Otherwise not replaced, because same word.

	    checkResult = spellChecker.checkNext();
	    showResult();
	} catch (SpellException e) {
	    spellCheckerError(e);
	}
    }

    public void ignore() {
	try {
	    checkResult = spellChecker.checkNext();
	    showResult();
	} catch (SpellException e) {
	    spellCheckerError(e);
	}
    }

    private void ignoreAll() {
	try {
	    if (checkResult == SpellChecker.ERR_UNKNOWN_WORD ||
		checkResult == SpellChecker.ERR_WRONG_CAP) {
		String badWord = spellChecker.getWord();
		ignoredWords.put(badWord, badWord);
	    }

	    checkResult = spellChecker.checkNext();
	    showResult();
	} catch (SpellException e) {
	    spellCheckerError(e);
	}
    }

    private void learn() {
	try {
	    if (checkResult == SpellChecker.ERR_UNKNOWN_WORD) {
		String badWord = spellChecker.getWord();
		spellChecker.learnWord(badWord, SpellChecker.PERSONAL_DICT);
	    }

	    checkResult = spellChecker.checkNext();
	    showResult();
	} catch (SpellException e) {
	    spellCheckerError(e);
	}
    }

    private void selectLanguage() {
	int index = language.getSelectedIndex();

	try {
	    spellChecker.setSelectedLanguage(languages[index]);

	    if (checkResult == SpellChecker.ERR_UNKNOWN_WORD) {
		// Restart checking before bad word ---

		int pos = spellChecker.getPosition();
		TextOffset offset1 = charInput.getTextOffset(pos);

		charInput = 
		    new CheckableCharSequence(checkedElement, offset1, null, 
					      /*emptyTextContent*/ null,
					      markParagraphs,
					      /*paragraphMark*/ null);
		spellChecker.setInput(charInput);

		checkResult = spellChecker.checkNext();
		showResult();
	    }
	} catch (Exception e) {
	    spellCheckerError(e);
	}
    }

    private void stop() {
	setDone(true, "");

	hide();
	dispose();
    }

    // -----------------------------------------------------------------------
    // Helpers
    // -----------------------------------------------------------------------

    private void spellCheckerError(Exception e) {
	e.printStackTrace();

	String error = msg("error");

	JOptionPane.showMessageDialog(
	    frame, 
	    msg("spellCheckerError", e.getClass().getName(), e.getMessage()),
	    error, JOptionPane.ERROR_MESSAGE);

	setDone(true, error);
    }

    private void setDone(boolean done, String message) {
	word.setText("");
	word.setEnabled(!done);

	suggestions.setListData(NO_SUGGESTIONS);
	suggestions.setEnabled(!done);

	language.setEnabled(!done);

	replace.setEnabled(false);
	replaceAll.setEnabled(false);
	ignore.setEnabled(!done);
	ignoreAll.setEnabled(false);
	learn.setEnabled(false);

	status.setText(message);

	if (done) {
	    reset();
	    stop.setText(msg("done"));
	} else {
	    getRootPane().setDefaultButton(replace);
	    stop.setText(msg("stop"));
	}
    }

    private void reset() {
	replaceEnabled = false;
	checkResult = SpellChecker.ERR_NONE;
	checkedElement = null;
	charInput = null;    
	ignoredWords.clear();
	replacedWords.clear();
    }
}
