/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.border.TitledBorder;
import fr.pixware.util.StringUtil;
import com.xmlmind.structformat.*;

public class StructuredFormatsDialog extends JDialog 
                                     implements ActionListener, 
                                                ListSelectionListener {
    private Frame frame;
    private JList classNameList;
    private JButton removeButton;
    private JButton configureButton;
    private JTextField mimeTypeField;
    private JTextField extensionsField;
    private JTextArea descriptionText;

    private App app;

    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.StructFormatsDialogRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0, Object arg1) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0, arg1 });
    }

    // -----------------------------------------------------------------------

    public StructuredFormatsDialog(Frame frame) {
	super(frame, msg("title"), /*modal*/ true); 

	this.frame = frame;

   	addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        close();
   	    }
   	});

        // workArea is a JPanel with a BorderLayout.
	JPanel workArea = (JPanel) getContentPane();
	workArea.setLayout(new GridBagLayout());

	// Should be implemented by all LAFs.
	workArea.registerKeyboardAction(
	    this, "close", KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), 
	    JPanel.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
        constraints.gridx = 0;
        constraints.gridy = 0;

	// Class names ---

	classNameList = new JList();
	classNameList.setVisibleRowCount(5);
	classNameList.addListSelectionListener(this);
	classNameList.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		if (e.getClickCount() == 2) {
		    configure();
		}
	    }
	});

	JScrollPane scroller = new JScrollPane(classNameList);
	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
	workArea.add(scroller, constraints);
	++constraints.gridy;

	// Buttons ---

	JPanel buttonFrame = new JPanel(new GridLayout(1, 3, 10, 0));
	constraints.fill = GridBagConstraints.NONE;
	constraints.weightx = 0;
	constraints.weighty = 0;
	workArea.add(buttonFrame, constraints);
	++constraints.gridy;

        JButton button = new JButton(msg("add") + "...");
	button.setActionCommand("add");
	button.addActionListener(this);
        buttonFrame.add(button);

        removeButton = new JButton(msg("remove"));
	removeButton.setActionCommand("remove");
	removeButton.addActionListener(this);
        buttonFrame.add(removeButton);

        configureButton = new JButton(msg("configure") + "...");
	configureButton.setActionCommand("configure");
	configureButton.addActionListener(this);
        buttonFrame.add(configureButton);

	// Info ---

	JPanel infoFrame = new JPanel(new GridBagLayout());
	infoFrame.setBorder(new TitledBorder(msg("info")));
	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
	workArea.add(infoFrame, constraints);
	++constraints.gridy;

        GridBagConstraints constraints2 = new GridBagConstraints();
        constraints2.insets.top = 5;
        constraints2.insets.left = 5;
        constraints2.gridx = 0;
        constraints2.gridy = 0;

	JLabel label = new JLabel(msg("mimeType"));
	constraints2.anchor = GridBagConstraints.EAST;
        constraints2.insets.right = 0;
	infoFrame.add(label, constraints2);
	++constraints2.gridx;

	mimeTypeField = new JTextField(40);
	mimeTypeField.setEditable(false);
	constraints2.anchor = GridBagConstraints.WEST;
	constraints2.fill = GridBagConstraints.HORIZONTAL;
	constraints2.weightx = 1;
        constraints2.insets.right = 5;
	infoFrame.add(mimeTypeField, constraints2);
	++constraints2.gridy;
	constraints2.gridx = 0;

	label = new JLabel(msg("extensions"));
	constraints2.anchor = GridBagConstraints.EAST;
	constraints2.fill = GridBagConstraints.NONE;
	constraints2.weightx = 0;
        constraints2.insets.right = 0;
	infoFrame.add(label, constraints2);
	++constraints2.gridx;

	extensionsField = new JTextField(40);
	extensionsField.setEditable(false);
	constraints2.anchor = GridBagConstraints.WEST;
	constraints2.fill = GridBagConstraints.HORIZONTAL;
	constraints2.weightx = 1;
        constraints2.insets.right = 5;
	infoFrame.add(extensionsField, constraints2);
	++constraints2.gridy;
	constraints2.gridx = 0;

	label = new JLabel(msg("description"));
	constraints2.anchor = GridBagConstraints.NORTHEAST;
	constraints2.fill = GridBagConstraints.NONE;
	constraints2.weightx = 0;
        constraints2.insets.right = 0;
	infoFrame.add(label, constraints2);
	++constraints2.gridx;

	descriptionText = new JTextArea(10, 40);
	descriptionText.setEditable(false);
	descriptionText.setBackground(extensionsField.getBackground());
	descriptionText.setLineWrap(true);
	descriptionText.setWrapStyleWord(true);

	scroller = new JScrollPane(descriptionText);
	constraints2.anchor = GridBagConstraints.WEST;
	constraints2.fill = GridBagConstraints.BOTH;
	constraints2.weightx = 1;
	constraints2.weighty = 1;
        constraints2.insets.right = 5;
        constraints2.insets.bottom = 5;
	infoFrame.add(scroller, constraints2);

	// Close ---

	JButton closeButton = new JButton(msg("close"));
	constraints.fill = GridBagConstraints.NONE;
	constraints.weightx = 0;
	constraints.weighty = 0;
        constraints.insets.bottom = 5;
	workArea.add(closeButton, constraints);
	closeButton.setActionCommand("close");
	closeButton.addActionListener(this);
	closeButton.setMnemonic(closeButton.getText().charAt(0));

        pack();
	setResizable(true);
    }

    public void edit(App app) {
	this.app = app;

	updateWorkArea();

	setLocationRelativeTo(frame);
	pack();
        show();
        // Modal means blocked here until Close is pressed.
    }

    private void updateWorkArea() {
	StructuredFormat[] formats = app.getStructuredFormats();
	String[] classNames = new String[formats.length];
	for (int i = 0; i < formats.length; ++i)
	    classNames[i] = formats[i].getClass().getName();

	classNameList.setListData(classNames);
	classNameList.ensureIndexIsVisible(0);
	setUnselected();
    }

    // -----------------------------------------------------------------------

    public void valueChanged(ListSelectionEvent e) {
	int index = classNameList.getSelectedIndex();
	if (index < 0)
	    setUnselected();
	else
	    setSelected(index);
    }

    private void setUnselected() {
	removeButton.setEnabled(false);
	configureButton.setEnabled(false);

	mimeTypeField.setText("");
	extensionsField.setText("");
	descriptionText.setText("");
    }

    private void setSelected(int index) {
	StructuredFormat format = (app.getStructuredFormats())[index];

	removeButton.setEnabled(true);
	configureButton.setEnabled(format.getParameterUsage() != null);

	mimeTypeField.setText(format.getMimeType());
	mimeTypeField.setCaretPosition(0);

	String extensions;
	if (format.getFileNameExtensions() == null)
	    extensions = "*";
	else
	    extensions = StringUtil.join(format.getFileNameExtensions(), ", ");
	extensionsField.setText(extensions);
	extensionsField.setCaretPosition(0);

	descriptionText.setText(format.getDescription());
	descriptionText.setCaretPosition(0);
    }

    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();

        if (command.equals("add"))
            add();
        else if (command.equals("remove"))
            remove();
        else if (command.equals("configure"))
            configure();
        else if (command.equals("close"))
            close();
    }

    private void add() {
	String className = (String)
	    JOptionPane.showInputDialog(this, msg("className"), msg("add"), 
					JOptionPane.QUESTION_MESSAGE, 
					null, null, null);
	if (className == null)
	    return;

	try {
	    app.addStructuredFormat(className);
	} catch (Exception e) {
	    String reason = e.getClass().getName();
	    if (e.getMessage() != null)
		reason += ": " + e.getMessage();

	    JOptionPane.showMessageDialog(
		frame, msg("cannotLoadFormat", className, reason), 
		msg("error"), JOptionPane.ERROR_MESSAGE);

	    return;
	}

	updateWorkArea();
    }

    private void remove() {
	String className = (String) classNameList.getSelectedValue();
	if (className == null) {
	    this.getToolkit().beep();
	    return;
	}
	
	if (app.removeStructuredFormat(className) != null)
	    updateWorkArea();
    }

    private void configure() {
	int index = classNameList.getSelectedIndex();
	if (index < 0) {
	    this.getToolkit().beep();
	    return;
	}

	StructuredFormat format = (app.getStructuredFormats())[index];
	String[][] parameterUsage = format.getParameterUsage();
	String[] parameters = format.getParameters();
	if (parameterUsage == null || parameters == null) {
	    this.getToolkit().beep();
	    return;
	}

	StructFormatParamDialog dialog = 
	    new StructFormatParamDialog(frame, parameterUsage, parameters);
	parameters = dialog.getParameters();

	if (parameters != null && parameters.length > 0) {
	    String error = null;
	    try {
		format.setParameters(parameters);
	    } catch (ParameterUsageException e) {
		error = msg("cannotConfigureFormat", 
			    format.getClass().getName(), e.getMessage());
	    }

	    if (error != null) {
		JOptionPane.showMessageDialog(
		    frame, error, msg("error"), JOptionPane.ERROR_MESSAGE);
		return;
	    }
	}
    }

    private void close() {
	hide();
	dispose();
    }
}





