/*
 * Copyright (c) 2001-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.awt.*;
import java.awt.event.*;
import com.xmlmind.xmledit.model.*;
import com.xmlmind.awtutil.AWTUtil;

public class SubelementPath extends Component implements ItemSelectable {
    public static final int RELIEF_FLAT   = 0;
    public static final int RELIEF_SUNKEN = 1;
    public static final int RELIEF_RAISED = 2;

    public static final String INSERT_BEFORE_ACTION="insertLikeSelectedBefore";
    public static final String INSERT_AFTER_ACTION = "insertLikeSelectedAfter";

    private static class Item {
	public SubelementModel subelement;
	public String tag;
	public int x;
	public int width;
    }

    private SubelementModel displayedSubelement = null;
    private SubelementModel selectedSubelement = null;
    private ItemListener itemListener = null;
    private ActionListener actionListener = null;
    private Dimension intrinsicSize = null;
    private Item[] items = null;

    private int marginWidth = 1;
    private int marginHeight = 1;
    private int separatorSize = 7;
    private Color separatorColor = Color.black;
    private int separatorRelief = RELIEF_FLAT;
    private int borderRelief = RELIEF_FLAT;

    // -----------------------------------------------------------------------

    public SubelementPath() {
	enableEvents(AWTEvent.MOUSE_EVENT_MASK);
    }

    public void setSubelement(SubelementModel sub) {
	displayedSubelement = sub;
	selectedSubelement = null;
	invalidate();
	repaint();
    }

    public void invalidate() {
	intrinsicSize = null;
	items = null;

	super.invalidate();
    }

    public SubelementModel getSubelement() {
	return displayedSubelement;
    }

    public void setFont(Font font) {
	super.setFont(font);
	invalidate();
	repaint();
    }

    public void setBackground(Color color) {
	super.setBackground(color);
	repaint();
    }

    public void setForeground(Color color) {
	super.setForeground(color);
	repaint();
    }

    public void setMarginWidth(int margin) {
	marginWidth = margin;
	invalidate();
	repaint();
    }

    public int getMarginWidth() {
	return marginWidth;
    }

    public void setMarginHeight(int margin) {
	marginHeight = margin;
	invalidate();
	repaint();
    }

    public int getMarginHeight() {
	return marginHeight;
    }

    public void setSeparatorSize(int size) {
	separatorSize = size;
	invalidate();
	repaint();
    }

    public int getSeparatorSize() {
	return separatorSize;
    }

    public void setSeparatorColor(Color color) {
	separatorColor = color;
	repaint();
    }

    public Color getSeparatorColor() {
	return separatorColor;
    }

    public void setSeparatorRelief(int relief) {
	separatorRelief = relief;
	repaint();
    }

    public int getSeparatorRelief() {
	return separatorRelief;
    }

    public void setBorderRelief(int relief) {
	borderRelief = relief;
	repaint();
    }

    public int getBorderRelief() {
	return borderRelief;
    }

    public Dimension getMinimumSize() {
	return getPreferredSize();
    }
    
    public Dimension getPreferredSize() {
	if (intrinsicSize == null) 
	    relayout();
	return intrinsicSize;
    }

    private void relayout() {
	intrinsicSize = new Dimension();

	FontMetrics metrics = getFontMetrics(getFont());
	intrinsicSize.height = metrics.getAscent() + metrics.getDescent();

	if (displayedSubelement == null) {
	    items = null;
	    intrinsicSize.width = 2*marginWidth;
	} else {
	    int depth = 0;
	    SubelementModel sub = displayedSubelement;
	    while (sub != null) {
		++depth;
		sub = sub.getParentElement();
	    }

	    items = new Item[depth];

	    sub = displayedSubelement;
	    while (sub != null) {
		Item item = new Item();
		items[--depth] = item;

		item.subelement = sub;
		item.tag = 
		    ((sub.getSubelementType() == SubelementModel.TEXT)?
		     "#PCDATA" : ((ElementModel) sub).getElementTypeName());

		sub = sub.getParentElement();
	    }

	    int x = marginWidth;
	    for (int i = 0; i < items.length; ++i) {
		Item item = items[i];

		if (i > 0)
		    x += separatorSize;
		item.x = x;
		item.width = metrics.stringWidth(item.tag);
		x += item.width;
	    }
	    intrinsicSize.width = x + marginWidth;
	}

	intrinsicSize.height += 2*marginHeight;
    }

    public void paint(Graphics g) {
	if (intrinsicSize == null)
	    relayout();
	Dimension dim = getSize();

	g.setColor(getBackground());
	g.fillRect(0, 0, dim.width, dim.height);

	if (borderRelief != RELIEF_FLAT && 
	    marginWidth > 0 && marginHeight > 0)
	    g.draw3DRect(0, 0, dim.width-1, dim.height-1, 
			 (borderRelief == RELIEF_RAISED));

	if (items == null)
	    return;

	Font font = getFont();
	g.setFont(font);

	Color foreground = getForeground();
	FontMetrics metrics = getFontMetrics(font);
	int y = marginHeight + metrics.getAscent();
	int x = Math.min(intrinsicSize.width, dim.width) - marginWidth;

	int y2 = intrinsicSize.height/2 - separatorSize/2 + 1;
	int separatorSize2 = separatorSize-2;

	for (int i = items.length-1; i >= 0; --i) {
	    Item item = items[i];

	    x -= item.width;
	    g.setColor(foreground);
	    g.drawString(item.tag, x, y);

	    if (i > 0) {
		x -= separatorSize;
		g.setColor(separatorColor);
		switch (separatorRelief) {
		case RELIEF_SUNKEN:
		    g.fill3DRect(x+1, y2, separatorSize2, separatorSize2, 
				 false);
		    break;
		case RELIEF_RAISED:
		    g.fill3DRect(x+1, y2, separatorSize2, separatorSize2, 
				 true);
		    break;
		default:
		    g.fillRect(x+1, y2, separatorSize2, separatorSize2);
		}
	    }
	}
    }

    protected void processMouseEvent(MouseEvent e) {
	int modifiers = e.getModifiers();
	if (e.getID() == MouseEvent.MOUSE_PRESSED && 
	    (modifiers & MouseEvent.BUTTON1_MASK) != 0)
	    selectSubelement(e.getX(), e.getY(), modifiers);

	super.processMouseEvent(e);
    }

    private void selectSubelement(int x, int y, int modifiers) {
	if (items == null)
	    return;

	if (intrinsicSize == null)
	    relayout();
	Dimension dim = getSize();

	if (intrinsicSize.width > dim.width)
	    x += (intrinsicSize.width - dim.width);

	int separatorHalfSize = separatorSize/2;
	SubelementModel sub = null;
	for (int i = items.length-1; i >= 0; --i) {
	    Item item = items[i];

	    if (x >= item.x-separatorHalfSize && 
		x <= item.x+item.width+separatorHalfSize) {
		sub = item.subelement;
		break;
	    }
	}

	if (sub != null && sub != selectedSubelement) {
	    SubelementModel previousSelection = selectedSubelement;
	    selectedSubelement = sub;
	    notifyListeners(previousSelection, modifiers);
	}
    }

    private void notifyListeners(SubelementModel previousSelection, 
				 int modifiers) {
	if (itemListener != null) {
	    if (previousSelection != null) 
		itemListener.itemStateChanged(
		    new ItemEvent(this, ItemEvent.ITEM_STATE_CHANGED,
				  previousSelection, ItemEvent.DESELECTED));

	    if (selectedSubelement != null) {
		itemListener.itemStateChanged(
		    new ItemEvent(this, ItemEvent.ITEM_STATE_CHANGED,
				  selectedSubelement, ItemEvent.SELECTED));

		if ((modifiers & MouseEvent.SHIFT_MASK) != 0) {
		    if (actionListener != null)
			actionListener.actionPerformed(
			    new ActionEvent(this, ActionEvent.ACTION_PERFORMED,
					    INSERT_BEFORE_ACTION));
		} else if ((modifiers & AWTUtil.MOUSE_MODIFIER) != 0) {
			if (actionListener != null)
			    actionListener.actionPerformed(
				new ActionEvent(this, 
						ActionEvent.ACTION_PERFORMED,
						INSERT_AFTER_ACTION));
		}
	    }
	}
    }

    // -----------------------------------------------------------------------

    public Object[] getSelectedObjects() {
	if (selectedSubelement == null)
	    return new Object[0];
	else
	    return new Object[] { selectedSubelement };
    }

    public void addItemListener(ItemListener l) {
	itemListener = AWTEventMulticaster.add(itemListener, l);
    }

    public void removeItemListener(ItemListener l) {
	itemListener = AWTEventMulticaster.remove(itemListener, l);
    }

    public void addActionListener(ActionListener l) {
	actionListener = AWTEventMulticaster.add(actionListener, l);
    }

    public void removeActionListener(ActionListener l) {
	actionListener = AWTEventMulticaster.remove(actionListener, l);
    }
}
