/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.util.Hashtable;
import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import fr.pixware.util.HashtableUtil;
import fr.pixware.util.QuickSort;
import com.xmlmind.awtutil.ImageResource;
import com.xmlmind.xmledit.modeltool.XMLUtil;

public class TagChooserDialog extends JDialog 
                              implements ActionListener, 
                                         ListSelectionListener {
    public static final String PCDATA = "#PCDATA";

    private Frame frame;
    private JTextField tagField;
    private JList tagList;
    private Hashtable tagHistory = new Hashtable();
    private Vector autoCompletionList = new Vector();
    private String[] tags;
    private boolean strict;
    private String result;

    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.TagChooserDialogRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    private static String msg(String id, Object arg0) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return MessageFormat.format(message, new Object[] { arg0 });
    }

    // -----------------------------------------------------------------------

    public TagChooserDialog(Frame frame) {
	super(frame, "", /*modal*/ true); 

	this.frame = frame;

   	addWindowListener(new WindowAdapter() {
   	    public void windowClosing(WindowEvent e) {
   	        cancel();
   	    }
   	});

        // workArea is a JPanel with a BorderLayout.
	JPanel workArea = (JPanel) getContentPane();
	workArea.setLayout(new GridBagLayout());

	// Should be implemented by all LAFs.
	workArea.registerKeyboardAction(
	    this, "cancel", KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), 
	    JPanel.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.gridx = 0;
        constraints.gridy = 0;

	tagField = new JTextField(15);
        constraints.insets.top = 5;
        constraints.insets.left = 5;
	constraints.fill = GridBagConstraints.HORIZONTAL;
	constraints.weightx = 1;
	workArea.add(tagField, constraints);
	++constraints.gridx;
	tagField.setActionCommand("ok");
	tagField.addActionListener(this);
	tagField.addKeyListener(new KeyAdapter() { 
	    public void keyTyped(KeyEvent e) {
		if (e.getKeyChar() == ' ') {
		    autoCompletion();
		    e.consume();
		}
	    }
	});

	JButton okButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "ok.gif")));
	Insets margin = new Insets(1, 1, 1, 1);
	okButton.setMargin(margin);
        constraints.insets.left = 2;
	constraints.fill = GridBagConstraints.NONE;
	constraints.weightx = 0;
	workArea.add(okButton, constraints);
	++constraints.gridx;
	okButton.setActionCommand("ok");
	okButton.addActionListener(this);
	getRootPane().setDefaultButton(okButton);
	
	JButton cancelButton = new JButton(new ImageIcon(
	    ImageResource.get(getClass(), "cancel.gif")));
	cancelButton.setMargin(margin);
        constraints.insets.left = 2;
        constraints.insets.right = 5;
	workArea.add(cancelButton, constraints);
        constraints.gridx = 0;
        ++constraints.gridy;
	cancelButton.setActionCommand("cancel");
	cancelButton.addActionListener(this);
	
	tagList = new JList();
	tagList.setVisibleRowCount(20);
	tagList.addListSelectionListener(this);
	tagList.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		if (e.getClickCount() == 2) {
		    ok();
		}
	    }
	});

	JScrollPane scroller = new JScrollPane(tagList);
	constraints.gridwidth = 3;
        constraints.insets.top = 2;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
        constraints.insets.bottom = 5;
	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
	workArea.add(scroller, constraints);
	
        pack();
	setResizable(true);
    }

    public void setTagHistory(String[] tags) {
	tagHistory.clear();
	if (tags != null) {
	    tagHistory.put(PCDATA, PCDATA);
	    for (int i = 0; i < tags.length; ++i)
		tagHistory.put(tags[i], tags[i]);
	}
    }

    public String[] getTagHistory() {
	return HashtableUtil.getAllElements(tagHistory);
    }

    public String selectTag(String message, String[] tags, boolean strict) {
	setTitle(message);

	if (tags == null)
	    tags = getTagHistory();

	if (tags.length == 1) {
	    tagField.setText(tags[0]);
	} else {
	    tagField.setText("");
	    QuickSort.sort(tags);
	}

	tagList.setListData(tags);
	tagList.ensureIndexIsVisible(0);

	this.tags = tags;
	this.strict = strict;

	setLocationRelativeTo(frame);
	pack();

	tagField.requestFocus();

        show();
        // Modal means blocked here until a button is pressed.

	if (result != null)
	    tagHistory.put(result, result);
	return result;
    }

    public void valueChanged(ListSelectionEvent e) {
	String selection  = (String) tagList.getSelectedValue();
	if (selection != null)
	    tagField.setText(selection);
    }

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	
	if (command.equals("ok")) 
	    ok();
	else if (command.equals("cancel")) 
	    cancel();
    }

    private void autoCompletion() {
	tagList.ensureIndexIsVisible(0);

	String prefix = tagField.getText().trim();
	int prefixLength = prefix.length();

	if (prefixLength == 0) {
	    if (tagList.getModel().getSize() != tags.length)
		tagList.setListData(tags);
	    return;
	}

	autoCompletionList.removeAllElements();
	for (int i = 0; i < tags.length; ++i) {
	    String tag = tags[i];
	    if (tag.startsWith(prefix)) 
		autoCompletionList.addElement(tag);
	}

	int autoCompletionCount = autoCompletionList.size();
	if (autoCompletionCount == 0) {
	    getToolkit().beep();
	    return;
	}

	String nextPrefix = prefix;
	String firstTag = (String) autoCompletionList.elementAt(0);
	int firstTagLength = firstTag.length();
	loop: for (int i = prefixLength; i < firstTagLength; ++i) {
	    String candidatePrefix = nextPrefix + firstTag.charAt(i);

	    for (int j = 1; j < autoCompletionCount; ++j) {
		String tag = (String) autoCompletionList.elementAt(j);
		if (!tag.startsWith(candidatePrefix))
		    break loop;
	    }

	    nextPrefix = candidatePrefix;
	}
	if (nextPrefix.length() > prefixLength) 
	    tagField.setText(nextPrefix);

	tagList.setListData(autoCompletionList);
    }

    private void ok() {
	result = tagField.getText().trim();
	if (result.length() == 0) {
	    getToolkit().beep();
	    tagField.selectAll();
	    tagField.requestFocus();
	    return;
	}

	if (strict) {
	    boolean valid = false;
	    for (int i = 0; i < tags.length; ++i) {
		if (result.equals(tags[i])) {
		    valid = true;
		    break;
		}
	    }

	    if (!valid) {
		getToolkit().beep();
		tagField.selectAll();
		tagField.requestFocus();
		return;
	    }
	} else {
	    // At least check that result is a valid element type name ---

	    if (!XMLUtil.isName(result) &&
		!result.equals(PCDATA)) {
		getToolkit().beep();
		JOptionPane.showMessageDialog(
		    frame, msg("invalidTag", result),
		    msg("error"), JOptionPane.ERROR_MESSAGE);
		tagField.selectAll();
		tagField.requestFocus();
		return;
	    }
	}

	hide();
	dispose();
    }

    private void cancel() {
	result = null;
	hide();
	dispose();
    }
}

