/*
 * Copyright (c) 2001-2002 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind XML Editor project.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package com.xmlmind.xmleditapp;

import java.io.StringReader;
import java.io.IOException;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.awt.*;
import javax.swing.*;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.StyleSheet;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.event.HyperlinkEvent;
import com.xmlmind.xmledit.JDocumentPane;
import com.xmlmind.xmledit.model.ElementModel;
import com.xmlmind.xmledit.modeltool.ElementCheckerDiagnostic;
import com.xmlmind.xmledit.modeltool.ElementStructureChecker;
import com.xmlmind.xmledit.modeltool.ElementAttributeChecker;
import com.xmlmind.xmledit.modeltool.XMLUtil;

public class ValidityDiagnosticsTool extends Tool 
                                     implements ActionListener,
                                                HyperlinkListener {
    private JDocumentPane docPane;
    private JEditorPane htmlPane;
    private ElementCheckerDiagnostic[] diagnostics = null;

    // -----------------------------------------------------------------------

    private static ResourceBundle resourceBundle = null;
    static {
	try {
	    resourceBundle = ResourceBundle.getBundle(
		"com.xmlmind.xmleditapp.ValidityDiagToolRes");
	} catch (MissingResourceException ignored) {}
    }

    private static String msg(String id) {
	String message = null;
	try {
	    message = resourceBundle.getString(id);
	} catch (Exception e) {
	    return id;
	}
	return message;
    }

    // -----------------------------------------------------------------------

    public ValidityDiagnosticsTool(Frame frame, JDocumentPane docPane) {
	super(frame, msg("title"), /*top*/ false, /*right*/ true); 
	
	this.docPane = docPane;

        // workArea is a JPanel with a BorderLayout.
	JPanel workArea = (JPanel) getContentPane();
	workArea.setLayout(new GridBagLayout());

        GridBagConstraints constraints = new GridBagConstraints();
        constraints.insets.top = 5;
        constraints.insets.left = 5;
        constraints.insets.right = 5;
        constraints.insets.bottom = 5;
        constraints.gridx = 0;
        constraints.gridy = 0;

	htmlPane = new JEditorPane();
	HTMLEditorKit editorKit = 
	    (HTMLEditorKit) htmlPane.getEditorKitForContentType("text/html");
	editorKit.getStyleSheet().addRule(
	    "body { font-family: sans-serif; font-size: 12pt; }");
	htmlPane.setEditorKit(editorKit);
	htmlPane.setEditable(false);
	htmlPane.addHyperlinkListener(this);

	JScrollPane scroller = new JScrollPane(htmlPane);
	scroller.setPreferredSize(new Dimension(450, 300));
	constraints.fill = GridBagConstraints.BOTH;
	constraints.weightx = 1;
	constraints.weighty = 1;
	workArea.add(scroller, constraints);
	++constraints.gridy;

	JButton closeButton = new JButton(msg("close"));
        constraints.insets.top = 0;
	constraints.fill = GridBagConstraints.NONE;
	constraints.weightx = 0;
	constraints.weighty = 0;
	workArea.add(closeButton, constraints);
	closeButton.addActionListener(this);
	closeButton.setMnemonic(closeButton.getText().charAt(0));

	pack();
	setResizable(true);
    }

    public void actionPerformed(ActionEvent e) {
	hide();
	dispose();
    }

    public void hyperlinkUpdate(HyperlinkEvent e) {
	if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
	    if (diagnostics == null) {
		getToolkit().beep();
		return;
	    }

	    int index;
	    try {
		index = Integer.parseInt(e.getDescription());
	    } catch (NumberFormatException ignored) {
		index = -1;
	    }
	    if (index < 0 || index >= diagnostics.length) {
		getToolkit().beep();
		return;
	    }

	    ElementModel element = diagnostics[index].invalidElement;
	    if (element.getDocument() == null) {
		// Detached from document.
		getToolkit().beep();
		return;
	    }

	    docPane.selectSubelement(element);
	}
    }

    public void setDiagnostics(ElementCheckerDiagnostic[] diagnostics) {
	if (diagnostics != null && diagnostics.length == 0)
	    diagnostics = null;
	this.diagnostics = diagnostics;

	if (isVisible())
	    showDiagnostics();
    }

    public ElementCheckerDiagnostic[] getDiagnostics() {
	return diagnostics;
    }

    public void showDiagnostics() {
	showTool();

	if (diagnostics == null) {
	    htmlPane.setDocument(
		htmlPane.getEditorKit().createDefaultDocument());
	} else {
	    StringBuffer html = new StringBuffer();

	    FontMetrics metrics = htmlPane.getFontMetrics(htmlPane.getFont());
	    int cellWidth = 
		10 + metrics.stringWidth("[" + diagnostics.length + "]");

	    html.append("<html><body><table border='0' cellspacing='1'>\n");

	    for (int i = 0; i < diagnostics.length; ++i) {
		ElementCheckerDiagnostic d = diagnostics[i];

		String background;
		if (d instanceof ElementStructureChecker.Diagnostic) {
		    background = "#FFC0C0";
		} else {
		    if (d.invalidReason >= ElementAttributeChecker.XREF_REASON)
			background = "#FFFFC0";
		    else
			background = "#FFE0C0";
		}

		html.append("<tr><td width='");
		html.append(cellWidth);
		html.append("' valign='top' bgcolor='");
		html.append(background);
		html.append("'><a href='");
		html.append(i);
		html.append("'>[");
		html.append(1+i);
		html.append("]</a></td>");

		html.append("<td valign='top' bgcolor='");
		html.append(background);
		html.append("'>");
		XMLUtil.escapeXML(d.message, html);
		html.append("</td></tr>\n");
	    }

	    html.append("</table></body></html>");

	    // setText(), while simple, doesn't work too well (unlike read,
	    // it doesn't replace the model of htmlPane.
	    try {
		htmlPane.read(new StringReader(html.toString()), null);
	    } catch (IOException ignored) {}
	    htmlPane.scrollRectToVisible(new Rectangle(0, 0, 10, 10));
	}
    }
}
