/*
 * Copyright (c) 1998,1999,2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the Pixware Java utilities.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package fr.pixware.util;

import java.io.*;
import java.net.*;
import java.util.Hashtable;
import java.util.zip.GZIPInputStream;

/**
 * A collection of utility functions (static methods) operating on files and
 * directories.
 */
public class FileUtil {
    /**
     * Converts a <tt>file:</tt> URL to a File.
     * 
     * @param url the URL to be converted
     * @return the result of the conversion or <code>null</code> if
     * <code>url</code> is not a <tt>file:</tt> URL
     */
    public static File urlToFile(URL url) {
	if (!url.getProtocol().equals("file"))
	    return null;

	String fileName = url.getFile();

	if (File.separatorChar != '/')
	    fileName = fileName.replace('/', File.separatorChar);

	return new File(fileName);
    }

    /**
     * Converts a <tt>file:</tt> URL name to a File.
     * 
     * @param urlName the URL name to be converted
     * @return the result of the conversion or <code>null</code> if
     * <code>urlName</code> is not a <tt>file:</tt> URL
     */
    public static File urlToFile(String urlName) {
	URL url;
	try {
	    url = new URL(urlName);
	} catch (MalformedURLException e) {
	    //e.printStackTrace();
	    url = null;
	}
	
	return (url == null)? null : urlToFile(url);
    }

    /**
     * Converts a <tt>file:</tt> URL name to a file name.
     * 
     * @param urlName the URL name to be converted
     * @return the result of the conversion or <code>null</code> if
     * <code>urlName</code> is not a <tt>file:</tt> URL
     */
    public static String urlToFileName(String urlName) {
	File file = urlToFile(urlName);
	return (file == null)? null : file.getAbsolutePath();
    }

    /**
     * Converts a File to a <tt>file:</tt> URL.
     * 
     * @param file the file to be converted
     * @return the result of the conversion
     */
    public static URL fileToURL(File file) {
	String fileName = file.getAbsolutePath();

	if (File.separatorChar != '/')
	    fileName = fileName.replace(File.separatorChar, '/');

	if (!fileName.startsWith("/"))
	    fileName = "/" + fileName;

	if (!fileName.endsWith("/") && file.isDirectory())
	    fileName = fileName + "/";

	URL url;
	try {
	    url = new URL("file", "", -1, fileName);
	} catch (MalformedURLException cannotHappen) {
	    //e.printStackTrace();
	    url = null;
	}
	return url;
    }

    /**
     * Converts a file name to a <tt>file:</tt> URL.
     * 
     * @param fileName the file name to be converted
     * @return the result of the conversion
     */
    public static URL fileToURL(String fileName) {
	return fileToURL(new File(fileName));
    }

    /**
     * Converts a file name to a <tt>file:</tt> URL name.
     * 
     * @param fileName the file name to be converted
     * @return the result of the conversion
     */
    public static String fileToURLName(String fileName) {
	URL url = fileToURL(fileName);
	return (url == null)? null : url.toExternalForm();
    }

    // -----------------------------------------------------------------------

    /**
     * Returns the directory part in a file path name. The directory part is
     * everything before the last file path name separator (that is,
     * <tt>'\'</tt> for Windows).
     * <p>On Windows, <tt>'/'</tt> is used as an alternate file path name
     * separator.
     * <p>Examples:
     * <table border="1">
     * <tr>
     * <th bgcolor="#C0C0C0" align="center">Path
     * <th bgcolor="#C0C0C0" align="center">Result
     * <tr>
     * <td>util/FileUtil.java
     * <td>util
     * <tr>
     * <td>FileUtil.java
     * <td>. (dot)
     * </table>
     * 
     * @param fileName a file path name
     * @return the directory part
     */
    public static String fileDirName(String fileName) {
	char separ = File.separatorChar;
	int slash = fileName.lastIndexOf(separ);
	if (slash < 0 && separ == '\\') {
	    separ = '/';
	    // On Windows, '/' is an alternate fileName separator.
	    slash = fileName.lastIndexOf(separ);
	}

	String name;
	if (slash < 0) {
	    name = ".";
	} else if (slash == 0) {
	    name = File.separator;
	} else {
	    name = fileName.substring(0, slash);
	    if (separ != File.separatorChar) 
		name = name.replace(separ, File.separatorChar);
	}
	return name;
    }

    /**
     * Returns the base name part in a file path name. The base name part is
     * everything after the last file name separator (that is <tt>'\'</tt> for
     * Windows).
     * <p>On Windows, <tt>'/'</tt> is used as an alternate file path name
     * separator.
     * <table border="1">
     * <tr>
     * <th bgcolor="#C0C0C0" align="center">Path
     * <th bgcolor="#C0C0C0" align="center">Result
     * <tr>
     * <td>util/FileUtil.java
     * <td>FileUtil.java
     * <tr>
     * <td>FileUtil.java
     * <td>FileUtil.java
     * </table>
     * 
     * @param fileName a file path name
     * @return the base name part
     */
    public static String fileBaseName(String fileName) {
	int slash = fileName.lastIndexOf(File.separatorChar);
	if (slash < 0 && File.separatorChar == '\\')
	    // On Windows, '/' is an alternate fileName separator.
	    slash = fileName.lastIndexOf('/');

	String name;
	if (slash < 0)
	    name = fileName;
	else
	    name = fileName.substring(slash+1);
	return name;
    }

    /**
     * Returns the file extension part in a file path name. The file extension
     * part is everything after the last dot.
     * <p>On Windows, <tt>'/'</tt> is used as an alternate file path name
     * separator.
     * <table border="1">
     * <tr>
     * <th bgcolor="#C0C0C0" align="center">Path
     * <th bgcolor="#C0C0C0" align="center">Result
     * <tr>
     * <td>util/FileUtil.java
     * <td>.java
     * <tr>
     * <td>makefile
     * <td>"" (empty string)
     * <tr>
     * <td>/home/hussein/.profile
     * <td>"" (empty string)
     * </table>
     * 
     * @param fileName a file path name
     * @return the file extension part; does not include the dot
     * <p>If the base name without its extension is empty, the path is
     * considered not to have an extension part. This is the case of
     * <tt>/home/hussein/.profile</tt> in the examples above.
     */
    public static String fileExtension(String fileName) {
	int slash = fileName.lastIndexOf(File.separatorChar);
	if (slash < 0 && File.separatorChar == '\\')
	    // On Windows, '/' is an alternate fileName separator.
	    slash = fileName.lastIndexOf('/');
	if (slash < 0) 
	    slash = 0;
	else 
	    ++slash;

	int dot = fileName.lastIndexOf('.');
	if (dot <= slash)
	    // '.profile' has no extension!
	    return "";
	else
	    return fileName.substring(dot+1);
    }

    /**
     * Returns a file path name without its file extension part. The file
     * extension part is everything after the last dot.
     * <p>On Windows, <tt>'/'</tt> is used as an alternate file path name
     * separator.
     * <table border="1">
     * <tr>
     * <th bgcolor="#C0C0C0" align="center">Path
     * <th bgcolor="#C0C0C0" align="center">Result
     * <tr>
     * <td>util/FileUtil.java
     * <td>util/FileUtil
     * <tr>
     * <td>makefile
     * <td>makefile
     * <tr>
     * <td>/home/hussein/.profile
     * <td>/home/hussein/.profile
     * </table>
     * 
     * @param fileName a file path name
     * @return the file path without extension part if any
     */
    public static String trimFileExtension(String fileName) {
	int slash = fileName.lastIndexOf(File.separatorChar);
	if (slash < 0 && File.separatorChar == '\\')
	    // On Windows, '/' is an alternate fileName separator.
	    slash = fileName.lastIndexOf('/');
	if (slash < 0) 
	    slash = 0;
	else 
	    ++slash;

	int dot = fileName.lastIndexOf('.');
	if (dot <= slash)
	    // '.profile' has no extension!
	    return fileName;
	else
	    return fileName.substring(0, dot);
    }

    /**
     * A simple test for all functions dealing with file path names.
     */
    public static void main(String[] args) {
	for (int i = 0; i < args.length; ++i) {
	    String arg = args[i];
	    System.out.println("'" + fileDirName(arg) + "'");
	    System.out.println("'" + fileBaseName(arg) + "'");
	    System.out.println("'" + trimFileExtension(arg) + "'");
	    System.out.println("'" + fileExtension(arg) + "'");
	    System.out.println("'" + fileToURLName(arg) + "'");
	    System.out.println("---");
	}
    }

    // -----------------------------------------------------------------------

    /**
     * Deletes a file or an empty directory.
     * 
     * @param fileName the name of the file or empty directory to be deleted
     * @return <code>true</code> if the file or directory has been
     * successfully deleted; <code>false</code> otherwise
     */
    public static boolean removeFile(String fileName) {
	return removeFile(fileName, /*force*/ false);
    }

    /**
     * Deletes a file or a directory, possibly emptying the directory before
     * deleting it.
     * 
     * @param fileName the name of the file or directory to be deleted
     * @param force if <code>true</code> and the file to be deleted is a
     * non-empty directory, empty it before attempting to delete it; if
     * <code>false</code>, do not empty directories
     * @return <code>true</code> if the file or directory has been
     * successfully deleted; <code>false</code> otherwise
     */
    public static boolean removeFile(String fileName, boolean force) {
	return removeFile(new File(fileName), force);
    }

    /**
     * Deletes a file or a directory, possibly emptying the directory before
     * deleting it.
     * 
     * @param file the file or directory to be deleted
     * @param force if <code>true</code> and the file to be deleted is a
     * non-empty directory, empty it before attempting to delete it; if
     * <code>false</code>, do not empty directories
     * @return <code>true</code> if the file or directory has been
     * successfully deleted; <code>false</code> otherwise
     */
    public static boolean removeFile(File file, boolean force) {
	if (file.isDirectory() && force) 
	    emptyDirectory(file);

	return file.delete();
    }

    /**
     * Recursively deletes all the entries of a directory.
     * 
     * @param dirName the name of the directory to be emptied
     */
    public static void emptyDirectory(String dirName) {
	emptyDirectory(new File(dirName));
    }

    /**
     * Recursively deletes all the entries of a directory.
     * 
     * @param dir the directory to be emptied
     */
    public static void emptyDirectory(File dir) {
	String[] children = dir.list();

	for (int i = 0; i < children.length; ++i) {
	    File child = new File(dir, children[i]);

	    if (child.isDirectory())
		removeFile(child, /*force*/ true);
	    else 
		child.delete();
	}
    }

    /**
     * Copy a file.
     * 
     * @param srcFileName the name of the copied file
     * @param dstFileName the name of the copy
     * @exception IOException if there is an IO problem
     */
    public static void copyFile(String srcFileName, String dstFileName) 
	throws IOException {
	FileInputStream src = new FileInputStream(srcFileName);
	FileOutputStream dst = new FileOutputStream(dstFileName);

	byte[] bytes = new byte[8192];
	int count;

	while ((count = src.read(bytes)) != -1) 
	    dst.write(bytes, 0, count);

	src.close();
	dst.flush();
	dst.close();
    }

    // -----------------------------------------------------------------------

    /**
     * Loads the content of a text file.
     * 
     * @param fileName the name of the text file
     * @return the loaded String
     * @exception IOException if there is an IO problem
     */
    public static String loadString(String fileName) 
	throws FileNotFoundException, IOException {
	return loadString(new FileInputStream(fileName));
    }

    /**
     * Loads the content of an URL containing text.
     * 
     * @param url the URL of the text resource
     * @return the loaded String
     * @exception IOException if there is an IO problem
     */
    public static String loadString(URL url) throws IOException {
	URLConnection connection = url.openConnection();
	connection.setDefaultUseCaches(false);
	connection.setUseCaches(false);
	connection.setIfModifiedSince(0);
	return loadString(connection.getInputStream());
    }

    /**
     * Loads the content of an InputStream returning text.
     * 
     * @param stream the text source
     * @return the loaded String
     * @exception IOException if there is an IO problem
     */
    public static String loadString(InputStream stream) throws IOException {
	InputStreamReader in = new InputStreamReader(stream);

	char[] chars = new char[8192];
	StringBuffer buffer = new StringBuffer(chars.length);
	int count;

	try {
	    while ((count = in.read(chars, 0, chars.length)) != -1) {
		if (count > 0)
		    buffer.append(chars, 0, count);
	    }
	} finally {
	    in.close();
	}

	return buffer.toString();
    }

    /**
     * Saves some text to a file.
     * 
     * @param string the text to be saved
     * @param fileName the name of the file
     * @exception IOException if there is an IO problem
     */
    public static void saveString(String string, String fileName) 
	throws IOException {
	saveString(string, new FileOutputStream(fileName));
    }

    /**
     * Saves some text to an OutputStream.
     * 
     * @param string the text to be saved
     * @param stream the text sink
     * @exception IOException if there is an IO problem
     */
    public static void saveString(String string, OutputStream stream) 
	throws IOException {
	OutputStreamWriter out = new OutputStreamWriter(stream);

	out.write(string, 0, string.length());
	out.flush();

	out.close();
    }

    // -----------------------------------------------------------------------

    /**
     * Loads the content of a binary file.
     * 
     * @param fileName the name of the binary file
     * @return the loaded bytes
     * @exception IOException if there is an IO problem
     */
    public static byte[] loadBytes(String fileName) throws IOException {
	return loadBytes(new FileInputStream(fileName));
    }

    /**
     * Loads the content of an URL containing binary data.
     * 
     * @param url the URL of the binary data
     * @return the loaded bytes
     * @exception IOException if there is an IO problem
     */
    public static byte[] loadBytes(URL url) throws IOException {
	URLConnection connection = url.openConnection();
	connection.setDefaultUseCaches(false);
	connection.setUseCaches(false);
	connection.setIfModifiedSince(0);
	return loadBytes(connection.getInputStream());
    }

    /**
     * Loads the content of an InputStream returning binary data.
     * 
     * @param source the binary data source
     * @return the loaded bytes
     * @exception IOException if there is an IO problem
     */
    public static byte[] loadBytes(InputStream source) throws IOException {
	BufferedInputStream in = new BufferedInputStream(source);
	ByteArrayOutputStream out = new ByteArrayOutputStream();
	byte[] bytes = new byte[8192];
	int count;

	while ((count = in.read(bytes)) >= 0)
	    out.write(bytes, 0, count);

	return out.toByteArray();
    }

    // -----------------------------------------------------------------------

    /**
     * Tests if a file has been compressed using gzip.
     * 
     * @param fileName the name of the file to be tested
     * @return <code>true</code> if the file has been gzip-ed,
     * <code>false</code> otherwise
     * @exception IOException if there is an IO problem
     */
    public static boolean isGzipped(String fileName) throws IOException {
	InputStream in = new FileInputStream(fileName);
	int magic1 = in.read();
	int magic2 = in.read();
	in.close();
	return (magic1 == 0037 && magic2 == 0213);
    }

    /**
     * Loads the content of a text file compressed using gzip.
     * 
     * @param fileName the name of the gzip-ed file; the encoding of the text
     * before compression is assumed to be the default encoding of the
     * platform
     * @return the loaded String
     * @exception IOException if there is an IO problem
     * @see #defaultEncoding
     */
    public static String loadGzippedString(String fileName) 
	throws IOException {
	return loadGzippedString(new FileInputStream(fileName), null);
    }

    /**
     * Loads the content of an URL containing text compressed using gzip.
     * 
     * @param url the URL of the gzip-ed data; the encoding of the text before
     * compression is assumed to be the default encoding of the platform
     * @return the loaded String
     * @exception IOException if there is an IO problem
     * @see #defaultEncoding
     */
    public static String loadGzippedString(URL url) throws IOException {
	return loadGzippedString(url.openStream(), null);
    }

    /**
     * Loads the content of an InputStream returning text compressed using
     * gzip.
     * 
     * @param source the gzip-ed data source
     * @param encoding the encoding of the text before compression
     * @return the loaded String
     * @exception IOException if there is an IO problem
     */
    public static String loadGzippedString(InputStream source, 
					   String encoding) 
	throws IOException {
	if (encoding == null)
	    encoding = defaultEncoding();

	Reader in = new InputStreamReader(new GZIPInputStream(source), 
					  encoding);
	char[] chars = new char[8192];
	StringBuffer buffer = new StringBuffer(chars.length);
	int count;

	try {
	    while ((count = in.read(chars, 0, chars.length)) != -1) {
		if (count > 0)
		    buffer.append(chars, 0, count);
	    }
	} finally {
	    in.close();
	}

	return buffer.toString();
    }

    // -----------------------------------------------------------------------

    public static final String[] KNOWN_ENCODINGS = {
	"ASCII", // American Standard Code for Information Interchange 
	"Cp1252", // Windows Latin-1 
	"ISO8859_1", // ISO 8859-1, Latin alphabet No. 1 
	"UnicodeBig", // Sixteen-bit Unicode Transformation Format, 
		      // big-endian byte order, with byte-order mark 
	"UnicodeBigUnmarked", // Sixteen-bit Unicode Transformation Format, 
			      // big-endian byte order 
	"UnicodeLittle", // Sixteen-bit Unicode Transformation Format, 
			 // little-endian byte order, with byte-order mark 
	"UnicodeLittleUnmarked", // Sixteen-bit Unicode Transformation Format,
				 // little-endian byte order 
	"UTF8", // Eight-bit Unicode Transformation Format 
	"UTF-16", //Sixteen-bit Unicode Transformation Format, byte order 
		  // specified by a mandatory initial byte-order mark 

	// Extended Encoding Set (contained in i18n.jar) ---

	"Big5", // Big5, Traditional Chinese 
	"Cp037", // USA, Canada (Bilingual, French), Netherlands, Portugal, 
		 // Brazil, Australia 
	"Cp273", // IBM Austria, Germany 
	"Cp277", // IBM Denmark, Norway 
	"Cp278", // IBM Finland, Sweden 
	"Cp280", // IBM Italy 
	"Cp284", // IBM Catalan/Spain, Spanish Latin America 
	"Cp285", // IBM United Kingdom, Ireland 
	"Cp297", // IBM France 
	"Cp420", // IBM Arabic 
	"Cp424", // IBM Hebrew 
	"Cp437", // MS-DOS United States, Australia, New Zealand, South Africa 
	"Cp500", // EBCDIC 500V1 
	"Cp737", // PC Greek 
	"Cp775", // PC Baltic 
	"Cp838", // IBM Thailand extended SBCS 
	"Cp850", // MS-DOS Latin-1 
	"Cp852", // MS-DOS Latin-2 
	"Cp855", // IBM Cyrillic 
	"Cp856", // IBM Hebrew 
	"Cp857", // IBM Turkish 
	"Cp858", // Variant of Cp850 with Euro character 
	"Cp860", // MS-DOS Portuguese 
	"Cp861", // MS-DOS Icelandic 
	"Cp862", // PC Hebrew 
	"Cp863", // MS-DOS Canadian French 
	"Cp864", // PC Arabic 
	"Cp865", // MS-DOS Nordic 
	"Cp866", // MS-DOS Russian 
	"Cp868", // MS-DOS Pakistan 
	"Cp869", // IBM Modern Greek 
	"Cp870", // IBM Multilingual Latin-2 
	"Cp871", // IBM Iceland 
	"Cp874", // IBM Thai 
	"Cp875", // IBM Greek 
	"Cp918", // IBM Pakistan (Urdu) 
	"Cp921", // IBM Latvia, Lithuania (AIX, DOS) 
	"Cp922", // IBM Estonia (AIX, DOS) 
	"Cp930", // Japanese Katakana-Kanji mixed with 4370 UDC, 
		 // superset of 5026 
	"Cp933", // Korean Mixed with 1880 UDC, superset of 5029 
	"Cp935", // Simplified Chinese Host mixed with 1880 UDC, 
		 // superset of 5031 
	"Cp937", // Traditional Chinese Host miexed with 6204 UDC, 
		 // superset of 5033 
	"Cp939", // Japanese Latin Kanji mixed with 4370 UDC, superset of 5035 
	"Cp942", // IBM OS/2 Japanese, superset of Cp932 
	"Cp942C", // Variant of Cp942 
	"Cp943", // IBM OS/2 Japanese, superset of Cp932 and Shift-JIS 
	"Cp943C", // Variant of Cp943 
	"Cp948", // OS/2 Chinese (Taiwan) superset of 938 
	"Cp949", // PC Korean 
	"Cp949C", // Variant of Cp949 
	"Cp950", // PC Chinese (Hong Kong, Taiwan) 
	"Cp964", // AIX Chinese (Taiwan) 
	"Cp970", // AIX Korean 
	"Cp1006", // IBM AIX Pakistan (Urdu) 
	"Cp1025", // IBM Multilingual Cyrillic: Bulgaria, Bosnia, 
		  // Herzegovinia, Macedonia (FYR) 
	"Cp1026", // IBM Latin-5, Turkey 
	"Cp1046", // IBM Arabic - Windows 
	"Cp1097", // IBM Iran (Farsi)/Persian 
	"Cp1098", // IBM Iran (Farsi)/Persian (PC) 
	"Cp1112", // IBM Latvia, Lithuania 
	"Cp1122", // IBM Estonia 
	"Cp1123", // IBM Ukraine 
	"Cp1124", // IBM AIX Ukraine 
	"Cp1140", // Variant of Cp037 with Euro character 
	"Cp1141", // Variant of Cp273 with Euro character 
	"Cp1142", // Variant of Cp277 with Euro character 
	"Cp1143", // Variant of Cp278 with Euro character 
	"Cp1144", // Variant of Cp280 with Euro character 
	"Cp1145", // Variant of Cp284 with Euro character 
	"Cp1146", // Variant of Cp285 with Euro character 
	"Cp1147", // Variant of Cp297 with Euro character 
	"Cp1148", // Variant of Cp500 with Euro character 
	"Cp1149", // Variant of Cp871 with Euro character 
	"Cp1250", // Windows Eastern European 
	"Cp1251", // Windows Cyrillic 
	"Cp1253", // Windows Greek 
	"Cp1254", // Windows Turkish 
	"Cp1255", // Windows Hebrew 
	"Cp1256", // Windows Arabic 
	"Cp1257", // Windows Baltic 
	"Cp1258", // Windows Vietnamese 
	"Cp1381", // IBM OS/2, DOS People's Republic of China (PRC) 
	"Cp1383", // IBM AIX People's Republic of China (PRC) 
	"Cp33722", // IBM-eucJP - Japanese (superset of 5050) 
	"EUC_CN", // GB2312, EUC encoding, Simplified Chinese 
	"EUC_JP", // JIS X 0201, 0208, 0212, EUC encoding, Japanese 
	"EUC_KR", // KS C 5601, EUC encoding, Korean 
	"EUC_TW", // CNS11643 (Plane 1-3), EUC encoding, Traditional Chinese 
	"GBK", // GBK, Simplified Chinese 
	"ISO2022CN", // ISO 2022 CN, Chinese (conversion to Unicode only) 
	"ISO2022CN_CNS", // CNS 11643 in ISO 2022 CN form, Traditional Chinese
			 // (conversion from Unicode only) 
	"ISO2022CN_GB", // GB 2312 in ISO 2022 CN form, Simplified Chinese 
			// (conversion from Unicode only) 
	"ISO2022JP", // JIS X 0201, 0208 in ISO 2022 form, Japanese 
	"ISO2022KR", // ISO 2022 KR, Korean 
	"ISO8859_2", // ISO 8859-2, Latin alphabet No. 2 
	"ISO8859_3", // ISO 8859-3, Latin alphabet No. 3 
	"ISO8859_4", // ISO 8859-4, Latin alphabet No. 4 
	"ISO8859_5", // ISO 8859-5, Latin/Cyrillic alphabet 
	"ISO8859_6", // ISO 8859-6, Latin/Arabic alphabet 
	"ISO8859_7", // ISO 8859-7, Latin/Greek alphabet 
	"ISO8859_8", // ISO 8859-8, Latin/Hebrew alphabet 
	"ISO8859_9", // ISO 8859-9, Latin alphabet No. 5 
	"ISO8859_13", // ISO 8859-13, Latin alphabet No. 7 
	"ISO8859_15_FDIS", // ISO 8859-15, Latin alphabet No. 9 
	"JIS0201", // JIS X 0201, Japanese 
	"JIS0208", // JIS X 0208, Japanese 
	"JIS0212", // JIS X 0212, Japanese 
	"JISAutoDetect", // Detects and converts from Shift-JIS, EUC-JP, 
			 // ISO 2022 JP (conversion to Unicode only) 
	"Johab", // Johab, Korean 
	"KOI8_R", // KOI8-R, Russian 
	"MS874", // Windows Thai 
	"MS932", // Windows Japanese 
	"MS936", // Windows Simplified Chinese 
	"MS949", // Windows Korean 
	"MS950", // Windows Traditional Chinese 
	"MacArabic", // Macintosh Arabic 
	"MacCentralEurope", // Macintosh Latin-2 
	"MacCroatian", // Macintosh Croatian 
	"MacCyrillic", // Macintosh Cyrillic 
	"MacDingbat", // Macintosh Dingbat 
	"MacGreek", // Macintosh Greek 
	"MacHebrew", // Macintosh Hebrew 
	"MacIceland", // Macintosh Iceland 
	"MacRoman", // Macintosh Roman 
	"MacRomania", // Macintosh Romania 
	"MacSymbol", // Macintosh Symbol 
	"MacThai", // Macintosh Thai 
	"MacTurkish", // Macintosh Turkish 
	"MacUkraine", // Macintosh Ukraine 
	"SJIS", // Shift-JIS, Japanese 
	"TIS620" // TIS620, Thai 
    };
    private static Hashtable knownEncodings = new Hashtable();
    private static String platformDefaultEncoding;
    static {
	for (int i = 0; i < KNOWN_ENCODINGS.length; ++i) {
	    String name = KNOWN_ENCODINGS[i];
	    knownEncodings.put(name, name);
	}

	platformDefaultEncoding = 
	    (new OutputStreamWriter(System.out)).getEncoding();
    }

    /**
     * Returns the default character encoding for this platform.
     * 
     * @return default character encoding for this platform
     */
    public static String defaultEncoding() {
	return platformDefaultEncoding;
    }

    /**
     * Returns true if the specified name is a character encoding known by
     * Java, false otherwise.
     * 
     * @param name the encoding name to be tested (case sensitive)
     * @return <code>true</code> if the specified name is a character encoding
     * known by Java, <code>false</code> otherwise
     */
    public static boolean isKnownEncoding(String name) {
	return knownEncodings.containsKey(name);
    }
}
