/*
 * Copyright (c) 2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the Pixware Java utilities.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package fr.pixware.util;

import java.io.*;
import java.util.Enumeration;

/**
 * A hashtable which is more compact and less efficient (modulo the fact that
 * java.util.Hashtable is thread safe and that <em>LinearHashtable is not
 * thread safe</em>) than {@link java.util.Hashtable}.
 * <p>LinearHashtable has exactly the same API as java.util.Hashtable,
 * therefore there is no need to document again this API here.
 * <p>This implementation is the one used by Smalltalk. It is useful if you
 * need to have hundreds or even thousands of very small hashtables (with up
 * to 3 entries for most instances).
 */
public final class LinearHashtable implements Cloneable, Serializable {
    private transient Object[] table;

    /**
     * Constructs a hashtable with an initial capacity of 3 key/value pairs.
     */
    public LinearHashtable() {
	this(3);
    }

    /**
     * Constructs a hashtable with the specified initial capacity.
     * 
     * @param capacity the initial capcity expressed in number of key/value
     * pairs
     */
    public LinearHashtable(int capacity) {
	table = new Object[2*capacity];
    }

    /**
     * Returns the number of key/value pairs currently stored in this
     * hashtable. Note that calling this method is more expensive than calling
     * {@link java.util.Hashtable#size} because the size is recomputed each
     * time this method is called.
     * 
     * @return the number of key/value pairs currently stored in this
     * hashtable
     */
    public int size() {
	int size = 0;
	for (int i = 0; i < table.length; i += 2) {
	    if (table[i] != null)
		++size;
	}
	return size;
    }

    public boolean isEmpty() {
	return (size() == 0);
    }

    public boolean contains(Object value) {
	for (int i = 1; i < table.length; i += 2) {
	    if (table[i] == value)
		return true;
	}
	return false;
    }

    public boolean containsKey(Object key) {
	return (get(key) != null);
    }

    public Object get(Object key) {
	int i = indexOf(table, key);
	return (i < 0)? null : table[i+1];
    }

    private static int indexOf(Object[] table, Object key) {
	int hash = key.hashCode();
	if (hash < 0)
	    hash = -hash;

	int i0 = 2*(hash % (table.length/2));
	int i = i0;

	for (;;) {
	    if (table[i] == null || table[i].equals(key))
		return i;

	    i += 2;
	    if (i == table.length)
		i = 0;

	    if (i == i0)
		// Table full.
		return -1;
	}
    }

    public Object put(Object key, Object value) {
	int i = indexOf(table, key);
	if (i < 0) {
	    // Grow table.
	    Object[] newTable = new Object[2*table.length];
	    
	    for (int j = 0; j < table.length; j += 2) {
		Object curKey = table[j];

		if (curKey != null) {
		    int k = indexOf(newTable, curKey);
		    newTable[k] = curKey;
		    newTable[k+1] = table[j+1];
		}
	    }

	    table = newTable;
	    i = indexOf(table, key);
	}

	Object oldValue = table[i+1];
	table[i] = key;
	table[i+1] = value;

	return oldValue;
    }

    public Object remove(Object key) {
	int i = indexOf(table, key);
	if (i < 0 || table[i] == null)
	    return null;

	Object oldValue = table[i+1];
	table[i] = null;
	table[i+1] = null;

	// Rehash from i.
	for (;;) {
	    i += 2;
	    if (i == table.length)
		i = 0;

	    // Any object which has collided with the deleted object is
	    // just after the deleted object. Thus, finding null means
	    // there is no more objects which *could* have collided with
	    // the deleted object.
	    // This loop is guaranteed to stop because null has replaced
	    // the deleted object.

	    if (table[i] == null)
		break;

	    Object curKey = table[i];
	    int j = indexOf(table, curKey); 

	    // If j == i, the slot will not be empty!
	    if (table[j] == null) {
		// Move to j because this slot is free.
		table[j] = curKey;
		table[j+1] = table[i+1];

		table[i] = null;
		table[i+1] = null;
	    }
	}

	return oldValue;
    }

    public void clear() {
	for (int i = 0; i < table.length; ++i) 
	    table[i] = null;
    }

    public Object clone() {
	LinearHashtable copy;

	try {
	    copy = (LinearHashtable) super.clone();
	} catch (CloneNotSupportedException cannotHappen) {
	    cannotHappen.printStackTrace();
	    return null;
	}

	copy.table = new Object[table.length];
	System.arraycopy(table, 0, copy.table, 0, table.length);
	return copy;
    }

    public String toString() {
	StringBuffer buffer = new StringBuffer();

	buffer.append(getClass().getName());
	buffer.append('[');
	for (int i = 0; i < table.length; i += 2) {
	    if (table[i] != null) {
		if (i > 0)
		    buffer.append(',');
		buffer.append(table[i]);
		buffer.append('=');
		buffer.append(table[i+1]);
	    }
	}
	buffer.append(']');

	return buffer.toString();
    }

    /**
     * Copies the keys contained into this hashtable into the specified array.
     * 
     * @param array the array where the keys are to be copied
     */
    public void copyKeysInto(Object[] array) {
	int count = 0;
	for (int i = 0; i < table.length; i += 2) {
	    if (table[i] != null) 
		array[count++] = table[i];
	}
    }

    /**
     * Copies the elements contained into this hashtable into the specified
     * array.
     * 
     * @param array the array where the elements are to be copied
     */
    public void copyElementsInto(Object[] array) {
	int count = 0;
	for (int i = 0; i < table.length; i += 2) {
	    if (table[i] != null) 
		array[count++] = table[i+1];
	}
    }

    public Enumeration keys() {
	return new KeyEnumeration();
    }

    public Enumeration elements() {
	return new ElementEnumeration();
    }

    private class KeyEnumeration implements Enumeration {
	private int index = 0;

	public boolean hasMoreElements() {
	    while (index < table.length) {
		if (table[index] != null)
		    return true;
		index += 2;
	    }
	    return false;
	}

	public Object nextElement() {
	    Object next;

	    if (index < table.length) {
		next = table[index];
		index += 2;
	    } else {
		next = null;
	    }

	    return next;
	}
    }

    private class ElementEnumeration implements Enumeration {
	private int index = 0;

	public boolean hasMoreElements() {
	    while (index < table.length) {
		if (table[index] != null)
		    return true;
		index += 2;
	    }
	    return false;
	}

	public Object nextElement() {
	    Object next;

	    if (index < table.length) {
		next = table[index+1];
		index += 2;
	    } else {
		next = null;
	    }

	    return next;
	}
    }

    // -----------------------------------------------------------------------

    private void writeObject(ObjectOutputStream stream)
	throws IOException {
	int length = table.length;
	stream.writeInt(length);

	stream.writeInt(2*size());
	for (int i = 0; i < length; i += 2) {
	    if (table[i] != null) {
		stream.writeObject(table[i]);
		stream.writeObject(table[i+1]);
	    }
	}
    }

    private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
	table = new Object[stream.readInt()];

	// It is absolutely mandatory to rehash the entries because
	// the hashCode() of an object is often its System.identityHashCode().

	int count = stream.readInt();
	for (int i = 0; i < count; i += 2) {
	    Object key = stream.readObject();
	    Object value = stream.readObject();

	    int index = indexOf(table, key);
	    table[index] = key;
	    table[index+1] = value;
	}
    }
}
