/*
 * Copyright (c) 1999,2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the Pixware Java utilities.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package fr.pixware.util;

import java.io.*;

/**
 * A collection of utility functions (static methods) returning information
 * which is platform dependant.
 */
public class PlatformUtil {
    /**
     * Returns the name of the platform Java is running on.
     * 
     * @return the name of the platform Java is running on: <tt>"windows"</tt>
     * or <tt>"unix"</tt>
     */
    public static String platform() {
	if (File.pathSeparatorChar == ';')
	    return "windows";
	else
	    return "unix";
    }

    /**
     * Returns the name of the home directory of current user. The system
     * property <tt>HOME</tt>, if set to an existing directory name, may be
     * used to force the value returned by this function.
     * 
     * @return the name of the home directory of current user or
     * <code>null</code> if the candidate directory found by this function
     * does not exist or is not a directory.
     */
    public static String homeDir() {
	String homeDir = System.getProperty("HOME");
	if (homeDir == null || homeDir.length() == 0) 
	    homeDir = System.getProperty("user.home");

	if (!(new File(homeDir)).isDirectory())
	    homeDir = null;

	return homeDir;
    }

    /**
     * Returns the name of the user preference file associated to the
     * specified application.
     * <p>Note that <b>RC</b> means <b>R</b>untime <b>C</b>onfiguration.
     * 
     * @param appName the application name
     * @return the name of the user preference file or <code>null</code> if
     * the user <tt>HOME</tt> directory is unknown
     * @see #homeDir
     */
    public static String rcFileName(String appName) {
	String homeDir = homeDir();
	if (homeDir == null)
	    return null;

	String rcFileName;
	if (platform().equals("windows"))
	    rcFileName = homeDir + File.separatorChar + appName + ".ini";
	else
	    rcFileName = homeDir + File.separatorChar + "." + appName;

	return rcFileName;
    }

    /**
     * Returns the name of a temporary file ending with extension
     * <tt>".tmp"</tt>. The directory of this file is the directory returned
     * by {@link #tmpDir}.
     * 
     * @return the name of a temporary file
     */
    public static String tmpFileName() {
	return tmpFileName(".tmp");
    }

    /**
     * Returns the name of a temporary file ending with the specified
     * extension. The directory of this file is the directory returned by
     * {@link #tmpDir}.
     * 
     * @param extension the desired extension for the file name; if a dot is
     * required, add it at the start of <code>extension</code>
     * @return the name of a temporary file
     */
    public static String tmpFileName(String extension) {
	String baseName = 
	    Long.toString(System.currentTimeMillis(), Character.MAX_RADIX) + 
	    extension;

	String tmpFile;
	String tmpDir = tmpDir();
	if (tmpDir == null)
	    tmpFile = baseName;
	else
	    tmpFile = (new File(tmpDir, baseName)).getPath();

	return tmpFile;
    }

    /**
     * Returns the name of directory where temporary files can be created
     * safely.
     * <p>Under Windows, the system property <tt>TMP</tt> or <tt>TEMP</tt>, if
     * set to an existing directory, may be used to force the value returned
     * by this function.
     * <p>Under Unix, this function always returns <tt>"/tmp"</tt>.
     * 
     * @return the name of the temporary directory or <code>null</code> if the
     * candidate directory found by this function does not exist or is not a
     * directory.
     */
    public static String tmpDir() {
	String tmpDir;
	boolean windows  = (platform().equals("windows"));
	if (windows) {
	    tmpDir = System.getProperty("TMP");
	    if (tmpDir == null || tmpDir.length() == 0) { 
		tmpDir = System.getProperty("TEMP");
		if (tmpDir == null || tmpDir.length() == 0) 
		    tmpDir = "C:\\";
	    }
	} else {
	    tmpDir = "/tmp";
	}

	if (!(new File(tmpDir)).isDirectory())
	    tmpDir = null;

	return tmpDir;
    }

    /**
     * Executes a command using the standard shell of the platform.
     * 
     * @param command the shell command to be executed
     * @return the exit status returned by the shell
     * @exception IOException if an I/O error occurs
     * @exception InterruptedException if the current thread is interrupted by
     * another thread while it is waiting the completion of the shell command
     */
    public static int shellExec(String command) 
	throws IOException, InterruptedException { 
	return shellExec(command, true);
    }

    private static class InputConsumer extends Thread {
	private InputStream in;
	private OutputStream echo;
	private byte[] buffer = new byte[4096];

	public InputConsumer(InputStream in, OutputStream echo) {
	    this.in = in;
	    this.echo = echo;
	}

	public void run() {
	    for (;;) {
		int count;
		try {
		    count = in.read(buffer);
		} catch (IOException e) {
		    count = -1;
		}
		if (count < 0)
		    break;

		if (count > 0 && echo != null) {
		    try {
			echo.write(buffer, 0, count);
		    } catch (IOException e) {
			echo = null;
		    }
		}
	    }
	    
	    //System.err.println(this.toString() + " terminated");
	}
    }

    /**
     * Executes a command using the standard shell of the platform.
     * 
     * @param command the shell command to be executed
     * @param verbose if true, the shell command output on
     * <code>System.out</code> and <code>System.err</code> is echoed; if
     * false, this output is discarded
     * @return the exit status returned by the shell
     * @exception IOException if an I/O error occurs
     * @exception InterruptedException if the current thread is interrupted by
     * another thread while it is waiting the completion of the shell command
     */
    public static int shellExec(String command, boolean verbose) 
	throws IOException, InterruptedException { 
	Process process;

	if (verbose)
	    System.out.println(command);

	if (platform().equals("unix")) 
	    process = Runtime.getRuntime().exec(
		new String[] { "/bin/sh", "-c", command });
	else 
	    process = Runtime.getRuntime().exec(
		new String[] { "cmd.exe", "/c", command });

	// Without these consumer threads, any shell command that outputs
	// something lengthy will block on Windows NT.

	InputConsumer consumer1 = 
	    new InputConsumer(process.getInputStream(),
			      (verbose? System.out : null));
	consumer1.start();

	InputConsumer consumer2 = 
	    new InputConsumer(process.getErrorStream(),
			      (verbose? System.err : null));
	consumer2.start();

	return process.waitFor();
    }

    /**
     * Returns the command separator of the standard shell of the platform.
     * Example: it is <tt>';'</tt> for the bourne shell <tt>/bin/sh</tt> of
     * Unix.
     * 
     * @return the command separator of the standard shell of the platform
     */
    public static String commandSeparator() {
	return (platform().equals("unix")? ";" : "&&");
    }
}
