/*
 * Copyright (c) 1999,2000-2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the Pixware Java utilities.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package fr.pixware.util;

/**
 * A quick sort function (static method). The objects to be compared (example:
 * String) are not required to implement a specific interface (example: {@link
 * java.lang.Comparable}). Instead the following design was chosen: the quick
 * sort function is passed a comparator (example: StringCompare), an object
 * which implements inner interface {@link Compare}.
 */
public class QuickSort {
    /**
     * Interface implemented by comparators, that is, objects able to compare
     * two objects.
     */
    public interface Compare {
	/**
	 * Compares two objects <code>o1</code> and <code>o2</code>.
	 * 
	 * @param o1 the object to be compared to <code>o2</code>
	 * @param o2 the object to be compared to <code>o1</code>
	 * @return a negative integer, zero, or a positive integer as
	 * <code>o1</code> is less than, equal to, or greater than
	 * <code>o2</code>
	 */
	int compare(Object o1, Object o2);
    }

    /**
     * The class of String comparators. Comparison is implemented using {@link
     * java.lang.String#compareTo} which compares two strings
     * lexicographically.
     */
    public static class StringCompare implements Compare {
	public int compare(Object o1, Object o2) {
	    return ((String) o1).compareTo((String) o2);
	}
    }
    /**
     * An instance of StringCompare.
     */
    public static StringCompare stringCompare = new StringCompare();

    /**
     * Sorts an array of Strings lexicographically.
     * 
     * @param strings the array of Strings to be sorted
     */
    public static void sort(String[] strings) {
	sort(strings, 0, strings.length-1, stringCompare);
    }

    /**
     * Sorts an array of objects.
     * 
     * @param values the array of Objects to be sorted
     * @param compare the Object comparator
     */
    public static void sort(Object[] values, Compare compare) {
	sort(values, 0, values.length-1, compare);
    }

    /**
     * Sorts a slice in an array of objects.
     * 
     * @param values the array of Objects to be sorted
     * @param first position of first object of the slice
     * @param last position of last object of the slice
     * @param compare the Object comparator
     */
    public static void sort(Object[] values, int first, int last, 
			    Compare compare) {
	int f = first;
	int l = last;
	Object pivot;

	if (last > first) {
	    pivot = values[(first + last) / 2];

	    while (f <= l) {
		while (f < last && compare.compare(values[f], pivot) < 0)
		    ++f;

		while (l > first && compare.compare(values[l], pivot) > 0)
		    --l;

		if (f <= l) {
		    swap(values, f, l);
		    ++f;
		    --l;
		}
	    }

	    if (first < l)
		sort(values, first, l, compare);

	    if (f < last)
		sort(values, f, last, compare);
	}
    }

    private static void swap(Object[] values, int i, int j) {
	Object temp;

	temp = values[i]; 
	values[i] = values[j];
	values[j] = temp;
    }

    /**
     * Simple test.
     */
    public static void main(String[] args) {
	QuickSort.sort(args);
	for (int i = 0; i < args.length; ++i)
	    System.out.println(args[i]);
    }
}
