/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the Pixware Java utilities.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package fr.pixware.util;

import java.io.*;

public class REFactoryImpl implements REFactory {
    private static final class ReaderAdapter extends Reader {
	private CharSequence chars;
	private int pos;

	public ReaderAdapter(CharSequence chars) {
	    super();
	    this.chars = chars;
	}

	public int read() throws IOException {
	    if (pos >= chars.length())
		return -1;
	    else
		return chars.charAt(pos++);
	}

	public int read(char buffer[], int offset, int length) 
	    throws IOException {
	    int maxCount = chars.length() - pos;
	    if (maxCount == 0)
		return -1;
	    int count = Math.min(length, maxCount);

	    int end = offset + count;
	    for (int i = offset; i < end; ++i)
		buffer[i] = chars.charAt(pos++);

	    return count;
	}

	public void close() throws IOException {}
    }

    private static final class REImpl implements RE, Serializable {
	private String pattern;
	private int flags;
	private transient gnu.regexp.RE impl;

	public REImpl(String pattern, int flags, gnu.regexp.RE re) {
	    this.pattern = pattern;
	    this.flags = flags;
	    impl = re;
	}

	private void readObject(ObjectInputStream stream)
	    throws IOException, ClassNotFoundException {
	    stream.defaultReadObject();

	    try {
		impl = new gnu.regexp.RE(pattern, flags, 
					 gnu.regexp.RESyntax.RE_SYNTAX_PERL5);
	    } catch (gnu.regexp.REException cannotHappen) {}
	}

	public String toString() {
	    return pattern;
	}

	public boolean search(String input) {
	    return (impl.getMatch(input) != null);
	}

	public REMatch[] getMatch(String input, int from) {
	    gnu.regexp.REMatch match = impl.getMatch(input, from);
	    if (match == null)
		return null;
	    return toREMatchList(match);
	}

	public REMatch[] getMatch(CharSequence input, int from) {
	    gnu.regexp.REMatch match = impl.getMatch(new ReaderAdapter(input),
						     from);
	    if (match == null)
		return null;
	    return toREMatchList(match);
	}

	private REMatch[] toREMatchList(gnu.regexp.REMatch match) {
	    REMatch[] result = new REMatch[1+impl.getNumSubs()];
	    int count = 0;

	    for (int i = 0; i < result.length; ++i) {
		int index = match.getStartIndex(i);
		if (index < 0)
		    break;

		result[count++] = new REMatch(index, match.getEndIndex(i),
					      match.toString(i));
	    }

	    if (count != result.length) {
		REMatch[] result2 = new REMatch[count];
		System.arraycopy(result, 0, result2, 0, count);
		result = result2;
	    }

	    return result;
	}
    }

    public RE createRE(String pattern, int options) throws RESyntaxException {
	gnu.regexp.RE re = null;

	int flags = 0x0;
	if ((options & CASE_INSENSITIVE) != 0)
	    flags |= gnu.regexp.RE.REG_ICASE;
	if ((options & MULTI_LINE) != 0)
	    flags |= gnu.regexp.RE.REG_MULTILINE;
	if ((options & SINGLE_LINE) != 0)
	    flags |= gnu.regexp.RE.REG_DOT_NEWLINE;
	
	try {
	    re = new gnu.regexp.RE(pattern, flags, 
				   gnu.regexp.RESyntax.RE_SYNTAX_PERL5);
	} catch (gnu.regexp.REException e) {
	    throw new RESyntaxException(e);
	}

	return new REImpl(pattern, flags, re);
    }
}

