/*
 * Copyright (c) 2001 Pixware. 
 *
 * Author: Hussein Shafie
 *
 * This file is part of the Pixware Java utilities.
 * For conditions of distribution and use, see the accompanying legal.txt file.
 */
package fr.pixware.util;

import java.io.*;

public class REFactoryImpl implements REFactory {
    private static final class CharacterIteratorAdapter 
	                 implements org.apache.regexp.CharacterIterator {
	private CharSequence chars;

	public CharacterIteratorAdapter(CharSequence chars) {
	    this.chars = chars;
	}

	public char charAt(int pos) {
	    return chars.charAt(pos);
	}

	public boolean isEnd(int pos) {
	    return (pos >= chars.length());
	}

	public String substring(int begin) {
	    return CharSequenceUtil.substring(chars, begin);
	}

	public String substring(int begin, int end) {
	    return CharSequenceUtil.substring(chars, begin, end);
	}
    }

    private static final class REImpl implements RE, Serializable {
	private String pattern;
	private int flags;
	private transient org.apache.regexp.RE impl;

	public REImpl(String pattern, int flags,
		      org.apache.regexp.REProgram program) {
	    this.pattern = pattern;
	    this.flags = flags;
	    impl = new org.apache.regexp.RE(program, flags);
	}

	private void readObject(ObjectInputStream stream)
	    throws IOException, ClassNotFoundException {
	    stream.defaultReadObject();

	    try {
		impl = new org.apache.regexp.RE(pattern, flags);
	    } catch (org.apache.regexp.RESyntaxException cannotHappen) {}
	}

	public String toString() {
	    return pattern;
	}

	public boolean search(String input) {
	    return impl.match(input);
	}
    
	public REMatch[] getMatch(String input, int from) {
	    if (!impl.match(input, from))
		return null;
	    return getMatch();
	}

	public REMatch[] getMatch(CharSequence input, int from) {
	    if (!impl.match(new CharacterIteratorAdapter(input), from))
		return null;
	    return getMatch();
	}

	private REMatch[] getMatch() {
	    REMatch[] result = new REMatch[impl.getParenCount()];

	    for (int i = 0; i < result.length; ++i)
		result[i] = new REMatch(impl.getParenStart(i), 
					impl.getParenEnd(i),
					impl.getParen(i));

	    return result;
	}
    }

    private org.apache.regexp.RECompiler compiler = 
        new org.apache.regexp.RECompiler();

    public RE createRE(String pattern, int options) throws RESyntaxException {
	org.apache.regexp.REProgram program = null;

	try {
	    program = compiler.compile(pattern);
	} catch (org.apache.regexp.RESyntaxException e) {
	    throw new RESyntaxException(e);
	}

	int flags = 0x0;
	if ((options & CASE_INSENSITIVE) != 0)
	    flags |= org.apache.regexp.RE.MATCH_CASEINDEPENDENT;
	if ((options & MULTI_LINE) != 0)
	    flags |= org.apache.regexp.RE.MATCH_MULTILINE;
	if ((options & SINGLE_LINE) != 0)
	    flags |= org.apache.regexp.RE.MATCH_SINGLELINE;

	return new REImpl(pattern, flags, program);
    }
}

